/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2020 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "base/threadpool.h"
#include "base/mainthread.h"
#include "abstractactionitem.h"


const int sh::actions::ActionDefaultPrecedenceValues::PRECEDENCE_OPENFILE = 300000;
const int sh::actions::ActionDefaultPrecedenceValues::PRECEDENCE_BUILTIN_SPECIAL_OPEN = 600000;

std::shared_ptr<sh::actions::AbstractActionItem> sh::actions::AbstractActionItem::NO_PARENT = std::make_shared<sh::actions::AbstractActionItem>();

sh::actions::AbstractActionItem::AbstractActionItem()
{
}

sh::actions::AbstractActionItem::AbstractActionItem(QString text, bool enabled, QString icon, int defaultActionPrecedence, bool checkable, bool ischecked) :
    QObject()
{
    moveToThread(QApplication::instance()->thread());
    _text = text;
    _enabled = enabled;
    _visible = true;
    _ischeckable = checkable;
    _ischecked = ischecked;
    _icon = icon;
    _defaultActionPrecedence = defaultActionPrecedence;
    _initialized = false;
    _initializing = false;
}

QMutex sh::actions::AbstractActionItem::mutex;

QMutex sh::actions::AbstractActionItem::initmutex;

QWaitCondition sh::actions::AbstractActionItem::initcondition;


QString sh::actions::AbstractActionItem::text()
{
    QMutexLocker _l(&mutex);
    return _text;
}

int sh::actions::AbstractActionItem::defaultActionPrecedence() const
{
    return _defaultActionPrecedence;
}

QString sh::actions::AbstractActionItem::icon()
{
    QMutexLocker _l(&mutex);
    return _icon;
}

bool sh::actions::AbstractActionItem::enabled()
{
    QMutexLocker _l(&mutex);
    QMutexLocker _l2(&initmutex);
    return _enabled && _initialized;
}

bool sh::actions::AbstractActionItem::isChecked()
{
    QMutexLocker _l(&mutex);
    return _ischecked;
}

bool sh::actions::AbstractActionItem::isCheckable()
{
    return _ischeckable;
}

void sh::actions::AbstractActionItem::setText(QString text)
{
    mutex.lock();
    _text = text;
    mutex.unlock();
    emit changed();
}

void sh::actions::AbstractActionItem::setIcon(QString icon)
{
    mutex.lock();
    _icon = icon;
    mutex.unlock();
    emit changed();
}

void sh::actions::AbstractActionItem::setEnabled(bool enabled)
{
    mutex.lock();
    _enabled = enabled;
    mutex.unlock();
    emit changed();
}

void sh::actions::AbstractActionItem::setVisible(bool visible)
{
    mutex.lock();
    _visible = visible;
    mutex.unlock();
    emit changed();
}

bool sh::actions::AbstractActionItem::visible()
{
    QMutexLocker _l(&mutex);
    return _visible;
}

void sh::actions::AbstractActionItem::setChecked(bool checked)
{
    mutex.lock();
    _ischecked = checked;
    mutex.unlock();
    emit changed();
}

void sh::actions::AbstractActionItem::initialize()
{
}

void sh::actions::AbstractActionItem::postinitialize()
{
}

void sh::actions::AbstractActionItem::initializeAsync(std::function<void()> oninitialized)
{
    {
        QMutexLocker _l(&initmutex);
        if (_initialized || _initializing)
            return;
    }
    sh::base::ThreadPool::enqueueForceBeyondAsyncCallBarrier([=](){
        FORBID_ASYNC_CALLS
        initializeSync();
        if (oninitialized)
            oninitialized();
    }, shared_from_this());
}

void sh::actions::AbstractActionItem::reinitialize()
{
    bool callinit = false;
    {
        QMutexLocker _l(&initmutex);
        if (_initializing)
            _reinitialize = true;
        else
        {
            callinit = true;
            _initialized = false;
        }
    }
    sh::base::MainThread::dispatcher()->invokeSync([&](){
        emit changed();
    });
    if (callinit)
        initializeAsync();
}

void sh::actions::AbstractActionItem::initializeSync()
{
    NOT_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    FORBID_ASYNC_CALLS
    {
        QMutexLocker _l(&initmutex);
        _reinitialize = false;
        if (_initialized)
            return;
        while(_initializing && !_initialized)
        {
            sh::base::ThreadPool::respectThreadAbort();
            initcondition.wait(&initmutex, 1000);
        }
        if (_initialized)
            return;
        _initializing = true;
    }
    sh::exceptions::Exception::executeGuarded([&](){
        setEnabled(true);
        initialize();
        postinitialize();
    }, sh::exceptions::ExecuteGuardFlag::ALL_ERRORS_KILL_SHALLOT);
    bool initialized;
    {
        QMutexLocker _l(&initmutex);
        _initializing = false;
        initialized = _initialized = (!_reinitialize);
        initcondition.wakeAll();
    }
    sh::base::MainThread::dispatcher()->invokeSync([&](){
        emit changed();
    });
    if (!initialized)
        initializeSync();
}

bool sh::actions::AbstractActionItem::isInitialized()
{
    QMutexLocker _l(&initmutex);
    return _initialized;
}

bool sh::actions::AbstractActionItem::isInitializing()
{
    QMutexLocker _l(&initmutex);
    return _initializing;
}

std::weak_ptr<sh::actions::AbstractActionItem> sh::actions::AbstractActionItem::parentAction()
{
    return _parentAction;
}

void sh::actions::AbstractActionItem::_setParentAction(std::shared_ptr<sh::actions::AbstractActionItem> parent)
{
    _parentAction = parent;
}

