/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2020 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef ABSTRACTACTIONITEM_H
#define ABSTRACTACTIONITEM_H

#include <functional>
#include <memory>
#include <QIcon>
#include <QMutex>
#include <QWaitCondition>


namespace sh{
    namespace actions{

        /** \brief
         * Reference values for calculating default precedence values of actions.
         */
        class ActionDefaultPrecedenceValues
        {
        public:
            /** \brief
             * Normal file open action.
             */
            static const int PRECEDENCE_OPENFILE;
            /** \brief
             * Special open action.
             */
            static const int PRECEDENCE_BUILTIN_SPECIAL_OPEN;
        };

        /** \brief
         * Abstract base class for executable actions, submenus of them, and more.
         *
         * See the subclasses of this class.
         *
         * They are used at various places. The toolbar and context menus provide them to the user.
         *
         * They can be registered e.g. in a sh::actions::ActionFactory or returned
         * from sh::filesystem::FilesystemHandler.getActions in order to make them available.
         */
        class AbstractActionItem : public QObject, public std::enable_shared_from_this<AbstractActionItem>
        {
            Q_OBJECT

        public:
            /** \brief
             * Returns the displayed text for this action.
             */
            QString text();
            /** \brief
             * Returns the icon for this action.
             */
            QString icon();
            /** \brief
             * Checks if this action is enabled.
             */
            bool enabled();
            /** \brief
             * Checks if this action is checked (has a cross in the ui).
             */
            bool isChecked();
            /** \brief
             * Checks if this action is checkable (can have a cross in the ui).
             */
            bool isCheckable();
            /** \brief
             * Returns the precedence for becoming the default action of a parent submenu.
             *
             * This e.g. leads to a bold label.
             *
             * The action with the highest value becomes the default. Only values >0 will
             * be considered. See ActionDefaultPrecedenceValues for reference values.
             */
            int defaultActionPrecedence() const;
            /** \brief
             * Sets the displayed text.
             */
            void setText(QString text);
            /** \brief
             * Sets the icon.
             */
            void setIcon(QString icon);
            /** \brief
             * Sets if the item is enabled.
             */
            void setEnabled(bool enabled);
            /** \brief
             * Sets if the item is checked (has a cross in the ui).
             */
            void setChecked(bool checked);
            /** \brief
             * Sets the visibility of this item.
             */
            void setVisible(bool visible);
            /** \brief
             * Checks the visibility of this item (non-recursively).
             */
            bool visible();
            /** \brief
             * Returns the parent action, if it is added to a container.
             */
            std::weak_ptr<AbstractActionItem> parentAction();
            /** \brief
             * Sets the parent action.
             * @a_infrastructure
             */
            void _setParentAction(std::shared_ptr<AbstractActionItem> parent);

            /** \brief
             * Asynchonously initializes the action.
             */
            void initializeAsync(std::function<void()> oninitialized=0);
            /** \brief
             * Synchonously initializes the action.
             */
            void initializeSync();
            /** \brief
             * Checks if this action is initialized.
             */
            bool isInitialized();
            /** \brief
             * Checks if this action is initializing.
             */
            bool isInitializing();

            AbstractActionItem();

        signals:
            /** \brief
             * Emits when some data changed.
             */
            void changed();

        public slots:

        protected:
            /** \brief Is (for subclasses) intended to be directly constructed from everywhere or by
             * registering a factory somewhere.
             */
            AbstractActionItem(QString text, bool enabled = true
                    , QString icon = QString()
                    , int defaultActionPrecedence = 0
                    , bool checkable = false, bool ischecked = false);
            /** \brief
             * Initialize the action. This should make the time-consuming parts, e.g. for determining a label or enabled state.
             */
            virtual void initialize();
            /** \brief
             * Late part of initialization for an action. Typically used for infrastructure.
             */
            virtual void postinitialize();
            /** \brief
             * Re-initializes the action asynchonously.
             */
            void reinitialize();

            static QMutex mutex; //static for saving memory
            static QMutex initmutex; //static for saving memory
            static std::shared_ptr<AbstractActionItem> NO_PARENT;

        private:
            QString _text;
            QString _icon;
            bool _enabled;
            bool _visible;
            bool _ischeckable;
            bool _ischecked;
            bool _initialized;
            bool _initializing;
            bool _reinitialize;
            std::weak_ptr<AbstractActionItem> _parentAction = NO_PARENT;
            int _defaultActionPrecedence;
            static QWaitCondition initcondition;
        };

    }
}

#endif // ABSTRACTACTIONITEM_H
