/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2020 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <QTimer>
#include "filesystem/eurl.h"
#include "filesystem/filesystemmodel.h"
#include "filesystem/operation.h"
#include "exceptions/argumentexception.h"
#include "base/mainthread.h"
#include "ui/actionexecutioninfodialog.h"
#include "ui/actionexecutioninfopanel.h"
#include "ui/mainwindow.h"
#include "actionexecutioninfo.h"


sh::actions::ActionExecutionInfo::~ActionExecutionInfo()
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
}

void sh::actions::ActionExecutionInfo::shutdown(bool success)
{
    setDetails(tr("finishing"), "", "", "");
    setProgressIndeterminate();
    if (success)
        _operation.commit();
    foreach(std::shared_ptr<const sh::filesystem::Eurl> changedEurl, changes)
        sh::filesystem::FilesystemModel::instance()->refreshData(changedEurl);
}

sh::actions::ActionExecutionInfo::ActionExecutionInfo(QObject *parent) :
    QObject(parent)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    setHead(tr("Processing"));
    setDetails("", "", "", "");
    setProgressIndeterminate();
    infodialog = createDialog();
    infopanel = createPanel(infodialog.get());
    QTimer::singleShot(4000, this, SLOT(_checkVisibilityOnce()));
}

sh::filesystem::Operation* sh::actions::ActionExecutionInfo::operation()
{
    return &_operation;
}

void sh::actions::ActionExecutionInfo::_checkVisibilityOnce()
{
    mutex.lock();
    bool makevisible = _wasAlwaysInvisible;
    mutex.unlock();
    if (makevisible)
        setVisualProcessFeedbackActive(true);
}

void sh::actions::ActionExecutionInfo::setDetails(QString fromverb, QString fromobj, QString toverb, QString toobj)
{
    mutex.lock();
    _fromverb = fromverb;
    _fromobj = fromobj;
    _toverb = toverb;
    _toobj = toobj;
    mutex.unlock();
    emit detailsChanged();
}

QString sh::actions::ActionExecutionInfo::fromVerb()
{
    mutex.lock();
    QString result = _fromverb;
    mutex.unlock();
    return result;
}

QString sh::actions::ActionExecutionInfo::fromObjectName()
{
    QMutexLocker _l(&mutex);
    return _fromobj;
}

QString sh::actions::ActionExecutionInfo::toVerb()
{
    QMutexLocker _l(&mutex);
    return _toverb;
}

QString sh::actions::ActionExecutionInfo::toObjectName()
{
    QMutexLocker _l(&mutex);
    return _toobj;
}

void sh::actions::ActionExecutionInfo::setVisualProcessFeedbackActive(bool active)
{
    mutex.lock();
    bool changed = active != _visualProcessFeedbackActive;
    _visualProcessFeedbackActive = active;
    _wasAlwaysInvisible = false; // we do this for active=false as well; it offers new ways ;)
    mutex.unlock();
    if (changed)
        emit visualProcessFeedbackActiveChanged();
}

bool sh::actions::ActionExecutionInfo::isVisualProcessFeedbackActive()
{
    QMutexLocker _l(&mutex);
    return _visualProcessFeedbackActive;
}

void sh::actions::ActionExecutionInfo::setManualInterventionNeeded(bool isneeded)
{
    mutex.lock();
    bool changed = isneeded != _manualInterventionNeeded;
    _manualInterventionNeeded = isneeded;
    _wasAlwaysInvisible = false; // we do this for isneeded=false as well; it offers new ways ;)
    mutex.unlock();
    if (changed)
        emit manualInterventionNeededChanged();
}

bool sh::actions::ActionExecutionInfo::isManualInterventionNeeded()
{
    QMutexLocker _l(&mutex);
    return _manualInterventionNeeded;
}

void sh::actions::ActionExecutionInfo::setHead(QString text)
{
    mutex.lock();
    _head = text;
    mutex.unlock();
    emit headChanged();
}

QString sh::actions::ActionExecutionInfo::head()
{
    QMutexLocker _l(&mutex);
    return _head;
}

void sh::actions::ActionExecutionInfo::setProgressIndeterminate(QString text)
{
    _progressDeterminate = false;
    _progressText = text;
    emit progressChanged();
}

void sh::actions::ActionExecutionInfo::setProgress(quint64 done, quint64 all, QString text)
{
    mutex.lock();
    _progressDeterminate = true;
    _progressDone = done;
    _progressAll = all;
    _progressText = text;
    mutex.unlock();
    emit progressChanged();
}

bool sh::actions::ActionExecutionInfo::isProgressDeterminate()
{
    QMutexLocker _l(&mutex);
    return _progressDeterminate;
}

quint64 sh::actions::ActionExecutionInfo::progressDone()
{
    QMutexLocker _l(&mutex);
    return _progressDone;
}

quint64 sh::actions::ActionExecutionInfo::progressAll()
{
    QMutexLocker _l(&mutex);
    return _progressAll;
}

QString sh::actions::ActionExecutionInfo::progressText()
{
    QMutexLocker _l(&mutex);
    return _progressText;
}

bool sh::actions::ActionExecutionInfo::isCancelled()
{
    QMutexLocker _l(&mutex);
    return _iscancelled;
}

void sh::actions::ActionExecutionInfo::cancel()
{
    mutex.lock();
    _iscancelled = true;
    mutex.unlock();
    emit wasCancelled();
}

sh::actions::ActionExecutionUserFeedback* sh::actions::ActionExecutionInfo::userfeedback()
{
    return infodialog.get();
}

std::shared_ptr<sh::ui::ActionExecutionInfoDialog> sh::actions::ActionExecutionInfo::createDialog()
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    auto dlg = sh::ui::MainWindow::mainWindow()->createActionExecutionInfoDialog(this);
    std::function<void()> visibilityChanged = [=](){
        sh::base::MainThread::dispatcher()->invokeSync([=](){
            dlg->setLogicallyVisible(isVisualProcessFeedbackActive());
        });
    };
    std::function<void()> detailsChanged = [=](){
        sh::base::MainThread::dispatcher()->invokeSync([=](){
            dlg->setDetails(fromVerb(), fromObjectName(), toVerb(), toObjectName());
        });
    };
    std::function<void()> headChanged = [=](){
        sh::base::MainThread::dispatcher()->invokeSync([=](){
            dlg->setHead(head());
        });
    };
    std::function<void()> progressChanged = [=](){
        sh::base::MainThread::dispatcher()->invokeSync([=](){
            dlg->setProgress(isProgressDeterminate(), progressDone(), progressAll(), progressText());
        });
    };
    std::function<void()> manualInterventionNeededChanged = [=](){
        sh::base::MainThread::dispatcher()->invokeSync([=](){
            dlg->setForceForeground(isManualInterventionNeeded());
        });
    };
    connect(this, &sh::actions::ActionExecutionInfo::detailsChanged, this, detailsChanged);
    connect(this, &sh::actions::ActionExecutionInfo::headChanged, this, headChanged);
    connect(this, &sh::actions::ActionExecutionInfo::progressChanged, this, progressChanged);
    connect(this, &sh::actions::ActionExecutionInfo::visualProcessFeedbackActiveChanged, this, visibilityChanged);
    connect(this, &sh::actions::ActionExecutionInfo::manualInterventionNeededChanged, this, manualInterventionNeededChanged);
    detailsChanged();
    headChanged();
    progressChanged();
    return dlg;
}

std::shared_ptr<sh::ui::ActionExecutionInfoPanel> sh::actions::ActionExecutionInfo::createPanel(sh::ui::ActionExecutionInfoDialog* dialog)
{
    std::shared_ptr<sh::ui::ActionExecutionInfoPanel> panel = sh::ui::MainWindow::mainWindow()->addInfoPanel(this);
    std::function<void()> visibilityChanged = [=](){
        sh::base::MainThread::dispatcher()->invokeSync([=](){
            panel->setPanelVisible(isVisualProcessFeedbackActive() || isManualInterventionNeeded());
        });
    };
    std::function<void()> headChanged = [=](){
        sh::base::MainThread::dispatcher()->invokeSync([=](){
            panel->setLabel(head());
        });
    };
    std::function<void()> progressChanged = [=](){
        sh::base::MainThread::dispatcher()->invokeSync([=](){
            panel->setProgress(isProgressDeterminate(), progressDone(), progressAll());
        });
    };
    std::function<void()> panelClicked = [=](){
        sh::base::MainThread::dispatcher()->invokeSync([=](){
            dialog->setBackground(!dialog->isBackground()); // not used exactly this way by web ui
        });
    };
    std::function<void()> manualInterventionNeededChanged = [=](){
        sh::base::MainThread::dispatcher()->invokeSync([=](){
            panel->setForceForeground(isManualInterventionNeeded());
            visibilityChanged();
        });
    };
    connect(this, &sh::actions::ActionExecutionInfo::headChanged, this, headChanged);
    connect(this, &sh::actions::ActionExecutionInfo::progressChanged, this, progressChanged);
    connect(this, &sh::actions::ActionExecutionInfo::visualProcessFeedbackActiveChanged, this, visibilityChanged);
    connect(this, &sh::actions::ActionExecutionInfo::manualInterventionNeededChanged, this, manualInterventionNeededChanged);
    panel->onClicked(panelClicked, this);
    detailsChanged();
    headChanged();
    progressChanged();
    panel->setPanelVisible(false);
    return panel;
}

void sh::actions::ActionExecutionInfo::withExecuteGuards_infodlg(std::function<void()> fct, int flags)
{
    sh::exceptions::Exception::HandlerSettings* handlerSettings = sh::exceptions::ExceptionHandlerSettingsManager::instance()->handlerSettings();
    auto x1 = sh::exceptions::Exception::createRegisterHandler<std::function<void(sh::exceptions::Exception&)> >([=](sh::exceptions::Exception& ex)
    {
        QString msg = ex.message()+"\n\n";
        if (ex.isDetailsAreInteresting())
            msg += ex.details()+"\n\n";
        msg += tr("The action is aborted.");
        userfeedback()->simpleMessageBox(msg, (int)sh::actions::ActionExecutionUserFeedback::MessageBoxButton::Continue,
                                              "dialog-error", sh::actions::ActionExecutionUserFeedback::MessageBoxButton::Continue, sh::actions::ActionExecutionUserFeedback::MessageBoxButton::Continue);
    }, &handlerSettings->exceptionHandler_resumeable);
    auto x2 = sh::exceptions::Exception::createRegisterHandler<std::function<bool(sh::exceptions::Exception&)> >([=](sh::exceptions::Exception& ex)
    {
        QString msg = ex.message()+"\n\n";
        if (ex.isDetailsAreInteresting())
            msg += ex.details()+"\n\n";
        msg += tr("Do you want to retry?");
        int answer = userfeedback()->simpleMessageBox(msg,
                                                (int)sh::actions::ActionExecutionUserFeedback::MessageBoxButton::Retry | (int)sh::actions::ActionExecutionUserFeedback::MessageBoxButton::Continue
                                                        | ((handlerSettings->exceptionHandler_cancel.isEmpty()) ? 0 : (int)sh::actions::ActionExecutionUserFeedback::MessageBoxButton::Cancel),
                                                "dialog-error", sh::actions::ActionExecutionUserFeedback::MessageBoxButton::Retry, sh::actions::ActionExecutionUserFeedback::MessageBoxButton::Cancel);
        if (answer == sh::actions::ActionExecutionUserFeedback::MessageBoxButton::Cancel)
        {
            cancel();
            respectCancel();
        }
        return answer == sh::actions::ActionExecutionUserFeedback::MessageBoxButton::Retry;
    }, &handlerSettings->exceptionHandler_retryable);
    sh::exceptions::Exception::executeGuarded(fct, flags);
}

void sh::actions::ActionExecutionInfo::respectCancel()
{
    if (isCancelled() && !sh::exceptions::ExceptionHandlerSettingsManager::instance()->handlerSettings()->exceptionHandler_cancel.isEmpty()) // should be thread-safe
        throw sh::exceptions::CancelException();
}


void sh::actions::ActionExecutionInfo::addChangedEurl(std::shared_ptr<const sh::filesystem::Eurl> eurl)
{
    changes.append(eurl);
}

void sh::actions::ActionExecutionInfo::startExecution(std::shared_ptr<sh::actions::AbstractActionItem> action)
{
    stack.push(action);
}

void sh::actions::ActionExecutionInfo::finishExecution()
{
    stack.pop();
}
