/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2020 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef ACTIONEXECUTIONINFO_H
#define ACTIONEXECUTIONINFO_H

#include <functional>
#include <memory>
#include <QMutex>
#include <QStack>
#include "filesystem/operation.h"


namespace sh{
    namespace filesystem{class Eurl;}
    namespace ui{
        class ActionExecutionInfoDialog;
    }

    namespace actions{
        class AbstractActionItem;
        class ActionExecutionUserFeedback;

        /** \brief
         * Programming interface for giving status infos and communication with
         * the user in an action execution.
         *
         * An instance is always available during an execution.
         */
        class ActionExecutionInfo : public QObject
        {
            Q_OBJECT
        public:

            /** \brief Constructed only by the infrastructure and made available otherwise.
             */
            explicit ActionExecutionInfo(QObject *parent = 0);

            virtual ~ActionExecutionInfo();

            /** \brief Specifies if there must be a visual process feedback.
             *
             * The visual process feedback is typically realized as a dialog showing progress information.
             *
             * Note: It can also appear in some situations, even if this property is set to false.
             */
            void setVisualProcessFeedbackActive(bool active);

            /** \brief Determines if there must be a visual process feedback.
             */
            bool isVisualProcessFeedbackActive();

            /** \brief Specifies if manual input is currently required.
             */
            void setManualInterventionNeeded(bool isneeded);

            /** \brief Determines if manual input is currently required.
             */
            bool isManualInterventionNeeded();

            /** \brief Triggers the beginning of the actual execution.
             * @a_infrastructure
             */
            void startExecution(std::shared_ptr<sh::actions::AbstractActionItem> action);

            /** \brief Triggers the finishing of the execution.
             * @a_infrastructure
             */
            void finishExecution();

            /** \brief Sets some detail texts.
             */
            void setDetails(QString fromverb, QString fromobj, QString toverb = QString(), QString toobj = QString());

            /** \brief Gets detail text 'from-verb'.
             */
            QString fromVerb();

            /** \brief Gets detail text 'from-objectname'.
             */
            QString fromObjectName();

            /** \brief Gets detail text 'to-verb'.
             */
            QString toVerb();

            /** \brief Gets detail text 'to-objectname'.
             */
            QString toObjectName();

            /** \brief Sets the header text.
             */
            void setHead(QString text);

            /** \brief Gets the header text.
             */
            QString head();

            /** \brief Sets progress information to 'indeterminate'.
             */
            void setProgressIndeterminate(QString text = QString());

            /** \brief Sets progress information.
             */
            void setProgress(quint64 done, quint64 all, QString text);

            /** \brief Gets if progress information is 'indeterminate'.
             */
            bool isProgressDeterminate();

            /** \brief Gets count of finished items.
             */
            quint64 progressDone();

            /** \brief Gets count of all items.
             */
            quint64 progressAll();

            /** \brief Gets progress information text.
             */
            QString progressText();

            /** \brief Program interface for interactive user input.
             */
            sh::actions::ActionExecutionUserFeedback* userfeedback();

            /** \brief Gets if the current action execution was cancelled.
             */
            bool isCancelled();

            /** \brief Cancel the current action execution.
             *
             * Used from outside, e.g. as handler for a 'Cancel' button.
             */
            void cancel();

            /** \brief Called regularly from an action execution code in order
             * to cancel execution at that place, if the user has cancelled it.
             */
            void respectCancel();

            /** \brief Executes code with different exception handlers for presenting
             * exceptions in the action execution inside the action dialog (instead of
             * the default exception dialog).
             */
            void withExecuteGuards_infodlg(std::function<void()>, int flags=0);

            /** \brief Program interface for operations on the filesystem (reading or writing).
             */
            sh::filesystem::Operation* operation();

            /** \brief Reminds an important change of a certain node in the filesystem (file created, modified, removed?).
             *
             * The infrastructure will force the ui to refresh those nodes.
             */
            void addChangedEurl(std::shared_ptr<const sh::filesystem::Eurl> eurl);

            /** \brief Shuts down this execution info and releases resources.
             */
            void shutdown(bool success);

        signals:

            /** \brief Signal for changed execution details.
             */
            void detailsChanged();

            /** \brief Signal for changed head.
             */
            void headChanged();

            /** \brief Signal for changed progress.
             */
            void progressChanged();

            /** \brief Signal for changed visibility-required.
             */
            void visualProcessFeedbackActiveChanged();

            /** \brief Signal for cancellation on user behalf.
             */
            void wasCancelled();

            /** \brief Signal for changed manual-intervention-needed.
             */
            bool manualInterventionNeededChanged();

        private slots:
            void _checkVisibilityOnce();

        private:
            QMutex mutex;
            QString _head;
            QString _fromverb, _fromobj, _toverb, _toobj;
            bool _progressDeterminate;
            quint64 _progressDone, _progressAll;
            bool _visualProcessFeedbackActive = false;
            bool _wasAlwaysInvisible = true;
            bool _manualInterventionNeeded = false;
            QString _progressText;
            std::shared_ptr<sh::ui::ActionExecutionInfoDialog> infodialog = 0;
            std::shared_ptr<sh::ui::ActionExecutionInfoPanel> infopanel = 0;
            std::shared_ptr<sh::ui::ActionExecutionInfoDialog> createDialog();
            std::shared_ptr<ui::ActionExecutionInfoPanel> createPanel(sh::ui::ActionExecutionInfoDialog *dialog);
            bool _iscancelled = false;
            sh::filesystem::Operation _operation;
            QList <std::shared_ptr<const sh::filesystem::Eurl>> changes;
            QStack<std::shared_ptr<sh::actions::AbstractActionItem>> stack;

        };

    }
}

#endif // ACTIONEXECUTIONINFO_H
