/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2020 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef ACTIONEXECUTIONUSERFEEDBACK_H
#define ACTIONEXECUTIONUSERFEEDBACK_H

#include <QList>
#include <QPair>


namespace sh{
    namespace actions{

        /** \brief Methods for user interaction in an action execution.
         *
         * When an action executes, it may need to ask the user for some input at some time. An action
         * implementation can do so by calling those methods. It is available as a
         * member of the ActionExecutionInfo class. An instance of it is available in each action execution.
         */
        class ActionExecutionUserFeedback
        {
        public:

            /** \brief A data structure for a choice (in an GridformEntry).
             */
            class Choice
            {
            public:
                QString label;
                QString text;
                Choice(QString label, QString text=QString());
            };

            /** \brief A data structure for a list of Choice instances.
             */
            class Choices
            {
            public:
                QList<Choice*> list;
                int selected = -1;
            };

            /** \brief A data structure for an entry in a simple chooser grid form.
             */
            class GridformEntry
            {
            public:
                QString label;
                Choices choices;
                /** \brief Constructed only indirectly.
                 */
                GridformEntry(QString label, Choices choices);
                virtual ~GridformEntry();
                int addChoice(QString label, QString text=QString());
            };

            /** \brief A data structure for a list of entries in a simple chooser grid form.
             */
            class GridformEntries
            {
            public:
                /** \brief Is intended to be directly constructed from everywhere.
                 */
                GridformEntries();
                virtual ~GridformEntries();
                GridformEntry* newEntry(QString label);
                QList<GridformEntry*> list;
            };

            /** \brief
             * Buttons in a message box from ActionExecutionUserFeedback.
             */
            enum MessageBoxButton
            {
                NONE = 0,
                OK = 1 << 0,
                Continue = 1 << 1,
                Cancel = 1 << 2,
                Retry = 1 << 3,
                Yes = 1 << 4,
                No = 1 << 5,
            };

            virtual int messageBox(QString text, QList<QString> answers, QString icon=QString(), int defaultanswer=-1, int cancelanswer=-1, QList<QString> answericons=QList<QString>()) =0;
            virtual int inputBox(QString text, QList<QString> answers, QString *value, QString icon=QString(), int defaultanswer=-1, int cancelanswer=-1, int valuePreselectFrom=-1, int valuePreselectTo=-1) =0;
            virtual int multilineInputBox(QString text, QList<QString> answers, QString *value, QString icon=QString(), int defaultanswer=-1, int cancelanswer=-1) =0;
            virtual int simpleChooserGridform(QString text, GridformEntries* entries, QStringList answers, int defaultanswer=-1, int cancelanswer=-1) =0;
            virtual bool credentialsDialog(QString text, bool showDomain, bool showUsername, bool showPassword, bool showAnonymous, bool showRemember, QString* domain, QString* username, QString* password, bool* isAnonymous, bool* isRemember) =0;
            virtual bool unixPermissionsDialog(bool* userMayRead, bool* userMayWrite, bool* userMayExecute, bool* groupMayRead, bool* groupMayWrite, bool* groupMayExecute, bool* othersMayRead, bool* othersMayWrite, bool* othersMayExecute, bool* sticky, bool* setuid, bool* setgid, QStringList users, QStringList groups, QString* ownerUser, QString* ownerGroup) =0;

            QString simpleInputBox(QString text, QString deflt, int valuePreselectFrom=-1, int valuePreselectTo=-1);
            int simpleMessageBox(QString text, int buttons=(int)MessageBoxButton::OK, QString icon=QString(), int defaultbutton=(MessageBoxButton)0, int cancelbutton=(MessageBoxButton)0);

            virtual ~ActionExecutionUserFeedback();

        };

    }
}

#endif // ACTIONEXECUTIONUSERFEEDBACK_H
