/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2020 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "filesystem/filesystemnode.h"
#include "filesystem/filesystemhandler.h"
#include "filesystem/filesystemoperation.h"
#include "filesystem/operation.h"
#include "exceptions/argumentexception.h"
#include "base/mainthread.h"
#include "actionfactory.h"


sh::actions::ActionInstantiation::ActionInstantiation(QList<std::shared_ptr<sh::filesystem::FilesystemNode>> selectedNodes, std::shared_ptr<filesystem::FilesystemNode> currentDirectory)
            : selectedNodes(selectedNodes), currentDirectory(currentDirectory)
{
}

sh::actions::ActionInstantiation::ActionInstantiation(const ActionInstantiation& s, std::shared_ptr<sh::actions::AbstractActionItem> createdAction)
            : ActionInstantiation(s)
{
    this->createdAction = createdAction;
}

sh::actions::Predicate::Predicate()
{
}

void sh::actions::Predicate::prepare(ActionInstantiation*)
{
}

bool sh::actions::Predicate::evaluate(ActionInstantiation*)
{
    return true;
}

sh::actions::Predicate::~Predicate()
{
}

sh::actions::OnDirectoriesPredicate::OnDirectoriesPredicate()
    : sh::actions::Predicate::Predicate()
{
}

bool sh::actions::OnDirectoriesPredicate::evaluate(ActionInstantiation* instantiation)
{
    for(auto node : instantiation->selectedNodes) {
        auto nodetype = node->nodetype();
        if (nodetype != sh::filesystem::FilesystemNodeType::Directory && nodetype != sh::filesystem::FilesystemNodeType::SpecialTreeOnlyDirectory)
            return false;
    }
    return true;
}

sh::actions::OnFilesPredicate::OnFilesPredicate()
    : sh::actions::Predicate::Predicate()
{
}

bool sh::actions::OnFilesPredicate::evaluate(ActionInstantiation* instantiation)
{
    for(auto node : instantiation->selectedNodes)
        if (node->nodetype() != sh::filesystem::FilesystemNodeType::File)
            return false;
    return true;
}

sh::actions::OnLinksPredicate::OnLinksPredicate()
    : sh::actions::Predicate::Predicate()
{
}

bool sh::actions::OnLinksPredicate::evaluate(ActionInstantiation* instantiation)
{
    if (instantiation->resolveLinks)
        throw sh::exceptions::ArgumentException(sh::exceptions::ExceptionData().details("OnLinksPredicate is only allowed together with DontResolveLinksPredicate").resumeable(false));
    for(auto node : instantiation->selectedNodes)
        if (node->nodetype() != sh::filesystem::FilesystemNodeType::Link)
            return false;
    return true;
}

sh::actions::OnSingleEntrySelectionPredicate::OnSingleEntrySelectionPredicate()
    : sh::actions::Predicate::Predicate()
{
}

bool sh::actions::OnSingleEntrySelectionPredicate::evaluate(ActionInstantiation* instantiation)
{
    if (instantiation->selectedNodes.length() != 1)
        return false;
    return true;
}

sh::actions::ByRegExpPredicate::ByRegExpPredicate(QString regexp)
    : sh::actions::Predicate::Predicate(), filterRegExp(regexp)
{
}

bool sh::actions::ByRegExpPredicate::evaluate(ActionInstantiation* instantiation)
{
    for(auto node : instantiation->selectedNodes)
        if (!filterRegExp.exactMatch(node->eurl()->asString()))
            return false;
    return true;
}

sh::actions::DontResolveLinksPredicate::DontResolveLinksPredicate()
    : sh::actions::Predicate::Predicate()
{
}

void sh::actions::DontResolveLinksPredicate::prepare(ActionInstantiation* instantiation)
{
    instantiation->resolveLinks = false;
}

sh::actions::HideOnCurrentDirectoryLevelPredicate::HideOnCurrentDirectoryLevelPredicate()
    : sh::actions::Predicate::Predicate()
{
}

bool sh::actions::HideOnCurrentDirectoryLevelPredicate::evaluate(ActionInstantiation* instantiation)
{
    return !instantiation->isLookupOnCurrentDirectoryLevel;
}

sh::actions::HideOnSelectionLevelPredicate::HideOnSelectionLevelPredicate()
    : sh::actions::Predicate::Predicate()
{
}

bool sh::actions::HideOnSelectionLevelPredicate::evaluate(ActionInstantiation* instantiation)
{
    return instantiation->isLookupOnCurrentDirectoryLevel;
}

sh::actions::KeyShortcutPredicate::KeyShortcutPredicate(QKeySequence shortcut, bool triggersOnCurrentDirectoryLevel)
    : sh::actions::Predicate::Predicate(), shortcut(shortcut), triggersOnCurrentDirectoryLevel(triggersOnCurrentDirectoryLevel)
{
}

void sh::actions::KeyShortcutPredicate::prepare(ActionInstantiation* instantiation)
{
    if (triggersOnCurrentDirectoryLevel == instantiation->isLookupOnCurrentDirectoryLevel)
        instantiation->shortcut = shortcut;
}


sh::actions::PositionIndexPredicate::PositionIndexPredicate(int positionIndex)
    : sh::actions::Predicate::Predicate(), positionIndex(positionIndex)
{
}

void sh::actions::PositionIndexPredicate::prepare(ActionInstantiation* instantiation)
{
    instantiation->positionIndex = positionIndex;
}

sh::actions::AbstractActionFactory::AbstractActionFactory(std::shared_ptr<ActionCategory> category, QList<std::shared_ptr<Predicate> > predicates)
    : category(category), predicates(predicates)
{

}

sh::actions::AbstractActionFactory::~AbstractActionFactory()
{
}

std::shared_ptr<sh::actions::ActionInstantiation> sh::actions::AbstractActionFactory::actionAvailable(ActionInstantiation* _situation)
{
    NOT_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    std::shared_ptr<ActionInstantiation> resultsituation = std::make_shared<ActionInstantiation>(*_situation);
    resultsituation->actionfactory = this;
    resultsituation->category = category;
    for(std::shared_ptr<Predicate> predicate : predicates)
        predicate->prepare(resultsituation.get());
    if (resultsituation->resolveLinks) {
        auto oselectednodes = resultsituation->selectedNodes;
        resultsituation->selectedNodes.clear();
        sh::filesystem::Operation operation;
        foreach(std::shared_ptr<sh::filesystem::FilesystemNode> n, oselectednodes)
        {
            QList<std::shared_ptr<sh::filesystem::FilesystemNode>> ln = operation.filesystem()->resolveNodeLink(n);
            resultsituation->selectedNodes.append(ln.isEmpty() ? n : ln.first());
        }
    }
    for(std::shared_ptr<Predicate> predicate : predicates)
        if (!predicate->evaluate(resultsituation.get()))
            return 0;
    return resultsituation;
}

