/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2020 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <QShortcut>
#include <QStack>
#include "filesystem/filesystemhandler.h"
#include "filesystem/eurl.h"
#include "filesystem/filesystemnode.h"
#include "base/threadpool.h"
#include "tools/benchmarking.h"
#include "base/mainthread.h"
#include "abstractactionitem.h"
#include "actionactionitem.h"
#include "actionsmanager.h"
#include "separatoractionitem.h"
#include "submenuactionitem.h"

REGISTER_SINGLETON(sh::actions, ActionsManager, "base")


sh::actions::ActionsManager::ActionsManager()
{
}

void sh::actions::ActionsManager::registerActionFactory(std::shared_ptr<sh::actions::AbstractActionFactory> actionFactory)
{
    QMutexLocker _l(&_mutex_actionfactories);
    _actionfactories.append(actionFactory);
}

void sh::actions::ActionsManager::sortAndSeparateActions(QList<std::shared_ptr<sh::actions::ActionInstantiation> >* actions)
{
    qSort(actions->begin(), actions->end(), [](const std::shared_ptr<sh::actions::ActionInstantiation>& a, const std::shared_ptr<sh::actions::ActionInstantiation>& b){
        return a->positionIndex < b->positionIndex;
    });
    int lastgroup=-1;
    for (int i=actions->length()-1; i>=0; i--) {
        std::shared_ptr<sh::actions::ActionInstantiation> acsi = actions->at(i);
        int group = acsi->positionIndex / 100000;
        if (group != lastgroup) {
            auto sepsit = std::make_shared<sh::actions::ActionInstantiation>(QList<std::shared_ptr<sh::filesystem::FilesystemNode>>(), (std::shared_ptr<sh::filesystem::FilesystemNode>)0);
            sepsit->createdAction = std::make_shared<sh::actions::SeparatorActionItem>();
            actions->insert(i+1, sepsit);
            lastgroup = group;
        }
        else if (lastgroup == -1)
            lastgroup = group;
    }
}

void sh::actions::ActionsManager::_getDefaultAction_helper(QList<std::shared_ptr<sh::actions::AbstractActionItem>> actionList, std::shared_ptr<GDAHstruct> gdah)
{
    QList<std::shared_ptr<sh::actions::AbstractActionItem>> interestingList;
    foreach (auto x, actionList) {
        if (x->defaultActionPrecedence() > 0) {
            gdah->counter.doIncValue();
            interestingList.append(x);
        }
    }
    foreach (auto x, interestingList)
        sh::base::ThreadPool::enqueueForce([=](){
            if (!x->isInitialized())
                x->initializeSync();
            if (x->enabled() && x->visible()) {
                sh::actions::SubmenuActionItem* itmSubmenu = qobject_cast<sh::actions::SubmenuActionItem*>(x.get());
                sh::actions::ActionActionItem* itmAction = qobject_cast<sh::actions::ActionActionItem*>(x.get());
                if (itmSubmenu)
                    _getDefaultAction_helper(itmSubmenu->subitems(), gdah);
                else if (itmAction) {
                    if (x->defaultActionPrecedence() > gdah->currentValue) {
                        gdah->currentAction = x;
                        gdah->currentValue = x->defaultActionPrecedence();
                    }
                }
            }
            if (gdah->counter.doDecValue() == 0)
                gdah->callback(std::dynamic_pointer_cast<sh::actions::ActionActionItem>(gdah->currentAction));
        });
}

void sh::actions::ActionsManager::getDefaultAction(QList<std::shared_ptr<sh::actions::AbstractActionItem>> actionList, std::function<void(std::shared_ptr<sh::actions::ActionActionItem>)> callback)
{
    std::shared_ptr<GDAHstruct> gdah = std::make_shared<GDAHstruct>();
    gdah->callback = callback;
    gdah->initialActionList = actionList;
    _getDefaultAction_helper(actionList, gdah);
}

QString sh::actions::ActionsManager::selectionLabel(QList<std::shared_ptr<sh::filesystem::FilesystemNode>> selection)
{
    int sellen = selection.length();
    if (sellen == 1) {
        QString sellbl = selection.first()->eurl()->basename();
        if (sellbl.length() > 15)
            sellbl = sellbl.left(6) + " ⋯ " + sellbl.right(6);
        return QObject::tr("Selection '%1'").arg(sellbl);
    }
    else
        return QObject::tr("%1 selected items").arg(sellen);
}

QString sh::actions::ActionsManager::directoryLabel(std::shared_ptr<sh::filesystem::FilesystemNode> directory)
{
    QString dirlbl = directory->eurl()->basename();
    if (dirlbl.length() > 15)
        dirlbl = dirlbl.left(6) + " ⋯ " + dirlbl.right(6);
    return QObject::tr("Directory '%1'").arg(dirlbl);
}

QList<std::shared_ptr<sh::actions::ActionInstantiation>> sh::actions::ActionsManager::_getActions_raw(std::shared_ptr<sh::actions::ActionInstantiation> instantiation)
{
    NOT_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    QList<std::shared_ptr<sh::actions::ActionInstantiation>> result;
    // get actions from handler (if we have a common one)
    bool sameHandler = true;
    sh::filesystem::FilesystemHandler* handler = 0;
    foreach (std::shared_ptr<sh::filesystem::FilesystemNode> node, instantiation->selectedNodes) {
        if (handler == 0 && sameHandler)
            handler = node->handler();
        else if (handler != node->handler())
        {
            sameHandler = false;
            break;
        }
    }
    if (sameHandler && handler) {
        QList<std::shared_ptr<const sh::filesystem::Eurl>> eurls;
        foreach (std::shared_ptr<sh::filesystem::FilesystemNode> node, instantiation->selectedNodes)
            eurls.append(node->eurl());
        {
            FORBID_ASYNC_CALLS
            QList< std::shared_ptr<sh::actions::AbstractActionItem> > actions;
            sh::exceptions::Exception::executeGuarded_errorpanel([&](){
                actions = handler->getActions(eurls);
            });
            foreach (std::shared_ptr<sh::actions::AbstractActionItem> actionitem, actions) {
                auto hai = std::make_shared<ActionInstantiation>(*instantiation, actionitem);
                hai->category = sh::actions::ActionCategory::getByName("manage");
                auto aactionitem = std::dynamic_pointer_cast<ActionActionItem>(actionitem);
                if (aactionitem && instantiation->isLookupOnCurrentDirectoryLevel == aactionitem->shortcuthintTriggersOnCurrentDirectory())
                    hai->shortcut = aactionitem->shortcuthint();
                result.append(hai);
            }
        }
    }
    // get actions from the registered action factories
    QList<std::shared_ptr<sh::actions::ActionInstantiation>> flist;
    {
        QMutexLocker _l(&_mutex_actionfactories);
        foreach (std::shared_ptr<sh::actions::AbstractActionFactory> actionfactory, _actionfactories) {
            auto rsituation = actionfactory->actionAvailable(instantiation.get());
            if (rsituation)
                flist.append(rsituation);
        }
    }
    sh::base::MainThread::dispatcher()->invokeSync([&](){
        foreach(std::shared_ptr<sh::actions::ActionInstantiation> rsituation, flist)
        {
            sh::exceptions::Exception::executeGuarded([&](){
                std::shared_ptr<sh::actions::AbstractActionItem> act;
                {
                    FORBID_ASYNC_CALLS
                    rsituation->createdAction = rsituation->actionfactory->construct(rsituation->selectedNodes);
                    std::shared_ptr<sh::actions::ActionActionItem> itmAction = std::dynamic_pointer_cast<sh::actions::ActionActionItem>(rsituation->createdAction);
                    if (rsituation->shortcut.isEmpty() && itmAction && itmAction->shortcuthintTriggersOnCurrentDirectory() == rsituation->isLookupOnCurrentDirectoryLevel)
                        rsituation->shortcut = itmAction->shortcuthint();
                }
                if (rsituation->createdAction)
                    result.append(rsituation);
            }, sh::exceptions::ExecuteGuardFlag::ALL_ERRORS_KILL_SHALLOT);
        }
    });
    return result;
}

void sh::actions::ActionsManager::getActions(std::shared_ptr<sh::actions::ActionInstantiation> instantiation, std::function<void(QList<std::shared_ptr<sh::actions::ActionInstantiation>>, QList<std::shared_ptr<sh::actions::ActionInstantiation>>)> callback, std::function<void(std::shared_ptr<sh::actions::ActionInstantiation>)> initializedcallback)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    sh::base::ThreadPool::enqueueForce([=](){
        QList<std::shared_ptr<sh::actions::ActionInstantiation> > selacts = _getActions_raw(instantiation);
        auto dsituation = std::make_shared<sh::actions::ActionInstantiation>(*instantiation.get());
        dsituation->selectedNodes = {instantiation->currentDirectory};
        dsituation->isLookupOnCurrentDirectoryLevel = true;
        QList<std::shared_ptr<sh::actions::ActionInstantiation> > diracts = _getActions_raw(dsituation);
        for (auto actions: {&selacts, &diracts})
            for (std::shared_ptr<sh::actions::ActionInstantiation> action: *actions)
                action->createdAction->initializeAsync([=](){
                    sh::base::MainThread::dispatcher()->invokeSync([=](){
                        initializedcallback(action);
                    });
                });
        callback(selacts, diracts);
    });
}

void sh::actions::ActionsManager::recursivelyInitializeAction(std::shared_ptr<sh::actions::ActionInstantiation> actsit, std::function<void(std::shared_ptr<sh::actions::ActionInstantiation>)> callback, QList<std::shared_ptr<sh::actions::ActionInstantiation>> sits)
{
    sits.append(actsit);
    sh::base::ThreadPool::enqueueForce([=](){
        QList<std::shared_ptr<sh::actions::ActionInstantiation>> _sits = sits;  // just preserve the parent pointers until the bitter end
        std::shared_ptr<sh::actions::AbstractActionItem> action = actsit->createdAction;
        action->initializeSync();
        sh::actions::SubmenuActionItem* asubmenu = qobject_cast<sh::actions::SubmenuActionItem*>(action.get());
        sh::actions::ActionActionItem* aaction = qobject_cast<sh::actions::ActionActionItem*>(action.get());
        if (asubmenu) {
            for (auto subitm : asubmenu->subitems()) {
                auto subsit = std::make_shared<sh::actions::ActionInstantiation>(*actsit);
                subsit->createdAction = subitm;
                auto asubitm = std::dynamic_pointer_cast<sh::actions::ActionActionItem>(subitm);
                if (asubitm && actsit->isLookupOnCurrentDirectoryLevel == asubitm->shortcuthintTriggersOnCurrentDirectory())
                    subsit->shortcut = asubitm->shortcuthint();
                recursivelyInitializeAction(subsit, callback, sits);
            }
        }
        if (aaction || asubmenu)
            callback(actsit);
    });
}

void sh::actions::ActionsManager::getActionForShortcut(QList<std::shared_ptr<sh::filesystem::FilesystemNode>> selectedNodes, std::shared_ptr<sh::filesystem::FilesystemNode> folderNode,
                                          int key, Qt::KeyboardModifiers modifiers,
                                          std::function<void(std::shared_ptr<sh::actions::AbstractActionItem>)> callback)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    BENCHMARK_SCOPE("ActionsManager::getActionForShortcut")
    int keyInt = key;
    if (modifiers & Qt::ShiftModifier)
        keyInt += Qt::SHIFT;
    if (modifiers & Qt::ControlModifier)
        keyInt += Qt::CTRL;
    if (modifiers & Qt::AltModifier)
        keyInt += Qt::ALT;
    if (modifiers & Qt::MetaModifier)
        keyInt += Qt::META;
    QKeySequence pressedSequence(keyInt);
    auto instantiation = std::make_shared<sh::actions::ActionInstantiation>(selectedNodes, folderNode);
    getActions(instantiation, [](QList<std::shared_ptr<sh::actions::ActionInstantiation>>, QList<std::shared_ptr<sh::actions::ActionInstantiation>>){},
                                    [=](std::shared_ptr<sh::actions::ActionInstantiation> action){
        recursivelyInitializeAction(action, [=](std::shared_ptr<sh::actions::ActionInstantiation> x){
            if (qobject_cast<sh::actions::ActionActionItem*>(x->createdAction.get()) && x->shortcut.matches(pressedSequence) == QKeySequence::ExactMatch)
                callback(x->createdAction);
        });
    });
}

sh::actions::ActionsManager::RunningActionsCounter::RunningActionsCounter(ActionsManager* manager, std::shared_ptr<sh::actions::ActionActionItem> action) : _manager(manager), _action(action)
{
    {
        QMutexLocker _l1(&_manager->_mutex_runningactions);
        manager->_runningactions.append(action);
    }
    sh::base::MainThread::dispatcher()->invokeSync([=](){
        _manager->_emit_runningActionsChanged();
    });
}

sh::actions::ActionsManager::RunningActionsCounter::~RunningActionsCounter()
{
    {
        QMutexLocker _l1(&_manager->_mutex_runningactions);
        _manager->_runningactions.removeOne(_action);
    }
    sh::base::MainThread::dispatcher()->invokeSync([=](){
        _manager->_emit_runningActionsChanged();
    });
}

QList<std::shared_ptr<sh::actions::ActionActionItem>> sh::actions::ActionsManager::runningActions()
{
    QMutexLocker _l1(&_mutex_runningactions);
    return _runningactions;
}

std::shared_ptr<sh::actions::ActionsManager::RunningActionsCounter> sh::actions::ActionsManager::_actionExecutionStarted(std::shared_ptr<sh::actions::ActionActionItem> action)
{
    return std::make_shared<sh::actions::ActionsManager::RunningActionsCounter>(this, action);
}

void sh::actions::ActionsManager::_emit_runningActionsChanged()
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    emit runningActionsChanged();
}
