/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2020 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#ifndef ACTIONSMANAGER_H
#define ACTIONSMANAGER_H

#include <functional>
#include <memory>
#include <QMutex>
#include "actions/actionfactory.h"
#include "base/singletoninitializer.h"
#include "tools/atomiccounter.h"


namespace sh{
    namespace filesystem{class FilesystemNode;}

    namespace actions{
        class AbstractActionItem;
        class ActionActionItem;
        class SubmenuActionItem;
        class AbstractActionFactory;

        /** \brief
         * Factory for actions which are valid for a given list of FileSystemNode and tools for placing actions in some gui elements.
         */
        class ActionsManager : public QObject, public sh::base::Singleton
        {
            friend class ActionActionItem;
            Q_OBJECT
            DECLARE_SINGLETON(ActionsManager)
        public:

            /** \brief
             * Asychonously gets actions for a list of sh::filesystem::FilesystemNode.
             * \param nodes The list of nodes.
             * \param callback This callback is called once the list is fetched.
             * \param initializedcallback This callback is called for each result action after initialization.
             */
            void getActions(std::shared_ptr<ActionInstantiation> instantiation, std::function<void(QList<std::shared_ptr<ActionInstantiation>>, QList<std::shared_ptr<sh::actions::ActionInstantiation>>)> callback, std::function<void(std::shared_ptr<ActionInstantiation>)> initializedcallback = [](std::shared_ptr<ActionInstantiation>){});

            /** \brief
             * Determines the default action for a given list of actions.
             * \param actionList The list of all available actions.
             * \param callback This callback is called when the default action was found (with an already initialized action).
             */
            void getDefaultAction(QList<std::shared_ptr<sh::actions::AbstractActionItem>> actionList, std::function<void(std::shared_ptr<sh::actions::ActionActionItem>)> callback);

            /** \brief
             * Asychonously gets the action for a shortcut.
             * \param selectedNodes The nodes which are currently selected in the active fileview.
             * \param folderNode The folder which is shown by the active fileview.
             * \param key The keypress character code.
             * \param modifiers The keypress modifiers.
             * \param callback This callback is called when an action was found (with an already initialized action).
             */
            void getActionForShortcut(QList<std::shared_ptr<filesystem::FilesystemNode> > selectedNodes, std::shared_ptr<filesystem::FilesystemNode> folderNode,
                                      int key, Qt::KeyboardModifiers modifiers, std::function<void(std::shared_ptr<sh::actions::AbstractActionItem>)> callback);

            /** \brief
             * Registers an action factory.
             *
             * This makes an action implementation available in the toolbar and context menus.
             */
            void registerActionFactory(std::shared_ptr<sh::actions::AbstractActionFactory> actionFactory);

            /** \brief Returns a list of all currently running actions.
             */
            QList<std::shared_ptr<sh::actions::ActionActionItem> > runningActions();

            /** \brief
             * Returns a short header label text for a node selection.
             */
            static QString selectionLabel(QList<std::shared_ptr<sh::filesystem::FilesystemNode>> selection);

            /** \brief
             * Returns a short header label text for a current directory.
             */
            static QString directoryLabel(std::shared_ptr<sh::filesystem::FilesystemNode> directory);

            /** \brief
             * Sorts a lists of actionitems and adds separators between groups. Modifies the given list.
             */
            static void sortAndSeparateActions(QList<std::shared_ptr<ActionInstantiation> > *actions);

        signals:

            /** @brief Emitted when the list of running actions changed.
             */
            void runningActionsChanged();

        private:

            ActionsManager();

            QList<std::shared_ptr<ActionInstantiation> > _getActions_raw(std::shared_ptr<ActionInstantiation> instantiation);

            void _emit_runningActionsChanged();

            /** \brief
             * Recursively initializes a list of action items and calls a callback for each action item recursively afterwards.
             * \param actions List of all action items to traverse.
             * \param callback This callback is called for each action item (even indirectly in a subtree hierarchy).
             */
            void recursivelyInitializeAction(std::shared_ptr<sh::actions::ActionInstantiation> actsit, std::function<void(std::shared_ptr<sh::actions::ActionInstantiation>)> callback, QList<std::shared_ptr<sh::actions::ActionInstantiation> > sits = {});

            /** \brief
             * Internal data structure used for _getDefaultAction_helper();
             */
            struct GDAHstruct
            {
                GDAHstruct() = default;
                GDAHstruct(const GDAHstruct& o) = default;
                sh::tools::AtomicCounter counter;
                std::function<void(std::shared_ptr<sh::actions::ActionActionItem>)> callback;
                int currentValue = 0;
                QList<std::shared_ptr<sh::actions::AbstractActionItem> > initialActionList;  // just for preserving all the parents until the bitter end
                std::shared_ptr<sh::actions::AbstractActionItem> currentAction = 0;
            };

            /** \brief
             * Internal helper for the public getDefaultAction() method.
             */
            void _getDefaultAction_helper(QList<std::shared_ptr<sh::actions::AbstractActionItem>> actionList, std::shared_ptr<GDAHstruct> gdah);

            /** @brief Helper for updating _runningactions.
             * @a_infrastructure
             */
            class RunningActionsCounter
            {
            public:
                RunningActionsCounter(ActionsManager* manager, std::shared_ptr<sh::actions::ActionActionItem> action);
                ~RunningActionsCounter();
            private:
                ActionsManager* _manager;
                std::shared_ptr<sh::actions::ActionActionItem> _action;
            };

            /** @brief Helper for updating _runningactions.
             * @a_infrastructure
             */
            std::shared_ptr<sh::actions::ActionsManager::RunningActionsCounter> _actionExecutionStarted(std::shared_ptr<sh::actions::ActionActionItem> action);

            QList<std::shared_ptr<sh::actions::AbstractActionFactory>> _actionfactories;
            QMutex _mutex_actionfactories;
            QList<std::shared_ptr<sh::actions::ActionActionItem>> _runningactions;
            QMutex _mutex_runningactions;
        };

    }
}

#endif // ACTIONSMANAGER_H
