/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2020 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <memory>
#include <QStack>
#include "actions/actionexecutioninfo.h"
#include "actions/actionexecutionuserfeedback.h"
#include "filesystem/eurl.h"
#include "filesystem/filesystemhandler.h"
#include "filesystem/filesystemhandlerregister.h"
#include "filesystem/filesystemmodel.h"
#include "filesystem/filesystemoperation.h"
#include "filesystem/operation.h"
#include "exceptions/argumentexception.h"
#include "exceptions/runtimeexception.h"
#include "base/threaddispatcher.h"
#include "actionabstracttransfertree.h"

sh::actions::common::ActionAbstractTransferTree::ActionAbstractTransferTree(QList< std::shared_ptr<const sh::filesystem::Eurl>> sources, std::shared_ptr<const sh::filesystem::Eurl> destination) :
    sh::actions::ActionActionItem(), _sources(sources), _destination(destination)
{
}

void sh::actions::common::ActionAbstractTransferTree::makereadableunits(double* value, QString* vunit)
{
    *vunit = tr("byte");
    if (*value>1024)
    {
        *vunit = tr("kilobyte");
        *value = *value / 1024.0;
        if (*value>1024)
        {
            *vunit = tr("megabyte");
            *value = *value / 1024.0;
            if (*value>1024)
            {
                *vunit = tr("gigabyte");
                *value = *value / 1024.0;
                if (*value>1024)
                {
                    *vunit = tr("terabyte");
                    *value = *value / 1024.0;
                }
            }
        }
    }
}

sh::actions::common::ActionAbstractTransferTree::MyFilesystemOperationProgressMonitor::MyFilesystemOperationProgressMonitor(ActionExecutionInfo* info)
            : sh::filesystem::FilesystemOperationProgressMonitor(info)
{
}

void sh::actions::common::ActionAbstractTransferTree::MyFilesystemOperationProgressMonitor::changed()
{
    QString estimationstring = estimation();
    if (!estimationstring.isEmpty())
        estimationstring = "- " + estimationstring;
    if (hasBytesInfo())
    {
        double donebytes = doneBytes();
        double allbytes = allBytes();
        QString doneunit;
        QString allunit;
        makereadableunits(&donebytes, &doneunit);
        makereadableunits(&allbytes, &allunit);
        _actionExecution->setProgress(doneBytes(), allBytes(), tr("%1 %2 of %3 %4 %5").arg(donebytes, 0, 'f', 2).arg(doneunit).arg(allbytes, 0, 'f', 2).arg(allunit).arg(estimationstring));
    }
    else if (hasItemInfo())
        _actionExecution->setProgress(doneItems(), allItems(), tr("%1 of %2 items %3").arg(doneItems()).arg(allItems()).arg(estimationstring));
    else
        _actionExecution->setProgressIndeterminate("");
    QString fromobj = getItemInfoFrom();
    QString toobj = getItemInfoTo();
    QString fromverb = fromobj.isEmpty() ? "" : tr("from");
    QString toverb = toobj.isEmpty() ? "" : tr("to");
    _actionExecution->setDetails(fromverb, fromobj, toverb, toobj);
}

void sh::actions::common::ActionAbstractTransferTree::MyFilesystemOperationProgressMonitor::resolveConflicts(QList<sh::filesystem::FilesystemOperationTransfers::OperationStep*> conflictsteps)
{
    sh::actions::ActionExecutionUserFeedback::GridformEntries entries;
    QMap<sh::actions::ActionExecutionUserFeedback::GridformEntry*, QMap<int, sh::filesystem::FilesystemOperationTransfers::OperationStep::ConflictResolution>> guichoices;
    foreach(sh::filesystem::FilesystemOperationTransfers::OperationStep* step, conflictsteps)
    {
        sh::actions::ActionExecutionUserFeedback::GridformEntry* entry = entries.newEntry(step->conflictDescription());
        int entryOverwrite = -1;
        int entryRenameDest = -1;
        int entryDifferentDestName = -1;
        int entryMergeDirs = -1;
        guichoices[entry][entryOverwrite= entry->addChoice(tr("Overwrite destination"))] = sh::filesystem::FilesystemOperationTransfers::OperationStep::ConflictResolution::OverwriteDestination;
        guichoices[entry][entry->addChoice(tr("Skip"))] = sh::filesystem::FilesystemOperationTransfers::OperationStep::ConflictResolution::Skip;
        guichoices[entry][entryDifferentDestName= entry->addChoice(tr("Use different name"), (step->conflictResolution_differentDestinationNameTo().isEmpty()?step->originalDestination()->basename():step->conflictResolution_differentDestinationNameTo()))] = sh::filesystem::FilesystemOperationTransfers::OperationStep::ConflictResolution::UseDifferentDestinationName;
        guichoices[entry][entryRenameDest= entry->addChoice(tr("Rename old one before copying"), (step->conflictResolution_renameDestinationBeforeTo().isEmpty()?step->originalDestination()->basename():step->conflictResolution_renameDestinationBeforeTo()))] = sh::filesystem::FilesystemOperationTransfers::OperationStep::ConflictResolution::RenameDestinationBefore;
        if (step->setConflictResolve_MergeDirectories_checkAllowed())
            guichoices[entry][entryMergeDirs= entry->addChoice(tr("Merge together"))] = sh::filesystem::FilesystemOperationTransfers::OperationStep::ConflictResolution::MergeDirectories;
        if (step->conflictResolution() == sh::filesystem::FilesystemOperationTransfers::OperationStep::ConflictResolution::RenameDestinationBefore)
            entry->choices.selected = entryRenameDest;
        else if (step->conflictResolution() == sh::filesystem::FilesystemOperationTransfers::OperationStep::ConflictResolution::OverwriteDestination)
            entry->choices.selected = entryOverwrite;
        else if (step->conflictResolution() == sh::filesystem::FilesystemOperationTransfers::OperationStep::ConflictResolution::UseDifferentDestinationName)
            entry->choices.selected = entryDifferentDestName;
        else if (step->conflictResolution() == sh::filesystem::FilesystemOperationTransfers::OperationStep::ConflictResolution::MergeDirectories)
            entry->choices.selected = entryMergeDirs;
    }
    if (!entries.list.isEmpty()) {
        int ro = _actionExecution->userfeedback()->simpleChooserGridform(tr("There are conflicts in your transfer request. Please decide how to proceed."), &entries, QStringList({"Proceed", "Cancel"}), -1, 1);
        if (ro==1)
        {
            _actionExecution->cancel();
            _actionExecution->respectCancel();
        }
    }
    foreach(sh::filesystem::FilesystemOperationTransfers::OperationStep* step, conflictsteps)
    {
        sh::actions::ActionExecutionUserFeedback::GridformEntry* entry = entries.list.takeFirst();
        sh::filesystem::FilesystemOperationTransfers::OperationStep::ConflictResolution selectedAction =
                        guichoices[entry].value(entry->choices.selected, sh::filesystem::FilesystemOperationTransfers::OperationStep::ConflictResolution::Unresolved);
        if (selectedAction == sh::filesystem::FilesystemOperationTransfers::OperationStep::ConflictResolution::OverwriteDestination)
            step->setConflictResolve_OverwriteDestination();
        else if (selectedAction == sh::filesystem::FilesystemOperationTransfers::OperationStep::ConflictResolution::Skip)
            step->setConflictResolve_Skip();
        else if (selectedAction == sh::filesystem::FilesystemOperationTransfers::OperationStep::ConflictResolution::UseDifferentDestinationName)
            step->setConflictResolve_UseDifferentDestinationName(entry->choices.list[entry->choices.selected]->text);
        else if (selectedAction == sh::filesystem::FilesystemOperationTransfers::OperationStep::ConflictResolution::RenameDestinationBefore)
            step->setConflictResolve_RenameDestinationBefore(entry->choices.list[entry->choices.selected]->text);
        else if (selectedAction == sh::filesystem::FilesystemOperationTransfers::OperationStep::ConflictResolution::MergeDirectories)
            step->setConflictResolve_MergeDirectories();
    }
}

void sh::actions::common::ActionAbstractTransferTree::MyFilesystemOperationProgressMonitor::respectCancel()
{
    _actionExecution->respectCancel();
}

sh::actions::common::ActionAbstractTransferTree::~ActionAbstractTransferTree()
{
}
