/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2020 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <QMimeData>
#include <QApplication>
#include <QClipboard>
#include "actions/actionexecutionuserfeedback.h"
#include "actions/actionexecutioninfo.h"
#include "actions/actionfactory.h"
#include "actions/actionsmanager.h"
#include "filesystem/eurl.h"
#include "filesystem/filesystemnode.h"
#include "filesystem/filesystemhandler.h"
#include "filesystem/filesystemoperation.h"
#include "exceptions/ioexception.h"
#include "tools/dataexchange.h"
#include "base/iconmanager.h"
#include "base/singletoninitializer.h"
#include "base/mainthread.h"
#include "actionclipboard.h"

REGISTER_STATICINIT(sh::actions::common, ActionClipboardCut, "base")
REGISTER_STATICINIT(sh::actions::common, ActionClipboardCopy, "base")
REGISTER_STATICINIT(sh::actions::common, ActionClipboardPaste, "base")
REGISTER_STATICINIT(sh::actions::common, ActionClipboardPasteAsLink, "base")


void sh::actions::common::ActionClipboardCut::doInitialize()
{
    sh::actions::ActionsManager::instance()->registerActionFactory(
                std::make_shared<sh::actions::ActionFactory<sh::actions::common::ActionClipboardCut>>(
                    sh::actions::ActionCategory::getByName("manage"), QList<std::shared_ptr<Predicate>>({
                    std::make_shared<sh::actions::DontResolveLinksPredicate>(),
                    std::make_shared<sh::actions::HideOnCurrentDirectoryLevelPredicate>(),
                    std::make_shared<sh::actions::KeyShortcutPredicate>(QKeySequence(Qt::CTRL + Qt::Key_X), false),
                    std::make_shared<sh::actions::PositionIndexPredicate>(6010000)
    })));
}

void sh::actions::common::ActionClipboardCut::doShutdown()
{
}

void sh::actions::common::ActionClipboardCopy::doInitialize()
{
    sh::actions::ActionsManager::instance()->registerActionFactory(
                std::make_shared<sh::actions::ActionFactory<sh::actions::common::ActionClipboardCopy>>(
                    sh::actions::ActionCategory::getByName("manage"), QList<std::shared_ptr<Predicate>>({
                    std::make_shared<sh::actions::DontResolveLinksPredicate>(),
                    std::make_shared<sh::actions::HideOnCurrentDirectoryLevelPredicate>(),
                    std::make_shared<sh::actions::KeyShortcutPredicate>(QKeySequence(Qt::CTRL + Qt::Key_C), false),
                    std::make_shared<sh::actions::PositionIndexPredicate>(6000000)
    })));
}

void sh::actions::common::ActionClipboardCopy::doShutdown()
{
}

void sh::actions::common::ActionClipboardPaste::doInitialize()
{
    sh::actions::ActionsManager::instance()->registerActionFactory(
                std::make_shared<sh::actions::ActionFactory<sh::actions::common::ActionClipboardPaste>>(
                    sh::actions::ActionCategory::getByName("create"), QList<std::shared_ptr<Predicate>>({
                    std::make_shared<sh::actions::OnDirectoriesPredicate>(),
                    std::make_shared<sh::actions::OnSingleEntrySelectionPredicate>(),
                    std::make_shared<sh::actions::HideOnSelectionLevelPredicate>(),
                    std::make_shared<sh::actions::KeyShortcutPredicate>(QKeySequence(Qt::CTRL + Qt::Key_V), true),
                    std::make_shared<sh::actions::PositionIndexPredicate>(6000000)
    })));
}

void sh::actions::common::ActionClipboardPaste::doShutdown()
{
}

void sh::actions::common::ActionClipboardPasteAsLink::doInitialize()
{
    sh::actions::ActionsManager::instance()->registerActionFactory(
                std::make_shared<sh::actions::ActionFactory<sh::actions::common::ActionClipboardPasteAsLink>>(
                    sh::actions::ActionCategory::getByName("create"), QList<std::shared_ptr<Predicate>>({
                    std::make_shared<sh::actions::OnDirectoriesPredicate>(),
                    std::make_shared<sh::actions::OnSingleEntrySelectionPredicate>(),
                    std::make_shared<sh::actions::HideOnSelectionLevelPredicate>(),
                    std::make_shared<sh::actions::PositionIndexPredicate>(6010000)
    })));
}

void sh::actions::common::ActionClipboardPasteAsLink::doShutdown()
{
}

sh::actions::common::ActionClipboardCopy::ActionClipboardCopy(QList<std::shared_ptr<filesystem::FilesystemNode> > nodes)
    :sh::actions::ActionActionItem( tr("Copy to clipboard"), true, "edit-copy")
{
    _nodes = nodes;
}

void sh::actions::common::ActionClipboardCopy::initialize()
{
    sh::filesystem::Operation op;
    foreach(std::shared_ptr<sh::filesystem::FilesystemNode> node, _nodes)
    {
        if (!node->handler() || ( node->nodetype() == sh::filesystem::FilesystemNodeType::File && !node->handler()->canGetFileContent(&op, node->eurl())))
        {
            setEnabled(false);
            break;
        }
    }
}

void sh::actions::common::ActionClipboardCut::initialize()
{
    sh::filesystem::Operation op;
    foreach(std::shared_ptr<sh::filesystem::FilesystemNode> node, _nodes)
    {
        if (!node->handler() || ( node->nodetype() == sh::filesystem::FilesystemNodeType::File && !node->handler()->canGetFileContent(&op, node->eurl())))
        {
            setEnabled(false);
            break;
        }
    }
}

void sh::actions::common::ActionClipboardCopy::action(sh::actions::ActionExecutionInfo*)
{
    sh::base::MainThread::dispatcher()->invokeSync([&](){
        QMimeData* mimeData = sh::tools::DataExchange::instance()->getMimeDataFromFilesystemNodes(_nodes, sh::tools::DataExchange::DataExchangeType::COPY);
        QApplication::clipboard()->setMimeData(mimeData); // takes ownership
    });
}

sh::actions::common::ActionClipboardCut::ActionClipboardCut(QList<std::shared_ptr<filesystem::FilesystemNode> > nodes)
    :sh::actions::ActionActionItem( tr("Cut to clipboard"), true, "edit-cut")
{
    _nodes = nodes;
}

void sh::actions::common::ActionClipboardCut::action(sh::actions::ActionExecutionInfo*)
{
    sh::base::MainThread::dispatcher()->invokeSync([&](){
        QMimeData* mimeData = sh::tools::DataExchange::instance()->getMimeDataFromFilesystemNodes(_nodes, sh::tools::DataExchange::DataExchangeType::MOVE);
        QApplication::clipboard()->setMimeData(mimeData); // takes ownership
    });
}

sh::actions::common::ActionClipboardPaste::ActionClipboardPaste(QList<std::shared_ptr<filesystem::FilesystemNode> > nodes)
    :sh::actions::ActionActionItem( tr("Paste from clipboard"), true, "edit-paste")
{
    const QMimeData* mimeData = QApplication::clipboard()->mimeData();
    _nodes = nodes;
    setEnabled( nodes.length() == 1 && sh::tools::DataExchange::instance()->containsFileEntries(mimeData) );
}

void sh::actions::common::ActionClipboardPaste::initialize()
{
    sh::filesystem::Operation op;
    foreach(std::shared_ptr<sh::filesystem::FilesystemNode> node, _nodes)
    {
        if (!node->handler() || !node->handler()->canCreateFile(&op, node->eurl()))
        {
            setEnabled(false);
            break;
        }
    }
}

void sh::actions::common::ActionClipboardPasteAsLink::initialize()
{
    sh::filesystem::Operation op;
    foreach(std::shared_ptr<sh::filesystem::FilesystemNode> node, _nodes)
    {
        if (!node->handler() || !node->handler()->canCreateFile(&op, node->eurl()))
        {
            setEnabled(false);
            break;
        }
    }
}

void sh::actions::common::ActionClipboardPaste::action(sh::actions::ActionExecutionInfo* info)
{
    std::shared_ptr<sh::actions::AbstractActionItem> action = 0;
    sh::base::MainThread::dispatcher()->invokeSync([&](){
        const QMimeData* mimeData = QApplication::clipboard()->mimeData();
        if (sh::tools::DataExchange::instance()->containsFileEntries(mimeData))
        {
            std::shared_ptr<sh::filesystem::FilesystemNode>  _node = _nodes[0];
            action = sh::tools::DataExchange::instance()->createPasteAction(mimeData, _node->eurl());
        }
    });
    if (action)
        qobject_cast<sh::actions::ActionActionItem*>(action.get())->execute(info);
}

sh::actions::common::ActionClipboardPasteAsLink::ActionClipboardPasteAsLink(QList<std::shared_ptr<filesystem::FilesystemNode> > nodes)
    :sh::actions::ActionActionItem( tr("Paste as link..."), true, "edit-paste+islink")
{
    const QMimeData* mimeData = QApplication::clipboard()->mimeData();
    _nodes = nodes;
    setEnabled( nodes.length() == 1 && sh::tools::DataExchange::instance()->containsFileEntries(mimeData) );
}

QString sh::actions::common::ActionClipboardPasteAsLink::_linktarget(std::shared_ptr<const sh::filesystem::Eurl> link, std::shared_ptr<const sh::filesystem::Eurl> destination, bool absolute)
{
    if (absolute)
        return destination->path();
    else
    {
        std::shared_ptr<const sh::filesystem::Eurl> compar1 = link->parentSegment();
        std::shared_ptr<const sh::filesystem::Eurl> compar2 = destination;
        QString addpath1;
        QString addpath2;
        while (compar1 != compar2)
        {
            if (compar1->asString().length() > compar2->asString().length())
            {
                addpath1 += "../";
                compar1 = compar1->parentSegment();
            }
            else
            {
                addpath2 = compar2->basename()+"/"+addpath2;
                compar2 = compar2->parentSegment();
            }
        }
        QString result = addpath1 + addpath2;
        while (result.endsWith("/"))
            result=result.mid(0, result.length()-1);
        if (result=="")
            result=".";
        return result;
    }
}

void sh::actions::common::ActionClipboardPasteAsLink::action(sh::actions::ActionExecutionInfo* info)
{
    bool proceed = false;
    QList<std::shared_ptr<const sh::filesystem::Eurl>> destinations;
    std::shared_ptr<sh::filesystem::FilesystemNode>  _node;
    sh::base::MainThread::dispatcher()->invokeSync([&](){
        const QMimeData* mimeData = QApplication::clipboard()->mimeData();
        if (sh::tools::DataExchange::instance()->containsFileEntries(mimeData))
        {
            info->setHead("Creating links...");
            info->setVisualProcessFeedbackActive(true);
            _node = _nodes[0];
            bool _dummy = false;
            destinations = sh::tools::DataExchange::instance()->getSources(mimeData, &_dummy);
            proceed = true;
        }
    });
    if (proceed) {
        QString text = tr("Do you want to create relative links or absolute links?\n\nUse relative links if you plan to copy the original(s) and the link(s) as one single unit. In most cases absolute links are fine.");
        QStringList answers;
        answers << tr("Create absolute links") << tr("Create relative links") << tr("Cancel");
        int mres = info->userfeedback()->messageBox(text, answers, "dialog-question", 0, 2);
        if (mres != 2)
        {
            foreach(std::shared_ptr<const sh::filesystem::Eurl> destination, destinations)
            {
                sh::exceptions::Exception::executeGuarded([&](){
                    if (destination->root() == _node->eurl()->root())
                    {
                        std::shared_ptr<const sh::filesystem::Eurl> newlink = _node->eurl()->withAppendedSegment(destination->basename());
                        QString target = _linktarget(newlink, destination, mres==0);
                        if (_node->handler()->getType(info->operation(), newlink) != sh::filesystem::FilesystemNodeType::NONE)
                            throw sh::exceptions::IOException(sh::exceptions::ExceptionData().detailsAreInteresting().details(tr("Path '%1' already exists.").arg(newlink->asString())));
                        info->operation()->filesystem()->createLink(newlink, target);
                    }
                    else
                        throw sh::exceptions::IOException(sh::exceptions::ExceptionData().detailsAreInteresting().details(tr("Creating a link across this boundary is not allowed.")).retryable(false));
                }, sh::exceptions::ExecuteGuardFlag::MANUAL_RETRY_ENABLED|sh::exceptions::ExecuteGuardFlag::RESUMEABLE_PROGRAM_ERROR_END_WITH_USER_FEEDBACK_HERE);
            }
        }
    }
}

