/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2020 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "actions/actionexecutioninfo.h"
#include "actions/actionexecutionuserfeedback.h"
#include "actions/actionfactory.h"
#include "actions/actionsmanager.h"
#include "filesystem/filesystemnode.h"
#include "filesystem/filesystemoperation.h"
#include "filesystem/filesystemhandler.h"
#include "filesystem/operation.h"
#include "base/iconmanager.h"
#include "base/singletoninitializer.h"
#include "ui/mainwindow.h"
#include "actiondeleteitems.h"

REGISTER_STATICINIT(sh::actions::common, ActionDeleteItems, "base")


void sh::actions::common::ActionDeleteItems::doInitialize()
{
    sh::actions::ActionsManager::instance()->registerActionFactory(
                std::make_shared<sh::actions::ActionFactory<sh::actions::common::ActionDeleteItems>>(
                    sh::actions::ActionCategory::getByName("manage"), QList<std::shared_ptr<Predicate>>({
                    std::make_shared<sh::actions::DontResolveLinksPredicate>(),
                    std::make_shared<sh::actions::HideOnCurrentDirectoryLevelPredicate>(),
                    std::make_shared<sh::actions::KeyShortcutPredicate>(QKeySequence(Qt::Key_Delete), false),
                    std::make_shared<sh::actions::PositionIndexPredicate>(-500000)
    })));
}

void sh::actions::common::ActionDeleteItems::doShutdown()
{
}

sh::actions::common::ActionDeleteItems::ActionDeleteItems(QList<std::shared_ptr<sh::filesystem::FilesystemNode>> nodes)
    :sh::actions::ActionActionItem( tr("Delete"), true, "edit-delete")
{
    _nodes = nodes;
}

void sh::actions::common::ActionDeleteItems::initialize()
{
    sh::filesystem::Operation op;
    foreach(std::shared_ptr<sh::filesystem::FilesystemNode> node, _nodes)
    {
        if (!node->handler() || !node->handler()->canDeleteItem(&op, node->eurl()))
        {
            setEnabled(false);
            break;
        }
    }
}

void sh::actions::common::ActionDeleteItems::action(sh::actions::ActionExecutionInfo* info)
{
    QStringList items;
    foreach(std::shared_ptr<sh::filesystem::FilesystemNode> node, _nodes)
        items << node->eurl()->asString();
    items = _stripCommonPrefix(items);
    QString qu;
    if (items.length()==1)
        qu = tr("Do you really want to delete this item?");
    else
        qu = tr("Do you really want to delete this items?");
    qu+="\n\n" + (items.join("\n"));
    int res = info->userfeedback()->simpleMessageBox(qu, (int)sh::actions::ActionExecutionUserFeedback::MessageBoxButton::Yes|(int)sh::actions::ActionExecutionUserFeedback::MessageBoxButton::No,
                                                     "dialog-question", sh::actions::ActionExecutionUserFeedback::MessageBoxButton::Yes, sh::actions::ActionExecutionUserFeedback::MessageBoxButton::No);
    if (res == sh::actions::ActionExecutionUserFeedback::MessageBoxButton::Yes)
    {
        info->setHead(tr("Deleting items..."));
        int i = 0;
        info->setVisualProcessFeedbackActive(true);
        foreach(std::shared_ptr<sh::filesystem::FilesystemNode> node, _nodes)
        {
            std::shared_ptr<const sh::filesystem::Eurl> nodeeurl = node->eurl();
            info->respectCancel();
            info->setProgress(i, _nodes.length(), tr("%1 of %2").arg(i).arg(_nodes.length()));
            info->setDetails(tr("deleting"), nodeeurl->asString(), "", "");
            info->operation()->filesystem()->deleteItem(nodeeurl);
            i++;
        }
        info->respectCancel();
    }
}

QStringList sh::actions::common::ActionDeleteItems::_stripCommonPrefix(QStringList list)
{
    QString commonprefix;
    if (list.length() <= 1) {
        return list;
    }
    commonprefix = list[0];
    for(int i=1;i<list.length();i++)
    {
        QString o = list[i];
        int prefixlen=0;
        for(int j=std::min(o.length(),commonprefix.length());j>0;j--)
        {
            if (commonprefix.mid(0,j) == o.mid(0,j))
            {
                prefixlen = j;
                break;
            }
        }
        commonprefix = commonprefix.mid(0,prefixlen);
        if (prefixlen==0)
            break;
    }
    int lastslash = commonprefix.lastIndexOf("/");
    if (lastslash>-1 && lastslash<commonprefix.length()-1)
    {
        commonprefix=commonprefix.mid(0, lastslash+1);
    }
    int prefixlen=commonprefix.length();
    if (prefixlen == 0)
    {
        return list;
    }
    else
    {
        QStringList res;
        res << commonprefix+" ...";
        foreach(QString s, list)
            res << "... " + s.mid(prefixlen);
        return res;
    }
}

