/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2020 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "actions/actionexecutioninfo.h"
#include "actions/actionexecutionuserfeedback.h"
#include "actions/actionfactory.h"
#include "actions/actionsmanager.h"
#include "filesystem/filesystemnode.h"
#include "filesystem/filesystemhandler.h"
#include "filesystem/operation.h"
#include "detailcolumns/detailcolumncustomattributes.h"
#include "filesystemhandlers/localfilesystemhandler.h"
#include "base/iconmanager.h"
#include "base/singletoninitializer.h"
#include "ui/mainwindow.h"
#include "actionexecute.h"

#ifdef Q_OS_UNIX
REGISTER_STATICINIT(sh::actions::common, ActionExecute, "base")
#include <sys/stat.h>
#include <unistd.h>
#include <sys/types.h>
#endif


void sh::actions::common::ActionExecute::doInitialize()
{
    sh::actions::ActionsManager::instance()->registerActionFactory(
                std::make_shared<sh::actions::ActionFactory<sh::actions::common::ActionExecute>>(
                    sh::actions::ActionCategory::getByName("open"), QList<std::shared_ptr<Predicate>>({
                    std::make_shared<sh::actions::OnFilesPredicate>(),
                    std::make_shared<sh::actions::OnSingleEntrySelectionPredicate>(),
                    std::make_shared<sh::actions::PositionIndexPredicate>(-520000)
    })));
}

void sh::actions::common::ActionExecute::doShutdown()
{
}

sh::actions::common::ActionExecute::ActionExecute(QList<std::shared_ptr<sh::filesystem::FilesystemNode>> nodes)
    :sh::actions::ActionActionItem( tr("Execute"), true, "media-playback-start", sh::actions::ActionDefaultPrecedenceValues::PRECEDENCE_OPENFILE+10000)
{
    _nodes = nodes;
    setVisible(false);
}

void sh::actions::common::ActionExecute::initialize()
{
    sh::filesystem::Operation op;
    if (_nodes.length() == 1) {
#ifdef Q_OS_UNIX
        static uid_t myuid = geteuid();
        static gid_t mygid = getgid();
        std::shared_ptr<sh::filesystem::FilesystemNode> node = _nodes.first();
        if (qobject_cast<sh::filesystemhandlers::LocalFilesystemHandler*>(node->handler())) {
            sh::exceptions::Exception::executeGuarded([&](){
                QVariant vv = sh::detailcolumns::DetailColumnCustomAttributes::instance()->requestValue(node, &op);
                QMap<QString, QString> vals = sh::detailcolumns::DetailColumnCustomAttributes::getMapByValue(vv);
                auto iuperms = vals.find("unix_permissions");
                if (iuperms != vals.end())
                {
                    QStringList uperms = iuperms->split(";");
                    int flags = uperms[0].toInt();
                    int uid = uperms[1].toInt();
                    int gid = uperms[2].toInt();
                    bool userMayExecute = (flags & S_IXUSR);
                    bool groupMayExecute = (flags & S_IXGRP);
                    bool othersMayExecute = (flags & S_IXOTH);
                    if (othersMayExecute || (groupMayExecute && mygid == (gid_t)gid) || (userMayExecute && myuid == (uid_t)uid))
                        setVisible(true);
                }
            }, sh::exceptions::ExecuteGuardFlag::IGNORE_ALL_RESUMEABLE_ERRORS_SILENTLY);
        }
#endif
    }
}

void sh::actions::common::ActionExecute::action(sh::actions::ActionExecutionInfo* info)
{
    info->setHead(tr("Executing..."));
    if (_nodes.length() == 1) {
        std::shared_ptr<sh::filesystem::FilesystemNode> node = _nodes.first();
        info->setDetails(tr("from"), node->eurl()->asString());
        QString execfile = qobject_cast<sh::filesystemhandlers::LocalFilesystemHandler*>(node->handler())->eurlToLocal(node->eurl());
        QString execdir = QFileInfo(execfile).path();
        QProcess::startDetached(execfile, QStringList(), execdir);
    }
}


