/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2020 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <QDesktopServices>
#include "actions/actionexecutioninfo.h"
#include "actions/actionfactory.h"
#include "actions/actionsmanager.h"
#include "filesystem/eurl.h"
#include "filesystem/filesystemnode.h"
#include "filesystem/operation.h"
#include "detailcolumns/detailcolumnmimetype.h"
#include "filesystemhandlers/localfilesystemhandler.h"
#include "settings/settingsmanager.h"
#include "tools/filetypes/filetypemanager.h"
#include "base/mainthread.h"
#include "base/singletoninitializer.h"
#include "base/threadpool.h"
#include "actionopenfile.h"
#include "actionopenfilewith.h"

REGISTER_STATICINIT(sh::actions::common, ActionOpenFile, "base")


void sh::actions::common::ActionOpenFile::doInitialize()
{
    sh::actions::ActionsManager::instance()->registerActionFactory(
                std::make_shared<sh::actions::ActionFactory<sh::actions::common::ActionOpenFile>>(
                    sh::actions::ActionCategory::getByName("open"), QList<std::shared_ptr<Predicate>>({
                    std::make_shared<sh::actions::OnFilesPredicate>(),
                    std::make_shared<sh::actions::PositionIndexPredicate>(-510000)
    })));
}

void sh::actions::common::ActionOpenFile::doShutdown()
{
}

sh::actions::common::ActionOpenFile::ActionOpenFile(QList<std::shared_ptr<sh::filesystem::FilesystemNode>> nodes)
    :sh::actions::ActionActionItem( tr("Open"), true, QString(), sh::actions::ActionDefaultPrecedenceValues::PRECEDENCE_OPENFILE)
{
    _nodes = nodes;
}

void sh::actions::common::ActionOpenFile::initialize()
{
    sh::filesystem::Operation operation;
    foreach(std::shared_ptr<sh::filesystem::FilesystemNode> node, _nodes)
    {
        if (!node->handler() || !node->handler()->canGetFileContent(&operation, node->eurl()))
        {
            setEnabled(false);
            break;
        }
    }
    _openmethod = 0;
    std::shared_ptr<sh::detailcolumns::DetailColumnMimetype> detcolMimetype = sh::detailcolumns::DetailColumnMimetype::instance();
    bool first=true;
    QString mimetype;
    foreach (std::shared_ptr<sh::filesystem::FilesystemNode> node, _nodes) {
        QString nodemimetype;
        {
            IGNORE_FORBIDDEN_ASYNC_CALLS
            nodemimetype = detcolMimetype->requestValue(node, &operation).toString();
        }
        if (first)
            mimetype = nodemimetype;
        else if (mimetype != nodemimetype) {
            mimetype = "";
            break;
        }
        first = false;
    }
    operation.abort();
    bool visible = false;
#ifdef Q_OS_WIN
    visible = true;
#endif
    if (!mimetype.isEmpty()) {
        auto openmethods = sh::tools::filetypes::FileTypeManager::instance()->getOpenMethods(mimetype, _nodes);
        if (!openmethods.isEmpty()) {
            auto rememberedActions = sh::actions::common::ActionOpenFileWith::getRememberedOpenActions(mimetype);
            _openmethod = openmethods.first();
            if (_openmethod->precedence <= 0)
                _openmethod = 0;
            if (!_openmethod) {
                foreach(auto x, rememberedActions) {
                    foreach(auto y, openmethods) {
                        if (x.command == y->command && x.arguments == y->arguments) {
                            _openmethod = y;
                            break;
                        }
                    }
                    if (_openmethod)
                        break;
                }
            }
            if (_openmethod) {
                setText(tr("Open with %1").arg(_openmethod->name));
                visible = true;
            }
        }
    }
    setEnabled(visible);
}

void sh::actions::common::ActionOpenFile::action(sh::actions::ActionExecutionInfo* info)
{
    QStringList filelist;
    if (_openmethod) {
        std::shared_ptr<sh::actions::AbstractActionItem> _a(new sh::actions::common::ActionOpenFileWith::_ActionOpenFileWithEntry(
                                                                _nodes, _openmethod->name, _openmethod->command, _openmethod->arguments, {}));
        qobject_cast<sh::actions::ActionActionItem*>(_a.get())->execute(info);
    }
    else {
        foreach(std::shared_ptr<sh::filesystem::FilesystemNode> node, _nodes){
            sh::exceptions::Exception::executeGuarded([&](){
                std::shared_ptr<const sh::filesystem::Eurl> nodeeurl = node->eurl();
                info->addChangedEurl(nodeeurl);
                filelist.append(nodeeurl->basename());
                QString localmirror = info->operation()->fetchFile(nodeeurl, nodeeurl->basename());
                QDesktopServices::openUrl(QUrl(sh::filesystemhandlers::LocalFilesystemHandler::instance()->localToEurl(localmirror)->asString()));
            }, sh::exceptions::ExecuteGuardFlag::MANUAL_RETRY_ENABLED);
        }
        sh::actions::common::ActionOpenFileWith::waitProgramClosedIfNeeded(info, filelist);
    }
}
