/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2020 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "actions/actionexecutioninfo.h"
#include "actions/separatoractionitem.h"
#include "actions/actionexecutionuserfeedback.h"
#include "actions/actionfactory.h"
#include "actions/actionsmanager.h"
#include "base/logger.h"
#include "base/mainthread.h"
#include "base/shallotprocess.h"
#include "base/singletoninitializer.h"
#include "base/threadpool.h"
#include "configuration/configurationmanager.h"
#include "detailcolumns/detailcolumnmimetype.h"
#include "filesystem/filesystemnode.h"
#include "filesystem/filesystemhandler.h"
#include "filesystem/filesystemmodel.h"
#include "filesystem/operation.h"
#include "filesystemhandlers/localfilesystemhandler.h"
#include "tools/userdirlock.h"
#include "tools/filetypes/filetypemanager.h"
#include "tools/filetypes/userdefinedopenmethoddeterminationstrategy.h"
#include "ui/mainwindow.h"
#include "ui/openwithdialog.h"
#include "ui/qt/qtmainwindow.h"
#include "actionopenfilewith.h"

#ifdef Q_OS_WIN
#define NTDDI_VERSION NTDDI_WIN7
#include <shlobj.h>
#endif

REGISTER_STATICINIT(sh::actions::common, ActionOpenFileWith, "base")


void sh::actions::common::ActionOpenFileWith::doInitialize()
{
    sh::actions::ActionsManager::instance()->registerActionFactory(
                std::make_shared<sh::actions::ActionFactory<sh::actions::common::ActionOpenFileWith>>(
                    sh::actions::ActionCategory::getByName("open"), QList<std::shared_ptr<Predicate>>({
                    std::make_shared<sh::actions::OnFilesPredicate>(),
                    std::make_shared<sh::actions::PositionIndexPredicate>(-500000)
    })));
}

void sh::actions::common::ActionOpenFileWith::doShutdown()
{
}

sh::actions::common::ActionOpenFileWith::ActionOpenFileWith(QList<std::shared_ptr<sh::filesystem::FilesystemNode>> nodes) :
    sh::actions::SubmenuActionItem(tr("Open with"), QList<std::shared_ptr<sh::actions::AbstractActionItem> >(), true, QString(), sh::actions::ActionDefaultPrecedenceValues::PRECEDENCE_OPENFILE-10000)
{
    _nodes = nodes;
}

void sh::actions::common::ActionOpenFileWith::initialize()
{
    std::shared_ptr<sh::detailcolumns::DetailColumnMimetype> detcolMimetype = sh::detailcolumns::DetailColumnMimetype::instance();
    QString mimetype;
    bool first=true;
    sh::filesystem::Operation operation;
    foreach (std::shared_ptr<sh::filesystem::FilesystemNode> node, _nodes) {
        if (!node->handler() || !node->handler()->canGetFileContent(&operation, node->eurl())) {
            setEnabled(false);
            return;
        }
    }
    foreach (std::shared_ptr<sh::filesystem::FilesystemNode> node, _nodes) {
        QString nodemimetype;
        {
            IGNORE_FORBIDDEN_ASYNC_CALLS
            nodemimetype = detcolMimetype->requestValue(node, &operation).toString();
        }
        if (first)
            mimetype = nodemimetype;
        else if (mimetype != nodemimetype) {
            mimetype = "";
            break;
        }
        first = false;
    }
    operation.abort();
    QList<std::shared_ptr<sh::actions::AbstractActionItem> > items;
    items.append(std::shared_ptr<sh::actions::AbstractActionItem>(new _ActionOpenFileWithUI(_nodes, mimetype)));
    if (!mimetype.isEmpty()) {
        items.append(std::shared_ptr<sh::actions::AbstractActionItem>(new sh::actions::SeparatorActionItem()));
        foreach (std::shared_ptr<sh::tools::filetypes::OpenMethod> method, sh::tools::filetypes::FileTypeManager::instance()->getOpenMethods(mimetype, _nodes)){
            sh::tools::filetypes::OpenMethod* m = method.get();
            items.append(std::shared_ptr<sh::actions::AbstractActionItem>(new _ActionOpenFileWithEntry(_nodes, m->name, m->command, m->arguments, {mimetype/*we are lazy, so this single mimetype is enough*/})));
        }
    }
    setSubitems(items);
}

sh::actions::common::ActionOpenFileWith::_ActionOpenFileWithEntry::_ActionOpenFileWithEntry(QList<std::shared_ptr<sh::filesystem::FilesystemNode>> nodes, QString name, QString cmd, QStringList arguments, QStringList rememberformimetype) :
    sh::actions::ActionActionItem(name, true, QString())
{
    _command = cmd;
    _nodes = nodes;
    _commandArguments = arguments;
    _rememberformimetype = rememberformimetype;
}

QString sh::actions::common::ActionOpenFileWith::_ActionOpenFileWithEntry::node2cmdfile(sh::actions::ActionExecutionInfo* info, std::shared_ptr<sh::filesystem::FilesystemNode> node, bool urlmode)
{
    sh::filesystemhandlers::LocalFilesystemHandler* localhandler = qobject_cast<sh::filesystemhandlers::LocalFilesystemHandler*>(node->handler());
    if (localhandler)
        return localhandler->eurlToLocal(node->eurl());
    else if (urlmode && !node->eurl()->hasInnerUrls() && node->handler()->isWellKnownScheme())
        return node->eurl()->asString();
    else
        return info->operation()->fetchFile(node->eurl(), node->eurl()->basename());
}

void sh::actions::common::ActionOpenFileWith::_ActionOpenFileWithEntry::action(sh::actions::ActionExecutionInfo* info)
{
    foreach (QString rmimetype, _rememberformimetype)
        rememberOpenAction(rmimetype, _command, _commandArguments);
    QStringList filelist;
    bool urlmode = (_commandArguments.contains("%U") || _commandArguments.contains("%u"));
    if (_commandArguments.contains("%U") || _commandArguments.contains("%F")) {
        // multi file mode
        QStringList files;
        QString lastfile;
        foreach (std::shared_ptr<sh::filesystem::FilesystemNode> node, _nodes) {
            std::shared_ptr<const sh::filesystem::Eurl> nodeeurl = node->eurl();
            info->addChangedEurl(nodeeurl);
            filelist.append(nodeeurl->basename());
            files << (lastfile = node2cmdfile(info, node, urlmode));
        }
        QStringList myCommandArguments;
        foreach (QString s, _commandArguments) {
            if (s == "%F" || s == "%U")
                myCommandArguments << files;
            else
                myCommandArguments << s;
        }
        QProcess::startDetached(_command, myCommandArguments, QFileInfo(lastfile).path());
    }
    else
    {
        // single file mode
        foreach (std::shared_ptr<sh::filesystem::FilesystemNode> node, _nodes) {
            std::shared_ptr<const sh::filesystem::Eurl> nodeeurl = node->eurl();
            info->addChangedEurl(nodeeurl);
            filelist.append(nodeeurl->basename());
            QString file = node2cmdfile(info, node, urlmode);
            QStringList myCommandArguments;
            foreach (QString s, _commandArguments) {
                if (s == "%f" || s == "%u")
                    myCommandArguments << file;
                else
                    myCommandArguments << s;
            }
            QProcess::startDetached(_command, myCommandArguments, QFileInfo(file).path());
        }
    }
    sh::actions::common::ActionOpenFileWith::waitProgramClosedIfNeeded(info, filelist);
}

void sh::actions::common::ActionOpenFileWith::waitProgramClosedIfNeeded(sh::actions::ActionExecutionInfo* info, QStringList filelist)
{
    if (!filelist.isEmpty() && !info->operation()->pendingCommits().isEmpty()) {
        QString sfiles;
        if (filelist.length() == 1)
            sfiles = "'" + filelist.at(0) + "'";
        else {
            QString last = filelist.takeLast();
            sfiles = "'" + filelist.join("', '") + "'";
            sfiles += " and '" + last + "'";
        }
        int result = info->userfeedback()->simpleMessageBox(tr("Please continue here when your changes in %1 are done and saved there (or if you do not want to make any changes).").arg(sfiles), (int)sh::actions::ActionExecutionUserFeedback::MessageBoxButton::Continue|(int)sh::actions::ActionExecutionUserFeedback::MessageBoxButton::Cancel, QString(),
                                                            sh::actions::ActionExecutionUserFeedback::MessageBoxButton::NONE, sh::actions::ActionExecutionUserFeedback::MessageBoxButton::Cancel);
        if (result == sh::actions::ActionExecutionUserFeedback::MessageBoxButton::Cancel)
            info->cancel();
    }
}

sh::actions::common::ActionOpenFileWith::_ActionOpenFileWithUI::_ActionOpenFileWithUI(QList<std::shared_ptr<sh::filesystem::FilesystemNode>> nodes, QString mimetype) :
    sh::actions::ActionActionItem(QObject::tr("Choose..."), true, QString(), 1),
    _nodes(nodes), _mimetype(mimetype)
{
}

void sh::actions::common::ActionOpenFileWith::_ActionOpenFileWithUI::action(sh::actions::ActionExecutionInfo* info)
{
    info->setVisualProcessFeedbackActive(false);
    QString name, command;
    QStringList arguments;
#ifndef Q_OS_WIN
    QList<std::shared_ptr<sh::tools::filetypes::OpenMethod>> openMethods = sh::tools::filetypes::FileTypeManager::instance()->getAllOpenMethods();
    std::shared_ptr<sh::ui::OpenWithDialog> dialog = sh::ui::MainWindow::mainWindow()->showOpenWithDialog(_nodes.isEmpty()?nullptr:_nodes.first(), openMethods);
    dialog->waitClosed();
    std::shared_ptr<sh::tools::filetypes::OpenMethod> chosenmethod = nullptr;
    bool rememberForFile = false;
    std::shared_ptr<const sh::filesystem::Eurl> rememberForDirectory = nullptr;
    bool rememberForMimetype = false;
    sh::base::MainThread::dispatcher()->invokeSync([&](){
        if (dialog->wasAccepted()) {
            chosenmethod = dialog->chosenMethod();
            rememberForFile = dialog->rememberForFile();
            rememberForDirectory = dialog->rememberForDirectory();
            rememberForMimetype = dialog->rememberForMimetype();
        }
    });
    if (chosenmethod) {
        name = chosenmethod->name;
        command = chosenmethod->command;
        arguments = chosenmethod->arguments;
        sh::tools::filetypes::UserDefinedOpenMethodDeterminationStrategy::instance()->storeCustomOpenMethod(_nodes, _mimetype, chosenmethod, rememberForMimetype, rememberForDirectory, rememberForFile);
    }
#else
    sh::base::MainThread::dispatcher()->invokeSync([&](){
        foreach(std::shared_ptr<sh::filesystem::FilesystemNode> node, _nodes) {
            OPENASINFO oOAI;
            sh::filesystemhandlers::LocalFilesystemHandler* lhandler = qobject_cast<sh::filesystemhandlers::LocalFilesystemHandler*>(node->handler());
            if (lhandler) {
                QString fname = lhandler->eurlToLocal(node->eurl());
                std::wstring _wsfname = fname.toStdWString();
                oOAI.pcszFile = _wsfname.c_str();
                oOAI.pcszClass = L"";
                oOAI.oaifInFlags = OAIF_EXEC;
                HWND hwnd = (HWND)(sh::ui::qt::QtMainWindow::mainWindow() ? sh::ui::qt::QtMainWindow::mainWindow()->winId() : 0);
                HRESULT r = SHOpenWithDialog(hwnd, &oOAI);
                if (r != S_OK)
                    sh::exceptions::Exception::executeGuarded([&](){
                        throw sh::exceptions::ProgramException(sh::exceptions::ExceptionData().details(QString("SHOpenWithDialog on %1 failed with %2.").arg(fname).arg(r)));
                    }, sh::exceptions::ALL_ERRORS_KILL_SHALLOT);
            }
        }
    });
#endif
    if (!command.isEmpty()) {
        std::shared_ptr<sh::actions::AbstractActionItem> _a(new _ActionOpenFileWithEntry(_nodes, name, command, arguments, {_mimetype}));
        qobject_cast<sh::actions::ActionActionItem*>(_a.get())->execute(info);
    }
}

void sh::actions::common::ActionOpenFileWith::rememberOpenAction(QString mimetype, QString cmd, QStringList arguments)
{
    QStringList s;
    s << "mimetype" << mimetype << QUrl::toPercentEncoding(cmd);
    foreach(QString arg, arguments)
        s << QUrl::toPercentEncoding(arg);
    sh::tools::UserDirLock userdirlock;
    QString cdir = sh::base::ShallotProcess::userDataDir() + "/openwith/";
    QDir dcdir(cdir);
    dcdir.mkpath(".");
    QString lastmatchingfile;
    foreach(QString chld, dcdir.entryList()) {
        if (chld.endsWith(".roa")) {
            QString ff = cdir + chld;
            QString _mimetype;
            QString _cmd;
            QStringList _arguments;
            readOpenAction(ff, &_mimetype, &_cmd, &_arguments);
            if (_mimetype == mimetype && !lastmatchingfile.isEmpty()) {
                QFile(lastmatchingfile).remove();
                lastmatchingfile = QString();
            }
            if (_mimetype == mimetype && _cmd == cmd && _arguments == arguments)
                lastmatchingfile = ff;
        }
    }
    if (!lastmatchingfile.isEmpty())
        return;
    QFile fnextid(cdir + "/nextid");
    fnextid.open(QFile::ReadOnly);
    int myidx = QString::fromUtf8(fnextid.readAll()).toInt();
    fnextid.close();
    fnextid.open(QFile::WriteOnly);
    QByteArray bnextid = QString::number(myidx + 1).toLocal8Bit();
    fnextid.write(bnextid);
    fnextid.close();
    QString newname = QString::number(myidx).rightJustified(10, '0') + ".roa";
    QFile f(cdir+newname);
    f.open(QFile::WriteOnly);
    f.write(s.join("\n").toLocal8Bit());
}

void sh::actions::common::ActionOpenFileWith::readOpenAction(QString path, QString* mimetype, QString* cmd, QStringList* arguments)
{
    QFile fcnt(path);
    fcnt.open(QFile::ReadOnly);
    QByteArray cnt = fcnt.readAll();
    QStringList scnt = QString::fromUtf8(cnt).split("\n");
    if (scnt[0] == "mimetype") {  // nothing else is currently supported
        *mimetype = scnt[1];
        *cmd = QUrl::fromPercentEncoding(scnt[2].toLocal8Bit());
        for (int i=3; i<scnt.length(); i++)
            arguments->append(QUrl::fromPercentEncoding(scnt[i].toLocal8Bit()));
    }
}

QList<sh::actions::common::ActionOpenFileWith::RememberedOpenAction> sh::actions::common::ActionOpenFileWith::getRememberedOpenActions(QString mimetype)
{
    QList<sh::actions::common::ActionOpenFileWith::RememberedOpenAction> result;
    sh::tools::UserDirLock userdirlock;
    QString cdir = sh::base::ShallotProcess::userDataDir() + "/openwith/";
    QDir dcdir(cdir);
    QStringList ents = dcdir.entryList();
    qSort(ents.begin(), ents.end(), [](QString a, QString b){return b<a;});
    foreach (QString chld, ents) {
        if (chld.endsWith(".roa")) {
            QString ff = cdir + chld;
            QString _mimetype;
            QString _cmd;
            QStringList _arguments;
            readOpenAction(ff, &_mimetype, &_cmd, &_arguments);
            if (_mimetype == mimetype)
                result.append(sh::actions::common::ActionOpenFileWith::RememberedOpenAction(_cmd, _arguments));
        }
    }
    return result;
}

sh::actions::common::ActionOpenFileWith::RememberedOpenAction::RememberedOpenAction(QString command, QStringList arguments)
                    : command(command), arguments(arguments)
{
}

bool sh::actions::common::ActionOpenFileWith::RememberedOpenAction::operator==(const RememberedOpenAction& a2) const
{
    return (command == a2.command) && (arguments == a2.arguments);
}
