/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2020 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "actions/actionexecutioninfo.h"
#include "actions/actionexecutionuserfeedback.h"
#include "actions/actionfactory.h"
#include "actions/actionsmanager.h"
#include "filesystem/eurl.h"
#include "filesystem/operation.h"
#include "filesystem/filesystemoperation.h"
#include "filesystem/filesystemhandler.h"
#include "filesystem/filesystemnode.h"
#include "exceptions/permissiondeniedexception.h"
#include "base/mainthread.h"
#include "base/singletoninitializer.h"
#include "ui/mainwindow.h"
#include "actionrenameitem.h"

REGISTER_STATICINIT(sh::actions::common, ActionRenameItem, "base")


void sh::actions::common::ActionRenameItem::doInitialize()
{
    sh::actions::ActionsManager::instance()->registerActionFactory(
                std::make_shared<sh::actions::ActionFactory<sh::actions::common::ActionRenameItem>>(
                    sh::actions::ActionCategory::getByName("manage"), QList<std::shared_ptr<Predicate>>({
                    std::make_shared<sh::actions::DontResolveLinksPredicate>(),
                    std::make_shared<sh::actions::HideOnCurrentDirectoryLevelPredicate>(),
                    std::make_shared<sh::actions::KeyShortcutPredicate>(QKeySequence("F2"), false),
                    std::make_shared<sh::actions::PositionIndexPredicate>(-300000)
    })));
}

void sh::actions::common::ActionRenameItem::doShutdown()
{
}

sh::actions::common::ActionRenameItem::ActionRenameItem(QList<std::shared_ptr<sh::filesystem::FilesystemNode>> nodes)
    :sh::actions::ActionActionItem(tr("Rename..."))
{
    _nodes = nodes;
}

void sh::actions::common::ActionRenameItem::initialize()
{
    sh::filesystem::Operation op;
    foreach(std::shared_ptr<sh::filesystem::FilesystemNode> node, _nodes)
    {
        if (!op.filesystem()->canMoveItem(node->eurl()))
        {
            setEnabled(false);
            break;
        }
    }
}

void sh::actions::common::ActionRenameItem::action(sh::actions::ActionExecutionInfo* info)
{
    int i = 0;
    info->setHead(tr("Renaming items..."));
    bool dobreak = false;
    foreach(std::shared_ptr<sh::filesystem::FilesystemNode> node, _nodes)
    {
        info->setProgress(i, _nodes.length(), tr("%1 of %2").arg(i).arg(_nodes.length()));
        i++;
        sh::exceptions::Exception::executeGuarded([&](){
            QString oldname;
            sh::base::MainThread::dispatcher()->invokeSync([&]() {
                oldname = node->eurl()->basename();
            });
            int ild = oldname.lastIndexOf('.');
            int lwoe = (ild > -1) ? ild : oldname.length();
            QString newname = info->userfeedback()->simpleInputBox(tr("Please enter the new name."), oldname, 0, lwoe);
            if (!newname.isNull())
            {
                info->setVisualProcessFeedbackActive(true);
                if (newname != oldname) {
                    sh::filesystem::Eurl::filenameCheck(newname);
                    std::shared_ptr<const sh::filesystem::Eurl> nodeeurl = node->eurl();
                    std::shared_ptr<const sh::filesystem::Eurl> desteurl = nodeeurl->parentSegment()->withAppendedSegment(newname);
                    if (node->handler()->getType(info->operation(), desteurl) != sh::filesystem::FilesystemNodeType::NONE)
                        throw sh::exceptions::IOException(sh::exceptions::ExceptionData().detailsAreInteresting().details(tr("There already is an item with that name.")));
                    info->setDetails(tr("renaming"), nodeeurl->asString(), tr("to"), desteurl->asString());
                    info->operation()->filesystem()->moveItem(nodeeurl, desteurl, std::make_shared<sh::filesystem::FilesystemOperationProgressMonitor>(info));
                }
            }
            else
                dobreak = true;
        }, sh::exceptions::ExecuteGuardFlag::MANUAL_RETRY_ENABLED|sh::exceptions::ExecuteGuardFlag::RESUMEABLE_PROGRAM_ERROR_END_WITH_USER_FEEDBACK_HERE);
        if (dobreak)
            break;
    }
}
