/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2020 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "actions/actionexecutioninfo.h"
#include "actions/actionfactory.h"
#include "actions/actionsmanager.h"
#include "actions/separatoractionitem.h"
#include "actions/common/actionabstracttransfertree.h"
#include "filesystem/filesystemnode.h"
#include "filesystem/filesystemoperation.h"
#include "tools/bookmarkmanager.h"
#include "base/mainthread.h"
#include "base/singletoninitializer.h"
#include "ui/fileview.h"
#include "ui/mainwindow.h"
#include "actiontransferto.h"

REGISTER_STATICINIT(sh::actions::common, ActionTransferToCopy, "base")
REGISTER_STATICINIT(sh::actions::common, ActionTransferToMove, "base")


sh::actions::common::ActionAbstractTransferTo::ActionAbstractTransferTo(QString text, QString icon, QList<std::shared_ptr<sh::filesystem::FilesystemNode>> nodes)
    : SubmenuActionItem(text, {}, true, icon)
{
    _nodes = nodes;
}

void sh::actions::common::ActionAbstractTransferTo::initialize()
{
    QList<std::shared_ptr<sh::actions::AbstractActionItem>> subitems;
    // other file views
    auto mainwnd = sh::ui::MainWindow::mainWindow();
    sh::base::MainThread::dispatcher()->invokeSync([&](){
        int fvcnt = mainwnd->fileViewCount();
        int imyfv = mainwnd->currentFileView()->index();
        for(int ifv=0; ifv<fvcnt; ifv++) {
            if (ifv != imyfv) {
                auto neurl = mainwnd->getFileView(ifv)->node()->eurl();
                QString atext = neurl->asString();
                if (atext.length() > 33)
                    atext = "⋯" + atext.right(32);
                atext = tr("View #%1: %2").arg(ifv+1).arg(atext);
                subitems.append(std::make_shared<ActionTransferToHelper>(atext, "folder", neurl, std::dynamic_pointer_cast<ActionAbstractTransferTo>(shared_from_this())));
            }
        }
    });
    // bookmarks
    bool needsSeparator = !subitems.isEmpty();
    if (sh::tools::BookmarkManager::instance()->hasBookmarks())
    {
        foreach(std::shared_ptr<sh::tools::Bookmark> bookmark, sh::tools::BookmarkManager::instance()->getBookmarks())
        {
            if (bookmark->folder().isEmpty()) {
                if (needsSeparator) {
                    subitems.append(std::make_shared<sh::actions::SeparatorActionItem>());
                    needsSeparator = false;
                }
                subitems.append(std::make_shared<ActionTransferToHelper>(bookmark->label(), "folder", bookmark->eurl(), std::dynamic_pointer_cast<ActionAbstractTransferTo>(shared_from_this())));
            }
        }
    }
    if (subitems.isEmpty())
        setVisible(false);
    else
        setSubitems(subitems);
}

sh::actions::common::ActionTransferToHelper::ActionTransferToHelper(QString text, QString icon, std::shared_ptr<const sh::filesystem::Eurl> transferto, std::shared_ptr<ActionAbstractTransferTo> root)
            : ActionActionItem(text, true, icon), _transferto(transferto), _root(root)
{
}

void sh::actions::common::ActionTransferToHelper::action(sh::actions::ActionExecutionInfo* info)
{
    _root->doTransfer(info, _transferto);
}

sh::actions::common::ActionTransferToCopy::ActionTransferToCopy(QList<std::shared_ptr<sh::filesystem::FilesystemNode>> nodes)
            : ActionAbstractTransferTo(tr("Copy to"), "arrow-to-right+list-add", nodes)
{
}

void sh::actions::common::ActionTransferToCopy::doInitialize()
{
    sh::actions::ActionsManager::instance()->registerActionFactory(
                std::make_shared<sh::actions::ActionFactory<sh::actions::common::ActionTransferToCopy>>(
                    sh::actions::ActionCategory::getByName("manage"), QList<std::shared_ptr<Predicate>>({
                    std::make_shared<sh::actions::DontResolveLinksPredicate>(),
                    std::make_shared<sh::actions::HideOnCurrentDirectoryLevelPredicate>(),
                    std::make_shared<sh::actions::PositionIndexPredicate>(7010000),
    })));
}

void sh::actions::common::ActionTransferToCopy::doShutdown()
{
}

void sh::actions::common::ActionTransferToCopy::doTransfer(ActionExecutionInfo *info, std::shared_ptr<const sh::filesystem::Eurl> to)
{
    info->setHead(tr("Copying..."));
    info->setVisualProcessFeedbackActive(true);
    QList<std::shared_ptr<const sh::filesystem::Eurl> > lfrom;
    foreach(auto n, _nodes)
        lfrom.append(n->eurl());
    info->operation()->filesystem()->copyItems(lfrom, to, std::make_shared<sh::actions::common::ActionAbstractTransferTree::MyFilesystemOperationProgressMonitor>(info));
}

sh::actions::common::ActionTransferToMove::ActionTransferToMove(QList<std::shared_ptr<sh::filesystem::FilesystemNode>> nodes)
            : ActionAbstractTransferTo(tr("Move to"), "arrow-to-right", nodes)
{
}

void sh::actions::common::ActionTransferToMove::doInitialize()
{
    sh::actions::ActionsManager::instance()->registerActionFactory(
                std::make_shared<sh::actions::ActionFactory<sh::actions::common::ActionTransferToMove>>(
                    sh::actions::ActionCategory::getByName("manage"), QList<std::shared_ptr<Predicate>>({
                    std::make_shared<sh::actions::DontResolveLinksPredicate>(),
                    std::make_shared<sh::actions::HideOnCurrentDirectoryLevelPredicate>(),
                    std::make_shared<sh::actions::PositionIndexPredicate>(7010100),
    })));
}

void sh::actions::common::ActionTransferToMove::doShutdown()
{
}

void sh::actions::common::ActionTransferToMove::doTransfer(sh::actions::ActionExecutionInfo* info, std::shared_ptr<const sh::filesystem::Eurl> to)
{
    info->setHead(tr("Moving..."));
    info->setVisualProcessFeedbackActive(true);
    QList<std::shared_ptr<const sh::filesystem::Eurl> > lfrom;
    foreach(auto n, _nodes)
        lfrom.append(n->eurl());
    info->operation()->filesystem()->moveItems(lfrom, to, std::make_shared<sh::actions::common::ActionAbstractTransferTree::MyFilesystemOperationProgressMonitor>(info));
}
