/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2020 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "actions/actionexecutioninfo.h"
#include "actions/headeractionitem.h"
#include "actions/separatoractionitem.h"
#include "actions/actionexecutionuserfeedback.h"
#include "actions/mainmenu/actionabout.h"
#include "actions/mainmenu/actionapplyprofile.h"
#include "actions/mainmenu/actionhelp.h"
#include "actions/mainmenu/actionmanagesavedsettings.h"
#include "actions/mainmenu/actionquit.h"
#include "actions/mainmenu/actionrefresh.h"
#include "actions/mainmenu/actionsavesettings.h"
#include "actions/mainmenu/actionsetwindowtitlepattern.h"
#include "actions/mainmenu/actionshowlog.h"
#include "actions/mainmenu/actiontuning.h"
#include "actions/mainmenu/actionsearch.h"
#include "actions/mainmenu/actionfileviewchangeicondimension.h"
#include "actions/mainmenu/actionfileviewfilesizeformatting.h"
#include "actions/mainmenu/actionfileviewshowhiddenfiles.h"
#include "actions/mainmenu/actionfileviewview.h"
#include "actions/mainmenu/actionnumberfileviews.h"
#include "actions/mainmenu/actionshowfoldertree.h"
#include "actions/mainmenu/actiontreestickyness.h"
#include "actions/mainmenu/actionshowdetailpanel.h"
#include "actions/mainmenu/actionselect.h"
#include "actions/mainmenu/actiongotodirectory.h"
#include "filesystem/eurl.h"
#include "filesystem/operation.h"
#include "filesystem/filesystemhandler.h"
#include "filesystem/filesystemnode.h"
#include "exceptions/permissiondeniedexception.h"
#include "base/mainthread.h"
#include "base/threadpool.h"
#include "ui/mainwindow.h"
#include "actionmain.h"

REGISTER_SINGLETON(sh::actions::mainmenu, ActionMain, "base")


sh::actions::mainmenu::ActionMain::ActionMain()
    : sh::actions::SubmenuActionItem("&S", QList<std::shared_ptr<sh::actions::AbstractActionItem>>(), true, "configure")
{
}

void sh::actions::mainmenu::ActionMain::initialize()
{
    QList<std::shared_ptr<sh::actions::AbstractActionItem>> subitems;
    sh::base::MainThread::dispatcher()->invokeSync([&](){
        subitems.append(std::shared_ptr<sh::actions::AbstractActionItem>(new sh::actions::HeaderActionItem( tr("View"), "view" )));
        subitems.append(std::shared_ptr<sh::actions::AbstractActionItem>(new sh::actions::mainmenu::ActionNumberFileviews()));
        subitems.append(std::shared_ptr<sh::actions::AbstractActionItem>(new sh::actions::SeparatorActionItem()));
        subitems.append(std::shared_ptr<sh::actions::AbstractActionItem>(new sh::actions::mainmenu::ActionFileviewShowHiddenFiles()));
        subitems.append(std::shared_ptr<sh::actions::AbstractActionItem>(new sh::actions::mainmenu::ActionFileviewView()));
        subitems.append(std::shared_ptr<sh::actions::AbstractActionItem>(new sh::actions::mainmenu::ActionFileViewIncreaseIconDimension()));
        subitems.append(std::shared_ptr<sh::actions::AbstractActionItem>(new sh::actions::mainmenu::ActionFileViewDecreaseIconDimension()));
        subitems.append(std::shared_ptr<sh::actions::AbstractActionItem>(new sh::actions::mainmenu::ActionRefresh()));
        subitems.append(std::shared_ptr<sh::actions::AbstractActionItem>(new sh::actions::SeparatorActionItem()));
        auto viewMore = std::make_shared<sh::actions::SubmenuActionItem>(tr("More"), QList<std::shared_ptr<sh::actions::AbstractActionItem> >{
            std::shared_ptr<sh::actions::AbstractActionItem>(new sh::actions::mainmenu::ActionShowFolderTree()),
            std::shared_ptr<sh::actions::AbstractActionItem>(new sh::actions::mainmenu::ActionTreeStickyness()),
            std::shared_ptr<sh::actions::AbstractActionItem>(new sh::actions::mainmenu::ActionShowDetailPanel()),
            std::shared_ptr<sh::actions::AbstractActionItem>(new sh::actions::mainmenu::ActionFileviewFilesizeFormatting()),
            std::shared_ptr<sh::actions::AbstractActionItem>(new sh::actions::mainmenu::ActionSetWindowTitlePattern()),
        });
        subitems.append(viewMore);
        subitems.append(std::shared_ptr<sh::actions::AbstractActionItem>(new sh::actions::HeaderActionItem( tr("Selection"), "tool-pointer" )));
        subitems.append(std::shared_ptr<sh::actions::AbstractActionItem>(new sh::actions::mainmenu::ActionSelectAllNodes()));
        subitems.append(std::shared_ptr<sh::actions::AbstractActionItem>(new sh::actions::mainmenu::ActionInvertNodeSelection()));
        subitems.append(std::shared_ptr<sh::actions::AbstractActionItem>(new sh::actions::HeaderActionItem( tr("Settings"), "format-node-line" )));
        subitems.append(std::shared_ptr<sh::actions::AbstractActionItem>(new sh::actions::mainmenu::ActionSaveSettings()));
        subitems.append(std::shared_ptr<sh::actions::AbstractActionItem>(new sh::actions::mainmenu::ActionManageSavedSettings()));
        subitems.append(std::shared_ptr<sh::actions::AbstractActionItem>(new sh::actions::SeparatorActionItem()));
        subitems.append(std::shared_ptr<sh::actions::AbstractActionItem>(new sh::actions::mainmenu::ActionApplyProfile()));
        subitems.append(std::shared_ptr<sh::actions::AbstractActionItem>(new sh::actions::HeaderActionItem( tr("Main"), "run-build" )));
        subitems.append(std::shared_ptr<sh::actions::AbstractActionItem>(new sh::actions::mainmenu::ActionSearch()));
        subitems.append(std::shared_ptr<sh::actions::AbstractActionItem>(new sh::actions::mainmenu::ActionGotoDirectory()));
        subitems.append(std::shared_ptr<sh::actions::AbstractActionItem>(new sh::actions::SeparatorActionItem()));
        auto mainMore = std::make_shared<sh::actions::SubmenuActionItem>(tr("More"), QList<std::shared_ptr<sh::actions::AbstractActionItem> >{
            std::shared_ptr<sh::actions::AbstractActionItem>(new sh::actions::mainmenu::ActionTuning()),
            std::shared_ptr<sh::actions::AbstractActionItem>(new sh::actions::mainmenu::ActionShowLog()),
            std::shared_ptr<sh::actions::AbstractActionItem>(new sh::actions::mainmenu::ActionAbout()),
        });
        subitems.append(mainMore);
        subitems.append(std::shared_ptr<sh::actions::AbstractActionItem>(new sh::actions::SeparatorActionItem()));
        subitems.append(std::shared_ptr<sh::actions::AbstractActionItem>(new sh::actions::mainmenu::ActionHelp()));
        if (sh::ui::MainWindow::uiMode() != "web")
            subitems.append(std::shared_ptr<sh::actions::AbstractActionItem>(new sh::actions::mainmenu::ActionQuit()));
    });
    setSubitems(subitems);
}
