/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2020 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <QPainter>
#include "base/mainthread.h"
#include "configuration/configurationmanager.h"
#include "iconmanager.h"
#include "shallotprocess.h"

REGISTER_SINGLETON(sh::base, IconManager, "early")


sh::base::IconManager::IconManager()
{
    cfgvalPreferredStrategy = sh::configuration::ConfigurationManager::instance()->registerConfigValue(
                "sh::base::IconManager.PreferredStrategy", "builtin",
                sh::configuration::ConfigurationValue::valueTypeString(), QObject::tr("The preferred icon source"),
                sh::configuration::ConfigurationCategory::CategoryGUI, QString(), QObject::tr("'builtin' or 'system'."));
    if (cfgvalPreferredStrategy->getConfigValueString() == "system") {
        getIconStrategies.append(new QIconFromThemeGetIconStrategy());
        getIconStrategies.append(new IncludedGetIconStrategy());
    }
    else {
        getIconStrategies.append(new IncludedGetIconStrategy());
        getIconStrategies.append(new QIconFromThemeGetIconStrategy());
    }
}

sh::base::IconManager::~IconManager()
{
    foreach(GetIconStrategy* gis, getIconStrategies)
        delete gis;
}

QIcon sh::base::IconManager::_getIcon(QString name)
{
    QIcon result;
    foreach (GetIconStrategy* strat, getIconStrategies) {
        result = strat->getIcon(name);
        if (!result.isNull())
            break;
    }
    if (result.isNull())
    {
        QPixmap emp(32,32);
        emp.fill(Qt::transparent);
        result.addPixmap(emp);
    }
    return result;
}

QImage sh::base::IconManager::_colourImage(QImage img, QColor color)
{
    QSize size = img.size();
    QImage dimg = img;
    QPainter pemb(&dimg);
    pemb.fillRect(0, 0, size.width(), size.height(), color);
    pemb.setCompositionMode(QPainter::CompositionMode_DestinationIn);
    pemb.drawImage(0, 0, img);
    return dimg;
}

QIcon sh::base::IconManager::getIcon(QIcon mainicon, QString emblemname, QString miniemblemname, QStringList)
{
    QIcon _result;
    QIcon emblemicon;
    QIcon miniemblemicon;
    if (!emblemname.isEmpty())
        emblemicon = _getIcon(emblemname);
    if (!miniemblemname.isEmpty())
        miniemblemicon = _getIcon(miniemblemname);
    for(const QIcon::Mode mode : {QIcon::Normal, QIcon::Disabled, QIcon::Active, QIcon::Selected})
    {
        for(const QIcon::State state : {QIcon::Off, QIcon::On})
        {
            int embfactor = (miniemblemicon.isNull())? 75 : 70;
            int miniembfactor = 40;
            QList<QSize> _sizes;
            foreach(QSize size, mainicon.availableSizes(mode, state))
                if (!_sizes.contains(size))
                    _sizes.append(size);
            if (!emblemicon.isNull())
                foreach(QSize size, emblemicon.availableSizes(mode, state))
                {
                    QSize effsize(size.width()*100/embfactor, size.height()*100/embfactor);
                    if (!_sizes.contains(effsize))
                        _sizes.append(effsize);
                }
            if (!miniemblemicon.isNull())
                foreach(QSize size, miniemblemicon.availableSizes(mode, state))
                {
                    QSize effsize(size.width()*100/miniembfactor, size.height()*100/miniembfactor);
                    if (!_sizes.contains(effsize))
                        _sizes.append(effsize);
                }
            foreach(QSize size, _sizes)
            {
                QImage mainimg = mainicon.pixmap(size, mode, state).toImage();
                QImage resimg(size, QImage::Format_ARGB32_Premultiplied);
                QPainter painter(&resimg);
                painter.setCompositionMode(QPainter::CompositionMode_Source);
                painter.fillRect(resimg.rect(), Qt::transparent);
                painter.setCompositionMode(QPainter::CompositionMode_SourceOver);
                painter.drawImage(0, 0, mainimg);
                painter.setCompositionMode(QPainter::CompositionMode_SourceOver);
                if (!emblemicon.isNull()) {
                    QImage emblemimg = _colourImage(
                            emblemicon.pixmap(size.width()*embfactor/100, size.height()*embfactor/100, mode, state).toImage(),
                            QColor(0, 20, 20, 90));
                    painter.drawImage(size.width()*(100-embfactor)/100, size.height()*(100-embfactor)/100, emblemimg);
                }
                if (!miniemblemicon.isNull()) {
                    QImage emblemimg = _colourImage(
                            miniemblemicon.pixmap(size.width()*miniembfactor/100, size.height()*miniembfactor/100, mode, state).toImage(),
                            QColor(0, 20, 20, 90));
                    painter.drawImage(0, size.height()*(100-miniembfactor)/100, emblemimg);
                }
                painter.end();
                QPixmap pmres = QPixmap::fromImage(resimg);
                _result.addPixmap(pmres, mode, state);
            }
        }
    }
    return _result;
}

QIcon sh::base::IconManager::getIcon(QString mainname, QString emblemname, QString miniemblemname, QStringList tags)
{
    QMutexLocker _l(&cachemutex);
    QString name = mainname;
    if (!emblemname.isEmpty())
        name += "+" + emblemname;
    if (!miniemblemname.isEmpty())
        name += "*" + miniemblemname;
    QIcon result;
    if (cache.contains(name))
        result = cache[name];
    else
    {
        result = _getIcon(mainname);
        if (!emblemname.isEmpty() || !miniemblemname.isEmpty())
            result = getIcon(result, emblemname, miniemblemname, tags);
        cache.insert(name, result);
    }
    return result;
}

QIcon sh::base::IconManager::getIconByFullname(QString fullname)
{
    {
        QMutexLocker _l(&cachemutex);
        auto itr = cache.constFind(fullname);
        if (itr != cache.cend())
            return *itr;
    }
    QString emblemname;
    QString miniemblemname;
    QStringList tags;
    tags = fullname.split("#");
    fullname = tags.takeFirst();
    int i = fullname.indexOf("*");
    if (i > -1)
    {
        miniemblemname = fullname.mid(i+1);
        fullname = fullname.mid(0,i);
    }
    int j = fullname.indexOf("+");
    if (j > -1)
    {
        emblemname = fullname.mid(j+1);
        fullname = fullname.mid(0,j);
    }
    return getIcon(fullname, emblemname, miniemblemname, tags);
}

QPixmap sh::base::IconManager::getPixmap(QString name, int size)
{
    return getIcon(name).pixmap(size, size);
}

sh::base::IconManager::QIconFromThemeGetIconStrategy::QIconFromThemeGetIconStrategy()
                : sh::base::IconManager::GetIconStrategy()
{
    _aliases["adjustlevels"] = "finish";
    _aliases["arrow-up"] = "go-up";
    _aliases["arrow-down"] = "go-down";
    _aliases["arrow-left"] = "go-previous";
    _aliases["arrow-right"] = "go-next";
    _aliases["arrow-right-double"] = "player_fwd";
    _aliases["arrow-left-double"] = "player_rew";
    _aliases["bookmarks-organize"] = "user-bookmarks";
    _aliases["cell_layout"] = "window_nofullscreen";
    _aliases["configure"] = "administration";
    _aliases["edit-clear-list"] = "edit-clear";
    _aliases["error-critical"] = "dialog-error";
    _aliases["error-normal"] = "dialog-error";
    _aliases["error-retryable"] = "dialog-warning";
    _aliases["folder-bookmark"] = "user-bookmarks";
    _aliases["format-insert-node"] = "tag-new";
    _aliases["format-join-node"] = "object-inverse";
    _aliases["format-node-line"] = "menu-editor";
    _aliases["format-remove-node"] = "remove";
    _aliases["offline"] = "network-offline";
    _aliases["online"] = "network-idle";
    _aliases["plugins"] = "extension";
    _aliases["rating"] = "user-bookmarks";
    _aliases["rating-unrated"] = "bookmark-missing";
    _aliases["tool-pointer"] = "mouse";
    _aliases["transform-move"] = "object-rotate-left";
    _aliases["view"] = "cell_layout";
    _aliases["circular-ligh-bulb"] = "games-hint";
    _aliases["pushpin"] = "tool-pointer";
    _aliases["small-pin-shape"] = "tool-pointer";
    _aliases["task-delegate"] = "media-repeat-all";
}

QIcon sh::base::IconManager::QIconFromThemeGetIconStrategy::getIcon(QString name)
{
    QIcon result = QIcon::fromTheme(name);
    if (result.isNull()) {
        auto ialias = _aliases.find(name);
        if (ialias != _aliases.end())
            result = getIcon(*ialias);
    }
    return result;
}

sh::base::IconManager::IncludedGetIconStrategy::IncludedGetIconStrategy()
                : sh::base::IconManager::GetIconStrategy()
{
    QColor c = sh::base::ShallotProcess::instance()->brandingColor();
    brandingcolor1 = QColor::fromHsl(c.hue(), c.saturation()/8, 160);
    brandingcolor2 = QColor::fromHsl(c.hue(), c.saturation(), 60);
}

QIcon sh::base::IconManager::IncludedGetIconStrategy::getIcon(QString name)
{
    static QString iconpath = sh::base::ShallotProcess::shallotDataDir() + "/icons";
    QIcon result;
    QDirIterator it(iconpath, QStringList({name+".*png"}));
    while (it.hasNext()) {
        QImage img(it.next());
        img = img.convertToFormat(QImage::Format_ARGB32);
        if (name != "islink") {
            QImage white(img);
            white.invertPixels();
            QImage front(img.size(), img.format());
            front.fill(Qt::transparent);
            QPainter pfront(&front);
            pfront.fillRect(front.rect(), brandingcolor1);
            pfront.setCompositionMode(QPainter::CompositionMode_DestinationIn);
            pfront.drawImage(0, 0, white);
            int ipadd = img.width() * 0.05;
            int opadd = img.width() * 0.15;
            QImage back(img.width()+2*opadd+2*ipadd, img.height()+2*opadd+2*ipadd, img.format());
            back.fill(Qt::transparent);
            {
                QPainter pback(&back);
                pback.fillRect(opadd, opadd, white.width()+2*ipadd, white.height()+2*ipadd, brandingcolor2);
                pback.setCompositionMode(QPainter::CompositionMode_DestinationIn);
                pback.drawImage(QRect(opadd, opadd, white.width()+2*ipadd, white.height()+2*ipadd), white, white.rect());
            }
            QImage backblurred(back.size(), back.format());
            backblurred.fill(Qt::transparent);
            QPainter pbackblurred(&backblurred);
            qt_blurImage(&pbackblurred, back, opadd-1, true, false);
            QImage bbresult(backblurred);
            QPainter presult(&bbresult);
            presult.drawImage(opadd+ipadd, opadd+ipadd, front);
            img = bbresult;
        }
        result.addPixmap(QPixmap::fromImage(img));
    }
    return result;
}
