/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2020 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef ICONMANAGER_H
#define ICONMANAGER_H

#include <memory>
#include <QApplication>
#include <QIcon>
#include <QMutex>
#include <QObject>
#include <QPainter>
#include <QPixmap>
#include "singletoninitializer.h"


namespace sh{
    namespace configuration {class ConfigurationValue;}

    namespace base{

        /** \brief
         * Fetches icons according to theme availability and settings.
         */
        class IconManager : public QObject, public Singleton
        {
            Q_OBJECT
            DECLARE_SINGLETON(IconManager)
        public:

            /** \brief
             * Creates a QIcon from the icon name(s). Uses a cache.
             */
            QIcon getIcon(QString mainname, QString emblemname = QString(), QString miniemblemname = QString(), QStringList tags = QStringList());

            /** \brief
             * Creates a QIcon from the icon fullname. Uses a cache.
             */
            QIcon getIconByFullname(QString fullname);

            /** \brief
             * Creates a QIcon based on an existing one. Uncached.
             */
            QIcon getIcon(QIcon mainicon, QString emblemname, QString miniemblemname, QStringList tags);

            /** \brief
             * Creates a QPixmap for an icon name. Uncached.
             */
            QPixmap getPixmap(QString name, int size = 22);

            virtual ~IconManager();

        private:

            class GetIconStrategy
            {
            public:
                virtual QIcon getIcon(QString name) = 0;
                virtual ~GetIconStrategy() {}
            };

            class QIconFromThemeGetIconStrategy : public GetIconStrategy
            {
            public:
                QIconFromThemeGetIconStrategy();
                QIcon getIcon(QString name);
            private:
                QHash<QString, QString> _aliases;
            };

            class IncludedGetIconStrategy : public GetIconStrategy
            {
            public:
                IncludedGetIconStrategy();
                QIcon getIcon(QString name);
            private:
                QColor brandingcolor1;
                QColor brandingcolor2;
            };

            IconManager();
            QIcon _getIcon(QString name);
            QImage _colourImage(QImage img, QColor color);
            QMutex cachemutex;
            QHash<QString, QIcon> cache;
            QList<GetIconStrategy*> getIconStrategies;
            std::shared_ptr<sh::configuration::ConfigurationValue> cfgvalPreferredStrategy;
        };

    }
}

QT_BEGIN_NAMESPACE
  extern Q_WIDGETS_EXPORT void qt_blurImage( QPainter *p, QImage &blurImage, qreal radius, bool quality, bool alphaOnly, int transposed = 0 );
QT_END_NAMESPACE

#endif // ICONMANAGER_H
