/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2020 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <iostream>
#include "base/mainthread.h"
#include "exceptions/runtimeexception.h"
#include "exceptions/programexception.h"
#include "logger.h"
#include "shallotprocess.h"

REGISTER_SINGLETON(sh::base, Logger, "logger")


sh::base::Logger::Logger() : QObject(0)
{
    _logPrefix = ShallotProcess::instance()->logPrefix();
    _logPrefix = _logPrefix.replace("{{pid}}", QString("%1").arg(QCoreApplication::applicationPid()));
    auto _logInfo = [&](QString msg){ log(msg, LogSeverity::INFO); };
    _logInfo("Starting Shallot.");
#ifdef QT_DEBUG
    _logInfo("This is a debug build.");
#else
    _logInfo("This is a release build.");
#endif
#ifdef Q_OS_UNIX
    _logInfo("Target platform: Unix/Linux.");
#elif defined(Q_OS_WIN)
    _logInfo("Target platform: Windows.");
#else
    _logInfo("Target platform: Unknown.");
#endif
#ifdef PYTHONSCRIPT
    _logInfo("Python scripting support: yes.");
#else
    _logInfo("Python scripting support: no.");
#endif
#ifdef GIO
    _logInfo("Gnome-IO support: yes.");
#else
    _logInfo("Gnome-IO support: no.");
#endif
#ifdef INOTIFY
    _logInfo("inotify support: yes.");
#else
    _logInfo("inotify support: no.");
#endif
#ifdef LIBSECRET
    _logInfo("libsecret support: yes.");
#else
    _logInfo("libsecret support: no.");
#endif
#ifdef WEBUI
    _logInfo("Web ui mode support: yes.");
#else
    _logInfo("Web ui mode support: no.");
#endif
}

sh::base::Logger::~Logger()
{
    foreach(sh::base::Logger::LogMessage* lm, _logMessages)
        delete lm;
}

void sh::base::Logger::logException(sh::exceptions::Exception& ex, sh::base::LogSeverity severity, QString source)
{
    log(QString("Exception info.") +
        "\nName: " + ex.name() +
        ", Resumeable: " + QString(ex.isResumeable()?"Yes":"No") +
        ", Retryable: " + QString(ex.isRetryable()?"Yes":"No") +
        ", ProgramException: " + QString(ex.isProgramException()?"Yes":"No") +
        "\nClasses: " + ex.classes() +
        "\nCallstack:\n" + ex.callstack() +
        ( ex.auxiliary().isEmpty() ? "" : ("\nAuxiliary information:\n" + ex.auxiliary()) )
        , LogSeverity::DEBUG, source);
    log(ex.message() +
        "\nDetails: " + ex.details()
        , severity, source);
}

void sh::base::Logger::log(QString message, sh::base::LogSeverity severity, QString source)
{
    QDateTime time = QDateTime::currentDateTime();
    int lastslash = source.lastIndexOf("/");
    int lastslash2 = source.lastIndexOf("\\", lastslash+1);
    if (lastslash2>-1)
        lastslash = lastslash2;
    QString _source = source;
    if (lastslash>-1)
        _source=_source.mid(lastslash+1);
    if (_source.length() > 25)
        _source = ":" + _source.right(24);
    else
        _source = (_source + QString(" ").repeated(25)).left(25);
    QStringList bmsg;
    if (!_logPrefix.isEmpty())
        bmsg << _logPrefix << ": ";
    bmsg << time.time().toString(Qt::ISODate) << QString("    " + QString("+").repeated(((int)severity)/100)).right(4) << " " << _source << "  " << message << "\n";
    QByteArray foo = bmsg.join("").toUtf8();
    // log to stdout
    std::cout << foo.data();
    std::cout.flush();
    // log to internal db
    sh::base::Logger::LogMessage* logMessage= new sh::base::Logger::LogMessage();
    logMessage->message = message;
    logMessage->source = source;
    logMessage->severity = severity;
    logMessage->time=time;
    mutex.lock();
    _logMessages.append(logMessage);
    mutex.unlock();
    emit newLogMessageArrived();
}

QList<sh::base::Logger::LogMessage*> sh::base::Logger::logMessages()
{
    QMutexLocker _l(&mutex);
    return _logMessages;
}

QString sh::base::Logger::logPrefix()
{
    QMutexLocker _l(&mutex);
    return _logPrefix;
}

QString sh::base::Logger::getLogAsText(LogSeverity minseverity)
{
    QStringList result;
    foreach(sh::base::Logger::LogMessage* msg, logMessages())
    {
        if (minseverity <= msg->severity)
        {
            QString ssev;
            switch(msg->severity)
            {
            case sh::base::LogSeverity::DEBUG:
                ssev="debug";
                break;
            case sh::base::LogSeverity::INFO:
                ssev="info";
                break;
            case sh::base::LogSeverity::WARNING:
                ssev="warning";
                break;
            case sh::base::LogSeverity::ERROR:
                ssev="error";
                break;
            default:
                ssev="unknown";
                break;
            }
            result << QString("%1 (%2) %3:\n%4\n")
                    .arg(msg->time.time().toString(Qt::ISODate))
                    .arg(msg->source)
                    .arg(ssev)
                    .arg(msg->message);
        }
    }
    return result.join("\n");
}
