/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2020 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef LOGGER_H
#define LOGGER_H

#include <QDateTime>
#include <QMutex>
#include <QObject>
#include "exceptions/exception.h"
#include "singletoninitializer.h"


namespace sh{
    namespace base{

        /** \brief The severity of a log message.
         */
        enum class LogSeverity {
            DEBUG = 0,
            INFO = 100, _DEFAULT = INFO,
            WARNING = 200,
            ERROR = 300, E_RROR /* ERROR because preprocessor replaces it otherwise in stupid win32 envs */ = ERROR
        };

        /** \brief
         * The logging manager.
         *
         * Use it for writing messages to the Shallot log and for reading from it.
         *
         * Note: Logging is easiest by the LOG_* macros like SH_LOG_INFO (and it provides more meta data!).
         */
        class Logger : public QObject, public Singleton
        {
            Q_OBJECT
            DECLARE_SINGLETON(Logger)
        public:

            /** \brief A single log message.
             */
            struct LogMessage{
                QString message;
                LogSeverity severity;
                QString source;
                QDateTime time;
            };

            /** \brief Logs an exception.
             */
            void logException(sh::exceptions::Exception& ex, LogSeverity severity = LogSeverity::ERROR, QString source = QString());

            /** \brief Logs a message.
             */
            void log(QString message, LogSeverity severity, QString source = QString());

            /** \brief Returns all logged messages as one text.
             */
            QString getLogAsText(LogSeverity minseverity = LogSeverity::DEBUG);

            /** \brief Returns the list of logged messages.
             */
            QList<LogMessage*> logMessages();

            /** \brief Returns the log prefix, i.e. an optional string all log output begins with.
             */
            QString logPrefix();

            virtual ~Logger();

        signals:

            /** \brief Emitted when a new log message was written.
             */
            void newLogMessageArrived();

        private:
            Logger();
            QMutex mutex;
            QList<LogMessage*> _logMessages;
            QString _logPrefix;
        };

    }
}

#define __LOG_HELPER(msg, sev, file, line) sh::base::Logger::instance()->log(msg, sev, QString(file) + ":" + QString("%1").arg(line));

#ifdef QT_DEBUG
#define SH_LOG_DEBUG(msg) __LOG_HELPER(msg, sh::base::LogSeverity::DEBUG, __FILE__, __LINE__)
#else
#define SH_LOG_DEBUG(msg) {}
#endif

#define SH_LOG_INFO(msg) __LOG_HELPER(msg, sh::base::LogSeverity::INFO, __FILE__, __LINE__)
#define SH_LOG_WARNING(msg) __LOG_HELPER(msg, sh::base::LogSeverity::WARNING, __FILE__, __LINE__)
#define SH_LOG_ERROR(msg) __LOG_HELPER(msg, sh::base::LogSeverity::E_RROR, __FILE__, __LINE__)


#endif // LOGGER_H
