/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2020 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <iostream>
#include "base/logger.h"
#include "configuration/configurationmanager.h"
#include "exceptions/argumentexception.h"
#include "ui/mainwindow.h"
#include "shallotprocess.h"

REGISTER_SINGLETON(sh::base, ShallotProcess, "early")

struct _ProjectInfo {
#include "_projectinfo.h"
};


sh::base::ShallotProcess::ShallotProcess()
{
    if (!QCoreApplication::instance())
        sh::exceptions::Exception::executeGuarded([=](){
            throw sh::exceptions::ProgramException(sh::exceptions::ExceptionData().details("ShallotProcess used too early."));
        });
    QCommandLineParser preparser; // we have to find out dynamic (i.e. unknown) option names before
    preparser.parse(QCoreApplication::arguments());
    QCommandLineParser parser;
    parser.setApplicationDescription(QObject::tr("Starts the Shallot file manager."));
    parser.addPositionalArgument("dir", QObject::tr("the directory to start in"));
    parser.addHelpOption();
    QCommandLineOption uiOption(QStringList() << "ui", QObject::tr("user interface mode (e.g. 'web', 'qt')"), "ui-mode", "qt");
    parser.addOption(uiOption);
    QCommandLineOption cfgOption(QStringList() << "cfg", QObject::tr("sets configuration values (also multiple times) like 'key:value'"), "assignment");
    parser.addOption(cfgOption);
    QCommandLineOption cfgfileOption(QStringList() << "cfgfile", QObject::tr("sets configuration values from an additional config file"), "filepath");
    parser.addOption(cfgfileOption);
    QCommandLineOption portOption(QStringList() << "port", QObject::tr("sets the tcp port to listen to (for web ui)"), "port");
    parser.addOption(portOption);
    QCommandLineOption workerminportOption(QStringList() << "workerminport", QObject::tr("sets the minimum tcp port for internal workers (for web ui)"), "port", "50000");
    parser.addOption(workerminportOption);
    QCommandLineOption workermaxportOption(QStringList() << "workermaxport", QObject::tr("sets the maximum tcp port for internal workers (for web ui)"), "port", "54000");
    parser.addOption(workermaxportOption);
    QCommandLineOption maxnumberworkersOption(QStringList() << "maxnumberworkers", QObject::tr("Maximum number of workers (for web ui janitor)"), "value");
    parser.addOption(maxnumberworkersOption);
    QCommandLineOption maxnumberworkersperhostOption(QStringList() << "maxnumberworkersperhost", QObject::tr("Maximum number of workers per host (for web ui janitor)"), "value");
    parser.addOption(maxnumberworkersperhostOption);
    QCommandLineOption maxmemorypercentOption(QStringList() << "maxmemorypercent", QObject::tr("Maximum memory in percent (for web ui janitor)"), "value");
    parser.addOption(maxmemorypercentOption);
    QCommandLineOption maxmemorypercentglobalOption(QStringList() << "maxmemorypercentglobal", QObject::tr("Maximum memory in percent globally (for web ui janitor)"), "value");
    parser.addOption(maxmemorypercentglobalOption);
    QCommandLineOption maxbrowserawaysecondsOption(QStringList() << "maxbrowserawayseconds", QObject::tr("Maximum browser away time in seconds (for web ui janitor)"), "value");
    parser.addOption(maxbrowserawaysecondsOption);
    QCommandLineOption maxidlenesssecondsOption(QStringList() << "maxidlenessseconds", QObject::tr("Maximum user idleness time in seconds (for web ui janitor)"), "value");
    parser.addOption(maxidlenesssecondsOption);
    foreach(QString k, preparser.unknownOptionNames())
        parser.addOption(QCommandLineOption(k, k, k));
    parser.process(QCoreApplication::arguments());
    _ui = parser.value(uiOption);
    foreach (QString scfg, parser.values(cfgOption)) {
        int icl = scfg.indexOf(":=");
        if (icl == -1)
            sh::exceptions::Exception::executeGuarded([&](){
                throw sh::exceptions::ArgumentException(sh::exceptions::ExceptionData().details("The given config value assignment has bad format: " + scfg));
            });
        else
            _cfgvalassignments[scfg.mid(0, icl)] = scfg.mid(icl+2);
    }
    foreach (QString scfgfile, parser.values(cfgfileOption))
        _cfgvalfiles.append(scfgfile);
    QStringList positionals = parser.positionalArguments();
    _workdir = positionals.isEmpty() ? QDir::currentPath() : parser.positionalArguments().first();
    cfgvalBrandingColor = sh::configuration::ConfigurationManager::instance()->registerConfigValue("sh::base::ShallotProcess.BrandingColor", "#3c001e");
    foreach (QString k, parser.optionNames())
        _cmdlineargs[k] = parser.values(k);
    auto portlst = parser.values(portOption);
    auto milst = parameterValueList("minport");
    if (!milst.isEmpty())
        _minport = milst.last().toInt();
    if (!_minport && !portlst.isEmpty())
        _minport = portlst.last().toInt();
    if (!_minport)
        _minport = 8100;
    auto malst = parameterValueList("maxport");
    if (!malst.isEmpty())
        _maxport = malst.last().toInt();
    if (!_maxport && !portlst.isEmpty())
        _maxport = portlst.last().toInt();
    if (!_maxport)
        _maxport = 8499;
    _workerminport = parser.value(workerminportOption).toInt();
    _workermaxport = parser.value(workermaxportOption).toInt();
}

int sh::base::ShallotProcess::minport()
{
    return _minport;
}

int sh::base::ShallotProcess::maxport()
{
    return _maxport;
}

int sh::base::ShallotProcess::workerminport()
{
    return _workerminport;
}

int sh::base::ShallotProcess::workermaxport()
{
    return _workermaxport;
}

int sh::base::ShallotProcess::maxnumberworkers(int deflt)
{
    return parameterValue("maxnumberworkers", QString("%1").arg(deflt)).toInt(); // easy, short, waterproof
}

int sh::base::ShallotProcess::maxnumberworkersperhost(int deflt)
{
    return parameterValue("maxnumberworkersperhost", QString("%1").arg(deflt)).toInt(); // easy, short, waterproof
}

int sh::base::ShallotProcess::maxmemorypercent(int deflt)
{
    return parameterValue("maxmemorypercent", QString("%1").arg(deflt)).toInt(); // easy, short, waterproof
}

int sh::base::ShallotProcess::maxmemorypercentglobal(int deflt)
{
    return parameterValue("maxmemorypercentglobal", QString("%1").arg(deflt)).toInt(); // easy, short, waterproof
}

int sh::base::ShallotProcess::maxbrowserawayseconds(int deflt)
{
    return parameterValue("maxbrowserawayseconds", QString("%1").arg(deflt)).toInt(); // easy, short, waterproof
}

int sh::base::ShallotProcess::maxidlenessseconds(int deflt)
{
    return parameterValue("maxidlenessseconds", QString("%1").arg(deflt)).toInt(); // easy, short, waterproof
}

QString sh::base::ShallotProcess::uiMode()
{
    return _ui;
}

QMap<QString, QString> sh::base::ShallotProcess::configurationValueAssignments()
{
    return _cfgvalassignments;
}

QList<QString> sh::base::ShallotProcess::configurationValueFiles()
{
    return _cfgvalfiles;
}

QString sh::base::ShallotProcess::parameterValue(QString key, QString deflt)
{
    QStringList sl = _cmdlineargs.value(key);
    return (sl.isEmpty() ? deflt : sl.first());
}

QStringList sh::base::ShallotProcess::parameterValueList(QString key)
{
    return _cmdlineargs.value(key);
}

qint64 sh::base::ShallotProcess::parameterValueInt(QString key, qint64 deflt)
{
    QString sv = parameterValue(key);
    bool ok = false;
    qint64 result = sv.toLongLong(&ok);
    return (ok ? result : deflt);
}

QStringList sh::base::ShallotProcess::parameters()
{
    return _cmdlineargs.keys();
}

QString sh::base::ShallotProcess::initialWorkDirectory()
{
    return _workdir;
}

QString sh::base::ShallotProcess::logPrefix()
{
    return parameterValue("logprefix");
}

QStringList sh::base::ShallotProcess::lang()
{
    return parameterValueList("lang");
}

QString sh::base::ShallotProcess::userDataDir()
{
    return QDir::homePath()+"/.shallot/";
}

QString sh::base::ShallotProcess::shallotDataDir()
{
#ifdef SHALLOT_DATA_DIR
#define _TO_STR(x) #x
#define TO_STR(x) _TO_STR(x)
    return TO_STR(SHALLOT_DATA_DIR);
#else
    return QCoreApplication::applicationDirPath();
#endif
}

QString sh::base::ShallotProcess::revisionString()
{
    return _ProjectInfo().version;
}

QString sh::base::ShallotProcess::homepage()
{
    return _ProjectInfo().homepage;
}

QDateTime sh::base::ShallotProcess::buildtime()
{
    QDateTime result = QDateTime::fromString(_ProjectInfo().buildtimeutc, "yyyy-MM-dd hh:mm:ss");
    result.setTimeSpec(Qt::UTC);
    return result.toLocalTime();
}

QColor sh::base::ShallotProcess::brandingColor()
{
    QMutexLocker _l1(&_brandingcolormutex);
    if (!_brandingcolor.isValid()) {
        if (sh::ui::MainWindow::uiMode() == "web")
            _brandingcolor = "#5500ff";
        else
            _brandingcolor = QColor(cfgvalBrandingColor->getConfigValueString());
    }
    return _brandingcolor;
}

