/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2020 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef SHALLOTPROCESS_H
#define SHALLOTPROCESS_H

#include <memory>
#include <QColor>
#include <QCommandLineParser>
#include <QMap>
#include <QMutex>
#include "singletoninitializer.h"


namespace sh{
    namespace configuration {
        class ConfigurationValue;
    }
    namespace base {

        /**
         * \brief Provides parameters given from the user by command line and some more simple infos.
         */
        class ShallotProcess : public Singleton
        {
            DECLARE_SINGLETON(ShallotProcess)
        public:

            /**
             * \brief The ui mode (e.g. "qt", "web") specified on command line.
             */
            QString uiMode();

            /**
             * \brief The configuration value assignments set on command line.
             */
            QMap<QString, QString> configurationValueAssignments();

            /**
             * \brief The configuration value files set on command line.
             */
            QList<QString> configurationValueFiles();

            /**
             * \brief Returns the command line parameter value for a key as string.
             */
            QString parameterValue(QString key, QString deflt=QString());

            /**
             * \brief Returns the command line parameter value for a key as list of strings (for multiple parameter usage).
             */
            QStringList parameterValueList(QString key);

            /**
             * \brief Returns the command line parameter value for a key as integer.
             */
            qint64 parameterValueInt(QString key, qint64 deflt=0);

            /**
             * \brief Returns the list of keys of all specified command line parameter names.
             */
            QStringList parameters();

            /**
             * \brief Returns the initial directory specified on command line.
             */
            QString initialWorkDirectory();

            /**
             * \brief Returns the log prefix (i.e. an optional string all log output begins with) specified on command line.
             */
            QString logPrefix();

            /**
             * \brief Returns the ui language specified on command line.
             */
            QStringList lang();

            /** \brief
             * Returns the path to a storage for per-user data. It resides somewhere within the user home directory.
             */
            static QString userDataDir();

            /** \brief
             * Returns the directory of the Shallot program data. It contains the static files which are part
             * of the Shallot program.
             */
            static QString shallotDataDir();

            /** \brief
             * Returns the Shallot revision string, i.e. the exact version number.
             * Note: This is only updated by a special build tool (anise) and not by directly compiling via qmake.
             */
            static QString revisionString();

            /** \brief
             * Returns the url to the Shallot homepage.
             */
            static QString homepage();

            /** \brief
             * Returns the time when this Shallot build was created.
             * Note: This is only updated by a special build tool (anise) and not by directly compiling via qmake.
             */
            static QDateTime buildtime();

            /** \brief
             * Returns the Shallot branding color (typically a dark red).
             */
            QColor brandingColor();

            /** @brief Return the minimal allowed port specified on command line (for web ui).
             */
            int minport();

            /** @brief Return the maximal allowed port specified on command line (for web ui).
             */
            int maxport();

            /** @brief Return the minimal allowed port specified on command line for spawing web workers with (for web ui).
             */
            int workerminport();

            /** @brief Return the maximal allowed port specified on command line for spawing web workers with (for web ui).
             */
            int workermaxport();

            /** @brief Maximum number of workers specified on command line.
             */
            int maxnumberworkers(int deflt);

            /** @brief Maximum number of workers per host specified on command line.
             */
            int maxnumberworkersperhost(int deflt);

            /** @brief Maximum memory in percent specified on command line.
             */
            int maxmemorypercent(int deflt);

            /** @brief Maximum memory in percent globally specified on command line.
             */
            int maxmemorypercentglobal(int deflt);

            /** @brief Maximum browser away time in seconds specified on command line.
             */
            int maxbrowserawayseconds(int deflt);

            /** @brief Maximum user idleness time in seconds specified on command line.
             */
            int maxidlenessseconds(int deflt);

        private:
            ShallotProcess();
            QMutex _brandingcolormutex;
            QMap<QString, QStringList> _cmdlineargs;
            QString _workdir;
            QString _ui;
            QMap<QString, QString> _cfgvalassignments;
            QList<QString> _cfgvalfiles;
            QColor _brandingcolor;
            std::shared_ptr<sh::configuration::ConfigurationValue> cfgvalBrandingColor;
            int _minport = 0;
            int _maxport = 0;
            int _workerminport = 0;
            int _workermaxport = 0;

        };

    }
}

#endif // SHALLOTPROCESS_H
