/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2020 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "logger.h"
#include "mainthread.h"
#include "singletoninitializer.h"
#include "threadpool.h"


sh::base::SingletonInitializer::SingletonInitializer()
{
}

QList<sh::base::SingletonInitializer::SingletonTuple>* sh::base::SingletonInitializer::_singletons;
QHash<QString, sh::base::SingletonInitializer::SingletonGroupTuple>* sh::base::SingletonInitializer::_groups;

bool sh::base::SingletonInitializer::_isshutdown = false;
QMutex sh::base::SingletonInitializer::shutdownmutex;

char sh::base::SingletonInitializer::registerGroup(QString groupname, QStringList groupdeps)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    if (!_groups) // this guarantees that we actually have a list even at very early times
        _groups = new QHash<QString, SingletonGroupTuple>();
#ifdef QT_DEBUG
    if (_groups->contains(groupname))
        sh::exceptions::Exception::executeGuarded([&](){
            throw sh::exceptions::ProgramException(sh::exceptions::ExceptionData().details("Group registered twice: " + groupname));
        });
#endif
    _groups->insert(groupname, std::make_tuple(groupname, groupdeps));
    return 0;
}

char sh::base::SingletonInitializer::registerSingleton(QString name, std::function<std::shared_ptr<Singleton>()> instancefct, QString groupname)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    if (!_singletons) // this guarantees that we actually have a list even at very early times
        _singletons = new QList<SingletonTuple>();
    _singletons->append(std::make_tuple(groupname, name, instancefct));
    return 0;
}

bool sh::base::SingletonInitializer::dependsOn(QString group, QString depgroup) {
    sh::exceptions::Exception::executeGuarded([&](){
        if (!_groups->contains(group))
            throw sh::exceptions::ProgramException(sh::exceptions::ExceptionData().details("Group not found: " + group));
        if (!_groups->contains(depgroup))
            throw sh::exceptions::ProgramException(sh::exceptions::ExceptionData().details("Group not found: " + depgroup));
    });
    QStringList groupsdeps;
    groupsdeps.append(std::get<0>(_groups->value(group)));
    while (!groupsdeps.isEmpty()) {
        QString gg = groupsdeps.takeFirst();
        if (gg == depgroup)
            return true;
        else
            groupsdeps.append(std::get<1>(_groups->value(gg)));
    }
    return false;
}

void sh::base::SingletonInitializer::callInitializers()
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    if (!_singletons)
        return;
    qSort(_singletons->begin(), _singletons->end(), [](const SingletonTuple& x, const SingletonTuple& y){
        QString xngrp = std::get<0>(x);
        QString yngrp = std::get<0>(y);
        if (xngrp != yngrp)
            return !sh::base::SingletonInitializer::dependsOn(xngrp, yngrp);
        else
            return std::get<1>(x) < std::get<1>(y);
    });
    foreach(SingletonTuple tup, *_singletons)
        std::get<2>(tup)();
    foreach(SingletonTuple tup, *_singletons) {
        SH_LOG_DEBUG(QString("Initializing singleton %1.").arg(std::get<1>(tup)))
        std::get<2>(tup)()->doInitialize();
    }
}

void sh::base::SingletonInitializer::shutdown()
{
    QMutexLocker _l1(&shutdownmutex);
    if (!_isshutdown) {
        _isshutdown = true;
        if (_singletons) {
            std::reverse (_singletons->begin(), _singletons->end());
            foreach (SingletonTuple tup, *_singletons) {
                SH_LOG_DEBUG(QString("Shutting down singleton %1.").arg(std::get<1>(tup)))
                std::get<2>(tup)()->shutdown();
            }
            delete _singletons;
            delete _groups;
        }
    }
}

bool sh::base::SingletonInitializer::isShutdown()
{
    QMutexLocker _l1(&shutdownmutex);
    return _isshutdown;
}

sh::base::SingletonInitializer::~SingletonInitializer()
{
    shutdown();
}

void sh::base::Singleton::doInitialize()
{
}

void sh::base::Singleton::doShutdown()
{
}

void sh::base::Singleton::shutdown()
{
    {
        QMutexLocker _l1(&_mutex_shutdown);
        _shutdown = true;
    }
    doShutdown();
}

bool sh::base::Singleton::isAlive() {
    QMutexLocker _l1(&_mutex_shutdown);
    return !_shutdown;
}

sh::base::Singleton::~Singleton()
{
}

sh::base::Singleton::Singleton()
{
}
