/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2020 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef SINGLETONINITIALIZER_H
#define SINGLETONINITIALIZER_H

#include <functional>
#include <memory>
#include <tuple>
#include <QList>
#include <QMutex>
#include <QTimer>
#include <QtConcurrent>


namespace sh{
    namespace base{

        /** @brief A singleton with initialization on Shallot startup and shutdown in the end.
         *
         * See SingletonInitializer.
         */
        class Singleton {
        public:

            /** @brief Executes singleton initialization.
             *
             * @a_dooverride_optional
             */
            virtual void doInitialize();

            /** @brief Executes singleton shutdown.
             *
             * @a_dooverride_optional
             */
            virtual void doShutdown();

            /** @brief Shutdown down this singleton.
             *
             * @a_infrastructure
             */
            void shutdown();

            /** @brief Returns if this singleton is alive (true until its shutdown begins).
             */
            bool isAlive();

            virtual ~Singleton();

            Singleton(const Singleton&) = delete;
            Singleton(Singleton &&) = delete;

        protected:
            Singleton();

        private:
            QMutex _mutex_shutdown;
            bool _shutdown = false;

        };


        /** \brief
         * Takes care of initialization and shutdown of infrastructure singletons.
         *
         * It is a very early part of shallot core infrastructure.
         * Singletons typically participate by using those macros:
         *
         * `DECLARE_SINGLETON(STYPE)`:
         * Used inside the singleton class definition.
         * STYPE is the singleton's class. It must be a subclass of sh::base::Singleton.
         * It will get a `static std::shared_ptr<STYPE> instance()` method by that.
         *
         * `REGISTER_SINGLETON(NS, STYPE, GROUPNAME)`:
         * Used inside the source file.
         * NS is the namespace of the singleton.
         * STYPE is the singleton's class.
         * GROUPNAME is the singleton group name (used for dependency handling, see later).
         *
         * For singleton groups, there is `REGISTER_SINGLETON_GROUP(GROUPNAME, ...)`:
         * GROUPNAME is the group name used for grouping singletons together.
         * Additional arguments are other group names; those groups are considered as dependencies, which must be fulfilled before this group can.
         *
         * For just initializing stuff via static methods (without a singleton instance), use `REGISTER_STATICINIT(NS, STYPE, GROUPNAME)`:
         * NS is the namespace of the class to initialize.
         * STYPE is the class to initialize. It must provide `public` `static void doInitialize()` and `static void doShutdown()`.
         * GROUPNAME is the group name.
         */
        class SingletonInitializer
        {
        public:

            /** @brief Registers a singleton group.
             *
             * You should typically use REGISTER_SINGLETON_GROUP. See SingletonInitializer.
             */
            static char registerGroup(QString groupname, QStringList groupdeps);

            /** @brief Registers a singleton.
             *
             * You should typically use DECLARE_SINGLETON and REGISTER_SINGLETON. See SingletonInitializer.
             */
            static char registerSingleton(QString name, std::function<std::shared_ptr<Singleton>()> instancefct, QString groupname);

            /** \brief
             * Executes all create-callbacks at first, then all init-callbacks. In both runs, the callbacks with lower `index` come first.
             * Afterwards, the singletons are considered as up and running.
             */
            void callInitializers();

            static QThread* initializerThread();

            /** \brief Constructed only by the infrastructure and made available otherwise.
             */
            SingletonInitializer();

            ~SingletonInitializer();

            /** \brief
             * Executes all shutdown-callbacks at first, then all remove-callbacks. The callback order is the reversed one of callInitializers.
             * Afterwards, the singletons are considered as shut down and removed.
             */
            void shutdown();

            /** \brief
             * If the singletons are completely shut down and removed.
             */
            static bool isShutdown();

        private:
            static bool dependsOn(QString group, QString depgroup);
            typedef std::tuple<QString, QString, std::function<std::shared_ptr<Singleton>()>> SingletonTuple;
            typedef std::tuple<QString, QStringList> SingletonGroupTuple;
            static QList<SingletonTuple>* _singletons;
            static QHash<QString, SingletonGroupTuple>* _groups;
            static bool _isshutdown;
            static QMutex shutdownmutex;

        };

    }
}

#define REGISTER_SINGLETON_GROUP(GROUPNAME, ...) char _shsingleton_group_ ## GROUPNAME = sh::base::SingletonInitializer::registerGroup(#GROUPNAME, QStringList{__VA_ARGS__});

#define DECLARE_SINGLETON(STYPE) public: static std::shared_ptr<STYPE> instance(); private:

#define REGISTER_SINGLETON(NS, STYPE, GROUPNAME) \
    char _shsingleton_ ## STYPE = sh::base::SingletonInitializer::registerSingleton(#STYPE, [](){ return NS::STYPE::instance();}, GROUPNAME); \
    std::shared_ptr<NS::STYPE> NS::STYPE::instance() { /*meyers singleton*/ \
        static std::shared_ptr<NS::STYPE> _i = std::shared_ptr<NS::STYPE>(new NS::STYPE(), [](NS::STYPE* x){ sh::base::MainThread::dispatcher()->invokeAsync([=](){delete x;}); }); \
        return _i; \
    }

#define REGISTER_STATICINIT(NS, STYPE, GROUPNAME) \
    namespace sh { \
        /** \brief Internal helper class for initializing NS::STYPE. */ \
        class _ShStaticInit_ ## STYPE : public sh::base::Singleton { \
            DECLARE_SINGLETON(_ShStaticInit_ ## STYPE) \
            void doInitialize() override { NS::STYPE::doInitialize(); }; \
            void doShutdown() override { NS::STYPE::doShutdown(); }; \
        }; \
    } \
    REGISTER_SINGLETON(sh, _ShStaticInit_ ## STYPE, GROUPNAME)

#endif // SINGLETONINITIALIZER_H

//TODO # shutdown hangs (thumbnails?!)
