/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2020 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <QThread>
#include <QMutex>
#include <QWaitCondition>
#include <functional>
#include "threaddispatcher.h"
#include "threadpool.h"


sh::base::ThreadDispatcher::ThreadDispatcher(QObject *parent, QThread* thread) :
    QObject(parent)
{
    _thread = thread;
    this->moveToThread(thread);
    if (parent)
        parent->moveToThread(thread);
    connect(this, SIGNAL(_invoked()), this, SLOT(slot_invoked()), Qt::QueuedConnection);
}


void sh::base::ThreadDispatcher::invokeSync(std::function<void ()> fct)
{
    if (inThread())
        sh::exceptions::Exception::executeGuarded([&](){
            fct();
        });
    else {
        QMutex _mutex;
        QWaitCondition _condWait;
        bool _done = false;
        sh::exceptions::Exception exception;
        bool genericException=false;
        bool cancelException=false;
        bool threadAbortException=false;
        bool cancellable=!sh::exceptions::ExceptionHandlerSettingsManager::instance()->handlerSettings()->exceptionHandler_cancel.isEmpty();
        _invokeAsync([&](){
            std::shared_ptr<sh::exceptions::Exception::RegisterHandler<std::function<void ()>>> cancelmock = 0;
            if (cancellable)
                // we create a pseudo cancel handler in the foreign thread if we have one here. So the 'cancelable detection' works.
                cancelmock = sh::exceptions::Exception::createRegisterHandler<std::function<void()> >([&]()
                { }, &sh::exceptions::ExceptionHandlerSettingsManager::instance()->handlerSettings()->exceptionHandler_cancel);
            try {
                fct();
            }
            catch (sh::exceptions::CancelException) {
                cancelException=true;
            }
            catch (sh::exceptions::ThreadAbortException) {
                threadAbortException=true;
            }
            catch (sh::exceptions::Exception& ex) {
                exception=sh::exceptions::Exception(ex.data());
            }
            catch (...) {
                genericException=true;
            }
            _mutex.lock();
            _done = true;
            _condWait.wakeOne();
            _mutex.unlock();
        });
        _mutex.lock();
        if (!_done)
            _condWait.wait(&_mutex);
        _mutex.unlock();
        sh::exceptions::Exception::executeGuarded([&](){
            if (!exception.data().isEmpty())
                throw exception;
            if (genericException)
                throw sh::exceptions::ProgramException(sh::exceptions::ExceptionData().details("Undefined error in a thread dispatcher client routine."));
            if (cancelException)
                throw sh::exceptions::CancelException();
            if (threadAbortException)
                throw sh::exceptions::ThreadAbortException();
        });
    }
}

void sh::base::ThreadDispatcher::invokeAsync(std::function<void ()> fct)
{
    _FORBID_ASYNC_CALLS_BARRIER
    _invokeAsync(fct);
}

void sh::base::ThreadDispatcher::invokeAsync(std::function<void ()> fct, std::shared_ptr<void> obj1, std::shared_ptr<void> obj2, std::shared_ptr<void> obj3)
{
    invokeAsync([=](){
        std::shared_ptr<void> _obj1=obj1, _obj2=obj2, _obj3=obj3;
        fct();
    });
}

bool sh::base::ThreadDispatcher::inThread()
{
    return QThread::currentThread() == _thread;
}

void sh::base::ThreadDispatcher::_invokeAsync(std::function<void ()> fct)
{
    QMutexLocker _l(&callsmutex);
    _queue.enqueue(fct);
    emit _invoked();
}

void sh::base::ThreadDispatcher::slot_invoked()
{
    callsmutex.lock();
    std::function<void ()> fct = _queue.dequeue();
    callsmutex.unlock();
    sh::exceptions::Exception::executeGuarded([&](){
        fct();
    }, sh::exceptions::ExecuteGuardFlag::ALL_ERRORS_KILL_SHALLOT);
}
