/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2020 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef THREADDISPATCHER_H
#define THREADDISPATCHER_H

#include <QMutex>
#include <QObject>
#include <QQueue>
#include <QThread>
#include "exceptions/threadingexception.h"


namespace sh{
    namespace base{

        /** \brief
         * Dispatcher for sync/async invocation of functions into the associated thread.
         */
        class ThreadDispatcher : public QObject
        {
            Q_OBJECT
        public:

            /** \brief Constructed only by the infrastructure and made available otherwise.
             */
            explicit ThreadDispatcher(QObject *parent, QThread *thread);

            /** \brief Synchronously executes a function in the dispatcher's thread.
             *
             * If it already runs in the current thread, it just executes the function.
             * It returns when the function is completely executed.
             */
            void invokeSync( std::function<void ()>);

            /** \brief Asynchronously executes a function in the dispatcher's thread.
             *
             * It directly returns. The actual execution will take place in a later message loop iteration of that thread.
             */
            void invokeAsync(std::function<void ()>);

            /** \brief Like the other variant, but it can also conserve some shared pointers until after execution.
             */
            void invokeAsync(std::function<void ()>, std::shared_ptr<void> obj1, std::shared_ptr<void> obj2=0, std::shared_ptr<void> obj3=0);

            /** \brief Returns if the current thread is already equal to the dispatcher's thread (so dispatching isn't required).
             */
            bool inThread();

        signals:
            void _invoked();

        private:
            void _invokeAsync(std::function<void ()> );
            QMutex callsmutex;
            QThread* _thread;
            QQueue< std::function<void ()> > _queue;

        private slots:
            void slot_invoked();
        };

    }
}

#ifdef QT_DEBUG
#define CHECK_CORRECT_CALLER_THREAD if (QThread::currentThread()!=this->thread()) sh::exceptions::Exception::executeGuarded([&](){ throw sh::exceptions::ThreadingException(sh::exceptions::ExceptionData().details("A method is called from a foreign thread, which is forbidden by that method.")); });
#else
#define CHECK_CORRECT_CALLER_THREAD
#endif

#endif // THREADDISPATCHER_H
