/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2020 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <cmath>
#include <cstdlib>
#include "logger.h"
#include "mainthread.h"
#include "threadpool.h"

REGISTER_STATICINIT(sh::base, ThreadPool, "foundation")


void sh::base::ThreadPool::doInitialize()
{
    for (int i=0; i<THREAD_COUNT; i++) {
        ThreadPoolThread* t = new ThreadPoolThread();
        t->start();
        _threads.append(t);
    }
}

sh::base::ThreadPoolThread::ThreadPoolThread()
{
}

void sh::base::ThreadPoolThread::run()
{
    ThreadPool::poolmutex.lock();
    ThreadPool::_threadsalive++;
    while (!ThreadPool::_shutsdown) {
        ThreadPool::pooladdedcondition.wait(&ThreadPool::poolmutex);
        if (ThreadPool::_shutsdown)
            break;
        bool foundstuff = true;
        while (foundstuff) {
            if (ThreadPool::tasks.isEmpty())
                foundstuff = false;
            else {
                std::function<void ()> f = ThreadPool::tasks.takeFirst();
                ThreadPool::poolmutex.unlock();
                try {
                    sh::exceptions::Exception::executeGuarded([&](){
                        f();
                    }, sh::exceptions::ExecuteGuardFlag::ALL_ERRORS_KILL_SHALLOT);
                }
                catch(sh::exceptions::ThreadAbortException&) {
                }
                ThreadPool::poolmutex.lock();
            }
        }
    }
    ThreadPool::_threadsalive--;
    ThreadPool::poolmutex.unlock();
}

void sh::base::ThreadPool::doShutdown()
{
    {
        QMutexLocker l1(&poolmutex);
        _shutsdown = true;
        pooladdedcondition.wakeAll();
    }
    while (!isShutdown())
        QCoreApplication::processEvents();
    foreach (ThreadPoolThread* t, _threads)
        delete t;
}

QMutex sh::base::ThreadPool::poolmutex;

QWaitCondition sh::base::ThreadPool::pooladdedcondition;

QList<std::function<void ()>> sh::base::ThreadPool::tasks;

bool sh::base::ThreadPool::_shutsdown = false;

int sh::base::ThreadPool::_threadsalive = 0;

QList<sh::base::ThreadPoolThread*> sh::base::ThreadPool::_threads;

#ifdef QT_DEBUG
QList<QThread*> sh::base::AsyncForbid::restrictedThreads;
QMutex sh::base::AsyncForbid::mutex;
QList<QThread*> sh::base::AsyncForbidIgnore::restrictionIgnoredThreads;
QMutex sh::base::AsyncForbidIgnore::mutex;
#endif
