/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2020 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "base/mainthread.h"
#include "base/logger.h"
#include "base/shallotprocess.h"
#include "filesystem/eurl.h"
#include "ui/mainwindow.h"
#include "tools/benchmarking.h"
#include "tools/userdirlock.h"
#include "configurationmanager.h"
#include "configurationvalue.h"

REGISTER_SINGLETON(sh::configuration, ConfigurationManager, "early")


sh::configuration::ConfigurationManager::ConfigurationManager() : QObject()
{
}

sh::configuration::ConfigurationManager::ConfigurationValueImpl::ConfigurationValueImpl(QString name, QVariant deflt, QString desc, QString longdesc, sh::configuration::ConfigurationCategory cat, sh::configuration::ConfigurationValueType* valuetype, QString changehint, sh::configuration::ConfigurationManager *mgr)
   :sh::configuration::ConfigurationValue(name, deflt, desc, longdesc, cat, valuetype, changehint), mgr(mgr)
{
}

QVariant sh::configuration::ConfigurationManager::ConfigurationValueImpl::getConfigValueVariant()
{
    QVariant result = mgr->getFixedConfigValue(this->name());
    if (result.isNull())
    {
        sh::tools::UserDirLock userdirlock;
        QSettings settings(sh::base::ShallotProcess::userDataDir()+"cfg", QSettings::IniFormat);
        result = settings.value(name(), defaultValue());
    }
    return result;
}

void sh::configuration::ConfigurationManager::ConfigurationValueImpl::setConfigValue(QVariant value)
{
    if (!mayChange())
        return;
    {
        sh::tools::UserDirLock userdirlock;
        QSettings settings(sh::base::ShallotProcess::userDataDir()+"cfg", QSettings::IniFormat);
        settings.setValue(name(), value);
    }
    foreach(std::function<void()> f, callbacks())
        f();
}

bool sh::configuration::ConfigurationManager::ConfigurationValueImpl::mayChange()
{
    return mgr->getFixedConfigValue(this->name()).isNull();
}

QList<std::shared_ptr<sh::configuration::ConfigurationValue> > sh::configuration::ConfigurationManager::getAllConfigurationValues()
{
    QMutexLocker _l1(&_mutex);
    return _configvalues.values();
}

QVariant sh::configuration::ConfigurationManager::getFixedConfigValue(QString key)
{
    QMutexLocker _l1(&_mutex);
    if (!_fixed_configvalues_initialized) {
        // load what's set by command line
        foreach (QString cvf, sh::base::ShallotProcess::instance()->configurationValueFiles()) {
            QSettings settings(cvf, QSettings::IniFormat);
            foreach (QString cvfk, settings.allKeys())
                _fixed_configvalues[cvfk] = settings.value(cvfk);
        }
        foreach (auto cva, sh::base::ShallotProcess::instance()->configurationValueAssignments().toStdMap())
            _fixed_configvalues[cva.first] = cva.second;
        _fixed_configvalues_initialized = true;
    }
    return _fixed_configvalues[key];
}

std::shared_ptr<sh::configuration::ConfigurationValue> sh::configuration::ConfigurationManager::registerConfigValue(QString name, QVariant deflt, sh::configuration::ConfigurationValueType* valuetype, QString desc, sh::configuration::ConfigurationCategory cat, QString longdesc, QString changehint)
{
    QMutexLocker _l1(&_mutex);
    if (!valuetype)
        valuetype = sh::configuration::ConfigurationValue::valueTypeString();
    std::shared_ptr<sh::configuration::ConfigurationValue> result = std::make_shared<ConfigurationValueImpl>(name, deflt, desc, longdesc, cat, valuetype, changehint, this);
    _configvalues[result->name()] = result;
    return result;
}
