/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2020 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef CONFIGURATIONMANAGER_H
#define CONFIGURATIONMANAGER_H

#include <memory>
#include <QMap>
#include <QVariant>
#include "base/singletoninitializer.h"
#include "configurationvalue.h"


namespace sh{
    namespace filesystem{ class Eurl; }

    namespace configuration {

        /** \brief
         * Manages the Shallot configuration.
         *
         * This is the more static part of Shallot.
         * Much more stuff, i.e. everything you see in 'Manage saved settings', is in
         * sh::settings::SettingsManager.
         */
        class ConfigurationManager : public QObject, public sh::base::Singleton
        {
            Q_OBJECT
            DECLARE_SINGLETON(ConfigurationManager)
        public:

            /** \brief
             * Creates and registers a new ConfigurationValue.
             * This is typically done once at the beginning.
             * Each registered instance is shown in the 'Tuning' dialog and can be set (for making changes) and observed (for applying changes) in code as well.
             * Useful e.g. for some internal bookkeeping or for exotic machine-wide configuration values (path to some tool, ...).
             */
            std::shared_ptr<ConfigurationValue> registerConfigValue(QString name, QVariant deflt, ConfigurationValueType* valuetype=0, QString desc="", ConfigurationCategory cat=CategoryNone, QString longdesc="", QString changehint="");

            /** \brief
             * Returns a list of all registered configuration value instances.
             */
            QList<std::shared_ptr<ConfigurationValue> > getAllConfigurationValues();

        private:

            /** \brief
             * Returns a non-null value if 'key' was set in a fixed way (e.g. via command line).
             */
            QVariant getFixedConfigValue(QString key);

            class ConfigurationValueImpl : public ConfigurationValue
            {
            public:
                ConfigurationValueImpl(QString name, QVariant deflt, QString desc, QString longdesc, ConfigurationCategory cat, ConfigurationValueType* valuetype, QString changehint, ConfigurationManager* mgr);
                void setConfigValue(QVariant value) override;
                QVariant getConfigValueVariant() override;
                bool mayChange() override;
            private:
                ConfigurationManager* mgr;
            };

            explicit ConfigurationManager();
            QHash<QString, std::shared_ptr<ConfigurationValue>> _configvalues;
            QHash<QString, QVariant> _fixed_configvalues;
            bool _fixed_configvalues_initialized = false;
            QMutex _mutex;
        };

    }
}

#endif // CONFIGURATIONMANAGER_H
