/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2020 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "base/threadpool.h"
#include "configurationmanager.h"
#include "configurationvalue.h"
#include "configurationvaluetype.h"

REGISTER_EARLYMUTEX(sh::configuration::ConfigurationValue, _typemutex)


sh::configuration::ConfigurationValueType* sh::configuration::ConfigurationValue::valueTypeString()
{
    QMutexLocker l1(_typemutex());
    static sh::configuration::ConfigurationValueType* _valueTypeString = new sh::configuration::ConfigurationValueTypeString();
    return _valueTypeString;
}

sh::configuration::ConfigurationValueType* sh::configuration::ConfigurationValue::valueTypeInteger()
{
    QMutexLocker l1(_typemutex());
    static sh::configuration::ConfigurationValueType* _valueTypeInteger = new sh::configuration::ConfigurationValueTypeInteger();
    return _valueTypeInteger;
}

sh::configuration::ConfigurationValueType* sh::configuration::ConfigurationValue::valueTypeFloat()
{
    QMutexLocker l1(_typemutex());
    static sh::configuration::ConfigurationValueType* _valueTypeFloat = new sh::configuration::ConfigurationValueTypeFloat();
    return _valueTypeFloat;
}

sh::configuration::ConfigurationValueType* sh::configuration::ConfigurationValue::valueTypeBoolean()
{
    QMutexLocker l1(_typemutex());
    static sh::configuration::ConfigurationValueType* _valueTypeBoolean = new sh::configuration::ConfigurationValueTypeBoolean();
    return _valueTypeBoolean;
}

sh::configuration::ConfigurationValueType* sh::configuration::ConfigurationValue::valueTypeLocalFilePath()
{
    QMutexLocker l1(_typemutex());
    static sh::configuration::ConfigurationValueType* _valueTypeLocalFilePath = new sh::configuration::ConfigurationValueTypeLocalFilePath();
    return _valueTypeLocalFilePath;
}

sh::configuration::ConfigurationValue::ConfigurationValue(QString name, QVariant deflt, QString desc, QString longdesc, sh::configuration::ConfigurationCategory cat, sh::configuration::ConfigurationValueType* valuetype, QString changehint)
    : _name(name), _desc(desc), _longdesc(longdesc), _changehint(changehint), _category(cat), _valuetype(valuetype), _deflt(deflt), _callbacksmutex(QMutex::Recursive)
{
}

int sh::configuration::ConfigurationValue::getConfigValueInt()
{
    return getConfigValueVariant().value<int>();
}

double sh::configuration::ConfigurationValue::getConfigValueFloat()
{
    return getConfigValueVariant().value<double>();
}

bool sh::configuration::ConfigurationValue::getConfigValueBool()
{
    return getConfigValueVariant().value<bool>();
}

QString sh::configuration::ConfigurationValue::getConfigValueString()
{
    return getConfigValueVariant().value<QString>();
}

void sh::configuration::ConfigurationValue::consumeValue(QObject* o, std::function<void()> f)
{
    QMutexLocker _l1(&_callbacksmutex);
    static int i = 0;
    int _i = i++;
    _callbacks.insert(_i, f);
    QObject::connect(o, &QObject::destroyed, [=](){
        _callbacks.remove(_i);
    });
    f();
}

QList<std::function<void()>> sh::configuration::ConfigurationValue::callbacks()
{
    QMutexLocker _l1(&_callbacksmutex);
    return _callbacks.values();
}

QString sh::configuration::ConfigurationValue::name()
{
    return _name;
}

QString sh::configuration::ConfigurationValue::description()
{
    return (_desc.isEmpty() ? _name : _desc);
}

QString sh::configuration::ConfigurationValue::longDescription()
{
    return _longdesc;
}

sh::configuration::ConfigurationCategory sh::configuration::ConfigurationValue::category()
{
    return _category;
}

QString sh::configuration::ConfigurationValue::changeHint()
{
    return _changehint;
}

QVariant sh::configuration::ConfigurationValue::defaultValue()
{
    return _deflt;
}

sh::configuration::ConfigurationValueType* sh::configuration::ConfigurationValue::valueType()
{
    return _valuetype;
}
