/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2020 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef CONFIGURATIONVALUE_H
#define CONFIGURATIONVALUE_H

#include "tools/earlymutex.h"
#include <functional>
#include <QMutex>
#include <QVariant>


namespace sh{
    namespace configuration {

        /** \brief
         * Categories of Configursh::configuration::ConfigurationValuentations.
         * They are used for grouping them in the dialogs. There is no
         * difference in behavior implied by this choice.
         */
        enum ConfigurationCategory
        {
            /** \brief
             * No category.
             */
            CategoryNone=0,
            /** \brief
             * User interface category.
             */
            CategoryGUI=1,
            /** \brief
             * Behavioral configuration category.
             */
            CategoryBehavior=2,
            /** \brief
             * External tools category.
             */
            CategoryExternalTools=3
        };

        class ConfigurationValueType;

        /* TODO# monitor for external changes? */

        /** \brief
         * Abstract base class for a configuration value which is managed in the *Tuning* dialog.
         *
         * Each instance can get and set the value associated to it.
         *
         * You should typically not need to override it. See sh::configuration::ConfigurationManager.
         * See Shallot documentation for more details.
         */
        class ConfigurationValue
        {
            friend class ConfigurationManager;
            friend class ConfigurationValueType;
        public:

            /** @a_infrastructure
             */
            ConfigurationValue(QString name, QVariant deflt, QString desc, QString longdesc, ConfigurationCategory cat, ConfigurationValueType* valuetype, QString changehint);

            /** \brief Returns the current value of this instance as integer.
             */
            int getConfigValueInt();

            /** \brief Returns the current value of this instance as floating point number.
             */
            double getConfigValueFloat();

            /** \brief Returns the current value of this instance as boolean.
             */
            bool getConfigValueBool();

            /** \brief Returns the current value of this instance as string.
             */
            QString getConfigValueString();

            /** \brief Returns the current value of this instance as variant.
             *
             * @a_dooverride
             */
            virtual QVariant getConfigValueVariant() =0;

            /** \brief Sets a new configuration value.
             *
             * @a_dooverride
             */
            virtual void setConfigValue(QVariant value) =0;

            /** \brief Returns if changes are allowed.
             *
             * @a_dooverride
             */
            virtual bool mayChange() =0;

            /** \brief Adds a consumer function, triggered directly and whenever the value changes.
             * \param o For lifetime monitoring.
             */
            void consumeValue(QObject* o, std::function<void()> h);

            /** \brief Returns the internal name of this configuration.
             */
            QString name();

            /** \brief Returns the short description of this configuration.
             */
            QString description();

            /** \brief Returns the long description of this configuration.
             */
            QString longDescription();

            /** \brief Returns the category of this configuration (mainly for UI structuring).
             */
            ConfigurationCategory category();

            /** \brief Returns the change hint text of this configuration.
             */
            QString changeHint();

            /** \brief Returns the default value of this configuration.
             */
            QVariant defaultValue();

            /** \brief Returns the value type of this configuration.
             */
            ConfigurationValueType* valueType();

            /** \brief Returns the string type for configuration values.
             */
            static ConfigurationValueType* valueTypeString();

            /** \brief Returns the integer type for configuration values.
             */
            static ConfigurationValueType* valueTypeInteger();

            /** \brief Returns the floating point number type for configuration values.
             */
            static ConfigurationValueType* valueTypeFloat();

            /** \brief Returns the boolean type for configuration values.
             */
            static ConfigurationValueType* valueTypeBoolean();

            /** \brief Returns the filepath type for configuration values.
             */
            static ConfigurationValueType* valueTypeLocalFilePath();

        protected:
            QList<std::function<void()>> callbacks();
            QString _name;
            QString _desc;
            QString _longdesc;
            QString _changehint;
            ConfigurationCategory _category;
            ConfigurationValueType* _valuetype;
            QVariant _deflt;
            QMap<int, std::function<void()>> _callbacks;
            QMutex _callbacksmutex;

            DECLARE_EARLYMUTEX(_typemutex)
        };

    }
}

#endif // CONFIGURATIONVALUE_H
