/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2020 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "actions/common/actionabstracttransfertree.h"
#include "filesystem/filesystemnode.h"
#include "filesystem/filesystemoperation.h"
#include "filesystem/eurl.h"
#include "filesystem/operation.h"
#include "detailcolumns/detailcolumnresolvedsymlink.h"
#include "detailcolumns/detailcolumnfilesize.h"
#include "detailcolumns/detailcolumnmimetype.h"
#include "detailcolumns/detailcolumnmtime.h"
#include "base/singletoninitializer.h"
#include "base/threadpool.h"
#include "base/mainthread.h"
#include "filepropertydialogtabgeneral.h"
#include "ui/filepropertydialogtabactionsview.h"

REGISTER_STATICINIT(sh::filepropertydialogtabs, FilePropertyDialogTabGeneral, "base")


sh::filepropertydialogtabs::FilePropertyDialogTabGeneral::FilePropertyDialogTabGeneral()
{
}

QString sh::filepropertydialogtabs::FilePropertyDialogTabGeneral::title()
{
    return QObject::tr("&General");
}

QList<QString> sh::filepropertydialogtabs::FilePropertyDialogTabGeneral::properties()
{
    QStringList result;
    result << QObject::tr("Location");
    result << QObject::tr("Resolved location");
    result << QObject::tr("Type");
    result << QObject::tr("Size");
    result << QObject::tr("Last modified");
    return result;
}

void sh::filepropertydialogtabs::FilePropertyDialogTabGeneral::populateWidget(int, sh::ui::FilePropertyDialogTabActionsView* widget)
{
    widget->setContent(dialog()->createTabViewText());
}

void sh::filepropertydialogtabs::FilePropertyDialogTabGeneral::updateWidget(int i, sh::ui::FilePropertyDialogTabActionsView* widget, sh::filesystem::Operation* op)
{
    QString result;
    bool visible = true;
    if (nodes().length() == 1)
        switch(i)
        {
        case 0: // Location
        {
            result = nodes().first()->eurl()->asString();
            break;
        }
        case 1: // Resolved location
        {
            QList<std::shared_ptr<sh::filesystem::FilesystemNode>> resnodes = op->filesystem()->resolveNodeLink(nodes().first(), true, true);
            if (resnodes.isEmpty())
                visible = false;
            else
                result = resnodes.first()->eurl()->asString();
            break;
        }
        case 2: // Type
        {
            result = sh::detailcolumns::DetailColumnMimetype::instance()->requestValue(nodes().first(), op).toString();
            break;
        }
        case 3: // Size
        {
            qint64 size = (qint64) sh::detailcolumns::DetailColumnFilesize::instance()->requestValue(nodes().first(), op).toLongLong();
            double dsize = size;
            QString unit;
            sh::actions::common::ActionAbstractTransferTree::makereadableunits(&dsize, &unit);
            result = QObject::tr("%1 %2").arg(dsize, 0, 'f', 2).arg(unit);
            if (dsize != size)
                result = QObject::tr("%1 (%2 byte)").arg(result).arg(size);
            break;
        }
        case 4: // Last modified
        {
            result = sh::detailcolumns::DetailColumnMtime::instance()->requestValue(nodes().first(), op).toDateTime().toString(Qt::DefaultLocaleLongDate);
            break;
        }
        }
    sh::base::MainThread::dispatcher()->invokeSync([&](){
        ((sh::ui::FilePropertyDialogTabTextView*)((sh::ui::FilePropertyDialogTabActionsView*)widget)->content())->setContent(result);
        widget->setVisible(visible);
    });
}

void sh::filepropertydialogtabs::FilePropertyDialogTabGeneral::doInitialize()
{
    REGISTER_FILEPROPERTYDIALOGTAB(sh::ui::FilePropertyDialogTabFactory::REGISTER_FILEPROPERTYDIALOGTAB_INDEX_CORE, sh::filepropertydialogtabs::FilePropertyDialogTabGeneral);
}

void sh::filepropertydialogtabs::FilePropertyDialogTabGeneral::doShutdown()
{
}
