/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2020 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "actions/common/actionabstracttransfertree.h"
#include "actions/actionexecutioninfo.h"
#include "actions/actionexecutionuserfeedback.h"
#include "filesystem/filesystemnode.h"
#include "filesystem/eurl.h"
#include "filesystem/operation.h"
#include "detailcolumns/detailcolumncustomattributes.h"
#include "base/singletoninitializer.h"
#include "base/threadpool.h"
#include "base/mainthread.h"
#include "ui/filepropertydialogtabactionsview.h"
#include "filepropertydialogtabunixpermissions.h"

#ifdef Q_OS_UNIX
#include <sys/stat.h>
#include <sys/types.h>
#include <pwd.h>
#include <grp.h>
#endif

REGISTER_STATICINIT(sh::filepropertydialogtabs, FilePropertyDialogTabUnixPermissions, "base")


sh::filepropertydialogtabs::FilePropertyDialogTabUnixPermissions::FilePropertyDialogTabUnixPermissions()
{
}

QString sh::filepropertydialogtabs::FilePropertyDialogTabUnixPermissions::title()
{
    return QObject::tr("&Permissions");
}

QList<QString> sh::filepropertydialogtabs::FilePropertyDialogTabUnixPermissions::properties()
{
    QStringList result;
    result << "";
    return result;
}

void sh::filepropertydialogtabs::FilePropertyDialogTabUnixPermissions::populateWidget(int, sh::ui::FilePropertyDialogTabActionsView* widget)
{
    widget->setContent(dialog()->createTabViewText());
    widget->setButtons(QStringList({QObject::tr("Change permissions")}));
    widget->onButtonTriggered([=](int i){slot_buttontriggered(i);});
}

void sh::filepropertydialogtabs::FilePropertyDialogTabUnixPermissions::slot_buttontriggered(int)
{
    auto me = std::static_pointer_cast<sh::filepropertydialogtabs::FilePropertyDialogTabUnixPermissions>(shared_from_this());
    sh::base::ThreadPool::enqueueForce([=](){
        std::shared_ptr<sh::actions::ActionActionItem> a = std::make_shared<ActionChange>(me);
        a->initializeSync();
        a->execute();
    });
}

void sh::filepropertydialogtabs::FilePropertyDialogTabUnixPermissions::updateWidget(int, sh::ui::FilePropertyDialogTabActionsView* widget, sh::filesystem::Operation* op)
{
#ifdef Q_OS_UNIX
    QString desc = "";
    if (nodes().length() == 1)
    {
        QVariant vv = sh::detailcolumns::DetailColumnCustomAttributes::instance()->requestValue(nodes().first(), op);
        QMap<QString, QString> vals = sh::detailcolumns::DetailColumnCustomAttributes::getMapByValue(vv);
        if (vals.contains("unix_permissions"))
        {
            QStringList vl = vals["unix_permissions"].split(";");
            int flags = vl[0].toInt();
            int uid = vl[1].toInt();
            int gid = vl[2].toInt();
            if (flags & S_ISUID)
                desc += QObject::tr("set user on execute") + "\n";
            if (flags & S_ISGID)
                desc += QObject::tr("set group on execute") + "\n";
            if (flags & S_ISVTX)
                desc += QObject::tr("sticky") + "\n";
            QString usermay = userPermissionsDescription(flags, S_IRUSR, S_IWUSR, S_IXUSR);
            QString groupmay = userPermissionsDescription(flags, S_IRGRP, S_IWGRP, S_IXGRP);
            QString othersmay = userPermissionsDescription(flags, S_IROTH, S_IWOTH, S_IXOTH);
            desc += QObject::tr("user '%1' may: %2").arg(getUserName(uid)).arg(usermay) + "\n";
            desc += QObject::tr("group '%1' may: %2").arg(getGroupName(gid)).arg(groupmay) + "\n";
            desc += QObject::tr("others may: %1").arg(othersmay);
        }
    }
    sh::base::MainThread::dispatcher()->invokeSync([&](){
        dynamic_cast<sh::ui::FilePropertyDialogTabTextView*>(dynamic_cast<sh::ui::FilePropertyDialogTabActionsView*>(widget)->content())->setContent(desc);
        widget->setVisible(!desc.isEmpty());
    });
#endif
}

QString sh::filepropertydialogtabs::FilePropertyDialogTabUnixPermissions::userPermissionsDescription(int flags, int rflag, int wflag, int xflag)
{
    QStringList result;
    if (flags & rflag)
        result << QObject::tr("read");
    if (flags & wflag)
        result << QObject::tr("write");
    if (flags & xflag)
        result << QObject::tr("execute");
    if (result.isEmpty())
        result << QObject::tr("nothing");
    return result.join(", ");
}

QString sh::filepropertydialogtabs::FilePropertyDialogTabUnixPermissions::getUserName(int uid)
{
#ifdef Q_OS_UNIX
    struct passwd *pw;
    pw = getpwuid((uid_t)uid);
    if (pw)
        return QString(pw->pw_name);
#endif
    return QString::number(uid);
}

QString sh::filepropertydialogtabs::FilePropertyDialogTabUnixPermissions::getGroupName(int gid)
{
#ifdef Q_OS_UNIX
    struct group *gr;
    gr = getgrgid((gid_t)gid);
    if (gr)
        return QString(gr->gr_name);
#endif
    return QString::number(gid);
}

QMap<int, QString> sh::filepropertydialogtabs::FilePropertyDialogTabUnixPermissions::getAllUsers()
{
    QMap<int, QString> result;
#ifdef Q_OS_UNIX
    sh::base::MainThread::dispatcher()->invokeSync([&](){
        struct passwd *grpp;
        setpwent();
        while (true) {
            grpp = getpwent();
            if (!grpp)
                break;
            result[grpp->pw_uid] = grpp->pw_name;
        }
        endpwent();
    });
#endif
    return result;
}

QMap<int, QString> sh::filepropertydialogtabs::FilePropertyDialogTabUnixPermissions::getAllGroups()
{
    QMap<int, QString> result;
#ifdef Q_OS_UNIX
    sh::base::MainThread::dispatcher()->invokeSync([&](){
        struct group *grpp;
        setgrent();
        while (true) {
            grpp = getgrent();
            if (!grpp)
                break;
            result[grpp->gr_gid] = grpp->gr_name;
        }
        endgrent();
    });
#endif
    return result;
}

void sh::filepropertydialogtabs::FilePropertyDialogTabUnixPermissions::doInitialize()
{
    REGISTER_FILEPROPERTYDIALOGTAB(sh::ui::FilePropertyDialogTabFactory::REGISTER_FILEPROPERTYDIALOGTAB_INDEX_CORE+1000, sh::filepropertydialogtabs::FilePropertyDialogTabUnixPermissions);
}

void sh::filepropertydialogtabs::FilePropertyDialogTabUnixPermissions::doShutdown()
{
}

sh::filepropertydialogtabs::FilePropertyDialogTabUnixPermissions::ActionChange::ActionChange(std::shared_ptr<FilePropertyDialogTabUnixPermissions> tab)
            : tab(tab)
{
}

void sh::filepropertydialogtabs::FilePropertyDialogTabUnixPermissions::ActionChange::action(sh::actions::ActionExecutionInfo* info)
{
#ifdef Q_OS_UNIX
    info->setHead(tr("Changing permissions"));
    std::shared_ptr<sh::ui::FilePropertyDialog> _dlg = tab->dialog();
    if (!_dlg)
        return;
    std::shared_ptr<sh::filesystem::FilesystemNode> node = tab->nodes().first();
    QVariant vv = sh::detailcolumns::DetailColumnCustomAttributes::instance()->requestValue(node, info->operation());
    QMap<QString, QString> vals = sh::detailcolumns::DetailColumnCustomAttributes::getMapByValue(vv);
    QStringList vl = vals["unix_permissions"].split(";");
    int flags = vl[0].toInt();
    int uid = vl[1].toInt();
    int gid = vl[2].toInt();
    bool userMayRead = (flags & S_IRUSR);
    bool userMayWrite = (flags & S_IWUSR);
    bool userMayExecute = (flags & S_IXUSR);
    bool groupMayRead = (flags & S_IRGRP);
    bool groupMayWrite = (flags & S_IWGRP);
    bool groupMayExecute = (flags & S_IXGRP);
    bool othersMayRead = (flags & S_IROTH);
    bool othersMayWrite = (flags & S_IWOTH);
    bool othersMayExecute = (flags & S_IXOTH);
    bool sticky = (flags & S_ISVTX);
    bool setuid = (flags & S_ISUID);
    bool setgid = (flags & S_ISGID);
    QMap<int, QString> musers = getAllUsers();
    QMap<int, QString> mgroups = getAllGroups();
    QString currentuser = musers.value(uid);
    QString currentgroup = mgroups.value(gid);
    bool dresult = info->userfeedback()->unixPermissionsDialog(&userMayRead, &userMayWrite, &userMayExecute, &groupMayRead, &groupMayWrite, &groupMayExecute, &othersMayRead, &othersMayWrite, &othersMayExecute, &sticky, &setuid, &setgid, musers.values(), mgroups.values(), &currentuser, &currentgroup);
    if (dresult)
    {
        if (!currentuser.isNull())
            uid = musers.key(currentuser);
        if (!currentgroup.isNull())
            gid = mgroups.key(currentgroup);
        flags = 0;
        if (userMayRead)
            flags |= S_IRUSR;
        if (userMayWrite)
            flags |= S_IWUSR;
        if (userMayExecute)
            flags |= S_IXUSR;
        if (groupMayRead)
            flags |= S_IRGRP;
        if (groupMayWrite)
            flags |= S_IWGRP;
        if (groupMayExecute)
            flags |= S_IXGRP;
        if (othersMayRead)
            flags |= S_IROTH;
        if (othersMayWrite)
            flags |= S_IWOTH;
        if (othersMayExecute)
            flags |= S_IXOTH;
        if (sticky)
            flags |= S_ISVTX;
        if (setuid)
            flags |= S_ISUID;
        if (setgid)
            flags |= S_ISGID;
        QStringList vperm;
        vperm << "unix_permissions";
        vperm << QString("%1;%2;%3!").arg(flags).arg(uid).arg(gid);
        sh::detailcolumns::DetailColumnCustomAttributes::instance()->applyValueByNode(node, info->operation(), vperm);
        tab->refresh();
    }
#endif
}
