/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2020 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef DETAILCOLUMN_H
#define DETAILCOLUMN_H

#include <memory>
#include <QMap>
#include <QMutex>
#include <QVariant>


namespace sh{
    namespace filesystem{
        class FilesystemModelFileviewProxy;
        class FilesystemNode;
        class Operation;
        class Eurl;

        /** \brief
         * Abstract base class for a detail column (on filesystem nodes).
         *
         * Those can e.g. be seen in the file list view, but can also be queried
         * internally by other places in code.
         *
         * It encapsulates the retrieval logic and metadata for one piece of additional
         * information a sh::filesystem::FilesystemNode can have (e.g. the filesize). Retrieving the values
         * is designed to be asynchronous. Each instance represents
         * one column, while the actual logic is implemented in subclasses.
         * For a new detail column, subclass this class and implement at least
         * determineValue.
         */
        class DetailColumn : public std::enable_shared_from_this<DetailColumn>
        {
            friend class sh::filesystem::FilesystemNode;
        public:

            /** \brief
             * The internal unique name.
             */
            QString name() { return _name; }

            /** \brief
             * The display name.
             */
            QString displayName() { return _displayName; }

            /** \brief
             * Checks if a value for this detail is available for one given node.
             */
            bool isValueAvailable(std::shared_ptr<FilesystemNode> node);

            /** \brief
             * Returns the value for this detail for one given node.
             * \param ignoreAged If no value should be returned which is older than the last request (not useful for nearly all cases).
             */
            QVariant value(std::shared_ptr<FilesystemNode> node, bool ignoreAged=false);

            /** \brief
             * A special value expressing that the value is not yet available.
             */
            static inline QVariant VALUE_UNAVAILABLE()
            {
                // safely initialized in main thread at the beginning
                static QVariant result = QVariant::fromValue( new QObject());
                return result;
            }

            /** \brief
             * Returns the stringified value for this details for one given node
             * considering the configuration of a view.
             */
            virtual QString displayValue(std::shared_ptr<FilesystemNode> node, const sh::filesystem::FilesystemModelFileviewProxy* viewmodel);

            /** \brief
             * If this detail shall be a visible column in the view.
             */
            virtual bool isVisible();

            /** \brief
             * Requests to determine the value for this detail for one given node.
             * Blocks until the value is available. Get it with the value method
             * afterwards.
             */
            QVariant requestValue(std::shared_ptr<FilesystemNode> node, sh::filesystem::Operation* op = 0, bool withNodeValues = false, bool withOperationsCache = true);

            /** \brief
             * If sorting should separate files and directories.
             */
            bool sort_doTypediff();

            /** \brief
             * The sorting order.
             */
            virtual int sort_order(std::shared_ptr<FilesystemNode> n1, std::shared_ptr<FilesystemNode> n2);

            /** \brief
             * Controls which place this column should get in the user interface.
             */
            uint displayIndex();

            /**
             * \brief Returns a freshly determined value for this column and the given node.
             *        It neither asks nor populates any caches or storages.
             */
            QVariant computeValue(std::shared_ptr<FilesystemNode> node, sh::filesystem::Operation* op);

            int defaultWidth();

            bool isRightAligned();

            virtual ~DetailColumn() {}

            static const uint DISPLAYINDEX_CORE;
            static const uint DISPLAYINDEX_INTERESTING;
            static const uint DISPLAYINDEX_EXOTIC;

            /**
             * \brief Applies the detail value to a given eurl (without using any caches).
             *
             * Used for transferring some details when a file transfer occurs.
             *
             * Optionally implement the one of the applyValueBy... methods and register the instance with
             * sh::filesystem::FilesystemOperation::addTransferrableDetailColumn. For performance reasons, you
             * should decide to implement applyValueByEurl.
             */
            virtual void applyValueByEurl(std::shared_ptr<const sh::filesystem::Eurl> eurl, sh::filesystem::Operation* op, QVariant value);
            /**
             * \brief Applies the detail value to a given node (without using any caches).
             *
             * Used for transferring some details when a file transfer occurs.
             *
             * Optionally implement the one of the applyValueBy... methods and register the instance with
             * sh::filesystem::FilesystemOperation::addTransferrableDetailColumn. For performance reasons, you
             * should decide to implement applyValueByEurl.
             */
            virtual void applyValueByNode(std::shared_ptr<FilesystemNode> node, sh::filesystem::Operation* op, QVariant value);

            static void registerKnownDetailColumn(QString name, std::shared_ptr<DetailColumn> column);
            static std::shared_ptr<DetailColumn> findKnownDetailColumn(QString name);

        protected:
            /** \brief It is intended to provide implementations as singletons.
             */
            DetailColumn(QString name, QString displayName, uint displayIndex=DISPLAYINDEX_EXOTIC+20000, bool sort_doTypediff=false, int defaultWidth=-1, bool isRightAligned=false);
            /** \brief
             * This method holds the retrieval logic. It returns the actual value
             * for this value for a given node.
             */
            virtual QVariant determineValue(std::shared_ptr<sh::filesystem::FilesystemNode> node, sh::filesystem::Operation* op) = 0;

        private:
            static QMap<QString, std::shared_ptr<DetailColumn>> _knownDetailColumns;
            static QMutex _knownDetailColumnsMutex;
            QString _displayName;
            QString _name;
            uint _displayIndex;
            bool _sort_doTypediff;
            int _defaultWidth;
            bool _isRightAligned;
        };

    }
}

#endif // DETAILCOLUMN_H
