/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2020 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef EURL_H
#define EURL_H

#include <QMap>
#include <QMutex>
#include "exceptions/argumentexception.h"


namespace sh {
    namespace filesystem{

        /** \brief Shallot exception for invalid input in Eurl creation methods.
         */
        class EurlMisformattedException : public sh::exceptions::ArgumentException
        {
        public:
            EurlMisformattedException(sh::exceptions::ExceptionData data);
        };

        /** \brief Filesystem paths.
         *
         * A Extended Uniform Resource Locator is something like a
         * [URL](https://de.wikipedia.org/wiki/Uniform_Resource_Locator).
         *
         * It is more general since it can also be nested. It is something like
         * `zip:/[ftp://ftpserver/foo/bar.zip]//zippedfolder/zippedfile`.
         *
         * You get Eurl instances only from the factory methods. There will never
         * be two instances with the same textual value.
         *
         * Please note: Instances can be associated with any kind of elements in
         * the filesystem (files, directories, links, ...). It might also point to something
         * which does not exist at all.
         * The documentation sometimes explicitly makes a difference
         * between those kinds (files, directories, links, ...; often called 'node type').
         * But it often uses the term 'file' implicitly while meaning all kinds of elements;
         * assuming that e.g. a directory is just a special kind of a file. It should be clear from
         * the particular context which meaning applies.
         */
        class Eurl : public std::enable_shared_from_this<Eurl>
        {
        public:

            /** \brief
             * Constructs a new Eurl by string (what Eurl.asString would return).
             */
            static std::shared_ptr<const Eurl> fromString(QString eurlstring);

            /** \brief
             * Constructs a new Eurl by scheme name, hostname and a path.
             */
            static std::shared_ptr<const Eurl> create(QString scheme, QString hostname, QString path);

            /** \brief
             * Constructs a new Eurl by scheme name, an inner eurl, a hostname and a path.
             */
            static std::shared_ptr<const Eurl> create(QString scheme, const Eurl* inner, QString hostname, QString path);

            /** \brief
             * Returns the textual value. This is what Eurl.fromString would expect as parameter.
             */
            QString asString() const;

            /** \brief
             * Returns the hostname part (from the outer url of this Eurl).
             * Examples:
             * `"livingroom-pc"` for `smb://livingroom-pc/foo/bar/baz`.
             * `""` for `file:///foo/bar/baz`.
             */
            QString hostname() const;

            /** \brief
             * Returns the path part (from the outer url of this Eurl).
             * Examples:
             * `"/foo/bar/baz"` for `smb://livingroom-pc/foo/bar/baz`.
             * `"/foo/bar/baz"` for `file:///foo/bar/baz`.
             * `"/"` for `file:///`.
             */
            QString path() const;

            /** \brief
             * Returns the last path segment. This is the text behind the last slash.
             * Examples:
             * `"baz"` for `file:///foo/bar/baz`.
             * `""` for `file:///`.
             */
            QString basename() const;

            /** \brief
             * Returns the scheme (from the outer url of this shallot.Eurl). This is what comes before the `://`.
             * Example:
             * `"file"` for `file:///foo/bar/baz`.
             */
            QString scheme() const;

            /** \brief
             * Returns a new Eurl containing only the outer part of this one (strips the embeddings).
             * Example:
             * `foobar://host/foo/bar/baz` for `foobar:/[zip:/[file:///a/b/c.zip]//d/e]//foo/bar/baz`.
             */
            std::shared_ptr<const Eurl> outerUrl() const;

            /** \brief
             * Returns a new Eurl containing only the embedding of this one.
             * Example:
             * `zip:/[file:///a/b/c.zip]//d/e` for `foobar:/[zip:/[file:///a/b/c.zip]//d/e]//foo/bar/baz`.
             */
            std::shared_ptr<const Eurl> outermostInnerEurl() const;

            /** \brief
             * Returns a new Eurl from this one with "/basename" appended.
             * The parameter is assumed to be a single path segment.
             */
            std::shared_ptr<const Eurl> withAppendedSegment(QString basename) const;

            /** \brief
             * Returns a new Eurl with path segments "/pa/t/h/" appended.
             * The parameter may contain `"/"`s for dividing path segments.
             */
            std::shared_ptr<const Eurl> withAppendedSegments(QString path) const;

            /** \brief
             * Returns the root Eurl from this one.
             * Example:
             * `zip:/[file:///a/b/c.zip]//` for `zip:/[file:///a/b/c.zip]//foo/bar/baz`.
             */
            std::shared_ptr<const Eurl> root() const;

            /** \brief
             * Returns a new Eurl containing this one packed as embedding and new outer
             * parts `scheme`, `hostname` and `path`.
             * Example:
             * `scheme:/[file:///a/b/c.zip]/hostname/p/a/t/h` for `file:///a/b/c.zip`.
             */
            std::shared_ptr<const Eurl> enwrapWithOuterUrl(QString scheme, QString hostname, QString path) const;

            /** \brief
             * Returns the parent Eurl. At first, this traverses path segments. For a root path
             * eurl with embeddings, it returns the embedding.
             * If none are available, it returns `0`.
             * Examples:
             * `foo://host/foo` for `foo://host/foo/bar`.
             * `foo://host/` for `foo://host/boo`.
             * `foo:/[bar://host/goo]/anotherhost/` for `foo:/[bar://host/goo]/anotherhost/boo`.
             * `bar://host/foo` for `foo:/[bar://host/foo]/host/`.
             * `0` for `foo://host/`.
             */
            std::shared_ptr<const Eurl> parentSegment() const;

            /** \brief
             * Checks if this Eurl has embeddings.
             * Examples:
             * `true` for `foo:/[bar:///foo]/host/`.
             * `false` for `foo://host/`.
             */
            bool hasInnerUrls() const;

            /** \brief
             * Checks if this Eurl is a root path (with or without embeddings).
             * Examples:
             * `true` for `foo://host/`.
             * `false` for `foo://host/a`.
             * `true` for `foo:/[bar:///goo]/host/`.
             * `false` for `foo:/[bar:///goo]/host/a`.
             * `true` for `foo:/[bar:///goo]//`.
             * `false` for `foo:/[bar:///goo]//a`.
             */
            bool outerUrlIsRootDirectory() const;

            /** \brief
             * Checks if this Eurl has a parent segment.
             * This indicates if Eurl.parentSegment would return `0`.
             */
            bool hasParentSegment() const;


            /** \brief
             * Checks if this Eurl is a prefix of another one.
             * This is not an equivalent to a string comparison but
             * it checks parent relationships according to
             * Eurl.parentSegment.
             */
            bool isPrefixOf(std::shared_ptr<const Eurl> longer) const;

            /** \brief
             * Checks if a name is a valid filename. If not, EurlMisformattedException is thrown.
             */
            static void filenameCheck(QString filename);

            /** \brief
             * The character marking the begin of an embedding.
             */
            static const QChar WRAPPER_BEGIN;

            /** \brief
             * The character marking the end of an embedding.
             */
            static const QChar WRAPPER_END;

            /** \brief
             * Characters which are forbidden in filenames.
             */
            static const QString FORBIDDEN_FILENAME_CHARACTERS;

            /** \brief Constructed only by the infrastructure and made available otherwise.
             */
            Eurl(QString eurlstring);
            ~Eurl();

            /** @a_infrastructure
             */
            static void doInitialize();

            /** @a_infrastructure
             */
            static void doShutdown();

        private:
            static QString check_scheme(QString scheme);
            static QString check_inner(QString inner);
            static QString check_hostname(QString hostname);
            static QString check_path(QString path);
            static QString check_filename(QString filename);
            static QString _escape(QString s);
            static QString _unescape(QString s);

            static QHash<QChar, QString> _escapemap;
            static QMutex _escapemapmutex;

            static std::shared_ptr<const Eurl> createNOCHECK(QString scheme, const Eurl *inner, QString hostname, QString path);

            static std::shared_ptr<const Eurl> fromStringNOCHECK(QString eurlstring);
            const QString _eurlstring;
            mutable std::shared_ptr<const Eurl> _cache_parentsegment = 0;
            mutable std::shared_ptr<const Eurl> _cache_outerurl = 0;
            mutable bool _cache_outerurl_isthis = false;
            mutable std::shared_ptr<const Eurl> _cache_outermostinnereurl = 0;
            mutable std::shared_ptr<const Eurl> _cache_root = 0;
            mutable bool _cache_root_isthis = false;
            mutable QString _cache_basename;
            mutable QString _cache_hostname;
            mutable QString _cache_path;
            mutable QString _cache_scheme;
            static QMutex _mutex; // use one single static mutex here, it makes many things simpler :)

            static QHash<QString, std::weak_ptr<const Eurl>> _eurluniverse;

        };

    }
}

#endif // EURL_H
