/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2020 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <QFile>
#include "exceptions/ioexception.h"
#include "search/searchcriterion.h"
#include "base/mainthread.h"
#include "filesystemhandler.h"
#include "filesystemhandlerregister.h"
#include "filesystemmodel.h"
#include "filesystemoperation.h"
#include "operation.h"


sh::filesystem::FilesystemHandler::FilesystemHandler(sh::filesystem::FilesystemModel *model) :
    QObject()
{
    this->_model = model;
}

void sh::filesystem::FilesystemHandler::deleteDirectoryIfEmpty(sh::filesystem::Operation* op, std::shared_ptr<const sh::filesystem::Eurl> eurl)
{
    NOT_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    if (getType(op, eurl) != sh::filesystem::FilesystemNodeType::Directory)
        throw sh::exceptions::IOException(sh::exceptions::ExceptionData().detailsAreInteresting().details(tr("'%1' is not a directory.").arg(eurl->asString())));
    if (!op->filesystem()->itemlist(eurl).isEmpty())
        throw sh::exceptions::IOException(sh::exceptions::ExceptionData().detailsAreInteresting().details(tr("'%1' is not empty.").arg(eurl->asString())));
    deleteItem(op, eurl);
}

QList<QString> sh::filesystem::FilesystemHandler::listExtendedAttributes(sh::filesystem::Operation*, std::shared_ptr<const sh::filesystem::Eurl>)
{
    return QList<QString>();
}

quint64 sh::filesystem::FilesystemHandler::getExtendedAttributeSize(sh::filesystem::Operation*, std::shared_ptr<const sh::filesystem::Eurl>, QString)
{
    return 0;
}

QByteArray sh::filesystem::FilesystemHandler::getExtendedAttribute(sh::filesystem::Operation*, std::shared_ptr<const sh::filesystem::Eurl>, QString)
{
    return QByteArray();
}

void sh::filesystem::FilesystemHandler::setExtendedAttribute(sh::filesystem::Operation*, std::shared_ptr<const sh::filesystem::Eurl>, QString, QByteArray)
{
}

void sh::filesystem::FilesystemHandler::removeExtendedAttribute(sh::filesystem::Operation*, std::shared_ptr<const sh::filesystem::Eurl>, QString)
{
}

sh::filesystem::FilesystemModel* sh::filesystem::FilesystemHandler::model()
{
    return _model;
}

void sh::filesystem::FilesystemHandler::configureItems(sh::filesystem::Operation*, QList<std::shared_ptr<sh::filesystem::FilesystemNode>>)
{
}

QMap<QString, QString> sh::filesystem::FilesystemHandler::getCustomAttributes(sh::filesystem::Operation*, std::shared_ptr<const sh::filesystem::Eurl>)
{
    return QMap<QString, QString>();
}

void sh::filesystem::FilesystemHandler::setCustomAttribute(sh::filesystem::Operation*, std::shared_ptr<const sh::filesystem::Eurl>, QString, QString)
{
}

bool sh::filesystem::FilesystemHandler::isWellKnownScheme()
{
    return false;
}

void sh::filesystem::FilesystemHandler::search(sh::filesystem::Operation* op, std::shared_ptr<const sh::filesystem::Eurl> eurl, std::function<void(std::shared_ptr<const sh::filesystem::Eurl>, QList<std::shared_ptr<sh::search::SearchCriterion>>, sh::filesystem::FilesystemNodeType ftype)> addfile, std::function<void(std::shared_ptr<const sh::filesystem::Eurl>)> visitdir, QList<std::shared_ptr<sh::search::SearchCriterion>> criteria)
{
    foreach(std::shared_ptr<const sh::filesystem::Eurl> item, op->filesystem()->itemlist(eurl))
    {
        sh::filesystem::FilesystemNodeType t = op->filesystem()->getType(item);
        if (t == sh::filesystem::FilesystemNodeType::Directory)
            visitdir(item);
        else
            addfile(item, criteria, t);
    }
}

void sh::filesystem::FilesystemHandler::customizeUi(std::shared_ptr<sh::filesystem::FilesystemNode>, bool*)
{
}

void sh::filesystem::FilesystemHandler::HandlerTransfer::respectCancel()
{
}

void sh::filesystem::FilesystemHandler::HandlerTransfer::incrementTransferredBytes(qint64)
{
}

sh::filesystem::FilesystemHandler::HandlerTransfer::~HandlerTransfer()
{
}
