/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2020 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <functional>
#include "filesystemmodel.h"
#include <QCoreApplication>
#include <QStack>
#include "exceptions/ioexception.h"
#include "exceptions/runtimeexception.h"
#include "exceptions/argumentexception.h"
#include "tools/benchmarking.h"
#include "base/mainthread.h"
#include "base/threadpool.h"
#include "ui/mainwindow.h"
#include "detailcolumn.h"
#include "filesystemhandler.h"
#include "filesystemhandlerregister.h"
#include "filesystemnode.h"
#include "filesystemmodeldirectorytreeproxy.h"
#include "filesystemmodelfileviewproxy.h"
#include "filesystemmodelsubtreeproxy.h"
#include "operation.h"

REGISTER_SINGLETON(sh::filesystem, FilesystemModel, "filesystem")


sh::filesystem::FilesystemModel::FilesystemModel() :
    QAbstractItemModel(), _nodeDataMutex(QMutex::Recursive)
{
    rootnode = std::shared_ptr<sh::filesystem::FilesystemNode>(new sh::filesystem::FilesystemNode(this));
    rootnode->_isRootNode = true;
    _subitemFetchingStateChanged(rootnode, false);

    addOpenNodeHelper(1000, [=](std::shared_ptr<const sh::filesystem::Eurl> eurl){ // helps opening a new node by openRootEurl
        QList<std::shared_ptr<sh::filesystem::FilesystemNode>> result;
        if (eurl->outerUrlIsRootDirectory() && eurl->hasInnerUrls())
            result = openRootEurl(eurl);
        return result;
    });

    addOpenNodeHelper(7000, [=](std::shared_ptr<const sh::filesystem::Eurl> eurl){ // helps opening a new node by refreshData
        refreshData(eurl);
        return tryGetNodesForEurl(eurl);
    });

#ifdef BENCHMARKING
    sh::tools::Benchmarking::instance()->addReporter([=](){
        QMap<QString, QString> result;
        int allnodes = 0, deadnodes = 0;
        QMutexLocker lock(&eurl2nodemutex);
        for(auto iter=eurl2node.begin(); iter!=eurl2node.end(); iter++)
        {
            auto node = iter->lock();
            if (node) {
                allnodes++;
                if (!node->isAlive())
                    deadnodes++;
            }
        }
        result.insert("FilesystemModel node count", QString("%1 (%2 dead)").arg(allnodes).arg(deadnodes));
        return result;
    });
#endif
}

sh::filesystem::FilesystemModel::~FilesystemModel()
{
}

QVariant sh::filesystem::FilesystemModel::data(const QModelIndex &index, int role) const
{
    CHECK_CORRECT_CALLER_THREAD
    QVariant result;
    if (index.isValid()){
        sh::filesystem::FilesystemNode *item = static_cast<sh::filesystem::FilesystemNode*>(index.internalPointer());
        if (index.column()!=0)
        {
            // Index column != 0 must be implemented by a proxy model.
            result = QVariant();
        }
        else if (role == Qt::DisplayRole)
        {
            item->requestDetails(false);
            result = QVariant(item->displayName());
        }
        else if (role == Qt::DecorationRole)
            result = item->icon();
    }
    return result;
}

Qt::ItemFlags sh::filesystem::FilesystemModel::flags(const QModelIndex &index) const
{
    CHECK_CORRECT_CALLER_THREAD
    if (!index.isValid())
        return 0;
    if (!static_cast<sh::filesystem::FilesystemNode*>(index.internalPointer())->eurl()) // if this is the loading placeholder
        return 0;
    return Qt::ItemIsEnabled | Qt::ItemIsSelectable;
}

QVariant sh::filesystem::FilesystemModel::headerData(int, Qt::Orientation orientation, int role) const
{
    CHECK_CORRECT_CALLER_THREAD
    if (orientation == Qt::Horizontal && role == Qt::DisplayRole)
        return QVariant(tr("Name"));
    else
        return QVariant();
}

QModelIndex sh::filesystem::FilesystemModel::index(int row, int column, const QModelIndex &parent) const
{
    CHECK_CORRECT_CALLER_THREAD
    if (!hasIndex(row, 0, parent))
         return QModelIndex();
    std::shared_ptr<sh::filesystem::FilesystemNode> parentItem = getNodeForIndex(parent);
    std::shared_ptr<sh::filesystem::FilesystemNode> childItem = 0;
    int childcount = parentItem->childnodeCount();
    if (row < childcount)
        childItem = parentItem->childnode(row);
    else if (row == childcount)
        childItem = parentItem->_loadnode; // just fyi: might also be 0 if we aren't loading
    if (childItem)
        return createIndex(row, column, childItem.get());
    else
        return QModelIndex();
}

QModelIndex sh::filesystem::FilesystemModel::parent(const QModelIndex &index) const
{
    CHECK_CORRECT_CALLER_THREAD
    if (!index.isValid())
        return QModelIndex();
    std::shared_ptr<sh::filesystem::FilesystemNode> parentItem = static_cast<sh::filesystem::FilesystemNode*>(index.internalPointer())->parentnode();
    if (parentItem == rootnode)
        return QModelIndex();
    return createIndex(parentItem->index(), 0, parentItem.get());
}

int sh::filesystem::FilesystemModel::rowCount(const QModelIndex &parent) const
{
    CHECK_CORRECT_CALLER_THREAD
    std::shared_ptr<sh::filesystem::FilesystemNode> parentItem = getNodeForIndex(parent);
    int res = parentItem->childnodeCount();
    if (parentItem->isFetchingSubitems())
        res++;
    return res;
}

int sh::filesystem::FilesystemModel::columnCount(const QModelIndex& parent) const
{
    CHECK_CORRECT_CALLER_THREAD
    return 1 + getNodeForIndex(parent)->_childrenDetailCols.length();
}

void sh::filesystem::FilesystemModel::_subitemFetchingStateChanged(std::shared_ptr<sh::filesystem::FilesystemNode> node, bool value)
{
    CHECK_CORRECT_CALLER_THREAD
    if ( (!node->_loadnode) == value)
    {
        bool isAliveAndInModel = node->parentnode() != 0;
        if (isAliveAndInModel)
        {
            QModelIndex idx = getIndexForNode(node);
            int c = node->childnodeCount();
            if (value)
                beginInsertRows(idx, c, c);
            else
                beginRemoveRows(idx, c, c);
        }
        node->_loadnode = value ? std::make_shared<LoadOnDemandPlaceholderFilesystemNode>(node) : 0;
        if (isAliveAndInModel)
        {
            if (value)
                endInsertRows();
            else
                endRemoveRows();
        }
    }
}

sh::filesystem::FilesystemModelFileviewProxy *sh::filesystem::FilesystemModel::createFileviewProxy(QModelIndex root, bool withtooltip,
                                                                   std::function<void(std::shared_ptr<sh::filesystem::FilesystemNode>)> onBecomesInvalid)
{
    CHECK_CORRECT_CALLER_THREAD
    sh::filesystem::FilesystemModelFileviewProxy* proxy = new sh::filesystem::FilesystemModelFileviewProxy(root, withtooltip);
    sh::filesystem::FilesystemModelSubtreeProxy* pproxy = new sh::filesystem::FilesystemModelSubtreeProxy(root, proxy);
    proxy->setSourceModel(pproxy);
    connect(pproxy, &sh::filesystem::FilesystemModelSubtreeProxy::becomesInvalid, onBecomesInvalid);
    return proxy;
}

QModelIndex sh::filesystem::FilesystemModel::getIndexForNode(std::shared_ptr<sh::filesystem::FilesystemNode> node)
{
    CHECK_CORRECT_CALLER_THREAD
    BENCHMARK_SCOPE("FilesystemModel::getIndexForNode")
    QStack<std::shared_ptr<sh::filesystem::FilesystemNode>> h;
    std::shared_ptr<sh::filesystem::FilesystemNode> s = node;
    QModelIndex result;
    while (s != rootnode)
    {
        h.push(s);
        s = s->parentnode();
        if (!s)
            return result;  // the node isn't alive anymore
    }
    while (!h.isEmpty()) {
        std::shared_ptr<sh::filesystem::FilesystemNode> n = h.pop();
        result = index(n->index(), 0, result);
    }
    return result;
}

void sh::filesystem::FilesystemModel::addOpenNodeHelper(int index, std::function<QList<std::shared_ptr<sh::filesystem::FilesystemNode>>(std::shared_ptr<const sh::filesystem::Eurl>)> openNodeHelper)
{
    QMutexLocker lock(&_openNodeHelpersMutex);
    _openNodeHelpersMap.insertMulti(index, openNodeHelper);
    QList<int> _keys = _openNodeHelpersMap.keys();
    qSort(_keys);
    _openNodeHelpers.clear();
    foreach(int key, _keys)
        _openNodeHelpers.append(_openNodeHelpersMap.values(key));
}

QList<std::shared_ptr<sh::filesystem::FilesystemNode>> sh::filesystem::FilesystemModel::findNodesForEurl(std::shared_ptr<const sh::filesystem::Eurl> eurl)
{
    BENCHMARK_SCOPE("FilesystemModel::findNodesForEurl")
    return _findNodesForEurl_helper(eurl);
}

QList<std::shared_ptr<sh::filesystem::FilesystemNode>> sh::filesystem::FilesystemModel::_findNodesForEurl_helper(std::shared_ptr<const sh::filesystem::Eurl> eurl)
{
    NOT_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    QList<std::shared_ptr<sh::filesystem::FilesystemNode>> result = tryGetNodesForEurl(eurl);
    if (result.isEmpty())
    {
        if (eurl == rootNode()->eurl())
            result = QList<std::shared_ptr<sh::filesystem::FilesystemNode>>({rootNode()});
        else
        {
            if (eurl->hasParentSegment())
                _findNodesForEurl_helper(eurl->parentSegment());
            QList<std::function<QList<std::shared_ptr<sh::filesystem::FilesystemNode>>(std::shared_ptr<const sh::filesystem::Eurl>)> > openNodeHelpers;
            {
                QMutexLocker _l1(&_openNodeHelpersMutex);
                openNodeHelpers = _openNodeHelpers;
            }
            foreach(std::function<QList<std::shared_ptr<sh::filesystem::FilesystemNode>>(std::shared_ptr<const sh::filesystem::Eurl>)> fct, openNodeHelpers)
            {
                result = fct(eurl);
                if (!result.isEmpty())
                    break;
            }
        }
    }
    return result;
}

QList<QPersistentModelIndex> sh::filesystem::FilesystemModel::findIndexesForEurl(std::shared_ptr<const sh::filesystem::Eurl>eurl)
{
    NOT_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    BENCHMARK_SCOPE("FilesystemModel::findIndexesForEurl")
    QList<std::shared_ptr<sh::filesystem::FilesystemNode>> nodes = findNodesForEurl(eurl);
    QList<QPersistentModelIndex> result;
    sh::base::MainThread::dispatcher()->invokeSync([&](){
        foreach(std::shared_ptr<sh::filesystem::FilesystemNode> node, nodes)
            result.append(QPersistentModelIndex(getIndexForNode(node)));
    });
    return result;
}

std::shared_ptr<sh::filesystem::FilesystemNode> sh::filesystem::FilesystemModel::getOrCreateFilesystemNode(std::shared_ptr<const sh::filesystem::Eurl> neweurl, sh::filesystem::FilesystemHandler* handler,
                                     sh::filesystem::FilesystemNodeType nodetype, bool isHidden, std::shared_ptr<sh::filesystem::FilesystemNode> parent, bool doinsert, bool showInitialLoadingLabel, bool* pIsNew)
{
    CHECK_CORRECT_CALLER_THREAD
    //check if there is already such a node in the model
    QList<std::shared_ptr<sh::filesystem::FilesystemNode>> nodes = tryGetNodesForEurl(neweurl);
    std::shared_ptr<sh::filesystem::FilesystemNode> node = 0;
    foreach(std::shared_ptr<sh::filesystem::FilesystemNode> _n, nodes)
    {
        if (_n->parentnode() == parent)
        {
            node = _n;
            break;
        }
    }
    if (node) // if we found one
    {
        if (pIsNew)
            (*pIsNew) = false;
        bool changetype;
        {
            QMutexLocker _l1(&_nodeDataMutex);
            changetype = (node->_nodetype != nodetype);
            if (changetype)
            {
                node->_nodetype = nodetype;
                node->_detailvalues.clear();
                node->_last_detailvalues.clear();
                node->_detailsRequested = false;
                node->_initialize(true);
            }
        }
        if (changetype) // do we have to change its type
        {
            std::shared_ptr<sh::filesystem::FilesystemNode> pnode = node->parentnode();
            if (pnode)
            {
                pnode->removeChild(node);
                node->_isConfigured = false;
                pnode->addChild(node);
            }
        }
        return node;
    }
    else
    {
        if (pIsNew)
            (*pIsNew) = true;
        return createFilesystemNode(neweurl, handler, nodetype, isHidden, parent, doinsert, showInitialLoadingLabel);
    }
}

std::shared_ptr<sh::filesystem::FilesystemNode> sh::filesystem::FilesystemModel::createFilesystemNode(std::shared_ptr<const sh::filesystem::Eurl>eurl,
                                             sh::filesystem::FilesystemHandler* handler, sh::filesystem::FilesystemNodeType nodetype, bool isHidden,
                                             std::shared_ptr<sh::filesystem::FilesystemNode> parent, bool doinsert, bool showInitialLoadingLabel
                                        )
{
    CHECK_CORRECT_CALLER_THREAD
    QMutexLocker lock(&eurl2nodemutex);
    std::shared_ptr<sh::filesystem::FilesystemNode> result;
    std::shared_ptr<sh::filesystem::FilesystemNode> existingnode = 0;
    if (parent) // if we create a real model node
    {
        auto iter = eurl2node.find(eurl); // search if this node already exists (due to references to it from somewhere, but must not already be in the model)
        while (iter != eurl2node.end() && iter.key() == eurl)
        {
            std::shared_ptr<sh::filesystem::FilesystemNode> siter = iter->lock();
            if (siter && siter->_origparent == parent)
            {
                existingnode = siter;
                break;
            }
            iter++;
        }
    }
    if (existingnode) // if there already is such a node
    {
        result = existingnode;
        {
            QMutexLocker _l1(&_nodeDataMutex);
#ifdef QT_DEBUG
            if (result->_parent)
                sh::exceptions::Exception::executeGuarded([&](){
                    throw sh::exceptions::ProgramException(sh::exceptions::ExceptionData().details("Creating a node which already exists: " + eurl->asString()));
                });
#endif
            result->_handler = handler;
            result->_nodetype = nodetype;
            result->_ishidden = isHidden;
            result->_childrenDetailCols.clear();// some of those lists are already empty in most cases now, but we want to be sure :)
            result->_detailvalues.clear();
            result->_last_detailvalues.clear();
            result->_linkdestinations.clear();
            result->_childnodes->clear();
            result->_initialize(true);
        }
        // some other stuff was already cleared in FilesystemNode::unmount
    }
    else
    {
        QString displayname = eurl->basename();
        if (displayname.isEmpty())
            displayname = "/";
        result = std::make_shared<sh::filesystem::FilesystemNode>(this, eurl, displayname, handler, nodetype, isHidden);
        result->_origparent = parent;
        if (parent) // if we create a real model node
        {
            std::weak_ptr<sh::filesystem::FilesystemNode> _weakresult = result;
            eurl2node.insertMulti(eurl, _weakresult);
            connect(result.get(), &QObject::destroyed, this, [=](){
                QMutexLocker lock(&eurl2nodemutex);
                auto i = eurl2node.find(eurl);
                bool removed = false;
                while (i != eurl2node.end() && i.key() == eurl) {
                    if (!i.value().owner_before(_weakresult) && !_weakresult.owner_before(i.value()))
                    {
                        eurl2node.erase(i);
                        removed = true;
                        break;
                    }
                    i++;
                }
                if (!removed)
                    sh::exceptions::Exception::executeGuarded([&](){
                        throw sh::exceptions::ProgramException(sh::exceptions::ExceptionData().details("Internal error in a node data structure."));
                    });
            }, Qt::DirectConnection);
        }
    }
    if (parent && doinsert)
        parent->addChild(result);
    _subitemFetchingStateChanged(result, showInitialLoadingLabel);
    return result;
}

QList<std::shared_ptr<sh::filesystem::FilesystemNode>> sh::filesystem::FilesystemModel::tryGetNodesForEurl(std::shared_ptr<const sh::filesystem::Eurl> eurl)
{
    QMutexLocker lock(&eurl2nodemutex);
    auto iter = eurl2node.find(eurl);
    QList<std::shared_ptr<sh::filesystem::FilesystemNode>> result;
    while (iter != eurl2node.end() && iter.key() == eurl)
    {
        std::shared_ptr<sh::filesystem::FilesystemNode> siter = iter->lock();
        if (siter && siter->isAlive())
            result.append(siter);
        iter++;
    }
    return result;
}

void sh::filesystem::FilesystemModel::refreshData(std::shared_ptr<const sh::filesystem::Eurl> eurl, bool forceFindParent, bool withDetails)
{
    sh::exceptions::Exception::executeGuarded([&](){
        if (!eurl->hasParentSegment())
        {
            // since we mostly check own existence here, for the root directory is nothing to do...
        }
        else
        {
            QList<std::shared_ptr<sh::filesystem::FilesystemNode>> parents;
            if (forceFindParent)
                parents = findNodesForEurl(eurl->parentSegment());
            else
                parents = tryGetNodesForEurl(eurl->parentSegment());
            Operation op;
            std::shared_ptr<sh::filesystem::FilesystemHandler> handler = sh::filesystem::FilesystemHandlerRegister::instance()->findHandler(eurl->scheme());
            if (handler)
            {
                sh::filesystem::FilesystemNodeType type = handler->getType(&op, eurl);
                foreach(std::shared_ptr<sh::filesystem::FilesystemNode> parent, parents)
                {
                    sh::base::MainThread::dispatcher()->invokeSync([&](){
                        QList<std::shared_ptr<sh::filesystem::FilesystemNode>> mynodes = tryGetNodesForEurl(eurl);
                        std::shared_ptr<sh::filesystem::FilesystemNode> mynode = 0;
                        foreach (std::shared_ptr<sh::filesystem::FilesystemNode> _n, mynodes)
                        {
                            if (_n->parentnode() == parent)
                            {
                                mynode = _n;
                                break;
                            }
                        }
                        if (type == sh::filesystem::FilesystemNodeType::NONE && mynode)
                            parent->removeChild(mynode);
                        else if (type != sh::filesystem::FilesystemNodeType::NONE && type != sh::filesystem::FilesystemNodeType::Invalid && !mynode)
                            createFilesystemNode(eurl, handler.get(), type, false, parent);
                        else if (mynode && withDetails)
                            mynode->requestDetails(true);
                    });
                }
            }
        }
    }, sh::exceptions::ExecuteGuardFlag::ALL_ERRORS_KILL_SHALLOT);
}

QList<std::shared_ptr<sh::filesystem::FilesystemNode> > sh::filesystem::FilesystemModel::openRootEurl(std::shared_ptr<const sh::filesystem::Eurl> eurl)
{
    NOT_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    QList<std::shared_ptr<sh::filesystem::FilesystemNode> > result;
    sh::exceptions::Exception::executeGuarded([&](){
        std::shared_ptr<sh::filesystem::FilesystemHandler> handler = 0;
        std::shared_ptr<const sh::filesystem::Eurl> neurl = 0;
        QHash<std::shared_ptr<sh::filesystem::FilesystemNode>, std::shared_ptr<sh::filesystem::FilesystemNode>> parentnodes;
        sh::base::MainThread::dispatcher()->invokeSync([&](){
            handler = sh::filesystem::FilesystemHandlerRegister::instance()->findHandler(eurl->scheme());
            if (!handler)
                throw sh::exceptions::ArgumentException(sh::exceptions::ExceptionData().details(QString("No handler for '%1' found.").arg(eurl->scheme())).detailsAreInteresting());
            foreach (std::shared_ptr<sh::filesystem::FilesystemNode> _peernode, tryGetNodesForEurl(eurl->outermostInnerEurl()))
            {
                std::shared_ptr<sh::filesystem::FilesystemNode> _parentnode = _peernode->parentnode();
                if (_parentnode)
                    parentnodes[_parentnode] = _peernode;
            }
            if (parentnodes.isEmpty())
                throw sh::exceptions::ArgumentException(sh::exceptions::ExceptionData().details(QString("No internal node for '%1'.").arg(eurl->outermostInnerEurl()->asString())));
            neurl = eurl->root();
            result = tryGetNodesForEurl(neurl);
        });
        if (!result.isEmpty())
            return;

        // test the access
        Operation op;
        handler->getType(&op, neurl);

        //create a new node
        sh::base::MainThread::dispatcher()->invokeSync([&](){
            result = sh::filesystem::FilesystemModel::instance()->tryGetNodesForEurl(neurl);
            if (!result.isEmpty()) // meanwhile there already is such a node
                return;

            foreach(std::shared_ptr<sh::filesystem::FilesystemNode> _parentnode, parentnodes.keys())
            {
                std::shared_ptr<sh::filesystem::FilesystemNode> _peernode = parentnodes[_parentnode];
                std::shared_ptr<sh::filesystem::FilesystemNode> rnode = createFilesystemNode(neurl, handler.get(), sh::filesystem::FilesystemNodeType::SpecialTreeOnlyDirectory, false, _parentnode);
                rnode->setDisplayName(eurl->outermostInnerEurl()->basename());
                std::weak_ptr<sh::filesystem::FilesystemNode> wresult = rnode;
                sh::filesystem::FilesystemNode* __parentnode = _parentnode.get();
                connect(_peernode.get(), &sh::filesystem::FilesystemNode::removed, rnode.get(), [=](){
                    sh::base::MainThread::dispatcher()->invokeAsync([=](){ //remove must be async here
                        std::shared_ptr<sh::filesystem::FilesystemNode> _result = wresult.lock();
                        if (_result && _result->isAlive())
                            __parentnode->removeChild(_result);
                    });
                });
                result.append(rnode);
            }
        });
    }, sh::exceptions::ExecuteGuardFlag::ALL_ERRORS_KILL_SHALLOT);
    return result;
}

std::shared_ptr<sh::filesystem::FilesystemNode> sh::filesystem::FilesystemModel::getNodeForIndex(const QModelIndex index) const
{
    CHECK_CORRECT_CALLER_THREAD
    if (!index.isValid())
        return rootnode;
    return (static_cast<sh::filesystem::FilesystemNode*>(index.internalPointer()))->shared_from_this();
}

std::shared_ptr<sh::filesystem::FilesystemNode> sh::filesystem::FilesystemModel::rootNode()
{
    return rootnode;
}

