/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2020 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "filesystem/filesystemmodel.h"
#include "base/mainthread.h"
#include "filesystemmodeldirectorytreeproxy.h"

REGISTER_SINGLETON(sh::filesystem, FilesystemModelDirectoryTreeProxy, "filesystem")


void sh::filesystem::FilesystemModelDirectoryTreeProxy::doInitialize()
{
    instance()->setSourceModel(sh::filesystem::FilesystemModel::instance().get());
}

sh::filesystem::FilesystemModelDirectoryTreeProxy::FilesystemModelDirectoryTreeProxy()
{
    this->sort(0, Qt::AscendingOrder);
}

bool sh::filesystem::FilesystemModelDirectoryTreeProxy::filterAcceptsRow (int row, const QModelIndex & parent) const
{
    CHECK_CORRECT_CALLER_THREAD
    if (!parent.isValid())
        return true;
    std::shared_ptr<sh::filesystem::FilesystemNode> node = sh::filesystem::FilesystemModel::instance()->getNodeForIndex(parent);
    if (row < node->childnodeCount())
    {
        std::shared_ptr<FilesystemNode> childnode = node->childnode(row);
        int type = childnode->nodetype();
        bool result = true;
        result = (type == sh::filesystem::FilesystemNodeType::Directory || type == sh::filesystem::FilesystemNodeType::SpecialTreeOnlyDirectory);
        if (!result)
        {
            if (!childnode->linkDestinationNodes().isEmpty())
            {
                type = childnode->linkDestinationNodes().first()->nodetype();
                result = (type == sh::filesystem::FilesystemNodeType::Directory || type == sh::filesystem::FilesystemNodeType::SpecialTreeOnlyDirectory);
            }
        }
        if (result) {
            if (!childnode->isHidden())
                return true;
            auto ifvnode=forcedvisible.begin();
            while (ifvnode!=forcedvisible.end()) {
                auto sifvnode = (*ifvnode).lock();
                if (!sifvnode)
                    ifvnode = forcedvisible.erase(ifvnode);
                else if (sifvnode == childnode)
                    return true;
                else
                    ifvnode++;
            }
        }
        return false;
    }
    else
        return true;
}

bool sh::filesystem::FilesystemModelDirectoryTreeProxy::filterAcceptsColumn (int column, const QModelIndex&  ) const
{
    CHECK_CORRECT_CALLER_THREAD
    return (column == 0);
}

bool sh::filesystem::FilesystemModelDirectoryTreeProxy::lessThan (const QModelIndex & left, const QModelIndex & right ) const
{
    CHECK_CORRECT_CALLER_THREAD
    if (!left.isValid())
            return true;
    if (!right.isValid())
        return false;
    std::shared_ptr<sh::filesystem::FilesystemNode> nleft = sh::filesystem::FilesystemModel::instance()->getNodeForIndex(left);
    std::shared_ptr<sh::filesystem::FilesystemNode> nright = sh::filesystem::FilesystemModel::instance()->getNodeForIndex(right);

    int typediff = (int)nright->nodetype() - (int)nleft->nodetype();
    if (typediff!=0) return (typediff<0);

    return QSortFilterProxyModel::lessThan(left, right);
}

void sh::filesystem::FilesystemModelDirectoryTreeProxy::enforceVisibility(std::shared_ptr<FilesystemNode> node)
{
    CHECK_CORRECT_CALLER_THREAD
    forcedvisible.append(node);
    invalidateFilter();
}

void sh::filesystem::FilesystemModelDirectoryTreeProxy::deEnforceVisibility(std::shared_ptr<FilesystemNode> node)
{
    CHECK_CORRECT_CALLER_THREAD
    auto i = forcedvisible.begin();
    while(i != forcedvisible.end()) {
        auto n = (*i).lock();
        if (!n || (n == node))
            i = forcedvisible.erase(i);
        else
            i++;
    }
    invalidateFilter();
}
