/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2020 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "base/threaddispatcher.h"
#include "tools/thumbnailmanager.h"
#include "detailcolumn.h"
#include "filesystemmodelfileviewproxy.h"
#include "filesystemmodelsubtreeproxy.h"
#include "filesystemmodel.h"
#include "filesystemnode.h"


sh::filesystem::FilesystemModelFileviewProxy::FilesystemModelFileviewProxy(QModelIndex root, bool withtooltip, QObject *parent) :
    QSortFilterProxyModel(parent), _sizeFormattingMode(SizeFormattingModePrefixes), _root(root), _withtooltip(withtooltip)
{
    QObject::connect(sh::tools::ThumbnailManager::instance().get(), &sh::tools::ThumbnailManager::thumbnailAvailable, this, [=](std::shared_ptr<sh::filesystem::FilesystemNode> _n){
        QModelIndex si = sh::filesystem::FilesystemModel::instance()->getIndexForNode(_n);
        if (si.isValid()) {
            QSortFilterProxyModel* sopromo = qobject_cast<QSortFilterProxyModel*>(sourceModel());
            sh::filesystem::FilesystemModelSubtreeProxy* fimosupro = qobject_cast<sh::filesystem::FilesystemModelSubtreeProxy*>(sourceModel());
            QModelIndex interm = sopromo ? sopromo->mapFromSource(si) : (fimosupro ? fimosupro->mapFromSource(si) : si);
            QModelIndex di = mapFromSource(interm);
            if (di.isValid())
                emit dataChanged(di, di);
        }
    });

}

void sh::filesystem::FilesystemModelFileviewProxy::triggerReloadData()
{
    CHECK_CORRECT_CALLER_THREAD
    emit dataChanged(index(0, 0), index(rowCount(), columnCount()));
}

QVariant sh::filesystem::FilesystemModelFileviewProxy::data(const QModelIndex &index, int role) const
{
    CHECK_CORRECT_CALLER_THREAD
    QModelIndex mi = mapToSource(index);
    if (role == Qt::DecorationRole) {
        if (index.isValid() && index.column() == 0) {
            sh::filesystem::FilesystemNode* item = static_cast<sh::filesystem::FilesystemNode*>(mi.internalPointer());
            if (_thumbnailmode) {
                QIcon result;
                bool outdated;
                bool refreshRequested;
                std::shared_ptr<sh::filesystem::FilesystemNode> sitem = item->shared_from_this();
                if (!sh::tools::ThumbnailManager::instance()->getThumbnail(sitem, _thumbnailsize, _thumbnailsize, &result, &outdated, &refreshRequested) || (outdated&&!refreshRequested))
                    sh::tools::ThumbnailManager::instance()->requestThumbnail(sitem, _thumbnailsize, _thumbnailsize);
                if (result.isNull())
                    result = item->icon();
                return result;
            }
            else
                return item->icon();
        }
        return QVariant();
    }
    else if (_withtooltip && role == Qt::ToolTipRole)
        return sourceModel()->data(mi, Qt::DisplayRole);
    else
        return sourceModel()->data(mi, role);
}

void sh::filesystem::FilesystemModelFileviewProxy::setThumbnail(bool enabled, int size)
{
    CHECK_CORRECT_CALLER_THREAD
    _thumbnailmode = enabled;
    _thumbnailsize = size;
}

bool sh::filesystem::FilesystemModelFileviewProxy::getThumbnailEnabled()
{
    CHECK_CORRECT_CALLER_THREAD
    return _thumbnailmode;
}

int sh::filesystem::FilesystemModelFileviewProxy::getThumbnailSize()
{
    CHECK_CORRECT_CALLER_THREAD
    return _thumbnailsize;
}

void sh::filesystem::FilesystemModelFileviewProxy::setSizeFormattingMode(SizeFormatting mode)
{
    CHECK_CORRECT_CALLER_THREAD
    _sizeFormattingMode = mode;
    triggerReloadData();
}

sh::filesystem::SizeFormatting sh::filesystem::FilesystemModelFileviewProxy::getSizeFormattingMode() const
{
    return _sizeFormattingMode;
}

void sh::filesystem::FilesystemModelFileviewProxy::setHiddenFilesVisible(bool v)
{
    _hiddenFilesVisible = v;
    invalidateFilter();
}

bool sh::filesystem::FilesystemModelFileviewProxy::getHiddenFilesVisible() const
{
    return _hiddenFilesVisible;
}

bool sh::filesystem::FilesystemModelFileviewProxy::filterAcceptsRow (int row, const QModelIndex & parent) const
{
    CHECK_CORRECT_CALLER_THREAD
    std::shared_ptr<sh::filesystem::FilesystemNode> node;
    if (parent.isValid())
        node = sh::filesystem::FilesystemModel::instance()->getNodeForIndex(parent);
    else
        // yes, this is a dirty hack :-/
        node = sh::filesystem::FilesystemModel::instance()->getNodeForIndex(_root);
    if (row < node->childnodeCount()) {
        std::shared_ptr<sh::filesystem::FilesystemNode> cn = node->childnode(row);
        if (cn->isHidden() && !_hiddenFilesVisible)
            return false;
        if (cn->nodetype() == sh::filesystem::FilesystemNodeType::SpecialTreeOnlyDirectory)
            return false;
        return true;
    }
    else
        return true;
}

bool sh::filesystem::FilesystemModelFileviewProxy::lessThan (const QModelIndex & left, const QModelIndex & right ) const
{
    CHECK_CORRECT_CALLER_THREAD
    if (!left.isValid()) return true;
    if (!right.isValid()) return false;
    std::shared_ptr<sh::filesystem::FilesystemNode> nleft = sh::filesystem::FilesystemModel::instance()->getNodeForIndex(left);
    std::shared_ptr<sh::filesystem::FilesystemNode> nright = sh::filesystem::FilesystemModel::instance()->getNodeForIndex(right);
    int col = sortColumn();
    std::shared_ptr<sh::filesystem::DetailColumn> detcol = 0;
    bool doTypediff=true;
    if (col>0) {
        int idx = col-1;
        if (nleft->parentnode()->_childrenDetailCols.length()>idx) {
            detcol = nleft->parentnode()->_childrenDetailCols[idx];
            doTypediff = detcol->sort_doTypediff();
        }
    }
    if (doTypediff) {
        std::shared_ptr<sh::filesystem::FilesystemNode> absright = nright;
        std::shared_ptr<sh::filesystem::FilesystemNode> absleft = nleft;
        if (absright->nodetype()==sh::filesystem::FilesystemNodeType::Link && !absright->linkDestinationNodes().isEmpty())
            absright = absright->linkDestinationNodes().first();
        if (absleft->nodetype()==sh::filesystem::FilesystemNodeType::Link && !absleft->linkDestinationNodes().isEmpty())
            absleft = absleft->linkDestinationNodes().first();
        int righttype = absright->nodetype();
        int lefttype = absleft->nodetype();
        if (righttype==sh::filesystem::FilesystemNodeType::Link)
            righttype=sh::filesystem::FilesystemNodeType::File;
        if (lefttype==sh::filesystem::FilesystemNodeType::Link)
            lefttype=sh::filesystem::FilesystemNodeType::File;
        int typediff = (int)righttype - (int)lefttype;
        if (sortOrder()==Qt::DescendingOrder)
            typediff *= -1;
        if (typediff!=0) return (typediff<0);
    }
    if (col == 0)
        return nleft->displayName() < nright->displayName();
    else if (detcol != 0) {
        int comp = detcol->sort_order(nleft, nright);
        if (comp==0) {
            QString leftdn = nleft->displayName();
            QString rightdn = nright->displayName();
            if (leftdn!=rightdn)
                comp = (leftdn < rightdn) ? -1 : 1;
            else
                comp = nleft->index() - nright->index();
            if (sortOrder()==Qt::DescendingOrder)
                comp *= -1;
        }
        return comp < 0;
    }
    return QSortFilterProxyModel::lessThan(left, right);
}
