/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2020 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <algorithm>
#include "tools/benchmarking.h"
#include "base/mainthread.h"
#include "detailcolumn.h"
#include "filesystemhandler.h"
#include "filesystemmodel.h"
#include "filesystemmodelfileviewproxy.h"
#include "filesystemmodelsubtreeproxy.h"
#include "filesystemnode.h"


sh::filesystem::FilesystemModelSubtreeProxy::FilesystemModelSubtreeProxy(QModelIndex root, sh::filesystem::FilesystemModelFileviewProxy *upperproxy) :
    QAbstractItemModel(upperproxy), _inChangeTransaction(false),
    _rootindex(root), mainmodel(sh::filesystem::FilesystemModel::instance().get()), _upperproxy(upperproxy)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    _rootnode = mainmodel->getNodeForIndex(root);
    connect(mainmodel, SIGNAL(rowsAboutToBeInserted(QModelIndex,int,int)), this, SLOT(slot_rowsAboutToBeInserted(QModelIndex,int,int)));
    connect(mainmodel, SIGNAL(rowsInserted(QModelIndex,int,int)), this, SLOT(slot_rowsInserted(QModelIndex,int,int)));
    connect(mainmodel, SIGNAL(columnsAboutToBeInserted(QModelIndex,int,int)), this, SLOT(slot_columnsAboutToBeInserted(QModelIndex,int,int)));
    connect(mainmodel, SIGNAL(columnsInserted(QModelIndex,int,int)), this, SLOT(slot_columnsInserted(QModelIndex,int,int)));
    connect(mainmodel, SIGNAL(rowsAboutToBeRemoved(QModelIndex,int,int)), this, SLOT(slot_rowsAboutToBeRemoved(QModelIndex,int,int)));
    connect(mainmodel, SIGNAL(rowsRemoved(QModelIndex,int,int)), this, SLOT(slot_rowsRemoved(QModelIndex,int,int)));
    connect(mainmodel, SIGNAL(columnsAboutToBeRemoved(QModelIndex,int,int)), this, SLOT(slot_columnsAboutToBeRemoved(QModelIndex,int,int)));
    connect(mainmodel, SIGNAL(columnsRemoved(QModelIndex,int,int)), this, SLOT(slot_columnsRemoved(QModelIndex,int,int)));
    connect(mainmodel, SIGNAL(dataChanged(QModelIndex,QModelIndex)), this, SLOT(slot_dataChanged(QModelIndex,QModelIndex)));
}

void sh::filesystem::FilesystemModelSubtreeProxy::slot_rowsAboutToBeInserted(const QModelIndex & parent, int start, int end)
{
    _inChangeTransaction = true;
    if (mainmodel->getNodeForIndex(parent)==_rootnode)
        this->beginInsertRows(QModelIndex(), start, end);
}

void sh::filesystem::FilesystemModelSubtreeProxy::slot_rowsInserted(const QModelIndex& parent, int, int)
{
    if (_inChangeTransaction) {
        if (mainmodel->getNodeForIndex(parent)==_rootnode)
            this->endInsertRows();
        _inChangeTransaction = false;
    }
}

void sh::filesystem::FilesystemModelSubtreeProxy::slot_columnsAboutToBeInserted(const QModelIndex & parent, int start, int end)
{
    _inChangeTransaction = true;
    if (mainmodel->getNodeForIndex(parent)==_rootnode)
        this->beginInsertColumns(QModelIndex(), start, end);
}

void sh::filesystem::FilesystemModelSubtreeProxy::slot_columnsInserted(const QModelIndex& parent, int, int)
{
    if (_inChangeTransaction) {
        if (mainmodel->getNodeForIndex(parent)==_rootnode)
            this->endInsertColumns();
        _inChangeTransaction = false;
    }
}

void sh::filesystem::FilesystemModelSubtreeProxy::slot_rowsAboutToBeRemoved(const QModelIndex & parent, int start, int end)
{
    _inChangeTransaction = true;
    std::shared_ptr<sh::filesystem::FilesystemNode> parentnode = mainmodel->getNodeForIndex(parent);

    // check if we went invalid by that removal
    bool wentInvalid = false;
    std::shared_ptr<sh::filesystem::FilesystemNode> bestValid = 0;
    for(int i=start; i<= std::min(end, parentnode->childnodeCount()-1) /* ... since it can contain our 'unsubstantial' loading node */ ; i++)
    {
        if (wentInvalid)
            break;
        std::shared_ptr<sh::filesystem::FilesystemNode> removedNode = parentnode->childnode(i);
        if (removedNode->nodetype() != sh::filesystem::FilesystemNodeType::File) { // further checks aren't really needed for simple files
            std::shared_ptr<sh::filesystem::FilesystemNode> me = _rootnode;
            while (me != 0)
            {
                if (me == removedNode)
                {
                    wentInvalid = true;
                    bestValid = removedNode->parentnode();
                    break;
                }
                me = me->parentnode();
            }
        }
    }
    if (wentInvalid)
        emit becomesInvalid(bestValid);
    else if (parentnode==_rootnode)
        this->beginRemoveRows(QModelIndex(), start, end);
}

void sh::filesystem::FilesystemModelSubtreeProxy::slot_rowsRemoved(const QModelIndex& parent, int, int)
{
    if (_inChangeTransaction) {
        if (mainmodel->getNodeForIndex(parent)==_rootnode)
            this->endRemoveRows();
        _inChangeTransaction = false;
    }
}

void sh::filesystem::FilesystemModelSubtreeProxy::slot_columnsAboutToBeRemoved(const QModelIndex & parent, int start, int end)
{
    _inChangeTransaction = true;
    if (mainmodel->getNodeForIndex(parent)==_rootnode)
        this->beginRemoveColumns(QModelIndex(), start, end);
}

void sh::filesystem::FilesystemModelSubtreeProxy::slot_columnsRemoved(const QModelIndex& parent, int, int)
{
    if (_inChangeTransaction) {
        if (mainmodel->getNodeForIndex(parent)==_rootnode)
            this->endRemoveColumns();
        _inChangeTransaction = false;
    }
}

void sh::filesystem::FilesystemModelSubtreeProxy::slot_dataChanged(const QModelIndex & topLeft, const QModelIndex & bottomRight)
{
    if (mainmodel->getNodeForIndex(topLeft)->parentnode()==_rootnode)
    {
        QModelIndex tl = index(topLeft.row(), topLeft.column(), QModelIndex());
        QModelIndex br = index(bottomRight.row(), bottomRight.column(), QModelIndex());
        emit dataChanged(tl, br);
    }
}

QVariant sh::filesystem::FilesystemModelSubtreeProxy::data(const QModelIndex &index, int role) const
{
    CHECK_CORRECT_CALLER_THREAD
    QVariant result;
    if (index.isValid()){
        std::shared_ptr<sh::filesystem::FilesystemNode> item = static_cast<sh::filesystem::FilesystemNode*>(index.internalPointer())->shared_from_this();
        if (index.column() > 0)
        {
            if (role == Qt::DisplayRole)
            {
                int detailCount = _rootnode->_childrenDetailCols.count();
                int idx = index.column() - 1;
                if (idx < detailCount)
                    result = item->_parent->_childrenDetailCols[idx]->displayValue(item, _upperproxy);
            }
            else if (role == Qt::TextAlignmentRole)
            {
                int detailCount = _rootnode->_childrenDetailCols.count();
                int idx = index.column() - 1;
                if (idx < detailCount)
                    result = (int)((item->_parent->_childrenDetailCols[idx]->isRightAligned() ? Qt::AlignRight : Qt::AlignLeft) | Qt::AlignVCenter);
            }
        }
        else
            result = mainmodel->data(index, role);
    }
    else
        result = mainmodel->data(_rootindex, role);
    return result;
}

QModelIndex sh::filesystem::FilesystemModelSubtreeProxy::mapFromSource(const QModelIndex mi) const
{
    if (mainmodel->getNodeForIndex(mi.parent()) == _rootnode)
        return index(mi.row(), mi.column());
    else
        return QModelIndex();
}

QModelIndex sh::filesystem::FilesystemModelSubtreeProxy::mapToSource(const QModelIndex mi) const
{
    return mainmodel->index(mi.row(), mi.column(), _rootindex);
}

Qt::ItemFlags sh::filesystem::FilesystemModelSubtreeProxy::flags(const QModelIndex &index) const
{
    CHECK_CORRECT_CALLER_THREAD
    if (!index.isValid())
        return mainmodel->flags(_rootindex);
    return mainmodel->flags(index);
}

QVariant sh::filesystem::FilesystemModelSubtreeProxy::headerData(int section, Qt::Orientation orientation, int role) const
{
    CHECK_CORRECT_CALLER_THREAD
    if (orientation == Qt::Horizontal && role == Qt::DisplayRole)
    {
        if (section <= 0)
            return mainmodel->headerData(0, orientation, role);
        else
        {
            int idx = section - 1;
            return QVariant(_rootnode->_childrenDetailCols[idx]->displayName());
        }
    }
    return QVariant();
}

QModelIndex sh::filesystem::FilesystemModelSubtreeProxy::index(int row, int column, const QModelIndex &parent) const
{
    CHECK_CORRECT_CALLER_THREAD
    QModelIndex myparent = (parent.isValid()) ? parent : (QModelIndex)_rootindex;
    return createIndex(row, column, mainmodel->index(row, 0, myparent).internalPointer());
}

QModelIndex sh::filesystem::FilesystemModelSubtreeProxy::parent(const QModelIndex &index) const
{
    CHECK_CORRECT_CALLER_THREAD
    QModelIndex op = mainmodel->parent(index);
    if (op == _rootindex)
        return QModelIndex();
    return op;
}

int sh::filesystem::FilesystemModelSubtreeProxy::rowCount(const QModelIndex &parent) const
{
    CHECK_CORRECT_CALLER_THREAD
    if (!parent.isValid())
        return mainmodel->rowCount(_rootindex);
    return mainmodel->rowCount(parent);
}

int sh::filesystem::FilesystemModelSubtreeProxy::columnCount(const QModelIndex &) const
{
    CHECK_CORRECT_CALLER_THREAD
    return 1 + _rootnode->_childrenDetailCols.count();
}
