/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2020 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <QSet>
#include "exceptions/programexception.h"
#include "base/mainthread.h"
#include "filesystemhandler.h"
#include "filesystemhandlerregister.h"
#include "filesystemmodel.h"
#include "filesystemnodelist.h"
#include "operation.h"


sh::filesystem::FilesystemNodeList::FilesystemNodeList() :
    QObject()
{
}

sh::filesystem::AdhocFilesystemNodeList::AdhocFilesystemNodeList() :
    sh::filesystem::FilesystemNodeList()
{
}

void sh::filesystem::AdhocFilesystemNodeList::resetItems(sh::filesystem::FilesystemNodeType type, QSet<std::shared_ptr<sh::filesystem::FilesystemNode>> nodes)
{
    for(int i=_nodes.length()-1;i>=0;i--)
        if (_nodes.at(i)->nodetype()==type)
            _nodes.removeAt(i);
    foreach(std::shared_ptr<sh::filesystem::FilesystemNode> newnode, nodes)
        if (newnode->nodetype()==type)
            _nodes.append(newnode);
}

void sh::filesystem::AdhocFilesystemNodeList::addItems(QSet<std::shared_ptr<sh::filesystem::FilesystemNode>> nodes)
{
    foreach(std::shared_ptr<sh::filesystem::FilesystemNode> node, nodes)
        _nodes.append(node);
}

void sh::filesystem::AdhocFilesystemNodeList::removeItems(QSet<std::shared_ptr<sh::filesystem::FilesystemNode>> nodes)
{
    foreach(std::shared_ptr<sh::filesystem::FilesystemNode> node, nodes)
        _nodes.removeOne(node);
}

void sh::filesystem::AdhocFilesystemNodeList::addItem(std::shared_ptr<sh::filesystem::FilesystemNode> node)
{
    _nodes.append(node);
}

void sh::filesystem::AdhocFilesystemNodeList::removeItem(std::shared_ptr<sh::filesystem::FilesystemNode> node)
{
    _nodes.removeOne(node);
}

std::shared_ptr<sh::filesystem::FilesystemNode> sh::filesystem::AdhocFilesystemNodeList::mynode()
{
    return 0;
}

sh::filesystem::AdhocFilesystemNodeList::~AdhocFilesystemNodeList()
{
}

sh::filesystem::ModelBackedFilesystemNodeList::ModelBackedFilesystemNodeList(FilesystemNode *node) :
    sh::filesystem::FilesystemNodeList(), _mynode(node), mymodel(node->model())
{
}

void sh::filesystem::ModelBackedFilesystemNodeList::resetItems(sh::filesystem::FilesystemNodeType type, QSet<std::shared_ptr<sh::filesystem::FilesystemNode>> pnewnodes)
{
    sh::base::MainThread::dispatcher()->invokeSync([&](){
        QSet<std::shared_ptr<sh::filesystem::FilesystemNode>> oldnodes;
        foreach(std::shared_ptr<sh::filesystem::FilesystemNode> oldnode, _nodes)
            if (oldnode->nodetype()==type)
                oldnodes.insert(oldnode);
        QSet<std::shared_ptr<sh::filesystem::FilesystemNode>> newnodes (pnewnodes);
        newnodes.subtract(oldnodes);
        oldnodes.subtract(pnewnodes);
        if (!newnodes.isEmpty())
            addItems(newnodes);
        if (!oldnodes.isEmpty())
            removeItems(oldnodes);
    });
}

void sh::filesystem::ModelBackedFilesystemNodeList::clear()
{
    CHECK_CORRECT_CALLER_THREAD
    _permanentnodes.clear();
    _nodes.clear();
#ifdef QT_DEBUG
    _childreneurls.clear();
#endif
}

void sh::filesystem::ModelBackedFilesystemNodeList::addItem_helper(std::shared_ptr<FilesystemNode> node)
{
    std::shared_ptr<FilesystemNode> __mynode = _mynode->shared_from_this();
#ifdef QT_DEBUG
    sh::exceptions::Exception::executeGuarded([&](){
        if (_mynode->_parent == 0 && !_mynode->_isRootNode)
            throw sh::exceptions::ProgramException(sh::exceptions::ExceptionData().details("adding a child to an unbound node is disallowed"));
        if(_childreneurls.contains(node->eurl()))
            throw sh::exceptions::ProgramException(sh::exceptions::ExceptionData().details(QString("some internal error occurred in the filesystem model: %1").arg(1)));
        _childreneurls.append(node->eurl());
        if (node->_origparent != __mynode)
            throw sh::exceptions::ProgramException(sh::exceptions::ExceptionData().details(QString("some internal error occurred in the filesystem model: %1").arg(10)));
    });
#endif
    node->_parent = __mynode;
    _nodes.append(node);
}

void sh::filesystem::ModelBackedFilesystemNodeList::addItems(QSet<std::shared_ptr<sh::filesystem::FilesystemNode>> nodes)
{
    sh::base::MainThread::dispatcher()->invokeSync([&](){
        if (!_mynode->isAlive())
            return;
        if (nodes.isEmpty())
            return;
        int first = _nodes.count();
        int last = nodes.count() + first - 1;
        QModelIndex parentidx = mymodel->getIndexForNode(_mynode->shared_from_this());
        QMap<sh::filesystem::FilesystemHandler*, QList<std::shared_ptr<sh::filesystem::FilesystemNode>>> unconfiguredNodes;
        mymodel->beginInsertRows(parentidx, first, last);
        {
            QMutexLocker locker(&mymodel->_nodeDataMutex);
            int ioff = 0;
            foreach (std::shared_ptr<sh::filesystem::FilesystemNode> node, nodes) {
                addItem_helper(node);
                node->_index = first + (ioff++);
                if (!node->_isConfigured)
                    unconfiguredNodes[node->handler()].append(node);
            }
        }
        if (!unconfiguredNodes.isEmpty()) {
            Operation op;
            sh::exceptions::Exception::executeGuarded([&](){
                foreach (sh::filesystem::FilesystemHandler* h, unconfiguredNodes.keys()) {
                    h->configureItems(&op, unconfiguredNodes[h]);
                    foreach (std::shared_ptr<sh::filesystem::FilesystemNode> n, unconfiguredNodes[h])
                        n->_isConfigured = true;
                }
            }, sh::exceptions::ExecuteGuardFlag::ALL_ERRORS_KILL_SHALLOT);
        }
        mymodel->endInsertRows();
    });
}

void sh::filesystem::ModelBackedFilesystemNodeList::removeItems(QSet<std::shared_ptr<sh::filesystem::FilesystemNode>> nodes)
{
    sh::base::MainThread::dispatcher()->invokeSync([&](){
         if (!_mynode->isAlive())
            return;
        foreach(std::shared_ptr<sh::filesystem::FilesystemNode> node, nodes)
            removeItem(node);
    });
}

void sh::filesystem::ModelBackedFilesystemNodeList::addItem(std::shared_ptr<sh::filesystem::FilesystemNode> node)
{
    sh::base::MainThread::dispatcher()->invokeSync([&](){
        if (!_mynode->isAlive())
            return;
        if (_mynode != sh::filesystem::FilesystemModel::instance()->rootnode.get() && !_mynode->parentnode())
            return;
        int first = _nodes.count();
        QModelIndex parentidx = mymodel->getIndexForNode(_mynode->shared_from_this());
        mymodel->beginInsertRows(parentidx, first, first);
        {
            QMutexLocker locker(&mymodel->_nodeDataMutex);
            addItem_helper(node);
            node->_index = first;
        }
        mymodel->endInsertRows();
        if (!node->_isConfigured)
        {
            Operation op;
            node->handler()->configureItems(&op, QList<std::shared_ptr<sh::filesystem::FilesystemNode>>({node}));
            node->_isConfigured = true;
        }
    });
}

void sh::filesystem::ModelBackedFilesystemNodeList::addPermanentItem(std::shared_ptr<sh::filesystem::FilesystemNode> node)
{
    sh::base::MainThread::dispatcher()->invokeSync([&](){
        if (!_mynode->isAlive())
            return;
        _permanentnodes.insert(node.get());
        addItem(node);
    });
}

void sh::filesystem::ModelBackedFilesystemNodeList::removeItem(std::shared_ptr<sh::filesystem::FilesystemNode> node)
{
    sh::base::MainThread::dispatcher()->invokeSync([&](){
        if (!_mynode->isAlive())
            return;
        if (_permanentnodes.contains(node.get()))
            return;
        int i = node->index();
        QModelIndex parentidx = mymodel->getIndexForNode(_mynode->shared_from_this());
        if (parentidx.isValid())
        {
            mymodel->beginRemoveRows(parentidx, i, i);
            {
                QMutexLocker _l1(&mymodel->_nodeDataMutex);
#ifdef QT_DEBUG
                if(!_childreneurls.contains(node->eurl()))
                    sh::exceptions::Exception::executeGuarded([&](){
                        throw sh::exceptions::ProgramException(sh::exceptions::ExceptionData().details(tr("some internal error occurred in the filesystem model: %1").arg(2)));
                    });
                _childreneurls.removeOne(node->eurl());
#endif
                _nodes.removeAt(i);
                for (int j = i; j < _nodes.count(); j++)
                    _nodes[j]->_index--;
                node->unmount();
            }
            mymodel->endRemoveRows();
        }
    });
}

void sh::filesystem::ModelBackedFilesystemNodeList::removePermanentItem(std::shared_ptr<sh::filesystem::FilesystemNode> node)
{
    sh::base::MainThread::dispatcher()->invokeSync([&](){
        if (!_mynode->isAlive())
            return;
        _permanentnodes.remove(node.get());
        removeItem(node);
    });
}

std::shared_ptr<sh::filesystem::FilesystemNode> sh::filesystem::ModelBackedFilesystemNodeList::mynode()
{
    return _mynode->shared_from_this();
}

const QList<std::shared_ptr<sh::filesystem::FilesystemNode> > *sh::filesystem::FilesystemNodeList::nodes()
{
    return &_nodes;
}

bool sh::filesystem::FilesystemNodeList::contains(std::shared_ptr<sh::filesystem::FilesystemNode> node)
{
    return nodes()->contains(node);
}

QSet<sh::filesystem::FilesystemNodeList*> sh::filesystem::FilesystemNodeListEditor::_pendingIterativeAddings;

QMutex sh::filesystem::FilesystemNodeListEditor::_pendingMutex;

QWaitCondition sh::filesystem::FilesystemNodeListEditor::_iterativeAddingPossibleCondition;

sh::filesystem::FilesystemNodeListEditor::FilesystemNodeListEditor(sh::filesystem::FilesystemNodeList* list, std::shared_ptr<const sh::filesystem::Eurl>parenteurl, sh::filesystem::FilesystemNodeType nodestype)
        : _list(list), _parenteurl(parenteurl), _nodestype(nodestype), _model(sh::filesystem::FilesystemModel::instance().get())
{
    _parentnode = list->mynode();
    _handler = _parentnode ? _parentnode->handler() : sh::filesystem::FilesystemHandlerRegister::instance()->findHandler(parenteurl->scheme()).get();
}

void sh::filesystem::FilesystemNodeListEditor::setItems(QList<QString> items)
{
    QMutexLocker _l(&_pendingMutex);
    while (_pendingIterativeAddings.contains(_list))
        _iterativeAddingPossibleCondition.wait(&_pendingMutex);
    QSet<std::shared_ptr<sh::filesystem::FilesystemNode>> nitems;
    sh::base::MainThread::dispatcher()->invokeSync([&](){
        foreach (QString item, items) {
            nitems.insert(_model->getOrCreateFilesystemNode(getChildEurl(_parenteurl, item), _handler, _nodestype, false, _parentnode, false, true));
        }
    });
    _list->resetItems(_nodestype, nitems);
}

void sh::filesystem::FilesystemNodeListEditor::beginIterativeAdding()
{
    QMutexLocker _l(&_pendingMutex);
    while (_pendingIterativeAddings.contains(_list))
        _iterativeAddingPossibleCondition.wait(&_pendingMutex);
    _pendingIterativeAddings.insert(_list);
    foreach(std::shared_ptr<sh::filesystem::FilesystemNode> node, *(_list->nodes()))
        if (node->nodetype() == _nodestype)
            _iterativeAddingsBeforeState.insert(node);
    _beganAddingIteratively = true;
}

void sh::filesystem::FilesystemNodeListEditor::addItem(QString item)
{
    addCustomItem(getChildEurl(_parenteurl, item), item);
}

std::shared_ptr<sh::filesystem::FilesystemNode> sh::filesystem::FilesystemNodeListEditor::addCustomItem(std::shared_ptr<const sh::filesystem::Eurl> eurl, QString displayname, sh::filesystem::FilesystemHandler* handler)
{
    std::shared_ptr<sh::filesystem::FilesystemNode> newnode;
#ifdef QT_DEBUG
    if (handler && handler != FilesystemHandlerRegister::instance()->findHandler(eurl->scheme()).get())
        sh::exceptions::Exception::executeGuarded([&](){
            throw sh::exceptions::ProgramException(sh::exceptions::ExceptionData().details("invalid call to addCustomItem: 1"));
        });
#endif
    sh::base::MainThread::dispatcher()->invokeSync([&](){
        newnode = _model->getOrCreateFilesystemNode(eurl, handler?handler:_handler, _nodestype, false, _parentnode, false, true);
        if (!displayname.isNull())
            newnode->setDisplayName(displayname);
    });
    addExistingNodeItem(newnode);
    return newnode;
}

void sh::filesystem::FilesystemNodeListEditor::addExistingNodeItem(std::shared_ptr<FilesystemNode> node)
{
    if (_iterativeAddingsBeforeState.contains(node))
        _iterativeAddingsBeforeState.remove(node);
    else
        _list->addItem(node);
}

void sh::filesystem::FilesystemNodeListEditor::endIterativeAdding()
{
    QMutexLocker _l(&_pendingMutex);
    _pendingIterativeAddings.remove(_list);
    _list->removeItems(_iterativeAddingsBeforeState);
    _beganAddingIteratively = false;
    _iterativeAddingsBeforeState.clear();
    _iterativeAddingPossibleCondition.wakeOne();
}

void sh::filesystem::FilesystemNodeListEditor::setItemsAreHosts()
{
    _itemsAreHosts = true;
}

sh::filesystem::FilesystemNodeList* sh::filesystem::FilesystemNodeListEditor::rawlist()
{
    return _list;
}

std::shared_ptr<const sh::filesystem::Eurl> sh::filesystem::FilesystemNodeListEditor::getChildEurl(std::shared_ptr<const sh::filesystem::Eurl> eurl, QString item)
{
    if (_itemsAreHosts)
    {
#ifdef QT_DEBUG
        if (!eurl->outerUrlIsRootDirectory())
            sh::exceptions::Exception::executeGuarded([&](){
                throw sh::exceptions::ProgramException(sh::exceptions::ExceptionData().details("Added hostname '"+item+"' as child in a bad way: " + eurl->asString()));
            });
#endif
        if (eurl->hasInnerUrls())
            return eurl->outermostInnerEurl()->enwrapWithOuterUrl(eurl->scheme(), item, "");
        else
            return sh::filesystem::Eurl::create(eurl->scheme(), item, "");
    }
    else
        return eurl->withAppendedSegment(item);
}

sh::filesystem::FilesystemNodeListEditor::~FilesystemNodeListEditor()
{
    QMutexLocker _l(&_pendingMutex);
    if (_beganAddingIteratively)
        endIterativeAdding();
}
