/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2020 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef ARCHIVEFILESYSTEMHANDLER_H
#define ARCHIVEFILESYSTEMHANDLER_H

#include <functional>
#include "actions/common/actionopenarchive.h"
#include "filesystem/detailcolumn.h"
#include "filesystem/filesystemhandler.h"


namespace sh{
    namespace filesystemhandlers {

        /** \brief
         * A filesystem handler for archive files of some formats.
         *
         * This implementation uses process calls to some external utility tools.
         */
        class ArchiveFilesystemHandler : public sh::filesystem::FilesystemHandler
        {
            Q_OBJECT
        public:
            explicit ArchiveFilesystemHandler(sh::filesystem::FilesystemModel *model);

            virtual void itemlist(sh::filesystem::Operation* op, std::shared_ptr<const sh::filesystem::Eurl> eurl, sh::filesystem::FilesystemNodeType type, sh::filesystem::FilesystemNodeListEditor *list);
            virtual sh::filesystem::FilesystemNodeType getType(sh::filesystem::Operation* op, std::shared_ptr<const sh::filesystem::Eurl> eurl);

            virtual qint64 getSize(sh::filesystem::Operation* op, std::shared_ptr<const sh::filesystem::Eurl> eurl);

            virtual QDateTime getMtime(sh::filesystem::Operation* op, std::shared_ptr<const sh::filesystem::Eurl> eurl);
            virtual void setMtime(sh::filesystem::Operation* op, std::shared_ptr<const sh::filesystem::Eurl> eurl, QDateTime time);

            virtual QString getMimetype(sh::filesystem::Operation* op, std::shared_ptr<const sh::filesystem::Eurl> eurl);

            virtual QString getLinkTarget(sh::filesystem::Operation* op, std::shared_ptr<const sh::filesystem::Eurl> eurl);

            virtual bool canGetFileContent(sh::filesystem::Operation* op, std::shared_ptr<const sh::filesystem::Eurl> eurl);
            virtual std::shared_ptr<QIODevice> getFileContent(sh::filesystem::Operation* op, std::shared_ptr<const sh::filesystem::Eurl> eurl);

            virtual bool canCreateDirectory(sh::filesystem::Operation* op, std::shared_ptr<const sh::filesystem::Eurl> eurl);
            virtual void createDirectory(sh::filesystem::Operation* op, std::shared_ptr<const sh::filesystem::Eurl> eurl);

            virtual bool canCreateLink(sh::filesystem::Operation* op, std::shared_ptr<const sh::filesystem::Eurl> eurl);
            virtual void createLink(sh::filesystem::Operation* op, std::shared_ptr<const sh::filesystem::Eurl> eurl, QString target);

            virtual bool canCreateFile(sh::filesystem::Operation* op, std::shared_ptr<const sh::filesystem::Eurl> eurl);
            virtual void createFile(sh::filesystem::Operation* op, std::shared_ptr<const sh::filesystem::Eurl> eurl, QIODevice* content, HandlerTransfer* handlertransfer);

            virtual bool canDeleteItem(sh::filesystem::Operation* op, std::shared_ptr<const sh::filesystem::Eurl> eurl);
            virtual void deleteItem(sh::filesystem::Operation* op, std::shared_ptr<const sh::filesystem::Eurl> eurl);

            virtual bool canRenameItem(sh::filesystem::Operation* op, std::shared_ptr<const sh::filesystem::Eurl> src);
            virtual void renameItem(sh::filesystem::Operation* op, std::shared_ptr<const sh::filesystem::Eurl> src, QString destpath);

            virtual QList<std::shared_ptr<sh::actions::AbstractActionItem> > getActions(const QList<std::shared_ptr<const sh::filesystem::Eurl>> eurls);

            /** @a_infrastructure
             */
            static void doInitialize();

            /** @a_infrastructure
             */
            static void doShutdown();

        signals:

        public slots:

        protected:

            static bool testexecute(QString cmd);
            QString execute(QString cmd, QStringList args, bool throwOnBadExitCode = true, QByteArray *rawanswer=0, QByteArray *bstdin=0);

            void _parseZipEntryListOutput(sh::filesystem::Operation *op, std::shared_ptr<const sh::filesystem::Eurl>eurl, std::function<void(qint64, QDateTime, QString)> onitem);
            void _parseTarEntryListOutput(sh::filesystem::Operation *op, std::shared_ptr<const sh::filesystem::Eurl>eurl, std::function<void(qint64, QDateTime, QString)> onitem);
            void _tarOperationAfterCompression(sh::filesystem::Operation *op, QString container, std::function<void(QString)> ontarfile);
            QString _zipMaskFile(QString path);
            static QRegularExpression parsere;

            class ActionOpenZipArchive : public sh::actions::common::ActionOpenArchive
            {
            public:
                ActionOpenZipArchive(QList<std::shared_ptr<sh::filesystem::FilesystemNode>> l) : sh::actions::common::ActionOpenArchive(l) {}
                virtual QString label() { return QObject::tr("Open zip archive"); }
                virtual QString scheme() { return "zip"; }
            };

            class ActionOpenTarArchive : public sh::actions::common::ActionOpenArchive
            {
            public:
                ActionOpenTarArchive(QList<std::shared_ptr<sh::filesystem::FilesystemNode>> l) : sh::actions::common::ActionOpenArchive(l) {}
                virtual QString label() { return QObject::tr("Open tar archive"); }
                virtual QString scheme() { return "tar"; }
            };

            static QList<QObject*> actionfactoryregistrations;
        };

    }
}

#endif // ARCHIVEFILESYSTEMHANDLER_H
