/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2020 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifdef GIO

#include <gio/gio.h>
#include "actions/actionexecutioninfo.h"
#include "filesystem/filesystemhandlerregister.h"
#include "base/mainthread.h"
#include "base/threadpool.h"
#include "base/iconmanager.h"
#include "gnomeiodevicesfilesystemhandler.h"
#include "gnomeiogenericfilesystemhandler.h"

REGISTER_SINGLETON(sh::filesystemhandlers, GnomeIODevicesFilesystemHandler, "filesystem")


void sh::filesystemhandlers::GnomeIODevicesFilesystemHandler::doInitialize()
{
    sh::filesystem::FilesystemHandlerRegister::instance()->addHandler("devices", sh::filesystemhandlers::GnomeIODevicesFilesystemHandler::instance());
}

void gvolumechangedcallback (GVolumeMonitor*, GVolume*, gpointer)
{
    sh::base::ThreadPool::enqueueForce([=](){
        sh::filesystemhandlers::GnomeIODevicesFilesystemHandler::instance()->refresh();
    });
}

sh::filesystemhandlers::GnomeIODevicesFilesystemHandler::GnomeIODevicesFilesystemHandler() :
    sh::filesystemhandlers::GnomeIOFilesystemHandler(), masternodeslock(QMutex::Recursive)
{
    rootnode = sh::filesystem::FilesystemModel::instance()->createFilesystemNode(sh::filesystem::Eurl::create("devices", "", "/"), this, sh::filesystem::FilesystemNodeType::Directory, false, sh::filesystem::FilesystemModel::instance()->rootNode());
    rootnode->setDisplayName(tr("Devices"));
    rootnode->setIcon(sh::base::IconManager::instance()->getIcon("drive-removable-media"));
    childlist = rootnode->childnodes();
    volumemonitor = g_volume_monitor_get ();
    g_signal_connect(volumemonitor, "volume-added", G_CALLBACK(gvolumechangedcallback), NULL);
    g_signal_connect(volumemonitor, "volume-removed", G_CALLBACK(gvolumechangedcallback), NULL);
    g_signal_connect(volumemonitor, "volume-changed", G_CALLBACK(gvolumechangedcallback), NULL);
}

sh::filesystemhandlers::GnomeIODevicesFilesystemHandler::~GnomeIODevicesFilesystemHandler()
{
    g_object_unref(volumemonitor);
}

void sh::filesystemhandlers::GnomeIODevicesFilesystemHandler::itemlist(sh::filesystem::Operation*, std::shared_ptr<const sh::filesystem::Eurl> eurl, sh::filesystem::FilesystemNodeType type, sh::filesystem::FilesystemNodeListEditor *list)
{
    GVolumeMonitor* volmon = (GVolumeMonitor*) volumemonitor;
    if (eurl->outerUrlIsRootDirectory())
    {
        if (type == sh::filesystem::FilesystemNodeType::Directory)
        {
            QList<GnomeIODevice> devices;
            QHash<QString, QString> _devicenames;
            GList* vollist = g_volume_monitor_get_volumes (volmon);
            sh::base::MainThread::dispatcher()->invokeSync([&](){
                GList *vollisti;
                for (vollisti = vollist; vollisti != NULL; vollisti = vollisti->next)
                {
                    GVolume* vol = (GVolume*) vollisti->data;
                    QString volname = g_volume_get_name(vol);
                    QString uuid = g_volume_get_identifier(vol, G_VOLUME_IDENTIFIER_KIND_UNIX_DEVICE);
                    if (!uuid.isEmpty())
                    {
                        GMount* mount = g_volume_get_mount(vol);
                        QString mountroot;
                        if (mount)
                        {
                            GFile* froot = g_mount_get_root(mount);
                            mountroot = g_file_get_uri(froot);
                            g_object_unref(froot);
                        }
                        bool isMounted = (mount != 0);
                        g_object_unref(mount);
                        devices.append(GnomeIODevice(uuid, volname, mountroot, isMounted));
                        _devicenames[uuid] = volname;
                    }
                    g_object_unref(vol);
                }
                g_list_free(vollist);
            });
            {
                list->beginIterativeAdding();
                QMutexLocker _l1(&masternodeslock);
                devicenames = _devicenames;
                foreach(GnomeIODevice entry, devices)
                {
                    std::shared_ptr<sh::filesystem::FilesystemNode> mnode;
                    if (entry.isMounted)
                        mnode = _getMountedNode(entry.name, entry.url, list->rawlist()->mynode());
                    else
                        mnode = _getUnmountedNode(entry.name, list->rawlist()->mynode());
                    list->addExistingNodeItem(mnode);
                }
                list->endIterativeAdding();
            }
        }
    }
}

void sh::filesystemhandlers::GnomeIODevicesFilesystemHandler::requestNewNode(std::shared_ptr<const sh::filesystem::Eurl> eurl, std::shared_ptr<sh::filesystem::FilesystemNode>* nnode)
{
    if (!sh::filesystem::FilesystemHandlerRegister::instance()->findHandler(eurl->scheme()))
        sh::filesystem::FilesystemHandlerRegister::instance()->addHandler(eurl->scheme(), sh::filesystemhandlers::GnomeIODevicesFilesystemHandler::instance());
    *nnode = sh::filesystem::FilesystemModel::instance()->createFilesystemNode(eurl,
                                     sh::filesystem::FilesystemHandlerRegister::instance()->findHandler(eurl->scheme()).get(), sh::filesystem::FilesystemNodeType::Directory, false,
                                     sh::filesystemhandlers::GnomeIODevicesFilesystemHandler::rootnode, false);
    (*nnode)->setDisplayName(eurl->asString());
    (*nnode)->setIcon(sh::base::IconManager::instance()->getIcon("drive-removable-media"));
    sh::filesystemhandlers::GnomeIODevicesFilesystemHandler::rootnode->childnodes()->addPermanentItem(*nnode);
}

sh::filesystem::FilesystemNodeType sh::filesystemhandlers::GnomeIODevicesFilesystemHandler::getType(sh::filesystem::Operation*, std::shared_ptr<const sh::filesystem::Eurl>)
{
    return sh::filesystem::FilesystemNodeType::Unknown;
}

QList<std::shared_ptr<sh::actions::AbstractActionItem> > sh::filesystemhandlers::GnomeIODevicesFilesystemHandler::getActions(const QList<std::shared_ptr<const sh::filesystem::Eurl>>)
{
    QList<std::shared_ptr<sh::actions::AbstractActionItem> > result;
    return result;
}

std::shared_ptr<sh::filesystem::FilesystemNode> sh::filesystemhandlers::GnomeIODevicesFilesystemHandler::_getMountedNode(QString gvolumeuuid, QString grooturl, std::shared_ptr<sh::filesystem::FilesystemNode> parentnode)
{
    QMutexLocker _l1(&masternodeslock);
    std::shared_ptr<const sh::filesystem::Eurl> rooteurl = sh::filesystemhandlers::GnomeIOFilesystemHandler::gurl2eurl(grooturl);
    std::shared_ptr<sh::filesystem::FilesystemHandler> _handler = sh::filesystem::FilesystemHandlerRegister::instance()->findHandler(rooteurl->scheme());
    if (!_handler)
    {
        _handler = std::make_shared<sh::filesystemhandlers::GnomeIOGenericFilesystemHandler>(sh::filesystem::FilesystemModel::instance().get());
        sh::filesystem::FilesystemHandlerRegister::instance()->addHandler(rooteurl->scheme(), _handler);
    }
    std::shared_ptr<sh::filesystem::FilesystemNode> newnode;
    QString devicename;
    {
        QMutexLocker _l1(&masternodeslock);
        auto it = devicenames.find(gvolumeuuid);
        if (it != devicenames.end())
            devicename = (*it);
        if (devicename.isEmpty())
            devicename = gvolumeuuid;
    }
    sh::base::MainThread::dispatcher()->invokeSync([&](){
        newnode = sh::filesystem::FilesystemModel::instance()->getOrCreateFilesystemNode(rooteurl, _handler.get(), sh::filesystem::FilesystemNodeType::Directory, false, parentnode, false);
        newnode->setDisplayName(devicename);
        sh::filesystemhandlers::ActionUnmountGnomeIOLocation::offerUnmountLocation(newnode, grooturl);
    });
    return newnode;
}

std::shared_ptr<sh::filesystem::FilesystemNode> sh::filesystemhandlers::GnomeIODevicesFilesystemHandler::_getUnmountedNode(QString gvolumeuuid, std::shared_ptr<sh::filesystem::FilesystemNode> parentnode)
{
    QMutexLocker _l1(&masternodeslock);
    std::shared_ptr<sh::filesystem::FilesystemNode> newnode;
    QString gurl = QString("devices:///") + QUrl::toPercentEncoding(gvolumeuuid).constData();
    std::shared_ptr<const sh::filesystem::Eurl> rooteurl = sh::filesystem::Eurl::fromString(gurl);
    QString devicename;
    {
        QMutexLocker _l1(&masternodeslock);
        auto it = devicenames.find(gvolumeuuid);
        if (it != devicenames.end())
            devicename = (*it);
        if (devicename.isEmpty())
            devicename = gvolumeuuid;
    }
    sh::base::MainThread::dispatcher()->invokeSync([&](){
        newnode = sh::filesystem::FilesystemModel::instance()->getOrCreateFilesystemNode(rooteurl, this, sh::filesystem::FilesystemNodeType::Directory, false, parentnode, false);
        newnode->setDisplayName(devicename);
        sh::filesystemhandlers::ActionMountGnomeIOLocation::offerMountGVolumeUuid(newnode, gvolumeuuid);
    });
    return newnode;
}

void sh::filesystemhandlers::GnomeIODevicesFilesystemHandler::refresh()
{
    rootnode->requestContainedItems(sh::filesystem::FilesystemNodeType::Directory);
}

sh::filesystemhandlers::GnomeIODevicesFilesystemHandler::GnomeIODevice::GnomeIODevice(QString name, QString displayname, QString url, bool isMounted)
        : name(name), displayname(displayname), url(url), isMounted(isMounted)
{
}

#endif
