/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2020 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifdef GIO

#include "actions/actionexecutioninfo.h"
#include "actions/actionexecutionuserfeedback.h"
#include "filesystem/filesystemhandlerregister.h"
#include "tools/bookmarkmanager.h"
#include "base/mainthread.h"
#include "base/iconmanager.h"
#include "base/threadpool.h"
#include "ui/mainwindow.h"
#include "gnomeionetworkfilesystemhandler.h"

REGISTER_SINGLETON(sh::filesystemhandlers, GnomeIONetworkFilesystemHandler, "filesystem")


sh::filesystemhandlers::GnomeIONetworkFilesystemHandler::GnomeIONetworkFilesystemHandler() :
    sh::filesystemhandlers::GnomeIOFilesystemHandler()
{
}

void sh::filesystemhandlers::GnomeIONetworkFilesystemHandler::requestNewNode(std::shared_ptr<const sh::filesystem::Eurl> eurl, std::shared_ptr<sh::filesystem::FilesystemNode>* nnode)
{
    static QStringList schemes({"smb", "ftp", "ftps", "ssh", "sftp", "dav", "davs", "obex"});
    if (schemes.contains(eurl->scheme()))
    {
        if (!sh::filesystem::FilesystemHandlerRegister::instance()->findHandler(eurl->scheme()))
            sh::filesystem::FilesystemHandlerRegister::instance()->addHandler(eurl->scheme(), sh::filesystemhandlers::GnomeIONetworkFilesystemHandler::instance());
        *nnode = sh::filesystem::FilesystemModel::instance()->createFilesystemNode(eurl,
                                         sh::filesystem::FilesystemHandlerRegister::instance()->findHandler(eurl->scheme()).get(), sh::filesystem::FilesystemNodeType::Directory, false,
                                         sh::filesystemhandlers::GnomeIONetworkFilesystemHandler::rootnode, false);
        (*nnode)->setDisplayName(eurl->asString());
        (*nnode)->setIcon(sh::base::IconManager::instance()->getIcon("network-server"));
        sh::filesystemhandlers::GnomeIONetworkFilesystemHandler::rootnode->childnodes()->addPermanentItem(*nnode);
    }
}

QList<std::shared_ptr<sh::actions::AbstractActionItem> > sh::filesystemhandlers::GnomeIONetworkFilesystemHandler::getActions(const QList<std::shared_ptr<const sh::filesystem::Eurl>> eurls){
    QList<std::shared_ptr<sh::actions::AbstractActionItem> > result = sh::filesystemhandlers::GnomeIOFilesystemHandler::getActions(eurls);
    if (eurls.length() == 1 && eurls.first() == rootnode->eurl())
        result.append(std::make_shared<ActionMountNetworkGnomeIOLocation>());
    return result;
}

void sh::filesystemhandlers::GnomeIONetworkFilesystemHandler::doInitialize()
{
    sh::filesystem::FilesystemHandlerRegister::instance()->addHandler("network", sh::filesystemhandlers::GnomeIONetworkFilesystemHandler::instance());
    rootnode = sh::filesystem::FilesystemModel::instance()->createFilesystemNode(sh::filesystem::Eurl::create("network", "", "/"), sh::filesystemhandlers::GnomeIONetworkFilesystemHandler::instance().get(), sh::filesystem::FilesystemNodeType::Directory, false, sh::filesystem::FilesystemModel::instance()->rootNode());
    rootnode->setDisplayName(tr("Network"));
    rootnode->setIcon(sh::base::IconManager::instance()->getIcon("network-server"));
    sh::filesystem::FilesystemModel::instance()->addOpenNodeHelper(3000, [=](std::shared_ptr<const sh::filesystem::Eurl> eurl){ // helps opening a new node by requesting smb:/// subnodes (without returning anything)
        if (eurl->outerUrlIsRootDirectory())
        {
            QString scheme = eurl->scheme();
            if (scheme == "smb")
                rootnode->requestContainedItems();
        }
        return QList<std::shared_ptr<sh::filesystem::FilesystemNode>>();
    });
}

sh::filesystemhandlers::ActionMountNetworkGnomeIOLocation::ActionMountNetworkGnomeIOLocation()
                : sh::actions::ActionActionItem(tr("Mount network service"), true, "network-server")
{
}

void sh::filesystemhandlers::ActionMountNetworkGnomeIOLocation::action(sh::actions::ActionExecutionInfo* info)
{
    info->setHead(tr("Mounting network service"));
    sh::actions::ActionExecutionUserFeedback::GridformEntries entries;
    sh::actions::ActionExecutionUserFeedback::GridformEntry* entry = entries.newEntry(tr("Host (optionally incl. :port)"));
    int chSsh = entry->addChoice(tr("SSH/SFTP"), tr("host"));
    int chDav = entry->addChoice(tr("WebDAV"), tr("host"));
    int chDavs = entry->addChoice(tr("WebDAVS"), tr("host"));
    int chFtp = entry->addChoice(tr("FTP"), tr("host"));
    int chFtps = entry->addChoice(tr("FTPS"), tr("host"));
    int chSmb = entry->addChoice(tr("SMB file share (Windows network)"), tr("host"));
    int ianswer = info->userfeedback()->simpleChooserGridform(tr("Please choose the network service you want to connect to."), &entries, QStringList({tr("Connect"), tr("Connect && Bookmark for later"), tr("Cancel")}), 0, 2);
    if (ianswer != 2 && entry->choices.selected > -1)
    {
        QString h = entry->choices.list[entry->choices.selected]->text;
        QString host = h;
        QString path = "/";
        int islash = h.indexOf("/");
        if (islash > -1)
        {
            host = h.mid(0, islash);
            path = h.mid(islash);
        }
        std::shared_ptr<const sh::filesystem::Eurl> eurl;
        if (entry->choices.selected == chSsh)
            eurl = sh::filesystem::Eurl::create("sftp", host, path);
        else if (entry->choices.selected == chDav)
            eurl = sh::filesystem::Eurl::create("dav", host, path);
        else if (entry->choices.selected == chDavs)
            eurl = sh::filesystem::Eurl::create("davs", host, path);
        else if (entry->choices.selected == chFtp)
            eurl = sh::filesystem::Eurl::create("ftp", host, path);
        else if (entry->choices.selected == chFtps)
            eurl = sh::filesystem::Eurl::create("ftps", host, path);
        else if (entry->choices.selected == chSmb)
            eurl = sh::filesystem::Eurl::create("smb", host, path);
        if (ianswer == 1)
            sh::tools::BookmarkManager::instance()->addBookmark({tr("Network Services")}, eurl->asString(), eurl);
        {
            IGNORE_FORBIDDEN_ASYNC_CALLS
            sh::ui::MainWindow::mainWindow()->jumpToEurl(eurl);
        }
    }
}

std::shared_ptr<sh::filesystem::FilesystemNode> sh::filesystemhandlers::GnomeIONetworkFilesystemHandler::rootnode;

#endif
