/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2020 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <QBuffer>
#include <QFile>
#include "base/mainthread.h"
#include "exceptions/ioexception.h"
#include "filesystem/filesystemmodel.h"
#include "filesystem/filesystemhandlerregister.h"
#include "filesystem/operation.h"
#include "tools/filetypes/filetypemanager.h"
#include "sharcfilesystemhandler.h"

REGISTER_SINGLETON(sh::filesystemhandlers, SharcFilesystemHandler, "filesystem")


sh::filesystemhandlers::SharcFilesystemHandler::SharcFilesystemHandler() :
    sh::filesystem::FilesystemHandler(sh::filesystem::FilesystemModel::instance().get())
{
    _detailColumnNameArchivedSize = std::make_shared<ArchivedSizeDetailColumn>();
}

void sh::filesystemhandlers::SharcFilesystemHandler::doInitialize()
{
    sh::filesystem::FilesystemHandlerRegister::instance()->addHandler("sharc", sh::filesystemhandlers::SharcFilesystemHandler::instance());
}

void sh::filesystemhandlers::SharcFilesystemHandler::_parse(sh::filesystem::Operation* op, std::shared_ptr<const sh::filesystem::Eurl> eurl, bool doesOutput, QStringList append, std::function<QString(QString,QString)> _decide, bool *peekData, std::function<void(QString)> _data)
{
    NOT_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    QString container = op->fetchContainerFile(eurl);
    QFile inputFile( container );
    QStringList outbytes;
    outbytes << "SHARC";
    if (inputFile.open(QIODevice::ReadOnly))
    {
       QTextStream in(&inputFile);
       if (!in.atEnd() && in.readLine()!="SHARC")
           throw sh::exceptions::IOException(sh::exceptions::ExceptionData().detailsAreInteresting().details(tr("This file is not a sharc archive.")).retryable(false));
       bool skipContent = true;
       while ( !in.atEnd() )
       {
           QString line = in.readLine();
           if (line.trimmed().isEmpty())
               continue;
           if (!line.startsWith(" "))
           {
               if (peekData)
                   *peekData = false;
               QString prefix = line.mid(0,4);
               QString object = line.mid(5).trimmed();
               QString decision = _decide(prefix, object);
               if (decision.isEmpty())
                   skipContent = true;
               else
               {
                   skipContent = false;
                   if (doesOutput)
                       outbytes << decision;
               }
           }
           else
           {
               if (peekData && *peekData)
                   _data(line);
               if (!skipContent)
               {
                   if (doesOutput)
                       outbytes << line;
               }
           }
       }
       inputFile.close();
    }
    if (doesOutput) {
        outbytes << append;
        QFile outputFile(container);
        if (!outputFile.open(QIODevice::WriteOnly | QIODevice::Truncate))
            throw sh::exceptions::IOException(sh::exceptions::ExceptionData().detailsAreInteresting().details(QObject::tr("Error while opening: ") + container));
        QTextStream out(&outputFile);
        foreach(QString line, outbytes)
            out << line << "\n";
        if (outputFile.error() != QFile::NoError)
            throw sh::exceptions::IOException(sh::exceptions::ExceptionData().detailsAreInteresting().details(tr("Unable to write to file '%1': %2").arg(eurl->asString()).arg(outputFile.errorString())));
        outputFile.close();
    }
}

void sh::filesystemhandlers::SharcFilesystemHandler::itemlist(sh::filesystem::Operation* op, std::shared_ptr<const sh::filesystem::Eurl> eurl, sh::filesystem::FilesystemNodeType type, sh::filesystem::FilesystemNodeListEditor *list)
{
    QString dir = _eurlToLocal(eurl->outerUrl());
    if (!dir.endsWith("/")) dir=dir+"/";
    QStringList ql;
    QString _prefix;
    if (type == sh::filesystem::FilesystemNodeType::Directory)
        _prefix = "FLDR";
    else if (type == sh::filesystem::FilesystemNodeType::File)
        _prefix = "FILE";
    else if (type == sh::filesystem::FilesystemNodeType::Link)
        _prefix = "LINK";
    else
        return;
    _parse(op, eurl, false, QStringList(), [&](QString prefix, QString name){
        if (name.startsWith(dir) && prefix==_prefix)
        {
            QString relname = name.mid(dir.length());
            if (!relname.contains("/")) {
                ql << relname;
            }
        }
        return QString();
    });
    list->setItems(ql);
}

void sh::filesystemhandlers::SharcFilesystemHandler::configureItems(sh::filesystem::Operation*, QList<std::shared_ptr<sh::filesystem::FilesystemNode>> nodes)
{
    foreach(std::shared_ptr<sh::filesystem::FilesystemNode> node, nodes)
        node->addDetail(_detailColumnNameArchivedSize);
}

QString sh::filesystemhandlers::SharcFilesystemHandler::getLinkTarget(sh::filesystem::Operation* op, std::shared_ptr<const sh::filesystem::Eurl> eurl)
{
    QString itm = _eurlToLocal(eurl->outerUrl());
    QString result;
    bool peekData = false;
    _parse(op, eurl, false, QStringList(), [&](QString prefix, QString name){
        if (name == itm && prefix == "LINK") {
            peekData = true;
        }
        return QString();
    }, &peekData, [&](QString l){
        result = l.mid(1);
    });
    return result;
}

sh::filesystem::FilesystemNodeType sh::filesystemhandlers::SharcFilesystemHandler::getType(sh::filesystem::Operation* op, std::shared_ptr<const sh::filesystem::Eurl> eurl)
{
    QString itm = _eurlToLocal(eurl->outerUrl());
    if (itm == "/")
        return sh::filesystem::FilesystemNodeType::Directory;
    sh::filesystem::FilesystemNodeType result = sh::filesystem::FilesystemNodeType::NONE;
    _parse(op, eurl, false, QStringList(), [&](QString prefix, QString name){
        if (name == itm) {
            if (prefix == "FLDR") result = sh::filesystem::FilesystemNodeType::Directory;
            else if (prefix == "FILE") result = sh::filesystem::FilesystemNodeType::File;
            else if (prefix == "LINK") result = sh::filesystem::FilesystemNodeType::Link;
            else result = sh::filesystem::FilesystemNodeType::Unknown;
        }
        return QString();
    });
    return result;
}

qint64 sh::filesystemhandlers::SharcFilesystemHandler::getSize(sh::filesystem::Operation* op, std::shared_ptr<const sh::filesystem::Eurl> eurl)
{
    qint64 result = 0;
    QString itm = _eurlToLocal(eurl->outerUrl());
    bool peekData = false;
    _parse(op, eurl, false, QStringList(), [&](QString prefix, QString name){
        if (name == itm && prefix == "FILE") {
            peekData = true;
        }
        return QString();
    }, &peekData, [&](QString l){
        QByteArray base64 = l.trimmed().toUtf8();
        result += QByteArray::fromBase64(base64).length();
    });
    return result;
}

QString sh::filesystemhandlers::SharcFilesystemHandler::getMimetype(sh::filesystem::Operation* op, std::shared_ptr<const sh::filesystem::Eurl> eurl)
{
    return sh::tools::filetypes::FileTypeManager::instance()->determineMimetype(op, eurl);
}

bool sh::filesystemhandlers::SharcFilesystemHandler::canGetFileContent(sh::filesystem::Operation*, std::shared_ptr<const sh::filesystem::Eurl>)
{
    return true;
}

std::shared_ptr<QIODevice> sh::filesystemhandlers::SharcFilesystemHandler::getFileContent(sh::filesystem::Operation* op, std::shared_ptr<const sh::filesystem::Eurl> eurl)
{
    QByteArray* bytes = new QByteArray();
    QString itm = _eurlToLocal(eurl->outerUrl());
    bool peekData = false;
    _parse(op, eurl, false, QStringList(), [&](QString prefix, QString name){
        if (name == itm && prefix == "FILE") {
            peekData = true;
        }
        return QString();
    }, &peekData, [&](QString l){
        bytes->append(QByteArray::fromBase64(l.trimmed().toUtf8()));
    });
    return std::make_shared<QBuffer2>(bytes);
}

bool sh::filesystemhandlers::SharcFilesystemHandler::canCreateDirectory(sh::filesystem::Operation*, std::shared_ptr<const sh::filesystem::Eurl>)
{
    return true;
}

bool sh::filesystemhandlers::SharcFilesystemHandler::canCreateFile(sh::filesystem::Operation*, std::shared_ptr<const sh::filesystem::Eurl>)
{
    return true;
}

void sh::filesystemhandlers::SharcFilesystemHandler::createDirectory(sh::filesystem::Operation* op, std::shared_ptr<const sh::filesystem::Eurl> eurl)
{
    QStringList a; a << ("FLDR " + _eurlToLocal( eurl->outerUrl() ));
    _parse(op, eurl, true, a, [&](QString prefix, QString name){ return prefix+" "+name; });
}

bool sh::filesystemhandlers::SharcFilesystemHandler::canCreateLink(sh::filesystem::Operation*, std::shared_ptr<const sh::filesystem::Eurl>)
{
    return true;
}

void sh::filesystemhandlers::SharcFilesystemHandler::createLink(sh::filesystem::Operation* op, std::shared_ptr<const sh::filesystem::Eurl> eurl, QString target)
{
    QStringList a; a << "LINK " + _eurlToLocal( eurl->outerUrl()) << " " + target.replace("\n","");
    _parse(op, eurl, true, a, [&](QString prefix, QString name){ return prefix+" "+name; });
}

void sh::filesystemhandlers::SharcFilesystemHandler::createFile(sh::filesystem::Operation* op, std::shared_ptr<const sh::filesystem::Eurl> eurl, QIODevice *content, HandlerTransfer *handlertransfer)
{
    deleteItem(op, eurl);
    QStringList a; a << "FILE " + _eurlToLocal(eurl->outerUrl());
    const int BUFFER_SIZE = 2048;
    char buffer[BUFFER_SIZE];
    while (qint64 l = content->read(buffer, BUFFER_SIZE))
    {
        if (l == -1)
            throw sh::exceptions::IOException(sh::exceptions::ExceptionData().detailsAreInteresting().details(tr("Unable to read from a stream.")));
        QByteArray barr(buffer, l);
        a << " " + barr.toBase64();
        handlertransfer->respectCancel();
        handlertransfer->incrementTransferredBytes(l);
    }
    _parse(op, eurl, true, a, [&](QString prefix, QString name){ return prefix+" "+name; });
}

QList<std::shared_ptr<sh::actions::AbstractActionItem> > sh::filesystemhandlers::SharcFilesystemHandler::getActions(const QList<std::shared_ptr<const sh::filesystem::Eurl>>)
{
    return QList<std::shared_ptr<sh::actions::AbstractActionItem> >();
}


QString sh::filesystemhandlers::SharcFilesystemHandler::_eurlToLocal(std::shared_ptr<const sh::filesystem::Eurl>eurl)
{
   return eurl->asString().mid(8).replace("\n","");
}

std::shared_ptr<const sh::filesystem::Eurl> sh::filesystemhandlers::SharcFilesystemHandler::localToEurl(const QString local)
{
   return sh::filesystem::Eurl::fromString("sharc:/"+local);
}

bool sh::filesystemhandlers::SharcFilesystemHandler::canDeleteItem(sh::filesystem::Operation*, std::shared_ptr<const sh::filesystem::Eurl>)
{
    return true;
}

bool sh::filesystemhandlers::SharcFilesystemHandler::canRenameItem(sh::filesystem::Operation*, std::shared_ptr<const sh::filesystem::Eurl>)
{
    return true;
}

void sh::filesystemhandlers::SharcFilesystemHandler::renameItem(sh::filesystem::Operation* op, std::shared_ptr<const sh::filesystem::Eurl> src, QString destpath)
{
    QString _dir = _eurlToLocal(src->outerUrl());
    int _dirlen = _dir.length();
    QString _ndir = _eurlToLocal(src->outerUrl()->root()->withAppendedSegments(destpath));
    QString dir = (_dir.endsWith("/")) ? _dir : _dir+"/";
    _parse(op, src, true, QStringList(), [&](QString prefix, QString name){
        if (name.startsWith(dir) || name==_dir)
            return prefix+" "+_ndir+name.mid(_dirlen);
        else
            return prefix+" "+name;
    });
}

void sh::filesystemhandlers::SharcFilesystemHandler::deleteItem(sh::filesystem::Operation* op, std::shared_ptr<const sh::filesystem::Eurl> eurl)
{
    QString _dir = _eurlToLocal(eurl->outerUrl());
    QString dir = (_dir.endsWith("/")) ? _dir : _dir+"/";
    _parse(op, eurl, true, QStringList(), [&](QString prefix, QString name){
        if (name.startsWith(dir) || name==_dir)
            return QString();
        else
            return prefix+" "+name;
    });
}

sh::filesystemhandlers::SharcFilesystemHandler::ArchivedSizeDetailColumn::ArchivedSizeDetailColumn() : sh::filesystem::DetailColumn("SHARC_ARCHIVED_SIZE", tr("Archived size"), sh::filesystem::DetailColumn::DISPLAYINDEX_INTERESTING,  true)
{
}

QVariant sh::filesystemhandlers::SharcFilesystemHandler::ArchivedSizeDetailColumn::determineValue(std::shared_ptr<sh::filesystem::FilesystemNode> node, sh::filesystem::Operation* op)
{
    qint64 result = 0;
    QString itm = _eurlToLocal(node->eurl()->outerUrl());
    bool peekData = false;
    _parse(op, node->eurl(), false, QStringList(), [&](QString prefix, QString name){
        if (name == itm && prefix == "FILE") {
            peekData = true;
        }
        return QString();
    }, &peekData, [&](QString l){
        QByteArray loc = l.trimmed().toUtf8();
        result += loc.length();
    });
    return QVariant(result);
}

QDateTime sh::filesystemhandlers::SharcFilesystemHandler::getMtime(sh::filesystem::Operation*, std::shared_ptr<const sh::filesystem::Eurl>)
{
    return QDateTime();
}

void sh::filesystemhandlers::SharcFilesystemHandler::setMtime(sh::filesystem::Operation*, std::shared_ptr<const sh::filesystem::Eurl>, QDateTime)
{
}
