/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2020 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <QFileInfo>
#ifdef Q_OS_WIN
#include "filesystem/filesystemmodel.h"
#include "filesystem/filesystemhandlerregister.h"
#include "filesystem/filesystemoperation.h"
#include "filesystem/operation.h"
#include "base/iconmanager.h"
#include "base/mainthread.h"
#include "base/logger.h"
#include "windowsnetworkfilesystemhandler.h"
#define _WIN32_WINNT 0x0601
#define INITGUID
#include <SdkDdkVer.h>
#include <windows.h>
#include <objbase.h>
#include <Shobjidl.h>
#include <shlguid.h>
#include <shlobj.h>
#include <shlwapi.h>

REGISTER_SINGLETON(sh::filesystemhandlers, WindowsNetworkFilesystemHandler, "filesystem")


sh::filesystemhandlers::WindowsNetworkFilesystemHandler::WindowsNetworkFilesystemHandler()
        : sh::filesystemhandlers::LocalFilesystemHandler()
{
}

void sh::filesystemhandlers::WindowsNetworkFilesystemHandler::createRootNode()
{
    sh::base::MainThread::dispatcher()->invokeSync([&](){
        rootnode = _model->createFilesystemNode(sh::filesystem::Eurl::create("network", "", "/"), this, sh::filesystem::FilesystemNodeType::Directory, false, sh::filesystem::FilesystemModel::instance()->rootNode());
        rootnode->setDisplayName(tr("Network"));
        rootnode->setIcon(sh::base::IconManager::instance()->getIcon("network-server"));
    });
}

QString sh::filesystemhandlers::WindowsNetworkFilesystemHandler::eurlToLocal(std::shared_ptr<const sh::filesystem::Eurl> eurl)
{
    if (eurl->hasInnerUrls() || eurl->scheme()!="smb")
        return QString();
    return QString("\\\\%1\%2").arg(eurl->hostname().toLower()).arg(eurl->path().replace("/", "\\"));
}

std::shared_ptr<const sh::filesystem::Eurl> sh::filesystemhandlers::WindowsNetworkFilesystemHandler::localToEurl(const QString local)
{
    std::shared_ptr<const sh::filesystem::Eurl> result;
    if (local.startsWith("\\\\")) {
        int ibslash = local.indexOf("\\", 2);
        result = sh::filesystem::Eurl::fromString(QString("smb://%1/%2").arg(local.mid(2, ibslash-2).toLower()).arg(local.mid(ibslash+1).replace("\\", "/")));
    }
    return result;
}

void sh::filesystemhandlers::WindowsNetworkFilesystemHandler::itemlist(sh::filesystem::Operation* op, std::shared_ptr<const sh::filesystem::Eurl> eurl, sh::filesystem::FilesystemNodeType type, sh::filesystem::FilesystemNodeListEditor *list)
{
    if (eurl->outerUrlIsRootDirectory() && eurl->scheme()=="network") {
        list->beginIterativeAdding();
        QStringList names;
        if (type == sh::filesystem::FilesystemNodeType::Directory) {
            CoInitializeEx(NULL, COINIT_MULTITHREADED);
            IShellItem* pItem = nullptr;
            HRESULT hr = SHCreateItemInKnownFolder(FOLDERID_NetworkFolder, 0, nullptr, IID_PPV_ARGS(&pItem));
            if (SUCCEEDED(hr)) {
                IEnumShellItems* pEnum = nullptr;
                hr = pItem->BindToHandler(nullptr, BHID_EnumItems, IID_PPV_ARGS(&pEnum));
                if (SUCCEEDED(hr)) {
                     IShellItem* pChildItem = nullptr;
                     ULONG ulFetched = 0;
                     do {
                         hr = pEnum->Next(1, & pChildItem, &ulFetched);
                         if (FAILED(hr))
                             break;
                         if (ulFetched != 0) {
                             LPWSTR szChildName = nullptr;
                             pChildItem->GetDisplayName(SIGDN_NORMALDISPLAY, &szChildName);
                             QString name = QString::fromWCharArray(szChildName);
                             names << name;
                             CoTaskMemFree(szChildName);
                             pChildItem ->Release();
                         }
                     } while (hr != S_FALSE);
                     pEnum->Release();
                }
                pItem->Release();
            }
        }
        foreach (QString name, names)
            list->addCustomItem(sh::filesystem::Eurl::create("smb", name, "/"), name, this);
        list->endIterativeAdding();
    }
    else
        sh::filesystemhandlers::LocalFilesystemHandler::itemlist(op, eurl, type, list);
}

sh::filesystem::FilesystemNodeType sh::filesystemhandlers::WindowsNetworkFilesystemHandler::getType(sh::filesystem::Operation* op, std::shared_ptr<const sh::filesystem::Eurl> eurl)
{
    if (!eurl->outerUrlIsRootDirectory() && eurl->parentSegment()->outerUrlIsRootDirectory() && eurl->scheme()=="network") {
        QList<std::shared_ptr<const sh::filesystem::Eurl>> l = op->filesystem()->itemlist(eurl->parentSegment());
        foreach (std::shared_ptr<const sh::filesystem::Eurl> e, l) {
            if (e->basename().toLower() == eurl->basename().toLower())
                return sh::filesystem::FilesystemNodeType::Directory;
        }
        return sh::filesystem::FilesystemNodeType::NONE;
    }
    else
        return sh::filesystemhandlers::LocalFilesystemHandler::getType(op, eurl);
}

void sh::filesystemhandlers::WindowsNetworkFilesystemHandler::doInitialize()
{
    sh::filesystem::FilesystemHandlerRegister::instance()->addHandler("network", instance());
    sh::filesystem::FilesystemHandlerRegister::instance()->addHandler("smb", instance());
    sh::filesystemhandlers::WindowsNetworkFilesystemHandler::instance()->createRootNode();
    sh::filesystem::FilesystemModel::instance()->addOpenNodeHelper(11000, [=](std::shared_ptr<const sh::filesystem::Eurl> eurl){ // helps opening a new smb host node via windows api
        QList<std::shared_ptr<sh::filesystem::FilesystemNode>> result;
        if (QFileInfo(eurlToLocal(eurl)).exists()) {
            std::shared_ptr<sh::filesystem::FilesystemNode> nnode = sh::filesystem::FilesystemModel::instance()->createFilesystemNode(eurl,
                                             instance().get(), sh::filesystem::FilesystemNodeType::Directory,
                                             false, rootnode, false);
            nnode->setDisplayName(eurl->hostname().toUpper());
            nnode->setIcon(sh::base::IconManager::instance()->getIcon("network-server"));
            qobject_cast<sh::filesystem::ModelBackedFilesystemNodeList*>(rootnode->childnodes())->addPermanentItem(nnode);
            result << nnode;
        }
        return result;
    });

}

void sh::filesystemhandlers::WindowsNetworkFilesystemHandler::doShutdown()
{
}

#endif
