/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2020 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <QApplication>
#include <QTranslator>
#include <QTime>
#include <functional>
#include "filesystem/filesystemhandlerregister.h"
#include "filesystemhandlers/archivefilesystemhandler.h"
#include "scripting/scriptingengine.h"
#include "tools/benchmarking.h"
#include "base/iconmanager.h"
#include "tools/filetypes/filetypemanager.h"
#include "base/mainthread.h"
#include "base/logger.h"
#include "base/shallotprocess.h"
#include "base/singletoninitializer.h"
#include "ui/mainwindow.h"

#ifdef Q_OS_UNIX
#include <iostream>
#include <signal.h>
#endif

/** \namespace Ui
 * \brief Qt-internal for ui.
 */

/** \class _ProjectInfo
 * \brief Some internal program information.
 *
 * This is auto-generated content only used by the core.
 */

/** \namespace sh
 * \brief Shallot root.
 */

/** \namespace sh::actions
 * \brief Infrastructure for actions (submenus, executable ones, and more), as shown in the toolbar and menus and used at some other places.
 *
 * See the abstract base class sh::actions::AbstractActionItem for more.
 */

/** \namespace sh::actions::common
 * \brief Implementations of actions.
 *
 * Subclasses of sh::actions::AbstractActionItem (and possibly some auxiliary stuff).
 * Many action implementations used in Shallot are here.
 * You can find other actions in the sibling namespaces and a few very special ones at different places.
 */

/** \namespace sh::actions::mainmenu
 * \brief Implementation of the main menu.
 *
 * Subclasses of sh::actions::AbstractActionItem (and possibly some auxiliary stuff).
 * You can find other actions in the sibling namespaces and a few very special ones at different places.
 */

/** \namespace sh::base
 * \brief Core logic.
 *
 * Includes common data structures and some basic infrastructure.
 * Some other namespaces in sh contain parts of the infrastructure as well.
 */

/** \namespace sh::configuration
 * \brief Configuration, so everything you see in 'Finetuning', and some other data (e.g. get shallot program directory).
 */

/** \namespace sh::detailcolumns
 * \brief Implementations of detail columns.
 *
 * Subclasses of sh::filesystem::DetailColumn (and possibly some auxiliary stuff).
 * They are shown in the fileviews and can be queried in code.
 */

/** \namespace sh::exceptions
 * \brief Exceptions.
 *
 * The class hierarchy of Shallot exceptions as well as some utilities for
 * exception handling.
 */

/** \namespace sh::filepropertydialogtabs
 * \brief Implementations of tabs in a file property dialog.
 *
 * Subclasses of sh::ui::FilePropertyDialogTab (and possibly some auxiliary stuff).
 * They implement content in file property dialogs.
 */

/** \namespace sh::filesystem
 * \brief Filesystem model and auxiliary logic.
 */

/** \namespace sh::filesystemhandlers
 * \brief Implementations of filesystem handlers.
 *
 * Subclasses of sh::filesystem::FilesystemHandler (and possibly some auxiliary stuff).
 * They implement how to access various types of filesystems.
 *
 * Some special purpose handlers reside outside of this namespace.
 */

/** \namespace sh::paneldetails
 * \brief Infrastructure for panel details.
 *
 * Can be shown in the bottom part of the main window.
 * Beyond the infrastructure, it also includes some very basic detail providers.
 *
 * See sh::paneldetails::PanelDetailManager for more.
 */

/** \namespace sh::scripting
 * \brief The Shallot scripting interface.
 */

/** \namespace sh::scripting::api
 * \brief Adapter functions and classes for interaction with scripting.
 *
 * Some programming interfaces in the core are not directly usable
 * for the scripting api. Scripting-friendly surrogates for them are here.
 */

/** \namespace sh::search
 * \brief Infrastructure for searches in the filesystem.
 *
 * See sh::search::SearchManager for more.
 */

/** \namespace sh::search::criteria
 * \brief Implementations of search criteria.
 *
 * Subclasses of sh::search::SearchCriterion.
 */

/** \namespace sh::settings
 * \brief Infrastructure for settings, so everything you see in 'Manage saved settings'.
 *
 * See the abstract base class sh::settings::Setting and sh::settings::SettingsManager for more.
 */

/** \namespace sh::settings::common
 * \brief Implementations of settings.
 *
 * Subclasses of sh::settings::Setting (and possibly some auxiliary stuff).
 * Everything here could be listed in the 'Save settings' dialog.
 */

/** \namespace sh::tools
 * \brief Auxiliary stuff.
 */

/** \namespace sh::tools::accounts.
 * \brief Account and password storage.
 *
 * See sh::tools::accounts::AccountsManager for more.
 */

/** \namespace sh::tools::filetypes
 * \brief Tools for determining a file's type (pdf, image, mp3, et al) and ways how to deal with that.
 *
 * See sh::tools::filetypes::FileTypeManager for more.
 */

/** \namespace sh::tools::thumbnailproviders
 * \brief Thumbnail providers.
 *
 * Subclasses of sh::tools::ThumbnailProvider (and possibly some auxiliary stuff).
 * Those classes are used for generating thumbnail pictures for files.
 */

/** \namespace sh::ui
 * \brief User interface.
 *
 * This is the presentation layer. The classes directly from here are abstract; often
 * even technically, but clearly from conceptual perspective.
 * Find non-abstract user interface implementations in sub-namespaces.
 */

/** \namespace sh::ui::qt
 * \brief The Qt user interface implementation.
 *
 * This is the default UI. It builds a desktop application based on the Qt gui components.
 */

/** \namespace sh::ui::qt::feedbackpanels
 * \brief Qt user interface parts for user feedback in action execution dialogs.
 *
 * Implementations of sh::ui::qt::feedbackpanels::FeedbackPanel are used for implementing parts of
 * sh::ui::ActionExecutionInfoDialog.
 */

/** \namespace sh::ui::web
 * \brief The web user interface implementation.
 *
 * This optional UI provides an application running in a web browser.
 */

/** \namespace sh::ui::web::engines
 * \brief Web server low-level engine implementations.
 */

REGISTER_SINGLETON_GROUP(logger)
REGISTER_SINGLETON_GROUP(foundation, "logger")
REGISTER_SINGLETON_GROUP(filesystem, "foundation")
REGISTER_SINGLETON_GROUP(mainwindow, "filesystem")
REGISTER_SINGLETON_GROUP(early, "mainwindow")
REGISTER_SINGLETON_GROUP(scripting, "early")
REGISTER_SINGLETON_GROUP(base, "scripting")
REGISTER_SINGLETON_GROUP(sh_main, "base")

namespace sh {

    /** @brief (Un)Initialization for main app.
     */
    class MainInit
    {
        public:
        static void doInitialize();
        static void doShutdown();
        static sh::base::SingletonInitializer* singletonInitializer;
        static volatile sig_atomic_t caughtSigterm;
    };

}

volatile sig_atomic_t sh::MainInit::caughtSigterm = 0;

void sh::MainInit::doInitialize()
{
    auto w = sh::ui::MainWindow::mainWindow();
    if (w) {
        w->_initialize(singletonInitializer);
        w->setCurrentProfile("main");
    }
    else if (!sh::ui::MainWindow::runsHeadless())
        sh::exceptions::Exception::executeGuarded([&](){
            throw sh::exceptions::ArgumentException(sh::exceptions::ExceptionData().message(QObject::tr("Unsupported ui mode '%1'.").arg(sh::base::ShallotProcess::instance()->uiMode())).resumeable(false));
        });
#ifdef Q_OS_UNIX
    // register SIGTERM handler
    struct sigaction action;
    memset(&action, 0, sizeof(struct sigaction));
    action.sa_handler = [](int){ caughtSigterm = 1; };
    sigaction(SIGTERM, &action, NULL);
    QTimer* timerCaughtSigterm = new QTimer(QCoreApplication::instance());
    QObject::connect(timerCaughtSigterm, &QTimer::timeout, [=](){
        if (caughtSigterm) {
            caughtSigterm = 0;
            if (w) {
                QString preventCloseMessage;
                if (!w->isCloseable(&preventCloseMessage)) {
                    std::cout << preventCloseMessage.toStdString() << "\n";
                    return;
                }
            }
            sh::ui::MainWindow::_closeDirectly(singletonInitializer);
        }
    });
    timerCaughtSigterm->start(1000);
#endif
}

void sh::MainInit::doShutdown()
{
}

sh::base::SingletonInitializer* sh::MainInit::singletonInitializer = nullptr;

REGISTER_STATICINIT(sh, MainInit, "sh_main")

int main(int argc, char *argv[])
{
    QApplication a(argc, argv);
    QCoreApplication::setAttribute(Qt::AA_UseHighDpiPixmaps, true);
    a.setQuitOnLastWindowClosed(false);

    // load correct translation
    QStringList locales;
    foreach (QString locale, sh::base::ShallotProcess::instance()->lang()) {
        locale = locale.trimmed().replace("-", "_");
        if (!locale.isEmpty())
            locales.append(locale);
    }
    locales << QLocale::system().name();
    QTranslator translator;
    foreach (QString locale, locales) {
        if (translator.load(QString("shallot_") + locale, sh::base::ShallotProcess::shallotDataDir() + "/i18n")) {
            a.installTranslator(&translator);
            break;
        }
    }

    sh::base::MainThread::dispatcher();
    sh::filesystem::DetailColumn::VALUE_UNAVAILABLE();
    sh::exceptions::ExceptionHandlerSettingsManager::instance()->handlerSettings();
    sh::base::SingletonInitializer singletonInitializer;
    sh::MainInit::singletonInitializer = &singletonInitializer;
    sh::base::ShallotProcess::instance();

    // initialize all singletons (most of them are stored in sh::base::SingletonInitializer)
    singletonInitializer.callInitializers();

    // execute the main loop in an exception-aware way
    bool succ = false;
    int result = 100;
    while (!succ)
        sh::exceptions::Exception::executeGuarded([&](){
            result = a.exec();
            succ = true;
        }, sh::exceptions::ExecuteGuardFlag::ALL_ERRORS_KILL_SHALLOT);

    auto w = sh::ui::MainWindow::mainWindow();
    if (w)
        delete w;
    return result;
}
