/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2020 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <QLocale>
#include "base/mainthread.h"
#include "filesystem/filesystemnode.h"
#include "detailcolumns/detailcolumnmimetype.h"
#include "detailcolumns/detailcolumnfilesize.h"
#include "detailcolumns/detailcolumnmtime.h"
#include "paneldetails/paneldetailmanager.h"
#include "base/singletoninitializer.h"
#include "tools/operationscache.h"
#include "base/threadpool.h"
#include "commonpaneldetails.h"

REGISTER_STATICINIT(sh::paneldetails, CommonPanelDetails, "base")


void sh::paneldetails::CommonPanelDetails::NumberOfSelectedItems(std::shared_ptr<sh::paneldetails::PanelDetail> detail, QList<std::shared_ptr<sh::filesystem::FilesystemNode>> nodes, sh::filesystem::Operation*)
{
    detail->setRows({ sh::paneldetails::PanelDetailRow(
                   QList<sh::paneldetails::PanelDetailRowValueElement>({sh::paneldetails::PanelDetailRowValueElement(QObject::tr("%1 items selected.").arg(nodes.length()))})
                      ) });
}

void sh::paneldetails::CommonPanelDetails::TotalFileSize(std::shared_ptr<PanelDetail> detail, QList<std::shared_ptr<sh::filesystem::FilesystemNode>> nodes, sh::filesystem::Operation*)
{
    sh::base::ThreadPool::enqueueForce([=](){
        sh::exceptions::Exception::executeGuarded_errorpanel([&](){
            quint64 totalsize = 0;
            foreach(std::shared_ptr<sh::filesystem::FilesystemNode> node, nodes)
            {
                int ntype = node->nodetype();
                if (ntype == sh::filesystem::FilesystemNodeType::File || ntype == sh::filesystem::FilesystemNodeType::Link)
                {
                    std::shared_ptr<sh::filesystem::Operation> op = sh::tools::OperationsCache::instance()->getOperationForContainer(node->eurl()->outermostInnerEurl());
                    totalsize += sh::detailcolumns::DetailColumnFilesize::instance()->requestValue(node, op.get()).toLongLong();
                }
            }
            static QLocale l;
            detail->setRows({ sh::paneldetails::PanelDetailRow(QObject::tr("Total size"),
                              QList<sh::paneldetails::PanelDetailRowValueElement>({sh::paneldetails::PanelDetailRowValueElement(l.toString(totalsize) + QObject::tr(" bytes"))})
                              )});
        });
    });
}

void sh::paneldetails::CommonPanelDetails::_ViaDetailColumn(std::shared_ptr<sh::paneldetails::PanelDetail> detail, std::shared_ptr<sh::filesystem::FilesystemNode> node, std::shared_ptr<sh::filesystem::DetailColumn> column, QString keyname, std::function<QString(QVariant)> valtostring)
{
    detail->setRows({ sh::paneldetails::PanelDetailRow(keyname,
                      QList<sh::paneldetails::PanelDetailRowValueElement>({sh::paneldetails::PanelDetailRowValueElement()})
                      ) });
    sh::base::ThreadPool::enqueueForce([=](){
        sh::exceptions::Exception::executeGuarded_errorpanel([&](){
            std::shared_ptr<sh::filesystem::Operation> op = sh::tools::OperationsCache::instance()->getOperationForContainer(node->eurl()->outermostInnerEurl());
            detail->setRows({ sh::paneldetails::PanelDetailRow(keyname,
                              QList<sh::paneldetails::PanelDetailRowValueElement>({sh::paneldetails::PanelDetailRowValueElement(valtostring(column->requestValue(node, op.get())))})
                              ) });
        });
    });
}

void sh::paneldetails::CommonPanelDetails::FileType(std::shared_ptr<sh::paneldetails::PanelDetail> detail, std::shared_ptr<sh::filesystem::FilesystemNode> node, sh::filesystem::Operation*)
{
    if (node->nodetype() == sh::filesystem::FilesystemNodeType::File || node->nodetype() == sh::filesystem::FilesystemNodeType::Link)
        _ViaDetailColumn(detail, node, sh::detailcolumns::DetailColumnMimetype::instance(), QObject::tr("Type"), [](QVariant v){return v.toString();});
}

void sh::paneldetails::CommonPanelDetails::FileSize(std::shared_ptr<sh::paneldetails::PanelDetail> detail, std::shared_ptr<sh::filesystem::FilesystemNode> node, sh::filesystem::Operation*)
{
    if (node->nodetype() == sh::filesystem::FilesystemNodeType::File || node->nodetype() == sh::filesystem::FilesystemNodeType::Link)
        _ViaDetailColumn(detail, node, sh::detailcolumns::DetailColumnFilesize::instance(), QObject::tr("Size"), [](QVariant v){ static QLocale l; return l.toString(v.toLongLong()) + QObject::tr(" bytes");});
}

void sh::paneldetails::CommonPanelDetails::FileModificationTime(std::shared_ptr<sh::paneldetails::PanelDetail> detail, std::shared_ptr<sh::filesystem::FilesystemNode> node, sh::filesystem::Operation*)
{
    if (node->nodetype() == sh::filesystem::FilesystemNodeType::File || node->nodetype() == sh::filesystem::FilesystemNodeType::Link)
        _ViaDetailColumn(detail, node, sh::detailcolumns::DetailColumnMtime::instance(), QObject::tr("Last modified"), [](QVariant v){return v.toDateTime().toString(Qt::DefaultLocaleShortDate);});
}

void sh::paneldetails::CommonPanelDetails::doInitialize()
{
    auto dm = sh::paneldetails::PanelDetailManager::instance();
    dm->registerFactoryMulti(NumberOfSelectedItems, [](QList<std::shared_ptr<sh::filesystem::FilesystemNode>>,QString){}, sh::paneldetails::PanelDetailPositionIndex::Interesting + 100, 5);
    dm->registerFactoryMulti(TotalFileSize, [](QList<std::shared_ptr<sh::filesystem::FilesystemNode>>,QString){}, sh::paneldetails::PanelDetailPositionIndex::Interesting + 120, 5);
    dm->registerFactorySingle(FileType, [](std::shared_ptr<sh::filesystem::FilesystemNode>,QString){}, sh::paneldetails::PanelDetailPositionIndex::Interesting + 120, 4);
    dm->registerFactorySingle(FileSize, [](std::shared_ptr<sh::filesystem::FilesystemNode>,QString){}, sh::paneldetails::PanelDetailPositionIndex::Interesting + 140, 5);
    dm->registerFactorySingle(FileModificationTime, [](std::shared_ptr<sh::filesystem::FilesystemNode>,QString){}, sh::paneldetails::PanelDetailPositionIndex::Interesting + 160, 3);
}

void sh::paneldetails::CommonPanelDetails::doShutdown()
{

}
