/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2020 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "filesystem/filesystemnode.h"
#include "filesystem/operation.h"
#include "base/mainthread.h"
#include "paneldetailmanager.h"

REGISTER_SINGLETON(sh::paneldetails, PanelDetailManager, "base")


sh::paneldetails::PanelDetailManager::PanelDetailManager() :
    QObject()
{
}

std::shared_ptr<sh::paneldetails::PanelDetailFactory> sh::paneldetails::PanelDetailManager::registerFactorySingle(std::shared_ptr<sh::paneldetails::PanelDetailFactory> factory)
{
    QMutexLocker _l1(&_mutex);
    _factoriessingle.append(factory);
    return factory;
}

std::shared_ptr<sh::paneldetails::PanelDetailFactory> sh::paneldetails::PanelDetailManager::registerFactoryMulti(std::shared_ptr<sh::paneldetails::PanelDetailFactory> factory)
{
    QMutexLocker _l1(&_mutex);
    _factoriesmulti.append(factory);
    return factory;
}

std::shared_ptr<sh::paneldetails::PanelDetailFactory> sh::paneldetails::PanelDetailManager::registerFactorySingle(
        std::function<void(std::shared_ptr<PanelDetail>, std::shared_ptr<sh::filesystem::FilesystemNode>, sh::filesystem::Operation* op)> fctsetvalue,
        std::function<void(std::shared_ptr<sh::filesystem::FilesystemNode>, QString link)> fctlinktriggered,
        int position, int valueWidthHint)
{
    std::shared_ptr<sh::paneldetails::PanelDetailFactory> factory = std::make_shared<sh::paneldetails::PanelDetailFactoryForFunctionSingle>(fctsetvalue, fctlinktriggered, position, valueWidthHint);
    registerFactorySingle(factory);
    return factory;
}

std::shared_ptr<sh::paneldetails::PanelDetailFactory> sh::paneldetails::PanelDetailManager::registerFactoryMulti(
        std::function<void(std::shared_ptr<sh::paneldetails::PanelDetail>, QList<std::shared_ptr<sh::filesystem::FilesystemNode>>, sh::filesystem::Operation* op)> fctsetvalue,
        std::function<void(QList<std::shared_ptr<sh::filesystem::FilesystemNode>>, QString link)> fctlinktriggered,
        int position, int valueWidthHint)
{
    std::shared_ptr<sh::paneldetails::PanelDetailFactory> factory = std::make_shared<sh::paneldetails::PanelDetailFactoryForFunctionMulti>(fctsetvalue, fctlinktriggered, position, valueWidthHint);
    registerFactoryMulti(factory);
    return factory;
}

sh::paneldetails::PanelDetailRow::PanelDetailRow(QString key, QList<sh::paneldetails::PanelDetailRowValueElement> value)
            : _key(key), _value(value)
{
}

sh::paneldetails::PanelDetailRow::PanelDetailRow(QList<PanelDetailRowValueElement> value)
            : sh::paneldetails::PanelDetailRow(QString(), value)
{
}

QString sh::paneldetails::PanelDetailRow::key()
{
    return _key;
}

QList<sh::paneldetails::PanelDetailRowValueElement> sh::paneldetails::PanelDetailRow::value()
{
    return _value;
}

sh::paneldetails::PanelDetail::PanelDetail(int positionIndex, int valueWidthHint)
                : _positionIndex(positionIndex), _valueWidthHint(valueWidthHint)
{
}

QList<sh::paneldetails::PanelDetailRow> sh::paneldetails::PanelDetail::rows()
{
    QMutexLocker _l1(&_rowsmutex);
    return _rows;
}

int sh::paneldetails::PanelDetail::positionIndex()
{
    return _positionIndex;
}

int sh::paneldetails::PanelDetail::valueWidthHint()
{
    return _valueWidthHint;
}

void sh::paneldetails::PanelDetail::setRows(QList<sh::paneldetails::PanelDetailRow> rows)
{
    {
        QMutexLocker _l1(&_rowsmutex);
        _rows = rows;
    }
    emit changed();
}

void sh::paneldetails::PanelDetail::addRow(sh::paneldetails::PanelDetailRow row)
{
    {
        QMutexLocker _l1(&_rowsmutex);
        _rows.append(row);
    }
    emit changed();
}

void sh::paneldetails::PanelDetail::addOrReplaceRow(PanelDetailRow row)
{
    {
        QMutexLocker _l1(&_rowsmutex);
        auto itr = _rows.begin();
        QString key = row.key();
        while(itr != _rows.end())
        {
            if ((*itr).key() == key)
            {
                _rows.insert(_rows.erase(itr), row);
                break;
            }
            else
                itr++;
        }
        if (itr == _rows.end())
            _rows.append(row);
    }
    emit changed();
}

void sh::paneldetails::PanelDetail::_emit_linkTriggered(QString link)
{
    emit linkTriggered(link);
}

QList<std::shared_ptr<sh::paneldetails::PanelDetail> > sh::paneldetails::PanelDetailManager::getDetails(QList<std::shared_ptr<sh::filesystem::FilesystemNode>> nodes)
{
    NOT_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    QMutexLocker _l1(&_mutex);
    QList<std::shared_ptr<sh::paneldetails::PanelDetail> > result;
    sh::filesystem::Operation op;
    if (nodes.length() == 1)
    {
        if (nodes[0]->eurl())
        {
            foreach(std::shared_ptr<sh::paneldetails::PanelDetailFactory> detailfactory, _factoriessingle)
                result.append(detailfactory->construct(nodes, &op));
        }
    }
    else if (nodes.length() > 1)
    {
        foreach(std::shared_ptr<sh::paneldetails::PanelDetailFactory> detailfactory, _factoriesmulti)
            result.append(detailfactory->construct(nodes, &op));
    }
    qSort(result.begin(), result.end(), [](const std::shared_ptr<sh::paneldetails::PanelDetail>& a, const std::shared_ptr<sh::paneldetails::PanelDetail>& b){
        return a->positionIndex() < b->positionIndex();
    });
    return result;
}

sh::paneldetails::PanelDetailFactoryForFunctionSingle::PanelDetailFactoryForFunctionSingle(std::function<void(std::shared_ptr<sh::paneldetails::PanelDetail> detail, std::shared_ptr<sh::filesystem::FilesystemNode> node, sh::filesystem::Operation* op)> fctsetvalue, std::function<void(std::shared_ptr<sh::filesystem::FilesystemNode>, QString link)> fctlinktriggered, int position, int valueWidthHint)
{
    _fctsetvalue = fctsetvalue;
    _fctlinktriggered = fctlinktriggered;
    _position = position;
    _valueWidthHint = valueWidthHint;
}

std::shared_ptr<sh::paneldetails::PanelDetail> sh::paneldetails::PanelDetailFactoryForFunctionSingle::construct(QList<std::shared_ptr<sh::filesystem::FilesystemNode>> nodes, sh::filesystem::Operation* op)
{
    PanelDetail* _result = new PanelDetail(_position, _valueWidthHint);
    std::shared_ptr<PanelDetail> result(_result);
    _fctsetvalue(result, nodes[0], op);
    QObject::connect(_result, &PanelDetail::linkTriggered, [=](QString linktarget){
        sh::exceptions::Exception::executeGuarded([&](){
            _fctlinktriggered(nodes[0], linktarget);
        }, sh::exceptions::MANUAL_RETRY_ENABLED|sh::exceptions::AUTOMATIC_RETRY_ENABLED|sh::exceptions::RESUMEABLE_PROGRAM_ERROR_END_WITH_USER_FEEDBACK_HERE);
    });
    return result;
}

sh::paneldetails::PanelDetailFactoryForFunctionMulti::PanelDetailFactoryForFunctionMulti(std::function<void(std::shared_ptr<sh::paneldetails::PanelDetail> detail, QList<std::shared_ptr<sh::filesystem::FilesystemNode>> node, sh::filesystem::Operation* op)> fctsetvalue, std::function<void(QList<std::shared_ptr<sh::filesystem::FilesystemNode>>, QString link)> fctlinktriggered, int position, int valueWidthHint)
{
    _fctsetvalue = fctsetvalue;
    _fctlinktriggered = fctlinktriggered;
    _position = position;
    _valueWidthHint = valueWidthHint;
}

std::shared_ptr<sh::paneldetails::PanelDetail> sh::paneldetails::PanelDetailFactoryForFunctionMulti::construct(QList<std::shared_ptr<sh::filesystem::FilesystemNode>> nodes, sh::filesystem::Operation* op)
{
    PanelDetail* _result = new PanelDetail(_position, _valueWidthHint);
    std::shared_ptr<PanelDetail> result(_result);
    _fctsetvalue(result, nodes, op);
    QObject::connect(_result, &PanelDetail::linkTriggered, [=](QString linktarget){
        sh::exceptions::Exception::executeGuarded([&](){
            _fctlinktriggered(nodes, linktarget);
        }, sh::exceptions::MANUAL_RETRY_ENABLED|sh::exceptions::AUTOMATIC_RETRY_ENABLED|sh::exceptions::RESUMEABLE_PROGRAM_ERROR_END_WITH_USER_FEEDBACK_HERE);
    });
    return result;
}

sh::paneldetails::PanelDetailRowValueElement::PanelDetailRowValueElement() : _text(QObject::tr("...")), _iswaiting(true)
{
}

sh::paneldetails::PanelDetailRowValueElement::PanelDetailRowValueElement(QString text) : _text(text)
{
}

sh::paneldetails::PanelDetailRowValueElement::PanelDetailRowValueElement(QIcon icon) : _icon(icon)
{
}

sh::paneldetails::PanelDetailRowValueElement::PanelDetailRowValueElement(QString text, QString linktarget, bool autohide)
        : _text(text), _linktarget(linktarget), _linkautohide(autohide)
{
}

QString sh::paneldetails::PanelDetailRowValueElement::text()
{
    return _text;
}

QIcon sh::paneldetails::PanelDetailRowValueElement::icon()
{
    return _icon;
}

QString sh::paneldetails::PanelDetailRowValueElement::linktarget()
{
    return _linktarget;
}

bool sh::paneldetails::PanelDetailRowValueElement::linkautohide()
{
    return _linkautohide;
}

bool sh::paneldetails::PanelDetailRowValueElement::isWaiting()
{
    return _iswaiting;
}
