# Shallot is a Qt based file manager.
# Copyright (C) 2014-2020 Josef Hahn
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


##@namespace pluginmanager A Shallot plugin which allows graphical plugin management and creation of new plugin from many sample templates.

import datetime
import re
import shallot
import shlex
import subprocess
import sys
import time
import os


Strings = shallot.IntlStringMap(
    BuiltinPlugins = {"en":"built-in plugins", "de":"Kern-Plugins"},
    SystemPlugins = {"en":"system plugins", "de":"System-Plugins"},
    UserPlugins = {"en":"user plugins", "de":"Benutzer-Plugins"},
    PluginManagement = {"en":"Plugin Management", "de":"Plugin Management"},
    FindInFilesystem = {"en":"Find in filesystem", "de":"Im Dateisystem finden"},
    CreatePlugin = {"en":"Create plugin", "de":"Plugin erzeugen"},
    ChooseTemplateFromSamples = {"en":"Please choose one of the sample plugins as template.\n\nExperiment with them step by step for learning by doing, but also read more about the classes and methods in the scripting help.\n\nIt takes effect when you start Shallot next time.", "de":"Bitte wählen Sie ein Beispiel-Plugin als Vorlage aus.\n\nProbieren Sie dieses Schritt-für-Schritt aus, lesen Sie aber auch mehr über die Klassen und Methoden in der Skripting-Hilfe.\n\nEs wird beim nächsten Start von Shallot aktiv."},
    Cancel = {"en":"Cancel", "de":"Abbrechen"},
    InputPluginName = {"en":"Please choose a name for the new plugin.", "de":"Bitte wählen Sie einen Namen für das neue Plugin."},
    PluginAlreadyExists = {"en":"A plugin with this name already exists.", "de":"Ein Plugin mit diesem Namen existiert bereits."},
    EditSource = {"en":"Edit source", "de":"Quellcode editieren"},
    ScriptingHelp = {"en":"Scripting help", "de":"Skripting-Hilfe"},
    PluginManagerActivated = {"en":"The Plugin Manager is activated. Find it in the root node 'Plugin Management'.", "de":"Der Plugin-Manager ist aktiviert und im Wurzelknoten 'Plugin Management' verfügbar."},
    EnablePluginManagement = {"en":"Enable the plugin management", "de":"Plugin Management aktivieren"},
    CreatingDirectoryForbidden = {"en":"Creating directories is not allowed.", "de":"Das Anlegen von Verzeichnissen ist nicht erlaubt."},
    FileOperationForbidden = {"en":"This file operation is not allowed.", "de":"Diese Dateioperation ist nicht erlaubt."},
    CreatingLinksForbidden = {"en":"Creating links is not allowed.", "de":"Das Anlegen von Links ist nicht erlaubt."},
    PluginState = {"en":"State", "de":"Zustand"},
    StateEnabled = {"en":"Enabled", "de":"Aktiviert"},
    StateDisabled = {"en":"Disabled", "de":"Deaktiviert"},
    EnablePlugin = {"en":"Enable", "de":"Aktivieren"},
    DisablePlugin = {"en":"Disable", "de":"Deaktivieren"},
)

class MyFSH(shallot.Filesystem.Handler):

    def __init__(self):
        shallot.Filesystem.Handler.__init__(self)

        self._builtinplugins = Strings.BuiltinPlugins
        self._systemplugins = Strings.SystemPlugins
        self._userplugins = Strings.UserPlugins
        self.mainlocations = {
            self._builtinplugins: (shallot.Environment.shallot_builtinplugin_dir,),
            self._systemplugins: (shallot.Environment.shallot_systemplugin_dir,),
            self._userplugins: (shallot.Environment.shallot_userplugin_dir,),
        }

        self.ispluginenableddetailcolumn=IsPluginEnabledDetailColumn(self)

        self.rootnode = shallot.Filesystem.create_node(
                       shallot.Eurl.create("plugins","","/"),
                       self,
                       shallot.Filesystem.NodeType.Directory,
                       shallot.Filesystem.rootnode)
        self.rootnode.set_icon("plugins")
        self.rootnode.set_displayname(Strings.PluginManagement)


    def get_mimetype(self, operation, eurl):
        return "application/shallot-plugin"
    def get_type(self, operation, eurl):
        if eurl.path()=="/":
            return shallot.Filesystem.NodeType.Directory
        elif eurl.parent_segment().path()=="/":
            if eurl.basename() in self.mainlocations.keys():
                return shallot.Filesystem.NodeType.Directory
        elif eurl.parent_segment().parent_segment().path()=="/":
            for mainlockey in self.mainlocations:
                mainlocpath, = self.mainlocations[mainlockey]
                if os.path.exists(mainlocpath):
                    if os.path.isfile(self.getactualname(mainlocpath + "/" + eurl.basename())):
                        return shallot.Filesystem.NodeType.File
        return shallot.Filesystem.NodeType.NONE
    def _getLocalPath(self, eurl):
        if eurl.parent_segment() and eurl.parent_segment().parent_segment() and eurl.parent_segment().parent_segment().path()=="/":
            return self.getactualname(self.mainlocations[eurl.parent_segment().basename()][0] + "/" + eurl.basename())
    def get_size(self, operation, eurl):
        localpath = self._getLocalPath(eurl)
        try:
            return os.path.getsize(localpath) if localpath else 0
        except IOError:
            return 0
    def get_mtime(self, operation, eurl):
        localpath = self._getLocalPath(eurl)
        try:
            return int(os.path.getmtime(localpath)*1000) if localpath else None
        except IOError:
            return None
    def set_mtime(self, operation, eurl, mtime):
        pass
    def can_create_directory(self, operation, eurl):
        return False
    def create_directory(self, operation, eurl):
        raise shallot.Exceptions.IOException(message=Strings.CreatingDirectoryForbidden,detailsAreInteresting=True)
    def can_create_file(self, operation, eurl):
        return False
    def can_delete_item(self, operation, eurl):
        return (eurl.parent_segment() and eurl.parent_segment().parent_segment() and (eurl.parent_segment().parent_segment().parent_segment() is None) and (eurl.parent_segment().basename() in [self._systemplugins, self._userplugins]))
    def delete_item(self, operation, eurl):
        plugin = eurl.basename()
        branch = eurl.parent_segment().basename()
        if branch == self._systemplugins:
            fpath = shallot.Environment.shallot_systemplugin_dir
        elif branch == self._userplugins:
            fpath = shallot.Environment.shallot_userplugin_dir
        fpath += "/" + plugin
        os.unlink(self.getactualname(fpath))
    def can_rename_item(self, operation, src):
        return False
    def rename_item(self, operation, src, destpath):
        return
    def getactualname(self,path):
        if os.path.isfile(path):
            return path
        else:
            return path+".disabled"
    def pluginlist(self, path):
        result=[]
        for x in os.listdir(path):
            if os.path.isfile(path + "/" + x):
                isenabled = True
                if x.lower().endswith(".disabled"):
                    isenabled = False
                    x = x[:-9]
                result.append((x, isenabled))
        return result
    def itemlist(self, operation, eurl, nodetype, nodelist):
        result=[]
        path=eurl.path()
        if path=="/":
            if nodetype==shallot.Filesystem.NodeType.Directory:
                for mainlockey in self.mainlocations:
                    result.append(mainlockey)
        else:
            if nodetype==shallot.Filesystem.NodeType.File:
                for mainlockey in self.mainlocations:
                    mainlocpath, = self.mainlocations[mainlockey]
                    if path=="/"+mainlockey:
                        if os.path.exists(mainlocpath):
                            for pname, pisenabled in self.pluginlist(mainlocpath):
                                result.append(pname)
        nodelist.set_items(result)
    def configure_item(self, operation, node):
        if node.eurl().basename() in self.mainlocations.keys():
            node.set_icon("folder*plugin")
        else:
            node.add_detail(self.ispluginenableddetailcolumn)
            node.set_icon("plugins")
    def can_get_filecontent(self, operation, eurl):
        return False
    def get_file_content(self, operation, eurl):
        return None
    def create_file(self, operation, eurl, outstream, handlertransfer):
        lpath = None
        if eurl.parent_segment() and eurl.parent_segment().parent_segment() and eurl.parent_segment().parent_segment().parent_segment() is None:
            parentbase = eurl.parent_segment().basename()
            if parentbase in [self._userplugins, self._systemplugins]:
                lpath = self.mainlocations[parentbase][0] + "/"
        if lpath is None:
            raise shallot.Exceptions.IOException(message=Strings.FileOperationForbidden,detailsAreInteresting=True)
        lpath += eurl.basename()
        with open(lpath, "wb") as f:
            f.write(outstream.readall())
    def get_actions(self, eurls):
        if len(eurls) == 1:
            eurl = eurls[0]
            eb = eurl.basename()
            actlist = [NewPluginAction(self, shallot.Environment.shallot_userplugin_dir), HelpAction()]
            if eurl.parent_segment() is not None and eurl.parent_segment().path()=="/":
                if eb == self._userplugins:
                    actlist += [FindInFilesystemAction(self, shallot.Environment.shallot_userplugin_dir)]
                elif eb == self._systemplugins:
                    actlist += [FindInFilesystemAction(self, shallot.Environment.shallot_systemplugin_dir)]
                elif eb == self._builtinplugins:
                    actlist += [FindInFilesystemAction(self, shallot.Environment.shallot_builtinplugin_dir)]
            elif eurl.parent_segment() and eurl.parent_segment().parent_segment() and eurl.parent_segment().parent_segment().path()=="/":
                if eurl.parent_segment().basename() in [self._systemplugins, self._userplugins]:
                    actlist += [EditPluginAction(self, eurl)]
                    if self._getLocalPath(eurl).lower().endswith(".disabled"):
                        actlist += [EnablePluginAction(self, eurl)]
                    else:
                        actlist += [DisablePluginAction(self, eurl)]
            return actlist
        return []
    def can_create_link(self, operation, eurl):
        return False
    def get_linktarget(self, operation, eurl):
        return None
    def create_link(self, operation, eurl, tgt):
        raise shallot.Exceptions.IOException(message=Strings.CreatingLinksForbidden,detailsAreInteresting=True)


class IsPluginEnabledDetailColumn(shallot.DetailColumn):

    def __init__(self, handler):
        shallot.DetailColumn.__init__(self, "ispluginenabled", Strings.PluginState, positionGroup=shallot.DetailColumn.INDEX_CORE, positionIndex=10)
        self.handler = handler

    def determine_value(self, node, op):
        if node.nodetype()==shallot.Filesystem.NodeType.File:
            if self.handler._getLocalPath(node.eurl()).lower().endswith(".disabled"):
                return Strings.StateDisabled
            else:
                return Strings.StateEnabled
        else:
            return ""


class FindInFilesystemAction(shallot.Actions.ActionAction):

    def __init__(self, handler, dir):
        self.handler = handler
        self.dir = dir
        shallot.Actions.ActionAction.__init__(self, Strings.FindInFilesystem, icon="plugin+folder")

    def action(self, info):
        if not os.path.exists(self.dir):
            try:
                os.makedirs(self.dir)
            except:
                pass
        shallot.MainWindow.current().jump_to_eurl(shallot.Eurl.from_string("file://"+self.dir))


class NewPluginAction(shallot.Actions.ActionAction):

    def __init__(self, handler, dir):
        shallot.Actions.ActionAction.__init__(self, Strings.CreatePlugin, icon="plugin+list-add")
        self.handler = handler
        self.dir = dir
        self.euserplugins = shallot.Eurl.from_string("plugins:///" + handler._userplugins)

    def action(self, info):
        info.set_head(Strings.CreatePlugin)
        spath = shallot.Environment.shallot_data_dir + "/scriptsamples/"
        pluginlist = sorted([x for x in os.listdir(spath) if os.path.isfile(spath+x)])
        r = info.userfeedback().messagebox(Strings.ChooseTemplateFromSamples,
                        pluginlist + [Strings.Cancel], "plugins", -1, len(pluginlist))
        if r < len(pluginlist):
            fpath = spath + pluginlist[r]
            with open(fpath, "r") as f:
                cnt = f.readlines()
            destname = info.userfeedback().simple_inputbox(Strings.InputPluginName, "my_" + pluginlist[r][3:-3].replace(".", "_") + "_plugin")
            if destname is not None:
                currlst = info.operation().filesystem().itemlist(self.euserplugins)
                i = 2
                destname = _destname = destname.replace("/", "_").replace("\\", "_").replace(".", "_").replace(":", "_")
                while [x for x in currlst if x.basename().lower() == destname.lower()+".py"] != []:
                    destname = _destname + "_" + str(i)
                    i += 1
                destname += ".py"
                tpath = self.dir + "/" + destname
                if os.path.exists(tpath) or os.path.exists(tpath+".disabled"):
                    raise shallot.Exceptions.IOException(message=Strings.PluginAlreadyExists,detailsAreInteresting=True)
                if not os.path.exists(self.dir):
                    os.makedirs(self.dir)
                with open(tpath, "w") as f:
                    for l in cnt:
                        f.write(l)
                info.add_changed_eurl(self.euserplugins.with_appended_segment(destname))


class EditPluginAction(shallot.Actions.ActionAction):

    def __init__(self, handler, eurl):
        shallot.Actions.ActionAction.__init__(self, Strings.EditSource, icon="text-x-generic")
        self.handler = handler
        self.eurl = eurl

    def action(self, info):
        global editsourcecommandvar
        editsourcecommand = editsourcecommandvar.value()
        lpath = self.handler._getLocalPath(self.eurl)
        if editsourcecommand:
            llpath = shlex.quote(lpath)
            if len(llpath) >= 2 and llpath[0] == "'" and llpath[-1] == "'":
                llpath = '"' + llpath[1:-1].replace('"', '\\"') + '"'
            editsourcecommand = re.sub(r'\b\s*"?%f"?\s*\b', " " + llpath + " ", editsourcecommand, flags=re.IGNORECASE)
            info.set_visualprocessfeedback_active(False)
            oldpypath = os.getenv("PYTHONPATH", "")
            os.putenv("PYTHONPATH", oldpypath + os.path.pathsep + shallot.Environment.shallot_data_dir)
            subprocess.call([editsourcecommand], shell=True)
            os.putenv("PYTHONPATH", oldpypath)
        else:
            shallot.MainWindow.open_items([shallot.Eurl.create("file", None, lpath)])


class EnablePluginAction(shallot.Actions.ActionAction):

    def __init__(self, handler, eurl):
        shallot.Actions.ActionAction.__init__(self, Strings.EnablePlugin, icon="media-playback-start")
        self.handler = handler
        self.eurl = eurl

    def action(self, info):
        realname = self.handler._getLocalPath(self.eurl)
        if realname.lower().endswith(".disabled"):
            os.rename(realname, realname[:-9])
            info.add_changed_eurl(self.eurl)


class DisablePluginAction(shallot.Actions.ActionAction):

    def __init__(self, handler, eurl):
        shallot.Actions.ActionAction.__init__(self, Strings.DisablePlugin, icon="media-playback-stop")
        self.handler = handler
        self.eurl = eurl

    def action(self, info):
        realname = self.handler._getLocalPath(self.eurl)
        if not realname.lower().endswith(".disabled"):
            os.rename(realname, realname+".disabled")
            info.add_changed_eurl(self.eurl)


class HelpAction(shallot.Actions.ActionAction):

    def __init__(self):
        shallot.Actions.ActionAction.__init__(self, Strings.ScriptingHelp, icon="help-hint")

    def action(self, info):
        shallot.MainWindow.open_items([shallot.Eurl.create("file", None, shallot.Environment.shallot_data_dir+"/scripting.pdf")])


class ShowWelcomeMessageAction(shallot.Actions.ActionAction):

    def __init__(self):
        shallot.Actions.ActionAction.__init__(self, "x")

    def action(self, info):
        info.set_head(Strings.PluginManagement)
        info.userfeedback().simple_messagebox(Strings.PluginManagerActivated, shallot.Actions.ExecutionUserFeedback.MessageBoxButton.OK, None, shallot.Actions.ExecutionUserFeedback.MessageBoxButton.OK, 0)


class PluginManagerEnabledConfigurationValue(shallot.ConfigurationValue):

    def __init__(self):
        shallot.ConfigurationValue.__init__(self, "pluginmanager.enable", defaultvalue=False,
                        category=shallot.ConfigurationValue.Category.Behavior,
                        description=Strings.EnablePluginManagement)


class PluginManagerWelcomeShownConfigurationValue(shallot.ConfigurationValue):

    def __init__(self):
        shallot.ConfigurationValue.__init__(self, "pluginmanager.welcomeshown", defaultvalue=False)


class PluginManagerEditSourceCommandConfigurationValue(shallot.ConfigurationValue):

    def __init__(self):
        shallot.ConfigurationValue.__init__(self, "pluginmanager.editsourcecommand", defaultvalue="")


if PluginManagerEnabledConfigurationValue().value():
    shallot.Filesystem.Handler.register("plugins", MyFSH())
    welcomeshownvar = PluginManagerWelcomeShownConfigurationValue()
    editsourcecommandvar = PluginManagerEditSourceCommandConfigurationValue()
    if not welcomeshownvar.value():
        a = ShowWelcomeMessageAction()
        a.initialize_sync()
        a.execute()
        welcomeshownvar.set_value(True)
