/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2020 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef API_H
#define API_H

#include <typeinfo>
#include <QMap>
#include <QString>
#include "scripting/api/apihelpermethods.h"

#ifdef PYTHONSCRIPT
#include "pythonscriptinterpreter.h"
#define REGISTERMETHOD_PYTHON(apiclss,name,nativemethod)  sh::scripting::PythonScriptInterpreter::registerMethod( apiclss, name, nativemethod );
#else
#define REGISTERMETHOD_PYTHON(...)  ;
#endif

#define REGISTERMETHOD(apiclss, cclss, name, retval, meth, ...) \
    registerMethod(apiclss, name); \
    REGISTERMETHOD_PYTHON( apiclss, name, ((PyCFunction) &sh::scripting::pythoninterop::CallNativeMethodWithPyObjectArguments< retval(cclss::*)(__VA_ARGS__), &meth >::call) );

#define REGISTERMETHODCONST(apiclss, cclss, name, retval, meth, ...) \
    registerMethod(apiclss, name); \
    REGISTERMETHOD_PYTHON( apiclss, name, ((PyCFunction) &sh::scripting::pythoninterop::CallNativeMethodWithPyObjectArguments< retval(cclss::*)(__VA_ARGS__) const, &meth >::call) );

#define REGISTERHELPERMETHOD(apiclss, cclss, name, retval, meth, ...) \
    registerMethod(apiclss, name); \
    REGISTERMETHOD_PYTHON( apiclss, name, ((PyCFunction) &sh::scripting::pythoninterop::CallNativeHelperMethodWithPyObjectArguments< retval(*)(cclss*, ##__VA_ARGS__), &meth >::call) ); // gnu special feature: ##__VA_ARGS__

#define REGISTERSCRIPTMETHOD(apiclss, cclss, name, retval, meth, ...) \
    registerMethod(apiclss, name); \
    REGISTERMETHOD_PYTHON( apiclss, name, ((PyCFunction) &sh::scripting::pythoninterop::ScriptedMethodProxy< cclss, std::function<retval(__VA_ARGS__)> cclss::*, &meth >::pyObjectFunctionSetImplementation) );


namespace sh{
    namespace scripting {

        /** \brief Data structure for one api-aware core method.
         */
        class ApiMethod
        {
        public:
            /** \brief Constructed only indirectly.
             */
            ApiMethod(QString name);
            QString name();
        protected:
            QString _name;
        };

        template<class V>
        using StringMap = QMap<QString, V>;

        /** \brief Data structure for one api-aware core class.
         */
        class ApiClass
        {
        public:
            /** \brief Constructed only indirectly.
             */
            ApiClass(QString name);
            ~ApiClass();
            QString name();
            QStringList methods();
            void addMethod(ApiMethod* m);
            ApiMethod *getMethod(QString name);
        private:
            QString _name;
            QHash<QString, ApiMethod*> _methods;
        };

        /** \brief Interoperation layer between the Shallot core and a plugin interpreter.
         */
        class Api
        {
        public:
            Api();
            ~Api();

            QStringList classes();
            QStringList rootMembers();

            ApiClass* getClass(QString name);
            ApiClass* getClassByTypeIdName(QString typeidName);
            void* rootMember(QString name);
            ApiClass* rootMemberClass(QString name);

        private:
            void registerClass(QString name, ApiClass** apiclass, const std::type_info& nativeType);
            void registerMethod(ApiClass* apiclass, QString name);
            void registerRootObjectMember(QString name, ApiClass* cls, void* member);
            QHash<QString, ApiClass*> _classes;
            QHash<QString, ApiClass*> _classesByNativeType;
            QHash<QString, void*> _rootMembers;
            QHash<QString, ApiClass*> _rootMembersClasses;

        };

    }
}

#endif // API_H
