/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2020 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "actions/common/actionabstracttransfertree.h"
#include "filesystem/filesystemnode.h"
#include "filesystem/eurl.h"
#include "filesystem/operation.h"
#include "detailcolumns/detailcolumnextendedattributes.h"
#include "base/singletoninitializer.h"
#include "base/threadpool.h"
#include "base/mainthread.h"
#include "base/iconmanager.h"
#include "ui/filepropertydialogtabactionsview.h"
#include "apifilepropertydialogtab.h"


const int sh::scripting::api::ApiFilePropertyDialogTab::PROPERTY_TYPE_STRING = 1;

const int sh::scripting::api::ApiFilePropertyDialogTab::PROPERTY_TYPE_STRINGMAP = 2;

const int sh::scripting::api::ApiFilePropertyDialogTab::PROPERTY_TYPE_ICONTEXTBANNER = 3;

sh::scripting::api::ApiFilePropertyDialogTab::ApiFilePropertyDialogTab(QString title, QList<PropertyConfig> properties)
                : _title(title), _properties(properties)
{
}

sh::scripting::api::ApiFilePropertyDialogTab::PropertyConfig::PropertyConfig(QString title, int propertytype, QList<ButtonConfig> buttons)
                : title(title), propertytype(propertytype), buttons(buttons)
{
}

QString sh::scripting::api::ApiFilePropertyDialogTab::title()
{
    return _title;
}

QList<QString> sh::scripting::api::ApiFilePropertyDialogTab::properties()
{
    QList<QString> result;
    foreach(PropertyConfig p, _properties)
        result.append(p.title);
    return result;
}

void sh::scripting::api::ApiFilePropertyDialogTab::populateWidget(int i, ui::FilePropertyDialogTabActionsView *widget)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    int ptype = _properties[i].propertytype;
    sh::ui::FilePropertyDialogTabViewContent* content = 0;
    if (ptype == PROPERTY_TYPE_STRING)
        content = dialog()->createTabViewText();
    else if (ptype == PROPERTY_TYPE_STRINGMAP)
        content = dialog()->createTabViewTable();
    else if (ptype == PROPERTY_TYPE_ICONTEXTBANNER)
        content = dialog()->createTabViewIconTextBanner();
    widget->setContent(content);
    QStringList buttons;
    foreach(PropertyConfig::ButtonConfig btncfg, _properties[i].buttons)
        buttons << btncfg.first;
    widget->setButtons(buttons);
    widget->onButtonTriggered([=](int j){
        sh::base::ThreadPool::enqueueForce([=](){
            sh::exceptions::Exception::executeGuarded_errorpanel([&](){
                _buttonTriggered( _properties[i].buttons[j].second);
            });
        });
    });
}

void sh::scripting::api::ApiFilePropertyDialogTab::updateWidget(int i, sh::ui::FilePropertyDialogTabActionsView* widget, sh::filesystem::Operation* op)
{
    sh::ui::FilePropertyDialogTabActionsView* awidget = dynamic_cast<sh::ui::FilePropertyDialogTabActionsView*>(widget);
    sh::ui::FilePropertyDialogTabTextView* wlabel = dynamic_cast<sh::ui::FilePropertyDialogTabTextView*>(awidget->content());
    sh::ui::FilePropertyDialogTabTableView* wgrid = dynamic_cast<sh::ui::FilePropertyDialogTabTableView*>(awidget->content());
    sh::ui::FilePropertyDialogTabIconTextBannerView* witbanner = dynamic_cast<sh::ui::FilePropertyDialogTabIconTextBannerView*>(awidget->content());
    QList<QString> lvals;
    sh::exceptions::Exception::executeGuarded_errorpanel([&](){
        lvals = _updateWidget(i, op, (qintptr)widget);
    });
    QString curkey;
    QList<QPair<QString, QString>> newval;
    double sizefact = 24;
    foreach(QString s, lvals)
    {
        if (curkey.isNull())
            curkey = s;
        else
        {
            newval.append(QPair<QString,QString>(curkey, s));
            curkey = QString();
        }
    }
    sh::base::MainThread::dispatcher()->invokeSync([&](){
        if (wlabel) {
            foreach (StringPair p, newval)
                if (p.first == "text")
                    wlabel->setContent(p.second);
        }
        else if (wgrid) {
            QList<QPair<QString,QString>> content;
            foreach (StringPair p, newval)
                if (p.first.startsWith("val_"))
                    content.append(QPair<QString,QString>(p.first.mid(4), p.second));
            wgrid->setContent(content);
        }
        else if (witbanner) {
            witbanner->clear();
            foreach (StringPair p, newval) {
                if (p.first.startsWith("val_")) {
                    QString type = p.first.mid(4);
                    QString value = p.second;
                    if (type == "icon") {
                        int iund = value.indexOf(";");
                        int size = value.mid(0, iund).toFloat() * sizefact;
                        value = value.mid(iund+1);
                        witbanner->insertIcon(sh::base::IconManager::instance()->getIconByFullname(value), size);
                    }
                    else if (type == "text")
                        witbanner->insertText(value);
                }
            }
        }
        bool visible = false;
        foreach (StringPair p, newval)
            if (p.first == "visible") {
                visible = p.second == "1";
                break;
            }
        widget->setVisible(visible);
    });
}

sh::scripting::api::ApiFilePropertyDialogTabFactory::ApiFilePropertyDialogTabFactory()
            : sh::ui::FilePropertyDialogTabFactory()
{
}

std::shared_ptr<sh::ui::FilePropertyDialogTab> sh::scripting::api::ApiFilePropertyDialogTabFactory::construct(std::shared_ptr<sh::ui::FilePropertyDialog> dialog)
{
    std::shared_ptr<sh::ui::FilePropertyDialogTab> result;
    sh::exceptions::Exception::executeGuarded_errorpanel([&](){
        result = _construct();
    });
    result->_dialog = dialog;
    return result;
}
