/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2020 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "actions/common/actionopenfile.h"
#include "actions/actionactionitem.h"
#include "actions/actionsmanager.h"
#include "filesystem/filesystemnode.h"
#include "filesystem/filesystemoperation.h"
#include "filesystem/operation.h"
#include "filesystem/filesystemhandlerregister.h"
#include "configuration/configurationmanager.h"
#include "configuration/configurationvalue.h"
#include "scripting/pythonscriptinterpreter.h"
#include "search/searchmanager.h"
#include "search/searchcriterionfactoryfromfunction.h"
#include "base/logger.h"
#include "base/mainthread.h"
#include "base/shallotprocess.h"
#include "base/threadpool.h"
#include "tools/thumbnailmanager.h"
#include "paneldetails/paneldetailmanager.h"
#include "ui/mainwindow.h"
#include "apiglobalobject.h"
#include "apipaneldetailfactory.h"
#include "apifilesystemhandler.h"
#include "apireaddatadevice.h"
#include "apifilesystemoperationprogressmonitor.h"
#include "apiaction.h"
#include "apidetailcolumn.h"
#include "apipredicatedactionfactory.h"
#include "apisetting.h"
#include "apithread.h"
#include "apitimer.h"
#include "apisearchcriterion.h"
#include "apithumbnailprovider.h"
#include "apifilepropertydialogtab.h"

REGISTER_SINGLETON(sh::scripting::api, ApiGlobalObject, "base")


sh::scripting::api::ApiGlobalObject::ApiGlobalObject()
{
}

#ifdef QT_DEBUG
void sh::scripting::api::ApiGlobalObject::logDebug(QString msg)
{
    sh::base::Logger::instance()->log(msg, sh::base::LogSeverity::DEBUG, "plugin");
}
#else
void sh::scripting::api::ApiGlobalObject::logDebug(QString)
{
}
#endif

void sh::scripting::api::ApiGlobalObject::logInfo(QString msg)
{
    sh::base::Logger::instance()->log(msg, sh::base::LogSeverity::INFO, "plugin");
}

void sh::scripting::api::ApiGlobalObject::logWarning(QString msg)
{
    sh::base::Logger::instance()->log(msg, sh::base::LogSeverity::WARNING, "plugin");
}

void sh::scripting::api::ApiGlobalObject::logError(QString msg)
{
    sh::base::Logger::instance()->log(msg, sh::base::LogSeverity::ERROR, "plugin");
}

void sh::scripting::api::ApiGlobalObject::register_predicatedactionfactory(std::shared_ptr<sh::scripting::api::ApiActionFactory> f)
{
    sh::base::MainThread::dispatcher()->invokeSync([&]() {
        sh::actions::ActionsManager::instance()->registerActionFactory(f);
    });
}

bool sh::scripting::api::ApiGlobalObject::isDebugBuild()
{
#ifdef QT_DEBUG
    return true;
#else
    return false;
#endif
}

std::shared_ptr<sh::scripting::api::ApiFilesystemHandler> sh::scripting::api::ApiGlobalObject::create_FilesystemHandler()
{
    std::shared_ptr<sh::scripting::api::ApiFilesystemHandler> result;
    sh::base::MainThread::dispatcher()->invokeSync([&]() {
        result = std::make_shared<sh::scripting::api::ApiFilesystemHandler>(sh::filesystem::FilesystemModel::instance().get());
    });
    return result;
}

std::shared_ptr<sh::scripting::api::ApiDetailColumn> sh::scripting::api::ApiGlobalObject::create_DetailColumn(QString name, QString displayName, int displayIndex, bool sort_doTypediff, int defaultWidth, bool isRightAligned)
{
    return std::make_shared<sh::scripting::api::ApiDetailColumn>(name, displayName, displayIndex, sort_doTypediff, defaultWidth, isRightAligned);
}

std::shared_ptr<sh::scripting::api::ApiFilePropertyDialogTab> sh::scripting::api::ApiGlobalObject::create_FilePropertyDialogTab(QString title, QList<QString> properties)
{
    QList<ApiFilePropertyDialogTab::PropertyConfig> props;
    while(!properties.isEmpty())
    {
        QString key = properties.takeFirst();
        int ptype = properties.takeFirst().toInt();
        int actioncnt = properties.takeFirst().toInt();
        QList<ApiFilePropertyDialogTab::PropertyConfig::ButtonConfig> buttons;
        for(int ia=0; ia<actioncnt; ia++)
        {
            QString actionname = properties.takeFirst();
            QString actionid = properties.takeFirst();
            buttons << ApiFilePropertyDialogTab::PropertyConfig::ButtonConfig(actionname, actionid);
        }
        props << sh::scripting::api::ApiFilePropertyDialogTab::PropertyConfig(key, ptype, buttons);
    }
    return std::make_shared<sh::scripting::api::ApiFilePropertyDialogTab>(title, props);
}

std::shared_ptr<sh::scripting::api::ApiFilePropertyDialogTabFactory> sh::scripting::api::ApiGlobalObject::create_FilePropertyDialogTabFactory()
{
    return std::make_shared<sh::scripting::api::ApiFilePropertyDialogTabFactory>();
}

std::shared_ptr<sh::scripting::api::ApiActionActionItem> sh::scripting::api::ApiGlobalObject::create_ActionActionItem(QString text, bool enabled, QString icon, int defaultActionPrecedence)
{
    std::shared_ptr<sh::scripting::api::ApiActionActionItem> r;
    sh::base::MainThread::dispatcher()->invokeSync([&]() {
        r = std::make_shared<sh::scripting::api::ApiActionActionItem>(text, enabled, icon, defaultActionPrecedence);
    });
    return r;
}

std::shared_ptr<sh::scripting::api::ApiSubmenuActionItem> sh::scripting::api::ApiGlobalObject::create_SubmenuActionItem(QString text, bool enabled, QString icon, int defaultActionPrecedence)
{
    return std::make_shared<sh::scripting::api::ApiSubmenuActionItem>(text, enabled, icon, defaultActionPrecedence);
}

std::shared_ptr<sh::filesystem::FilesystemNode> sh::scripting::api::ApiGlobalObject::_createFilesystemNode(std::shared_ptr<sh::filesystem::Eurl> eurl, sh::scripting::api::ApiFilesystemHandler* handler, int nodetype, std::shared_ptr<sh::filesystem::FilesystemNode> parent, bool doinsert, bool showInitialLoadingLabel, bool isHidden)
{
    std::shared_ptr<sh::filesystem::FilesystemNode> result=0;
    sh::base::MainThread::dispatcher()->invokeSync([&]() {
        result = sh::filesystem::FilesystemModel::instance()->createFilesystemNode(std::const_pointer_cast<const sh::filesystem::Eurl>(eurl), handler, (sh::filesystem::FilesystemNodeType)nodetype, isHidden, parent, doinsert, showInitialLoadingLabel);
    });
    return result;
}

std::shared_ptr<sh::scripting::api::ApiReadDataDevice> sh::scripting::api::ApiGlobalObject::_createReadDataDevice()
{
    return std::make_shared<sh::scripting::api::ApiReadDataDevice>();
}

std::shared_ptr<sh::scripting::api::ApiActionFactory> sh::scripting::api::ApiGlobalObject::create_ActionFactory(QString category, QList<std::shared_ptr<sh::actions::Predicate>> predicates)
{
    return std::make_shared<sh::scripting::api::ApiActionFactory>(category, predicates);
}

std::shared_ptr<sh::scripting::api::ApiSetting> sh::scripting::api::ApiGlobalObject::createSetting(QString name, QString description, int group, bool isAdvancedSetting, bool isGlobal, bool isPerFileview)
{
    return std::make_shared<sh::scripting::api::ApiSetting>(name, description, (sh::settings::SettingGroup) group, isAdvancedSetting, isGlobal, isPerFileview);
}

std::shared_ptr<sh::scripting::api::ApiPanelDetailFactorySingle> sh::scripting::api::ApiGlobalObject::create_DetailFactorySingle(int position, int valueWidthHint)
{
    return std::make_shared<sh::scripting::api::ApiPanelDetailFactorySingle>(position, valueWidthHint);
}

std::shared_ptr<sh::scripting::api::ApiPanelDetailFactoryMulti> sh::scripting::api::ApiGlobalObject::create_DetailFactoryMulti(int position, int valueWidthHint)
{
    return std::make_shared<sh::scripting::api::ApiPanelDetailFactoryMulti>(position, valueWidthHint);
}

std::shared_ptr<sh::scripting::api::ApiFilesystemOperationProgressMonitor> sh::scripting::api::ApiGlobalObject::create_FilesystemOperationProgressMonitor(sh::actions::ActionExecutionInfo* actionExecution)
{
    return std::make_shared<sh::scripting::api::ApiFilesystemOperationProgressMonitor>(actionExecution);
}

std::shared_ptr<sh::filesystem::FilesystemNode> sh::scripting::api::ApiGlobalObject::_getOrCreateFilesystemNode(std::shared_ptr<sh::filesystem::Eurl> eurl, sh::scripting::api::ApiFilesystemHandler* handler, int nodetype, std::shared_ptr<sh::filesystem::FilesystemNode> parent, bool doinsert, bool showInitialLoadingLabel, bool isHidden)
{
    std::shared_ptr<sh::filesystem::FilesystemNode> result=0;
    sh::base::MainThread::dispatcher()->invokeSync([&]() {
        result = sh::filesystem::FilesystemModel::instance()->getOrCreateFilesystemNode(std::const_pointer_cast<const sh::filesystem::Eurl>(eurl), handler, (sh::filesystem::FilesystemNodeType)nodetype, isHidden, parent, doinsert, showInitialLoadingLabel);
    });
    return result;
}

void sh::scripting::api::ApiGlobalObject::_register_FilesystemHandler(std::shared_ptr<sh::scripting::api::ApiFilesystemHandler> handler, QString scheme)
{
    sh::base::MainThread::dispatcher()->invokeSync([&]() {
        sh::filesystem::FilesystemHandlerRegister::instance()->addHandler(scheme, handler);
    });
}

QList<std::shared_ptr<sh::filesystem::FilesystemNode>> sh::scripting::api::ApiGlobalObject::_findFilesystemNodesForEurl(std::shared_ptr<const sh::filesystem::Eurl> eurl)
{
    return sh::filesystem::FilesystemModel::instance()->findNodesForEurl(eurl);
}

QList<std::shared_ptr<sh::filesystem::FilesystemNode>> sh::scripting::api::ApiGlobalObject::_tryGetFilesystemNodesForEurl(std::shared_ptr<const sh::filesystem::Eurl> eurl)
{
    return sh::filesystem::FilesystemModel::instance()->tryGetNodesForEurl(eurl);
}

std::shared_ptr<sh::filesystem::FilesystemNode> sh::scripting::api::ApiGlobalObject::modelRootNode()
{
    return sh::filesystem::FilesystemModel::instance()->rootNode();
}

void sh::scripting::api::ApiGlobalObject::refreshData(std::shared_ptr<const sh::filesystem::Eurl> eurl, bool forceFindParent, bool withDetails)
{
    sh::filesystem::FilesystemModel::instance()->refreshData(eurl, forceFindParent, withDetails);
}

std::shared_ptr<sh::filesystem::Eurl> sh::scripting::api::ApiGlobalObject::createEurl(QString scheme, QString hostname, QString path)
{
    return std::const_pointer_cast<sh::filesystem::Eurl>(sh::filesystem::Eurl::create(scheme, hostname, path));
}

std::shared_ptr<sh::filesystem::Operation> sh::scripting::api::ApiGlobalObject::createOperation()
{
    return std::make_shared<sh::filesystem::Operation>();
}

sh::scripting::StringMap<QString> sh::scripting::api::ApiGlobalObject::createException()
{
    sh::exceptions::ExceptionData odata;
    sh::exceptions::Exception e(odata);
    return e.data();
}

sh::scripting::StringMap<QString> sh::scripting::api::ApiGlobalObject::createArgumentException()
{
    sh::exceptions::ExceptionData odata;
    sh::exceptions::ArgumentException e(odata);
    return e.data();
}

sh::scripting::StringMap<QString> sh::scripting::api::ApiGlobalObject::createIOException()
{
    sh::exceptions::ExceptionData odata;
    sh::exceptions::IOException e(odata);
    return e.data();
}

sh::scripting::StringMap<QString> sh::scripting::api::ApiGlobalObject::createRuntimeException()
{
    sh::exceptions::ExceptionData odata;
    sh::exceptions::RuntimeException e(odata);
    return e.data();
}

sh::scripting::StringMap<QString> sh::scripting::api::ApiGlobalObject::createProgramException()
{
    sh::exceptions::ExceptionData odata;
    sh::exceptions::ProgramException e(odata);
    return e.data();
}

std::shared_ptr<sh::configuration::ConfigurationValue> sh::scripting::api::ApiGlobalObject::register_ConfigurationValueInt(QString name, int deflt, QString desc, int category, QString longdesc, QString changehint)
{
    return sh::configuration::ConfigurationManager::instance()->registerConfigValue(name, QVariant(deflt), sh::configuration::ConfigurationValue::valueTypeInteger(), desc, (sh::configuration::ConfigurationCategory) category, longdesc, changehint);
}

std::shared_ptr<sh::configuration::ConfigurationValue> sh::scripting::api::ApiGlobalObject::register_ConfigurationValueFloat(QString name, int deflt, QString desc, int category, QString longdesc, QString changehint)
{
    return sh::configuration::ConfigurationManager::instance()->registerConfigValue(name, QVariant(deflt), sh::configuration::ConfigurationValue::valueTypeFloat(), desc, (sh::configuration::ConfigurationCategory) category, longdesc, changehint);
}

std::shared_ptr<sh::configuration::ConfigurationValue> sh::scripting::api::ApiGlobalObject::register_ConfigurationValueBool(QString name, bool deflt, QString desc, int category, QString longdesc, QString changehint)
{
    return sh::configuration::ConfigurationManager::instance()->registerConfigValue(name, QVariant(deflt), sh::configuration::ConfigurationValue::valueTypeBoolean(), desc, (sh::configuration::ConfigurationCategory) category, longdesc, changehint);
}

std::shared_ptr<sh::configuration::ConfigurationValue> sh::scripting::api::ApiGlobalObject::register_ConfigurationValueString(QString name, QString deflt, QString desc, int category, QString longdesc, QString changehint)
{
    return sh::configuration::ConfigurationManager::instance()->registerConfigValue(name, QVariant(deflt), sh::configuration::ConfigurationValue::valueTypeString(), desc, (sh::configuration::ConfigurationCategory) category, longdesc, changehint);
}

void sh::scripting::api::ApiGlobalObject::registerTransferrableDetailColumn(int i, std::shared_ptr<sh::filesystem::DetailColumn> detailColumn)
{
    sh::filesystem::FilesystemOperation::addTransferrableDetailColumn(i, detailColumn);
}

sh::ui::MainWindow* sh::scripting::api::ApiGlobalObject::mainwindow()
{
    return sh::ui::MainWindow::mainWindow();
}

QString sh::scripting::api::ApiGlobalObject::shallotDataDir()
{
    return sh::base::ShallotProcess::shallotDataDir();
}

QString sh::scripting::api::ApiGlobalObject::shallotBuiltinPluginDir()
{
    return sh::base::ShallotProcess::shallotDataDir() + "/plugins";
}

QString sh::scripting::api::ApiGlobalObject::shallotSystemPluginDir()
{
    return "/etc/shallot/plugins";
}

QString sh::scripting::api::ApiGlobalObject::shallotUserPluginDir()
{
    return sh::base::ShallotProcess::userDataDir() + "/plugins";
}

QString sh::scripting::api::ApiGlobalObject::shallotLanguage()
{
    return QLocale::system().name();
}

void sh::scripting::api::ApiGlobalObject::registerSetting(std::shared_ptr<sh::settings::Setting> setting)
{
    new sh::settings::SettingsRegistration(setting);
}

void sh::scripting::api::ApiGlobalObject::openItems(QList<std::shared_ptr<const sh::filesystem::Eurl>> eurls)
{
    QList<std::shared_ptr<sh::filesystem::FilesystemNode>> nodes;
    foreach(std::shared_ptr<const sh::filesystem::Eurl> eurl, eurls)
    {
        QList<std::shared_ptr<sh::filesystem::FilesystemNode>> _nodes = sh::filesystem::FilesystemModel::instance()->findNodesForEurl(eurl);
        if (_nodes.isEmpty())
            throw sh::exceptions::ArgumentException(sh::exceptions::ExceptionData().detailsAreInteresting().details(QObject::tr("Unable to find node for: ") + eurl->asString()));
        nodes.append(_nodes.first());
    }
    std::shared_ptr<sh::actions::AbstractActionItem> a = std::shared_ptr<sh::actions::AbstractActionItem>(new sh::actions::common::ActionOpenFile(nodes));
    a->initializeSync();
    {
        IGNORE_FORBIDDEN_ASYNC_CALLS
        dynamic_cast<sh::actions::ActionActionItem*>(a.get())->execute();
    }
}

void sh::scripting::api::ApiGlobalObject::registerPanelDetailFactorySingle(std::shared_ptr<sh::scripting::api::ApiPanelDetailFactorySingle> detail)
{
    sh::paneldetails::PanelDetailManager::instance()->registerFactorySingle(detail);
}

void sh::scripting::api::ApiGlobalObject::registerPanelDetailFactoryMulti(std::shared_ptr<sh::scripting::api::ApiPanelDetailFactoryMulti> detail)
{
    sh::paneldetails::PanelDetailManager::instance()->registerFactoryMulti(detail);
}

void sh::scripting::api::ApiGlobalObject::registerFilePropertyDialogTabFactory(int index, std::shared_ptr<sh::scripting::api::ApiFilePropertyDialogTabFactory> factory)
{
    sh::ui::FilePropertyDialog::addTabFactory(index, factory);
}

std::shared_ptr<sh::scripting::api::ApiThread> sh::scripting::api::ApiGlobalObject::create_Thread()
{
    return std::make_shared<sh::scripting::api::ApiThread>();
}

std::shared_ptr<sh::scripting::api::ApiTimer> sh::scripting::api::ApiGlobalObject::create_Timer()
{
    std::shared_ptr<sh::scripting::api::ApiTimer> result;
    sh::base::MainThread::dispatcher()->invokeSync([&](){
        result = std::make_shared<sh::scripting::api::ApiTimer>();
    });
    return result;
}

std::shared_ptr<sh::scripting::api::ApiSearchCriterion> sh::scripting::api::ApiGlobalObject::create_SearchCriterion(std::shared_ptr<sh::scripting::api::ApiSearchCriterionFactory> factory)
{
    return std::make_shared<sh::scripting::api::ApiSearchCriterion>(factory);
}

std::shared_ptr<sh::scripting::api::ApiSearchCriterionFactory> sh::scripting::api::ApiGlobalObject::create_SearchCriterionFactory(QString key, QString description)
{
    return std::make_shared<sh::scripting::api::ApiSearchCriterionFactory>(key, description);
}

std::shared_ptr<sh::scripting::api::ApiThumbnailProvider> sh::scripting::api::ApiGlobalObject::create_ThumbnailProvider()
{
    return std::make_shared<sh::scripting::api::ApiThumbnailProvider>();
}

std::shared_ptr<sh::filesystem::DetailColumn> sh::scripting::api::ApiGlobalObject::findDetailColumnByName(QString name)
{
    return sh::filesystem::DetailColumn::findKnownDetailColumn(name);
}

void sh::scripting::api::ApiGlobalObject::registerSearchCriterionFactory(int idx, std::shared_ptr<sh::scripting::api::ApiSearchCriterionFactory> apicrit)
{
    sh::search::SearchManager::instance()->addFactory(idx, apicrit);
}

void sh::scripting::api::ApiGlobalObject::registerThumbnailProvider(int idx, std::shared_ptr<sh::scripting::api::ApiThumbnailProvider> thumbnailprovider)
{
    sh::tools::ThumbnailManager::instance()->addThumbnailProvider(idx, thumbnailprovider);
}

sh::tools::BookmarkManager* sh::scripting::api::ApiGlobalObject::bookmarkManager()
{
    return sh::tools::BookmarkManager::instance().get();
}

std::shared_ptr<sh::actions::Predicate> sh::scripting::api::ApiGlobalObject::create_OnDirectoriesPredicate()
{
    return std::make_shared<sh::actions::OnDirectoriesPredicate>();
}

std::shared_ptr<sh::actions::Predicate> sh::scripting::api::ApiGlobalObject::create_OnFilesPredicate()
{
    return std::make_shared<sh::actions::OnFilesPredicate>();
}

std::shared_ptr<sh::actions::Predicate> sh::scripting::api::ApiGlobalObject::create_OnLinksPredicate()
{
    return std::make_shared<sh::actions::OnLinksPredicate>();
}

std::shared_ptr<sh::actions::Predicate> sh::scripting::api::ApiGlobalObject::create_OnSingleEntrySelectionPredicate()
{
    return std::make_shared<sh::actions::OnSingleEntrySelectionPredicate>();
}

std::shared_ptr<sh::actions::Predicate> sh::scripting::api::ApiGlobalObject::create_DontResolveLinksPredicate()
{
    return std::make_shared<sh::actions::DontResolveLinksPredicate>();
}

std::shared_ptr<sh::actions::Predicate> sh::scripting::api::ApiGlobalObject::create_HideOnCurrentDirectoryLevelPredicate()
{
    return std::make_shared<sh::actions::HideOnCurrentDirectoryLevelPredicate>();
}

std::shared_ptr<sh::actions::Predicate> sh::scripting::api::ApiGlobalObject::create_HideOnSelectionLevelPredicate()
{
    return std::make_shared<sh::actions::HideOnSelectionLevelPredicate>();
}

std::shared_ptr<sh::actions::Predicate> sh::scripting::api::ApiGlobalObject::create_ByRegExpPredicate(QString pattern)
{
    return std::make_shared<sh::actions::ByRegExpPredicate>(pattern);
}

std::shared_ptr<sh::actions::Predicate> sh::scripting::api::ApiGlobalObject::create_KeyShortcutPredicate(QString shortcut, bool triggersOnCurrentDirectoryLevel)
{
    return std::make_shared<sh::actions::KeyShortcutPredicate>(shortcut, triggersOnCurrentDirectoryLevel);
}

std::shared_ptr<sh::actions::Predicate> sh::scripting::api::ApiGlobalObject::create_PositionIndexPredicate(int index)
{
    return std::make_shared<sh::actions::PositionIndexPredicate>(index);
}

