/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2020 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "actions/actionexecutionuserfeedback.h"
#include "filesystem/filesystemnodelist.h"
#include "filesystem/filesystemoperation.h"
#include "exceptions/ioexception.h"
#include "configuration/configurationvalue.h"
#include "base/mainthread.h"
#include "base/iconmanager.h"
#include "base/threadpool.h"
#include "paneldetails/paneldetailmanager.h"
#include "ui/mainwindow.h"
#include "ui/filepropertydialogtabactionsview.h"
#include "apiaction.h"
#include "apihelpermethods.h"
#include "apireaddatadevice.h"
#include "apifilesystemoperationprogressmonitor.h"
#include "apifilepropertydialogtab.h"


QByteArray sh::scripting::api::ApiHelperMethods::QIODevice_read(QIODevice *self)
{
    const int BUFSIZE = 4096;
    char buf[BUFSIZE];
    int l = self->read(buf, BUFSIZE);
    if (l==-1)
        throw sh::exceptions::IOException(sh::exceptions::ExceptionData().detailsAreInteresting().details(QObject::tr("Errors while reading from IO device.")));
    return QByteArray(buf, l);
}

QByteArray sh::scripting::api::ApiHelperMethods::QIODevice_readall(QIODevice *self)
{
    QByteArray res;
    QByteArray a("");
    while(!(a=QIODevice_read(self)).isEmpty())
        res.append(a);
    return res;
}

QByteArray sh::scripting::api::ApiHelperMethods::ApiReadDataDevice_read(sh::scripting::api::ApiReadDataDevice* self)
{
    return QIODevice_read(self);
}

QByteArray sh::scripting::api::ApiHelperMethods::ApiReadDataDevice_readall(sh::scripting::api::ApiReadDataDevice* self)
{
    return QIODevice_readall(self);
}

void sh::scripting::api::ApiHelperMethods::FilesystemNodeList_resetItems(sh::filesystem::FilesystemNodeList* self, int type, QList<std::shared_ptr<sh::filesystem::FilesystemNode>> list)
{
    QSet<std::shared_ptr<sh::filesystem::FilesystemNode>> set;
    foreach(std::shared_ptr<sh::filesystem::FilesystemNode> n, list)
        set.insert(n);
    self->resetItems((sh::filesystem::FilesystemNodeType)type, set);
}

void sh::scripting::api::ApiHelperMethods::FilesystemNode_addDetail(sh::filesystem::FilesystemNode* self, std::shared_ptr<sh::filesystem::DetailColumn> col)
{
    sh::base::MainThread::dispatcher()->invokeSync([&](){
        self->addDetail(col);
    });
}

void sh::scripting::api::ApiHelperMethods::FilesystemNode_setIcon(sh::filesystem::FilesystemNode* self, QString icon)
{
    QIcon _icon = sh::base::IconManager::instance()->getIconByFullname(icon);
    sh::base::MainThread::dispatcher()->invokeSync([&](){
        self->setIcon(_icon);
    });
}

void sh::scripting::api::ApiHelperMethods::FilesystemNode_setDisplayName(sh::filesystem::FilesystemNode* self, QString displayname)
{
    sh::base::MainThread::dispatcher()->invokeSync([&](){
        self->setDisplayName(displayname);
    });
}

bool sh::scripting::api::ApiHelperMethods::FilesystemNode_isHidden(filesystem::FilesystemNode *self)
{
    bool _v = false;
    sh::base::MainThread::dispatcher()->invokeSync([&](){
        _v = self->isHidden();
    });
    return _v;
}

void sh::scripting::api::ApiHelperMethods::FilesystemNode_setHidden(filesystem::FilesystemNode *self, bool v)
{
    sh::base::MainThread::dispatcher()->invokeSync([&](){
        self->setHidden(v);
    });
}

void sh::scripting::api::ApiHelperMethods::ConfigurationValue_setConfigValueInt(sh::configuration::ConfigurationValue* self, int v)
{
    self->setConfigValue(v);
}

void sh::scripting::api::ApiHelperMethods::ConfigurationValue_setConfigValueFloat(sh::configuration::ConfigurationValue* self, double v)
{
    self->setConfigValue(v);
}

void sh::scripting::api::ApiHelperMethods::ConfigurationValue_setConfigValueBool(sh::configuration::ConfigurationValue* self, bool v)
{
    self->setConfigValue(v);
}

void sh::scripting::api::ApiHelperMethods::ConfigurationValue_setConfigValueString(sh::configuration::ConfigurationValue* self, QString v)
{
    self->setConfigValue(v);
}

QList<std::shared_ptr<sh::filesystem::Eurl> > sh::scripting::api::ApiHelperMethods::FilesystemOperation_itemlistByType(sh::filesystem::FilesystemOperation* self, std::shared_ptr<const sh::filesystem::Eurl> eurl, int ntype)
{
    QList<std::shared_ptr<sh::filesystem::Eurl>> res;
    foreach(std::shared_ptr<const sh::filesystem::Eurl> e, self->itemlist(eurl, (sh::filesystem::FilesystemNodeType)ntype))
        res.append(std::const_pointer_cast<sh::filesystem::Eurl>(e));
    return res;
}

QList<std::shared_ptr<sh::filesystem::Eurl> > sh::scripting::api::ApiHelperMethods::FilesystemOperation_itemlist(sh::filesystem::FilesystemOperation* self, std::shared_ptr<const sh::filesystem::Eurl> eurl)
{
    QList<std::shared_ptr<sh::filesystem::Eurl>> res;
    foreach(std::shared_ptr<const sh::filesystem::Eurl> e, self->itemlist(eurl))
        res.append(std::const_pointer_cast<sh::filesystem::Eurl>(e));
    return res;
}

int sh::scripting::api::ApiHelperMethods::FilesystemOperation_getType(sh::filesystem::FilesystemOperation *self, std::shared_ptr<const sh::filesystem::Eurl> eurl)
{
    return (int) self->getType(eurl);
}

void sh::scripting::api::ApiHelperMethods::FilesystemOperation_createFile(sh::filesystem::FilesystemOperation *self, std::shared_ptr<const sh::filesystem::Eurl> eurl, sh::scripting::api::ApiReadDataDevice* content, std::shared_ptr<sh::filesystem::FilesystemOperationProgressMonitor> progressmon)
{
    self->createFile(eurl, content, progressmon);
}

QList<std::shared_ptr<sh::filesystem::FilesystemNode>> sh::scripting::api::ApiHelperMethods::FilesystemOperation_resolveNodeLink(sh::filesystem::FilesystemOperation *self, std::shared_ptr<sh::filesystem::FilesystemNode> node)
{
    return self->resolveNodeLink(node);
}

QList<std::shared_ptr<sh::filesystem::FilesystemNode>> sh::scripting::api::ApiHelperMethods::FilesystemOperation_resolveNodeLink_NonRecursive(sh::filesystem::FilesystemOperation *self, std::shared_ptr<sh::filesystem::FilesystemNode> node)
{
    return self->resolveNodeLink(node, false);
}

std::shared_ptr<sh::filesystem::Eurl> sh::scripting::api::ApiHelperMethods::FilesystemOperation_resolveEurlLink(sh::filesystem::FilesystemOperation *self, std::shared_ptr<const sh::filesystem::Eurl> eurl)
{
    return std::const_pointer_cast<sh::filesystem::Eurl>(self->resolveEurlLink(eurl));
}

std::shared_ptr<sh::filesystem::Eurl> sh::scripting::api::ApiHelperMethods::FilesystemOperation_resolveEurlLink_NonRecursive(sh::filesystem::FilesystemOperation *self, std::shared_ptr<const sh::filesystem::Eurl> eurl)
{
    return std::const_pointer_cast<sh::filesystem::Eurl>(self->resolveEurlLink(eurl, false));
}

void sh::scripting::api::ApiHelperMethods::MainWindow_jumpToEurl(sh::ui::MainWindow* self, std::shared_ptr<const sh::filesystem::Eurl> eurl)
{
    IGNORE_FORBIDDEN_ASYNC_CALLS
    self->jumpToEurl(eurl);
}

std::shared_ptr<sh::filesystem::FilesystemNode> sh::scripting::api::ApiHelperMethods::MainWindow_getCurrentDirectoryNode(sh::ui::MainWindow* self)
{
    std::shared_ptr<sh::filesystem::FilesystemNode> result;
    sh::base::MainThread::dispatcher()->invokeSync([&](){
        result = self->currentDirectory();
    });
    return result;
}

void sh::scripting::api::ApiHelperMethods::PanelDetail_setRow(sh::paneldetails::PanelDetail* self, QString key, QList<QString> value)
{
    QList<sh::paneldetails::PanelDetailRowValueElement> elems;
    while(!value.isEmpty())
    {
        QString type = value.takeFirst();
        QString val = value.takeFirst();
        if (type=="text")
            elems.append(sh::paneldetails::PanelDetailRowValueElement(val));
        else if (type=="icon")
            elems.append(sh::paneldetails::PanelDetailRowValueElement(sh::base::IconManager::instance()->getIconByFullname(val)));
        else if (type=="link")
        {
            QStringList _l = val.split("\n");
            elems.append(sh::paneldetails::PanelDetailRowValueElement(_l[2], _l[0], _l[1]=="1"));
        }
        else if (type=="wait")
        {
            elems.append(sh::paneldetails::PanelDetailRowValueElement());
        }
    }
    self->addOrReplaceRow(sh::paneldetails::PanelDetailRow(key, elems));
}

void sh::scripting::api::ApiHelperMethods::ApiActionActionItem_setEnabled(sh::scripting::api::ApiActionActionItem* self, bool enabled)
{
    self->setEnabled(enabled);
}

bool sh::scripting::api::ApiHelperMethods::ApiActionActionItem_enabled(sh::scripting::api::ApiActionActionItem* self)
{
    return self->enabled();
}

void sh::scripting::api::ApiHelperMethods::ApiActionActionItem_setVisible(sh::scripting::api::ApiActionActionItem* self, bool visible)
{
    self->setVisible(visible);
}

bool sh::scripting::api::ApiHelperMethods::ApiActionActionItem_visible(sh::scripting::api::ApiActionActionItem* self)
{
    return self->visible();
}

void sh::scripting::api::ApiHelperMethods::ApiSubmenuActionItem_setSubitems(sh::scripting::api::ApiSubmenuActionItem* self, QList<std::shared_ptr<sh::actions::AbstractActionItem>> subitems)
{
    self->setSubitems(subitems);
}

void sh::scripting::api::ApiHelperMethods::ApiSubmenuActionItem_setEnabled(sh::scripting::api::ApiSubmenuActionItem* self, bool enabled)
{
    self->setEnabled(enabled);
}

bool sh::scripting::api::ApiHelperMethods::ApiSubmenuActionItem_enabled(sh::scripting::api::ApiSubmenuActionItem* self)
{
    return self->enabled();
}

void sh::scripting::api::ApiHelperMethods::ApiSubmenuActionItem_setVisible(sh::scripting::api::ApiSubmenuActionItem* self, bool visible)
{
    self->setVisible(visible);
}

bool sh::scripting::api::ApiHelperMethods::ApiSubmenuActionItem_visible(sh::scripting::api::ApiSubmenuActionItem* self)
{
    return self->visible();
}

int sh::scripting::api::ApiHelperMethods::OperationStep_conflictResolution(sh::filesystem::FilesystemOperationTransfers::OperationStep* self)
{
    return (int)self->conflictResolution();
}

bool sh::scripting::api::ApiHelperMethods::FilesystemOperationProgressMonitor_hasItemInfo(sh::scripting::api::ApiFilesystemOperationProgressMonitor *self)
{
    return self->hasItemInfo();
}

quint64 sh::scripting::api::ApiHelperMethods::FilesystemOperationProgressMonitor_doneItems(sh::scripting::api::ApiFilesystemOperationProgressMonitor *self)
{
    return self->doneItems();
}

quint64 sh::scripting::api::ApiHelperMethods::FilesystemOperationProgressMonitor_allItems(sh::scripting::api::ApiFilesystemOperationProgressMonitor *self)
{
    return self->allItems();
}

bool sh::scripting::api::ApiHelperMethods::FilesystemOperationProgressMonitor_hasBytesInfo(sh::scripting::api::ApiFilesystemOperationProgressMonitor *self)
{
    return self->hasBytesInfo();
}

quint64 sh::scripting::api::ApiHelperMethods::FilesystemOperationProgressMonitor_doneBytes(sh::scripting::api::ApiFilesystemOperationProgressMonitor *self)
{
    return self->doneBytes();
}

quint64 sh::scripting::api::ApiHelperMethods::FilesystemOperationProgressMonitor_allBytes(sh::scripting::api::ApiFilesystemOperationProgressMonitor *self)
{
    return self->allBytes();
}

QString sh::scripting::api::ApiHelperMethods::FilesystemOperationProgressMonitor_getItemInfoFrom(sh::scripting::api::ApiFilesystemOperationProgressMonitor *self)
{
    return self->getItemInfoFrom();
}

QString sh::scripting::api::ApiHelperMethods::FilesystemOperationProgressMonitor_getItemInfoTo(sh::scripting::api::ApiFilesystemOperationProgressMonitor *self)
{
    return self->getItemInfoTo();
}

QString sh::scripting::api::ApiHelperMethods::FilesystemOperationProgressMonitor_estimation(sh::scripting::api::ApiFilesystemOperationProgressMonitor *self)
{
    return self->estimation();
}

QList<std::shared_ptr<sh::filesystem::FilesystemNode>> sh::scripting::api::ApiHelperMethods::ApiFilePropertyDialogTab_nodes(sh::scripting::api::ApiFilePropertyDialogTab* self)
{
    return self->nodes();
}

int sh::scripting::api::ApiHelperMethods::FilePropertyDialogTab_selectedIndexForProperty(ApiFilePropertyDialogTab*, qintptr widget)
{
    int result = -1;
    sh::base::MainThread::dispatcher()->invokeSync([&](){
        sh::ui::FilePropertyDialogTabActionsView* aw = ((sh::ui::FilePropertyDialogTabActionsView*)widget);
        sh::ui::FilePropertyDialogTabTableView* wt = ((sh::ui::FilePropertyDialogTabTableView*) aw->content());
        auto selitms = wt->getSelectedItems();
        if (!selitms.empty())
            result = selitms.first().first;
    });
    return result;
}

void sh::scripting::api::ApiHelperMethods::FilePropertyDialogTab_refresh(ApiFilePropertyDialogTab* self)
{
    self->refresh();
}

int sh::scripting::api::ApiHelperMethods::ActionExecutionUserFeedback_messageBox(sh::actions::ActionExecutionUserFeedback* self, QString m, QList<QString> l, QString s , int i, int j)
{
    return self->messageBox(m, l, s, i, j);
}

QString sh::scripting::api::ApiHelperMethods::ActionExecutionUserFeedback_simpleInputBox(sh::actions::ActionExecutionUserFeedback* self, QString text, QString deflt)
{
    return self->simpleInputBox(text, deflt);
}
