/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2020 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef PYTHONSCRIPTINTERPRETER_H
#define PYTHONSCRIPTINTERPRETER_H

#include <functional>
#include <typeinfo>
#include <QList>
#include <QMap>
#include <QString>
#include "exceptions/exception.h"
#include "scripting/api/apihelpermethods.h"
#include "base/singletoninitializer.h"
#include "base/threaddispatcher.h"
#include "scriptingengine.h"

#ifndef PyObject_HEAD
struct _object;
typedef _object PyObject;
typedef PyObject* (*PyCFunction)(PyObject*, PyObject*);
#endif


namespace sh{
    namespace scripting {
        class Api;
        class ApiClass;

        template<class V>
        using StringMap = QMap<QString, V>;

#ifdef PYTHONSCRIPT

        /** \brief
         * Interoperation layer between the Shallot core (C++) and Python plugins.
         */
        namespace pythoninterop
        {

            PyObject* getPyNone();
            bool isPyNone(PyObject* o);
            void* getPointerFromShallotPyObject(PyObject* o);
            std::shared_ptr<void> getSharedPointerFromShallotPyObject(PyObject* o);
            PyObject* getShallotPyObjectFromPointer(void* o, sh::scripting::ApiClass *apiclass);
            PyObject* getShallotPyObjectFromSharedPointer(std::shared_ptr<void> o, sh::scripting::ApiClass *apiclass);
            sh::scripting::ApiClass* getApiClass(QString typeIdName);

            PyObject* PyTupleGetItem(PyObject* o, int itm);
            PyObject* PyListGetItem(PyObject* o, int itm);
            int PyTupleCount(PyObject* o);
            int PyListCount(PyObject* o);
            bool PyIsTuple(PyObject* o);
            bool PyIsList(PyObject* o);
            bool PyIsUnicode(PyObject* o);
            bool PyIsBytes(PyObject* o);
            bool PyIsDict(PyObject* o);
            bool PyIsBool(PyObject* o);
            bool PyIsFloat(PyObject* o);
            bool PyIsLong(PyObject* o);
            void PyIncRef(PyObject* o);
            void PyDecRef(PyObject* o);
            void PyTupleSetItem(PyObject* o, int itm, PyObject* val);
            PyObject* PyObjectCall(PyObject *fct, PyObject* args);
            PyObject* PyTupleNew(int cnt);
            PyObject* PyDictNew();
            PyObject* PyDictKeys(PyObject* o);
            PyObject* PyDictGetItem(PyObject* dict, PyObject* key);
            void PyDictSetItem(PyObject* dict, PyObject* key, PyObject* val);

            bool PythonToC_ExecuteGuarded(std::function<void()> fct);
            void invokePython(std::function<void()> fct);

            class Converters
            {
            public:
                static QString               getStringFromPyObject        (PyObject* o);
                static QVariant              getVariantFromPyObject       (PyObject* o);
                static int                   getIntFromPyObject           (PyObject* o);
                static double                getDoubleFromPyObject        (PyObject* o);
                static qint64                getInt64FromPyObject         (PyObject* o);
                static quint64               getUint64FromPyObject        (PyObject* o);
                static bool                  getBoolFromPyObject          (PyObject* o);
                static QByteArray            getByteArrayFromPyObject     (PyObject* o);
                static void*                 getPointerFromPyObject       (PyObject* o);
                static std::shared_ptr<void> getSharedPointerFromPyObject (PyObject* o);
                static PyObject* getPyObjectFromString    (QString v);
                static PyObject* getPyObjectFromVariant   (QVariant v);
                static PyObject* getPyObjectFromInt       (int v);
                static PyObject* getPyObjectFromDouble    (double v);
                static PyObject* getPyObjectFromInt64     (qint64 v);
                static PyObject* getPyObjectFromUint64    (quint64 v);
                static PyObject* getPyObjectFromBool      (bool v);
                static PyObject* getPyObjectFromByteArray (QByteArray v);
            };

            template<class V>
            using StringMap = QMap<QString, V>;

            /// Helps to return a c++ value from a Python object (borrows the reference)
            template<class V> class PyObjectToNative;

            /// Helps to return a Python object from a native c++ value (with reference ownership)
            template<class V> class NativeToPyObject;


            template<>
            class NativeToPyObject<QString>
            {
            public:
                static PyObject* toPyObject(QString v)
                {
                    return Converters::getPyObjectFromString(v);
                }
            };

            template<>
            class NativeToPyObject<QVariant>
            {
            public:
                static PyObject* toPyObject(QVariant v)
                {
                    return Converters::getPyObjectFromVariant(v);
                }
            };

            template<>
            class NativeToPyObject<int>
            {
            public:
                static PyObject* toPyObject(int v)
                {
                    return Converters::getPyObjectFromInt(v);
                }
            };

            template<>
            class NativeToPyObject<double>
            {
            public:
                static PyObject* toPyObject(double v)
                {
                    return Converters::getPyObjectFromDouble(v);
                }
            };

            template<>
            class NativeToPyObject<qint64>
            {
            public:
                static PyObject* toPyObject(qint64 v)
                {
                    return Converters::getPyObjectFromInt64(v);
                }
            };

            template<>
            class NativeToPyObject<quint64>
            {
            public:
                static PyObject* toPyObject(quint64 v)
                {
                    return Converters::getPyObjectFromUint64(v);
                }
            };

            template<>
            class NativeToPyObject<bool>
            {
            public:
                static PyObject* toPyObject(bool v)
                {
                    return Converters::getPyObjectFromBool(v);
                }
            };

            template<>
            class NativeToPyObject<QByteArray>
            {
            public:
                static PyObject* toPyObject(QByteArray v)
                {
                    return Converters::getPyObjectFromByteArray(v);
                }
            };

            template<class V>
            class NativeToPyObject<V*>
            {
            public:
                static PyObject* toPyObject(V* v)
                {
                    QString classname = typeid(V).name();
                    sh::scripting::ApiClass* apiclass = getApiClass(classname);
                    return pythoninterop::getShallotPyObjectFromPointer(v, apiclass);
                }
            };

            template<class V>
            class NativeToPyObject<std::shared_ptr<V>>
            {
            public:
                static PyObject* toPyObject(std::shared_ptr<V> v)
                {
                    QString classname = typeid(V).name();
                    sh::scripting::ApiClass* apiclass = getApiClass(classname);
                    return pythoninterop::getShallotPyObjectFromSharedPointer(v, apiclass);
                }
            };

            template<class V>
            class NativeToPyObject<std::shared_ptr<const V>>
            {
            public:
                static PyObject* toPyObject(std::shared_ptr<const V> v)
                {
                    QString classname = typeid(V).name();
                    sh::scripting::ApiClass* apiclass = getApiClass(classname);
                    return pythoninterop::getShallotPyObjectFromSharedPointer(std::const_pointer_cast<V>(v), apiclass);
                }
            };

            template<typename V>
            class NativeToPyObject<QList<V> >
            {
            public:
                static PyObject* toPyObject(QList<V> v)
                {
                    PyObject *tuple = pythoninterop::PyTupleNew(v.length());
                    for(int i=0;i<v.length();i++)
                    {
                        PyObject* innerpy = NativeToPyObject<V>::toPyObject(v[i]);
                        pythoninterop::PyTupleSetItem(tuple, i, innerpy); // this steals the reference, which is perfect
                    }
                    return tuple;
                }
            };

            template<typename V>
            class NativeToPyObject<QMap<QString,V> >
            {
            public:
                static PyObject* toPyObject(QMap<QString,V> v)
                {
                    PyObject *dict = pythoninterop::PyDictNew();
                    foreach(QString key, v.keys())
                    {
                        PyObject* keypy = NativeToPyObject<QString>::toPyObject(key);
                        PyObject* valpy = NativeToPyObject<V>::toPyObject(v[key]);
                        pythoninterop::PyDictSetItem(dict, keypy, valpy); // this does not steal the references
                        pythoninterop::PyDecRef(keypy);
                        pythoninterop::PyDecRef(valpy);
                    }
                    return dict;
                }
            };


            template<>
            class PyObjectToNative<int>
            {
            public:
                static int toNative(PyObject* , PyObject* o)
                {
                    return Converters::getIntFromPyObject(o);
                }
            };

            template<>
            class PyObjectToNative<double>
            {
            public:
                static double toNative(PyObject* , PyObject* o)
                {
                    return Converters::getDoubleFromPyObject(o);
                }
            };

            template<>
            class PyObjectToNative<qint64>
            {
            public:
                static qint64 toNative(PyObject* , PyObject* o)
                {
                    return Converters::getInt64FromPyObject(o);
                }
            };

            template<>
            class PyObjectToNative<quint64>
            {
            public:
                static qint64 toNative(PyObject* , PyObject* o)
                {
                    return Converters::getUint64FromPyObject(o);
                }
            };

            template<>
            class PyObjectToNative<bool>
            {
            public:
                static bool toNative(PyObject* , PyObject* o)
                {
                    return Converters::getBoolFromPyObject(o);
                }
            };

            template<>
            class PyObjectToNative<QByteArray>
            {
            public:
                static QByteArray toNative(PyObject* , PyObject* o)
                {
                    return Converters::getByteArrayFromPyObject(o);
                }
            };

            template<>
            class PyObjectToNative<QString>
            {
            public:
                static QString toNative(PyObject* , PyObject* o)
                {
                    return Converters::getStringFromPyObject(o);
                }
            };

            template<>
            class PyObjectToNative<QVariant>
            {
            public:
                static QVariant toNative(PyObject* , PyObject* o)
                {
                    return Converters::getVariantFromPyObject(o);
                }
            };

            template<class V>
            class PyObjectToNative<QList<V>>
            {
            public:
                static QList<V> toNative(PyObject* a, PyObject* o)
                {
                    QList<V> res;
                    if (pythoninterop::PyIsTuple(o))
                    {
                        for(int i=0;i<pythoninterop::PyTupleCount(o);i++) {
                            PyObject *pyinner = pythoninterop::PyTupleGetItem(o,i);
                            res.append( PyObjectToNative<V>::toNative(a, pyinner) );
                        }
                    }
                    else if (pythoninterop::PyIsList(o))
                    {
                        for(int i=0;i<pythoninterop::PyListCount(o);i++) {
                            PyObject *pyinner = pythoninterop::PyListGetItem(o,i);
                            res.append( PyObjectToNative<V>::toNative(a, pyinner) );
                        }
                    }
                    else
                        throw sh::scripting::ScriptingEngine::ScriptedException( sh::exceptions::ExceptionData().details("A list or tuple was expected but something else was found.").resumeable());
                    return res;
                }
            };

            template<class V>
            class PyObjectToNative<StringMap<V> >
            {
            public:
                static StringMap<V> toNative(PyObject* a, PyObject* o)
                {
                    StringMap<V> res;
                    if (pythoninterop::PyIsDict(o))
                    {
                        PyObject* keys = pythoninterop::PyDictKeys(o);
                        for(int i=0;i<pythoninterop::PyListCount(keys);i++) {
                            PyObject *key = pythoninterop::PyListGetItem(keys,i);
                            PyObject *value = pythoninterop::PyDictGetItem(o, key);
                            QString skey = pythoninterop::Converters::getStringFromPyObject(key);
                            V vval = PyObjectToNative<V>::toNative(a, value);
                            res[skey]=vval;
                        }
                        pythoninterop::PyDecRef(keys);
                    }
                    else
                        throw sh::scripting::ScriptingEngine::ScriptedException(sh::exceptions::ExceptionData().details("A string-dictionary was expected but something else was found.").program().resumeable());
                    return res;
                }
            };

            template<int pos, class ... Args> class _PyObjectFunctionToNativeFunction_SetTupleValue;

            template<int pos>
            class _PyObjectFunctionToNativeFunction_SetTupleValue<pos>
            {
            public:
                static void setTupleValue(PyObject*)
                {
                }
            };


            template<int pos, class Arg, class ... Args>
            class _PyObjectFunctionToNativeFunction_SetTupleValue<pos, Arg, Args...>
            {
            public:
                static void setTupleValue(PyObject* tup, Arg arg, Args... args)
                {
                    PyObject* v= NativeToPyObject<Arg>::toPyObject(arg);
                    pythoninterop::PyTupleSetItem(tup, pos, v);
                    _PyObjectFunctionToNativeFunction_SetTupleValue<pos+1, Args...>::setTupleValue(tup, args...);
                }
            };

            /// This class is a workaround for gcc<4.9 bug 55914. It was not able
            /// to call ...::setTupleValue directly from within the invokePython lambda
            template<int pos, class ... Args>
            class _PyObjectFunctionToNativeFunction_TupleValueSetter
            {
            public:
                _PyObjectFunctionToNativeFunction_TupleValueSetter<pos, Args...>(Args... args) {
                    _set = std::bind(__set, args..., std::placeholders::_1);
                }
                void set(PyObject* tuple){
                    _set(tuple);
                }
            private:
                std::function<void(PyObject*)> _set;
                static void __set(Args... args, PyObject* tuple){
                    _PyObjectFunctionToNativeFunction_SetTupleValue<1,Args...>::setTupleValue(tuple, args...);
                }
            };

            template<class R, class ... Args>
            class PyObjectFunctionToNativeFunction
            {
            public:
                static std::function<R(Args...)> toNativeFunction(PyObject* self, PyObject* fct)
                {
                    return [=](Args... args){
                        R r;
                        _PyObjectFunctionToNativeFunction_TupleValueSetter<1, Args...> tvs(args...);
                        pythoninterop::invokePython([&](){
                            PyObject* vals=pythoninterop::PyTupleNew(sizeof...(Args)+1);
                            pythoninterop::PyTupleSetItem(vals, 0, self);
                            pythoninterop::PyIncRef(self); //PyTupleSetItem steals the reference
                            //PyObjectFunctionToNativeFunction_SetTupleValue<1,Args...>::setTupleValue(vals, args...);
                            tvs.set(vals);
                            PyObject* ret=pythoninterop::PyObjectCall(fct,vals);
                            r=PyObjectToNative<R>::toNative(self, ret);
                            pythoninterop::PyDecRef(ret);
                            pythoninterop::PyDecRef(vals);
                        });
                        return r;
                    };
                }
            };

            template<class ... Args>
            class PyObjectFunctionToNativeFunction<void,Args...>
            {
            public:
                static std::function<void(Args...)> toNativeFunction(PyObject* self, PyObject* fct)
                {
                    return [=](Args... args){
                        _PyObjectFunctionToNativeFunction_TupleValueSetter<1, Args...> tvs(args...);
                        pythoninterop::invokePython([&](){
                            PyObject* vals=pythoninterop::PyTupleNew(sizeof...(Args)+1);
                            pythoninterop::PyTupleSetItem(vals, 0, self);
                            pythoninterop::PyIncRef(self); //PyTupleSetItem steals the reference
                            //PyObjectFunctionToNativeFunction_SetTupleValue<1,Args...>::setTupleValue(vals, args...);
                            tvs.set(vals);
                            PyObject* ret=pythoninterop::PyObjectCall(fct,vals);
                            pythoninterop::PyDecRef(ret);
                            pythoninterop::PyDecRef(vals);
                        });
                        return;
                    };
                }
            };

            template<class V>
            class PyObjectToNative<std::shared_ptr<V>>
            {
            public:
                static std::shared_ptr<V> toNative(PyObject*, PyObject* o)
                {
                    return std::static_pointer_cast<V>(Converters::getSharedPointerFromPyObject(o));
                }
            };

            template<class V>
            class PyObjectToNative<V*>
            {
            public:
                static V* toNative(PyObject*, PyObject* o)
                {
                    return (V*) Converters::getPointerFromPyObject(o);
                }
            };

            template<class V, class ... Args>
            class PyObjectToNative<std::function<V(Args...)>>
            {
            public:
                static std::function<V(Args...)> toNative(PyObject* self, PyObject* io)
                {
                    return PyObjectFunctionToNativeFunction<V,Args...>::toNativeFunction(self,io);
                }
            };

            template < int mode, uint N >
            struct FunctionBindTuple;

            template <uint N >
            struct FunctionBindTuple<0,N>
            {
              template < typename R, typename T, typename... ArgsF, typename... ArgsT, typename... Args >
              static R applyTuple( T* pObj, R (T::*f)( ArgsF... ), const std::tuple<ArgsT...>& t, Args... args )
              {
                  return FunctionBindTuple<0, N-1>::applyTuple( pObj, f, t, std::get<N-1>( t ), args... );
              }
            };

            template <uint N >
            struct FunctionBindTuple<1,N>
            {
              template < typename R, typename T, typename... ArgsF, typename... ArgsT, typename... Args >
              static R applyTuple( T* pObj, R (*f)(T*, ArgsF... ), const std::tuple<ArgsT...>& t, Args... args )
              {
                  return FunctionBindTuple<1, N-1>::applyTuple( pObj, f, t, std::get<N-1>( t ), args... );
              }
            };


            template <>
            struct FunctionBindTuple<0,0>
            {
              template < typename R, typename T, typename... ArgsF, typename... ArgsT, typename... Args >
              static R applyTuple( T* pObj, R (T::*f)( ArgsF... ), const std::tuple<ArgsT...>&, Args... args )
              {
                  return (pObj->*f)( args... );
              }
            };

            template <>
            struct FunctionBindTuple<1,0>
            {
              template < typename R, typename T, typename... ArgsF, typename... ArgsT, typename... Args >
              static R applyTuple( T* pObj, R (*f)( T*, ArgsF... ), const std::tuple<ArgsT...>&, Args... args )
              {
                  return (*f)( pObj, args... );
              }
            };

            template < typename R, typename T, typename... ArgsF, typename... ArgsT >
            R CallNativeFunctionWithTuple( T* pObj, R (T::*f)( ArgsF... ), std::tuple<ArgsT...> const& t )
            {
                return FunctionBindTuple<0,sizeof...(ArgsT)>::applyTuple( pObj, f, t );
            }

            template < typename R, typename T, typename... ArgsF, typename... ArgsT >
            R CallNativeFunctionWithTuple( T* pObj, R (T::*f)( ArgsF... ) const, std::tuple<ArgsT...> const& t )
            {
                return FunctionBindTuple<0,sizeof...(ArgsT)>::applyTuple( pObj, (R(T::*)(ArgsF...)) f, t );
            }

            template < typename R, typename T, typename... ArgsF, typename... ArgsT >
            R CallNativeHelperFunctionWithTuple( T* pObj, R (*f)( T*, ArgsF... ), std::tuple<ArgsT...> const& t )
            {
                return FunctionBindTuple<1,sizeof...(ArgsT)>::applyTuple( pObj, f, t );
            }

            template <int pos, typename ... Args> struct PyObjectTupleToNativeTuple;

            template <int pos, class T, class ... Args>
            struct PyObjectTupleToNativeTuple<pos, T, Args...>
            {
                static std::tuple<T, Args...> toNativeTuple(PyObject* a, PyObject* b){
                    PyObject* io = pythoninterop::PyTupleGetItem(b, pos);
                    T v = (T) PyObjectToNative<T>::toNative(a, io);
                    return std::tuple_cat( std::tuple<T>(v), PyObjectTupleToNativeTuple<pos+1, Args...>::toNativeTuple(a,b) );
                }
            };

            template <int pos>
            struct PyObjectTupleToNativeTuple<pos>
            {
                static std::tuple<> toNativeTuple(PyObject*,PyObject*){
                    return std::tuple<>();
                }
            };


            template <typename T, T> struct CallNativeHelperMethodWithPyObjectArguments;

            template <typename T, typename R, typename ...Args, R (*MF)(T*, Args...)>
            struct CallNativeHelperMethodWithPyObjectArguments<R (*)(T*,Args...), MF>
            {
                static PyObject* call(PyObject* a, PyObject* b)
                {
                    T* tc = (T*) getPointerFromShallotPyObject(a);
                    auto tuple = PyObjectTupleToNativeTuple<0, Args...>::toNativeTuple(a,b);
                    R ret;
                    if (!pythoninterop::PythonToC_ExecuteGuarded([&](){
                        ret = CallNativeHelperFunctionWithTuple(tc, MF, tuple);
                    }))
                        return 0;
                    PyObject* pyret = NativeToPyObject<R>::toPyObject(ret);
                    return pyret;
                }
            };

            template <typename T, typename ...Args, void (*MF)(T*, Args...)>
            struct CallNativeHelperMethodWithPyObjectArguments<void (*)(T*,Args...), MF>
            {
                static PyObject* call(PyObject* a, PyObject* b)
                {
                    T* tc = (T*) getPointerFromShallotPyObject(a);
                    auto tuple = PyObjectTupleToNativeTuple<0, Args...>::toNativeTuple(a,b);
                    if (!pythoninterop::PythonToC_ExecuteGuarded([&](){
                        CallNativeHelperFunctionWithTuple(tc, MF, tuple);
                    }))
                        return 0;
                    return getPyNone();
                }
            };

            template <typename T, T> struct CallNativeMethodWithPyObjectArguments;

            template <typename T, typename R, typename ...Args, R (T::*MF)(Args...)>
            struct CallNativeMethodWithPyObjectArguments<R (T::*)(Args...), MF>
            {
                static PyObject* call(PyObject* a, PyObject* b)
                {
                    T* tc = (T*) getPointerFromShallotPyObject(a);
                    auto tuple = PyObjectTupleToNativeTuple<0, Args...>::toNativeTuple(a,b);
                    R ret;
                    if (!pythoninterop::PythonToC_ExecuteGuarded([&](){
                        ret = CallNativeFunctionWithTuple(tc, MF, tuple);
                    }))
                        return 0;
                    PyObject* pyret = NativeToPyObject<R>::toPyObject(ret);
                    return pyret;
                }
            };

            template <typename T, typename ...Args, void (T::*MF)(Args...)>
            struct CallNativeMethodWithPyObjectArguments<void (T::*)(Args...), MF>
            {
                static PyObject* call(PyObject* a, PyObject* b)
                {
                    T* tc = (T*) getPointerFromShallotPyObject(a);
                    auto tuple = PyObjectTupleToNativeTuple<0, Args...>::toNativeTuple(a,b);
                    if (!pythoninterop::PythonToC_ExecuteGuarded([&](){
                        CallNativeFunctionWithTuple(tc, MF, tuple);
                    }))
                        return 0;
                    return getPyNone();
                }
            };

            template <typename T, typename R, typename ...Args, R (T::*MF)(Args...) const>
            struct CallNativeMethodWithPyObjectArguments<R (T::*)(Args...) const, MF>
            {
                static PyObject* call(PyObject* a, PyObject* b)
                {
                    T* tc = (T*) getPointerFromShallotPyObject(a);
                    auto tuple = PyObjectTupleToNativeTuple<0, Args...>::toNativeTuple(a,b);
                    R ret;
                    if (!pythoninterop::PythonToC_ExecuteGuarded([&](){
                        ret = CallNativeFunctionWithTuple(tc, MF, tuple);
                    }))
                        return 0;
                    PyObject* pyret = NativeToPyObject<R>::toPyObject(ret);
                    return pyret;
                }
            };

            template <typename T, typename ...Args, void (T::*MF)(Args...) const>
            struct CallNativeMethodWithPyObjectArguments<void (T::*)(Args...) const, MF>
            {
                static PyObject* call(PyObject* a, PyObject* b)
                {
                    T* tc = (T*) getPointerFromShallotPyObject(a);
                    auto tuple = PyObjectTupleToNativeTuple<0, Args...>::toNativeTuple(a,b);
                    if (!pythoninterop::PythonToC_ExecuteGuarded([&](){
                        CallNativeFunctionWithTuple(tc, MF, tuple);
                    }))
                        return 0;
                    return getPyNone();
                }
            };

            template <typename D, typename T, T> struct ScriptedMethodProxy;

            template <typename T, typename R, typename ...Args, std::function<R(Args...)> (T::* MF) >
            struct ScriptedMethodProxy<T,std::function<R(Args...)> T::*, MF>
            {
                static void setImplementation(T* inst, std::function<R(Args...)> vfct)
                {
                    inst->*MF=vfct;
                }

                static PyObject* pyObjectFunctionSetImplementation(PyObject* a, PyObject* b)
                {
                    CallNativeHelperMethodWithPyObjectArguments< void(*)(T*, std::function<R(Args...)>), setImplementation >::call(a, b);
                    return getPyNone();
                }
            };

        }

#endif

        /** \brief
         * Integration of the Python language for scripting.
         *
         * Find also the scripting manual for all details about the scripting interface.
         */
        class PythonScriptInterpreter : public sh::scripting::ScriptInterpreter, public sh::base::Singleton
        {
            DECLARE_SINGLETON(PythonScriptInterpreter)
        protected:
            void doShutdown() override;
        private:
            PythonScriptInterpreter();
#ifdef PYTHONSCRIPT
        public:
            ~PythonScriptInterpreter();
            bool isAvailable();
            void initializeInterpreter();
            QString filesuffix();
            void execute(QString script);
            static void registerMethod(sh::scripting::ApiClass* clss, QString name, PyCFunction fctptr);
            static void throwNativeExceptionFromPythonTraceback();
            PyObject *_Exception;
            PyObject *_CancelException;
            PyObject *_ThreadAbortException;
            static QHash<void*, PyObject*> _buddies;
            static QHash<void*, std::weak_ptr<void>> _natives;

        private:
            static void* apiModuleDef;
            static PyObject *getApiModuleDef();
            static sh::scripting::Api* api;
            static QHash<QString, PyCFunction> _methods;
            static PyObject *modtraceback;
#endif
        };

    }
}

#endif // PYTHONSCRIPTINTERPRETER_H
