/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2020 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "actions/actionexecutioninfo.h"
#include "actions/actionexecutionuserfeedback.h"
#include "configuration/configurationmanager.h"
#include "base/logger.h"
#include "base/mainthread.h"
#include "base/threadpool.h"
#include "base/shallotprocess.h"
#include "ui/mainwindow.h"
#include "pythonscriptinterpreter.h"
#include "scriptingengine.h"

REGISTER_SINGLETON(sh::scripting, ScriptingEngine, "scripting")


std::shared_ptr<sh::configuration::ConfigurationValue> sh::scripting::ScriptingEngine::cfgvalLastStartupFailed =
        sh::configuration::ConfigurationManager::instance()->registerConfigValue(
            "sh::scripting::ScriptingEngine.LastStartupFailed", false,
            sh::configuration::ConfigurationValue::valueTypeBoolean());

void sh::scripting::ScriptingEngine::doInitialize()
{
    QStringList pluginpaths;
    registerInterpreter(sh::scripting::PythonScriptInterpreter::instance());
    pluginpaths << sh::base::ShallotProcess::shallotDataDir() + "/plugins";
    pluginpaths << "/etc/shallot/plugins";
    pluginpaths << sh::base::ShallotProcess::userDataDir() + "/plugins";
    sh::base::ThreadPool::enqueueForce([=](){
        sh::exceptions::Exception::executeGuarded([&](){
            foreach(QString pluginpath, pluginpaths)
                foreach(QString plugin, QDir(pluginpath).entryList(QDir::Files))
                    try
                    {
                        loadScript(pluginpath + "/" + plugin);
                    }
                    catch(sh::exceptions::Exception& ex)
                    {
                        if (cfgvalLastStartupFailed->getConfigValueBool())
                        {
                            QMutex mutex;
                            mutex.lock();
                            std::shared_ptr<sh::actions::ActionActionItem> act = std::make_shared<ActionPluginLoadCrashedAgain>(pluginpath + "/" + plugin, &mutex);
                            act->initializeSync();
                            act->execute();
                            while (!mutex.tryLock(1000))
                                sh::base::ThreadPool::respectThreadAbort();
                            mutex.unlock();
                        }
                        else
                            cfgvalLastStartupFailed->setConfigValue(true);
                        throw ex;
                    }
        }, sh::exceptions::ExecuteGuardFlag::ALL_ERRORS_KILL_SHALLOT);
        cfgvalLastStartupFailed->setConfigValue(false);
    });
}

sh::scripting::ScriptingEngine::ScriptingEngine()
{
}

void sh::scripting::ScriptingEngine::registerInterpreter(std::shared_ptr<sh::scripting::ScriptInterpreter> interpreter)
{
    if (interpreter->isAvailable()) {
        interpreter->initializeInterpreter();
        _interpreters.insert(interpreter->filesuffix(), interpreter);
    }
}

void sh::scripting::ScriptingEngine::loadScript(QString script)
{
    QString ext = script.split(".").last().toLower();
    std::shared_ptr<sh::scripting::ScriptInterpreter> interpreter = _interpreters[ext];
    if (interpreter)
        interpreter->execute(script);
    else
        SH_LOG_INFO(QString("unable to load script '%1' because there is no interpreter for it available.").arg(script))
}

sh::scripting::ScriptingEngine::ScriptedException::ScriptedException(sh::exceptions::ExceptionData data) :
    sh::exceptions::Exception(data.name("ScriptedException").resumeable())
{
}

sh::scripting::ScriptingEngine::ActionPluginLoadCrashedAgain::ActionPluginLoadCrashedAgain(QString path, QMutex* mutex)
                : sh::actions::ActionActionItem(""), _path(path), _mutex(mutex)
{
}

void sh::scripting::ScriptingEngine::ActionPluginLoadCrashedAgain::action(sh::actions::ActionExecutionInfo* info)
{
    QFileInfo fi(_path);
    int res = info->userfeedback()->simpleMessageBox(tr("The Shallot startup failed again (details follow).\n\nDo you want to disable the plugin '%1'? After it stopped, you can then try to start Shallot again.").arg(fi.baseName()),
                                          (int)sh::actions::ActionExecutionUserFeedback::MessageBoxButton::Yes|(int)sh::actions::ActionExecutionUserFeedback::MessageBoxButton::No, "error-critical");
    if (res == sh::actions::ActionExecutionUserFeedback::MessageBoxButton::Yes)
        QFile(_path).rename(_path + ".disabled");
    _mutex->unlock();
}
