/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2020 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "actions/actionexecutioninfo.h"
#include "filesystem/operation.h"
#include "filesystem/eurl.h"
#include "search/searchcriterionfactoryfromfunction.h"
#include "base/threadpool.h"
#include "base/mainthread.h"
#include "ui/searchpanelconfiguration.h"
#include "abstractactiondrivensearchcriterion.h"


sh::search::criteria::AbstractActionDrivenSearchCriterion::AbstractActionDrivenSearchCriterion(std::shared_ptr<SearchCriterionFactory> factory)
    : sh::search::SearchCriterion(factory)
{
}

bool sh::search::criteria::AbstractActionDrivenSearchCriterion::match(sh::filesystem::Operation* op, std::shared_ptr<const sh::filesystem::Eurl> eurl)
{
    if (searchspec.isEmpty())
        return false;
    return match2(op, eurl);
}

QString sh::search::criteria::AbstractActionDrivenSearchCriterion::valuedescription()
{
    return searchspec.join(" ");
}

void sh::search::criteria::AbstractActionDrivenSearchCriterion::_createEditor(std::shared_ptr<sh::search::SearchCriterion> cfg, sh::ui::SearchPanelConfiguration* panelcfg)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    panelcfg->addLabel();
    panelcfg->addActionButton(QObject::tr("C&onfigure"), [=](){
        QMutex* _execmutex = new QMutex();
        std::shared_ptr<ActionAbstractActionDrivenSearchCriterionConfigure> action = std::make_shared<ActionAbstractActionDrivenSearchCriterionConfigure>(std::dynamic_pointer_cast<AbstractActionDrivenSearchCriterion>(cfg), _execmutex);
        sh::base::ThreadPool::enqueueForce([=](){
            action->initializeSync();
            action->execute();
            _execmutex->lock();
            _execmutex->unlock();
            delete _execmutex;
            sh::base::MainThread::dispatcher()->invokeSync([&](){
                updatelabel(cfg, panelcfg);
            });
        });
    });
    _aliveWidgets.insert(panelcfg);  //TODO # this mechanism is weird (and has imperfections)
    panelcfg->onDestroyed([=](){
        ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
        _aliveWidgets.remove(panelcfg);
    });
    updatelabel(cfg, panelcfg);
}

void sh::search::criteria::AbstractActionDrivenSearchCriterion::updatelabel(std::shared_ptr<sh::search::SearchCriterion> cfg, sh::ui::SearchPanelConfiguration* panelcfg)
{
    if (_aliveWidgets.contains(panelcfg))
    {
        sh::search::criteria::AbstractActionDrivenSearchCriterion* wcfg = dynamic_cast<sh::search::criteria::AbstractActionDrivenSearchCriterion*>(cfg.get());
        QString lbltext = "???";
        sh::exceptions::Exception::executeGuarded_errorpanel([&](){
            lbltext = wcfg->searchspec.isEmpty() ? QObject::tr("Not yet configured") : QObject::tr("Search for: %1").arg(wcfg->getSearchSpecDescription());
        });
        auto lbl = dynamic_cast<sh::ui::SearchPanelLabelEditor*>(panelcfg->getEditorAt(0));
        lbl->setTextContent(lbltext);
    }
}

void sh::search::criteria::AbstractActionDrivenSearchCriterion::_editorUpdateConfiguration(ui::SearchPanelConfiguration*, std::shared_ptr<sh::search::SearchCriterion>)
{
}

void sh::search::criteria::AbstractActionDrivenSearchCriterion::_deserialize(QString s)
{
    searchspec = sh::search::SearchCriterion::_deserializeStringList(s);
}

QString sh::search::criteria::AbstractActionDrivenSearchCriterion::_serialize()
{
    return sh::search::SearchCriterion::_serializeStringList(searchspec);
}

QString sh::search::criteria::AbstractActionDrivenSearchCriterion::getSearchSpecDescription()
{
    return searchspec.join(" ");
}

sh::search::criteria::ActionAbstractActionDrivenSearchCriterionConfigure::ActionAbstractActionDrivenSearchCriterionConfigure(std::shared_ptr<AbstractActionDrivenSearchCriterion> cfg, QMutex* mutex)
    : sh::actions::ActionActionItem(""), _cfg(cfg), _mutex(mutex)
{
    _mutex->lock();
}

void sh::search::criteria::ActionAbstractActionDrivenSearchCriterionConfigure::action(sh::actions::ActionExecutionInfo* info)
{
    _cfg->configure(info);
    QVariant v = info->operation()->getCustomData(0, "searchconfiguration");
    if (!v.isNull())
        _cfg->searchspec = v.toStringList();
    sh::base::MainThread::dispatcher()->invokeSync([&](){
        _mutex->unlock();
    });
}

QSet<sh::ui::SearchPanelConfiguration*> sh::search::criteria::AbstractActionDrivenSearchCriterion::_aliveWidgets;
