/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2020 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <QRegularExpression>
#include "filesystem/eurl.h"
#include "filesystem/filesystemmodel.h"
#include "detailcolumns/detailcolumnextendedattributes.h"
#include "search/criteria/abstractstringsearchcriterion.h"
#include "search/searchcriterionfactoryfromfunction.h"
#include "search/searchmanager.h"
#include "base/mainthread.h"
#include "base/singletoninitializer.h"
#include "ui/searchpanelconfiguration.h"
#include "extendedattributesearchcriterion.h"

REGISTER_STATICINIT(sh::search::criteria, ExtendedAttributeSearchCriterion, "base")


void sh::search::criteria::ExtendedAttributeSearchCriterion::doInitialize()
{
    REGISTER_CRITERION_FACTORY(sh::search::SearchManager::REGISTER_SEARCHCRITERION_INDEX_CORE+15000, "xattr", QObject::tr("Filesystem extended attributes"), sh::search::criteria::ExtendedAttributeSearchCriterion)
}

void sh::search::criteria::ExtendedAttributeSearchCriterion::doShutdown()
{
}

sh::search::criteria::ExtendedAttributeSearchCriterion::ExtendedAttributeSearchCriterion(std::shared_ptr<SearchCriterionFactory> factory)
                : sh::search::SearchCriterion(factory)
{
}

void sh::search::criteria::ExtendedAttributeSearchCriterion::createEditor(std::shared_ptr<sh::search::SearchCriterion> cfg, sh::ui::SearchPanelConfiguration* panelcfg)
{
    sh::search::criteria::ExtendedAttributeSearchCriterion* me = dynamic_cast<sh::search::criteria::ExtendedAttributeSearchCriterion*>(cfg.get());
    sh::ui::SearchPanelLabelEditor* lblfoo = panelcfg->addLabel(); // this hack lets focus jump to value field
    lblfoo->setTextContent("");
    sh::ui::SearchPanelTextEditor* edtkey = panelcfg->addTextEditor();
    edtkey->setPlaceholderDescription(QObject::tr("Attribute key"));
    edtkey->setTextContent(me->key);
    sh::ui::SearchPanelTextEditor* edtvalue = panelcfg->addTextEditor();
    edtvalue->setPlaceholderDescription(QObject::tr("Attribute value"));
    edtvalue->setTextContent(me->value);
    lblfoo->setFocusProxyEditor(edtvalue);
    QString keymode = QObject::tr("Key mode: ");
    QString valuemode = QObject::tr("Value mode: ");
    QStringList slkeymodes({ QObject::tr("No key search"), keymode + QObject::tr("Simple"), keymode + QObject::tr("Regular expression"), keymode + QObject::tr("Fuzzy") });
    QStringList slvaluemodes({ QObject::tr("No value search"), valuemode + QObject::tr("Simple"), valuemode + QObject::tr("Regular expression"), valuemode + QObject::tr("Fuzzy") });
    sh::ui::SearchPanelButton* btnKeyMode = panelcfg->addMenuButton(QObject::tr("(c&hange)"),  slkeymodes, [=](int i){
        me->keymode = (Mode)i;
        _checkEditFieldStates(panelcfg, me);
    });
    btnKeyMode->setMenuSelection((int)(me->keymode));
    sh::ui::SearchPanelButton* btnValueMode = panelcfg->addMenuButton(QObject::tr("(c&hange)"), slvaluemodes, [=](int i){
        me->valuemode = (Mode)i;
        _checkEditFieldStates(panelcfg, me);
    });
    btnValueMode->setMenuSelection((int)(me->valuemode));
    _checkEditFieldStates(panelcfg, me);
}

void sh::search::criteria::ExtendedAttributeSearchCriterion::_checkEditFieldStates(ui::SearchPanelConfiguration *panelcfg, sh::search::criteria::ExtendedAttributeSearchCriterion* cfg)
{
    auto txtKey = dynamic_cast<sh::ui::SearchPanelTextEditor*>(panelcfg->getEditorAt(1));
    auto txtValue = dynamic_cast<sh::ui::SearchPanelTextEditor*>(panelcfg->getEditorAt(2));
    txtKey->setWidgetEnabled(cfg->keymode != Mode::None);
    txtValue->setWidgetEnabled(cfg->valuemode != Mode::None);
}

void sh::search::criteria::ExtendedAttributeSearchCriterion::editorUpdateConfiguration(ui::SearchPanelConfiguration *panelcfg, std::shared_ptr<sh::search::SearchCriterion> c)
{
    auto txtKey = dynamic_cast<sh::ui::SearchPanelTextEditor*>(panelcfg->getEditorAt(1));
    auto txtValue = dynamic_cast<sh::ui::SearchPanelTextEditor*>(panelcfg->getEditorAt(2));
    sh::search::criteria::ExtendedAttributeSearchCriterion* me = dynamic_cast<sh::search::criteria::ExtendedAttributeSearchCriterion*>(c.get());
    me->key = (me->keymode==Mode::None) ? "" : txtKey->textContent();
    me->value = (me->valuemode==Mode::None) ? "" : txtValue->textContent();
}

bool sh::search::criteria::ExtendedAttributeSearchCriterion::match(sh::filesystem::Operation* op, std::shared_ptr<const sh::filesystem::Eurl> eurl)
{
    QList<std::shared_ptr<sh::filesystem::FilesystemNode>> nodes = sh::filesystem::FilesystemModel::instance()->findNodesForEurl(eurl);
    if (!nodes.isEmpty())
    {
        QVariant vmap = sh::detailcolumns::DetailColumnExtendedAttributes::instance()->computeValue(nodes.first(), op);
        QMap<QString, QByteArray> map = sh::detailcolumns::DetailColumnExtendedAttributes::getMapByValue(vmap);
        foreach(QString s, map.keys())
        {
            bool match = matchString(keymode, key, s, &reKey, &reKeyKey);
            if (!match)
                continue;
            if (valuemode != Mode::None)
            {
                QString val = QString::fromUtf8(map[s]);
                match = matchString(valuemode, value, val, &reValue, &reValueValue);
                if (!match)
                    continue;
            }
            return true;
        }
    }
    return false;
}

bool sh::search::criteria::ExtendedAttributeSearchCriterion::matchString(Mode mode, QString needle, QString haystack, QRegularExpression* re, QString* restring)
{
    if (mode == Mode::Simple)
    {
        return haystack.toLower().contains(needle.toLower());
    }
    else if (mode == Mode::Fuzzy)
    {
        return sh::search::criteria::AbstractStringSearchCriterion::relativeLevenshteinPartsSubstringDistance(needle, haystack) < 0.25;
    }
    else if (mode == Mode::RegExp)
    {
        if (*restring != needle)
        {
            re->setPattern(needle);
            re->setPatternOptions(QRegularExpression::DotMatchesEverythingOption|QRegularExpression::UseUnicodePropertiesOption|QRegularExpression::MultilineOption);
            *restring = needle;
        }
        return re->match(haystack).hasMatch();
    }
    else if (mode == Mode::None)
    {
        return true;
    }
    else
        return false;
}

QString sh::search::criteria::ExtendedAttributeSearchCriterion::valuedescription()
{
    return (valuemode==Mode::None) ? key : value;
}

void sh::search::criteria::ExtendedAttributeSearchCriterion::_deserialize(QString s)
{
    QStringList sl = SearchCriterion::_deserializeStringList(s);
    keymode = (Mode)ModeCodes.indexOf(sl[0]);
    key = sl[1];
    valuemode = (Mode)ModeCodes.indexOf(sl[2]);
    value = sl[3];
}

QString sh::search::criteria::ExtendedAttributeSearchCriterion::_serialize()
{
    QStringList l;
    l << ModeCodes[(int)keymode];
    l << key;
    l << ModeCodes[(int)valuemode];
    l << value;
    return SearchCriterion::_serializeStringList(l);
}

const QStringList sh::search::criteria::ExtendedAttributeSearchCriterion::ModeCodes({"none", "simple", "re", "fuzzy"});
