/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2020 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "base/mainthread.h"
#include "filesystem/eurl.h"
#include "filesystem/filesystemmodel.h"
#include "detailcolumns/detailcolumnmtime.h"
#include "search/searchcriterionfactoryfromfunction.h"
#include "search/searchmanager.h"
#include "base/singletoninitializer.h"
#include "ui/searchpanelconfiguration.h"
#include "mtimesearchcriterion.h"

REGISTER_STATICINIT(sh::search::criteria, MtimeSearchCriterion, "base")


void sh::search::criteria::MtimeSearchCriterion::doInitialize()
{
    REGISTER_CRITERION_FACTORY(sh::search::SearchManager::REGISTER_SEARCHCRITERION_INDEX_CORE+10000, "mtime", QObject::tr("Modification time"), sh::search::criteria::MtimeSearchCriterion)
}

void sh::search::criteria::MtimeSearchCriterion::doShutdown()
{
}

sh::search::criteria::MtimeSearchCriterion::MtimeSearchCriterion(std::shared_ptr<SearchCriterionFactory> factory)
                : sh::search::SearchCriterion(factory)
{
    QDateTime now = QDateTime::currentDateTime();
    from = now;
    to = now;
}

void sh::search::criteria::MtimeSearchCriterion::createEditor(std::shared_ptr<sh::search::SearchCriterion> cfg, sh::ui::SearchPanelConfiguration* panelcfg)
{
    sh::search::criteria::MtimeSearchCriterion* me = dynamic_cast<sh::search::criteria::MtimeSearchCriterion*>(cfg.get());
    sh::ui::SearchPanelLabelEditor* lblfrom = panelcfg->addLabel();
    lblfrom->setTextContent(QObject::tr("between"));
    sh::ui::SearchPanelDateTimeEditor* edtfrom = panelcfg->addDateTimeEditor();
    edtfrom->setDatetime(me->from);
    lblfrom->setFocusProxyEditor(edtfrom); // important for our focus magic
    sh::ui::SearchPanelLabelEditor* lblto = panelcfg->addLabel();
    lblto->setTextContent(QObject::tr("and"));
    sh::ui::SearchPanelDateTimeEditor* edtto = panelcfg->addDateTimeEditor();
    edtto->setDatetime(me->to);
    panelcfg->addSpacer();
}

void sh::search::criteria::MtimeSearchCriterion::editorUpdateConfiguration(ui::SearchPanelConfiguration *panelcfg, std::shared_ptr<sh::search::SearchCriterion> c)
{
    sh::search::criteria::MtimeSearchCriterion* me = dynamic_cast<sh::search::criteria::MtimeSearchCriterion*>(c.get());
    auto wfrom = dynamic_cast<sh::ui::SearchPanelDateTimeEditor*>(panelcfg->getEditorAt(1));
    auto wto = dynamic_cast<sh::ui::SearchPanelDateTimeEditor*>(panelcfg->getEditorAt(3));
    me->from = wfrom->datetime();
    me->to = wto->datetime();
}

bool sh::search::criteria::MtimeSearchCriterion::match(sh::filesystem::Operation* op, std::shared_ptr<const sh::filesystem::Eurl> eurl)
{
    QList<std::shared_ptr<sh::filesystem::FilesystemNode>> nodes = sh::filesystem::FilesystemModel::instance()->findNodesForEurl(eurl);
    if (!nodes.isEmpty())
    {
        QDateTime t= sh::detailcolumns::DetailColumnMtime::instance()->computeValue(nodes.first(), op).toDateTime();
        return (from <= t) && (t <= to);
    }
    return false;
}

QString sh::search::criteria::MtimeSearchCriterion::valuedescription()
{
    return from.toString(Qt::DefaultLocaleShortDate);
}

void sh::search::criteria::MtimeSearchCriterion::_deserialize(QString s)
{
    QStringList sl = SearchCriterion::_deserializeStringList(s);
    from = QDateTime::fromString(sl[0], "yyyyMMdd.HHmm");
    to = QDateTime::fromString(sl[1], "yyyyMMdd.HHmm");
}

QString sh::search::criteria::MtimeSearchCriterion::_serialize()
{
    QStringList l;
    l << from.toString("yyyyMMdd.HHmm");
    l << to.toString("yyyyMMdd.HHmm");
    return SearchCriterion::_serializeStringList(l);
}

