/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2020 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "filesystem/filesystemnode.h"
#include "filesystem/operation.h"
#include "filesystem/filesystemoperation.h"
#include "filesystem/eurl.h"
#include "filesystem/filesystemmodel.h"
#include "base/singletoninitializer.h"
#include "base/mainthread.h"
#include "base/threadpool.h"
#include "search.h"
#include "searchfilesystemhandler.h"


sh::search::Search::Search(std::shared_ptr<const sh::filesystem::Eurl> eurl, std::shared_ptr<sh::filesystem::FilesystemNode> searchnode)
            : QObject(), _eurl(eurl), _searchnode(searchnode)
{
    _config = sh::search::SearchConfiguration::deserialize(_eurl->hostname());
}

sh::search::Search::~Search()
{
    delete _config;
}

std::shared_ptr<const sh::filesystem::Eurl> sh::search::Search::eurl()
{
    return _eurl;
}

std::shared_ptr<sh::filesystem::FilesystemNode> sh::search::Search::searchnode()
{
    return _searchnode;
}

void sh::search::Search::search()
{
    std::shared_ptr<sh::search::Search> self = this->shared_from_this();
    if (_config->criteria.isEmpty())
        return;
    {
        QMutexLocker _l1(&_isSearchingMutex);
        if (_isSearching)
        {
            _isSearchAgain = true;
            return;
        }
        _isSearchAgain = false;
        _isSearching = true;
    }
    sh::base::MainThread::dispatcher()->invokeSync([&](){
        emit isSearchingChanged();
        _dirs.clear();
        _dirs[_eurl] = _searchnode;
        std::shared_ptr<const sh::filesystem::Eurl> seurl = _eurl->outermostInnerEurl();
        // remove old stuff
        QSet<std::shared_ptr<sh::filesystem::FilesystemNode>> cnodes;
        foreach(std::shared_ptr<sh::filesystem::FilesystemNode> cnode, *(_searchnode->childnodes()->nodes()))
            cnodes << cnode;
        foreach(std::shared_ptr<sh::filesystem::FilesystemNode> cnode, cnodes)
            _searchnode->childnodes()->removePermanentItem(cnode);
        // do the actual search
        sh::base::ThreadPool::enqueueForce([=](){
            std::shared_ptr<sh::search::Search> _self = self;
            search(seurl, "");
            {
                QMutexLocker _l1(&_isSearchingMutex);
                _isSearching = false;
                if (_isSearchAgain)
                    sh::base::ThreadPool::enqueueForce([=](){
                        std::shared_ptr<sh::search::Search> _self = self;
                        search();
                    });
            }
            emit isSearchingChanged();
        });
    });
}

bool sh::search::Search::isSearching()
{
    QMutexLocker _l1(&_isSearchingMutex);
    return _isSearching;
}

std::shared_ptr<sh::filesystem::FilesystemNode> sh::search::Search::getDirItem(std::shared_ptr<const sh::filesystem::Eurl> item)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    std::shared_ptr<sh::filesystem::FilesystemNode> result = _dirs.value(item);
    if (result)
        return result;
    std::shared_ptr<sh::filesystem::FilesystemNode> presult = getDirItem(item->parentSegment());
    sh::filesystem::FilesystemHandler* handler = sh::search::SearchFilesystemHandler::instance().get();
    result = sh::filesystem::FilesystemModel::instance()->createFilesystemNode(item, handler, sh::filesystem::FilesystemNodeType::Directory, false, presult, false);
    presult->childnodes()->addPermanentItem(result);
    _dirs.insert(item, result);
    return result;
}

void sh::search::Search::insertFileItem(std::shared_ptr<const sh::filesystem::Eurl> item, QString relpath, sh::filesystem::FilesystemNodeType ftype)
{
    sh::base::MainThread::dispatcher()->invokeSync([=](){
        sh::filesystem::FilesystemHandler* handler = sh::filesystem::FilesystemHandlerRegister::instance()->findHandler(item->scheme()).get();
        std::shared_ptr<const sh::filesystem::Eurl> direurl = _searchnode->eurl()->withAppendedSegments(relpath);
        std::shared_ptr<sh::filesystem::FilesystemNode> dir = _config->showAsTree ? getDirItem(direurl) : _searchnode;
        std::shared_ptr<sh::filesystem::FilesystemNode> result = sh::filesystem::FilesystemModel::instance()->createFilesystemNode(item, handler, ftype, false, dir, false);
        if (!_config->showAsTree)
            result->setDisplayName(relpath.mid(1) + "/" + item->basename());
        dir->childnodes()->addPermanentItem(result);
    });
}

void sh::search::Search::search(std::shared_ptr<const sh::filesystem::Eurl> eurl, QString relpath)
{
    sh::filesystem::Operation op;
    sh::filesystem::FilesystemHandler* handler = sh::filesystem::FilesystemHandlerRegister::instance()->findHandler(_eurl->scheme()).get();
    handler->search(&op, eurl, [&](std::shared_ptr<const sh::filesystem::Eurl> feurl, QList<std::shared_ptr<sh::search::SearchCriterion>> fcriteria, sh::filesystem::FilesystemNodeType ftype){
        // addfile handler
        if (ftype == sh::filesystem::FilesystemNodeType::File || ftype == sh::filesystem::FilesystemNodeType::Link)
        {
            bool match = true;
            foreach(std::shared_ptr<sh::search::SearchCriterion> criterion, fcriteria)
            {
                match = criterion->match(&op, feurl);
                if (!match)
                    break;
            }
            if (match)
                insertFileItem(feurl, relpath, ftype);
        }
    }, [&](std::shared_ptr<const sh::filesystem::Eurl> deurl){
        // visitdir handler
        search(deurl, relpath + "/" + deurl->basename());
    }, _config->criteria);
}
