/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2020 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef SEARCH_H
#define SEARCH_H

#include <memory>
#include <QMap>
#include <QMutex>
#include <QObject>
#include "filesystem/filesystemnode.h"
#include "searchconfiguration.h"


namespace sh{
    namespace search{

        /** \brief
         * A search object represents one search request the user made.
         */
        class Search : public QObject, public std::enable_shared_from_this<Search>
        {
            Q_OBJECT
        public:
            virtual ~Search();
            /** \brief Returns the eurl containg the search configuration.
             */
            std::shared_ptr<const sh::filesystem::Eurl> eurl();
            /** \brief Returns the root node of the search (the one named like 'Search ...').
             */
            std::shared_ptr<sh::filesystem::FilesystemNode> searchnode();
            /** \brief
             * @a_infrastructure
             * See sh::search::SearchManager::requestSearch.
             */
            Search(std::shared_ptr<const filesystem::Eurl> eurl, std::shared_ptr<sh::filesystem::FilesystemNode> searchnode);

            /** \brief Start the searching.
             */
            void search();
            /** \brief Returns if the search is currently in progress.
             */
            bool isSearching();
        signals:
            /** \brief Triggers when Search::isSearching changes.
             */
            void isSearchingChanged();
        private:
            std::shared_ptr<const sh::filesystem::Eurl> _eurl;
            std::shared_ptr<sh::filesystem::FilesystemNode> _searchnode;
            sh::search::SearchConfiguration* _config;
            void search(std::shared_ptr<const filesystem::Eurl> eurl, QString relpath);
            void insertFileItem(std::shared_ptr<const sh::filesystem::Eurl> item, QString relpath, sh::filesystem::FilesystemNodeType ftype);
            std::shared_ptr<sh::filesystem::FilesystemNode> getDirItem(std::shared_ptr<const sh::filesystem::Eurl> item);
            QHash<std::shared_ptr<const sh::filesystem::Eurl>, std::shared_ptr<sh::filesystem::FilesystemNode>> _dirs;
            QMutex _isSearchingMutex;
            bool _isSearching = false;
            bool _isSearchAgain = false;
        };

    }
}

#endif // SEARCH_H
