/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2020 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "filesystem/eurl.h"
#include "exceptions/argumentexception.h"
#include "exceptions/programexception.h"
#include "searchcriterion.h"
#include "searchcriterionfactory.h"
#include "searchmanager.h"


sh::search::SearchCriterion::SearchCriterion(std::shared_ptr<SearchCriterionFactory> factory)
            : _factory(factory)
{
}

QString sh::search::SearchCriterion::_serializeStringList(QStringList l)
{
    QStringList r;
    foreach(QString s, l)
        r << s.replace("%", "%25").replace(":", "%3A");
    return r.join(':');
}

QStringList sh::search::SearchCriterion::_deserializeStringList(QString s)
{
    if (s.isEmpty())
        return QStringList();
    QStringList r = s.split(':');
    QStringList l;
    foreach(QString s, r)
        l << QUrl::fromPercentEncoding(s.toUtf8());
    return l;
}

QString sh::search::SearchCriterion::serialize()
{
    return _factory->key() + "=" + _serialize();
}

std::shared_ptr<sh::search::SearchCriterion> sh::search::SearchCriterion::deserialize(QString s)
{
    int ieq = s.indexOf("=");
    if (ieq == -1)
        throw sh::exceptions::ArgumentException(sh::exceptions::ExceptionData().detailsAreInteresting().details(QObject::tr("Invalid format in search criterion token: '=' missing: ") + s));
    QString key = s.mid(0, ieq);
    std::shared_ptr<sh::search::SearchCriterion> result;
    foreach(std::shared_ptr<SearchCriterionFactory> fact, sh::search::SearchManager::instance()->factories())
    {
        if (fact->key()==key)
        {
            result = fact->construct();
            break;
        }
    }
    if (!result)
        throw sh::exceptions::ArgumentException(sh::exceptions::ExceptionData().detailsAreInteresting().details(QObject::tr("Unable to deserialize a search criterion: Key unknown: ") + key));
    result->_deserialize(s.mid(ieq+1));
    return result;
}

std::shared_ptr<sh::search::SearchCriterionFactory> sh::search::SearchCriterion::factory()
{
    return _factory;
}

sh::search::SearchCriterion::~SearchCriterion()
{
}
