/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2020 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef SEARCHCRITERION_H
#define SEARCHCRITERION_H

#include <memory>
#include <QObject>


namespace sh{
    namespace filesystem { class Eurl; class Operation; }

    namespace search{
        class SearchCriterionFactory;

        /** \brief Abstract base class for a search criterion.
         *
         * Each search criterion implements how a user can filter his files in a search.
         *
         * Register an implementation, e.g. with 'REGISTER_CRITERION_FACTORY' of "search/searchcriterionfactoryfromfunction.h".
         */
        class SearchCriterion
        {
        public:
            /** \brief
             * @a_infrastructure
             */
            SearchCriterion(std::shared_ptr<sh::search::SearchCriterionFactory> factory);
            virtual ~SearchCriterion();
            /** \brief Serializes this criterion to a string. See also SearchCriterion::deserialize.
             */
            QString serialize();
            /** \brief Deserializes this criterion from a string. See also SearchCriterion::serialize.
             */
            static std::shared_ptr<sh::search::SearchCriterion> deserialize(QString s);
            /** \brief Checks if a given file (by eurl) matches this criterion.
             * @a_dooverride
             */
            virtual bool match(sh::filesystem::Operation* op, std::shared_ptr<const sh::filesystem::Eurl> eurl) =0;
            /** \brief Returns a human readable text describing the current criterion shortly and precisely.
             * @a_dooverride
             */
            virtual QString valuedescription() =0;
            /** \brief Returns the factory this criterion has created (provides some more metadata).
             */
            std::shared_ptr<sh::search::SearchCriterionFactory> factory();
        protected:
            /** \brief Serializes a QStringList to a string (using ':' as separator and some percent encoding). See also SearchCriterion::_deserializeStringList.
             */
            static QString _serializeStringList(QStringList l);
            /** \brief Deserializes a string back to a QStringList. See also SearchCriterion::_serializeStringList.
             */
            static QStringList _deserializeStringList(QString s);
            /** \brief Serializes this criterion to a string. See also SearchCriterion::_deserialize.
             * @a_dooverride
             * Note: This is used by SearchCriterion::serialize, which just boxes it, adding some control information.
             */
            virtual QString _serialize() =0;
            /** \brief Deserializes this criterion from a string. See also SearchCriterion::_serialize.
             * @a_dooverride
             */
            virtual void _deserialize(QString s) =0;
            std::shared_ptr<sh::search::SearchCriterionFactory> _factory;
        };

    }
}

#endif // SEARCHCRITERION_H
