/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2020 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef SEARCHCRITERIONFACTORY_H
#define SEARCHCRITERIONFACTORY_H

#include <memory>
#include <QObject>


namespace sh{
    namespace ui { class SearchPanelConfiguration; }
    namespace filesystem { class Eurl; class Operation; class FilesystemNode; }

    namespace search {
        class SearchCriterion;

        /** \brief Abstract base class for a search criterion factory, which constructs some sh::search::SearchCriterion instances.
         *
         * It also provides some control methods and metadata (which is possible since there is a separate factory for each criterion class).
         *
         * Note: Though possible in some exotic situations, one should not override this class. Override and register sh::search::SearchCriterion instead!
         */
        class SearchCriterionFactory : public std::enable_shared_from_this<SearchCriterionFactory>
        {
        public:
            /** \brief
             * @a_infrastructure
             */
            SearchCriterionFactory();
            /** \brief Returns the string key for the associated search criterion class.
             * @a_dooverride
             */
            virtual QString key() =0;
            /** \brief Returns the description string for the associated search criterion class.
             * @a_dooverride
             */
            virtual QString description() =0;
            /** \brief Builds the search config ui in a search panel.
             *
             * Takes the data from `c` and populates `panelcfg` with it.
             * @a_dooverride
             */
            virtual void editor(std::shared_ptr<sh::search::SearchCriterion> c, sh::ui::SearchPanelConfiguration* panelcfg) =0;
            /** \brief Updates own data with the content from the search config ui in a search panel.
             *
             * Takes the data from `panelcfg` and updates `c` with it.
             * @a_dooverride
             */
            virtual void editorupdateconfig(sh::ui::SearchPanelConfiguration* panelcfg, std::shared_ptr<sh::search::SearchCriterion> c) =0;
            /** \brief Constructs an instance of the associated search criterion class.
             * @a_dooverride
             */
            virtual std::shared_ptr<sh::search::SearchCriterion> construct() =0;
            /** \brief Checks if the associated search criterion class should be offered to the user for a given node.
             * @a_dooverride_optional
             */
            virtual bool isVisibleFor(sh::filesystem::Operation* op, std::shared_ptr<sh::filesystem::FilesystemNode> node);
            virtual ~SearchCriterionFactory();
        };

    }
}

#endif // SEARCHCRITERIONFACTORY_H
