/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2020 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "base/mainthread.h"
#include "filesystem/filesystemnode.h"
#include "filesystem/filesystemmodel.h"
#include "filesystem/filesystemhandlerregister.h"
#include "base/iconmanager.h"
#include "searchfilesystemhandler.h"
#include "searchconfiguration.h"

REGISTER_SINGLETON(sh::search, SearchFilesystemHandler, "base")


sh::search::SearchFilesystemHandler::SearchFilesystemHandler() : FilesystemHandler(sh::filesystem::FilesystemModel::instance().get())
{
}

void sh::search::SearchFilesystemHandler::getreal(std::shared_ptr<const sh::filesystem::Eurl> seurl, std::shared_ptr<sh::filesystem::FilesystemHandler>* handler, std::shared_ptr<const sh::filesystem::Eurl>* eurl)
{
    *eurl = seurl->outermostInnerEurl()->withAppendedSegments(seurl->path());
    *handler = sh::filesystem::FilesystemHandlerRegister::instance()->findHandler((*eurl)->scheme());
}

void sh::search::SearchFilesystemHandler::itemlist(sh::filesystem::Operation*, std::shared_ptr<const sh::filesystem::Eurl>, sh::filesystem::FilesystemNodeType, sh::filesystem::FilesystemNodeListEditor*)
{
}

sh::filesystem::FilesystemNodeType sh::search::SearchFilesystemHandler::getType(sh::filesystem::Operation*, std::shared_ptr<const sh::filesystem::Eurl> eurl)
{
    QList<std::shared_ptr<sh::filesystem::FilesystemNode>> nl = sh::filesystem::FilesystemModel::instance()->tryGetNodesForEurl(eurl);
    if (!nl.isEmpty())
        return (sh::filesystem::FilesystemNodeType) nl.first()->nodetype();
    return sh::filesystem::FilesystemNodeType::NONE;
}

qint64 sh::search::SearchFilesystemHandler::getSize(sh::filesystem::Operation*, std::shared_ptr<const sh::filesystem::Eurl>)
{
    return 0;
}

QDateTime sh::search::SearchFilesystemHandler::getMtime(sh::filesystem::Operation* op, std::shared_ptr<const sh::filesystem::Eurl> eurl)
{
    std::shared_ptr<sh::filesystem::FilesystemHandler> ehandler;
    std::shared_ptr<const sh::filesystem::Eurl> eeurl;
    getreal(eurl, &ehandler, &eeurl);
    return ehandler->getMtime(op, eeurl);
}

void sh::search::SearchFilesystemHandler::setMtime(sh::filesystem::Operation*, std::shared_ptr<const sh::filesystem::Eurl>, QDateTime)
{
}

QString sh::search::SearchFilesystemHandler::getMimetype(sh::filesystem::Operation*, std::shared_ptr<const sh::filesystem::Eurl>)
{
    return "application/x-shallot-search-result";
}

QString sh::search::SearchFilesystemHandler::getLinkTarget(sh::filesystem::Operation*, std::shared_ptr<const sh::filesystem::Eurl>)
{
    return QString();
}

bool sh::search::SearchFilesystemHandler::canGetFileContent(sh::filesystem::Operation*, std::shared_ptr<const sh::filesystem::Eurl>)
{
    return false;
}

std::shared_ptr<QIODevice> sh::search::SearchFilesystemHandler::getFileContent(sh::filesystem::Operation*, std::shared_ptr<const sh::filesystem::Eurl>)
{
    return 0;
}

bool sh::search::SearchFilesystemHandler::canCreateDirectory(sh::filesystem::Operation*, std::shared_ptr<const sh::filesystem::Eurl>)
{
    return false;
}

void sh::search::SearchFilesystemHandler::createDirectory(sh::filesystem::Operation*, std::shared_ptr<const sh::filesystem::Eurl>)
{
}

bool sh::search::SearchFilesystemHandler::canCreateLink(sh::filesystem::Operation*, std::shared_ptr<const sh::filesystem::Eurl>)
{
    return false;
}

void sh::search::SearchFilesystemHandler::createLink(sh::filesystem::Operation*, std::shared_ptr<const sh::filesystem::Eurl>, QString)
{
}

bool sh::search::SearchFilesystemHandler::canCreateFile(sh::filesystem::Operation*, std::shared_ptr<const sh::filesystem::Eurl>)
{
    return false;
}

void sh::search::SearchFilesystemHandler::createFile(sh::filesystem::Operation*, std::shared_ptr<const sh::filesystem::Eurl>, QIODevice*, HandlerTransfer*)
{
}

bool sh::search::SearchFilesystemHandler::canDeleteItem(sh::filesystem::Operation*, std::shared_ptr<const sh::filesystem::Eurl>)
{
    return false;
}

void sh::search::SearchFilesystemHandler::deleteItem(sh::filesystem::Operation*, std::shared_ptr<const sh::filesystem::Eurl>)
{
}

bool sh::search::SearchFilesystemHandler::canRenameItem(sh::filesystem::Operation*, std::shared_ptr<const sh::filesystem::Eurl>)
{
    return false;
}

void sh::search::SearchFilesystemHandler::renameItem(sh::filesystem::Operation*, std::shared_ptr<const sh::filesystem::Eurl>, QString)
{
}

QList<std::shared_ptr<sh::actions::AbstractActionItem> > sh::search::SearchFilesystemHandler::getActions(const QList<std::shared_ptr<const sh::filesystem::Eurl>>)
{
    return QList<std::shared_ptr<sh::actions::AbstractActionItem> >();
}

void sh::search::SearchFilesystemHandler::customizeUi(std::shared_ptr<sh::filesystem::FilesystemNode> node, bool* showSearchPanel)
{
    if (node->eurl()->outerUrlIsRootDirectory())
        *showSearchPanel = true;
}

void sh::search::SearchFilesystemHandler::configureItems(sh::filesystem::Operation*, QList<std::shared_ptr<sh::filesystem::FilesystemNode> > nodes)
{
    foreach (std::shared_ptr<sh::filesystem::FilesystemNode> node, nodes) {
        if (node->eurl()->outerUrlIsRootDirectory()) {
            sh::search::SearchConfiguration* cfg = sh::search::SearchConfiguration::deserialize(node->eurl()->hostname());
            if (cfg->criteria.isEmpty())
                node->setDisplayName(tr("Search"));
            else {
                QString dsc = cfg->criteria.first()->valuedescription();
                if (dsc.length()>9)
                    dsc = dsc.left(7) + "...";
                QString andmore = (cfg->criteria.length()>1) ? tr(" &more") : "";
                node->setDisplayName(tr("Search '%1'%2").arg(dsc).arg(andmore));
            }
            node->setIcon(sh::base::IconManager::instance()->getIcon("edit-find"));
            delete cfg;
        }
    }
}

void sh::search::SearchFilesystemHandler::doInitialize()
{
    sh::filesystem::FilesystemHandlerRegister::instance()->addHandler("search", instance());
}
