/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2020 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "filesystem/eurl.h"
#include "filesystem/filesystemmodel.h"
#include "filesystem/filesystemnode.h"
#include "exceptions/argumentexception.h"
#include "base/mainthread.h"
#include "base/threadpool.h"
#include "searchmanager.h"
#include "search.h"
#include "searchfilesystemhandler.h"

REGISTER_SINGLETON(sh::search, SearchManager, "base")


sh::search::SearchManager::SearchManager() : QObject()
{
}

std::shared_ptr<sh::search::Search> sh::search::SearchManager::requestSearch(std::shared_ptr<const sh::filesystem::Eurl> eurl, QString searchconfig, std::shared_ptr<sh::filesystem::FilesystemNode> searchrootnode)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    std::shared_ptr<const sh::filesystem::Eurl> seurl = eurl->enwrapWithOuterUrl("search", searchconfig, "/");
    std::shared_ptr<sh::search::Search> result;
    std::shared_ptr<sh::filesystem::FilesystemNode> searchnode;
    // find existing one
    foreach (std::weak_ptr<sh::search::Search> wrs, _weak_searches) {
        std::shared_ptr<sh::search::Search> rs = wrs.lock();
        if (rs && rs->eurl() == seurl) {
            result = rs;
            searchnode = result->searchnode();
            break;
        }
    }
    bool weakresult = !_searches.contains(result);
    if (weakresult || !result) {
        // make new one
        bool isNew;
        searchnode = sh::filesystem::FilesystemModel::instance()->getOrCreateFilesystemNode(seurl, sh::search::SearchFilesystemHandler::instance().get(), sh::filesystem::FilesystemNodeType::SpecialTreeOnlyDirectory, false, searchrootnode, false, true, &isNew);
        if (isNew)
            searchrootnode->childnodes()->addPermanentItem(searchnode);
        if (!result) {
            result = std::make_shared<sh::search::Search>(seurl, searchnode);
            _weak_searches.append(result);
            connect(result.get(), &QObject::destroyed, this, [=](){
                auto itr = _weak_searches.begin();
                while (itr != _weak_searches.end()) {
                    if ((*itr).expired())
                        itr = _weak_searches.erase(itr);
                    else
                        itr++;
                }
            });
        }
        _searches.append(result);
    }
    result->search();
    return result;
}

void sh::search::SearchManager::removeSearch(std::shared_ptr<sh::search::Search> search)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    search->searchnode()->parentnode()->childnodes()->removePermanentItem(search->searchnode());
    _searches.removeOne(search);
}

std::shared_ptr<sh::search::Search> sh::search::SearchManager::findSearch(std::shared_ptr<const sh::filesystem::Eurl> eurl)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    foreach (std::weak_ptr<sh::search::Search> wrs, _weak_searches) {
        std::shared_ptr<sh::search::Search> rs = wrs.lock();
        if (rs && (rs->eurl() == eurl))
            return rs;
    }
    return 0;
}

void sh::search::SearchManager::doInitialize()
{
    sh::filesystem::FilesystemModel::instance()->addOpenNodeHelper(500, [=](std::shared_ptr<const sh::filesystem::Eurl> eurl){ // helps opening a new node for search:/// roots by search manager
        QList<std::shared_ptr<sh::filesystem::FilesystemNode>> result;
        if (eurl->outerUrlIsRootDirectory() && eurl->scheme() == "search") {
            std::shared_ptr<const sh::filesystem::Eurl> ieurl = eurl->outermostInnerEurl();
            QList<std::shared_ptr<sh::filesystem::FilesystemNode>> lsn = sh::filesystem::FilesystemModel::instance()->tryGetNodesForEurl(ieurl);
            if (!lsn.isEmpty()) {
                sh::base::MainThread::dispatcher()->invokeSync([&](){
                    std::shared_ptr<sh::search::Search> search = sh::search::SearchManager::instance()->requestSearch(ieurl, eurl->hostname(), lsn.first());
                    result.append(search->searchnode());
                });
            }
        }
        return result;
    });
}

void sh::search::SearchManager::addFactory(int index, std::shared_ptr<SearchCriterionFactory> f)
{
    _factories.insertMulti(index, f);
}

QList<std::shared_ptr<sh::search::SearchCriterionFactory>> sh::search::SearchManager::factories()
{
    return _factories.values();
}

sh::search::SearchManager::~SearchManager()
{
}

const int sh::search::SearchManager::REGISTER_SEARCHCRITERION_INDEX_CORE = 1000000;

const int sh::search::SearchManager::REGISTER_SEARCHCRITERION_INDEX_NORMAL = 2000000;

const int sh::search::SearchManager::REGISTER_SEARCHCRITERION_INDEX_EXOTIC = 3000000;
