/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2020 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef PROFILENODE_H
#define PROFILENODE_H

#include <memory>
#include <QMap>
#include <QString>
#include <QStringList>


namespace sh{
    namespace filesystem{ class Eurl; }

    namespace settings{
        class Setting;

        /** \brief
         * One entry in a Shallot settings profile, so one item as selectable
         * in the 'Manage settings' dialog.
         */
        class ProfileNode
        {
        public:
            /** \brief Constructed only by the infrastructure and made available otherwise.
             */
            ProfileNode();
            /** \brief An identifier name.
             */
            QString nodeId();
            /** \brief Sets the identifier name.
             */
            void setNodeId(QString nodeId);
            /** \brief The eurl of the directory, or 0 for global profile nodes.
             */
            std::shared_ptr<const filesystem::Eurl> eurl();
            /** \brief Sets the eurl.
             */
            void setEurl(std::shared_ptr<const filesystem::Eurl> eurl);
            /** \brief The profile name.
             */
            QString profilename();
            /** \brief Sets the profile name.
             */
            void setProfilename(QString profilename);
            /** \brief List of profile names from which this node inherits.
             */
            QStringList inheritsFrom();
            /** \brief Sets list of profile names from which this node inherits.
             */
            void setInheritsFrom(QStringList profilenames);
            /** \brief If this node also applies recursively on subfolders.
             */
            bool withSubfolders();
            /** \brief Sets of this node also applies recursively on subfolders.
             */
            void setWithSubfolders(bool v);
            /** \brief Stores a value for a setting for applying it later.
             */
            void setSetting(QString name, QString value, int onlyInFileview=-1);
            /** \brief List of settings names which are set in this node.
             */
            QList<QString> getSettingNames();
            /** \brief Gets the stored value of a setting by setting name.
             */
            QString getSettingValue(QString name);
            /** \brief Gets if a setting applies to a certain fileview or to the entire window.
             */
            int getSettingOnlyInFileviewIndex(QString name);

        private:
            QMap<QString, QString> _values;
            QMap<QString, int> _valuesOnlyInFileviewIndex;
            std::shared_ptr<const sh::filesystem::Eurl> _eurl;
            QString _profilename;
            QStringList _inheritsFrom;
            bool _withSubfolders;
            QString _nodeId;
        };

    }
}

#endif // PROFILENODE_H
