/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2020 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <QDateTime>
#include "base/mainthread.h"
#include "base/shallotprocess.h"
#include "base/threadpool.h"
#include "configuration/configurationmanager.h"
#include "exceptions/ioexception.h"
#include "exceptions/runtimeexception.h"
#include "exceptions/argumentexception.h"
#include "exceptions/programexception.h"
#include "filesystem/eurl.h"
#include "filesystem/filesystemnode.h"
#include "filesystem/filesystemmodel.h"
#include "tools/benchmarking.h"
#include "tools/userdirlock.h"
#include "ui/mainwindow.h"
#include "ui/fileview.h"
#include "profilenode.h"
#include "setting.h"
#include "settingsmanager.h"

REGISTER_SINGLETON(sh::settings, SettingsManager, "early")


sh::settings::SettingsManager::SettingsManager() :
    QObject()
{
    readStoredProfiles();
}

std::shared_ptr<sh::settings::Setting> sh::settings::SettingsManager::getSettingByName(QString name)
{
    CHECK_CORRECT_CALLER_THREAD
    auto i = _settings.find(name);
    if (i == _settings.end())
        return 0;
    else
        return *i;
}

QList<std::shared_ptr<sh::settings::Setting>> sh::settings::SettingsManager::getAllSettings()
{
    CHECK_CORRECT_CALLER_THREAD
    QList<std::shared_ptr<sh::settings::Setting>> result = _settings.values();
    qSort(result.begin(), result.end(), [](std::shared_ptr<sh::settings::Setting> arg1, std::shared_ptr<sh::settings::Setting> arg2) {
        if (arg1->group() != arg2->group())
            return arg1->group() < arg2->group();
        else
            return arg1->name() < arg2->name();
    });
    return result;
}

sh::settings::ProfileNode* sh::settings::SettingsManager::getNodeByNodeId(QString nodeId)
{
    CHECK_CORRECT_CALLER_THREAD
    sh::settings::ProfileNode* result=0;
    auto itr = _profileNodesById.find(nodeId);
    if (itr != _profileNodesById.end())
        result = *itr;
    return result;
}

void sh::settings::SettingsManager::readStoredProfiles()
{
    CHECK_CORRECT_CALLER_THREAD
    sh::tools::UserDirLock userdirlock;
    QString profileDirPath = sh::base::ShallotProcess::userDataDir() + "profiles/";
    QDir profileDir(profileDirPath);
    QStringList _filter;
    _filter << "*.xsp";
    QStringList profilenodefilenames = profileDir.entryList(_filter);
    QMap<QString, QList<sh::settings::ProfileNode*> > lists;
    foreach(QString profilenodefilename, profilenodefilenames) {
        QString fullprofilenodefilename=profileDirPath+profilenodefilename;
        sh::settings::ProfileNode* profilenode = new sh::settings::ProfileNode();
        profilenode->setNodeId(fullprofilenodefilename);
        QXmlSimpleReader xmlReader;
        QFile profilenodefile(fullprofilenodefilename);
        QXmlInputSource source(&profilenodefile);
        _MyHandler handler(profilenode, profilenodefilename);
        xmlReader.setContentHandler(&handler);
        xmlReader.setErrorHandler(&handler);
        xmlReader.parse(&source);
        lists[profilenode->profilename()].append(profilenode);
    }
    foreach (sh::settings::ProfileNode* n, _profileNodesById.values())
        delete n;
    _profileNodes.clear();
    _profileNodesById.clear();
    _profileNames.clear();
    foreach(QString pn, lists.keys()) {
        foreach (sh::settings::ProfileNode* node, lists[pn]) {
            _profileNodesById[node->nodeId()] = node;
            _profileNodes[node->eurl()][pn].append(node);
        }
        _profileNames.append(pn);
    }
    invalidateCache();
    emit profilesChanged();
}

void sh::settings::SettingsManager::storeProfile(std::shared_ptr<const sh::filesystem::Eurl>eurl, QString profilename, QMap<QString,QString> values, QMap<QString,int> onlyinfileviewindex, bool withSubfolders, QStringList inheritFrom)
{
    CHECK_CORRECT_CALLER_THREAD
    {
        sh::tools::UserDirLock userdirlock;
        if (!QRegExp("[^,/*?$\"']+").exactMatch(profilename))
            throw sh::exceptions::ArgumentException(sh::exceptions::ExceptionData().message(tr("The profile name contains unallowed characters.")));
        QString profileDirPath = sh::base::ShallotProcess::userDataDir()+"profiles/";
        QString eurlstring = "0";
        if (eurl)
            eurlstring = QString::fromLatin1( QUrl::toPercentEncoding(eurl->asString()) );
        if (!QDir(profileDirPath).exists())
            QDir(sh::base::ShallotProcess::userDataDir()).mkdir("profiles");
        QString tprofilefilename =
                profileDirPath +
                eurlstring + "," +
                profilename + ",%1.xsp";
        QString profilefilename;
        int i=0;
        while(profilefilename.isNull() || QFile::exists(profilefilename))
            profilefilename=tprofilefilename.arg(i++, (int)6, 10, QLatin1Char('0'));
        {
            QFile outfile(profilefilename);
            if (!outfile.open(QIODevice::WriteOnly))
                throw sh::exceptions::IOException(sh::exceptions::ExceptionData().detailsAreInteresting().details(QString(tr("unable to write to file '%1'.")).arg(profilefilename)).retryable(false));
            QXmlStreamWriter xmlWriter;
            xmlWriter.setAutoFormatting(true);
            xmlWriter.setDevice(&outfile);
            xmlWriter.writeStartDocument();
            xmlWriter.writeStartElement("profile");
            QString _eurlstring="";
            if (eurl)
                _eurlstring = eurl->asString();
            xmlWriter.writeAttribute("eurl", _eurlstring);
            xmlWriter.writeAttribute("name", profilename);
            QString sWithSubfolders = "1";
            if (!withSubfolders)
                sWithSubfolders = "0";
            xmlWriter.writeAttribute("subfolders-inherit", sWithSubfolders);
            foreach(QString valname, values.keys()) {
                xmlWriter.writeStartElement("setting");
                xmlWriter.writeAttribute("name", valname);
                xmlWriter.writeAttribute("value", values[valname]);
                int _onlyinfileviewindex = onlyinfileviewindex[valname];
                if (_onlyinfileviewindex != -1)
                    xmlWriter.writeAttribute("onlyinfileview", QVariant(_onlyinfileviewindex).toString());
                xmlWriter.writeEndElement();
            }
            foreach(QString inheritname, inheritFrom) {
                xmlWriter.writeStartElement("inherits-from");
                xmlWriter.writeAttribute("profilename", inheritname);
                xmlWriter.writeEndElement();
            }
            xmlWriter.writeEndElement();
            xmlWriter.writeEndDocument();
            if (outfile.error() != QFile::NoError)
                throw sh::exceptions::ProgramException(sh::exceptions::ExceptionData().message(tr("Errors while writing settings to '%1': %2").arg(profilefilename).arg(outfile.errorString())));
        }
    }
    readStoredProfiles();
}

void sh::settings::SettingsManager::invalidateCache()
{
    CHECK_CORRECT_CALLER_THREAD
    _getSettings_cache.clear();
}

void sh::settings::SettingsManager::applyToFileView(sh::ui::FileView *list, QMap<QString, QString> settings)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    foreach(QString settingname, settings.keys()) {
        QString settingvalue = settings[settingname];
        std::shared_ptr<sh::settings::Setting> setting = sh::settings::SettingsManager::instance()->getSettingByName(settingname);
        if (setting != 0)
            setting->setValue(list, settingvalue);
    }
}

void sh::settings::SettingsManager::applyToMainWindow(QMap<QString, QString> settings)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    foreach(QString settingname, settings.keys()) {
        QString settingvalue = settings[settingname];
        std::shared_ptr<sh::settings::Setting> setting = sh::settings::SettingsManager::instance()->getSettingByName(settingname);
        if (setting != 0)
            setting->setValue(settingvalue);
    }
}

void sh::settings::SettingsManager::applyPerEurlSettingsToFileView(sh::ui::FileView* list)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    QString profilename = sh::ui::MainWindow::mainWindow()->currentProfile();
    QMap<QString,QString> settings = _getSettings(list->node()->eurl(), profilename, list->index());
    applyToFileView(list, settings);
}

void sh::settings::SettingsManager::applyGlobalSettingsToFileView(sh::ui::FileView* list)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    QString profilename = sh::ui::MainWindow::mainWindow()->currentProfile();
    QMap<QString,QString> settings = _getSettings(0, profilename, list->index());
    applyToFileView(list, settings);
}

void sh::settings::SettingsManager::applyPerEurlSettingsToMainWindow()
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    QString profilename = sh::ui::MainWindow::mainWindow()->currentProfile();
    sh::ui::FileView* list = sh::ui::MainWindow::mainWindow()->currentFileView();
    QMap<QString,QString> settings = _getSettings(list->node()->eurl(), profilename, -1);
    applyToMainWindow(settings);
}

void sh::settings::SettingsManager::applyGlobalSettingsToMainWindow()
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    QString profilename = sh::ui::MainWindow::mainWindow()->currentProfile();
    QMap<QString,QString> settings = _getSettings(0, profilename, -1);
    applyToMainWindow(settings);
}

QMap<QString, QString> sh::settings::SettingsManager::_getSettings(std::shared_ptr<const sh::filesystem::Eurl> eurl, QString profilename, int fileviewindex, bool directNode)
{
    CHECK_CORRECT_CALLER_THREAD
    QString cachekey = QString("%1/%2/%3//%4").arg(directNode).arg(fileviewindex).arg(eurl?eurl->asString():"").arg(profilename);
    auto itr =  _getSettings_cache.find(cachekey);
    if (itr != _getSettings_cache.end())
        return *itr;
    QMap<QString, QString> settings;
    foreach (_SettingsFinderStructure _fsettings, _getSettings_flat(eurl, profilename)) {
        if (directNode || _fsettings._withSubfolders) {
            foreach (QString _par, _fsettings._inheritFrom) {
                QMap<QString, QString> inheritedsettings = _getSettings(eurl, _par, fileviewindex, directNode);
                foreach(QString _key, inheritedsettings.keys())
                    settings[_key]=inheritedsettings[_key];
            }
            foreach (QString _key, _fsettings._settings.keys()) {
                std::shared_ptr<sh::settings::Setting> s = getSettingByName(_key);
                if (s != 0 && (eurl==nullptr && directNode) == s->isGlobal()) { // for indirect nodes, we skip global settings, since we already have seen them
                    if ((fileviewindex!=-1) == s->isPerFileview()) {
                        int onlyinfileviewindex = _fsettings._onlyinfileviewindexes[_key];
                        if (onlyinfileviewindex == -1 || fileviewindex == -1 || onlyinfileviewindex == fileviewindex)
                            settings[_key]=_fsettings._settings[_key];
                    }
                }
            }
        }
    }
    if (eurl) {
        std::shared_ptr<const sh::filesystem::Eurl> rooteurl = sh::filesystem::FilesystemModel::instance()->rootNode()->eurl();
        std::shared_ptr<const sh::filesystem::Eurl> parent;
        if (eurl == rooteurl)
            parent = 0;
        else if (eurl->outerUrlIsRootDirectory())
            parent = rooteurl;
        else
            parent = eurl->parentSegment();
        QMap<QString, QString> _settings = _getSettings(parent, profilename, fileviewindex, false);
        foreach (QString _key, _settings.keys())
            if (!settings.contains(_key))
                settings[_key]=_settings[_key];
    }
    _getSettings_cache[cachekey] = settings;
    return settings;
}

QList<sh::settings::SettingsManager::_SettingsFinderStructure> sh::settings::SettingsManager::_getSettings_flat(std::shared_ptr<const sh::filesystem::Eurl>eurl, QString profilename)
{
    CHECK_CORRECT_CALLER_THREAD
    QList<sh::settings::SettingsManager::_SettingsFinderStructure> result;
    auto itr = _profileNodes.find(eurl);
    if (itr != _profileNodes.end()) {
        QMap<QString, QList<sh::settings::ProfileNode*> > x = *itr;
        auto it1 = x.constFind(profilename);
        if (it1 != x.constEnd()) {
            foreach (sh::settings::ProfileNode* node, *it1) {
                QMap<QString, QString> values;
                QMap<QString, int> onlyinfileview;
                foreach (QString settingname, node->getSettingNames()) {
                    values[settingname] = node->getSettingValue(settingname);
                    onlyinfileview[settingname] = node->getSettingOnlyInFileviewIndex(settingname);
                }
                const QStringList inheritFrom = node->inheritsFrom();
                const bool withSubfolders = node->withSubfolders();
                result.append(sh::settings::SettingsManager::_SettingsFinderStructure(values, onlyinfileview, inheritFrom, withSubfolders));
            }
        }
    }
    return result;
}

void sh::settings::SettingsManager::removeProfile(QString profile)
{
    CHECK_CORRECT_CALLER_THREAD
    {
        sh::tools::UserDirLock userdirlock;
        QDir pdir(sh::base::ShallotProcess::userDataDir()+"profiles/");
        QStringList filter;
        filter << "*.xsp";
        foreach (QString e, pdir.entryList(filter)) {
            QString fullprofilenodefilename=pdir.path()+"/"+e;
            sh::settings::ProfileNode profilenode;
            QXmlSimpleReader xmlReader;
            QFile profilenodefile(fullprofilenodefilename);
            QXmlInputSource source(&profilenodefile);
            _MyHandler handler(&profilenode, e);
            xmlReader.setContentHandler(&handler);
            xmlReader.setErrorHandler(&handler);
            xmlReader.parse(&source);
            QString fn = pdir.path()+"/"+e;
            if (profilenode.profilename() == profile)
                if (!QFile(fn).remove())
                    throw sh::exceptions::IOException(sh::exceptions::ExceptionData().detailsAreInteresting().details(tr("unable to remove file '%1'.").arg(fn)).retryable(false));
        }
    }
    readStoredProfiles();
}

void sh::settings::SettingsManager::removeNode(QString nodeId)
{
    CHECK_CORRECT_CALLER_THREAD
    {
        sh::tools::UserDirLock userdirlock;
        if(!QFile(nodeId).remove())
            throw sh::exceptions::IOException(sh::exceptions::ExceptionData().detailsAreInteresting().details(tr("unable to remove file '%1'.").arg(nodeId)).retryable(false));
    }
    readStoredProfiles();
}

QStringList sh::settings::SettingsManager::getProfileList()
{
    CHECK_CORRECT_CALLER_THREAD
    QStringList res = _profileNames;
    if (!res.contains("main"))
        res.append("main");
    return res;
}

QList<sh::settings::ProfileNode*> sh::settings::SettingsManager::getNodesForProfile(QString profile)
{
    CHECK_CORRECT_CALLER_THREAD
    QList<sh::settings::ProfileNode*> result;
    foreach (std::shared_ptr<const sh::filesystem::Eurl> eurl, _profileNodes.keys()) {
        auto o = _profileNodes[eurl];
        auto itr = o.find(profile);
        if (itr != o.end()) {
            QList<sh::settings::ProfileNode*> nd = *itr;
            foreach(sh::settings::ProfileNode* n, nd)
                result.append(n);
        }
    }
    return result;
}

sh::settings::SettingsManager::_SettingsFinderStructure::_SettingsFinderStructure(QMap<QString, QString> settings, QMap<QString, int> onlyinfileviewindexes, QStringList inheritFrom , bool withSubfolders)
    :_settings(settings), _onlyinfileviewindexes(onlyinfileviewindexes), _inheritFrom(inheritFrom), _withSubfolders(withSubfolders)
{
}

sh::settings::SettingsManager::_SettingsFinderStructure::_SettingsFinderStructure()
    : _SettingsFinderStructure(QMap<QString,QString>(), QMap<QString,int>(), QStringList(), false)
{
}

sh::settings::SettingsManager::_MyHandler::_MyHandler(sh::settings::ProfileNode* node, QString filename)
    :_node(node), _filename(filename)
{
}

void sh::settings::SettingsManager::_MyHandler::throwError()
{
    sh::exceptions::Exception::executeGuarded([&](){
        throw sh::exceptions::IOException(sh::exceptions::ExceptionData().detailsAreInteresting().details("Errors in loading settings file: "+_filename));
    }, sh::exceptions::ExecuteGuardFlag::ALL_ERRORS_KILL_SHALLOT);
}

bool sh::settings::SettingsManager::_MyHandler::startElement(const QString&, const QString&, const QString &qName, const QXmlAttributes &atts)
{
    if (qName == "profile") {
        if (!_stateStarted) throwError();
        QString seurl = atts.value("eurl");
        std::shared_ptr<const sh::filesystem::Eurl> eurl = 0;
        if (!seurl.isEmpty())
            eurl = sh::filesystem::Eurl::fromString(seurl);
        _node->setEurl(eurl);
        _node->setProfilename(atts.value("name"));
        _node->setWithSubfolders(atts.value("subfolders-inherit") == "1");
        _stateStarted = false;
        _stateInProfile = true;
    }
    else if (qName == "setting") {
        if (!_stateInProfile)
            throwError();
        QString settingname = atts.value("name");
        QString settingvalue = atts.value("value");
        QString settingonlyinfileview = atts.value("onlyinfileview");
        int onlyInFileview = -1;
        if (!settingonlyinfileview.isEmpty())
            onlyInFileview = settingonlyinfileview.toInt();
        _node->setSetting(settingname, settingvalue, onlyInFileview);
    }
    else if (qName == "inherits-from") {
        if (!_stateInProfile)
            throwError();
        QString profilename = atts.value("profilename");
        QStringList npn = _node->inheritsFrom();
        npn.append(profilename);
        _node->setInheritsFrom(npn);
    }
    return true;
}

bool sh::settings::SettingsManager::_MyHandler::endElement(const QString&, const QString&, const QString &qName)
{
    if (qName == "profile") {
        _stateStarted = true;
        _stateInProfile = false;
    }
    return true;
}

sh::settings::SettingsManager::~SettingsManager()
{
}
