/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2020 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef SETTINGSMANAGER_H
#define SETTINGSMANAGER_H

#include <memory>
#include <QMap>
#include <QMutex>
#include <QObject>
#include <QVariant>
#include <QXmlDefaultHandler>
#include <QXmlSimpleReader>
#include <QXmlStreamWriter>
#include "base/singletoninitializer.h"


namespace sh{
    namespace filesystem{ class Eurl; }
    namespace ui {
        class FileView;
    }

    namespace settings{

        class Setting;
        class SettingsRegistration;
        class ProfileNode;

        /** \brief
         * Manager for Shallot settings.
         *
         * This is the more flexible way of shallot customization.
         * It includes everything you see in the 'Manage saved settings' or 'Save settings'.
         *
         * See sh::settings::Setting for more.
         */
        class SettingsManager : public QObject, public sh::base::Singleton
        {
            Q_OBJECT
            DECLARE_SINGLETON(SettingsManager)
            friend class sh::settings::SettingsRegistration;
        public:

            /** @brief Applies all stored settings, which are per-eurl to a fileview.
             */
            void applyPerEurlSettingsToFileView(sh::ui::FileView *list);

            /** @brief Applies all stored settings, which are global to a fileview.
             */
            void applyGlobalSettingsToFileView(sh::ui::FileView *list);

            /** @brief Applies all stored settings, which are per-eurl to main window.
             */
            void applyPerEurlSettingsToMainWindow();

            /** @brief Applies all stored settings, which are global to a main window.
             */
            void applyGlobalSettingsToMainWindow();

            /** @brief Returns a list of all stored settings (as `ProfileNode`s) for a given profile.
             */
            QList<sh::settings::ProfileNode *> getNodesForProfile(QString profile);

            /** @brief Searches and returns a ProfileNode by id.
             */
            sh::settings::ProfileNode* getNodeByNodeId(QString nodeId);

            /** @brief Searches and returns a Setting by name.
             */
            std::shared_ptr<sh::settings::Setting> getSettingByName(QString name);

            /** \brief Returns a list of all available settings, primarily sorted by group.
             */
            QList<std::shared_ptr<sh::settings::Setting> > getAllSettings();

            /** @brief Returns a list of all existing profiles.
             */
            QStringList getProfileList();

            /** @brief Removes a profile (including all `ProfileNode`s inside it).
             */
            void removeProfile(QString profile);

            /** @brief Removes a `ProfileNode` by id.
             */
            void removeNode(QString nodeId);

            /** @brief Stores one setting into a profile.
             */
            void storeProfile(std::shared_ptr<const sh::filesystem::Eurl> eurl, QString profilename, QMap<QString,QString> values, QMap<QString,int> onlyinfileviewindex,
                              bool withSubfolders, QStringList inheritFrom);

            ~SettingsManager();

        signals:

            /** @brief Triggered when the list of profiles or the nodes inside it change.
             *
             * Can also be triggered when nothing really changed!
             */
            void profilesChanged();

        private:

            /** @brief Used internally by _getSettings().
             */
            class _SettingsFinderStructure
            {
            public:
                _SettingsFinderStructure( QMap<QString, QString> settings, QMap<QString, int> onlyinfileviewindexes, QStringList inheritFrom, bool withSubfolders );
                _SettingsFinderStructure( );
                QMap<QString, QString> _settings;
                QMap<QString, int> _onlyinfileviewindexes;
                QStringList _inheritFrom;
                bool _withSubfolders;
                int _onlyFileviewIndex;
            };

            /** @brief Used internally for reading profile node XMLs.
             */
            class _MyHandler : public QXmlDefaultHandler
            {
            public:
                _MyHandler(sh::settings::ProfileNode* node, QString filename);
                bool startElement(const QString &namespaceURI, const QString &localName, const QString &qName, const QXmlAttributes &atts);
                bool endElement(const QString &namespaceURI, const QString &localName, const QString &qName);
            private:
                void throwError();
                bool _stateStarted = true;
                bool _stateInProfile = false;
                sh::settings::ProfileNode* _node;
                QString _filename;
            };

            /** @brief Applies some settings (as string tuples) to a fileview.
             */
            void applyToFileView(sh::ui::FileView *list, QMap<QString, QString> settings);

            /** @brief Applies some settings (as string tuples) to main window.
             */
            void applyToMainWindow(QMap<QString, QString> settings);

            /** @brief Computes all effective settings (including inheritance and sub-directories) for a given
             * situation (global or eurl-bound; fileview- or mainwindow-bound) and returns it as string tuples.
             * \param eurl The directory to consider as current (or `nullptr` for global settings).
             * \param profilename The current profile.
             * \param fileviewindex The fileview index (or -1 for mainwindow-bound settings).
             * \param directNode If the given directory is to be considered as the real current directory (not e.g. a parent).
             */
            QMap<QString, QString> _getSettings(std::shared_ptr<const sh::filesystem::Eurl> eurl, QString profilename, int fileviewindex, bool directNode=true);

            /** @brief Used internally by _getSettings().
             */
            QList<_SettingsFinderStructure> _getSettings_flat(std::shared_ptr<const sh::filesystem::Eurl> eurl, QString profilename);

            /** @brief Invalidates the _getSettings cache.
             */
            void invalidateCache();

            /** @brief Reads all profiles (including all setting nodes) from filesystem.
             */
            void readStoredProfiles();

            explicit SettingsManager();

            QMap<QString, QMap<QString, QString>> _getSettings_cache;
            QMap<QString, std::shared_ptr<sh::settings::Setting>> _settings;
            QMap<std::shared_ptr<const sh::filesystem::Eurl>, QMap<QString, QList<sh::settings::ProfileNode*> > > _profileNodes;
            QMap<QString, sh::settings::ProfileNode*> _profileNodesById;
            QList<QString> _profileNames;

        };

    }
}

#endif // SETTINGSMANAGER_H
