# Shallot is a Qt based file manager.
# Copyright (C) 2014-2020 Josef Hahn
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import *
import datetime
import hashlib
import threading
import time

from _shallot import *


# WARNING WARNING WARNING WARNING WARNING WARNING WARNING WARNING WARNING WARNING WARNING WARNING #
#                                                                                                 #
# This file is an internal part of Shallot. It implements the Python part of the scripting api.   #
#                                                                                                 #
# Don't make changes here unless you know what you do!                                            #
#                                                                                                 #
# WARNING WARNING WARNING WARNING WARNING WARNING WARNING WARNING WARNING WARNING WARNING WARNING #

"""
The Shallot API for Python modules. The manual about the Shallot scripting API gives a full reference.
"""


# (*1) Some methods are defined just for helping IDEs and creating documentation. They contain _abstract() or _nonreal(), and maybe a
# return statement for giving IDEs some hints.


def _abstract():
    raise Exceptions.ProgramException(details="The called method from a custom shallot api class is not implemented.")


def _nonreal():
    raise Exceptions.ProgramException(details="This method can't be called this way.")


def _optional(fct):
    fct._unimplemented = True
    return fct


def _isimplemented(fct):
    return not getattr(fct, "_unimplemented", False)


def _computeindex(defaultcategory, category, clss, index):
    if isinstance(clss, str):
        classname = clss
    elif isinstance(clss, type):
        classname = clss.__name__
    else:
        classname = clss.__class__.__name__
    mycategory = category if (category is not None) else defaultcategory
    myindex = index if (index is not None) else ((int(hashlib.md5(classname.encode()).hexdigest()[:4],16)%1000)*10)
    return mycategory + myindex


class _ApiProxy:
    """
    Common superclass for many classes in this interface.
    
    This class mostly has internal meaning and does not directly offer any
    services to the plugin developer. Read :ref:`h_apiproxy`.
    """

    def __init__(self, native):
        self._native = native
        native._set_buddy(self)
        self._methods=[]

    def _native(self):
        return internalstuff.getNativeObjectForBuddy(self)

    def _addmethod(self, methodsetter, methodname):
        if hasattr(self,methodname):
            targetmethod=getattr(self,methodname)
            def _method(o, *a):
                return targetmethod(*a)
            self._methods.append(_method)
        else:
            def _method(o, *a):
                raise Exceptions.ProgramException(None, "Missing method '"+methodname+"'.")
            self._methods.append(_method)
        methodsetter(_method)


class Filesystem:
    """
    The :ref:`filesystem<h_filesystem>`.
    """

    @staticmethod
    def find_nodes_for_eurl(eurl):
        """
        Get a list of shallot.Filesystem.Node for a shallot.Eurl.
        If it is unknown to the model so far, it tries to build them.
        In typical cases, this list either contains one element, or
        is empty if the filesystem handlers decide that this file
        does not exist. But in some cases, there is also more than
        one node for one shallot.Eurl (when the Eurl appears on more
        than one place in the tree).
        It only returns nodes, which are 'alive', i.e. which have a living parent
        and which are a child of this parent.
        """
        return internalstuff.findFilesystemNodesForEurl(eurl)

    @staticmethod
    def try_get_nodes_for_eurl(eurl):
        """
        Returns a list of shallot.Filesystem.Node for a shallot.Eurl.
        It only considers the current state of the in-memory model. It will
        only return nodes which are already known to the model so far.
        This operation is cheaper and handling only the known nodes is sufficient
        in many situations.
        In typical cases, this list either contains one element, or
        is empty. But in some cases, there is also more than
        one node for one shallot.Eurl (when the Eurl appears on more
        than one place in the tree).
        It only returns nodes, which are 'alive', i.e. which have a living parent
        and which are a child of this parent.
        """
        return internalstuff.tryGetFilesystemNodesForEurl(eurl)

    #: The shallot.Filesystem.Node which is the root node of the entire model.
    #: It is the  parent for all toplevel nodes.
    rootnode: 'Filesystem.Node' = internalstuff.modelRootNode()

    @staticmethod
    def refresh(eurl, forceFindParent=False, withDetails=True):
        """
        Request to refresh the internal model information for a shallot.Eurl.
        This may be a place which is already known (then a change of some metadata
        or the removal is detected) or a formerly unknown place (then new nodes
        get inserted in the model).
        
        :param eurl: The shallot.Eurl of the node.
        :param forceFindParent: Forcefully create such model information if unknown before.
        :param withDetails: If detail columns assigned to an existing node should also be updated.
        """
        internalstuff.refreshData(eurl, forceFindParent, withDetails)

    @staticmethod
    def get_or_create_node(eurl, handler, nodetype, parentnode, doinsert=True, showInitialLoadingLabel=True, hidden=False):
        """
        Returns a shallot.Filesystem.Node for using it as a child node in the model.
        It either creates a new one, if there currently is no node for this
        eurl in this parentnode, or returns the existing one.
        Even for existing ones, this call can change the nodetype of that node.
        
        :param eurl: The shallot.Eurl of the new node.
        :param handler: The shallot.Filesystem.Handler of the new node. This must always be the same one as the handler registered for the scheme of the eurl!
        :param nodetype: The shallot.Filesystem.NodeType node type.
        :param parentnode: The shallot.Filesystem.Node parent node.
        :param doinsert: If this function shall actually add the node to the model (to be exact, in some situations, this will not actually take place nonetheless).
        :param showInitialLoadingLabel: If a 'loading...' label shall be visible at beginning.
        :param hidden: If the node shall be created as hidden one. See also shallot.Filesystem.Node.ishidden.
        """
        return internalstuff.getOrCreateFilesystemNode(eurl, handler, nodetype, parentnode, doinsert, showInitialLoadingLabel, hidden)

    @staticmethod
    def create_node(eurl, handler, nodetype, parentnode, doinsert=True, showInitialLoadingLabel=True, hidden=False):
        """
        Creates a new shallot.Filesystem.Node for placing it into the model.
        If such a node (with the same eurl for the same parent node) does not exist,
        it generates a new one. If there already is such a node alive,
        but currently not placed in the model, it recycles this one.
        This can happen when references exist to a node which is not yet inserted
        or which is removed meanwhile.
        It is not allowed
        to call this method when such a node already exists in the model.
        Use this function for getting a shallot.Filesystem.Node, which is
        to be added to the model now or later.
        It is typically used within a shallot.Filesystem.Handler implementation.
        
        :param eurl: The shallot.Eurl of the new item.
        :param handler: The shallot.Filesystem.Handler of the new node. This must always be the same one as the handler registered for the scheme of the eurl!
        :param nodetype: The shallot.Filesystem.NodeType node type.
        :param parentnode: The shallot.Filesystem.Node parent node.
        :param doinsert: If this function shall actually add the node to the model (to be exact, in some situations, this will not actually take place nonetheless).
        :param showInitialLoadingLabel: If a 'loading...' label shall be visible at beginning.
        :param hidden: If the node shall be created as hidden one. See also shallot.Filesystem.Node.ishidden.
        """
        return internalstuff.createFilesystemNode(eurl, handler, nodetype, parentnode, doinsert, showInitialLoadingLabel, hidden)
    
    class NodeType:
        """
        Enumeration of types a shallot.Filesystem.Node can have.
        """
        #: Usual file.
        File = internalstuff.filesystemnodetype_file()
        #: Directory.
        Directory = internalstuff.filesystemnodetype_directory()
        #: Link.
        Link = internalstuff.filesystemnodetype_link()
        #: Unknown.
        Unknown = internalstuff.filesystemnodetype_unknown()
        #: Points to the first type (used for generating int ranges).
        FIRSTTYPE = internalstuff.filesystemnodetype_firsttype()
        #: Points to the last physical type, excluding internal magic types (used for generating int ranges).
        LASTPHYSICALTYPE = internalstuff.filesystemnodetype_lastphysicaltype()
        #: Does not exist.
        NONE = internalstuff.filesystemnodetype_none()

    class Handler(_ApiProxy):
        """
        Abstract base class for a custom filesystem handler.
        Subclass and register it for implementing a new virtual filesystem,
        which lets new nodes appear somewhere in the filesystem tree and
        controls how to handle them.
        Use shallot.Filesystem.Handler.register for registration.
        """

        def __init__(self):
            _ApiProxy.__init__(self, internalstuff.create_FilesystemHandler())
            self._addmethod(self._native.set_getmimetype, "get_mimetype")
            self._addmethod(self._native.set_gettype, "get_type")
            self._addmethod(self._native.set_getsize, "get_size")
            self._addmethod(self._native.set_getmtime, "_getMtime")
            self._addmethod(self._native.set_setmtime, "_setMtime")
            self._addmethod(self._native.set_getlinktarget, "get_linktarget")
            self._addmethod(self._native.set_listextendedattributes, "list_extendedattributes")
            self._addmethod(self._native.set_getextendedattributesize, "get_extendedattribute_size")
            self._addmethod(self._native.set_getextendedattribute, "get_extendedattribute")
            self._addmethod(self._native.set_setextendedattribute, "set_extendedattribute")
            self._addmethod(self._native.set_removeextendedattribute, "remove_extendedattribute")
            self._addmethod(self._native.set_getcustomattributes, "_getCustomAttributes")
            self._addmethod(self._native.set_setcustomattribute, "set_customattribute")
            self._addmethod(self._native.set_cancreatedirectory, "can_create_directory")
            self._addmethod(self._native.set_createdirectory, "create_directory")
            self._addmethod(self._native.set_cancreatelink, "can_create_link")
            self._addmethod(self._native.set_createlink, "create_link")
            self._addmethod(self._native.set_cancreatefile, "can_create_file")
            self._addmethod(self._native.set_candeleteitem, "can_delete_item")
            self._addmethod(self._native.set_deleteitem, "delete_item")
            self._addmethod(self._native.set_canrenameitem, "can_rename_item")
            self._addmethod(self._native.set_renameitem, "rename_item")
            self._addmethod(self._native.set_itemlist, "itemlist")
            self._addmethod(self._native.set_configureitems, "_configureItems")
            self._addmethod(self._native.set_cangetfilecontent, "can_get_filecontent")
            self._addmethod(self._native.set_getfilecontent, "_getFileContent")
            self._addmethod(self._native.set_createfile, "create_file")
            self._addmethod(self._native.set_getactions, "_getActions")

        @staticmethod
        def register(scheme, handler):
            """
            Registers a shallot.Filesystem.Handler filesystem handler implementation.
            
            :param scheme: The scheme name (first part in a shallot.Eurl, like *file*).
            :param handler: The shallot.Filesystem.Handler filesystem handler.
            """
            return internalstuff.registerFilesystemHandler(handler, scheme)

        def _getFileContent(self, op, eurl):
            return self.get_file_content(op, eurl)._native

        def configure_item(self, operation, node):
            """
            Configure a newly created node (e.g. setting another icon or changing the display name).
            |dooverride_optional|
            
            :param operation: The shallot.Operations.Operation operation object.
            :param node: The shallot.Filesystem.Node filesystem node to configure.
            """
            pass

        def _configureItems(self, op, itemlist):
            for node in itemlist:
                self.configure_item(op, node)

        def list_extendedattributes(self, operation, eurl):
            """
            Returns a list of extended attributes which exist for a location.
            |dooverride_optional|
            Extended attributes are all kinds of properties of a file which aren't handled otherwise. It can contain filesystem's extended attributes, permission information and more.
            
            :param operation: The shallot.Operations.Operation operation object.
            :param eurl: The shallot.Eurl address this call is referred to.
            """
            return []

        def get_extendedattribute_size(self, operation, eurl, attribname):
            """
            Returns the size (in byte) of the value for a particular extended attribute.
            |dooverride_optional|
            Extended attributes are all kinds of properties of a file which aren't handled otherwise. It can contain filesystem's extended attributes, permission information and more.
            
            :param operation: The shallot.Operations.Operation operation object.
            :param eurl: The shallot.Eurl address this call is referred to.
            :param attribname: The attribute name.
            """
            return 0

        def get_extendedattribute(self, operation, eurl, attribname):
            """
            Gets the value for a particular extended attribute.
            |dooverride_optional|
            Extended attributes are all kinds of properties of a file which aren't handled otherwise. It can contain filesystem's extended attributes, permission information and more.
            
            :param operation: The shallot.Operations.Operation operation object.
            :param eurl: The shallot.Eurl address this call is referred to.
            :param attribname: The attribute name.
            """
            return b''

        def set_extendedattribute(self, operation, eurl, attribname, value):
            """
            Sets the value for a particular extended attribute.
            |dooverride_optional|
            Extended attributes are all kinds of properties of a file which aren't handled otherwise. It can contain filesystem's extended attributes, permission information and more.
            
            :param operation: The shallot.Operations.Operation operation object.
            :param eurl: The shallot.Eurl address this call is referred to.
            :param attribname: The attribute name.
            :param value: The new attribute value (as byte string).
            """
            pass

        def remove_extendedattribute(self, operation, eurl, attribname):
            """
            Removes an extended attribute.
            |dooverride_optional|
            Extended attributes are all kinds of properties of a file which aren't handled otherwise. It can contain filesystem's extended attributes, permission information and more.
            
            :param operation: The shallot.Operations.Operation operation object.
            :param eurl: The shallot.Eurl address this call is referred to.
            :param attribname: The attribute name.
            """
            pass

        def get_customattributes(self, operation, eurl):
            """
            Returns a dict<string,string> with custom attributes for a node.
            |dooverride_optional|
            In contrast to extended attributes, those ones are not directly
            stored in the filesystems, but are managed in a handler specific way
            (e.g. file permissions).
            
            :param operation: The shallot.Operations.Operation operation object.
            :param eurl: The shallot.Eurl address this call is referred to.
            """
            return {}

        def _getCustomAttributes(self, op, eurl):
            result = []
            m = self.get_customattributes(op, eurl)
            for k in m.keys():
                result.append(k)
                result.append(str(m[k]))
            return result

        def set_customattribute(self, operation, eurl, key, value):
            """
            Sets a custom attribute to an entry.
            |dooverride_optional|
            See also get_customattributes.
            
            :param operation: The shallot.Operations.Operation operation object.
            :param eurl: The shallot.Eurl address this call is referred to.
            :param key: The attribute key.
            :param value: The attribute value.
            """
            pass

        def _getActions(self, eurls):
            return [x._native for x in self.get_actions(eurls)]

        def _getMtime(self, op, eurl):
            res = self.get_mtime(op, eurl)
            if res is None:
                return 0
            elif type(res) is int:
                return res
            else:
                return time.mktime(res.timetuple())*1000

        def _setMtime(self, op, eurl, itime):
            dtime = None
            if itime:
                dtime = datetime.datetime.fromtimestamp(itime/1000.0)
            self.set_mtime(op, eurl, dtime)

        def get_mimetype(self, operation, eurl):
            """
            Determine mime type of a file.
            |dooverride|
            
            :param operation: The shallot.Operations.Operation operation object.
            :param eurl: The shallot.Eurl address this call is referred to.
            """
            _abstract()

        def get_type(self, operation, eurl):
            """
            Determine if a file is regular, or a dir, or a link, ...
            |dooverride|
            
            :param operation: The shallot.Operations.Operation operation object.
            :param eurl: The shallot.Eurl address this call is referred to.
            """
            _abstract()

        def get_size(self, operation, eurl):
            """
            Determine the size of a file.
            |dooverride|
            
            :param operation: The shallot.Operations.Operation operation object.
            :param eurl: The shallot.Eurl address this call is referred to.
            """
            _abstract()

        def get_mtime(self, operation, eurl):
            """
            Determine the mtime of a file.
            |dooverride|
            
            :param operation: The shallot.Operations.Operation operation object.
            :param eurl: The shallot.Eurl address this call is referred to.
            """
            _abstract()

        def set_mtime(self, operation, eurl, mtime):
            """
            Set the mtime of a file.
            |dooverride|
            
            :param operation: The shallot.Operations.Operation operation object.
            :param eurl: The shallot.Eurl address this call is referred to.
            :param mtime: The datetime.datetime, which specifies the new modification time.
            """
            _abstract()

        def can_create_directory(self, operation, eurl):
            """
            Is it allowed to create subdirectories in a certain directory?
            |dooverride|
            
            :param operation: The shallot.Operations.Operation operation object.
            :param eurl: The shallot.Eurl address this call is referred to.
            """
            _abstract()

        def create_directory(self, operation, eurl):
            """
            Create a directory.
            |dooverride|
            
            :param operation: The shallot.Operations.Operation operation object.
            :param eurl: The shallot.Eurl address this call is referred to.
            """
            _abstract()

        def can_create_file(self, operation, eurl):
            """
            Is it allowed to create files in a certain directory?
            |dooverride|
            
            :param operation: The shallot.Operations.Operation operation object.
            :param eurl: The shallot.Eurl address this call is referred to.
            """
            _abstract()

        def can_delete_item(self, operation, eurl):
            """
            Is it allowed to delete a certain file/directory/link/...?
            |dooverride|
            
            :param operation: The shallot.Operations.Operation operation object.
            :param eurl: The shallot.Eurl address this call is referred to.
            """
            _abstract()

        def delete_item(self, operation, eurl):
            """
            Delete a file/directory/link/...
            |dooverride|
            
            :param operation: The shallot.Operations.Operation operation object.
            :param eurl: The shallot.Eurl address this call is referred to.
            """
            _abstract()

        def can_rename_item(self, operation, src):
            """
            Is it allowed to move a certain file/directory/link/...?
            |dooverride|
            
            :param operation: The shallot.Operations.Operation operation object.
            :param src: The shallot.Eurl address pointing to the potential item to be moved.
            """
            _abstract()

        def rename_item(self, operation, src, destpath):
            """
            Move a file/directory/link/... to somewhere else (can be a simple renaming).
            |dooverride|
            
            :param operation: The shallot.Operations.Operation operation object.
            :param src: The shallot.Eurl address pointing to the item to be moved.
            :param destpath: The new path for the src item.
            """
            _abstract()

        def itemlist(self, operation, eurl, nodetype, nodelist):
            """
            Determine a list of subelements in a certain directory.
            |dooverride|
            
            :param operation: The shallot.Operations.Operation operation object.
            :param eurl: The shallot.Eurl address this call is referred to.
            :param nodetype: Which kind of items are requested. See shallot.Filesystem.NodeType.
            :param nodelist: A shallot.Filesystem.NodeList object which controls the entry list for the referred directory. Operate on this object for adding children.
            """
            _abstract()

        def can_get_filecontent(self, operation, eurl):
            """
            Is it allowed to get the content of a certain file?
            |dooverride|
            
            :param operation: The shallot.Operations.Operation operation object.
            :param eurl: The shallot.Eurl address this call is referred to.
            """
            _abstract()

        def get_file_content(self, operation, eurl):
            """
            Get the content of a file as a shallot.Streaming.ReadDataDevice.
            |dooverride|
            
            :param operation: The shallot.Operations.Operation operation object.
            :param eurl: The shallot.Eurl address this call is referred to.
            """
            _abstract()

        def create_file(self, operation, eurl, outstream, handlertransfer):
            """
            Creates a file with some content.
            |dooverride|
            
            :param operation: The shallot.Operations.Operation operation object.
            :param eurl: The shallot.Eurl address this call is referred to.
            :param outstream: the content stream to be transferred into the new file
            :param handlertransfer: shallot.Operations.HandlerTransfer may be used for some better integration, like cancel support and progress monitoring.
            """
            _abstract()

        def get_actions(self, eurls):
            """
            Which actions (see former example) are to be offered for certain files?
            |dooverride|
            
            :param eurls: A list of shallot.Eurl addresses this call is referred to.
            """
            _abstract()

        def can_create_link(self, operation, eurl):
            """
            Is it allowed to create a link in a certain directory?
            |dooverride|
            
            :param operation: The shallot.Operations.Operation operation object.
            :param eurl: The shallot.Eurl address this call is referred to.
            """
            _abstract()

        def create_link(self, operation, eurl, tgt):
            """
            Create a link.
            |dooverride|
            
            :param operation: The shallot.Operations.Operation operation object.
            :param eurl: The shallot.Eurl address this call is referred to.
            :param tgt: The link destination path (as string).
            """
            _abstract()

        def get_linktarget(self, operation, eurl):
            """
            Resolve a link.
            |dooverride|
            
            :param operation: The shallot.Operations.Operation operation object.
            :param eurl: The shallot.Eurl address this call is referred to.
            """
            _abstract()

    class NodeList:
        """
        A list editor for a filesystem node list. Thay are mainly used
        for specifying child nodes in shallot.Filesystem.Handler.itemlist and provide some help there.
        In most cases, you should just use shallot.Filesystem.NodeList.set_items.
        """

        def __init__(self):
            """
            Can't be constructed directly.
            """
            _nonreal()

        def set_items(self, items):
            """
            Sets a new list content. This function is all you need in most situations. For iteratively adding nodes, which makes the intermediate results visible in the user interface, see shallot.Filesystem.NodeList.add_item.
            
            :param items: A list of strings providing the new children nodes. The list must contain the basenames of all existing children (with the node type you got as argument).
            """
            _nonreal()

        def add_item(self, item):
            """
            Adds a new children to the list, directly showing that in the user interface, while you can proceed filling the list. Please read shallot.Filesystem.NodeList.begin_iterative_adding as well! In most cases, you don't need this function.
            
            :param item: A the new children node. This must be the basename.
            """
            _nonreal()

        def begin_iterative_adding(self):
            """
            Must be called before you begin to iteratively fill the children list (e.g. with shallot.Filesystem.NodeList.add_item). You must also call shallot.Filesystem.NodeList.end_iterative_adding afterwards.
            """
            _nonreal()

        def end_iterative_adding(self):
            """
            Must be called after you iteratively filled the children list with shallot.Filesystem.NodeList.add_item. This automatically removes all the old nodes, which you haven't added in this session.
            """
            _nonreal()

    class Node:
        """
        A model representation of a location in the filesystem tree. It represents
        stuff like a file or a directory. Those instances are used at many places
        for all kinds of operations.
        
        Please note: Instances can be associated with any kind of elements in
        the filesystem (files, directories, links, ...). It might also point to something
        which does not exist at all (see parentnode).
        The documentation sometimes explicitly makes a difference
        between those kinds (files, directories, links, ...; often called 'node type').
        But it often uses the term 'file' implicitly while meaning all kinds of elements;
        assuming that e.g. a directory is just a special kind of a file. It should be clear from
        the particular context which meaning applies.
        """

        def __init__(self):
            """
            Can't be constructed directly.
            """
            _nonreal()

        def eurl(self):
            """
            Return the shallot.Eurl entry address.
            """
            _nonreal()
            return Eurl()  # *1

        def nodetype(self):
            """
            Returns the shallot.Filesystem.NodeType node type.
            """
            _nonreal()

        def add_detail(self, column):
            """
            Adds a detail to this node.
            
            :param column: The shallot.DetailColumn instance to add.
            """
            _nonreal()

        def set_icon(self, icon):
            """
            Sets the node icon.
            """
            _nonreal()

        def set_displayname(self, displayname):
            """
            Sets the displayed name of the node.
            """
            _nonreal()

        def ishidden(self):
            """
            Returns if the node is hidden.

            Note: It's not difficult for the user to also show the hidden items.
            """
            _nonreal()

        def set_hidden(self, value):
            """
            Sets if the node is hidden.

            See also ishidden().
            
            :param value: If the node shall be visible (boolean).
            """
            _nonreal()


class ConfigurationValue(_ApiProxy):
    """
    Abstract base class for a :ref:`configuration value<h_configurationvalues>`.
    """

    def __init__(self, name, defaultvalue, category=0, description="", longdescription="", changehint=None):
        """
        :param name: The config value name.
        :param defaultvalue: The default value.
        :param category: The category. One of shallot.ConfigurationValue.Category.
        :param description: The short description.
        :param longdescription: The long description.
        :param changehint: A label hint for change buttons in user interfaces.
        """
        if type(defaultvalue) is int:
            self.vtype = int
            nat = internalstuff.register_ConfigurationValueInt(name, defaultvalue, description, category, longdescription, changehint)
        elif type(defaultvalue) is float:
            self.vtype = float
            nat = internalstuff.register_ConfigurationValueFloat(name, defaultvalue, description, category, longdescription, changehint)
        elif type(defaultvalue) is bool:
            self.vtype = bool
            nat = internalstuff.register_ConfigurationValueBool(name, defaultvalue, description, category, longdescription, changehint)
        else:
            self.vtype = None
            nat = internalstuff.register_ConfigurationValueString(name, defaultvalue, description, category, longdescription, changehint)
        _ApiProxy.__init__(self, nat)

    def value(self):
        """
        Returns the current value.
        """
        if self.vtype == int:
            return self._native.get_configvalue_int()
        elif self.vtype == float:
            return self._native.get_configvalue_float()
        elif self.vtype == bool:
            return self._native.get_configvalue_bool()
        else:
            return self._native.get_configvalue_string()

    def set_value(self, value):
        """
        Sets the value.
        
        :param value: The new value.
        """
        if self.vtype == int:
            self._native.set_configvalue_int(value)
        elif self.vtype == float:
            self._native.set_configvalue_float(value)
        elif self.vtype == bool:
            self._native.set_configvalue_bool(value)
        else:
            self._native.set_configvalue_string(value)

    class Category:
        """
        Categories of shallot.ConfigurationValue implementations.
        They are used for grouping them in the dialogs. There is no
        difference in behavior implied by this choice.
        """
        #: User interface category.
        GUI = internalstuff.configurationcategory_gui()
        #: Behavioral configuration category.
        Behavior = internalstuff.configurationcategory_behavior()
        #: External tools category.
        ExternalTools = internalstuff.configurationcategory_externaltools()


class Exceptions:
    """
    Everything about :ref:`exceptions<h_exceptions>`.
    """

    class ScriptedException(Exception):
        """
        The Shallot exception class.
        """

        def __init__(self,details=None,message=None,*,isRuntime=None,isResumeable=None,isRetryable=None,autoRetryRecommendedIn=-1,detailsAreInteresting=None,_class=""):
            """
            :param details: Detail text.
            :param message: Message text.
            :param isRuntime: Is a runtime error (instead of a program logic error).
            :param isResumeable: Is resumeable.
            :param isRetryable: Is retryable.
            :param detailsAreInteresting: If the details contain information which is directly interesting for (and consumable by) the end user.
            :param autoRetryRecommendedIn: Recommended retry interval in milliseconds.
            :param _class: Only used internally.
            """
            clss = ("|"+_class+"|") if (_class != "") else ""
            if "|IOException|" in clss:
                nativeexcdata=internalstuff.createIOException()
            elif "|ArgumentException|" in clss:
                nativeexcdata=internalstuff.createArgumentException()
            elif "|RuntimeException|" in clss:
                nativeexcdata=internalstuff.createRuntimeException()
            elif "|ProgramException|" in clss:
                nativeexcdata=internalstuff.createProgramException()
            else:
                nativeexcdata=internalstuff.createException()
            if details is not None:
                nativeexcdata["details"] = details
            if message is not None:
                nativeexcdata["message"] = message
            if isRuntime is not None:
                nativeexcdata["isruntime"] = "1" if isRuntime else "0"
            if isResumeable is not None:
                nativeexcdata["isresumeable"] = "1" if isResumeable else "0"
            if isRetryable is not None:
                nativeexcdata["isretryable"] = "1" if isRetryable else "0"
            if detailsAreInteresting is not None:
                nativeexcdata["detailsareinteresting"] = "1" if detailsAreInteresting else "0"
            if autoRetryRecommendedIn!=-1:
                nativeexcdata["autoretryrecommendedIn"] = str(autoRetryRecommendedIn)
            nativeexcdata["name"] = self.__class__.__name__
            nativeexcdata["classes"] += "|ScriptedException|" + self.__class__.__name__ + "|"
            Exception.__init__(self, nativeexcdata)

        def isExceptionClass(self,classname):
            """
            Checks if this exception is instance of a given exception class.
            
            :param classname: A exception class name (as string).
            """
            return ("|"+classname+"|") in ("|"+self.args[0]["classes"]+"|")

    class ProgramException(ScriptedException):
        """
        Shallot exception for program errors (typically logical stuff) in the program or plugin.
        It typically allows resume but no retry (special cases may override each of them).
        Read :ref:`more about Shallot Exceptions<h_exceptions>`.
        """
        def __init__(self,details=None,message=None,*,isResumeable=True,detailsAreInteresting=None,_class=""):
            """
            See shallot.Exceptions.ScriptedException.__init__ for details.
            """
            Exceptions.ScriptedException.__init__(self,details=details,message=message,isResumeable=isResumeable,detailsAreInteresting=detailsAreInteresting,_class=_class+"|ProgramException")

    class RuntimeException(ScriptedException):
        """
        Shallot exception for failed operation due to (often external) runtime effects.
        It allows resume and optionally allows retry (special cases may override each of them).
        Read :ref:`more about Shallot Exceptions<h_exceptions>`.
        """
        def __init__(self,details=None,message=None,*,isRetryable=None,autoRetryRecommendedIn=-1,detailsAreInteresting=None,_class=""):
            """
            See shallot.Exceptions.ScriptedException.__init__ for details.
            """
            Exceptions.ScriptedException.__init__(self,details=details,message=message,isRetryable=None,autoRetryRecommendedIn=-1,detailsAreInteresting=detailsAreInteresting,_class=_class+"|RuntimeException")

    class ArgumentException(ProgramException):
        """
        Shallot exception for failed operation due to invalid arguments given to some program part.
        It typically allows resume but not retry (special cases may override each of them).
        Read :ref:`more about Shallot Exceptions<h_exceptions>`.
        """
        def __init__(self,details=None,message=None,*,isResumeable=True,detailsAreInteresting=None,_class=""):
            """
            See shallot.Exceptions.ScriptedException.__init__ for details.
            """
            Exceptions.ScriptedException.__init__(self,details=details,message=message,isResumeable=None,detailsAreInteresting=detailsAreInteresting,_class=_class+"|ArgumentException")

    class IOException(RuntimeException):
        """
        Shallot exception in IO.
        It allows resume and typically allows retry (special cases may override each of them).
        Read :ref:`more about Shallot Exceptions<h_exceptions>`.
        """
        def __init__(self,details=None,message=None,*,isRetryable=None,autoRetryRecommendedIn=-1,detailsAreInteresting=None,_class=""):
            """
            See shallot.Exceptions.ScriptedException.__init__ for details.
            """
            Exceptions.ScriptedException.__init__(self,details=details,message=message,isRetryable=None,autoRetryRecommendedIn=-1,detailsAreInteresting=detailsAreInteresting,_class=_class+"|IOException")


class Logging:
    """
    Methods for :ref:`logging<h_logging>`.
    """

    @staticmethod
    def log_debug(s):
        """
        Logs a message with debug severity.
        
        :param s: The message.
        """
        internalstuff.log_debug(s)

    @staticmethod
    def log_info(s):
        """
        Logs a message with info severity.
        
        :param s: The message.
        """
        internalstuff.log_info(s)

    @staticmethod
    def log_warning(s):
        """
        Logs a message with warning severity.
        
        :param s: The message.
        """
        internalstuff.log_warning(s)

    @staticmethod
    def log_error(s):
        """
        Logs a message with message severity.
        
        :param s: The message.
        """
        internalstuff.log_error(s)


class Environment:
    """
    Environment information and some :ref:`utilities<h_utilities>`.
    """

    class Thread(_ApiProxy):
        """
        A thread executes code asynchronously.
        See also the Python documentation about threading. Although this thread
        class is not related to the Python thread class (in terms of object orientation), you should
        understand the general pitfalls which come with multithreading and you might find the
        Python synchronization tools useful for avoiding those pitfalls.
        Note: It is strictly forbidden to 'park' a thread and wait for some external event most of the time.
        Use a shallot.Environment.Timer for recurring tasks.
        """

        def __init__(self):
            _ApiProxy.__init__(self, internalstuff.create_Thread())
            self._addmethod(self._native.set_run, "run")

        def run(self):
            """
            Contains the code to execute in this thread.
            |dooverride|
            """
            _abstract()

        def start(self):
            """
            Starts the execution of this thread and calls 'run' in that thread.
            Returns immediately.
            """
            self._native.start()

        @staticmethod
        def execute_threaded(fct, *args, **kwargs):
            """
            Executes a function in a new shallot.Environment.Thread (and returns that).
            This avoids subclassing and makes the code more compact.
            'args' and 'kwargs' are additional parameters which become the parameters in the actual call of 'fct'.
            
            :param fct: The function to execute.
            """
            class MyThread(Environment.Thread):
                def __init__(self):
                    Environment.Thread.__init__(self)
                def run(self):
                    fct(*args, **kwargs)
            t = MyThread()
            t.start()
            return t

    class Timer(_ApiProxy):
        """
        A timers executes some code recurringly in some time interval.
        """

        def __init__(self):
            _ApiProxy.__init__(self, internalstuff.create_Timer())
            self._addmethod(self._native.set_run, "run")
            self._isstarted = False

        def run(self):
            """
            Contains the code to execute.
            |dooverride|
            """
            _abstract()

        def start(self, interval):
            """
            Starts the timer, which executes 'run' each 'interval' milliseconds.
            Returns immediately.
            
            :param interval: The interval in milliseconds.
            """
            if self._isstarted:
                raise Exceptions.ProgramException(details="Tried to start a timer which was already started.")
            self._isstarted = True
            self._native.start(interval)

        def stop(self):
            """
            Stops the timer.
            """
            if not self._isstarted:
                raise Exceptions.ProgramException(details="Tried to stop a timer which was already stopped.")
            self._isstarted = False
            self._native.stop()

        def is_started(self):
            """
            Returns if the timer is currently active.
            """
            return self._isstarted

        @staticmethod
        def start_in_timer(fct, interval, *args, **kwargs):
            """
            Starts a function in a shallot.Environment.Timer (and returns that).
            This avoids subclassing and makes the code more compact.
            'args' and 'kwargs' are additional parameters which become the parameters in the actual call of 'fct'.
            
            :param fct: The function to execute.
            :param interval: The interval in milliseconds.
            """
            class MyTimer(Environment.Timer):
                def __init__(self):
                    Environment.Timer.__init__(self)
                def run(self):
                    fct(*args, **kwargs)
            t = MyTimer()
            t.start(interval)
            return t

    #: If Shallot was built in debug mode.
    is_debug_build = internalstuff.is_debug_build()

    #: The directory containing the shallot data files.
    shallot_data_dir = internalstuff.shallotDataDir()

    #: The directory containing the shallot builtin plugin files.
    shallot_builtinplugin_dir = internalstuff.shallotBuiltinPluginDir()

    #: The directory containing the shallot system plugin files.
    shallot_systemplugin_dir = internalstuff.shallotSystemPluginDir()

    #: The directory containing the shallot user plugin files.
    shallot_userplugin_dir = internalstuff.shallotUserPluginDir()

    #: The language code of the current user interface language.
    shallot_language = internalstuff.shallotLanguage()


class MainWindow:
    """
    A Shallot :ref:`main window<h_mainwindow>`.
    """

    def jump_to_eurl(self, eurl):
        """
        Let the current view jump to another location.
        
        :param eurl: The shallot.Eurl address to jump to.
        """
        _nonreal()

    def get_current_directory_node(self):
        """
        Gets the shallot.Filesystem.Node selected in the current view.
        """
        _nonreal()
        return Filesystem.Node()  # *1

    @staticmethod
    def open_items(eurls):
        """
        Opens a list of files, largely as if the user had doubleclicked on them.
        
        :param eurls: a list of shallot.Eurl.
        """
        return internalstuff.open_items(eurls)

    _current=internalstuff.mainwindow()

    @staticmethod
    def current():
        """
        Returns the current shallot.MainWindow.
        """
        return MainWindow._current


class IntlStringMap:
    """
    A multi-language string map used for :ref:`localization<h_localization>` of plugins.
    
    Internally it uses shallot.IntlStringMap.IntlString instances, but is more convenient for dealing with a larger
    amount of strings.
    """

    class IntlString:
        """
        A multi-language string.
        """

        @staticmethod
        def _normalizecode(code):
            return code.upper().replace("-", "_")

        def __init__(self, strings):
            """
            Creates a multi-language string by a dictionary, considering
            the keys as language code and the values as the string in that language.
            
            Example: `shallot.IntlStringMap.IntlString({'en':'yes', 'de':'ja'})`
            """
            self._map = {IntlStringMap.IntlString._normalizecode(c): strings[c] for c in strings}

        def get(self):
            """
            Returns the best variant as plain Python string, depending on the current
            user interface language.
            """
            langcode = IntlStringMap.IntlString._normalizecode(Environment.shallot_language)
            result = self._map.get(langcode)
            if result is None:
                i = langcode.find("_")
                if i >= 0:
                    langcode = langcode[:i]
                    result = self._map.get(langcode)
            if result is None:
                langcode = "EN_US"
                result = self._map.get(langcode)
            if result is None:
                langcode = "EN"
                result = self._map.get(langcode)
            if result is None:
                if len(self._map) == 0:
                    raise Exceptions.ProgramException(details="A IntlStringMap.IntlString must not be empty.")
                else:
                    result = self._map.get(sorted(self._map.keys())[0])
            return result

    def __init__(self, **stringdicts):
        """
        Creates a map of internationalized strings. For each keyword-argument, this map will get a member
        with the key name as member name. The values are dictionaries in the same form as
        for shallot.IntlStringMap.IntlString. The members will be the best available language variants as
        plain Python strings.
        
        Example: `Strings = shallot.IntlStringMap(Foo = {"en":"foo", "it":"Fuh"}, Bar = {"en":"bar", "it":"Barra"})`
        
        Then `Strings.Foo` might contain the Python string "Fuh".
        """
        #: A map of shallot.IntlStringMap.IntlString instances.
        self.map = {}
        for kstringdict in stringdicts.keys():
            stringdict = stringdicts[kstringdict]
            intlstring = IntlStringMap.IntlString(stringdict)
            self.map[kstringdict] = intlstring
            setattr(self, kstringdict, intlstring.get())


class DetailColumn(_ApiProxy):
    """
    Abstract base class for a :ref:`detail column<h_detailcolumns>` for shallot.Filesystem.Node instances.
    
    Those can e.g. be seen in the file list view, but can also be used internally by other places.
    
    It encapsulates the retrieval logic and metadata for one piece of additional
    information a shallot.Filesystem.Node can have (e.g. the filesize). Retrieving the values
    is designed to be asynchronous. Each instance represents
    one column, while the actual logic is implemented in subclasses.
    For a new detail column, subclass this class and implement at least
    shallot.DetailColumn.determine_value.
    """

    #: A integer value to be used as base display index for very interesting columns.
    INDEX_CORE = internalstuff.displayindex_core()
    #: A integer value to be used as base display index for interesting columns.
    INDEX_INTERESTING = internalstuff.displayindex_interesting()
    #: A integer value to be used as base display index for not generally interesting columns.
    INDEX_EXOTIC = internalstuff.displayindex_exotic()

    def __init__(self, name, displayname, positionGroup=None, positionIndex=None, sort_doTypediff=True, defaultWidth=-1, isRightAligned=False):
        """
        :param name: The internal name (must be unique).
        :param displayname: The label text.
        :param positionGroup: Controls display order. See :ref:`Position Indexes<h_positionindexes>` for details. Use one of the `INDEX_*` values from shallot.DetailColumn.
        :param positionIndex: Controls display order. See :ref:`Position Indexes<h_positionindexes>` for details.
        :param sort_doTypediff: Shall sorting differ between files and directories?
        :param defaultWidth: The default width in pixels (optional).
        :param isRightAligned: If the column values are right aligned (optional).
        """
        _ApiProxy.__init__(self, internalstuff.create_DetailColumn(name, displayname, _computeindex(DetailColumn.INDEX_EXOTIC, positionGroup, self, positionIndex), sort_doTypediff, defaultWidth, isRightAligned))
        self._addmethod(self._native.set_determinevalue, "determine_value")
        self._addmethod(self._native.set_applyvalue, "apply_value")

    def determine_value(self, node, operation):
        """
        Determines the column value for a node.
        |dooverride|
        Only used internally. For querying foreign detail columns, use compute_value instead.
        
        :param node: The shallot.Filesystem.Node node for which the value must be determined.
        :param operation: The shallot.Operations.Operation operation object.
        """
        _abstract()

    def apply_value(self, eurl, operation, value):
        """
        Set the detail value for a given eurl.
        Used for transferring details in file transfers (see shallot.DetailColumn.register_as_transferrable).
        |dooverride_optional|
        
        :param eurl: The shallot.Eurl for which the value must be determined.
        :param operation: The shallot.Operations.Operation operation object.
        :param value: The detail value (as string).
        """
        _abstract()

    def compute_value(self, node, operation):
        """
        Queries the column value for a node.
        
        :param node: The shallot.Filesystem.Node node for which the value must be determined.
        :param operation: The shallot.Operations.Operation operation object.
        """
        return self._native.compute_value(node, operation)

    @staticmethod
    def find_by_name(name):
        """
        Finds a shallot.DetailColumn implementation by name.
        
        :param name: The detail column name.
        """
        return internalstuff.findDetailColumnByName(name)

    @staticmethod
    def register_as_transferrable(index, detailcolumn):
        """
        Registers a shallot.DetailColumn for transferring those details in
        file transfers (e.g. when the user copies files). This requires to
        implement some methods.
        
        :param index: An integer which controls the order of transferring.
        :param detailcolumn: The shallot.DetailColumn detail column.
        """
        return internalstuff.registerTransferrableDetailColumn(index, detailcolumn)


class Streaming:
    """
    Streams of binary data.
    """

    class ReadDataDevice(_ApiProxy):
        """
        A scripted source of binary content.
        Override this class for providing own binary content.
        You might often find it more convenient to use one of the non-abstract subclasses.
        """

        def __init__(self):
            _ApiProxy.__init__(self, internalstuff.create_ReadDataDevice())
            self._addmethod(self._native.set_getdata, "_getdata")

        def _getdata(self):
            res = self.getdata()
            if res is None:
                self._native._ended()
                return b""
            else:
                return res

        def getdata(self):
            """
            Returns a Python byte array containing the next available chunk of content data.
            This may either be the complete content, or just the next part in an arbitrary size which is convenient in your situation.
            It is allowed to return empty arrays whenever there is temporarily no data available. Return `None` when the end of stream is
            reached.
            |dooverride|
            """
            _abstract()

        def read(self):
            """
            Reads a chuck of data as Python byte array.
            An empty array signals the stream ended.
            """
            return self._native.read()

        def readall(self):
            """
            Reads the complete data as Python byte array.
            """
            return self._native.readall()

    class BlobReadDataDevice(ReadDataDevice):
        """
        A binary source backed by static binary content in a byte array.

        For really large content (many megabytes), you should avoid creating complete copies
        but use one of the other subclasses or implementing an own one.
        """

        def __init__(self, content):
            """
            :param content: The byte array containing the content.
            """
            Streaming.ReadDataDevice.__init__(self)
            self.content = content

        def getdata(self):
            """
            See shallot.Streaming.ReadDataDevice.
            """
            result = self.content
            self.content = None
            return result

    class StreamReadDataDevice(ReadDataDevice):
        """
        A binary source backed by a file object.

        The file object must be opened in blocking mode (which is the typical one).
        """

        def __init__(self, stream):
            """
            :param stream: The file object containing the content.
            """
            Streaming.ReadDataDevice.__init__(self)
            self.stream = stream

        def getdata(self):
            """
            See shallot.Streaming.ReadDataDevice.
            """
            res = self.stream.read(8092)
            if res == b"":
                return None
            else:
                return res

    class ThreadedReadDataDevice(ReadDataDevice):
        """
        A binary source dynamically created piecewise in a separate thread.
        Override this class and provide the content generator within it.
        """

        def __init__(self):
            Streaming.ReadDataDevice.__init__(self)
            self._content = []
            self._stopped = False
            self._mutex = threading.Lock()
            Environment.Thread.execute_threaded(lambda: (self.run(), self._stop()))

        def _stop(self):
            """
            |infrastructure|
            """
            with self._mutex:
                self._stopped = True

        def write(self, content):
            """
            Appends a piece of data.
            Call this function from within the run() method.
            
            :param content: The next piece of content as byte array.
            """
            with self._mutex:
                self._content.append(content)

        def run(self):
            """
            This is the content generator function.
            |dooverride|
            Generate the content in an arbitrary way and call write() once or iteratively until the complete content is written.
            This method will automatically be executed in a separate thread.
            """
            pass

        def getdata(self):
            """
            See shallot.Streaming.ReadDataDevice.
            """
            with self._mutex:
                if self._content:
                    r = self._content.pop(0)
                    return r
                elif self._stopped:
                    return None
                else:
                    return b""


class Actions:
    """
    Everything about :ref:`actions<h_actions>`.
    """

    class Predicate :
        """
        Controls when and how an action is created. Used in shallot.Actions.register.
        See base class for more information.
        """
        pass

    class OnDirectoriesPredicate(Predicate, _ApiProxy):
        """
        Shows actions only on directories.
        See base class for more information.
        """

        def __init__(self):
            _ApiProxy.__init__(self, internalstuff.create_OnDirectoriesPredicate())

    class OnFilesPredicate(Predicate, _ApiProxy):
        """
        Shows actions only on files.
        See base class for more information.
        """

        def __init__(self):
            _ApiProxy.__init__(self, internalstuff.create_OnFilesPredicate())

    class OnLinksPredicate(Predicate, _ApiProxy):
        """
        Shows actions only on links.
        See base class for more information.
        """

        def __init__(self):
            _ApiProxy.__init__(self, internalstuff.create_OnLinksPredicate())

    class OnSingleEntrySelectionPredicate(Predicate, _ApiProxy):
        """
        Shows actions only on single-entry selections.
        See base class for more information.
        """

        def __init__(self):
            _ApiProxy.__init__(self, internalstuff.create_OnSingleEntrySelectionPredicate())

    class DontResolveLinksPredicate(Predicate, _ApiProxy):
        """
        Disables links resolving.
        See base class for more information.
        """

        def __init__(self):
            _ApiProxy.__init__(self, internalstuff.create_DontResolveLinksPredicate())

    class HideOnCurrentDirectoryLevelPredicate(Predicate, _ApiProxy):
        """
        Shows actions only when not searching for 'current directory level' actions.
        See base class for more information.
        """

        def __init__(self):
            _ApiProxy.__init__(self, internalstuff.create_HideOnCurrentDirectoryLevelPredicate())

    class HideOnSelectionLevelPredicate(Predicate, _ApiProxy):
        """
        Shows actions only when not searching for 'selection level' actions.
        See base class for more information.
        """

        def __init__(self):
            _ApiProxy.__init__(self, internalstuff.create_HideOnSelectionLevelPredicate())

    class ByRegExpPredicate(Predicate, _ApiProxy):
        """
        Shows actions only when the selection paths matches a regular expression.
        See base class for more information.
        """

        def __init__(self, pattern):
            _ApiProxy.__init__(self, internalstuff.create_ByRegExpPredicate(pattern))

    class KeyShortcutPredicate(Predicate, _ApiProxy):
        """
        Sets a keyboard shortcut.
        See base class for more information.
        """

        def __init__(self, shortcut, triggers_on_currentdirectory_level):
            _ApiProxy.__init__(self, internalstuff.create_KeyShortcutPredicate(shortcut, triggers_on_currentdirectory_level))

    class PositionIndexPredicate(Predicate, _ApiProxy):
        """
        Sets a positioning information index.
        See base class for more information.
        """

        def __init__(self, index):
            _ApiProxy.__init__(self, internalstuff.create_PositionIndexPredicate(index))

    @staticmethod
    def _factory(*a):

        class _Factory(_ApiProxy):

            def __init__(self, actiontype, category, predicates):
                _ApiProxy.__init__(self, internalstuff.create_ActionFactory(category, predicates))
                self._addmethod(self._native.set_construct, "_construct")
                self.ActionType=actiontype

            def _construct(self, nodes):
                return self.ActionType(nodes)._native

        return _Factory(*a)

    @staticmethod
    def register(actiontype, category="manage", predicates=[]):
        """
        Registers a subclass of shallot.Actions.AbstractAction. This makes it permanently visible
        in the context menu of selections and/or in the toolbar. If and where the action is
        visible depends on the action implementation (see constructor parameters),
        the situation (what is selected in the user interface?)
        and the registration `mode`.
        
        :param actiontype: The class implementing shallot.Actions.AbstractAction. Its constructor
                           must be callable with just `actiontype(nodes)`. The `nodes` parameter contains
                           a list of shallot.Filesystem.Node containing the node selection.
        :param category: The action category. Used for grouping in menus. Typically available are `"create"`, `"open"` and `"manage"`.
        :param predicates: A list of shallot.Actions.Predicate controlling when and how the action is offered.
        """
        f = Actions._factory(actiontype, category, predicates)
        internalstuff.register_predicatedactionfactory(f._native)

    class ExecutionInfo:
        """
        An object for signalling action execution state changes between an
        action implementation and the Shallot core (mostly to the core).
        Most calls lead to changes in the information presented by the
        progress dialog.
        """

        def __init__(self):
            """
            Can't be constructed directly.
            """
            _nonreal()

        def set_visualprocessfeedback_active(self, v):
            """
            Sets the visibility of a visual feedback.
            """
            _nonreal()

        def is_visualprocessfeedback_active(self):
            """
            Is visual feedback visibility enforced?
            """
            _nonreal()

        def set_manual_intervention_needed(self, v):
            """
            Set if manual intervention is needed.
            """
            _nonreal()

        def is_manual_intervention_needed(self):
            """
            Is manual intervention needed?
            """
            _nonreal()

        def set_details(self, fromverb, fromobjectname, toverb, toobjectname):
            """
            Sets the progress details.
            
            :param fromverb: The verb on from-side (the source).
            :param fromobjectname: (file path) The objectname on from-side (the source).
            :param toverb: The verb on to-side (the destination).
            :param toobjectname: (file path) The objectname on to-side (the destination).
            """
            _nonreal()

        def from_verb(self):
            """
            Gets the current verb on from-side.
            """
            _nonreal()

        def from_objectname(self):
            """
            Gets the current object name on from-side.
            """
            _nonreal()

        def to_verb(self):
            """
            Gets the current verb on to-side.
            """
            _nonreal()

        def to_objectname(self):
            """
            Gets the current object name on to-side.
            """
            _nonreal()

        def set_head(self, txt):
            """
            Sets the head text.
            """
            _nonreal()

        def head(self):
            """
            Gets the head text.
            """
            _nonreal()

        def set_progress(self, done, all, label):
            """
            Sets a current determinate progress.
            
            :param done: The number of finished items.
            :param all: The gross number of items.
            :param label: Progress description text.
            """
            _nonreal()

        def set_progress_indeterminate(self, label):
            """
            Sets a current indeterminate progress.
            
            :param label: Progress description text.
            """
            _nonreal()

        def progress_done(self):
            """
            How many items are done?
            """
            _nonreal()

        def progress_all(self):
            """
            How many items are to do in sum?
            """
            _nonreal()

        def progress_text(self):
            """
            The textual representation of the progress.
            """
            _nonreal()

        def is_cancelled(self):
            """
            Is the action cancelled?
            """
            _nonreal()

        def cancel(self):
            """
            Cancel the action.
            """
            _nonreal()

        def respect_cancel(self):
            """
            Respect a cancel request. This should be called from time to time (inside loops for example).
            """
            _nonreal()

        def operation(self):
            """
            Gets the shallot.Operations.Operation transactional operation object.
            """
            _nonreal()
            return Operations.Operation()  # *1

        def add_changed_eurl(self, eurl):
            """
            Requests updating the filesystem model info for an item.
            
            :param eurl: The item location to be updated as shallot.Eurl.
            """
            _nonreal()

        def userfeedback(self):
            """
            The shallot.Actions.ExecutionUserFeedback user feedback object.
            """
            _nonreal()
            return Actions.ExecutionUserFeedback()  # *1

    class DefaultPrecedenceValues:
        """
        Reference values for calculating default precedence values of open actions.
        """
        #: Normal file open action.
        OpenFile = internalstuff.actiondefaultprecedencevalues_openfile()
        #: Special open action.
        BuiltinSpecialOpen = internalstuff.actiondefaultprecedencevalues_builtinspecialopen()

    class ExecutionUserFeedback:
        """
        An object for communication with the user in action implementations.
        It can be used to ask the user for some information or to just give some
        information to the user.
        """

        def __init__(self):
            """
            Can't be constructed directly.
            """
            _nonreal()

        def simple_messagebox(self, message, buttons, icon, defaultbutton, abortbutton):
            """
            A simple message box. Returns the chosen button.
            
            :param message: The message text.
            :param buttons: A list of buttons as or-combination of shallot.Actions.ExecutionUserFeedback.MessageBoxButton.
            :param icon: An icon name.
            :param defaultbutton: Answer for keyboard Enter as shallot.Actions.ExecutionUserFeedback.MessageBoxButton.
            May be 0 for no definition.
            :param abortbutton: Answer for keyboard Esc as shallot.Actions.ExecutionUserFeedback.MessageBoxButton.
            May be 0 for no definition.
            """
            _nonreal()

        def messagebox(self, message, buttons, icon, defaultbutton, abortbutton):
            """
            A message box. Returns the index of the chosen button.
            
            :param message: The message text.
            :param buttons: A list of buttons as string list.
            :param icon: An icon name.
            :param defaultbutton: Answer for keyboard Enter as shallot.Actions.ExecutionUserFeedback.MessageBoxButton.
            :param abortbutton: Answer for keyboard Esc as shallot.Actions.ExecutionUserFeedback.MessageBoxButton.
            """
            _nonreal()

        def simple_inputbox(self, question, defaulttext):
            """
            A simple input box.
            
            :param question: The question.
            :param defaulttext: The default answer.
            """
            _nonreal()

        class MessageBoxButton:
            """
            Buttons in a message box from shallot.Actions.ExecutionUserFeedback.
            """
            #: OK button.
            OK = internalstuff.messageboxbutton_ok()
            #: Continue button.
            Continue = internalstuff.messageboxbutton_continue()
            #: Cancel button.
            Cancel = internalstuff.messageboxbutton_cancel()
            #: Retry button.
            Retry = internalstuff.messageboxbutton_retry()
            #: Yes button.
            Yes = internalstuff.messageboxbutton_yes()
            #: No button.
            No = internalstuff.messageboxbutton_no()

    class AbstractAction:
        """
        Abstract base class for executable actions or submenu structures of them.
        See the subclasses of this class.
        They can be registered with shallot.Actions.register.
        They can be returned from shallot.Filesystem.Handler.get_actions.
        """

        def initialize(self):
            """
            Initialize the action. This should make the time-consuming parts, e.g. for determining a label or enabled state.
            |dooverride_optional|
            """
            pass

        def set_enabled(self, value):
            """
            Sets if the item is enabled.
            
            :param value: The new value.
            """
            self._native.set_enabled(value)

        def set_visible(self, value):
            """
            Sets the visibility of this item.

            :param value: The new value.
            """
            self._native.set_visible(value)

        def enabled(self):
            """
            Checks if this action is enabled.
            """
            return self._native.enabled()

        def visible(self):
            """
            Checks the visibility of this item (non-recursively).
            """
            return self._native.visible()

    class ActionAction(_ApiProxy, AbstractAction):
        """
        Abstract base class for an executable action, which can be made visible in
        menus or the toolbar or executed directly. See shallot.Actions.AbstractAction
        for more.
        """

        def __init__(self, text, enabled=True, icon="", defaultActionPrecedence=0):
            """
            :param text: A label text (can be "" if you want to directly execute it instead of adding it to some menu structure).
            :param enabled: If this action is enabled.
            :param icon: Name of an icon.
            :param defaultActionPrecedence: Precedence value for being a default action. This int value must be higher than all others for becoming the default action. See shallot.Actions.DefaultPrecedenceValues.
            """
            _ApiProxy.__init__(self, internalstuff.create_ActionActionItem(text, enabled, icon, defaultActionPrecedence))
            Actions.AbstractAction.__init__(self)
            self._addmethod(self._native.set_action, "action")
            self._addmethod(self._native.set_initialize, "initialize")

        def action(self, info):
            """
            The action implementation, i.e. what the actions should actually do.
            |dooverride|
            
            :param info: A shallot.Actions.ExecutionInfo execution info object.
            """
            _abstract()

        def execute(self):
            """
            Executes this action.
            """
            self._native._execute()

        def initialize_sync(self):
            """
            Initializes the action.
            This can be called from outside in order to do the initialization.
            """
            self._native._initializeSync()

    class SubmenuAction(_ApiProxy, AbstractAction):
        """
        Abstract base class for a submenu action, which can be made visible in menus or the toolbar.
        See shallot.Actions.AbstractAction for more.
        """

        def __init__(self, text, enabled=True, icon="", defaultActionPrecedence=0):
            """
            :param text: A label text.
            :param enabled: If this action is enabled.
            :param icon: Name of an icon.
            :param defaultActionPrecedence: Precedence value for being a default action. This int value must be higher than all others for becoming the default action. See shallot.Actions.DefaultPrecedenceValues.
            """
            _ApiProxy.__init__(self, internalstuff.create_SubmenuActionItem(text, enabled, icon, defaultActionPrecedence))
            Actions.AbstractAction.__init__(self)
            self._addmethod(self._native.set_initialize, "initialize")

        def set_subitems(self, subitems):
            """
            Sets the subitems.
            
            :param subitems: A list of shallot.Actions.AbstractAction instances.
            """
            self._native.set_subitems(subitems)


class Setting(_ApiProxy):
    """
    Abstract base class for a scripted Shallot setting (those who have to be stored manually).
    See Shallot documentation for details.
    """

    class GroupInfo:
        """
        Enumeration of groups to which a setting can belong. This is just a matter of grouping the for presentation.
        """
        #: Global aspects group.
        Global = internalstuff.settinggroup_global()
        #: User interface aspects group.
        GUI = internalstuff.settinggroup_gui()
        #: File handling aspects group.
        FileHandling = internalstuff.settinggroup_filehandling()
        #: Data view aspects group.
        DataView = internalstuff.settinggroup_dataview()
        #: Behavioral aspects group.
        Behavior = internalstuff.settinggroup_behavior()
        #: Special aspects group.
        Special = internalstuff.settinggroup_special()

    def __init__(self, name, description, group, isAdvancedSetting=False, isGlobal=False, isPerFileview=False):
        """
        :param name: Internal name (must be unique).
        :param description: Description text.
        :param group: Group. One of shallot.Setting.GroupInfo.
        :param isAdvancedSetting: Is this an advanced setting?
        :param isGlobal: Does this setting regard global aspects (instead of per-directory aspects)?
        :param isPerFileview: Does this setting regard per-fileview aspects (instead of per-mainwindow)?
        """
        _ApiProxy.__init__(self, internalstuff.createSetting(name, description, group, isAdvancedSetting, isGlobal, isPerFileview))
        self._addmethod(self._native.set_setvalue1, "_setvalue1")
        self._addmethod(self._native.set_setvalue2, "_setvalue2")
        self._addmethod(self._native.set_getvalue, "get_value")
        self._addmethod(self._native.set_valuedescription, "value_description")

    def _setvalue1(self, value):
        self.set_value(value)

    def _setvalue2(self, viewindex, value):
        self.set_value(viewindex, value)
        
    def set_value(self, value, viewindex=None):
        """
        Called from Shallot core when the value was set.
        |dooverride|
        
        :param value: The value.
        :param viewindex: View index. Will be `Nothing` for per-mainwindow settings. Otherwise a fileview index.
        """
        _abstract()
    
    def get_value(self, viewindex):
        """
        Get the currently set value.
        |dooverride|
        
        :param viewindex: View index. Will be `Nothing` for per-mainwindow settings. Otherwise a fileview index.
        """
        _abstract()
    
    def value_description(self, val):
        """
        Gets a human readable description text for a value.
        |dooverride|
        
        :param val: The value.
        """
        _abstract()
    
    @staticmethod
    def register(setting):
        """
        Registers a shallot.Setting instance in Shallot.
        
        :param setting: The shallot.Setting setting object.
        """
        return internalstuff.registerSetting(setting)


class Operations:
    """
    Everything about :ref:`operations on the filesystem<h_operations>`.
    """

    class Operation:
        """
        Transactional read and write filesystem accesses. Read :ref:`more<h_operations>`.
        """

        def get_free_cachespace():
            """
            Returns the free disk space (in bytes) available for operations like fetch_container_file() or fetch_file().
            """
            _nonreal()

        def fetch_container_file(self, eurl):
            """
            Fetches the container file for a eurl locally and returns the local path.
            Use this with care and only if needed. In many cases working with streams is the better way!
            It returns nothing for an eurl without embedded parts (like foo://[bar:///y]/x).
            If the outermost inner eurl represents a local file path, it is directly returned without copying.
            You should always regard the available disk space. See get_free_cachespace().
            
            :param eurl: The shallot.Eurl location to fetch the container for.
            """
            _nonreal()

        def fetch_file(self, eurl):
            """
            Fetches a file locally and returns the local path.
            Use this with care and only if needed. In many cases working with streams is the better way!
            If the eurl represents a local file path, it is directly returned without copying.
            You should always regard the available disk space. See get_free_cachespace().
            
            :param eurl: The shallot.Eurl location to fetch the container for.
            """
            _nonreal()

        def fetch_container_file2(self, eurl, namehint):
            """
            Fetches the container file for a eurl locally and returns the local path.
            This is like fetch_container_file() but with more options.
            
            :param eurl: The shallot.Eurl location to fetch the container for.
            :param namehint: A small file name prefix for temporary files, like `"tmp"`.
            """
            _nonreal()

        def fetch_file2(self, eurl, namehint):
            """
            Fetches a file locally and returns the local path.
            This is like fetch_file() but with more options.
            
            :param eurl: The shallot.Eurl location to fetch the container for.
            :param namehint: A small file name prefix for temporary files, like `"tmp"`.
            """
            _nonreal()

        def abort(self):
            """
            Aborts transaction dropping all pending changes.
            """
            _nonreal()

        def commit(self):
            """
            Commits transaction applying all pending changes.
            """
            _nonreal()

        def enable_detailscache(self):
            """
            Enable details caching.
            """
            _nonreal()

        def is_detailscache_enabled(self):
            """
            Is details caching enabled?
            """
            _nonreal()

        def store_detail_in_cache(self, eurl, column, value):
            """
            Stores a column detail in cache.
            
            :param eurl: The shallot.Eurl to store a detail value for.
            :param column: The shallot.DetailColumn to store a detail value for.
            :param value: The column value for this entry.
            """
            _nonreal()

        def get_detail_from_cache(self, eurl, column):
            """
            Gets a column value from cache.
            
            :param eurl: The shallot.Eurl to get a detail value for.
            :param column: The shallot.DetailColumn to get a detail value for.
            """
            _nonreal()

        def set_custom_data(self, eurl, key, value):
            """
            Stores custom data.
            
            :param eurl: The shallot.Eurl instance you want to store data about (or None).
            :param key: Custom data key name.
            :param value: Custom data value.
            """
            _nonreal()

        def get_custom_data(self, eurl, key):
            """
            Gets stored custom data.
            
            :param eurl: The shallot.Eurl instance you want to get data about (or None).
            :param key: Custom data key name.
            """
            _nonreal()

        def filesystem(self):
            """
            Gets the shallot.Operations.FilesystemOperation filesystem operation object.
            """
            _nonreal()

    @staticmethod
    def create():
        """
        Creates a new shallot.Operations.Operation.
        """
        return internalstuff.createOperation()

    class FilesystemOperationProgressMonitor(_ApiProxy):
        """
        Abstract base class for progress monitors, used for monitoring progress
        of some operations in shallot.Operations.FilesystemOperation and for conflict resolution.
        
        It makes sense to directly instantiate this class in some cases.
        """

        def __init__(self, actionExecution=None):
            """
            :param actionExecution: Optional instance of shallot.Actions.ExecutionInfo. If available, it provides some additional features
            (e.g. the user may cancel the transfer).
            """
            _ApiProxy.__init__(self, internalstuff.create_FilesystemOperationProgressMonitor(actionExecution))
            self._addmethod(self._native.set_changed, "changed")
            self._addmethod(self._native.set_resolveconflicts, "_resolveConflicts")

        def changed(self):
            """
            Reacts on progress changes.
            |dooverride_optional|
            """
            pass

        @_optional
        def resolve_conflicts(self, steps):
            """
            Resolves upcoming filesystem conflicts.
            |dooverride_optional|
            
            :param steps: A list of shallot.Operations.FilesystemOperationStep which stay in conflict.
            """
            _abstract()

        def _resolveConflicts(self, steps):
            if _isimplemented(self.resolve_conflicts):
                self.resolve_conflicts(steps)
                return True
            else:
                return False

        def has_item_info(self):
            """
            If the monitor currently has information about progress in terms of item counts.
            """
            return self._native.has_item_info()

        def done_items(self):
            """
            How many items are transferred so far.
            """
            return self._native.done_items()

        def all_items(self):
            """
            How many items are to be transferred in total.
            """
            return self._native.all_items()

        def has_bytes_info(self):
            """
            If the monitor currently has information about progress in terms of transferred bytes.
            """
            return self._native.has_bytes_info()

        def done_bytes(self):
            """
            How many bytes are transferred so far.
            """
            return self._native.done_bytes()

        def all_bytes(self):
            """
            How many bytes are to be transferred in total.
            """
            return self._native.all_bytes()

        def get_item_info_from(self):
            """
            The current source as textual representation.
            """
            return self._native.get_item_info_from()

        def get_item_info_to(self):
            """
            The current destination as textual representation.
            """
            return self._native.get_item_info_to()

        def estimation(self):
            """
            The current performance estimation as textual representation.
            """
            return self._native.estimation()

    class HandlerTransfer:
        """
        Used for some additional functionality in some transfer operations
        in shallot.Filesystem.Handler instances.
        """

        def __init__(self):
            """
            Can't be constructed directly.
            """
            _nonreal()

        def increment_transferred_bytes(self, donebytes):
            """
            Notify that a certain amount of date (in byte) are transferred.
            This is used for progress monitoring.
            """
            _nonreal()

        def respect_cancel(self):
            """
            Called from time to time for allowing the user to cancel the transfer.
            """
            _nonreal()

    class FilesystemOperationStep:
        """
        One step in a larger filesystem transfer running inside some operations in
        shallot.Operations.FilesystemOperation.
        """

        def __init__(self):
            """
            Can't be constructed directly.
            """
            _nonreal()

        class ConflictResolution:
            """
            Enumeration of conflict resolution strategies.
            """
            #: Merge source into destination directory.
            MergeDirectories = internalstuff.operationstep_conflictresolution_mergedirectories()
            #: Overwrite the destination.
            OverwriteDestination = internalstuff.operationstep_conflictresolution_overwritedestination()
            #: Rename the file at the destination before transferring.
            RenameDestinationBefore = internalstuff.operationstep_conflictresolution_renamedestinationbefore()
            #: Transfer to another destination filename.
            UseDifferentDestinationName = internalstuff.operationstep_conflictresolution_usedifferentdestinationname()
            #: Skip this element.
            Skip = internalstuff.operationstep_conflictresolution_skip()
            #: No strategy.
            Unresolved = internalstuff.operationstep_conflictresolution_unresolved()
            #: Indirectly solved. Only set by the engine.
            Indirect = internalstuff.operationstep_conflictresolution_indirect()

        def conflict_description(self):
            """
            The conflict descrition text.
            """
            _nonreal()

        def conflict_resolution_rename_destination_before_to(self):
            """
            The new destination name (if conflict_resolution is shallot.Operations.FilesystemOperationStep.ConflictResolution.RenameDestinationBefore)
            """
            _nonreal()

        def conflict_resolution_different_destination_name_to(self):
            """
            The new destination name (if conflict_resolution is shallot.Operations.FilesystemOperationStep.ConflictResolution.UseDifferentDestinationName)
            """
            _nonreal()

        def set_conflict_resolve_skip(self):
            """
            Set the conflict resolution to shallot.Operations.FilesystemOperationStep.ConflictResolution.Skip.
            """
            _nonreal()

        def set_conflict_resolve_overwrite_destination(self):
            """
            Set the conflict resolution to shallot.Operations.FilesystemOperationStep.ConflictResolution.OverwriteDestination.
            """
            _nonreal()

        def set_conflict_resolve_rename_destination_before(self, newname):
            """
            Set the conflict resolution to shallot.Operations.FilesystemOperationStep.ConflictResolution.RenameDestinationBefore.
            """
            _nonreal()

        def set_conflict_resolve_use_different_destination_name(self, newname):
            """
            Set the conflict resolution to shallot.Operations.FilesystemOperationStep.ConflictResolution.UseDifferentDestinationName.
            """
            _nonreal()

        def set_conflict_resolve_merge_directories(self):
            """
            Set the conflict resolution to shallot.Operations.FilesystemOperationStep.ConflictResolution.MergeDirectories.
            """
            _nonreal()

        def sourcetype(self):
            """
            The shallot.Filesystem.NodeType of the source.
            """
            _nonreal()

        def source(self):
            """
            The source shallot.Eurl location.
            """
            _nonreal()
            return Eurl()  # *1

        def original_destination(self):
            """
            The original destination shallot.Eurl location.
            """
            _nonreal()
            return Eurl()  # *1

        def effective_destination(self):
            """
            The effective destination shallot.Eurl location with applied conflict resolution strategies.
            """
            _nonreal()
            return Eurl()  # *1

        def set_conflict_resolve_merge_directories_check_allowed(self):
            """
            Checks if it is allowed to solve this conflict by merging directories.
            """
            _nonreal()

        def conflict_resolution(self):
            """
            The currently selected shallot.Operations.FilesystemOperationStep.ConflictResolution conflict resolution.
            """
            _nonreal()

    class FilesystemOperation:        
        """
        A high-level provider of filesystem operations. It is always based on the
        transaction of a shallot.Operations.Operation and also accessible from such an instance.
        
        Some operations allow to specify an shallot.Operations.FilesystemOperationProgressMonitor.
        You can specify an instance for getting some additional functionaly, like
        progress notification, conflict resolution and more. Please note that not each
        operation uses each aspect of shallot.Operations.FilesystemOperationProgressMonitor (e.g.
        some will not do any conflict resolution).
        """

        def __init__(self):
            """
            Can't be constructed directly.
            """
            _nonreal()

        def itemlist(self, eurl):
            """
            Gets a list of shallot.Eurl in a directory.
            
            :param eurl: The shallot.Eurl of the directory to list.
            """
            _nonreal()
            return [Eurl(), Eurl()]  # *1

        def itemlist_by_type(self, eurl, nodetype):
            """
            Gets a list of shallot.Eurl in a directory which have a certain type.
            
            :param eurl: The shallot.Eurl of the directory to list.
            :param nodetype: The shallot.Filesystem.NodeType node type to fetch.
            """
            _nonreal()
            return [Eurl(), Eurl()]  # *1

        def get_type(self, eurl):
            """
            Gets the shallot.Filesystem.NodeType node type for an entry.
            
            :param eurl: The shallot.Eurl of the considered item.
            """
            _nonreal()

        def get_filesize(self, eurl):
            """
            Gets the file size in bytes for an entry.
            
            :param eurl: The shallot.Eurl of the considered item.
            """
            _nonreal()

        def get_linktarget(self, eurl):
            """
            Gets the link target for an entry (if it is a link).
            
            :param eurl: The shallot.Eurl of the considered item.
            """
            _nonreal()

        def can_get_filecontent(self, eurl):
            """
            Can we get the file content for an entry?
            
            :param eurl: The shallot.Eurl of the considered item.
            """
            _nonreal()

        def get_file_content(self, eurl):
            """
            Gets the file content for an entry as shallot.Streaming.ReadDataDevice.
            
            :param eurl: The shallot.Eurl of the considered item.
            """
            _nonreal()
            return Streaming.ReadDataDevice()  # *1

        def can_create_directory(self, eurl):
            """
            Can we create a given directory?
            
            :param eurl: The shallot.Eurl of the considered item.
            """
            _nonreal()

        def create_directory(self, eurl, progressmon):
            """
            Creates a directory.
            
            :param eurl: The shallot.Eurl of the new directory.
            :param progressmon: A shallot.Operations.FilesystemOperationProgressMonitor instance for some additional functionality (or `None`). See shallot.Operations.FilesystemOperation for details.
            """
            _nonreal()

        def can_create_link(self, eurl):
            """
            Can we create a given link?
            
            :param eurl: The shallot.Eurl of the considered item.
            """
            _nonreal()

        def create_link(self, eurl, tgt, progressmon):
            """
            Creates a link.
            
            :param eurl: The shallot.Eurl of the new link.
            :param tgt: Target path of this link.
            :param progressmon: A shallot.Operations.FilesystemOperationProgressMonitor instance for some additional functionality (or `None`). See shallot.Operations.FilesystemOperation for details.
            """
            _nonreal()

        def can_create_file(self, eurl):
            """
            Can we create a given file?
            
            :param eurl: The shallot.Eurl of the considered item.
            """
            _nonreal()

        def create_file(self, eurl, contentdevice, progressmon):
            """
            Create a file.
            
            :param eurl: The shallot.Eurl of the new file.
            :param contentdevice: a shallot.Streaming.ReadDataDevice content device.
            :param progressmon: A shallot.Operations.FilesystemOperationProgressMonitor instance for some additional functionality (or `None`). See shallot.Operations.FilesystemOperation for details.
            """
            _nonreal()

        def can_delete_item(self, eurl):
            """
            Can we delete a given entry?
            
            :param eurl: The shallot.Eurl of the considered item.
            """
            _nonreal()

        def delete_item(self, eurl, progressmon):
            """
            Delete an entry.
            
            :param eurl: The shallot.Eurl of the item to delete.
            :param progressmon: A shallot.Operations.FilesystemOperationProgressMonitor instance for some additional functionality (or `None`). See shallot.Operations.FilesystemOperation for details.
            """
            _nonreal()

        def delete_directory_if_empty(self, eurl, progressmon):
            """
            Delete a directory entry if empty.
            
            :param eurl: The shallot.Eurl of the directory to delete.
            :param progressmon: A shallot.Operations.FilesystemOperationProgressMonitor instance for some additional functionality (or `None`). See shallot.Operations.FilesystemOperation for details.
            """
            _nonreal()

        def can_move_item(self, src, dest):
            """
            Can we move a given entry?
            
            :param src: The shallot.Eurl of the item to be moved.
            :param dest: The shallot.Eurl destination path. If it is not known, use `None`.
            """
            _nonreal()

        def move_items(self, src, tgt, progressmon):
            """
            Move entries.
            
            :param src: List of shallot.Eurl of the item to be moved.
            :param tgt: The shallot.Eurl of the new common parent destination.
            :param progressmon: A shallot.Operations.FilesystemOperationProgressMonitor instance for some additional functionality (or `None`). See shallot.Operations.FilesystemOperation for details.
            """
            _nonreal()

        def move_item(self, src, tgt, progressmon):
            """
            Move an entry.
            
            :param src: The shallot.Eurl of the item to be moved.
            :param tgt: The shallot.Eurl of the new destination.
            :param progressmon: A shallot.Operations.FilesystemOperationProgressMonitor instance for some additional functionality (or `None`). See shallot.Operations.FilesystemOperation for details.
            """
            _nonreal()

        def can_copy_item(self, src, dest):
            """
            Can we copy a given entry?
            
            :param src: The shallot.Eurl of the item to be copied.
            :param dest: The shallot.Eurl destination path. If it is not known, use `None`.
            """
            _nonreal()

        def copy_items(self, src, tgt, progressmon):
            """
            Copy entries.
            
            :param src: List of shallot.Eurl of the item to be copied.
            :param tgt: The shallot.Eurl of the new common parent destination.
            :param progressmon: A shallot.Operations.FilesystemOperationProgressMonitor instance for some additional functionality (or `None`). See shallot.Operations.FilesystemOperation for details.
            """
            _nonreal()

        def copy_item(self, src, tgt, progressmon):
            """
            Copy an entry.
            
            :param src: The shallot.Eurl of the item to be copied.
            :param tgt: The shallot.Eurl of the new destination.
            :param progressmon: A shallot.Operations.FilesystemOperationProgressMonitor instance for some additional functionality (or `None`). See shallot.Operations.FilesystemOperation for details.
            """
            _nonreal()

        def list_extendedattributes(self, eurl):
            """
            Lists all keys of extended attributes stored in the filesystem for an entry.
            
            :param eurl: The shallot.Eurl of the considered item.
            """
            _nonreal()

        def get_extendedattribute_size(self, eurl, attribute):
            """
            Returns the size of an extended attribute value stored in the filesystem for an entry.
            
            :param eurl: The shallot.Eurl of the considered item.
            :param attribute: The attribute key.
            """
            _nonreal()

        def get_extendedattribute(self, eurl, attribute):
            """
            Returns the value (as byte string) of an extended attribute value stored in the filesystem for an entry.
            
            :param eurl: The shallot.Eurl of the considered item.
            :param attribute: The attribute key.
            """
            _nonreal()

        def set_extendedattribute(self, eurl, attribute, value):
            """
            Store the value of an extended attribute in the filesystem for an entry.
            
            :param eurl: The shallot.Eurl of the considered item.
            :param attribute: The attribute key.
            :param value: The attribute value (as byte string).
            """
            _nonreal()

        def remove_extendedattribute(self, eurl, attribute):
            """
            Removes an extended attribute stored in the filesystem for an entry.
            
            :param eurl: The shallot.Eurl of the considered item.
            :param attribute: The attribute key.
            """
            _nonreal()

        def resolve_node_link(self, node):
            """
            Resolve a link as node.
            
            :param node: A shallot.Filesystem.Node to resolve.
            """
            _nonreal()

        def resolve_node_link_nonrecursive(self, node):
            """
            Resolve a link as node without recursion.
            
            :param node: A shallot.Filesystem.Node to resolve.
            """
            _nonreal()

        def resolve_eurl_link(self, eurl):
            """
            Resolve a link as Eurl.
            
            :param eurl: A shallot.Eurl to resolve.
            """
            _nonreal()
            return Eurl()  # *1

        def resolve_eurl_link_nonrecursive(self, eurl):
            """
            Resolve a link as Eurl without recursion.
            
            :param eurl: A shallot.Eurl to resolve.
            """
            _nonreal()
            return Eurl()  # *1


class PanelDetails:
    """
    Everything about :ref:`panel details<h_paneldetails>`.
    """

    class AbstractPanelDetailValueElement:
        """
        Abstract base class for an element of a panel detail row's value.
        This can be something like a text, an icon or a link button.
        See the subclasses.
        """

        def __init__(self, type, value):
            self.type = type
            self.value = value

    class PanelDetailValueElementString(AbstractPanelDetailValueElement):
        """
        A piece of text as element of a panel detail row's value.
        """

        def __init__(self, text):
            """
            :param text: The text.
            """
            PanelDetails.AbstractPanelDetailValueElement.__init__(self, "text", text)

    class PanelDetailValueElementIcon(AbstractPanelDetailValueElement):
        """
        An icon as element of a panel detail row's value.
        """

        def __init__(self, icon):
            """
            :param icon: The icon name.
            """
            PanelDetails.AbstractPanelDetailValueElement.__init__(self, "icon", icon)

    class PanelDetailValueElementButton(AbstractPanelDetailValueElement):
        """
        A link button (for executing some code on user behalf) as element of a panel detail row's value.
        Use shallot.PanelDetails.SingleSelectionPanelDetail.link_triggered or shallot.PanelDetails.MultiSelectionPanelDetail.link_triggered
        for handling user actions.
        """

        def __init__(self, text, target, autohide=True):
            """
            :param text: The text.
            :param target: The link target name (will be passed to the handler function).
            :param autohide: If the button shall only be visible when the mouse cursor is in the details panel.
            """
            PanelDetails.AbstractPanelDetailValueElement.__init__(self, "link", "{}\n{}\n{}".format(target.replace("\n", ""), "1" if autohide else "0", text.replace("\n", "")))

    class PanelDetailValueElementWaiting(AbstractPanelDetailValueElement):
        """
        A 'waiting' placeholder as element of a panel detail row's value.
        """

        def __init__(self):
            PanelDetails.AbstractPanelDetailValueElement.__init__(self, "wait", "")

    class PanelDetailInst:
        """
        The storage for detail value rows, which are the actual content of a shallot.PanelDetails.PanelDetail
        for an actual file selection.
        """

        def __init__(self, _detail):
            """
            Can't be constructed directly.
            """
            self._detail = _detail

        def set_row(self, row, val):
            """
            Sets the value for one row.
            
            :param row: The label of that row.
            :param val: The value (which is a list of shallot.PanelDetails.AbstractPanelDetailValueElement).
            """
            valll = []
            for v in val:
                valll.append(v.type)
                valll.append(v.value)
            self._detail.set_row(row, valll)

    class PanelDetail:
        """
        Abstract base class for a detail panel entry. See the subclasses.
        """

        #: A integer value to be used as base display index for very interesting details.
        INDEX_VERYINTERESTING = internalstuff.paneldetail_positionindex_veryinteresting()
        #: A integer value to be used as base display index for interesting details.
        INDEX_INTERESTING = internalstuff.paneldetail_positionindex_interesting()
        #: A integer value to be used as base display index for exotic details.
        INDEX_EXOTIC = internalstuff.paneldetail_positionindex_exotic()

        def __init__(self):
            pass

        def _setValue(self, detail, nodes, op):
            _detail = PanelDetails.PanelDetailInst(detail)
            valls = self.set_value(_detail, nodes, op)
            if valls:
                for row, vall in valls:
                    _detail.set_row(row, vall)

        @staticmethod
        def register(detail):
            """
            Registers a shallot.PanelDetails.PanelDetail in the Shallot core.
            
            :param detail: The shallot.PanelDetails.PanelDetail instance.
            """
            if isinstance(detail, PanelDetails.SingleSelectionPanelDetail):
                internalstuff.registerPanelDetailFactorySingle(detail)
            elif isinstance(detail, PanelDetails.MultiSelectionPanelDetail):
                internalstuff.registerPanelDetailFactoryMulti(detail)
            else:
                raise Exceptions.ArgumentException(details="Invalid panel detail object: "+str(detail))

    class SingleSelectionPanelDetail(_ApiProxy, PanelDetail):
        """
        Abstract base class for a detail panel entry, which occurs when one item is selected.
        See shallot.PanelDetails.PanelDetail.register for registering custom implementations to shallot.
        """

        def __init__(self, positionGroup=None, positionIndex=None, valueWidthHint=4):
            """
            :param positionGroup: Controls display order. See :ref:`Position Indexes<h_positionindexes>` for details. Use one of the `INDEX_*` values from shallot.PanelDetails.PanelDetail.
            :param positionIndex: Controls display order. See :ref:`Position Indexes<h_positionindexes>` for details.
            :param valueWidthHint: A width in centimeters to reserve for printing the values.
            """
            _ApiProxy.__init__(self, internalstuff.create_DetailFactorySingle(_computeindex(PanelDetails.PanelDetail.INDEX_EXOTIC, positionGroup, self, positionIndex), valueWidthHint))
            PanelDetails.PanelDetail.__init__(self)
            self._addmethod(self._native.set_setvalue, "_setValue")
            self._addmethod(self._native.set_linktriggered, "link_triggered")

        def set_value(self, detail, node, operation):
            """
            This method is called whenever an output must be determined for a certain node.
            It must return a list of 2-tuples (one for each row). Those are each name/value pairs with the label of that row and the value (which is a list of shallot.PanelDetails.PanelDetails.AbstractPanelDetailValueElement).
            For large waiting times, you should just return a value list only containing a shallot.PanelDetails.PanelDetailValueElementWaiting for your rows, start an asynchronous execution and use shallot.PanelDetailInst.set_row in the end.
            |dooverride|
            
            :param detail: A shallot.PanelDetailInst object for printing the details. Only used in advanced cases.
            :param node: The selected shallot.Filesystem.Node.
            :param operation: The shallot.Operations.Operation operation object.
            """
            _abstract()

        def link_triggered(self, node, linktarget):
            """
            This method is called whenever the user triggers a link.
            Override in custom implementations if links are used.
            
            :param node: The selected shallot.Filesystem.Node.
            :param linktarget: The link target string, as specified in shallot.PanelDetails.PanelDetailValueElementButton.__init__.
            """
            _abstract()

    class MultiSelectionPanelDetail(_ApiProxy, PanelDetail):
        """
        Abstract base class for a detail panel entry, which occurs when multiple items are selected.
        See shallot.PanelDetails.PanelDetail.register for registering custom implementations to shallot.
        """

        def __init__(self, positionGroup=None, positionIndex=None, valueWidthHint=4):
            """
            :param positionGroup: Controls display order. See :ref:`Position Indexes<h_positionindexes>` for details. Use one of the `INDEX_*` values from shallot.PanelDetails.PanelDetail.
            :param positionIndex: Controls display order. See :ref:`Position Indexes<h_positionindexes>` for details.
            :param valueWidthHint: A width in centimeters to reserve for printing the values.
            """
            _ApiProxy.__init__(self, internalstuff.create_DetailFactoryMulti(_computeindex(PanelDetails.PanelDetail.INDEX_EXOTIC, positionGroup, self, positionIndex), valueWidthHint))
            PanelDetails.PanelDetail.__init__(self)
            self._addmethod(self._native.set_setvalue, "_setValue")
            self._addmethod(self._native.set_linktriggered, "link_triggered")

        def set_value(self, detail, nodes, operation):
            """
            This method is called whenever an output must be determined for certain nodes.
            It must return a list of 2-tuples (one for each row). Those are each name/value pairs with the label of that row and the value (which is a list of shallot.PanelDetails.AbstractPanelDetailValueElement).
            For large waiting times, you should just return an empty value list for your rows, start an asynchronous execution and use shallot.PanelDetailInst.set_row in the end.
            |dooverride|
            
            :param detail: A shallot.PanelDetailInst object for printing the details. Only used in advanced cases.
            :param nodes: The selected list of shallot.Filesystem.Node object.
            :param operation: The shallot.Operations.Operation operation object.
            """
            _abstract()

        def link_triggered(self, nodes, linktarget):
            """
            This method is called whenever the user triggers a link.
            Override in custom implementations if links are used.
            
            :param nodes: The selected list of shallot.Filesystem.Node object.
            :param linktarget: The link target string, as specified in shallot.PanelDetails.PanelDetailValueElementButton.__init__.
            """
            _abstract()
    

class FilePropertyDialog:
    """
    The :ref:`file property dialog<h_filepropertydialog>`.
    """

    class TabPropertyIconTextBanner:
        """
        Represents values for image/text banners as used for shallot.FilePropertyDialog.Tab.PropertyType.IconTextBanner.
        """

        def __init__(self):
            self.items = []

        def add_text(self, text):
            """
            Adds text to the banner.
            
            :param text: The text to append.
            """
            self.items.append(("text", text))

        def add_icon(self, iconname, size=1.0):
            """
            Adds an icon to the banner.
            
            :param iconname: The name of the icon to append.
            :param size: Icon size (not in pixels but relative to the default).
            """
            self.items.append(("icon", "{};{}".format(size, iconname)))

    class Tab(_ApiProxy):
        """
        Abstract base class for a tab in the Properties dialog. It provides content and can also offer user interactions.
        See shallot.FilePropertyDialog.Tab.register for registering custom implementations to shallot.
        """

        #: An integer value to be used as base value for specifying order indexes of core (i.e. maximum important) properties.
        INDEX_CORE = internalstuff.filepropertydialogtab_index_core()
        #: An integer value to be used as base value for specifying order indexes of very important properties.
        INDEX_VERYIMPORTANT = internalstuff.filepropertydialogtab_index_veryimportant()
        #: An integer value to be used as base value for specifying order indexes of important properties.
        INDEX_IMPORTANT = internalstuff.filepropertydialogtab_index_important()
        #: An integer value to be used as base value for specifying order indexes of properties with normal importance.
        INDEX_NORMAL = internalstuff.filepropertydialogtab_index_normal()
        #: An integer value to be used as base value for specifying order indexes of exotic (i.e. less important) properties.
        INDEX_EXOTIC = internalstuff.filepropertydialogtab_index_exotic()

        class PropertyType:
            """
            A type of a single property. Used in shallot.FilePropertyDialog.Tab.PropertyConfig.
            """
            #: A single string. If chosen, return a string in update_widget.
            String = internalstuff.filepropertydialogtab_propertytype_string()
            #: A table of key/value string pairs. If chosen, return a list of (string,string) tuples in update_widget.
            StringMap = internalstuff.filepropertydialogtab_propertytype_stringmap()
            #: A mixed banner with texts and icons in a horizontal alignment. If chosen, return a shallot.FilePropertyDialog.TabPropertyIconTextBanner in update_widget.
            IconTextBanner = internalstuff.filepropertydialogtab_propertytype_icontextbanner()

        @staticmethod
        def register(tabclass, positionGroup=None, positionIndex=None):
            """
            Registers a shallot.FilePropertyDialog.Tab implementation for making
            its content available in the Properties dialog.
            
            :param positionGroup: Controls display order. See :ref:`Position Indexes<h_positionindexes>` for details. Use one of the `INDEX_*` values from shallot.FilePropertyDialog.Tab.
            :param positionIndex: Controls display order. See :ref:`Position Indexes<h_positionindexes>` for details.
            :param tabclass: A subclass of shallot.FilePropertyDialog.Tab.
            """
            factory = FilePropertyDialog._tabFactory(tabclass)
            internalstuff.registerFilePropertyDialogTabFactory(_computeindex(FilePropertyDialog.Tab.INDEX_EXOTIC, positionGroup, tabclass, positionIndex), factory)

        def __init__(self, title, properties):
            """
            :param title: The title text for this tab.
            :param properties: A list of shallot.FilePropertyDialog.Tab.PropertyConfig instances. Each instances specifies one piece of information
            you want to provide.
            """
            lproperties = []
            self.properties = properties
            self.btnid2btn = {}
            for x in properties:
                lproperties.append(str(x.title))
                lproperties.append(str(x.propertytype))
                lproperties.append(str(len(x.buttons)))
                for btn in x.buttons:
                    lproperties.append(btn.label)
                    lproperties.append(btn.id)
                    self.btnid2btn[btn.id] = btn
            _ApiProxy.__init__(self, internalstuff.create_FilePropertyDialogTab(title, lproperties))
            self._addmethod(self._native.set_updatewidget, "_updateWidget")
            self._addmethod(self._native.set_buttontriggered, "_buttonTriggered")
            self._widgets = {} # okay, this is a hack ;)

        def refresh(self):
            """
            Reloads the content in the dialog (typically called after the external situation has changed so the dialog shows outdated information).
            """
            self._native.refresh()

        def selected_index_for_property(self, index):
            """
            For a shallot.FilePropertyDialog.Tab.PropertyType.StringMap property, this returns the index of the row selected by the user in the dialog (or -1 for no selection).
            It is not allowed to call it for other properties.
            """
            return self._native.selected_index_for_property(self._widgets[index])

        def nodes(self):
            """
            Returns a list of shallot.Filesystem.Node containing the nodes to show (typically one).
            """
            return list(self._nodes)

        def _updateWidget(self, i, operation, widgetptr):
            self._nodes = self._native.nodes()
            self._widgets[i] = widgetptr
            v = self.update_widget(i, operation)
            if isinstance(v, FilePropertyDialog.TabPropertyIconTextBanner):
                v = v.items
            if v is None:
                return ["visible", "0"]
            elif isinstance(v, str):
                return ["text", v, "visible", "1"]
            elif isinstance(v, list):
                result = ["visible", "1"]
                for ol in v:
                    f,s = ol
                    result += ["val_"+str(f), str(s)]
                return result

        def update_widget(self, i, operation):
            """
            Provides the actual content. Returns a value depending on the type choice in shallot.FilePropertyDialog.Tab.PropertyConfig.
            |dooverride|
            
            :param i: The index (in the same order as the `properties` constructor parameter).
            :param operation: A shallot.Operations.Operation instance.
            """
            _abstract()

        def _buttonTriggered(self, k):
            self.btnid2btn[k].fct()

        class PropertyConfig:
            """
            Definitions for properties, which present a piece of information in a shallot.FilePropertyDialog.Tab.
            """

            def __init__(self, title, propertytype, buttons=[]):
                """
                :param title: The property title/name.
                :param propertytype: The value type. This controls what you have to return in shallot.FilePropertyDialog.Tab.update_widget and how it is displayed. See the PROPERTYTYPE_* constants.
                :param buttons: A list of shallot.FilePropertyDialog.Tab.PropertyButtonConfig specifying which buttons to show for offering user interaction.
                """
                self.title = title
                self.propertytype = propertytype
                self.buttons = list(buttons)

        class PropertyButtonConfig:
            """
            Definitions for buttons in a shallot.FilePropertyDialog.Tab.PropertyConfig.
            """

            _idcounter = 0

            def __init__(self, label, fct):
                """
                :param label: The button text.
                :param fct: The function to execute when the user clicks on the button.
                """
                self.label = label
                self.fct = fct
                self.id = "api" + str(FilePropertyDialog.Tab.PropertyButtonConfig._idcounter + 1)
                FilePropertyDialog.Tab.PropertyButtonConfig._idcounter+= 1

    @staticmethod
    def _tabFactory(tabclass):

        class _TabFactory(_ApiProxy):

            def __init__(self, tabclass):
                _ApiProxy.__init__(self, internalstuff.create_FilePropertyDialogTabFactory())
                self._addmethod(self._native.set_construct, "construct")
                self.tabclass = tabclass

            def construct(self):
                return self.tabclass()

        return _TabFactory(tabclass)


class FileSearch:
    """
    File :ref:`searches<h_filesearches>`.
    """

    class SearchCriterion(_ApiProxy):
        """
        Abstract base class for a search criterion. It implements custom file search filters.
        Implement this class as well as shallot.FileSearch.SearchCriterionFactory and register this pair with
        shallot.FileSearch.SearchCriterionFactory.register.
        """

        def __init__(self, factory):
            """
            Each implementation must offer a constructor with this signature. It must forward the `factory`
            parameter to this constructor.
            """
            _ApiProxy.__init__(self, internalstuff.create_SearchCriterion(factory))
            self._addmethod(self._native.set_configure, "_configure")
            self._addmethod(self._native.set_match, "match")

        def configure(self, info):
            """
            Asks some questions to the user (with `info` parameter) in order to determine a
            configuration and returns it as a list of strings. Some other methods must
            be able to interpret this list in order to get back this configuration.
            |dooverride|
            
            :param info: A shallot.Actions.ExecutionInfo execution info object.
            """
            _abstract()

        def _configure(self, info):
            res = self.configure(info)
            if res is not None:
                res = [str(x) for x in res]
            info.operation().set_custom_data(None, "searchconfiguration", res)

        def searchspec(self):
            """
            Returns the current search configuration as string list.
            """
            return self._native.searchspec()

        def match(self, operation, eurl):
            """
            Determines if a file matches the configured search criteria (returns bool).
            |dooverride|
            Use searchspec for getting the current configuration.
            
            :param operation: The shallot.Operations.Operation operation object.
            :param eurl: The shallot.Eurl on which the filter logic must act.
            """
            _abstract()

    class SearchCriterionFactory(_ApiProxy):
        """
        A factory for a shallot.FileSearch.SearchCriterion class.
        """

        #: Base index value for core (i.e. very interesting) search criteria.
        INDEX_CORE = internalstuff.searchcriterionfactory_index_core()
        #: Base index value for search criteria with normal importance.
        INDEX_NORMAL = internalstuff.searchcriterionfactory_index_normal()
        #: Base index value for exotic search criteria.
        INDEX_EXOTIC = internalstuff.searchcriterionfactory_index_exotic()

        def __init__(self, key, ctype, description):
            """
            Call this constructor from subclasses.
            
            :param key: A short string used as key for this criterion.
            :param ctype: The type of your shallot.FileSearch.SearchCriterion implementation to create.
            :param description: The short description (as shown in menus).
            """
            _ApiProxy.__init__(self, internalstuff.create_SearchCriterionFactory(key, description))
            self._addmethod(self._native.set_getsearchspecdesc, "get_searchspec_description")
            self._addmethod(self._native.set_construct, "_construct")
            self._addmethod(self._native.set_isvisiblefor, "is_visible_for")
            self._ctype = ctype

        def _construct(self):
            return self._ctype(self)

        def get_searchspec_description(self, val):
            """
            Returns the description of a search configuration (for displaying in the user interface).
            |dooverride_optional|
            
            :param val: The search configuration as string list.
            """
            return None

        def is_visible_for(self, operation, node):
            """
            Determines if this search criterion should be offered for a node.
            |dooverride_optional|
            
            :param operation: The shallot.Operations.Operation operation object.
            :param node: The shallot.Filesystem.Node node for which the visibility must be determined.
            """
            return True

        @staticmethod
        def register(criterionfactory, positionGroup=None, positionIndex=None):
            """
            Registers a custom shallot.FileSearch.SearchCriterionFactory instance for offering custom file searching filters.
            
            :param criterionfactory: An instance of shallot.FileSearch.SearchCriterionFactory.
            :param positionGroup: Controls display order. See :ref:`Position Indexes<h_positionindexes>` for details. Use one of the `INDEX_*` values from shallot.FileSearch.SearchCriterionFactory.
            :param positionIndex: Controls display order. See :ref:`Position Indexes<h_positionindexes>` for details.
            """
            internalstuff.registerSearchCriterionFactory(_computeindex(FileSearch.SearchCriterionFactory.INDEX_EXOTIC, positionGroup, criterionfactory, positionIndex), criterionfactory)


class ThumbnailProvider(_ApiProxy):
    """
    Abstract base class for a :ref:`thumbnail<h_thumbnails>`. It can implement new ways of generating thumbnail images for files.
    Implement this class and register an instance with shallot.ThumbnailProvider.register.
    """

    #: Base index value for core (i.e. very high priority) thumbnail providers.
    INDEX_CORE = internalstuff.thumbnailprovider_index_core()
    #: Base index value for thumbnail providers with normal importance.
    INDEX_NORMAL = internalstuff.thumbnailprovider_index_normal()
    #: Base index value for exotic thumbnail providers.
    INDEX_EXOTIC = internalstuff.thumbnailprovider_index_exotic()
    #: Base index value for fallback thumbnail providers (executed after everything else failed).
    INDEX_FALLBACK = internalstuff.thumbnailprovider_index_fallback()

    def __init__(self):
        _ApiProxy.__init__(self, internalstuff.create_ThumbnailProvider())
        self._addmethod(self._native.set_getthumbnail, "get_thumbnail")

    def get_thumbnail(self, operation, node, contenttype, width, height):
        """
        Returns the thumbnail. It must be a bytestring containing the image in a well known format (png, jpeg, svg, ...).
        |dooverride|
        
        :param operation: The shallot.Operations.Operation operation object.
        :param node: The shallot.Filesystem.Node filesystem node to get a thumbnail for.
        :param contenttype: The content type of the original file.
        :param width: The requested thumbnail width in pixels.
        :param height: The requested thumbnail height in pixels.
        """
        _abstract()

    @staticmethod
    def register(thumbnailprovider, positionGroup=None, positionIndex=None):
        """
        Registers a custom shallot.ThumbnailProvider instance for offering custom thumbnails.
        
        :param thumbnailprovider: An instance of shallot.ThumbnailProvider.
        :param positionGroup: Controls execution order. See :ref:`Position Indexes<h_positionindexes>` for details. Use one of the `INDEX_*` values from shallot.ThumbnailProvider.
        :param positionIndex: Controls execution order. See :ref:`Position Indexes<h_positionindexes>` for details.
        """
        internalstuff.registerThumbnailProvider(_computeindex(ThumbnailProvider.INDEX_EXOTIC, positionGroup, thumbnailprovider, positionIndex), thumbnailprovider)


class Eurl:
    """
    A :ref:`EURL<h_eurls>`.
    
    You can get instances with global methods like shallot.Eurl.create and
    shallot.Eurl.from_string.
    
    Please note: Instances can be associated with any kind of elements in
    the filesystem (files, directories, links, ...). It might also point to something
    which does not exist at all.
    The documentation sometimes explicitly makes a difference
    between those kinds (files, directories, links, ...; often called 'node type').
    But it often uses the term 'file' implicitly while meaning all kinds of elements;
    assuming that e.g. a directory is just a special kind of a file. It should be clear from
    the particular context which meaning applies.
    """

    def __init__(self):
        """
        Can't be constructed directly.
        """
        _nonreal()

    def as_string(self):
        """
        Returns the textual value. This is what shallot.Eurl.from_string would expect as parameter.
        """
        _nonreal()

    def __str__(self):
        return self.as_string()

    def basename(self):
        """
        Returns the last path segment. This is the text behind the last slash.
        Examples:
        `"baz"` for `file:///foo/bar/baz`.
        `""` for `file:///`.
        """
        _nonreal()

    def parent_segment(self):
        """
        Returns the parent shallot.Eurl. At first, this traverses path segments. For a root path
        eurl with embeddings, it returns the embedding.
        If none are available, it returns `0`.
        Examples:
        `foo://host/foo` for `foo://host/foo/bar`.
        `foo://host/` for `foo://host/boo`.
        `foo:/[bar://host/goo]/anotherhost/` for `foo:/[bar://host/goo]/anotherhost/boo`.
        `bar://host/foo` for `foo:/[bar://host/foo]/host/`.
        `0` for `foo://host/`.
        """
        _nonreal()
        return Eurl()  # *1

    def with_appended_segment(self, segment):
        """
        Returns a new shallot.Eurl from this one with "/basename" appended.
        The parameter is assumed to be a single path segment.
        
        :param segment: The basename to be appended.
        """
        _nonreal()
        return Eurl()  # *1

    def with_appended_segments(self, path):
        """
        Returns a new shallot.Eurl with path segments "/pa/t/h/" appended.
        The parameter may contain `"/"`s for dividing path segments.
        
        :param path: The path segments to be appended, like `"foo/bar"`.
        """
        _nonreal()
        return Eurl()  # *1

    def enwrap_with_outer_url(self, scheme, hostname, path):
        """
        Returns a new shallot.Eurl containing this one packed as embedding and new outer
        parts `scheme`, `hostname` and `path`.
        Example:
        `scheme:/[file:///a/b/c.zip]/hostname/p/a/t/h` for `file:///a/b/c.zip`.
        
        :param scheme: The scheme part of the new outer eurl.
        :param hostname: The hostname of the new outer eurl.
        :param path: The path of the new outer eurl.
        """
        _nonreal()
        return Eurl()  # *1

    def path(self):
        """
        Returns the path part (from the outer url of this shallot.Eurl).
        Examples:
        `"/foo/bar/baz"` for `smb://livingroom-pc/foo/bar/baz`.
        `"/foo/bar/baz"` for `file:///foo/bar/baz`.
        `"/"` for `file:///`.
        """
        _nonreal()

    def root(self):
        """
        Returns the root shallot.Eurl from this one.
        Example:
        `zip:/[file:///a/b/c.zip]//` for `zip:/[file:///a/b/c.zip]//foo/bar/baz`.
        """
        _nonreal()
        return Eurl()  # *1

    def hostname(self):
        """
        Returns the hostname part (from the outer url of this shallot.Eurl).
        Examples:
        `"livingroom-pc"` for `smb://livingroom-pc/foo/bar/baz`.
        `""` for `file:///foo/bar/baz`.
        """
        _nonreal()

    def scheme(self):
        """
        Returns the scheme (from the outer url of this shallot.Eurl). This is what comes before the `://`.
        Example:
        `"file"` for `file:///foo/bar/baz`.
        """
        _nonreal()

    def outer_url(self):
        """
        Returns a new shallot.Eurl containing only the outer part of this one (strips the embeddings).
        Example:
        `foobar://host/foo/bar/baz` for `foobar:/[zip:/[file:///a/b/c.zip]//d/e]//foo/bar/baz`.
        """
        _nonreal()
        return Eurl()  # *1

    def outermost_inner_eurl(self):
        """
        Returns a new shallot.Eurl containing only the embedding of this one.
        Example:
        `zip:/[file:///a/b/c.zip]//d/e` for `foobar:/[zip:/[file:///a/b/c.zip]//d/e]//foo/bar/baz`.
        """
        _nonreal()
        return Eurl()  # *1

    def has_inner_urls(self):
        """
        Checks if this shallot.Eurl has embeddings.
        Examples:
        `true` for `foo:/[bar:///foo]/host/`.
        `false` for `foo://host/`.
        """
        _nonreal()

    def outer_url_is_root_directory(self):
        """
        Checks if this shallot.Eurl is a root path (with or without embeddings).
        Examples:
        `True` for `foo://host/`.
        `False` for `foo://host/a`.
        `True` for `foo:/[bar:///goo]/host/`.
        `False` for `foo:/[bar:///goo]/host/a`.
        `True` for `foo:/[bar:///goo]//`.
        `False` for `foo:/[bar:///goo]//a`.
        """
        _nonreal()

    def has_parent_segment(self):
        """
        Checks if this shallot.Eurl has a parent segment.
        This indicates if shallot.Eurl.parent_segment would return `0`.
        """
        _nonreal()

    def is_prefix_of(self, eurl):
        """
        Checks if this shallot.Eurl is a prefix of another one.
        This is not an equivalent to a string comparison but
        it checks parent relationships according to
        shallot.Eurl.parent_segment.
        
        :param eurl: The longer shallot.Eurl.
        """
        _nonreal()

    @staticmethod
    def from_string(s):
        """
        Returns a shallot.Eurl from a string.
        
        :param s: The eurl string. This is what shallot.Eurl.as_string would return.
        """
        return internalstuff.getEurlFromString(s)

    @staticmethod
    def create(scheme, hostname, path):
        """
        Returns a shallot.Eurl from three parts.
        The structure is `scheme://hostname/pa/th`.
        
        :param scheme: The scheme part.
        :param hostname: The hostname.
        :param path: The path.
        """
        return internalstuff.createEurl(scheme, hostname, path)


class Bookmarks:
    """
    Everything about :ref:`TODO<TODO>`  h_bookmarks "Bookmarks".
    """

    @staticmethod
    def get_bookmarks():
        """
        Returns a list of all existing shallot.Bookmarks.Bookmark instances.
        """
        return internalstuff.bookmarkManager().get_bookmarks()

    @staticmethod
    def has_bookmarks():
        """
        Returns True if there are any bookmarks stored.
        """
        return internalstuff.bookmarkManager().has_bookmarks()

    @staticmethod
    def add_bookmark(eurl, *, folder=[], label=None, tags=None):
        """
        Adds a new bookmark.
        
        :param eurl: The location to bookmark.
        :param folder: A string list describing in which subcollection this bookmark should be placed.
        :param label: The textual description of the new bookmark.
        :param tags: A string for internal bookkeeping. Use it for recognizing your own bookmarks later on. See shallot.Bookmarks.Bookmark.tags.
        """
        return internalstuff.bookmarkManager().add_bookmark(folder, label, eurl, tags)

    @staticmethod
    def remove_bookmark(id):
        """
        Removes a bookmark.
        
        :param id: The bookmark id this call is about (See shallot.Bookmarks.Bookmark.id).
        """
        internalstuff.bookmarkManager().remove_bookmark(id)

    @staticmethod
    def change_bookmark(id, eurl, label=None):
        """
        Changes bookmark data.
        
        :param id: The bookmark id this call is about (See shallot.Bookmarks.Bookmark.id).
        :param eurl: The new location.
        :param label: The new textual description.
        """
        internalstuff.bookmarkManager().change_bookmark(id, eurl, label)

    @staticmethod
    def change_bookmark_tags(id, tags):
        """
        Changes bookmark tags. See shallot.Bookmarks.Bookmark.tags.
        
        :param id: The bookmark id this call is about (See shallot.Bookmarks.Bookmark.id).
        :param tags: The new tags string.s
        """
        internalstuff.bookmarkManager().change_bookmark_tags(id, tags)

    @staticmethod
    def move_bookmark_up(id):
        """
        Moves a bookmark upwards in its collection.
        
        :param id: The bookmark id this call is about (See shallot.Bookmarks.Bookmark.id).
        """
        internalstuff.bookmarkManager().move_bookmark_up(id)

    @staticmethod
    def move_bookmark_down(id):
        """
        Moves a bookmark downwards in its collection.
        
        :param id: The bookmark id this call is about (See shallot.Bookmarks.Bookmark.id).
        """
        internalstuff.bookmarkManager().move_bookmark_down(id)

    @staticmethod
    def move_bookmark_to_folder(id, folder):
        """
        Moves a bookmark to another subcollection.
        
        :param id: The bookmark id this call is about (See shallot.Bookmarks.Bookmark.id).
        :param folder: The new subcollection (as list of strings).
        """
        internalstuff.bookmarkManager().move_bookmark_to_folder(id, folder)

    class Bookmark:
        """
        A bookmark. Change it with the methods in shallot.Bookmarks.
        """

        def __init__(self):
            """
            Can't be constructed directly.
            """
            _nonreal()

        def eurl(self):
            """
            The location this bookmark points to.
            """
            _nonreal()

        def folder(self):
            """
            The subcollection (as list of strings).
            """
            _nonreal()

        def id(self):
            """
            The bookmark id. It is not display anywhere but only used for the
            management methods in shallot.Bookmarks.
            """
            _nonreal()

        def label(self):
            """
            A textual description of this bookmark.
            """
            _nonreal()

        def tags(self):
            """
            The tags strings. This value is solely used by plugins for their
            bookkeeping. It can allow to programmatically find a particular
            bookmark (if the creator wrote meaningful information in it).
            """
            _nonreal()
