/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2020 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <QSettings>
#include "base/shallotprocess.h"
#include "base/singletoninitializer.h"
#include "configuration/configurationmanager.h"
#include "tools/userdirlock.h"
#include "accountsmanager.h"
#include "fallbackaccountsprovider.h"

REGISTER_STATICINIT(sh::tools::accounts, FallbackAccountsProvider, "base")


sh::tools::accounts::FallbackAccountsProvider::FallbackAccountsProvider()
{
    accountsdir = sh::base::ShallotProcess::userDataDir() + "accounts/";
}

void sh::tools::accounts::FallbackAccountsProvider::doInitialize()
{
    AccountsManager::instance()->registerProvider(std::make_shared<FallbackAccountsProvider>());
}

void sh::tools::accounts::FallbackAccountsProvider::doShutdown()
{
}

QList<std::shared_ptr<sh::tools::accounts::Account>> sh::tools::accounts::FallbackAccountsProvider::_find_rem_account_helper(std::shared_ptr<Account> pattern, bool remove)
{
    QList<std::shared_ptr<sh::tools::accounts::Account>> result;
    foreach(QString e, QDir(accountsdir).entryList(QDir::Files))
    {
        auto taccount = std::make_shared<sh::tools::accounts::Account>();
        QSettings facts(accountsdir+e, QSettings::IniFormat);
        taccount->domain = facts.value("domain").toString();
        taccount->path = facts.value("path").toString();
        taccount->protocol = facts.value("protocol").toString();
        taccount->server = facts.value("server").toString();
        taccount->serverport = facts.value("serverport").toInt();
        taccount->username = facts.value("username").toString();
        if (!pattern->domain.isEmpty() && pattern->domain.toLower()!=taccount->domain.toLower())
            continue;
        if (!pattern->path.isEmpty() && pattern->path!=taccount->path)
            continue;
        if (!pattern->protocol.isNull() && pattern->protocol.toLower()!=taccount->protocol.toLower())
            continue;
        if (!pattern->server.isNull() && pattern->server.toLower()!=taccount->server.toLower())
            continue;
        if (pattern->serverport>0 && pattern->serverport!=taccount->serverport)
            continue;
        if (!pattern->username.isNull() && pattern->username!=taccount->username)
            continue;
        if (remove)
            QFile(accountsdir+e).remove();
        else
            result.append(taccount);
    }
    return result;
}

void sh::tools::accounts::FallbackAccountsProvider::findAccounts(std::shared_ptr<Account> pattern, QList<std::shared_ptr<Account>>& result)
{
    foreach (auto x, _find_rem_account_helper(pattern, false))
        result.append(x);
}

bool sh::tools::accounts::FallbackAccountsProvider::storeAccount(std::shared_ptr<Account> account)
{
    _find_rem_account_helper(account, true);
    {
        sh::tools::UserDirLock userdirlock;
        QString fname;
        while (fname.isEmpty() || QFile(fname).exists())
            fname = accountsdir + QString::number(qrand() % 10000000);
        QSettings facts(fname, QSettings::IniFormat);
        facts.setValue("domain", account->domain);
        facts.setValue("path", account->path);
        facts.setValue("protocol", account->protocol);
        facts.setValue("server", account->server);
        facts.setValue("serverport", account->serverport);
        facts.setValue("username", account->username);
    }
    return true;
}
