/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2020 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef ATOMICCOUNTER_H
#define ATOMICCOUNTER_H

#include <memory>
#include <QMutex>


namespace sh{
    namespace tools{

        /** \brief
         * Internal tool for reference counting.
         */
        class AtomicCounter
        {
        public:
            AtomicCounter() {}

            inline int value()
            {
                QMutexLocker l1(&_mutex);
                return _value;
            }

            class Increment
            {
                friend class AtomicCounter;
            public:
                Increment(AtomicCounter* ac)
                {
                    _ac = ac;
                    _ac->doIncValue();
                }

                ~Increment()
                {
                    _ac->doDecValue();
                }

            private:
                AtomicCounter* _ac;
            };

            inline std::shared_ptr<AtomicCounter::Increment> increment()
            {
                return std::shared_ptr<Increment>(new Increment(this));
            }

            inline int doIncValue()
            {
                QMutexLocker l1(&_mutex);
                return ++_value;
            }

            inline int doDecValue()
            {
                QMutexLocker l1(&_mutex);
                return --_value;
            }

        private:
            int _value = 0;
            QMutex _mutex;
        };

    }
}

#endif // ATOMICCOUNTER_H
