/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2020 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef BOOKMARKMANAGER_H
#define BOOKMARKMANAGER_H

#include <functional>
#include <memory>
#include <QMutex>
#include <QObject>
#include <QUrl>
#include "base/singletoninitializer.h"


namespace sh {
    namespace filesystem { class Eurl; }
    namespace configuration {class ConfigurationValue;}
    namespace tools {

        /**
         * \brief A bookmark
         */
        class Bookmark
        {
        public:
            /** \brief Constructed only by the infrastructure and made available otherwise.
             */
            Bookmark(QString id, QStringList folder, QString label, std::shared_ptr<const sh::filesystem::Eurl> eurl, QString tags);
            /**
             * \brief The bookmark id (used in some methods of sh::tools::BookmarkManager).
             */
            QString id();
            /**
             * \brief The bookmark folder (the user interface calls it 'Collections').
             */
            QList<QString> folder();
            /**
             * \brief The bookmark label.
             */
            QString label();
            /**
             * \brief The location this bookmark points to.
             */
            std::shared_ptr<const sh::filesystem::Eurl> eurl();
            /**
             * \brief Internal information for bookkeeping.
             *
             * Can be used by external code for keeping track of dynamically created bookmarks.
             */
            QString tags();
        private:
            QString _id;
            QStringList _folder;
            QString _label;
            std::shared_ptr<const sh::filesystem::Eurl> _eurl;
            QString _tags;
        };

        /**
         * \brief The bookmark manager.
         *
         * Use it for getting or modifying bookmarks.
         */
        class BookmarkManager : public QObject, public sh::base::Singleton
        {
            Q_OBJECT
            DECLARE_SINGLETON(BookmarkManager)
        public:

            /**
             * \brief Returns a list of all stored bookmarks.
             */
            QList<std::shared_ptr<sh::tools::Bookmark>> getBookmarks();

            /**
             * \brief Checks if any bookmarks exist.
             */
            bool hasBookmarks();

            /**
             * \brief Adds a new bookmark.
             * @returns The id of the new bookmark.
             */
            QString addBookmark(QList<QString> folder, QString label, std::shared_ptr<const sh::filesystem::Eurl> eurl, QString tags=QString());

            /**
             * \brief Removes a bookmarks.
             */
            void removeBookmark(QString id);

            /**
             * \brief Changes data of a bookmark.
             */
            void changeBookmark(QString id, QString label, std::shared_ptr<const sh::filesystem::Eurl> eurl);

            /**
             * \brief Changes the `tags` of a bookmark.
             */
            void changeBookmarkTags(QString id, QString tags);

            /**
             * \brief Moves a bookmark up within its folder.
             */
            void moveBookmarkUp(QString id);

            /**
             * \brief Moves a bookmark down within its folder.
             */
            void moveBookmarkDown(QString id);

            /**
             * \brief Changes to `folder` of a bookmark.
             */
            void moveBookmarkToFolder(QString id, QList<QString> folder);

        signals:
            void changed();

        private:
            BookmarkManager();
            void moveBookmark(QString id, int direction);
            void _changeBookmark(QString id, std::function<QString(std::shared_ptr<Bookmark>)> label,
                                 std::function<std::shared_ptr<const sh::filesystem::Eurl>(std::shared_ptr<Bookmark>)> eurl,
                                 std::function<QList<QString>(std::shared_ptr<Bookmark>)> folder,
                                 std::function<QString(std::shared_ptr<Bookmark>)> tags);
            std::shared_ptr<sh::configuration::ConfigurationValue> cfgvalBookmarks;
            void readBookmarks();
            void writeBookmarks();
            QList<std::shared_ptr<Bookmark>> _bookmarks;
            QMutex _bookmarksmutex;
        };

    }
}

#endif // BOOKMARKMANAGER_H
