/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2020 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef DATAEXCHANGE_H
#define DATAEXCHANGE_H

#include <memory>
#include <QObject>
#include <QMimeData>
#include <QMutex>
#include "base/singletoninitializer.h"


namespace sh{
    namespace filesystem{ class Eurl; class FilesystemNode; }
    namespace actions{ class AbstractActionItem; }

    namespace tools{

        /** \brief
         * Clipboard and DnD related tools.
         */
        class DataExchange : public QObject, public sh::base::Singleton
        {
            Q_OBJECT
            DECLARE_SINGLETON(DataExchange)
        public:

            /** \brief A kind of data exchange movement.
             */
            enum class DataExchangeType{
                COPY,
                MOVE
            };

            static QString FilelistTypeShallot;
            static QString FilelistTypeGnome;
            static QString FilelistTypeUrilist;
            static QString FilelistTypePlaintext;

            /** @brief Checks if a QMimeData contains any file entries we understand.
             */
            bool containsFileEntries(const QMimeData *);

            /** @brief Returns a new QMimeData for a list of filesystem nodes and an exchange type.
             */
            QMimeData *getMimeDataFromFilesystemNodes(QList<std::shared_ptr<sh::filesystem::FilesystemNode>>, DataExchangeType type);

            /** @brief Returns a copy action for a source by QMimeData and a destination eurl.
             */
            std::shared_ptr<sh::actions::AbstractActionItem> createCopyAction(const QMimeData *src, std::shared_ptr<const sh::filesystem::Eurl> dest);

            /** @brief Returns a move action for a source by QMimeData and a destination eurl.
             */
            std::shared_ptr<sh::actions::AbstractActionItem> createMoveAction(const QMimeData *src, std::shared_ptr<const sh::filesystem::Eurl> dest);

            /** @brief Returns a paste action for a source by QMimeData and a destination eurl (copy or move by QMimeData).
             */
            std::shared_ptr<sh::actions::AbstractActionItem> createPasteAction(const QMimeData *src, std::shared_ptr<const sh::filesystem::Eurl> dest);

            /** @brief Returns a list of eurls (and if it is a cut/move exchange) from the QMimeData.
             */
            QList<std::shared_ptr<const sh::filesystem::Eurl>> getSources(const QMimeData *src, bool *isCut = 0);

        private:
            static QString linebreak;
            explicit DataExchange();
        };

    }
}

#endif // DATAEXCHANGE_H
