/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2020 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef DATASTREAM_H
#define DATASTREAM_H

#include <QByteArray>
#include <QFile>
#include <QObject>


namespace sh{
    namespace tools{

        /** \brief
         * A QFile (and a QIODevice) for local file access.
         *
         * It also stores the information whether this is a temporary location or the permanent one.
         */
        class LocalFile : public QFile
        {
        public:
            LocalFile(QString& file, bool isTemp=false);
            bool isTemp();
        private:
            bool _istemp;
        };

        /** \brief
         * A QIODevice implementation, acting as an abstract base class for other subclasses.
         *
         * The provided interface is not as generic but easier than the original one for many
         * easier tasks.
         */
        class ReadDataDevice : public QIODevice
        {
        public:
            ReadDataDevice();
            /** \brief
             * This method returns the next available chunk of data. It may return empty arrays whenever
             * temporarily no data is available. Returning a null array (with `QByteArray::isNull()==true`)
             * marks the end of the stream.
             * @a_dooverride
             */
            virtual QByteArray getdata() = 0;

        protected:
            virtual qint64 readData(char* data, qint64 maxSize);
            virtual qint64 writeData(const char* data, qint64 maxSize);
            virtual bool isSequential() const { return true; }
            virtual bool isOpen() const { return true; }

        private:
            QByteArray current;
            int currentlen = 0;
            int currentconsumed = 0;
        };

    }
}

#endif // DATASTREAM_H
