/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2020 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef FILETYPEMANAGER_H
#define FILETYPEMANAGER_H

#include <memory>
#include <QIcon>
#include <QList>
#include <QMutex>
#include <QObject>
#include <QStringList>
#include "base/singletoninitializer.h"


namespace sh{
    namespace filesystem{ class Eurl; class FilesystemNode; class Operation; }
    namespace tools {
        namespace filetypes {

            /** \brief Commandline and infos for opening a file with an external program.
             */
            struct OpenMethod
            {
                OpenMethod(QString name, QString command, QStringList arguments, QIcon icon = QIcon(), int precedence = 0);
                const QString name;
                const QString command;
                const QStringList arguments;
                const QIcon icon;
                const int precedence;
            };

            /** \brief Utilities for dealing with file types.
             *
             * It can determine the type of a file (png, plaintext, html, ...), it can provide information about
             * how to open them with an external program, and more.
             *
             * For most tasks it uses a pluggable interface. Actual implementations of strategies for
             * those tasks reside in separate classes in this namespace.
             */
            class FileTypeManager : public QObject, public sh::base::Singleton
            {
                Q_OBJECT
                DECLARE_SINGLETON(FileTypeManager)
            public:

                /** \brief Abstract base class for a mimetype determination strategy.
                 */
                class MimetypeDeterminationStrategy
                {
                public:
                    virtual QHash<std::shared_ptr<const sh::filesystem::Eurl>, QString> determineMimetype(sh::filesystem::Operation* op, QList<std::shared_ptr<const sh::filesystem::Eurl>> items) = 0;
                    virtual ~MimetypeDeterminationStrategy();
                };

                /** \brief Abstract base class for a 'open method' determination strategy.
                 */
                class OpenMethodDeterminationStrategy
                {
                public:
                    virtual QList<std::shared_ptr<OpenMethod> > getOpenMethods(QString mimetype, QList<std::shared_ptr<sh::filesystem::FilesystemNode>> nodes) = 0;
                    virtual ~OpenMethodDeterminationStrategy();
                };

                /** \brief Abstract base class for mimetype information retrieval strategy.
                 */
                class MimetypeInformationRetrievalStrategy
                {
                public:
                    virtual QStringList getMimetypeSubclasses(QString mimetype);
                    virtual ~MimetypeInformationRetrievalStrategy();
                };

                /** \brief Determines the mimetype for one file.
                 */
                QString determineMimetype(sh::filesystem::Operation* op, std::shared_ptr<const sh::filesystem::Eurl> eurl);

                /** \brief Determines the mimetypes for a list of files.
                 */
                QHash<std::shared_ptr<const sh::filesystem::Eurl>, QString> determineMimetype(sh::filesystem::Operation* op, QList<std::shared_ptr<const sh::filesystem::Eurl>> items);

                /** \brief Determines how to open a file with a given mimetype with an external program.
                 */
                QList<std::shared_ptr<OpenMethod>> getOpenMethods(QString mimetype, QList<std::shared_ptr<sh::filesystem::FilesystemNode>> nodes);

                /** \brief Returns a list of all known infos how to open a file external programs.
                 *
                 * This has roughly one entry for each installed program on the user's system, which can graphically open files.
                 */
                QList<std::shared_ptr<OpenMethod>> getAllOpenMethods();

            private:
                explicit FileTypeManager(QObject *parent = 0);
                QList<std::shared_ptr<MimetypeDeterminationStrategy>> _mimetypeDeterminationMethods;
                QList<std::shared_ptr<OpenMethodDeterminationStrategy>> _openMethodDeterminationMethods;
                QList<std::shared_ptr<MimetypeInformationRetrievalStrategy>> _mimetypeInformationRetrievalMethods;
                QMutex _mutex;

            };

        }
    }
}

#endif // FILETYPEMANAGER_H
