/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2020 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "filesystem/filesystemnode.h"
#include "filesystem/eurl.h"
#include "filesystem/operation.h"
#include "filesystem/filesystemhandlerregister.h"
#include "configuration/configurationmanager.h"
#include "filesystemhandlers/localfilesystemhandler.h"
#include "freedesktoporgtoolsmimetypedeterminationmethod.h"


std::shared_ptr<sh::configuration::ConfigurationValue> sh::tools::filetypes::FreedesktopOrgToolsMimetypeDeterminationStrategy::cfgvalXdgmimePath =
        sh::configuration::ConfigurationManager::instance()->registerConfigValue(
            "sh::tools::filetypes::FreedesktopOrgToolsMimetypeDeterminationStrategy.XdgmimePath", "/usr/bin/xdg-mime",
            sh::configuration::ConfigurationValue::valueTypeLocalFilePath(),
            QObject::tr("Path to the 'xdg-mime' tool"),
            sh::configuration::ConfigurationCategory::CategoryExternalTools,
            QObject::tr("This tool is possibly used for file type detection."));

sh::tools::filetypes::FreedesktopOrgToolsMimetypeDeterminationStrategy::FreedesktopOrgToolsMimetypeDeterminationStrategy(sh::tools::filetypes::FileTypeManager* manager)
    : _reMimetype( QRegularExpression("^[^ /]+/[^ /]+$") )
{
    cfgvalXdgmimePath->consumeValue(manager, [=](){
        QMutexLocker _l1(&_mutex);
        _pathToFileTool = cfgvalXdgmimePath->getConfigValueString();
    });
}

QHash<std::shared_ptr<const sh::filesystem::Eurl>, QString> sh::tools::filetypes::FreedesktopOrgToolsMimetypeDeterminationStrategy::determineMimetype(sh::filesystem::Operation*, QList<std::shared_ptr<const sh::filesystem::Eurl>> items)
{
    QMutexLocker _l1(&_mutex);
    QHash<std::shared_ptr<const sh::filesystem::Eurl>, QString> result;
    auto reMimetype = _reMimetype;  // thread-local copy needed
    if (!_pathToFileTool.isEmpty())
    {
        foreach(std::shared_ptr<const sh::filesystem::Eurl> eurl, items) {
            sh::filesystemhandlers::LocalFilesystemHandler* ehandler = qobject_cast<sh::filesystemhandlers::LocalFilesystemHandler*>(sh::filesystem::FilesystemHandlerRegister::instance()->findHandler(eurl->scheme()).get());
            if (ehandler)
            {
                QString localpath = ehandler->eurlToLocal(eurl);
                if (!localpath.isEmpty()){
                    QProcess fileprocess;
                    fileprocess.setProgram(_pathToFileTool);
                    QStringList args;
                    args << "query" << "filetype" << localpath;
                    fileprocess.setArguments(args);
                    fileprocess.start();
                    if (fileprocess.waitForFinished())
                    {
                        QByteArray ba = fileprocess.readAll();
                        QString t = QString::fromUtf8(ba).trimmed();
                        if (reMimetype.match(t).hasMatch())
                            result[eurl] = t;
                    }
                }
            }
        }
    }
    return result;
}
