/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2020 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "filesystem/filesystemnode.h"
#include "filesystem/eurl.h"
#include <QDir>
#include <QFile>
#include <QRegularExpression>
#include "exceptions/ioexception.h"
#include "base/shallotprocess.h"
#include "suffixlistmimetypedeterminationmethod.h"


sh::tools::filetypes::SuffixListMimetypeDeterminationStrategy::SuffixListMimetypeDeterminationStrategy()
{
    QString shallotpath = sh::base::ShallotProcess::shallotDataDir();
    foreach(QString mimefile, QStringList({"mime.2.types", "mime.types"})) {
        QFile f(shallotpath + QDir::separator() + mimefile);
        sh::exceptions::Exception::executeGuarded([&](){
            if (!f.open(QIODevice::ReadOnly))
                throw sh::exceptions::IOException(sh::exceptions::ExceptionData().detailsAreInteresting().details(QObject::tr("Unable to open mime types file")).resumeable(false));
        }, sh::exceptions::ExecuteGuardFlag::ALL_ERRORS_KILL_SHALLOT);
        QByteArray b = f.readAll();
        sh::exceptions::Exception::executeGuarded([&](){
            if (f.error() != QFile::NoError)
                throw sh::exceptions::IOException(sh::exceptions::ExceptionData().detailsAreInteresting().details(QObject::tr("Unable to read mime types file")).resumeable(false));
        }, sh::exceptions::ExecuteGuardFlag::ALL_ERRORS_KILL_SHALLOT);
        QString m = QString::fromUtf8(b);
        foreach(QString _line, m.split("\n"))
        {
            QString line = _line.trimmed();
            if (!line.isEmpty() && !line.startsWith("#"))
            {
                QStringList ltoks = line.replace("\t", " ").replace(QRegularExpression("( )+"), " ").split(" ");
                for(int i=1; i<ltoks.length(); i++) {
                    QString ltok = ltoks[i].toLower();
                    if (!_suffixToMimetype.contains(ltok))
                        _suffixToMimetype[ltok] = ltoks[0];
                }
            }
        }
    }
}

QHash<std::shared_ptr<const sh::filesystem::Eurl>, QString> sh::tools::filetypes::SuffixListMimetypeDeterminationStrategy::determineMimetype(sh::filesystem::Operation*, QList<std::shared_ptr<const sh::filesystem::Eurl>> items)
{
    QMutexLocker _l1(&_mutex);
    QHash<std::shared_ptr<const sh::filesystem::Eurl>, QString> result;
    foreach (std::shared_ptr<const sh::filesystem::Eurl> eurl, items) {
        QString eurlstring = eurl->asString();
        int idx = eurlstring.lastIndexOf(".");
        if (idx>=0) {
            QString suffix = eurlstring.mid(idx+1).toLower();
            QString mimetype = _suffixToMimetype.value(suffix, QString());
            if (!mimetype.isNull())
                result[eurl] = mimetype;
        }
    }
    return result;
}
