/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2020 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "base/shallotprocess.h"
#include "configuration/configurationmanager.h"
#include "filesystem/filesystemnode.h"
#include "tools/userdirlock.h"
#include "userdefinedopenmethoddeterminationstrategy.h"

REGISTER_SINGLETON(sh::tools::filetypes, UserDefinedOpenMethodDeterminationStrategy, "base")


sh::tools::filetypes::UserDefinedOpenMethodDeterminationStrategy::UserDefinedOpenMethodDeterminationStrategy()
                        : sh::tools::filetypes::FileTypeManager::OpenMethodDeterminationStrategy()
{
}

QList<std::shared_ptr<sh::tools::filetypes::OpenMethod> > sh::tools::filetypes::UserDefinedOpenMethodDeterminationStrategy::getOpenMethods(QString mimetype, QList<std::shared_ptr<sh::filesystem::FilesystemNode>> nodes)
{
    QMutexLocker _l1(&_mutex);
    QList<std::shared_ptr<sh::tools::filetypes::OpenMethod>> result;
    sh::tools::UserDirLock userdirlock;
    QString cdir = sh::base::ShallotProcess::userDataDir() + "/openmethods/";
    QDir dir(cdir);
    foreach (QString x, dir.entryList()) {
        QSettings f(cdir + x, QSettings::IniFormat);
        bool matched = false;
        int iNode = 0;
        QString forMimetype = f.value("rememberForMimetype").toString();
        bool rememberForMimetype = !forMimetype.isEmpty();
        std::shared_ptr<const sh::filesystem::Eurl> rememberForDirectory;
        QString _srememberfordir = f.value("rememberForDirectory").toString();
        if (!_srememberfordir.isEmpty())
            rememberForDirectory = sh::filesystem::Eurl::fromString(_srememberfordir);
        bool rememberForFile = f.value("rememberForFile") == "1";
        while (true) {
            QVariant v = f.value(QString("node%1").arg(iNode));
            if (v.isNull())
                break;
            std::shared_ptr<const sh::filesystem::Eurl> e = sh::filesystem::Eurl::fromString(v.toString());
            if (rememberForMimetype && mimetype == forMimetype) {
                matched = true;
                break;
            }
            foreach (std::shared_ptr<sh::filesystem::FilesystemNode> node, nodes) {
                if (rememberForFile && e == node->eurl()) {
                    matched = true;
                    break;
                }
                if (rememberForDirectory) {
                    std::shared_ptr<sh::filesystem::FilesystemNode> pnode = node->parentnode();
                    while (pnode && pnode->eurl() != rememberForDirectory)
                        pnode = pnode->parentnode();
                    if (pnode) {
                        matched = true;
                        break;
                    }
                }
            }
            if (matched)
                break;
            iNode++;
        }
        if (matched) {
            QString name = f.value("name").toString();
            QString command = f.value("command").toString();
            QStringList arguments;
            int iArgument = 0;
            while(true) {
                QVariant v = f.value(QString("argument%1").arg(iArgument));
                if (v.isNull())
                    break;
                arguments << v.toString();
                iArgument++;
            }
            result.append(std::shared_ptr<sh::tools::filetypes::OpenMethod>(new sh::tools::filetypes::OpenMethod(name, command, arguments)));
        }
    }
    return result;
}

void sh::tools::filetypes::UserDefinedOpenMethodDeterminationStrategy::storeCustomOpenMethod(QList<std::shared_ptr<sh::filesystem::FilesystemNode>> nodes, QString mimetype, std::shared_ptr<sh::tools::filetypes::OpenMethod> m, bool rememberForMimetype, std::shared_ptr<const sh::filesystem::Eurl> rememberForDirectory, bool rememberForFile)
{
    QMutexLocker _l1(&_mutex);
    if (rememberForMimetype || rememberForDirectory || rememberForFile) {
        sh::tools::UserDirLock userdirlock;
        QUuid uuid = QUuid::createUuid();
        QString cdir = sh::base::ShallotProcess::userDataDir() + "/openmethods/";
        QString cfile = cdir + uuid.toString();
        if (!QDir(cdir).exists())
            QDir(sh::base::ShallotProcess::userDataDir()).mkdir("openmethods");
        QSettings f(cfile, QSettings::IniFormat);
        f.setValue("name", m->name);
        f.setValue("command", m->command);
        for (int i=0; i< m->arguments.length(); i++)
            f.setValue(QString("argument%1").arg(i), m->arguments[i]);
        f.setValue("rememberForMimetype", rememberForMimetype?mimetype:"");
        f.setValue("rememberForDirectory", rememberForDirectory?rememberForDirectory->asString():"");
        f.setValue("rememberForFile", rememberForFile?"1":"0");
        for (int i=0; i< nodes.length(); i++)
            f.setValue(QString("node%1").arg(i), nodes[i]->eurl()->asString());
    }
}
