/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2020 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#define _CRT_RAND_S // needed for stupid win32
#include "base/logger.h"
#include "base/mainthread.h"
#include "exceptions/runtimeexception.h"
#include "misc.h"

#if QT_VERSION >= QT_VERSION_CHECK(5, 10, 0)
#include <QRandomGenerator>
#elif defined(Q_OS_WIN)
#define _WIN32_WINNT 0x0601
#include <windows.h>
#include <stdlib.h>
#endif


sh::tools::Misc::Misc()
{
}

QByteArray sh::tools::Misc::iconToPngByteArray(QIcon icon, int sizeInPt)/* TODO # fixed size */
{
    QByteArray bArray;
    sh::base::MainThread::dispatcher()->invokeSync([&](){
        QPixmap icn = icon.pixmap(sizeInPt*1.5, sizeInPt*1.5);
        QBuffer buffer(&bArray);
        buffer.open(QIODevice::WriteOnly);
        icn.save(&buffer, "PNG");
    });
    return bArray;
}

QByteArray sh::tools::Misc::iconToBase64SrcEncoding(QIcon icon, int sizeInPt)
{
    if (icon.isNull())
        return QByteArray();
    static QByteArray pref("data:image/png;base64,");
    return pref + iconToPngByteArray(icon, sizeInPt).toBase64();
}

QByteArray sh::tools::Misc::hash(QStringList data)
{
    return _hash(data, true, QByteArray());
}

QByteArray sh::tools::Misc::hashUnsalted(QStringList data)
{
    return _hash(data, false, QByteArray());
}

QByteArray sh::tools::Misc::_hash(QStringList data, bool salted, QByteArray _withSalt)
{
    QCryptographicHash hashr(QCryptographicHash::Sha3_256);
    const int SALT_LEN = 256; // matching to the hash length
    foreach (QString d, data)
        hashr.addData(d.toLatin1());
    QByteArray salt;
    if (salted || !_withSalt.isEmpty()) {
        salt = _withSalt.isEmpty() ? randomBytes(SALT_LEN / 8) : _withSalt;
        hashr.addData(salt);
    }
    QByteArray result = hashr.result().toBase64();
    if (salted)
        result += "." + salt;
    return result;
}

bool sh::tools::Misc::compareHash(QByteArray bhash, QStringList data)
{
    QList<QByteArray> parts = bhash.split('.');
    if (parts.length() == 2) // salted
        return _hash(data, false, parts[1]) == parts[0];
    else //plain
        return _hash(data, false, QByteArray()) == bhash;
}

QByteArray sh::tools::Misc::qjsonToJson(QJsonObject o)
{
    QJsonDocument jd;
    jd.setObject(o);
    return jd.toJson();
}

QByteArray sh::tools::Misc::qjsonToJson(QJsonValue o)
{
    QJsonArray ao;
    ao.append(o);
    QByteArray result = qjsonToJson(ao).trimmed();
    int i1 = result.indexOf("[");
    int i2 = result.lastIndexOf("]");
    return result.mid(i1+1, i2-i1-1);
}

QByteArray sh::tools::Misc::qmapToJson(QVariantMap m)
{
    return qjsonToJson(QJsonObject::fromVariantMap(m));
}

QByteArray sh::tools::Misc::qjsonToJson(QJsonArray a)
{
    QJsonDocument jd;
    jd.setArray(a);
    return jd.toJson();
}

QByteArray sh::tools::Misc::jsonableToJson(Jsonable* a)
{
    if (a)
        return qjsonToJson(a->toJson());
    else
        return _jundefined;
}

QByteArray sh::tools::Misc::randomBytes(int length)
{
    QByteArray result(length, 0);
#if QT_VERSION >= QT_VERSION_CHECK(5, 10, 0)
    auto rnd = QRandomGenerator::system();
    for (int i=0; i<length; i++)
        result[i] = rnd->generate();
#elif defined(Q_OS_WIN)
    // my current env for windows building doesnt support QByteArray ^^
    unsigned int r;
    for (int i=0; i<length; i++) {
        if (rand_s(&r))
            throw sh::exceptions::RuntimeException(sh::exceptions::ExceptionData().details("RtlGenRandom returned false").resumeable(false));
        result[i] = r;
    }
#else
    SH_LOG_WARNING("Secure randomization is not supported on your platform. Falling back to pseudo random numbers.")
    for (int i=0; i<length; i++)
        result[i] = rand();
#endif
    return result;
}

QByteArray sh::tools::Misc::generateUniqueHash()
{
    QStringList l;
    l << randomBytes();
    l << QString("%1").arg(QDateTime::currentMSecsSinceEpoch());
    l << QHostInfo::localHostName() << QHostInfo::localDomainName();
    return hashUnsalted(l);
}

void sh::tools::Misc::makeHttpRequest(QUrl url, QByteArray* responseBody, QString* mimeType, int* httpStatus)
{
    QMutex m;
    QWaitCondition wc;
    bool finished = false;
    auto fctfinish = [&](QByteArray _responseBody, QString _mimeType, int _httpStatus){
        if (httpStatus)
            *httpStatus = _httpStatus;
        if (responseBody)
            *responseBody = _responseBody;
        if (mimeType)
            *mimeType = _mimeType;
        {
            QMutexLocker _l1(&m);
            finished = true;
            wc.wakeAll();
        }
    };
    QNetworkReply* reply = nullptr;
    auto deletereply = [&](){
        ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
        reply->deleteLater();
        reply = nullptr;
    };
    sh::base::MainThread::dispatcher()->invokeSync([&](){
        QNetworkRequest request;
        request.setUrl(url);
        reply = _qnetwork.get(request);
        if (reply->isRunning()) {
            QObject::connect(reply, QOverload<QNetworkReply::NetworkError>::of(&QNetworkReply::error), reply, [=](QNetworkReply::NetworkError) {
                // emitted when the reply detects an error in processing. The finished() signal will _probably_ follow
                // -> so we delete it, but deferred (so 'finished' can maybe grab details)
                QTimer* t = new QTimer(reply);
                QObject::connect(t, &QTimer::timeout, reply, [=](){
                    fctfinish("", "text/html", 0);
                    deletereply();
                    t->deleteLater();
                });
                t->setSingleShot(true);
                t->start(1000);
            });
            QObject::connect(reply, &QNetworkReply::finished, reply, [&,reply]() {
                fctfinish(reply->readAll(), reply->header(QNetworkRequest::ContentTypeHeader).toString(), reply->attribute(QNetworkRequest::HttpStatusCodeAttribute).toInt());
                deletereply();
            });
        }
        else {
            fctfinish("", "text/html", 0);
            reply->deleteLater();
        }
    });
    {
        qint64 beg = QDateTime::currentSecsSinceEpoch();
        QMutexLocker _l1(&m);
        while (!finished) {
            if (sh::base::MainThread::dispatcher()->inThread()) {
                m.unlock();
                QCoreApplication::processEvents();
                m.lock();
            }
            else
                wc.wait(&m, 2000);
            if (QDateTime::currentSecsSinceEpoch() - beg > 60 * 60 * 3) {
                sh::base::MainThread::dispatcher()->invokeSync([&](){
                    fctfinish("", "text/html", 0);
                    if (reply) {
                        reply->abort();
                        deletereply();
                    }
                });
            }
        }
    }
}

QByteArray sh::tools::Misc::_jundefined("undefined");

QNetworkAccessManager sh::tools::Misc::_qnetwork;
