/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2020 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef MISC_H
#define MISC_H

#include <QBuffer>
#include <QByteArray>
#include <QCryptographicHash>
#include <QDateTime>
#include <QHostInfo>
#include <QIcon>
#include <QJsonArray>
#include <QJsonDocument>
#include <QJsonObject>
#include <QNetworkAccessManager>
#include <QNetworkReply>
#include <QTimer>
#include <QWaitCondition>


namespace sh{
    namespace tools{

        /**
         * \brief An interface for objects which can return a json representation as QJsonObject.
         */
        class Jsonable
        {
        public:

            /**
             * \brief Returns the json respesentation as QJsonObject.
             */
            virtual QJsonObject toJson() =0;

        };

        class Misc
        {
        public:

            /** \brief Computes a png representation for an icon.
             */
            static QByteArray iconToPngByteArray(QIcon icon, int sizeInPt);

            /** \brief Computes a html-compatible 'data:image/png;base64,...' png representation for an icon.
             */
            static QByteArray iconToBase64SrcEncoding(QIcon icon, int sizeInPt);

            /** \brief Hashes string data in a cryptographical way (with a salt).
             *
             * Use with compareHash().
             */
            static QByteArray hash(QStringList data);

            /** \brief Hashes string data in a cryptographical way without a salt.
             *
             * Use with compareHash() or just compare strings.
             */
            static QByteArray hashUnsalted(QStringList data);

            /** \brief Check if a hash is matching to given string data.
             */
            static bool compareHash(QByteArray hash, QStringList data);

            /** \brief Returns a json string for a QJsonObject, QJsonValue or QJsonArray.
             */
            static QByteArray qjsonToJson(QJsonObject o);
            static QByteArray qjsonToJson(QJsonValue o);
            static QByteArray qjsonToJson(QJsonArray a);

            /** \brief Returns a json string for a QVariantMap.
             */
            static QByteArray qmapToJson(QVariantMap m);

            /** \brief Returns a json string from a Jsonable.
             */
            static QByteArray jsonableToJson(Jsonable* a);

            /** \brief Returns a random byte array.
             */
            static QByteArray randomBytes(int length=32);

            /** \brief Returns a (mostly) random generated unique hash.
             */
            static QByteArray generateUniqueHash();

            /** \brief Makes an http request (as a client).
             *
             * `httpStatus` can return `0` for network errors and similar.
             */
            static void makeHttpRequest(QUrl url, QByteArray *responseBody=nullptr, QString* mimeType=nullptr, int* httpStatus=nullptr);

        private:
            static QByteArray _hash(QStringList data, bool salted, QByteArray _withSalt);
            static QByteArray _jundefined;
            static QNetworkAccessManager _qnetwork;
            Misc();
        };

    }
}

#endif // MISC_H
