/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2020 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef THUMBNAILMANAGER_H
#define THUMBNAILMANAGER_H

#include <functional>
#include <memory>
#include <QDateTime>
#include <QIcon>
#include <QHash>
#include <QMutex>
#include <QObject>
#include "base/singletoninitializer.h"


namespace sh{
    namespace filesystem{class FilesystemNode;class Operation;}

    namespace tools {

        class ThumbnailProvider
        {
        public:
            virtual void getThumbnail(sh::filesystem::Operation* operation, std::shared_ptr<sh::filesystem::FilesystemNode> node, QString contentType, int width, int height, QIcon* icon) =0;
        };

        /** \brief
         * Creates thumbnail images for filesystem nodes.
         *
         * This is a cached source.
         */
        class ThumbnailManager : public QObject, public sh::base::Singleton
        {
            Q_OBJECT
            DECLARE_SINGLETON(ThumbnailManager)
        public:

            /** \brief Base value for display indexes of core (i.e. very important) thumbnail providers.
             *
             * Used in ThumbnailProvider::addThumbnailProvider.
             */
            static const int REGISTER_THUMBNAILPROVIDER_INDEX_CORE;

            /** \brief Base value for display indexes of thumbnail providers with normal importance.
             *
             * Used in ThumbnailProvider::addThumbnailProvider.
             */
            static const int REGISTER_THUMBNAILPROVIDER_INDEX_NORMAL;

            /** \brief Base value for display indexes of exotic thumbnail providers.
             *
             * Used in ThumbnailProvider::addThumbnailProvider.
             */
            static const int REGISTER_THUMBNAILPROVIDER_INDEX_EXOTIC;

            /** \brief Base value for display indexes of fallback thumbnail providers (those who try to get something when everything else failed).
             *
             * Used in ThumbnailProvider::addThumbnailProvider.
             */
            static const int REGISTER_THUMBNAILPROVIDER_INDEX_FALLBACK;

            /** @brief Requests a thumbnail in a given size for a given node.
             *
             * \param callback Called when the thumbnail is ready.
             */
            void requestThumbnail(std::shared_ptr<sh::filesystem::FilesystemNode> node, int width, int height, std::function<void(QIcon)> callback = 0);

            /** @brief Returns a thumbnail in a given size for a given node from the current cache.
             */
            bool getThumbnail(std::shared_ptr<sh::filesystem::FilesystemNode> node, int width, int height, QIcon *result, bool *outdated=0, bool *refreshRequested=0);

            /** @brief Invalidates the thumbnail cache.
             */
            void invalidateCache();

            /** @brief Adds a thumbnail provider.
             */
            void addThumbnailProvider(int index, std::shared_ptr<ThumbnailProvider> provider);

        signals:

            /** @brief Emitted when a new thumbnail is available.
             */
            void thumbnailAvailable(std::shared_ptr<sh::filesystem::FilesystemNode> node);

        protected:
            void doShutdown() override;

        private:

            class ThumbnailRequest
            {
            public:
                ThumbnailRequest(std::weak_ptr<sh::filesystem::FilesystemNode> node, int width, int height, QList<std::function<void(QIcon)>> callbacks);
                bool operator==(ThumbnailRequest const& b);
                std::weak_ptr<sh::filesystem::FilesystemNode> node;
                int width;
                int height;
                QList<std::function<void(QIcon)>> callbacks;
            };

            class Thumbnail
            {
            public:
                Thumbnail(QIcon icon, qint64 validUntil, qint64 accessTime, std::weak_ptr<sh::filesystem::FilesystemNode> node, int width, int height);
                QIcon icon;
                qint64 validUntil;
                qint64 accessTime;
                bool refreshRequested;
                std::weak_ptr<sh::filesystem::FilesystemNode> node;
                int width;
                int height;
            };

            class ThumbnailSortStruct
            {
            public:
                ThumbnailSortStruct(sh::filesystem::FilesystemNode* node, qint64 accessTime);
                sh::filesystem::FilesystemNode* node;
                qint64 accessTime;
            };

            ThumbnailManager();
            QHash<sh::filesystem::FilesystemNode*, Thumbnail*> thumbnails;
            QMutex thumbnailslock;
            QList<ThumbnailRequest> requestQueue;
            int maxworkers;
            int runningworkers;
            QMutex workermutex;
            int capacity;
            qint64 _curr_accessTime = 0;
            void worker();
            void _enforce_capacity();
            QHash<int, std::shared_ptr<ThumbnailProvider>> _thumbnailProviderMap;
            QList<std::shared_ptr<ThumbnailProvider>> _thumbnailProviders;
            QMutex _thumbnailProvidersMutex;
        };

    }
}

#endif // THUMBNAILMANAGER_H
