/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2020 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "filesystem/filesystemnode.h"
#include "filesystem/operation.h"
#include "configuration/configurationmanager.h"
#include "base/singletoninitializer.h"
#include "imagemagickpdfthumbnailprovider.h"

REGISTER_STATICINIT(sh::tools::thumbnailproviders, ImageMagickPdfThumbnailProvider, "base")


std::shared_ptr<sh::configuration::ConfigurationValue> sh::tools::thumbnailproviders::ImageMagickPdfThumbnailProvider::cfgvalImagemagickConvertPath =
        sh::configuration::ConfigurationManager::instance()->registerConfigValue(
            "sh::tools::thumbnailproviders::ImageMagickPdfThumbnailProvider.ImagemagickConvertPath", "/usr/bin/convert",
            sh::configuration::ConfigurationValue::valueTypeLocalFilePath(),
            QObject::tr("Path to the 'convert' tool of ImageMagick"),
            sh::configuration::ConfigurationCategory::CategoryExternalTools,
            QObject::tr("This tool is used for getting thumbnails of some file types."));

sh::tools::thumbnailproviders::ImageMagickPdfThumbnailProvider::ImageMagickPdfThumbnailProvider()
            : ThumbnailProvider()
{
    cfgvalImagemagickConvertPath->consumeValue(this, [=](){
        pathToImagemagickConvertTool = cfgvalImagemagickConvertPath->getConfigValueString();
    });
}

void sh::tools::thumbnailproviders::ImageMagickPdfThumbnailProvider::doInitialize()
{
    ThumbnailManager::instance()->addThumbnailProvider(ThumbnailManager::REGISTER_THUMBNAILPROVIDER_INDEX_CORE+1000, std::make_shared<ImageMagickPdfThumbnailProvider>());
}

void sh::tools::thumbnailproviders::ImageMagickPdfThumbnailProvider::doShutdown()
{
}

void sh::tools::thumbnailproviders::ImageMagickPdfThumbnailProvider::getThumbnail(sh::filesystem::Operation *operation, std::shared_ptr<sh::filesystem::FilesystemNode> node, QString contentType, int width, int height, QIcon* icon)
{
    if (contentType.startsWith("application/pdf"))
    {
        QString pdffile;
        sh::exceptions::Exception::executeGuarded([&](){
            pdffile = operation->fetchFile(node->eurl());
        }, sh::exceptions::ExecuteGuardFlag::AUTOMATIC_RETRY_ENABLED|sh::exceptions::ExecuteGuardFlag::IGNORE_ALL_RESUMEABLE_ERRORS_SILENTLY);
        if (!pdffile.isNull())
        {
            QProcess convertprocess;
            convertprocess.setProgram(pathToImagemagickConvertTool);
            QStringList args;
            args << pdffile + "[0]" << "-thumbnail" << QString("%1x%2>").arg(width).arg(height) << "jpeg:-";
            convertprocess.setArguments(args);
            convertprocess.start();
            if (convertprocess.waitForFinished())
            {
                QByteArray bba = convertprocess.readAll();
                QImage img;
                img.loadFromData(bba);
                *icon = QIcon(QPixmap::fromImage(img));
            }
        }
    }
}
