/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2020 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <QPainter>
#include "filesystem/filesystemnode.h"
#include "filesystem/operation.h"
#include "base/mainthread.h"
#include "base/shallotprocess.h"
#include "base/singletoninitializer.h"
#include "plaintextthumbnailprovider.h"

REGISTER_STATICINIT(sh::tools::thumbnailproviders, PlaintextThumbnailProvider, "base")


void sh::tools::thumbnailproviders::PlaintextThumbnailProvider::doInitialize()
{
    ThumbnailManager::instance()->addThumbnailProvider(ThumbnailManager::REGISTER_THUMBNAILPROVIDER_INDEX_CORE+1000, std::make_shared<PlaintextThumbnailProvider>());
}

void sh::tools::thumbnailproviders::PlaintextThumbnailProvider::doShutdown()
{
}

sh::tools::thumbnailproviders::PlaintextThumbnailProvider::PlaintextThumbnailProvider()
{
    QColor c;
    sh::base::MainThread::dispatcher()->invokeSync([&](){
        c = sh::base::ShallotProcess::instance()->brandingColor();
    });
    brandingcolor = QColor::fromHsl(c.hue(), c.saturation(), 50);
}

void sh::tools::thumbnailproviders::PlaintextThumbnailProvider::getThumbnail(sh::filesystem::Operation *operation, std::shared_ptr<sh::filesystem::FilesystemNode> node, QString contentType, int width, int height, QIcon* icon)
{
    if (contentType.startsWith("text/")) {
        QString txtfile;
        sh::exceptions::Exception::executeGuarded([&](){
            txtfile = operation->fetchFile(node->eurl());
        }, sh::exceptions::ExecuteGuardFlag::AUTOMATIC_RETRY_ENABLED|sh::exceptions::ExecuteGuardFlag::IGNORE_ALL_RESUMEABLE_ERRORS_SILENTLY);
        if (!txtfile.isNull()) {
            QString scnt;
            {
                QByteArray cnt;
                QFile fcnt(txtfile);
                fcnt.open(QIODevice::ReadOnly);
                const int BUFFER_SIZE = 4096;
                char buffer[BUFFER_SIZE];
                while (cnt.size() < 50000) {
                    qint64 l = fcnt.read(buffer, BUFFER_SIZE);
                    if (l <= 0)
                        break;
                    cnt.append(buffer, l);
                }
                scnt = QString::fromUtf8(cnt);
                if (!fcnt.atEnd())
                    scnt += "\n\n .\n .\n .";
            }
            QImage img2a(10, height, QImage::QImage::Format_ARGB32);
            img2a.fill(Qt::transparent);
            {
                QPainter pimg2a(&img2a);
                QLinearGradient grad(0, 0, 10, 0);
                grad.setColorAt(0.0, QColor::fromRgb(0, 0, 0, 0));
                grad.setColorAt(1.0, QColor::fromRgb(0, 0, 0));
                pimg2a.fillRect(0, 0, 10, height, grad);
            }
            QImage img2b(width, 10, QImage::QImage::Format_ARGB32);
            img2b.fill(Qt::transparent);
            {
                QPainter pimg2b(&img2b);
                QLinearGradient grad(0, 0, 0, 10);
                grad.setColorAt(0.0, QColor::fromRgb(0, 0, 0, 0));
                grad.setColorAt(1.0, QColor::fromRgb(0, 0, 0));
                pimg2b.fillRect(0, 0, width, 10, grad);
            }
            QImage img(width, height, QImage::QImage::Format_ARGB32);
            img.fill(Qt::transparent);
            {
                QPainter pimg(&img);
                pimg.setPen(brandingcolor);
                pimg.setFont(QFont("monospace", 6, QFont::Light));
                QTextOption textopts;
                textopts.setWrapMode(QTextOption::NoWrap);
                pimg.drawText(img.rect(), scnt, textopts);
                pimg.setCompositionMode(QPainter::CompositionMode_DestinationOut);
                pimg.drawImage(width-10, 0, img2a);
                pimg.drawImage(0, height-10, img2b);
            }
            *icon = QIcon(QPixmap::fromImage(img));
        }
    }
}
