/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2020 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <QObject>
#ifdef Q_OS_UNIX
#include <stdio.h>
#include <ctype.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/sem.h>
#endif

#include "base/shallotprocess.h"
#include "exceptions/runtimeexception.h"
#include "benchmarking.h"
#include "userdirlock.h"


sh::tools::UserDirLock::UserDirLock()
{
    BENCHMARK_SCOPE("UserDirLock::UserDirLock")
    QMutexLocker _l1(&mutex);
    if (QThread::currentThread() == currentThread)
        dounlock = false;
    else
    {
        dounlock = true;
#ifdef Q_OS_UNIX
        static QByteArray bmypath = sh::base::ShallotProcess::userDataDir().toLocal8Bit();
        static char* cmypath = bmypath.data();
        static key_t key = ftok(cmypath, 's');
        static bool _inited = false;
        if (!_inited) {
            if ((semset_id = semget(key, 1, IPC_CREAT|IPC_EXCL|0666)) == -1)
                semset_id = semget(key, 0, 0666);
            else
            {
                union semun semopts;
                semopts.val = 1;
                semctl(semset_id, 0, SETVAL, semopts);
            }
        }
        mutex.unlock();
        struct sembuf sem_lock = {0, -1, SEM_UNDO};
        sem_lock.sem_num = 0;
        int res = semop(semset_id, &sem_lock, 1);
        mutex.lock();
        if (res == -1)
            throw sh::exceptions::RuntimeException(sh::exceptions::ExceptionData().detailsAreInteresting().details(QObject::tr("Error while acquiring the config files lock.")));
#else
        QString mypath = sh::base::ShallotProcess::userDataDir(); //TODO # this mechanism now leads to 5 sec additional waiting time on first shallot start. in the past, we created '.lock' on startup, but this is incorrect (maybe another shallot just locked it!)
        slockpref = mypath + ".lock";
        bool succ = false;
        while(!succ)
        {
            int i = 0;
            qint64 nowms = QDateTime::currentMSecsSinceEpoch();
            QString _slockfile;
            while(!succ && (QDateTime::currentMSecsSinceEpoch() < nowms + 5000))
            {
                _slockfile = slockpref+QString(".%1").arg(i);
                succ = QFile::rename(slockpref, _slockfile);
                i++;
                if (!succ)
                {
                    mutex.unlock();
                    QDateTime now = QDateTime::currentDateTime();
                    foreach(QString e, QDir(mypath).entryList(QDir::Dirs))
                    {
                        if (e.startsWith(".lock.") && QFileInfo(mypath+e).lastModified().secsTo(now) > 30)
                            QFile::remove(mypath+e);
                    }
                    QThread::msleep(5);
                    mutex.lock();
                }
            }
            if (succ)
                slockfile = _slockfile;
            else
            {
                foreach(QString e, QDir(mypath).entryList(QDir::Dirs))
                {
                    if (e.startsWith(".lock."))
                        QFile::remove(mypath+e);
                }
                QDir(mypath).mkdir(".lock");
            }
        }
#endif
        currentThread = QThread::currentThread();
    }
}

QMutex sh::tools::UserDirLock::mutex;

QThread* sh::tools::UserDirLock::currentThread = 0;

sh::tools::UserDirLock::~UserDirLock()
{
    QMutexLocker _l1(&mutex);
    BENCHMARK_SCOPE("UserDirLock::UserDirLock")
    if (dounlock)
    {
        currentThread = 0;
        sh::exceptions::Exception::executeGuarded([&](){
#ifdef Q_OS_UNIX
            struct sembuf sem_unlock = {0, 1, SEM_UNDO};
            sem_unlock.sem_num = 0;
            if ((semop(semset_id, &sem_unlock, 1)) == -1)
                throw sh::exceptions::RuntimeException(sh::exceptions::ExceptionData().detailsAreInteresting().details(QObject::tr("Error while releasing the config files lock.")));
#else
            if (!QFile::rename(slockfile, slockpref))
                throw sh::exceptions::RuntimeException(sh::exceptions::ExceptionData().detailsAreInteresting().details(QObject::tr("The lock was stolen while saving settings.")));
#endif
        }, sh::exceptions::ExecuteGuardFlag::RESUMEABLE_PROGRAM_ERROR_END_WITH_USER_FEEDBACK_HERE);
    }
}
