/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2020 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "base/mainthread.h"
#include "dialog.h"
#include "mainwindow.h"


sh::ui::DialogManager::~DialogManager()
{
}

void sh::ui::DialogManager::initAndShowDialog(std::shared_ptr<Dialog> dialog)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    qint64 dialogId = _nextDialogId++;
    dialog->setup(this, dialogId);
    dialog->init(dialog);
    dialog->_inited = true;
    if (!_openDialogs.contains(dialogId)) {
        _openDialogs[dialogId] = dialog;
        showDialog(dialog);
    }
}

void sh::ui::DialogManager::closeDialog(std::shared_ptr<Dialog> dialog)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    auto dlgid = dialog->dialogId();
    if (_openDialogs.contains(dlgid)) {
        dialog->prepareClose();
        stopDialog(dialog);
        _openDialogs.remove(dlgid);
    }
}

std::shared_ptr<sh::ui::Dialog> sh::ui::DialogManager::getDialogById(qint64 id)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    return _openDialogs.value(id);
}

QList<std::shared_ptr<sh::ui::Dialog>> sh::ui::DialogManager::getAllDialogs()
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    return _openDialogs.values();
}

QList<qint64> sh::ui::DialogManager::getAllDialogIds()
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    return _openDialogs.keys();
}

sh::ui::Dialog::Dialog()
{
}

sh::ui::Dialog::~Dialog()
{
}

qint64 sh::ui::Dialog::dialogId()
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    verifyInited();
    return _dialogId;
}

void sh::ui::Dialog::init(std::shared_ptr<Dialog>)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    // remains empty in base class
}

void sh::ui::Dialog::setup(DialogManager* manager, qint64 dialogId)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    _manager = manager;
    _dialogId = dialogId;
}

void sh::ui::Dialog::prepareClose()
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    // remains empty in base class
}

void sh::ui::Dialog::verifyInited()
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
#ifdef QT_DEBUG
    if (!isInited())
        throw sh::exceptions::ProgramException(sh::exceptions::ExceptionData().details("WebDialog routines were used before initialization"));
#endif
}

bool sh::ui::Dialog::isInited()
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    return (_dialogId != -1);
}

bool sh::ui::Dialog::wasAccepted()
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    return manager()->wasAccepted(shared_from_this());
}

void sh::ui::Dialog::waitClosed()
{
    manager()->waitClosed(shared_from_this());
}

void sh::ui::Dialog::close()
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    verifyInited();
    manager()->closeDialog(shared_from_this());
    _closed = true;
}

bool sh::ui::Dialog::wasClosed()
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    return _closed;
}

sh::ui::DialogManager *sh::ui::Dialog::manager()
{
    return _manager;
}
