/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2020 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef DIALOG_H
#define DIALOG_H

#include <memory>
#include <QObject>
#include "base/mainthread.h"


namespace sh{
    namespace ui {
        class Dialog;

        /** \brief Creates and drives dialogs (in child windows), i.e. instances of Dialog.
         *
         * Subclass it for implementing the low level mechanisms for handling dialogs in a particular
         * ui mode (e.g. qt, web).
         *
         * The infrastructure will use one of this implementations for creating and managing most
         * kinds of dialogs.
         */
        class DialogManager
        {
            friend class Dialog;
        public:

            /** \brief Returns a Dialog by dialog id.
             *
             * Must be called in main thread.
             */
            std::shared_ptr<Dialog> getDialogById(qint64 id);

            /** \brief Returns a list of all dialogs currently shown (in order of creation).
             *
             * Must be called in main thread.
             */
            QList<std::shared_ptr<Dialog>> getAllDialogs();

            /** \brief Returns a list of dialog ids of all dialogs currently shown (in order of creation).
             *
             * Must be called in main thread.
             */
            QList<qint64> getAllDialogIds();

            /** \brief Creates and shows a Dialog by class and constructor parameters.
             *
             * Those dialogs (TDlg) have to implement Dialog (typically a subclass of
             * it, representing a particular dialog, like FilePropertyDialog), and also
             * some ui mode (e.g. qt, web) specific class (depends on that ui mode).
             *
             * You typically should not have to use it outside of MainWindow or subclasses.
             * The MainWindow interface allows to create all available dialogs in a ui mode independent way.
             *
             * May be called in any thread.
             *
             * @a_infrastructure
             * @returns The created Dialog instance.
             */
            template<class TDlg, typename... Args>
            std::shared_ptr<TDlg> createAndShowDialog(Args... args)
            {
                std::shared_ptr<TDlg> dlg;
                sh::base::MainThread::dispatcher()->invokeSync([&](){
                    dlg = std::shared_ptr<TDlg>(new TDlg(args...), [](TDlg* t){
                        sh::base::MainThread::dispatcher()->invokeSync([&](){
                            delete t;
                        });
                    });
                    initAndShowDialog(dlg);
                });
                return dlg;
            }

            virtual ~DialogManager();

        private:

            /** \brief Closes this dialog, including internal bookkeeping.
             *
             * Must be called in main thread.
             *
             * @a_infrastructure
             */
            void closeDialog(std::shared_ptr<Dialog> dialog);

            /** \brief Initializes and shows a freshly created dialog.
             *
             * Must be called in main thread.
             *
             * @a_infrastructure
             */
            void initAndShowDialog(std::shared_ptr<Dialog> dialog);

            /** \brief Returns if the dialog was 'accepted' by the user (typically by clicking some 'OK' button).
             *
             * The result is undefined before the dialog was closed.
             *
             * Must be called in main thread.
             *
             * @a_dooverride
             * @a_infrastructure
             */
            virtual bool wasAccepted(std::shared_ptr<Dialog> dialog) =0;

            /** \brief Wait until the user closed the dialog in some way.
             *
             * May be called in any thread.
             *
             * @a_dooverride
             * @a_infrastructure
             */
            virtual void waitClosed(std::shared_ptr<Dialog> dialog) =0;

            /** \brief This method implements the ui mode specific mechanism for stopping (i.e. closing) a dialog.
             *
             * Must be called in main thread.
             *
             * @a_dooverride
             * @a_infrastructure
             */
            virtual void stopDialog(std::shared_ptr<Dialog> dialog) =0;

            /** \brief This method implements the ui mode specific mechanism for showing a dialog.
             *
             * Must be called in main thread.
             *
             * @a_dooverride
             * @a_infrastructure
             */
            virtual void showDialog(std::shared_ptr<Dialog> dialog) =0;

            QHash<qint64, std::shared_ptr<Dialog>> _openDialogs;
            qint64 _nextDialogId = 0;
        };

        /** \brief Abstract subclass for Shallot dialogs (in child windows).
         *
         * Subclasses of Dialog represent particular dialog types (like ManageBookmarksDialog).
         *
         * For each ui mode, there typically is an implementation for all of these types, implementing
         * one of this dialog types subclasses (mentioned before) and also some ui mode specific class.
         *
         * Note: Unless stated otherwise, all methods must be called from main thread!
         *
         * Note: It's not allowed to show one web dialog instance more than once. After closing it, it's sole
         * use is to fetch answer data from it.
         *
         * Note: You should not create instances directly. It's required to use DialogManager::createAndShowDialog()
         * for actually showing those dialogs. Even this should not be used directly, because call depends on the current
         * ui mode. Use the show*() methods in sh::ui::MainWindow for creating dialogs in a ui mode independent way.
         */
        class Dialog : public std::enable_shared_from_this<Dialog>
        {
            friend class MainWindow;
            friend class DialogManager;
        public:

            /** @a_infrastructure
             */
            Dialog();
            virtual ~Dialog();

            /** \brief Returns the dialog id.
             *
             * Each instance has an id unique in the complete Shallot process lifetime.
             *
             * Must be called in main thread.
             */
            qint64 dialogId();

            /** \brief Returns if this dialog is initialized.
             *
             * Must be called in main thread.
             */
            bool isInited();

            /** \brief Returns if this dialog was 'accepted' by the user (typically by clicking some 'OK' button).
             *
             * Must be called in main thread.
             */
            bool wasAccepted();

            /** \brief Wait until the user closed the dialog in some way.
             *
             * May be called in any thread.
             */
            void waitClosed();

            /** \brief Closes the dialog.
             *
             * Must be called in main thread.
             */
            void close();

            /** \brief Returns if this dialog was closed.
             *
             * Must be called in main thread.
             */
            bool wasClosed();

            /** \brief Returns the DialogManager which hosts this dialog.
             */
            DialogManager* manager();

        protected:

            /** \brief Throws an exception if this dialog isn't initialized.
             *
             * Must be called in main thread.
             *
             * @a_infrastructure
             */
            void verifyInited();

            /** \brief Executes custom stuff on closing.
             *
             * @a_dooverride_optional
             */
            virtual void prepareClose();

            /** \brief Executes custom stuff on initialization, e.g. for populating the dialog.
             *
             * @a_dooverride_optional
             */
            virtual void init(std::shared_ptr<Dialog> dialog);

        private:

            /** \brief Sets manager and id for this dialog (right after creation).
             */
            void setup(DialogManager* manager, qint64 dialogId);

            DialogManager* _manager;
            qint64 _dialogId = -1;
            bool _closed = false;
            bool _inited = false;
        };

    }
}

#endif // DIALOG_H
