/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2020 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef EXCEPTIONDIALOG_H
#define EXCEPTIONDIALOG_H

#include "dialog.h"


namespace sh{
    namespace ui {

        /** \brief Enumeration of decisions the user can make in an ExceptionDialog.
         */
        enum class ExceptionDialogResult
        {
            Shutdown,
            Close,
            Cancel,
            Retry,
        };

        /** \brief The 'Exception' dialog.
         */
        class ExceptionDialog : public Dialog
        {
        public:

            /**
             * \param error1: The 1st error text.
             * \param error2: The 2nd error text.
             * \param details: The error details.
             * \param icon: An icon (as name resolveable by sh::base::IconManager).
             * \param mayRetry: If it's allowed to retry.
             * \param mayClose: If it's allowed to just close and continue without closing Shallot.
             * \param mayCancel: If it's allowed to cancel, i.e. throwing a sh::exceptions::CancelException.
             * \param showLoglabelAndDetails: If the dialog shall allow to read the details.
             * @a_infrastructure
             */
            ExceptionDialog(QString error1, QString error2, QString details, QString icon, bool mayRetry, bool mayClose, bool mayCancel, bool showLoglabelAndDetails);

            /** \brief Returns the 1st error text.
             */
            QString error1();

            /** \brief Returns the 2nd error text.
             */
            QString error2();

            /** \brief Returns the error details.
             */
            QString details();

            /** \brief Returns the icon (as name resolveable by sh::base::IconManager).
             */
            QString icon();

            /** \brief Returns if it's allowed to retry.
             */
            bool mayRetry();

            /** \brief Returns if it's allowed to just close and continue without closing Shallot.
             */
            bool mayClose();

            /** \brief Returns if it's allowed to cancel, i.e. throwing a sh::exceptions::CancelException.
             */
            bool mayCancel();

            /** \brief Returns if the dialog shall allow to read the details.
             */
            bool showLoglabelAndDetails();

            /** \brief Returns the answer the user has given in this dialog.
             */
            virtual ExceptionDialogResult answer() =0;

        private:
            QString _error1;
            QString _error2;
            QString _details;
            QString _icon;
            bool _mayRetry;
            bool _mayClose;
            bool _mayCancel;
            bool _showLoglabelAndDetails;
        };

    }
}

#endif // EXCEPTIONDIALOG_H
