/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2020 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "filesystem/operation.h"
#include "base/mainthread.h"
#include "base/singletoninitializer.h"
#include "base/threadpool.h"
#include "filepropertydialog.h"


sh::ui::FilePropertyDialog::FilePropertyDialog(QList<std::shared_ptr<sh::filesystem::FilesystemNode> > nodes) : _nodes(nodes)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
}

sh::ui::FilePropertyDialog::~FilePropertyDialog()
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
}

void sh::ui::FilePropertyDialog::addTabFactory(int i, std::shared_ptr<FilePropertyDialogTabFactory> factory)
{
    sh::base::MainThread::dispatcher()->invokeSync([=](){
        _propertytabs.insertMulti(i, factory);
    });
}

sh::ui::FilePropertyDialogTabFactoryByFunction::FilePropertyDialogTabFactoryByFunction(std::function<std::shared_ptr<sh::ui::FilePropertyDialogTab>()> fct)
                : fct(fct)
{
}

std::shared_ptr<sh::ui::FilePropertyDialogTab> sh::ui::FilePropertyDialogTabFactoryByFunction::construct(std::shared_ptr<sh::ui::FilePropertyDialog> dialog)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    auto tab = fct();
    tab->_dialog = dialog;
    return tab;
}

sh::ui::TabViewStructWidgets sh::ui::FilePropertyDialog::findWidgetsForTab(QList<TabViewStruct> list, std::shared_ptr<sh::ui::FilePropertyDialogTab> tab)
{
    foreach (auto tbs, list) {
        if (tbs.first == tab)
            return tbs.second;
    }
    return TabViewStructWidgets();
}

void sh::ui::FilePropertyDialog::updateWidgets()
{
    sh::base::MainThread::dispatcher()->invokeSync([=](){
        std::shared_ptr<sh::filesystem::Operation> op = std::make_shared<sh::filesystem::Operation>();
        foreach (auto tab, tabs()) {
            int wc = tab->widgetCount();
            for (int iw=0; iw<wc; iw++) {
                sh::ui::FilePropertyDialogTabActionsView* wiw = tab->widgetAt(iw);
                sh::base::ThreadPool::enqueueForce([=](){
                    sh::exceptions::Exception::executeGuarded_errorpanel([&](){
                        tab.get()->updateWidget(iw, wiw, op.get());
                    });
                }, tabs()[iw]);
            }
        }
    });
}

const int sh::ui::FilePropertyDialogTabFactory::REGISTER_FILEPROPERTYDIALOGTAB_INDEX_CORE = 10000;
const int sh::ui::FilePropertyDialogTabFactory::REGISTER_FILEPROPERTYDIALOGTAB_INDEX_VERYIMPORTANT = 20000;
const int sh::ui::FilePropertyDialogTabFactory::REGISTER_FILEPROPERTYDIALOGTAB_INDEX_IMPORTANT = 30000;
const int sh::ui::FilePropertyDialogTabFactory::REGISTER_FILEPROPERTYDIALOGTAB_INDEX_NORMAL = 40000;
const int sh::ui::FilePropertyDialogTabFactory::REGISTER_FILEPROPERTYDIALOGTAB_INDEX_EXOTIC = 50000;

sh::ui::FilePropertyDialogTabFactory::~FilePropertyDialogTabFactory()
{
}

QList<std::shared_ptr<sh::filesystem::FilesystemNode> > sh::ui::FilePropertyDialogTab::nodes()
{
    return _dialog.lock()->nodes();
}

std::shared_ptr<sh::ui::FilePropertyDialog> sh::ui::FilePropertyDialogTab::dialog()
{
    return _dialog.lock();
}

sh::ui::FilePropertyDialogTabActionsView* sh::ui::FilePropertyDialogTab::widgetAt(int i)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    return _dialog.lock()->widgetAt(shared_from_this(), i);
}

int sh::ui::FilePropertyDialogTab::widgetCount()
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    return _dialog.lock()->widgetCount(shared_from_this());
}

QString sh::ui::FilePropertyDialogTab::titleWithoutMnemonic()
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    QString titleWithoutMnemonic;
    bool lastWasAmp = false;
    foreach (QChar c, title()) {
        if (c != '&' || lastWasAmp)
            titleWithoutMnemonic += c;
        lastWasAmp = (c == '&');
    }
    return titleWithoutMnemonic;
}

void sh::ui::FilePropertyDialogTab::refresh()
{
    _dialog.lock()->refresh();
}

QList<sh::ui::FilePropertyDialogTabFactory*> sh::ui::FilePropertyDialog::getPropertyTabs()
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    QList<FilePropertyDialogTabFactory*> result;
    QList<int> keys = _propertytabs.keys();
    qSort(keys);
    foreach (int i, keys)
        foreach (auto f, _propertytabs.values(i))
            result.append(f.get());
    return result;
}

sh::ui::FilePropertyDialogTabActionsView* sh::ui::FilePropertyDialog::widgetAt(std::shared_ptr<FilePropertyDialogTab> tab, int i)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    return FilePropertyDialog::findWidgetsForTab(_tabs, tab).value(i, nullptr);
}

int sh::ui::FilePropertyDialog::widgetCount(std::shared_ptr<FilePropertyDialogTab> tab)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    return FilePropertyDialog::findWidgetsForTab(_tabs, tab).count();
}

QList<std::shared_ptr<sh::ui::FilePropertyDialogTab>> sh::ui::FilePropertyDialog::tabs()
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    QList<std::shared_ptr<sh::ui::FilePropertyDialogTab>> result;
    foreach (auto x, _tabs)
        result.append(x.first);
    return result;
}

void sh::ui::FilePropertyDialog::addTabstruct(std::shared_ptr<sh::ui::FilePropertyDialogTab> tab, TabViewStructWidgets widgets)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    _tabs.append(TabViewStruct(tab, widgets));
}

QList<sh::ui::TabViewStruct> sh::ui::FilePropertyDialog::tabstructs()
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    return _tabs;
}

QList<std::shared_ptr<sh::filesystem::FilesystemNode>> sh::ui::FilePropertyDialog::nodes()
{
    return _nodes;
}

sh::ui::FilePropertyDialogTab::~FilePropertyDialogTab()
{
}

QHash<int, std::shared_ptr<sh::ui::FilePropertyDialogTabFactory>> sh::ui::FilePropertyDialog::_propertytabs;

