/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2020 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef FILEPROPERTYDIALOG_H
#define FILEPROPERTYDIALOG_H

#include <functional>
#include <memory>
#include <QObject>
#include "dialog.h"


namespace sh {
    namespace filesystem {
        class FilesystemNode;
        class Operation;
    }
    namespace scripting {
        namespace api { class ApiFilePropertyDialogTabFactory; }
    }
    namespace ui {
        class FilePropertyDialogTabFactory;
        class FilePropertyDialogTab;
        class FilePropertyDialogTabActionsView;
        class FilePropertyDialogTabTableView;
        class FilePropertyDialogTabTextView;
        class FilePropertyDialogTabIconTextBannerView;

        typedef QList<sh::ui::FilePropertyDialogTabActionsView*> TabViewStructWidgets;
        typedef QPair<std::shared_ptr<sh::ui::FilePropertyDialogTab>, TabViewStructWidgets> TabViewStruct;

        /**
         * \brief The 'File Properties' dialog.
         */
        class FilePropertyDialog : public Dialog
        {
            friend class FilePropertyDialogTab;
        public:

            /**
             * \brief Adds a FilePropertyDialogTabFactory so the Properties dialogs show its
             * content.
             *
             * See also the REGISTER_FILEPROPERTYDIALOGTAB macro.
             *
             * \param i An index which controls the display order. Use one of the
             * REGISTER_FILEPROPERTYDIALOGTAB_INDEX_* values in FilePropertyDialogTabFactory as base values.
             */
            static void addTabFactory(int i, std::shared_ptr<FilePropertyDialogTabFactory> factory);

            virtual ~FilePropertyDialog();

            /** @a_infrastructure
             */
            explicit FilePropertyDialog(QList<std::shared_ptr<sh::filesystem::FilesystemNode>> nodes);

            /** \brief Refreshes the dialog content.
             */
            virtual void refresh() =0;

            /** \brief Returns a list of all tabs.
             */
            QList<std::shared_ptr<FilePropertyDialogTab>> tabs();

            /** \brief Returns the widget from a tab at a certain index.
             */
            sh::ui::FilePropertyDialogTabActionsView* widgetAt(std::shared_ptr<FilePropertyDialogTab> tab, int i);

            /** \brief Returns the number of widgets in a tab.
             */
            int widgetCount(std::shared_ptr<FilePropertyDialogTab> tab);

            /** \brief Creates a new tab view table sub-widget.
             */
            virtual sh::ui::FilePropertyDialogTabTableView* createTabViewTable() =0;

            /** \brief Creates a new tab view text sub-widget.
             */
            virtual sh::ui::FilePropertyDialogTabTextView* createTabViewText() =0;

            /** \brief Creates a new tab view icon text banner sub-widget.
             */
            virtual sh::ui::FilePropertyDialogTabIconTextBannerView* createTabViewIconTextBanner() =0;

        protected:

            /** \brief Returns the list of tab widgets (sh::ui::FilePropertyDialogTabActionsView) for a given tab from a given list of tabs.
             */
            static TabViewStructWidgets findWidgetsForTab(QList<TabViewStruct> list, std::shared_ptr<sh::ui::FilePropertyDialogTab> tab);

            /** \brief Asks all widgets in all tabs to update.
             */
            virtual void updateWidgets();

            /** \brief Returns the list of all registered tab factories.
             */
            static QList<sh::ui::FilePropertyDialogTabFactory*> getPropertyTabs();

            /** \brief Adds a new tab instance.
             */
            void addTabstruct(std::shared_ptr<sh::ui::FilePropertyDialogTab> tab, TabViewStructWidgets widgets);

            /** \brief Returns the list of tab structures (e.g. containing the actual content widgets).
             */
            QList<sh::ui::TabViewStruct> tabstructs();

            /** \brief Returns the list of filesystem nodes to show properties for.
             */
            QList<std::shared_ptr<sh::filesystem::FilesystemNode>> nodes();

        private:
            static QHash<int, std::shared_ptr<FilePropertyDialogTabFactory>> _propertytabs;
            QList<std::shared_ptr<sh::filesystem::FilesystemNode>> _nodes;
            QList<TabViewStruct> _tabs;

        };

        /**
         * \brief Abstract base class for one tab in the Properties dialog (containing a list of widgets).
         *
         * Subclass and register this class for providing additional content in the
         * Properties dialog.
         *
         * Register it by using the REGISTER_FILEPROPERTYDIALOGTAB macro (or by adding a
         * FilePropertyDialogTabFactory via FilePropertyDialog::addTabFactory).
         */
        class FilePropertyDialogTab : public std::enable_shared_from_this<FilePropertyDialogTab>
        {
            friend class FilePropertyDialog;
            friend class FilePropertyDialogTabFactoryByFunction;
            friend class sh::scripting::api::ApiFilePropertyDialogTabFactory;

        public:

            /** \brief The tab title.
             * @a_dooverride
             */
            virtual QString title() =0;

            /** \brief The tab title without mnemonic.
             */
            QString titleWithoutMnemonic();

            /** \brief Returns the list of property labels (one entry for each widget).
             *
             * Each property is displayed with a header and a widget (created in createWidget)
             * with some content (populated in updateWidget).
             * @a_dooverride
             */
            virtual QList<QString> properties() =0;

            /** \brief Populates the tab widget for the i-th property.
             *
             * Typically uses the FilePropertyDialog::create* methods to create sub-widgets.
             *
             * See also dialog().
             * @a_dooverride
             */
            virtual void populateWidget(int i, FilePropertyDialogTabActionsView* widget) =0;

            /** \brief Populates the widget for the i-th property with actual content.
             * @a_dooverride
             */
            virtual void updateWidget(int i, FilePropertyDialogTabActionsView* widget, sh::filesystem::Operation *op) =0;

            /**
             * \brief Refreshes the complete content.
             *
             * Typically called after some user actions in a property widget require that all the other content must be refreshed.
             */
            void refresh();

            /** \brief The nodes to show.
             */
            QList<std::shared_ptr<sh::filesystem::FilesystemNode>> nodes();

            /** \brief Returns the widget for the i-th property.
             */
            FilePropertyDialogTabActionsView* widgetAt(int i);

            /** \brief Returns the number of widgets.
             */
            int widgetCount();

            /** \brief Returns the associated dialog.
             */
            std::shared_ptr<FilePropertyDialog> dialog();

            virtual ~FilePropertyDialogTab();

        private:
            std::weak_ptr<FilePropertyDialog> _dialog;

        };

        /** \brief Abstract factory for creating property dialog tabs for the selected nodes.
         *
         * Register an instance of this class with FilePropertyDialog::addTabFactory in order to make a property tab implementation available.
         * See also FilePropertyDialogTabFactoryByFunction.
         */
        class FilePropertyDialogTabFactory
        {
        public:

            /** \brief Base value for display indexes of core (i.e. maximum important) tabs.
             *
             * Used in FilePropertyDialog::addTabFactory.
             */
            static const int REGISTER_FILEPROPERTYDIALOGTAB_INDEX_CORE;

            /** \brief Base value for display indexes of very important tabs.
             *
             * Used in FilePropertyDialog::addTabFactory.
             */
            static const int REGISTER_FILEPROPERTYDIALOGTAB_INDEX_VERYIMPORTANT;

            /** \brief Base value for display indexes of important tabs.
             *
             * Used in FilePropertyDialog::addTabFactory.
             */
            static const int REGISTER_FILEPROPERTYDIALOGTAB_INDEX_IMPORTANT;

            /** \brief Base value for display indexes of tabs with medium importance.
             *
             * Used in FilePropertyDialog::addTabFactory.
             */
            static const int REGISTER_FILEPROPERTYDIALOGTAB_INDEX_NORMAL;

            /** \brief Base value for display indexes of exotic (i.e. less important) tabs.
             *
             * Used in FilePropertyDialog::addTabFactory.
             */
            static const int REGISTER_FILEPROPERTYDIALOGTAB_INDEX_EXOTIC;

            /** \brief Constructs a fresh instance of a FilePropertyDialogTab implementation.
             * @a_dooverride
             */
            virtual std::shared_ptr<sh::ui::FilePropertyDialogTab> construct(std::shared_ptr<sh::ui::FilePropertyDialog> dialog) =0;

            virtual ~FilePropertyDialogTabFactory();

        };

        /**
         * \brief Implementation of FilePropertyDialogTabFactory for making custom FilePropertyDialogTab implementations available.
         *
         * This implementation relies on an external factory function which must be provided to the constructor.
         *
         * Used internally inside the REGISTER_FILEPROPERTYDIALOGTAB macro.
         */
        class FilePropertyDialogTabFactoryByFunction : public FilePropertyDialogTabFactory
        {
        public:
            FilePropertyDialogTabFactoryByFunction(std::function<std::shared_ptr<sh::ui::FilePropertyDialogTab>()> fct);
            virtual std::shared_ptr<sh::ui::FilePropertyDialogTab> construct(std::shared_ptr<sh::ui::FilePropertyDialog> dialog);
        private:
            std::function<std::shared_ptr<sh::ui::FilePropertyDialogTab>()> fct;
        };

    }
}

#define REGISTER_FILEPROPERTYDIALOGTAB(I,T) sh::ui::FilePropertyDialog::addTabFactory((I), std::make_shared<sh::ui::FilePropertyDialogTabFactoryByFunction>([](){ return (std::shared_ptr<sh::ui::FilePropertyDialogTab>) new T(); }))

#endif // FILEPROPERTYDIALOG_H
