/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2020 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef FILEPROPERTYDIALOGTABACTIONSVIEW_H
#define FILEPROPERTYDIALOGTABACTIONSVIEW_H

#include <functional>
#include <memory>
#include <QObject>


namespace sh {
    namespace ui {
        class FilePropertyDialogTabViewContent;

        /** \brief A helper for some signal/slot stuff of in FilePropertyDialogTabActionsView.
         */
        class _FilePropertyDialogTabActionsView_HelperQObject : public QObject
        {
            Q_OBJECT
        public:
            void _emit_buttonTriggered(int i);
        signals:
            void buttonTriggered(int i);
        };

        /** \brief A tab view which shows a main widget in the main part and some buttons below.
         */
        class FilePropertyDialogTabActionsView
        {
        public:

            /** \brief Is intended to be directly constructed from everywhere.
             */
            explicit FilePropertyDialogTabActionsView();

            /** \brief The main widget.
             */
            FilePropertyDialogTabViewContent* content();

            /** \brief Sets the main widget.
             * @a_dooverride
             */
            virtual void setContent(FilePropertyDialogTabViewContent* cnt);

            /** \brief The list of buttons.
             */
            QList<QString> buttons();

            /** \brief Sets the list of buttons.
             * @a_dooverride
             */
            virtual void setButtons(QList<QString> buttons);

            /** \brief Sets the view visible or hidden.
             * @a_dooverride
             */
            virtual void setVisible(bool v) =0;

            /** \brief Sets a handler for a click on one of the buttons (optionally bound to an owner lifetime).
             */
            void onButtonTriggered(std::function<void(int i)> fct, QObject* owner=0);

        protected:
            QList<QString> _buttons;
            FilePropertyDialogTabViewContent* _content=0;
            _FilePropertyDialogTabActionsView_HelperQObject myqobject;
            /** \brief Called by an implementation when a button was clicked.
             */
            void _emit_buttonTriggered(int i);

        };

        /** \brief Abstract subclass for a tab view content.
         */
        class FilePropertyDialogTabViewContent
        {
        public:
            FilePropertyDialogTabViewContent();
            virtual ~FilePropertyDialogTabViewContent();
        };

        /** \brief A tab view for showing key/value pairs.
         */
        class FilePropertyDialogTabTableView : public FilePropertyDialogTabViewContent
        {
        public:
            typedef QPair<int, int> ItemIndex;

            /** \brief Is intended to be directly constructed from everywhere.
             */
            FilePropertyDialogTabTableView();
            /** \brief Sets the content.
             */
            virtual void setContent(QList<QPair<QString, QString>> content) =0;
            /** \brief Returns the selected cells.
             */
            virtual QList<ItemIndex> getSelectedItems() =0;
            /** \brief Returns a value of a cell.
             */
            virtual QVariant getItemValue(int r, int c) =0;
        };

        /** \brief A tab view for showing a text.
         */
        class FilePropertyDialogTabTextView : public FilePropertyDialogTabViewContent
        {
        public:
            /** \brief Is intended to be directly constructed from everywhere.
             */
            FilePropertyDialogTabTextView();
            /** \brief Sets the content.
             */
            virtual void setContent(QString content) =0;
        };

        /** \brief A tab view for showing texts and icons combined.
         */
        class FilePropertyDialogTabIconTextBannerView : public FilePropertyDialogTabViewContent
        {
        public:
            /** \brief Is intended to be directly constructed from everywhere.
             */
            FilePropertyDialogTabIconTextBannerView();
            /** \brief Clears the contents.
             */
            virtual void clear() =0;
            /** \brief Adds an icon.
             */
            virtual void insertIcon(QIcon icon, int sizePt = 24) =0;
            /** \brief Adds a text.
             */
            virtual void insertText(QString text) =0;
        };

    }
}

#endif // FILEPROPERTYDIALOGTABACTIONSVIEW_H
