/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2020 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef FILEVIEW_H
#define FILEVIEW_H

#include <memory>
#include <QObject>
#include "filesystem/filesystemmodelfileviewproxy.h"
#include "tools/historytracker.h"


namespace sh{
    namespace filesystem{ class Eurl; class FilesystemNode; }

    namespace ui{
        class ColumnDimensions;

        enum class FileViewMode
        {
            IconView,
            ListView
        };

        /** \brief
         * Abstract base class for a file view implementation (i.e. which shows the content of a directory somehow).
         */
        class FileView : public QObject
        {
            Q_OBJECT
        public:

            /** \brief
             * @a_infrastructure
             */
            FileView();

            virtual ~FileView();

            /** \brief Returns the column dimensions handler.
             * @a_dooverride
             */
            virtual sh::ui::ColumnDimensions* columndimensions() =0;

            /** \brief Returns the history tracker.
             * @a_dooverride
             */
            virtual sh::tools::HistoryTracker<std::shared_ptr<const sh::filesystem::Eurl>>* historyTracker() =0;

            /** \brief Returns the view mode (icons, list, ... ?).
             * @a_dooverride
             */
            virtual FileViewMode viewmode() =0;

            /** \brief Sets the view mode.
             * @a_dooverride
             */
            virtual void setViewmode(FileViewMode m) =0;

            /** \brief Returns the position of this file view within the panel.
             * @a_dooverride
             */
            virtual int index() =0;

            /** \brief Returns the current directory.
             */
            std::shared_ptr<sh::filesystem::FilesystemNode> node();

            /** \brief Jumps to a new current directory.
             *
             * @a_dooverride
             * Implementations have to call the base class implementation inside.
             */
            virtual void gotoDir(std::shared_ptr<sh::filesystem::FilesystemNode> node);

            /** \brief Returns the mode how file sizes are formatted for displaying.
             * @a_dooverride
             */
            virtual sh::filesystem::SizeFormatting getSizeFormattingMode() =0;

            /** \brief Sets the mode how file sizes are formatted for displaying.
             * @a_dooverride
             */
            virtual void setSizeFormattingMode(sh::filesystem::SizeFormatting mode) =0;

            /** \brief Returns if hidden files are visible.
             * @a_dooverride
             */
            virtual bool hiddenFilesVisible() =0;

            /** \brief Sets the visibility of hidden files.
             * @a_dooverride
             */
            virtual void setHiddenFilesVisible(bool v) =0;

            /** \brief Returns the icon size (in pixels).
             * @a_dooverride
             */
            virtual int iconDimension() =0;

            /** \brief Sets the icon size (in pixels).
             * @a_dooverride
             */
            virtual void setIconDimension(int v) =0;

            /** \brief Sets how to sort this view.
             * @a_dooverride
             */
            virtual void setSort(int column, Qt::SortOrder order) =0;

            /** \brief Returns the index of the current sort column.
             * @a_dooverride
             */
            virtual int sortColumn() =0;

            /** \brief Returns the current sort order.
             * @a_dooverride
             */
            virtual Qt::SortOrder sortOrder() =0;

            /** \brief Returns the nodes which the user has selected in this fileview.
             * @a_dooverride
             */
            virtual QList<std::shared_ptr<sh::filesystem::FilesystemNode>> selectedNodes() =0;

            /** \brief Sets the node selection of this fileview.
             * @a_dooverride
             */
            virtual void setSelection(const QList<std::shared_ptr<sh::filesystem::FilesystemNode>> nodes) =0;

            /** \brief Returns a list of all nodes currently listed in this file view.
             * @a_dooverride
             */
            virtual QList<std::shared_ptr<sh::filesystem::FilesystemNode>> getAllVisibleNodes() =0;

            /** \brief Reloads the data inside this file view.
             */
            void reload(bool skipModel=false);

            /** \brief Returns the filesystem model for this view.
             * @a_infrastructure
             */
            sh::filesystem::FilesystemModelFileviewProxy* filemodel();

            /** \brief Sets the filesystem model for this view.
             * @a_infrastructure
             */
            void setFilemodel(sh::filesystem::FilesystemModelFileviewProxy* model);

        signals:

            /** \brief Triggered when the selection have changed.
             */
            void selectionChanged();

            /** \brief Triggered when some view options have changed.
             */
            void viewOptionChanged();

        private:
            sh::filesystem::FilesystemModelFileviewProxy* _model = nullptr;
            std::shared_ptr<QObject> _visibleviewslifetimecanary = nullptr;
            std::shared_ptr<sh::filesystem::FilesystemNode> _node = nullptr;

        };

    }
}

#endif // FILEVIEW_H
