/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2020 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "actions/actionactionitem.h"
#include "actions/actionsmanager.h"
#include "base/shallotprocess.h"
#include "base/threadpool.h"
#include "configuration/configurationmanager.h"
#include "filesystem/filesystemnode.h"
#include "filesystem/filesystemmodel.h"
#include "settings/settingsmanager.h"
#include "actionexecutioninfopanel.h"
#include "fileview.h"
#include "mainwindow.h"


std::shared_ptr<sh::configuration::ConfigurationValue> sh::ui::MainWindow::cfgvalFolderTreeIconSize =
        sh::configuration::ConfigurationManager::instance()->registerConfigValue(
            "sh::ui::MainWindow.FolderTreeIconSize", -1,
            sh::configuration::ConfigurationValue::valueTypeInteger(),
            QObject::tr("Icon size in folder tree"),
            sh::configuration::ConfigurationCategory::CategoryGUI, QString(),
            QObject::tr("For default size, set it to -1."));

std::shared_ptr<sh::configuration::ConfigurationValue> sh::ui::MainWindow::cfgvalToolbarLocation =
        sh::configuration::ConfigurationManager::instance()->registerConfigValue(
            "sh::ui::MainWindow.ToolbarLocation", "N",
            sh::configuration::ConfigurationValue::valueTypeString(),
            QObject::tr("Main toolbar location"),
            sh::configuration::ConfigurationCategory::CategoryGUI, QString(),
            QObject::tr("N, E, S or W."));

std::shared_ptr<sh::configuration::ConfigurationValue> sh::ui::MainWindow::cfgvalDetailPanelLocation =
        sh::configuration::ConfigurationManager::instance()->registerConfigValue(
            "sh::ui::MainWindow.DetailPanelLocation", "S",
            sh::configuration::ConfigurationValue::valueTypeString(),
            QObject::tr("Detail panel location"),
            sh::configuration::ConfigurationCategory::CategoryGUI, QString(),
            QObject::tr("N, E, S or W."));

std::shared_ptr<sh::configuration::ConfigurationValue> sh::ui::MainWindow::cfgvalDetailPanelDimensions =
        sh::configuration::ConfigurationManager::instance()->registerConfigValue(
            "sh::ui::MainWindow.DetailPanelDimensions", "-1",
            sh::configuration::ConfigurationValue::valueTypeString());

std::shared_ptr<sh::configuration::ConfigurationValue> sh::ui::MainWindow::cfgvalDetailPanelThumbnailDimensions =
        sh::configuration::ConfigurationManager::instance()->registerConfigValue(
            "sh::ui::MainWindow.DetailPanelThumbnailDimensions", "-1",
            sh::configuration::ConfigurationValue::valueTypeString());

sh::ui::MainWindow* sh::ui::MainWindow::_mainWindow = nullptr;

bool sh::ui::MainWindow::_mainWindow_runheadless = false;

sh::ui::MainWindow::KillHelperThread* sh::ui::MainWindow::killthread = nullptr;

void sh::ui::MainWindow::setMainWindow(MainWindow* mainWindow)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    if (isReady())
        return;
    if (mainWindow)
        _mainWindow = mainWindow;
    else
        _mainWindow_runheadless = true;
}

sh::ui::MainWindow* sh::ui::MainWindow::mainWindow()
{
    return _mainWindow;
}

bool sh::ui::MainWindow::isReady()
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    return _mainWindow || _mainWindow_runheadless;
}

bool sh::ui::MainWindow::runsHeadless()
{
    return _mainWindow_runheadless;
}

sh::ui::MainWindow::MainWindow()
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    _mainWindow = this;
    QObject::connect(sh::actions::ActionsManager::instance().get(), &actions::ActionsManager::runningActionsChanged, [=](){
        _refreshIsCloseable();
    });
}

void sh::ui::MainWindow::_refreshIsCloseable()
{
    QString iscloseable;
    if (!sh::actions::ActionsManager::instance()->runningActions().isEmpty())
        iscloseable = QObject::tr("Shallot cannot be closed while actions are running.");
    setCloseable(iscloseable.isNull(), iscloseable);
}

void sh::ui::MainWindow::_initialize(sh::base::SingletonInitializer *singletonInitializer)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    setTitlePattern("{dirname}");
    this->singletonInitializer = singletonInitializer;
    initialize();
}

void sh::ui::MainWindow::initialize()
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    setFileDetailsPanelVisible(true);
    jumpToEurl(sh::filesystem::Eurl::fromString(sh::base::ShallotProcess::instance()->initialWorkDirectory()));
}

QString sh::ui::MainWindow::_getWindowTitle()
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    QString pattern = titlePattern();
    std::shared_ptr<sh::filesystem::FilesystemNode> _node = currentDirectory();
    QString dir;
    QString dirname;
    if (!_node || _node == sh::filesystem::FilesystemModel::instance()->rootNode())
        dirname = dir = QObject::tr("(root)");
    else {
        dir = _node->eurl()->asString();
        dirname = _node->displayName();
    }
    QString t = "";
    while (!pattern.isEmpty()) {
        int a = pattern.indexOf("{");
        if (a > -1) {
            int b = pattern.indexOf("}", a);
            if (b > -1) {
                QString token = pattern.mid(a+1, b-a-1);
                t += pattern.left(a);
                if (token == "dir")
                    t += dir;
                else if (token == "dirname")
                    t += dirname;
                else
                    t += pattern.mid(a, b-a+1);
                pattern = pattern.mid(b+1);
                continue;
            }
        }
        t += pattern;
        break;
    }
    return t;
}

void sh::ui::MainWindow::_closeDirectly()
{
    _closeDirectly(singletonInitializer);
}

void sh::ui::MainWindow::_closeDirectly(sh::base::SingletonInitializer* singletonInitializer)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    killthread = new KillHelperThread();
    killthread->start();
    sh::exceptions::Exception::executeGuarded([&](){
        singletonInitializer->shutdown();
    }, sh::exceptions::ALL_ERRORS_KILL_SHALLOT);
    QApplication::instance()->exit();
}

bool sh::ui::MainWindow::closeApp()
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    QString msg;
    if (isCloseable(&msg)) {
        handleClosed();
        _closeDirectly();
        return true;
    }
    else {
        handleCloseAppRejected(msg);
        return false;
    }
}

bool sh::ui::MainWindow::isCloseable(QString* msg)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    if (msg)
        *msg = _isCloseable;
    return _isCloseable.isEmpty();
}

void sh::ui::MainWindow::handleClosed()
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    _closing = true;
}

bool sh::ui::MainWindow::isClosing()
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    return _closing;
}

QString sh::ui::MainWindow::uiMode()
{
    return sh::base::ShallotProcess::instance()->uiMode();
}

void sh::ui::MainWindow::jumpToEurl(std::shared_ptr<const sh::filesystem::Eurl> eurl)
{
    sh::base::ThreadPool::enqueueIfInMain([=](){
        QList<std::shared_ptr<sh::filesystem::FilesystemNode>> nodes = sh::filesystem::FilesystemModel::instance()->findNodesForEurl(eurl);
        if (!nodes.isEmpty()) {
            sh::base::MainThread::dispatcher()->invokeSync([&](){
                jumpToNode(nodes.first());
            });
        }
    });
}

void sh::ui::MainWindow::onCurrentDirectoryChanged(std::function<void()> fct, QObject* owner)
{
    if (!owner)
        owner = &myqobject;
    QObject::connect(&myqobject, &sh::ui::_MainWindow_HelperQObject::currentDirectoryChanged, owner, fct);
}

void sh::ui::MainWindow::onGlobalViewOptionsChanged(std::function<void()> fct, QObject* owner)
{
    if (!owner)
        owner = &myqobject;
    QObject::connect(&myqobject, &sh::ui::_MainWindow_HelperQObject::globalViewOptionsChanged, owner, fct);
}

void sh::ui::MainWindow::onCurrentProfileChanged(std::function<void()> fct, QObject* owner)
{
    if (!owner)
        owner = &myqobject;
    QObject::connect(&myqobject, &sh::ui::_MainWindow_HelperQObject::currentProfileChanged, owner, fct);
}

void sh::ui::MainWindow::onFileViewOptionsChanged(std::function<void(int)> fct, QObject* owner)
{
    if (!owner)
        owner = &myqobject;
    QObject::connect(&myqobject, &sh::ui::_MainWindow_HelperQObject::fileViewOptionsChanged, owner, fct);
}

void sh::ui::MainWindow::onCurrentFileViewChanged(std::function<void()> fct, QObject* owner)
{
    if (!owner)
        owner = &myqobject;
    QObject::connect(&myqobject, &sh::ui::_MainWindow_HelperQObject::currentFileViewChanged, owner, fct);
}

void sh::ui::MainWindow::onFileViewCountChanged(std::function<void()> fct, QObject* owner)
{
    if (!owner)
        owner = &myqobject;
    QObject::connect(&myqobject, &sh::ui::_MainWindow_HelperQObject::fileViewCountChanged, owner, fct);
}

void sh::ui::MainWindow::_emit_currentDirectoryChanged()
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    myqobject._emit_currentDirectoryChanged();
}

void sh::ui::MainWindow::_emit_globalViewOptionsChanged()
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    myqobject._emit_globalViewOptionsChanged();
}

void sh::ui::MainWindow::_emit_currentProfileChanged()
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    myqobject._emit_currentProfileChanged();
}

void sh::ui::MainWindow::_emit_fileViewOptionsChanged(int i)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    myqobject._emit_fileViewOptionsChanged(i);
}

void sh::ui::MainWindow::_emit_fileViewCountChanged()
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    myqobject._emit_fileViewCountChanged();
}

void sh::ui::MainWindow::_emit_currentFileViewChanged()
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    myqobject._emit_currentFileViewChanged();
}

sh::ui::MainWindow::~MainWindow()
{
    _mainWindow = 0;
}

void sh::ui::MainWindow::KillHelperThread::run()
{
    QThread::sleep(5*60);
    QCoreApplication::exit(128);
}

void sh::ui::_MainWindow_HelperQObject::_emit_currentDirectoryChanged()
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    emit currentDirectoryChanged();
}

void sh::ui::_MainWindow_HelperQObject::_emit_globalViewOptionsChanged()
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    emit globalViewOptionsChanged();
}

void sh::ui::_MainWindow_HelperQObject::_emit_currentProfileChanged()
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    emit currentProfileChanged();
}

void sh::ui::_MainWindow_HelperQObject::_emit_fileViewOptionsChanged(int i)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    emit fileViewOptionsChanged(i);
}

void sh::ui::_MainWindow_HelperQObject::_emit_fileViewCountChanged()
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    emit fileViewCountChanged();
}

void sh::ui::_MainWindow_HelperQObject::_emit_currentFileViewChanged()
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    emit currentFileViewChanged();
}

std::shared_ptr<sh::ui::ActionExecutionInfoPanel> sh::ui::MainWindow::showErrorPanel()
{
    std::shared_ptr<sh::ui::ActionExecutionInfoPanel> result;
    sh::base::MainThread::dispatcher()->invokeSync([&](){
        result = addInfoPanel(0, QColor(100,30,30));
        result->setProgress(false, 0, 0);
        result->setWidth(50);
        result->setLabel(QObject::tr("Problem"));
        result->onClicked([=](){
            sh::ui::MainWindow::mainWindow()->showLogViewDialog(QObject::tr("Some problems occurred.\n\nIf you encounter wrong program behavior, you should restart Shallot."),
                                                                QObject::tr("This can be a software error in Shallot or it is the fault of a plugin. If you are interested in technical details, please read the log."),
                                                                sh::base::LogSeverity::ERROR);
            _errorpanels.removeOne(result);
        });
    });
    _errorpanels.append(result);
    return result;
}

void sh::ui::MainWindow::activateNodes(QList<std::shared_ptr<sh::filesystem::FilesystemNode>> nodes, std::function<void(std::shared_ptr<sh::filesystem::FilesystemNode>)> directoryActivated)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    int nodetype = (nodes.length()==1) ? nodes.first()->nodetype() : sh::filesystem::FilesystemNodeType::NONE;
    if (nodetype == sh::filesystem::FilesystemNodeType::Directory || nodetype == sh::filesystem::FilesystemNodeType::SpecialTreeOnlyDirectory)
        directoryActivated(nodes[0]);
    else if (nodetype == sh::filesystem::FilesystemNodeType::Link && !nodes[0]->linkDestinationNodes().isEmpty() && nodes[0]->linkDestinationNodes().first()->nodetype() == sh::filesystem::FilesystemNodeType::Directory)
        directoryActivated(nodes[0]->linkDestinationNodes().first());
    else {
        auto instantiation = std::make_shared<sh::actions::ActionInstantiation>(nodes, currentDirectory());
        sh::actions::ActionsManager::instance()->getActions(instantiation, [&](QList<std::shared_ptr<sh::actions::ActionInstantiation> > selacts, QList<std::shared_ptr<sh::actions::ActionInstantiation>>){
            sh::base::MainThread::dispatcher()->invokeAsync([=](){
                QList<std::shared_ptr<sh::actions::AbstractActionItem>> acts;
                for(auto actsit : selacts)
                    acts.append(actsit->createdAction);
                sh::actions::ActionsManager::instance()->getDefaultAction(acts, [=](std::shared_ptr<sh::actions::ActionActionItem> action){
                    if (action)
                        action->execute();
                });

            });
        }, [](std::shared_ptr<sh::actions::ActionInstantiation>){
        });
    }
}

void sh::ui::MainWindow::setTreeVisible(bool v)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    if (v != _treeVisible) {
        _treeVisible = v;
        _emit_globalViewOptionsChanged();
    }
}

bool sh::ui::MainWindow::treeVisible()
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    return _treeVisible;
}

void sh::ui::MainWindow::setTreeSticky(bool v)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    if (v != _treeSticky) {
        _treeSticky = v;
        _emit_globalViewOptionsChanged();
    }
}

bool sh::ui::MainWindow::treeSticky()
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    return _treeSticky;
}

void sh::ui::MainWindow::setFileDetailsPanelVisible(bool v)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    if (v != _detailsPanelVisible)
    {
        _detailsPanelVisible = v;
        _emit_globalViewOptionsChanged();
    }
}

bool sh::ui::MainWindow::fileDetailsPanelVisible()
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    return _detailsPanelVisible;
}

void sh::ui::MainWindow::setTitlePattern(QString value)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    _titlePattern = value;
}

QString sh::ui::MainWindow::titlePattern()
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    return _titlePattern;
}

std::shared_ptr<sh::ui::ActionExecutionInfoPanel> sh::ui::MainWindow::addInfoPanel(sh::actions::ActionExecutionInfo* info, QColor color)
{
    return addInfoPanel(-1, info, color);
}

void sh::ui::MainWindow::setCurrentProfile(QString profile)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    _currentProfile = profile;
    sh::settings::SettingsManager::instance()->applyGlobalSettingsToMainWindow();
    sh::settings::SettingsManager::instance()->applyPerEurlSettingsToMainWindow();
    for(int i=0; i<fileViewCount(); i++) {
        sh::ui::FileView* viewctrl = getFileView(i);
        sh::settings::SettingsManager::instance()->applyGlobalSettingsToFileView(viewctrl);
        sh::settings::SettingsManager::instance()->applyPerEurlSettingsToFileView(viewctrl);
    }
    _emit_currentProfileChanged();
}

QString sh::ui::MainWindow::currentProfile()
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    return _currentProfile;
}

void sh::ui::MainWindow::reloadProfile()
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    setCurrentProfile(currentProfile());
}

void sh::ui::MainWindow::fileViewsReloadAll()
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    for (int il=0; il<fileViewCount(); il++) {
        sh::ui::FileView* list = getFileView(il);
        list->reload();
        for(int i=0; i<list->node()->childnodeCount(); i++)
            list->node()->childnode(i)->requestDetails();
    }
}

void sh::ui::MainWindow::setCloseable(bool closeable, QString message)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    _isCloseable = closeable ? QString() : (message.isEmpty() ? " " : message);
}


void sh::ui::MainWindow::requestDetailThumbnail(int width, int height, std::function<void()> onBeforeRequest, std::function<void(QIcon)> onArrived, bool force)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    if (_thumbnailrequestongoing && !force /* this is laziness; we could implement the same for forced requests as well */) {
        _thumbnailrequestfollowup = true;
        _thumbnailrequestfollowup_width = width;
        _thumbnailrequestfollowup_height = height;
        auto _c_thumbnailrequestfollowup_onBeforeRequest = _thumbnailrequestfollowup_onBeforeRequest;
        auto _c__thumbnailrequestfollowup_onArrived = _thumbnailrequestfollowup_onArrived;
        _thumbnailrequestfollowup_onBeforeRequest = [=](){
            if (_c_thumbnailrequestfollowup_onBeforeRequest)
                _c_thumbnailrequestfollowup_onBeforeRequest();
            onBeforeRequest();
        };
        _thumbnailrequestfollowup_onArrived = [=](QIcon result){
            if (_c__thumbnailrequestfollowup_onArrived)
                _c__thumbnailrequestfollowup_onArrived(result);
            onArrived(result);
        };
        return;
    }
    quint64 mythumbnailrequestid = ++_thumbnaillastrequestid;
    if (width > 0) {
        QList<std::shared_ptr<sh::filesystem::FilesystemNode>> selectednodes;
        if (currentFileView())
            selectednodes = currentFileView()->selectedNodes();
        if (selectednodes.length() == 1) {
            if (force || _thumbnaillastreqwidth != width || _thumbnaillastreqheight != height) {
                _thumbnaillastreqwidth = width;
                _thumbnaillastreqheight = height;
                _thumbnailrequestongoing = true;
                onBeforeRequest();
                sh::tools::ThumbnailManager::instance()->requestThumbnail(selectednodes.first(), width, height, [=](QIcon result){
                    sh::base::MainThread::dispatcher()->invokeSync([&](){
                        if (mythumbnailrequestid != _thumbnaillastrequestid)
                            return;
                        onArrived(result);
                        _thumbnailrequestongoing = false;
                        if (_thumbnailrequestfollowup) {
                            _thumbnailrequestfollowup = false;
                            requestDetailThumbnail(_thumbnailrequestfollowup_width, _thumbnailrequestfollowup_height, _thumbnailrequestfollowup_onBeforeRequest, _thumbnailrequestfollowup_onArrived, false);
                        }
                    });
                });
            }
        }
        else
            onArrived(QIcon());
    }
    else
        onArrived(QIcon());
}

void sh::ui::MainWindow::_cfgvalDetailPanelThumbnailDimensionsValueDimensions(int& width, int& height)
{
    width = -1;
    height = -1;
    QStringList lx = cfgvalDetailPanelThumbnailDimensions->getConfigValueString().split(";");
    if (lx.length() == 2) {
        bool ok = true;
        width = lx[0].toInt(&ok);
        if (!ok)
            width = -1;
        height = lx[1].toInt(&ok);
        if (!ok)
            height = -1;
    }
}
