/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2020 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef MAINWINDOW_H
#define MAINWINDOW_H

#include <functional>
#include <memory>
#include <QAction>
#include <QColor>
#include <QObject>
#include <QThread>
#include "base/logger.h"
#include "base/mainthread.h"
#include "base/singletoninitializer.h"
#include "tools/thumbnailmanager.h"
#include "dialog.h"


namespace Ui {
class QtMainWindow;
}

namespace sh{
    namespace actions { class ActionActionItem; class ActionExecutionInfo; }
    namespace filesystem{ class Eurl; class FilesystemNode;class FilesystemModelDirectoryTreeProxy; }
    namespace configuration{ class ConfigurationValue; }
    namespace exceptions{ class Exception; }
    namespace tools{
        namespace filetypes { class OpenMethod; }
        class SingletonInitializer;
    }

    namespace ui{ //TODO# add nicer type detail-column showing sth like "Image" or "Video"
        class AboutDialog;
        class ActionExecutionInfoDialog;
        class ActionExecutionInfoPanel;
        class Dialog;
        class ExceptionDialog;
        class FilePropertyDialog;
        class FileView;
        class LogViewDialog;
        class ManageBookmarksDialog;
        class ManageProfilesDialog;
        class OpenWithDialog;
        class StoreProfileDialog;
        class TuningDialog;

        /** \brief A helper for some signal/slot stuff of in MainWindow.
         */
        class _MainWindow_HelperQObject : public QObject
        {
            Q_OBJECT
        public:
            void _emit_currentDirectoryChanged();
            void _emit_globalViewOptionsChanged();
            void _emit_currentProfileChanged();
            void _emit_currentFileViewChanged();
            void _emit_fileViewCountChanged();
            void _emit_fileViewOptionsChanged(int i);
        signals:
            void currentDirectoryChanged();
            void globalViewOptionsChanged();
            void currentProfileChanged();
            void currentFileViewChanged();
            void fileViewCountChanged();
            void fileViewOptionsChanged(int);
        };

        /** \brief
         * The abstract main window class. There is one instance of it, and it is also used for other ui parts (e.g. creating some dialogs, ...).
         */
        class MainWindow
        {
            friend class sh::exceptions::Exception;

        public:
            explicit MainWindow();
            virtual ~MainWindow();

            /** \brief
             * Returns the instance of this singleton.
             */
            static MainWindow* mainWindow();

            /** \brief
             * Sets the main window instance.
             * @a_infrastructure
             */
            static void setMainWindow(MainWindow* mainWindow);

            /** \brief
             * Returns if this process ui is ready (i.e. is created or runs headless).
             */
            static bool isReady();

            /** \brief
             * Returns if this process runs headless, i.e. without any ui, just for a background process.
             */
            static bool runsHeadless();

            /** \brief Initializes the main window. For custom initialization logic, see MainWindow.initialize.
             * @a_infrastructure
             */
            void _initialize(sh::base::SingletonInitializer* singletonInitializer);

            /** \brief Returns the current number of file views.
             * @a_dooverride
             */
            virtual int fileViewCount() = 0;

            /** \brief Adds a new file view.
             * @a_dooverride
             */
            virtual void addFileView(bool reinitialize=false) = 0;

            /** \brief Removes a file view.
             * @a_dooverride
             */
            virtual void removeFileView(int i) = 0;

            /** \brief Returns the file view which is currently active (i.e. focussed).
             * @a_dooverride
             */
            virtual sh::ui::FileView* currentFileView() = 0;

            /** \brief Returns a file view by position index.
             * @a_dooverride
             */
            virtual sh::ui::FileView* getFileView(int i) = 0;

            /** \brief Reload all content in each file view.
             */
            void fileViewsReloadAll();

            /** \brief Sets a handler for some view options in a file view changed (optionally bound to an owner lifetime).
             */
            void onFileViewOptionsChanged(std::function<void(int)> fct, QObject* owner=0);

            /** \brief Sets a handler for the currently active file view changed (optionally bound to an owner lifetime).
             */
            void onCurrentFileViewChanged(std::function<void()> fct, QObject* owner=0);

            /** \brief Sets a handler for the number of file views changed (optionally bound to an owner lifetime).
             */
            void onFileViewCountChanged(std::function<void()> fct, QObject* owner=0);

            /** \brief Sets a handler for the current directory in the active file view changed (optionally bound to an owner lifetime).
             */
            void onCurrentDirectoryChanged(std::function<void()> fct, QObject* owner=0);

            /** \brief Sets a handler for some global view options changed (optionally bound to an owner lifetime).
             */
            void onGlobalViewOptionsChanged(std::function<void()> fct, QObject* owner=0);

            /** \brief Sets a handler for the current profile changed (optionally bound to an owner lifetime).
             */
            void onCurrentProfileChanged(std::function<void()> fct, QObject* owner=0);

            /** \brief
             * Let the current view jump to another location.
             * @a_dooverride_optional
             */
            virtual void jumpToEurl(std::shared_ptr<const sh::filesystem::Eurl> eurl);

            /** \brief
             * Let the current view jump to another location.
             * @a_dooverride
             */
            virtual void jumpToNode(std::shared_ptr<sh::filesystem::FilesystemNode> node) = 0;

            /** \brief Sets the window title pattern.
             * @a_dooverride
             */
            virtual void setTitlePattern(QString value);

            /** \brief Returns the window title pattern.
             */
            QString titlePattern();

            /** \brief Sets the current profile.
             * @a_dooverride_optional
             */
            virtual void setCurrentProfile(QString profile);

            /** \brief Returns the current profile.
             */
            QString currentProfile();

            /** \brief Reloads the profile data.
             */
            void reloadProfile();

            /** \brief Sets the visibility of the directory tree.
             * @a_dooverride
             */
            virtual void setTreeVisible(bool v);

            /** \brief Returns the visibility of the directory tree.
             */
            bool treeVisible();
            /** \brief Sets if the directory tree should follow the active file view.
             * @a_dooverride
             */
            virtual void setTreeSticky(bool v);

            /** \brief Returns if the directory tree follows the active file view.
             */
            bool treeSticky();

            /** \brief Sets the visibility of the file details panel.
             * @a_dooverride
             */
            virtual void setFileDetailsPanelVisible(bool v);

            /** \brief Returns the visibility of the file details panel.
             */
            bool fileDetailsPanelVisible();

            /** \brief
             * Gets the sh::filesystem::FilesystemNode selected in the current view.
             * @a_dooverride
             */
            virtual std::shared_ptr<sh::filesystem::FilesystemNode> currentDirectory() = 0;

            /** \brief Creates a new action execution panel.
             *
             * Let this smart pointer die for removing it.
             * @a_dooverride
             */
            virtual std::shared_ptr<sh::ui::ActionExecutionInfoPanel> addInfoPanel(int position, sh::actions::ActionExecutionInfo* info, QColor color=QColor()) = 0;

            /** \brief Creates a new action execution panel.
             *
             * Let this smart pointer die for removing it.
             * @a_dooverride
             */
            std::shared_ptr<sh::ui::ActionExecutionInfoPanel> addInfoPanel(sh::actions::ActionExecutionInfo* info, QColor color=QColor());

            /** \brief Sets the jumpbar to text input mode.
             * @a_dooverride
             */
            virtual void jumpbarSetTextMode() = 0;

            /** \brief Sets the jumpbar to button mode.
             * @a_dooverride
             */
            virtual void jumpbarSetButtonMode() = 0;

            /** \brief Returns if the jumpbar is in text input mode.
             * @a_dooverride
             */
            virtual bool jumpbarIsTextMode() = 0;

            /** \brief Returns the current toolbar position.
             * @a_dooverride
             */
            virtual QString toolbarPosition() = 0;

            /** \brief Sets the toolbar position.
             * @a_dooverride
             */
            virtual void setToolbarPosition(QString pos) = 0;

            /** \brief Returns the current detail panel position.
             * @a_dooverride
             */
            virtual QString detailPanelPosition() = 0;

            /** \brief Sets the detail panel position.
             * @a_dooverride
             */
            virtual void setDetailPanelPosition(QString pos) = 0;

            /** \brief Shows and returns an 'About' dialog.
             * @a_dooverride
             */
            virtual std::shared_ptr<AboutDialog> showAboutDialog() = 0;

            /** \brief Shows and returns a 'Manage Saved Settings' dialog.
             * @a_dooverride
             */
            virtual std::shared_ptr<ManageProfilesDialog> showManageProfilesDialog() = 0;

            /** \brief Shows and returns a 'Open With' dialog.
             * @a_dooverride
             */
            virtual std::shared_ptr<OpenWithDialog> showOpenWithDialog(std::shared_ptr<sh::filesystem::FilesystemNode> node, QList<std::shared_ptr<sh::tools::filetypes::OpenMethod>> openMethods) = 0;

            /** \brief Shows and returns a 'Save Settings' dialog.
             * @a_dooverride
             */
            virtual std::shared_ptr<StoreProfileDialog> showStoreProfileDialog(std::shared_ptr<const sh::filesystem::Eurl> eurl) = 0;

            /** \brief Shows and returns a 'Tuning' dialog.
             * @a_dooverride
             */
            virtual std::shared_ptr<TuningDialog> showTuningDialog() = 0;

            /** \brief Shows and returns a 'Log' dialog.
             * @a_dooverride
             */
            virtual std::shared_ptr<LogViewDialog> showLogViewDialog(QString headtext=QString(), QString subheadtxt=QString(), sh::base::LogSeverity minseverity=sh::base::LogSeverity::_DEFAULT) = 0;

            /** \brief Shows and returns a 'Manage Bookmarks' dialog.
             * @a_dooverride
             */
            virtual std::shared_ptr<ManageBookmarksDialog> showManageBookmarksDialog() = 0;

            /** \brief Shows and returns a 'File Properties' dialog.
             * @a_dooverride
             */
            virtual std::shared_ptr<FilePropertyDialog> showFilePropertyDialog(QList<std::shared_ptr<sh::filesystem::FilesystemNode>> nodes) = 0;

            /** \brief Creates an action execution dialog.
             * @a_dooverride
             */
            virtual std::shared_ptr<ActionExecutionInfoDialog> createActionExecutionInfoDialog(sh::actions::ActionExecutionInfo* info) = 0;

            /** \brief Shows an error panel.
             */
            virtual std::shared_ptr<ActionExecutionInfoPanel> showErrorPanel();

            /** \brief Returns the DialogManager which creates and drives `Dialog`s for this main window.
             *
             * You typically should not need to use it directly.
             */
            virtual std::shared_ptr<DialogManager> dialogManager() =0;

            /** \brief Directly begin application shutdown without checks.
             */
            void _closeDirectly();
            static void _closeDirectly(sh::base::SingletonInitializer* singletonInitializer);

            /** \brief Check if the application may shut down now, and if so, initiate it.
             */
            bool closeApp();

            /** \brief Check if the application may shut down now.
             *
             * @a_dooverride_optional
             */
            bool isCloseable(QString* msg=nullptr);

            /** \brief React on a rejected application close request (with some feedback message).
             * @a_dooverride
             */
            virtual void handleCloseAppRejected(QString rejectmsg) = 0;

            /** \brief Do some cleanup steps just when closing starts.
             * @a_dooverride_optional
             * Implementations must call base implementation!
             */
            virtual void handleClosed();

            /** \brief Returns the UI mode (e.g. qt, web).
             */
            static QString uiMode();

            /** \brief Returns if the application is currently closing.
             */
            bool isClosing();

        protected:
            static std::shared_ptr<sh::configuration::ConfigurationValue> cfgvalFolderTreeIconSize;
            static std::shared_ptr<sh::configuration::ConfigurationValue> cfgvalToolbarLocation;
            static std::shared_ptr<sh::configuration::ConfigurationValue> cfgvalDetailPanelLocation;
            static std::shared_ptr<sh::configuration::ConfigurationValue> cfgvalDetailPanelDimensions;
            static std::shared_ptr<sh::configuration::ConfigurationValue> cfgvalDetailPanelThumbnailDimensions;
            sh::base::SingletonInitializer* singletonInitializer;
            _MainWindow_HelperQObject myqobject;

            /** \brief Used by main window implementations for getting detail panel thumbnail dimensions by the config string value.
             */
            void _cfgvalDetailPanelThumbnailDimensionsValueDimensions(int& width, int& height);

            /** \brief Called by an implementation when the current directory in the active file view changed.
             */
            void _emit_currentDirectoryChanged();

            /** \brief Called by an implementation when some global view options have changed.
             */
            void _emit_globalViewOptionsChanged();

            /** \brief Called by an implementation when the current profile has changed.
             */
            void _emit_currentProfileChanged();

            /** \brief Called by an implementation when some view options of a file view have changed.
             */
            void _emit_fileViewOptionsChanged(int i);

            /** \brief Called by an implementation when the number of file views has changed.
             */
            void _emit_fileViewCountChanged();

            /** \brief Called by an implementation when the currently active file view has changed.
             */
            void _emit_currentFileViewChanged();

            /** \brief Initializes this main window.
             * @a_dooverride
             */
            virtual void initialize();

            /** \brief Shows and returns an 'Exception' dialog.
             * @a_dooverride
             */
            virtual std::shared_ptr<ExceptionDialog> showExceptionDialog(QString error1, QString error2, QString details, QString icon, bool mayRetry, bool mayClose, bool mayCancel, bool showLoglabelAndDetails) = 0;

            /** \brief Generates the current window title from a pattern.
             * Used by implementations for setting its title from a given pattern.
             */
            QString _getWindowTitle();

            /** \brief Can be used by implementations for activating (i.e. 'reacting on double-click') nodes.
             */
            void activateNodes(QList<std::shared_ptr<sh::filesystem::FilesystemNode>> nodes, std::function<void(std::shared_ptr<sh::filesystem::FilesystemNode>)> directoryActivated);

            /** \brief Sets the closeable state and message.
             *
             * @a_infrastructure
             * @a_dooverride_optional
             */
            virtual void setCloseable(bool closeable, QString message=QString());

            /**
             * @brief Convenience method used by a main window implementation for requesting a new thumbnail in the detail panel.
             */
            void requestDetailThumbnail(int width, int height, std::function<void()> onBeforeRequest, std::function<void(QIcon)> onArrived, bool force);

        private:

            /** \brief Helps application shutdown in some situations :).
             */
            class KillHelperThread : public QThread
            {
            public:
                void run();
            };

            void _refreshIsCloseable();
            static MainWindow* _mainWindow;
            static bool _mainWindow_runheadless;
            static KillHelperThread* killthread;
            bool _closing = false;
            bool _treeSticky = true;
            bool _treeVisible = true;
            bool _detailsPanelVisible = true;
            QString _titlePattern;
            QList<std::shared_ptr<sh::ui::ActionExecutionInfoPanel>> _errorpanels;
            QString _currentProfile;
            QString _isCloseable;
            bool _thumbnailrequestongoing = false;
            bool _thumbnailrequestfollowup = false;
            int _thumbnailrequestfollowup_width = 0;
            int _thumbnailrequestfollowup_height = 0;
            std::function<void()> _thumbnailrequestfollowup_onBeforeRequest = nullptr;
            std::function<void(QIcon)> _thumbnailrequestfollowup_onArrived = nullptr;
            int _thumbnaillastreqwidth = 0;
            int _thumbnaillastreqheight = 0;
            quint64 _thumbnaillastrequestid = 0;

        };

    }
}

#define REGISTER_UI_MAINWINDOW(NS, STYPE) \
    namespace sh { \
        /** \brief Internal helper class for main window creation for NS::STYPE. */ \
        class _ShMainWindow_ ## STYPE : public sh::base::Singleton { \
            DECLARE_SINGLETON(_ShMainWindow_ ## STYPE) \
            void doInitialize() override { \
                if (!sh::ui::MainWindow::isReady()) { \
                    sh::ui::MainWindow* w = nullptr; \
                    if (NS::STYPE::tryCreateMainWindow(w)) \
                        sh::ui::MainWindow::setMainWindow(w); \
                } \
            }; \
            void doShutdown() override {  }; \
        }; \
    } \
    REGISTER_SINGLETON(sh, _ShMainWindow_ ## STYPE, "mainwindow")

#endif // MAINWINDOW_H


//TODO # poetschhttp gallery view
//TODO #zz rich geo controls
/*TODO # cooler searches
  - new api must allow mixing combinations of parsleymetadatainterpreter criteria
  - must be easy:
   - fulltext/everything (incl. comment and tags, filename)
   - comment
   - tags
   - geo via center/distance (fully graphically)
   - filename
   - file type (also simplied; from ['audio', 'video', 'text', 'graphics', ...])
  - must be possible:
   - rating
   - and/or/not
*/
//TODO # (all shallot) apidocs should have threading information when it makes sense

/*TODO #zz restricting shallot
  - via hierarchical Configuration
   - which actions to show in menus
   - which detail columns to show
   - which file property dialog tabs to show
  - also via manager method
   - later on even scripting api
*/

/*TODO #zz filesystem restricted by root path
  - hard access restriction to a subdirectory (for poetschhttp2)
   - are there bypasses?
*/
