/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2020 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <QDesktopWidget>
#include "actions/actionexecutioninfo.h"
#include "base/mainthread.h"
#include "base/iconmanager.h"
#include "ui/qt/feedbackpanels/msgbox.h"
#include "ui/qt/feedbackpanels/gridform.h"
#include "ui/qt/feedbackpanels/credentials.h"
#include "ui/qt/feedbackpanels/unixpermissions.h"
#include "ui/qt/feedbackpanels/gridforminnersimplechooser.h"
#include "ui/qt/feedbackpanels/feedbackpanel.h"
#include "qtmainwindow.h"
#include "qtuistyle.h"
#include "qtactionexecutioninfodialog.h"
#include "ui_qtactionexecutioninfodialog.h"


sh::ui::qt::QtActionExecutionInfoDialog::QtActionExecutionInfoDialog(sh::actions::ActionExecutionInfo* info, QWidget *parent) :
    QDialog(parent), sh::ui::ActionExecutionInfoDialog(info), ui(new Ui::QtActionExecutionInfoDialog)
{
    ui->setupUi(this);
    ui->btnCancel->setSizeByFactor(100);
    ui->btnCancel->setIcon(sh::base::IconManager::instance()->getIcon("dialog-close"));
    ui->btnBackground->setSizeByFactor(100);
    ui->btnBackground->setIcon(sh::base::IconManager::instance()->getIcon("dialog-close"));
    auto uistyle = sh::ui::qt::QtMainWindow::mainWindow()->uiStyle();
    uistyle->createSheet().background_liketitlebar().applyTo(ui->panelHead);
    uistyle->createSheet().fontsize_header().applyTo(ui->label);
    uistyle->createSheet().textcolor_lighter1().applyTo(ui->labelfromverb).applyTo(ui->labeltoverb);
    uistyle->createSheet().background_highlighted().applyTo(ui->panelFeedback);
    ui->panelFeedback->setVisible(false);
    setWindowFlags( (windowFlags() | Qt::CustomizeWindowHint) & ~Qt::WindowCloseButtonHint);
}

sh::ui::qt::QtActionExecutionInfoDialog::~QtActionExecutionInfoDialog()
{
    delete ui;
}

void sh::ui::qt::QtActionExecutionInfoDialog::keyPressEvent(QKeyEvent *e)
{
    if (e->key()==Qt::Key_Escape && currentFeedbackPanel)
        currentFeedbackPanel->cancelRequested();
    else
        QDialog::keyPressEvent(e);
}

void sh::ui::qt::QtActionExecutionInfoDialog::reject()
{
    // ignore reject
}

void sh::ui::qt::QtActionExecutionInfoDialog::setDetails(QString fv, QString fob, QString tv, QString tob)
{
    QFontMetrics fmetrics(ui->labelfromobj->font());
    QString elidedfromObj = fmetrics.elidedText(fob, Qt::ElideMiddle, ui->labelfromobj->width());
    QFontMetrics tmetrics(ui->labeltoobj->font());
    QString elidedtoObj = tmetrics.elidedText(tob, Qt::ElideMiddle, ui->labeltoobj->width());
    ui->labelfromobj->setText( elidedfromObj );
    ui->labelfromverb->setText( fv );
    ui->labeltoobj->setText( elidedtoObj );
    ui->labeltoverb->setText( tv );
}

void sh::ui::qt::QtActionExecutionInfoDialog::setHead(QString txt)
{
    ui->label->setText(txt);
}

void sh::ui::qt::QtActionExecutionInfoDialog::setProgress(bool isDeterminate, quint64 done, quint64 all, QString text)
{
    if (isDeterminate)
    {
        if (ui->progressBar->maximum() != 1000)
            ui->progressBar->setMaximum(1000);
        double dpval = ((double)done) / ((double)all);
        int pval = (int)(dpval * 1000);
        ui->progressBar->setValue( pval );
    }
    else
    {
        if (ui->progressBar->maximum() != 0)
            ui->progressBar->setMaximum(0);
    }
    ui->labelestimation->setText(text);
}

void sh::ui::qt::QtActionExecutionInfoDialog::askFeedbackViaWidget(sh::ui::qt::feedbackpanels::FeedbackPanel* widget)
{
    bool manualInNeeded = false;
    sh::base::MainThread::dispatcher()->invokeSync([&](){
        if (currentFeedbackPanel)
            throw sh::exceptions::ProgramException(sh::exceptions::ExceptionData().details("Nested feedback panels are not supported.").resumeable(true));
        currentFeedbackPanel = widget;
        widget->setParent(ui->panelFeedback);
        ui->panelFeedback->layout()->addWidget(widget);
        ui->panelFeedback->setVisible(true);
        ui->progressBar->setVisible(false);
        ui->labelestimation->setVisible(false);
        widget->setVisible(true);
        manualInNeeded = _info->isManualInterventionNeeded();
        _info->setManualInterventionNeeded(true);
        int prefHeight = widget->preferredHeight();
        if (prefHeight != -1) {
            int outerHeight = height() - widget->height();
            int screenHeight = QApplication::desktop()->screenGeometry().height();
            int avail = screenHeight - outerHeight - 100;
            prefHeight *= 1.3;
            int useheight = (avail<prefHeight)?avail:prefHeight;
            int useouterheight = useheight + outerHeight;
            if (height() < useouterheight)
                resize(width(), useouterheight);
        }
        show();
    });
    widget->waitUntilClosed();
    sh::base::MainThread::dispatcher()->invokeSync([&](){
        _info->setManualInterventionNeeded(manualInNeeded);
        widget->setParent(ui->panelFeedback);
        ui->panelFeedback->setVisible(false);
        ui->progressBar->setVisible(true);
        ui->labelestimation->setVisible(true);
        currentFeedbackPanel = 0;
    });
}

int sh::ui::qt::QtActionExecutionInfoDialog::messageBox(QString text, QList<QString> answers, QString icon, int defaultanswer, int cancelanswer, QList<QString> answericons)
{
    sh::ui::qt::feedbackpanels::MsgBox *msgbox = 0;
    sh::base::MainThread::dispatcher()->invokeSync([&](){
        msgbox = new sh::ui::qt::feedbackpanels::MsgBox(text, answers, icon, 0, false, defaultanswer, cancelanswer, -1, -1, answericons);
    });
    askFeedbackViaWidget(msgbox);
    int result = msgbox->answer;
    sh::base::MainThread::dispatcher()->invokeSync([&](){
        delete msgbox;
    });
    return result;
}

int sh::ui::qt::QtActionExecutionInfoDialog::inputBox(QString text, QList<QString> answers, QString *value, QString icon, int defaultanswer, int cancelanswer, int valuePreselectFrom, int valuePreselectTo)
{
    sh::ui::qt::feedbackpanels::MsgBox *msgbox = 0;
    sh::base::MainThread::dispatcher()->invokeSync([&](){
        msgbox = new sh::ui::qt::feedbackpanels::MsgBox(text, answers, icon, *value, false, defaultanswer, cancelanswer, valuePreselectFrom, valuePreselectTo);
    });
    askFeedbackViaWidget(msgbox);
    *value = msgbox->text;
    int result = msgbox->answer;
    sh::base::MainThread::dispatcher()->invokeSync([&](){
        delete msgbox;
    });
    return result;
}

int sh::ui::qt::QtActionExecutionInfoDialog::multilineInputBox(QString text, QList<QString> answers, QString *value, QString icon, int defaultanswer, int cancelanswer)
{
    sh::ui::qt::feedbackpanels::MsgBox *msgbox = 0;
    sh::base::MainThread::dispatcher()->invokeSync([&](){
        msgbox = new sh::ui::qt::feedbackpanels::MsgBox(text, answers, icon, *value, true, defaultanswer, cancelanswer);
    });
    askFeedbackViaWidget(msgbox);
    *value = msgbox->text;
    int result = msgbox->answer;
    sh::base::MainThread::dispatcher()->invokeSync([&](){
        delete msgbox;
    });
    return result;
}

int sh::ui::qt::QtActionExecutionInfoDialog::gridform(QString text, QList<QPair<QString, QWidget *> > form, QStringList answers, int defaultanswer, int cancelanswer)
{
    sh::ui::qt::feedbackpanels::GridForm *gridform = 0;
    sh::base::MainThread::dispatcher()->invokeSync([&](){
        gridform = new sh::ui::qt::feedbackpanels::GridForm(text, form, answers, defaultanswer, cancelanswer);
    });
    askFeedbackViaWidget(gridform);
    int result = gridform->answer;
    sh::base::MainThread::dispatcher()->invokeSync([&](){
        delete gridform;
    });
    return result;
}

int sh::ui::qt::QtActionExecutionInfoDialog::simpleChooserGridform(QString text, sh::actions::ActionExecutionUserFeedback::GridformEntries *entries, QStringList answers, int defaultanswer, int cancelanswer)
{
    QList<QPair<QString, QWidget*>> form;
    sh::base::MainThread::dispatcher()->invokeSync([&](){
        foreach(sh::actions::ActionExecutionUserFeedback::GridformEntry* entry, entries->list)
        {
            auto w = new sh::ui::qt::feedbackpanels::GridFormInnerSimpleChooser(&entry->choices, this);
            QPair<QString, QWidget*> row(entry->label, w);
            form.append(row);
        }
    });
    int result = gridform(text, form, answers, defaultanswer, cancelanswer);
    return result;
}

void sh::ui::qt::QtActionExecutionInfoDialog::on_btnCancel_clicked()
{
    _info->cancel();
}

void sh::ui::qt::QtActionExecutionInfoDialog::on_btnBackground_clicked()
{
    setBackground(true);
}

void sh::ui::qt::QtActionExecutionInfoDialog::setLogicallyVisible(bool v)
{
    sh::ui::ActionExecutionInfoDialog::setLogicallyVisible(v);
    _computevisibility();
}

void sh::ui::qt::QtActionExecutionInfoDialog::_computevisibility()
{
    if ((_logicallyVisible && !_isbackground) || _isforeground_forced)
        show();
    else
        hide();
}

void sh::ui::qt::QtActionExecutionInfoDialog::setBackground(bool v)
{
    if (!_isforeground_forced)
    {
        sh::ui::ActionExecutionInfoDialog::setBackground(v);
        _computevisibility();
    }
}

void sh::ui::qt::QtActionExecutionInfoDialog::setForceForeground(bool v)
{
    sh::ui::ActionExecutionInfoDialog::setForceForeground(v);
    ui->panelButtons->setVisible(!v);
    _computevisibility();
}

bool sh::ui::qt::QtActionExecutionInfoDialog::credentialsDialog(QString text, bool showDomain, bool showUsername, bool showPassword, bool showAnonymous, bool showRemember, QString* domain, QString* username, QString* password, bool *isAnonymous, bool* isRemember)
{
    sh::ui::qt::feedbackpanels::Credentials *form = 0;
    sh::base::MainThread::dispatcher()->invokeSync([&](){
        form = new sh::ui::qt::feedbackpanels::Credentials(*domain, *username, *password, text, showDomain, showUsername, showPassword, showAnonymous, showRemember);
    });
    askFeedbackViaWidget(form);
    bool result = form->accepted;
    *domain = form->domain;
    *username = form->username;
    *password = form->password;
    *isAnonymous = form->anonymous;
    *isRemember = form->remember;
    sh::base::MainThread::dispatcher()->invokeSync([&](){
        delete form;
    });
    return result;
}

bool sh::ui::qt::QtActionExecutionInfoDialog::unixPermissionsDialog(bool* userMayRead, bool* userMayWrite, bool* userMayExecute, bool* groupMayRead, bool* groupMayWrite, bool* groupMayExecute, bool* othersMayRead, bool* othersMayWrite, bool* othersMayExecute, bool* sticky, bool* setuid, bool* setgid, QStringList users, QStringList groups, QString* ownerUser, QString* ownerGroup)
{
    sh::ui::qt::feedbackpanels::UnixPermissions *form = 0;
    sh::base::MainThread::dispatcher()->invokeSync([&](){
        form = new sh::ui::qt::feedbackpanels::UnixPermissions();
        form->setflags(*userMayRead, *userMayWrite, *userMayExecute, *groupMayRead, *groupMayWrite, *groupMayExecute, *othersMayRead, *othersMayWrite, *othersMayExecute, *sticky, *setuid, *setgid);
        form->setusers(users, *ownerUser);
        form->setgroups(groups, *ownerGroup);
    });
    askFeedbackViaWidget(form);
    bool result = form->accepted;
    *userMayRead = form->userMayRead;
    *userMayWrite = form->userMayWrite;
    *userMayExecute = form->userMayExecute;
    *groupMayRead = form->groupMayRead;
    *groupMayWrite = form->groupMayWrite;
    *groupMayExecute = form->groupMayExecute;
    *othersMayRead = form->othersMayRead;
    *othersMayWrite = form->othersMayWrite;
    *othersMayExecute = form->othersMayExecute;
    *sticky = form->sticky;
    *setuid = form->setuid;
    *setgid = form->setgid;
    *ownerUser = form->user;
    *ownerGroup = form->group;
    sh::base::MainThread::dispatcher()->invokeSync([&](){
        delete form;
    });
    return result;
}
