/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2020 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "base/shallotprocess.h"
#include "qtactionmenu.h"
#include "qtmainwindow.h"


sh::ui::qt::QtActionMenu::QtActionMenu(QWidget* parent) : QMenu(parent)
{
    _boldfont = _normalfont = font();
    _boldfont.setBold(true);
    QColor c = sh::base::ShallotProcess::instance()->brandingColor();
    brandingcolor = QColor::fromHsl(c.hue(), c.saturation(), 30);
    QLabel* _buddy = new QLabel(this);  // TODO # this is made a bit weird due to a qt bug (completely clearing a visible menu makes it disappear)
    _buddy->setText("");
    _buddy->setFixedHeight(1);
    QWidgetAction* wa = new QWidgetAction(this);
    wa->setDefaultWidget(_buddy);
    addAction(wa);
    if (!_header.isEmpty()) {
        QAction* hact = addAction(_header);
        hact->setEnabled(false);
        QFont hdfont = hact->font();
        hdfont.setItalic(true);
        hdfont.setPointSizeF(hdfont.pointSizeF()*1.1);
        hact->setFont(hdfont);
    }
    _cntInternalActions = actions().count();
}

void sh::ui::qt::QtActionMenu::paintEvent(QPaintEvent *e)
{
    QMenu::paintEvent(e);
    QPainter painter(this);
    QList<QAction*> lstactions = actions();
    int actioncount = lstactions.count();
    for (int iaction=0; iaction<actioncount; iaction++) {
        QAction* action = lstactions[iaction];
        if (action->font().italic()) {  // it's a header
            QRect r = this->actionGeometry(action);
            int iactiona = iaction;
            QRect ra;
            while (ra.isNull()) {
                iactiona++;
                QAction* actiona = lstactions.value(iactiona);
                if (!actiona || !actiona->isSeparator())
                    break;  // just to be sure
                ra = this->actionGeometry(actiona);
            }
            int iactionb = iaction;
            QRect rb;
            while (rb.isEmpty()) {
                iactionb--;
                QAction* actionb = lstactions.value(iactionb);
                if (!actionb || !actionb->isSeparator())
                    break;  // just to be sure
                rb = this->actionGeometry(actionb);
            }
            r.adjust(0, -rb.height()/(rb.top()>5?2:1), 0, ra.height()/2);
            QImage back(r.size(), QImage::Format_ARGB32_Premultiplied);
            back.fill(brandingcolor);
            painter.setCompositionMode(QPainter::RasterOp_NotDestination);
            painter.drawImage(r.x(), r.y(), back);
            painter.setCompositionMode(QPainter::CompositionMode_Screen);
            painter.drawImage(r.x(), r.y(), back);
        }
    }
}

void sh::ui::qt::QtActionMenu::setHeader(QString t)
{
    _header = t;
}

QList<QAction*> sh::ui::qt::QtActionMenu::allActions()
{
    QList<QAction*> res = actions();
    for(int r=0; r<_cntInternalActions; r++)
        res.removeFirst();
    return res;
}

void sh::ui::qt::QtActionMenu::clearAllActions()
{
    int aacnt = allActions().count();
    for (int i=aacnt-1; i>=0; i--)
        removeActionAt(i);
}

void sh::ui::qt::QtActionMenu::removeActionAt(int i)
{
    int j = i + _cntInternalActions;
    QAction* ra = actions().at(j);
    removeAction(ra);
}

QAction* sh::ui::qt::QtActionMenu::addNewAction(int i)
{
    QAction* result;
    int j = i + _cntInternalActions;
    if (i==-1 || j>=actions().count())
        result = addAction("");
    else {
        QAction* a = new QAction(this);
        insertAction(actions().at(j), a);
        result = a;
    }
    _observeAction(result);
    return result;
}

QAction* sh::ui::qt::QtActionMenu::addNewSubMenu(sh::ui::qt::QtActionMenu** qsubmenu, int i)
{
    QAction* result;
    int j = i + _cntInternalActions;
    *qsubmenu = new sh::ui::qt::QtActionMenu(this);
    if (i==-1 || j>=actions().count())
        result = addMenu(*qsubmenu);
    else {
        QAction* a = new QAction(this);
        insertMenu(actions().at(j), *qsubmenu);
        result = a;
    }
    _observeAction(result);
    return result;
}

QAction* sh::ui::qt::QtActionMenu::addNewSeparator(int i)
{
    QAction* result;
    int j = i + _cntInternalActions;
    if (i==-1 || j>=actions().count())
        result = addSeparator();
    else {
        QAction* a = new QAction(this);
        insertSeparator(actions().at(j));
        result = a;
    }
    _observeAction(result);
    return result;
}

QAction* sh::ui::qt::QtActionMenu::addNewHeader(int i)
{
    QAction* result = addNewAction(i);
    result->setEnabled(false);
    QFont hdfont = result->font();
    hdfont.setItalic(true);
    hdfont.setBold(true);
    hdfont.setPointSizeF(hdfont.pointSizeF()*1.1);
    result->setFont(hdfont);
    result->setData("is_header");
    return result;
}

bool sh::ui::qt::QtActionMenu::actionIsHeader(QAction* a)
{
    return a->data().toString() == "is_header";
}

void sh::ui::qt::QtActionMenu::_observeAction(QAction* a)
{
    connect(a, &QAction::changed, this, [=](){
        _correctMenuPosition();
    });
}

void sh::ui::qt::QtActionMenu::_correctMenuPosition()
{
    if (_origMenuPosX == -1 && isVisible())
    {
        _origMenuPosX = pos().x();
        _origMenuPosY = pos().y();
    }
    int x = _origMenuPosX;
    int y = _origMenuPosY;
    bool changed = false;
    QRect screen = QApplication::desktop()->availableGeometry(QApplication::desktop()->screenNumber(QPoint(x, y)));
    int maxX = screen.x() + screen.width();
    int maxY = screen.y() + screen.height();
    if (x + width() > maxX)
    {
        x = maxX - width();
        changed = true;
    }
    if (y + height() > maxY)
    {
        y = maxY - height();
        changed = true;
    }
    if (changed)
        move(x, y);
}
