/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2020 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "actions/abstractactionitem.h"
#include "actions/actionactionitem.h"
#include "actions/headeractionitem.h"
#include "actions/submenuactionitem.h"
#include "actions/separatoractionitem.h"
#include "actions/actionsmanager.h"
#include "base/iconmanager.h"
#include "base/mainthread.h"
#include "qtactionmenu.h"
#include "qtactionmenuhandler.h"


void sh::ui::qt::QtActionMenuHandler::_updateSubmenu(sh::actions::SubmenuActionItem* itm, QtActionMenu* menu, std::function<void()> onchanged)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    menu->clearAllActions();
    foreach(auto subitem, itm->subitems())
        _createAndConnectAction(subitem.get(), menu, onchanged, itm);
}

std::shared_ptr<sh::actions::ActionActionItem> sh::ui::qt::QtActionMenuHandler::_getDefaultAction(QList<std::shared_ptr<sh::actions::AbstractActionItem> > actionList)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    int lastPrecedence = 0;
    std::shared_ptr<sh::actions::ActionActionItem> result = std::shared_ptr<sh::actions::ActionActionItem>(0);
    foreach( std::shared_ptr<sh::actions::AbstractActionItem> item, actionList)
    {
        if (item->visible() && item->enabled() && item->defaultActionPrecedence() > lastPrecedence)
        {
            lastPrecedence = item->defaultActionPrecedence();
            sh::actions::SubmenuActionItem* itmSubmenu = qobject_cast<sh::actions::SubmenuActionItem*>(item.get());
            if (itmSubmenu)
                result = _getDefaultAction(itmSubmenu->subitems());
            else
                result = std::dynamic_pointer_cast<sh::actions::ActionActionItem>(item);
        }
    }
    return result;
}

void sh::ui::qt::QtActionMenuHandler::_markDefault(std::weak_ptr<sh::actions::SubmenuActionItem> itmSubmenu, QtActionMenu* menu)
{
    std::shared_ptr<sh::actions::SubmenuActionItem> sitmSubmenu = itmSubmenu.lock();
    if (sitmSubmenu) {
        auto subitms = sitmSubmenu->subitems();
        auto defact = _getDefaultAction(subitms);
        int idef = subitms.indexOf(defact);
        int inow = 0;
        QFont normalfont = menu->font();
        QFont boldfont = normalfont;
        boldfont.setBold(true);
        foreach(QAction* a, menu->allActions()) {
            if (!menu->actionIsHeader(a))
                a->setFont((idef==inow) ? boldfont : normalfont);
            inow++;
        }
    }
}

void sh::ui::qt::QtActionMenuHandler::_markDefault2()
{
    QList<std::shared_ptr<sh::actions::AbstractActionItem>> lacts;
    foreach (auto ai, selacts)
        lacts.append(ai->createdAction);
    std::shared_ptr<sh::actions::ActionActionItem> da = _getDefaultAction(lacts);
    auto mmenu = menu.lock().get();
    if (mmenu) {
        QFont normalfont = mmenu->font();
        QFont boldfont = normalfont;
        boldfont.setBold(true);
        foreach(QAction* a, mmenu->allActions()) {
            std::shared_ptr<sh::actions::AbstractActionItem> sa = qaction2action[a];
            if (sa)
                a->setFont((sa==da) ? boldfont : normalfont);
        }
    }
}

QAction* sh::ui::qt::QtActionMenuHandler::_createAndConnectAction(sh::actions::AbstractActionItem* itm, QtActionMenu* menu, std::function<void()> onchanged, QObject *onchangedbuddy)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    auto itmAction = dynamic_cast<sh::actions::ActionActionItem*>(itm);
    auto itmHeader = dynamic_cast<sh::actions::HeaderActionItem*>(itm);
    auto itmSeparator = dynamic_cast<sh::actions::SeparatorActionItem*>(itm);
    auto itmSubmenu = dynamic_cast<sh::actions::SubmenuActionItem*>(itm);
    QAction* action = 0;
    if (itmAction) {
        action = menu->addNewAction();
        QObject::connect(action, &QAction::triggered, itmAction, [=](){
            itmAction->execute();
        });
    }
    else if (itmHeader)
        action = menu->addNewHeader();
    else if (itmSeparator)
        action = menu->addNewSeparator();
    else if (itmSubmenu) {
        sh::ui::qt::QtActionMenu* qsubmenu;
        action = menu->addNewSubMenu(&qsubmenu);
        std::weak_ptr<sh::actions::SubmenuActionItem> witmSubmenu = std::dynamic_pointer_cast<sh::actions::SubmenuActionItem>(itmSubmenu->shared_from_this());
        QObject::connect(itmSubmenu, &sh::actions::SubmenuActionItem::subitemsChanged, action, [=](){
            _updateSubmenu(itmSubmenu, qsubmenu, [=](){ _markDefault(witmSubmenu, qsubmenu); });
        });
        _updateSubmenu(itmSubmenu, qsubmenu, [=](){ _markDefault(witmSubmenu, qsubmenu); });
    }
    QObject::connect(itm, &sh::actions::AbstractActionItem::changed, action, [=](){
        _applyPropertiesToQAction(itm, action);
        if (onchanged)
            onchanged();
    });
    if (!onchangedbuddy)
        onchangedbuddy = itm;
    QObject::connect(onchangedbuddy, &QObject::destroyed, action, [=](){
        delete action;
    });
    _applyPropertiesToQAction(itm, action);
    if (onchanged)
        onchanged();
    return action;
}

void sh::ui::qt::QtActionMenuHandler::_applyPropertiesToQAction(sh::actions::AbstractActionItem* itm, QAction* _widgetaction)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    sh::actions::SeparatorActionItem* itmSeparator = qobject_cast<sh::actions::SeparatorActionItem*>(itm);
    sh::actions::HeaderActionItem* itmHeader = qobject_cast<sh::actions::HeaderActionItem*>(itm);
    _widgetaction->setVisible(itm->visible());
    if (!itmSeparator) {
        _widgetaction->setText(itm->text());
        _widgetaction->setIcon(sh::base::IconManager::instance()->getIconByFullname(itm->icon()));
        if (!itmHeader) {
            _widgetaction->setEnabled(itm->enabled());
            _widgetaction->setCheckable(itm->isCheckable());
            _widgetaction->setChecked(itm->isChecked());
        }
    }
    auto itmAction = dynamic_cast<sh::actions::ActionActionItem*>(itm);
    if (itmAction && !itmAction->shortcuthint().isEmpty())
    {
        _widgetaction->setShortcut(dynamic_cast<sh::actions::ActionActionItem*>(itm)->shortcuthint());
        _widgetaction->setShortcutContext(Qt::WidgetShortcut); // we just want to display that
    }
}

void sh::ui::qt::QtActionMenuHandler::_append_actions(QtActionMenu* menu, QList<std::shared_ptr<sh::actions::ActionInstantiation>> acts, std::shared_ptr<sh::actions::ActionCategory> category)
{
    QList<std::shared_ptr<sh::actions::ActionInstantiation> > actions;
    for(std::shared_ptr<sh::actions::ActionInstantiation> item : acts)
        if (item->category == category)
            actions.append(item);
    sh::actions::ActionsManager::sortAndSeparateActions(&actions);
    menu->addNewSeparator();
    for(std::shared_ptr<sh::actions::ActionInstantiation> acs: actions) {
        QAction* qac = _createAndConnectAction(acs->createdAction.get(), menu, [=](){
            _markDefault2();
        });
        if (!acs->shortcut.isEmpty())
        {
            qac->setShortcut(acs->shortcut);
            qac->setShortcutContext(Qt::WidgetShortcut); // we just want to display that
        }
        qaction2action[qac] = acs->createdAction;
    }
}

sh::ui::qt::QtActionMenuHandler::QtActionMenuHandler(std::shared_ptr<sh::actions::ActionInstantiation> ai, std::shared_ptr<QtActionMenu> menu) : menu(menu)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    sh::actions::ActionsManager::instance()->getActions(ai, [=](QList<std::shared_ptr<sh::actions::ActionInstantiation>> _selacts, QList<std::shared_ptr<sh::actions::ActionInstantiation>> _diracts) {
        selacts = _selacts;
        diracts = _diracts;
        sh::base::MainThread::dispatcher()->invokeAsync([=](){
            if (!ai->selectedNodes.isEmpty()) {
                auto hdr = menu->addNewHeader();
                hdr->setText(sh::actions::ActionsManager::selectionLabel(ai->selectedNodes));
                hdr->setIcon(sh::base::IconManager::instance()->getIcon("tool-pointer"));
                for(auto category : sh::actions::ActionCategory::categories())
                    _append_actions(menu.get(), selacts, category);
            }
            auto hdr = menu->addNewHeader();
            hdr->setText(sh::actions::ActionsManager::directoryLabel(ai->currentDirectory));
            hdr->setIcon(sh::base::IconManager::instance()->getIcon("folder"));
            for(auto category : sh::actions::ActionCategory::categories())
                _append_actions(menu.get(), diracts, category);
        });

    });
}
