/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2020 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "base/logger.h"
#include "base/iconmanager.h"
#include "base/mainthread.h"
#include "base/threadpool.h"
#include "qtmainwindow.h"
#include "qtuistyle.h"
#include "qtexceptiondialog.h"
#include "ui_qtexceptiondialog.h"


sh::ui::qt::QtExceptionDialog::QtExceptionDialog(QString error1, QString error2, QString details, QString icon, bool mayRetry, bool mayClose, bool mayCancel, bool showLoglabelAndDetails) :
    ExceptionDialog(error1, error2, details, icon, mayRetry, mayClose, mayCancel, showLoglabelAndDetails), ui(new Ui::QtExceptionDialog)
{
    ui->setupUi(this);
    sh::exceptions::Exception::executeGuarded([&](){
        ui->btnClose->setIcon(sh::base::IconManager::instance()->getIcon("arrow-right-double"));
        ui->btnRetry->setIcon(sh::base::IconManager::instance()->getIcon("view-refresh"));
        ui->btnCancel->setIcon(sh::base::IconManager::instance()->getIcon("dialog-close"));
        ui->lblIcon->setPixmap(sh::base::IconManager::instance()->getPixmap(icon, 48));
    }, sh::exceptions::ExecuteGuardFlag::IGNORE_ALL_RESUMEABLE_ERRORS_SILENTLY);
    auto uistyle = sh::ui::qt::QtMainWindow::mainWindow()->uiStyle();
    uistyle->createSheet().background_liketitlebar().applyTo(ui->panelHead);
    uistyle->createSheet().fontsize_header().applyTo(ui->lblHead);
    uistyle->createSheet().fontsize_smaller().textcolor_lighter1().applyTo(ui->lblLogfile);
    ui->lblHead->setText(error1);
    ui->label->setText(error2);
    ui->txtDetails->setText(details);
    ui->btnClose->setVisible(mayClose);
    ui->btnExit->setVisible(!mayClose);
    ui->btnRetry->setVisible(mayRetry);
    ui->btnCancel->setVisible(mayCancel);
    ui->btnShowDetails->setVisible(showLoglabelAndDetails);
    resize(600, 450);
    on_btnShowDetails_toggled(false);
}

sh::ui::qt::QtExceptionDialog::~QtExceptionDialog()
{
    delete ui;
}

sh::ui::ExceptionDialogResult sh::ui::qt::QtExceptionDialog::answer()
{
    return _answer;
}

void sh::ui::qt::QtExceptionDialog::on_btnExit_clicked()
{
    _answer = ExceptionDialogResult::Shutdown;
    this->done(QDialog::Accepted);
}

void sh::ui::qt::QtExceptionDialog::on_btnClose_clicked()
{
    _answer = ExceptionDialogResult::Close;
    this->done(QDialog::Accepted);
}

void sh::ui::qt::QtExceptionDialog::on_btnCancel_clicked()
{
    _answer = ExceptionDialogResult::Cancel;
    this->done(QDialog::Accepted);
}

void sh::ui::qt::QtExceptionDialog::on_btnRetry_clicked()
{
    _answer = ExceptionDialogResult::Retry;
    this->done(QDialog::Accepted);
}

void sh::ui::qt::QtExceptionDialog::on_btnShowLog_clicked()
{
    on_btnClose_clicked();
    sh::base::ThreadPool::enqueueForce([](){
        sh::ui::MainWindow::mainWindow()->showLogViewDialog(QString(), QString(), sh::base::LogSeverity::ERROR);
    });
}

void sh::ui::qt::QtExceptionDialog::on_btnExitAndSaveLog_clicked()
{
    QString log = sh::base::Logger::instance()->getLogAsText();
    QString filename = QDir::homePath() + QString("/shallot.crashlog.%1.txt").arg(QDateTime::currentDateTime().toString(Qt::ISODate));
    QFile file(filename);
    if (file.open(QIODevice::ReadWrite)) {
        QTextStream stream(&file);
        stream << log << endl;
        file.close();
    }
    on_btnExit_clicked();
}

void sh::ui::qt::QtExceptionDialog::on_btnShowDetails_toggled(bool checked)
{
    ui->btnExitAndSaveLog->setVisible(checked && !mayClose());
    ui->btnShowLog->setVisible(checked && mayClose());
    ui->lblLogfile->setVisible(checked);
    ui->txtDetails->setVisible(checked);
    if (checked)
        ui->txtDetails->setFocus();
}
