/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2020 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "filesystem/filesystemnode.h"
#include "filesystem/eurl.h"
#include "paneldetails/paneldetailmanager.h"
#include "base/threadpool.h"
#include "base/mainthread.h"
#include "qtmainwindow.h"
#include "qtuistyle.h"
#include "qtfiledetailspanel.h"


sh::ui::qt::QtFileDetailsPanel::QtFileDetailsPanel(QWidget *parent) :
    QWidget(parent)
{
    int fontsize = sh::ui::qt::QtMainWindow::mainWindow()->uiStyle()->fontSizeInPt(90);
    fontNormal.setPointSize(fontsize);
    fontBold.setPointSize(fontsize);
    fontBold.setBold(true);
    PADDINGX = this->logicalDpiX() / 3;
    PADDINGY = this->logicalDpiY() / 10;
    setMouseTracking(true);
    connect(&_widgetimagecachetimer, &QTimer::timeout, [=](){
        _invalidate_widgetimagecache_ifrefreshdone(true);
        refreshview();
        repaint();
    });
}

void sh::ui::qt::QtFileDetailsPanel::setNodes(QList<std::shared_ptr<sh::filesystem::FilesystemNode>> nodes)
{
    _nodes = nodes;
    refresh();
}

void sh::ui::qt::QtFileDetailsPanel::setOrientation(Qt::Orientation orientation)
{
    _orientation = orientation;
    refresh();
}

void sh::ui::qt::QtFileDetailsPanel::showEvent(QShowEvent*)
{
    refresh();
}

void sh::ui::qt::QtFileDetailsPanel::refresh()
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    if (!isVisible())
        return;
    if (_widgetimagecache.isNull())
    {
        _widgetimagecachetimer.start(200);
        _widgetimagecache = grab();
        setEnabled(false);
    }
    foreach(std::shared_ptr<sh::paneldetails::PanelDetail> paneldetail, _panelDetails)
        QObject::disconnect(paneldetail.get());
    _panelDetails.clear();
    sh::base::ThreadPool::enqueueForce([=](){
        QList<std::shared_ptr<sh::paneldetails::PanelDetail>> paneldetails = sh::paneldetails::PanelDetailManager::instance()->getDetails(_nodes);
        sh::base::MainThread::dispatcher()->invokeSync([&](){
            _panelDetails = paneldetails;
            foreach(std::shared_ptr<sh::paneldetails::PanelDetail> paneldetail, paneldetails)
            {
                QObject::connect(paneldetail.get(), &sh::paneldetails::PanelDetail::changed, this, [=](){
                    _invalidate_widgetimagecache_ifrefreshdone();
                    refreshview();
                    repaint();
                });
            }
            _invalidate_widgetimagecache_ifrefreshdone();
            refreshview();
            repaint();
        });
    });
}

void sh::ui::qt::QtFileDetailsPanel::_invalidate_widgetimagecache_ifrefreshdone(bool force)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    bool finished = true;
    if (!force) {
        foreach (std::shared_ptr<sh::paneldetails::PanelDetail> detail, _panelDetails) {
            foreach(sh::paneldetails::PanelDetailRow row, detail->rows()) {
                if (row.value().length()>0 && row.value()[0].isWaiting()) {
                    finished = false;
                    break;
                }
            }
            if (!finished)
                break;
        }
    }
    if (finished)
    {
        _widgetimagecachetimer.stop();
        _widgetimagecache = QPixmap();
        setEnabled(true);
    }
}

QList<int> sh::ui::qt::QtFileDetailsPanel::computevalueelementgeometry(QList<sh::paneldetails::PanelDetailRowValueElement> value, int detailvaluewidth)
{
    QList<int> result;
    QList<int> ifixed;
    QFontMetrics fmNormal(fontNormal);
    int grossvar = 0;
    int grossfixed = 0;
    foreach(sh::paneldetails::PanelDetailRowValueElement e, value)
    {
        if (!e.icon().isNull())
        {
            int w = fmNormal.height();
            ifixed.append(result.length());
            result.append(w);
            grossfixed += w;
        }
        else if (!e.linktarget().isNull())
        {
            int w = fmNormal.boundingRect(e.text() + "   ").width();
            ifixed.append(result.length());
            result.append(w);
            grossfixed += w;
        }
        else if (!e.text().isNull())
        {
            int w = fmNormal.boundingRect(e.text() + "   ").width();
            result.append(w);
            grossvar += w;
        }
    }
    int availw = detailvaluewidth - grossfixed;
    for(int i=0; i<result.length(); i++)
    {
        if (!ifixed.contains(i))
            result[i] = std::min( (int)(availw*((double)(result[i])/grossvar)), result[i] );
    }
    return result;
}

void sh::ui::qt::QtFileDetailsPanel::refreshview()
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    if (!isVisible())
        return;
    if (!_widgetimagecache.isNull())
        return;
    QFontMetrics fmNormal(fontNormal);
    QFontMetrics fmBold(fontBold);
    int lineheight = fmNormal.height();
    QSize _size = size();
    int mywidth = _size.width();
    int myheight = _size.height();
    int xcur = 0;
    int ycur = 0;
    int curblockheight = 0;
    _placements.clear();
    int logicalDpcX = this->logicalDpiX()/2.54;
    int addlineheight = (_orientation == Qt::Vertical) ? lineheight : 0;
    foreach (std::shared_ptr<sh::paneldetails::PanelDetail> detail, _panelDetails)
    {
        int detailheight = detail->rows().length() * (lineheight + addlineheight);
        if (detailheight)
        {
            int detailvaluewidth = detail->valueWidthHint() * logicalDpcX;
            QList<DetailRowPlacement> rowplacements;
            int headerwidth = 0;
            foreach (sh::paneldetails::PanelDetailRow row, detail->rows())
            {
                int _w = row.key().isEmpty() ? 0 : fmBold.width(row.key()+": ");
                headerwidth = std::max(_w, headerwidth);
                int cox = (_orientation == Qt::Vertical) ? 0 : headerwidth;
                int coy = addlineheight;
                int rh = lineheight + addlineheight;
                rowplacements.append(DetailRowPlacement(rh, cox, coy, computevalueelementgeometry(row.value(), detailvaluewidth)));
            }
            int detailwidth = headerwidth + detailvaluewidth;
            if ((xcur+detailwidth < mywidth) && (ycur+detailheight < myheight))
            {
                // place in same line
                _placements.append(DetailPlacement(xcur, ycur, detailwidth, lineheight, rowplacements, detail));
                xcur += detailwidth + PADDINGX;
                curblockheight = std::max(curblockheight, detailheight);
            }
            else if ((detailwidth < mywidth) && (ycur+curblockheight+PADDINGY+detailheight < myheight))
            {
                // place in new line
                _placements.last().w += (mywidth - xcur);
                ycur += curblockheight + PADDINGY;
                xcur = 0;
                _placements.append(DetailPlacement(xcur, ycur, detailwidth, lineheight, rowplacements, detail));
                xcur += detailwidth + PADDINGX;
                curblockheight = detailheight;
            }
        }
    }
    if (!_placements.isEmpty())
        _placements.last().w += (mywidth - xcur);
}


void sh::ui::qt::QtFileDetailsPanel::mouseMoveEvent(QMouseEvent*){
    repaint();
}

void sh::ui::qt::QtFileDetailsPanel::leaveEvent(QEvent *)
{
    repaint();
}

void sh::ui::qt::QtFileDetailsPanel::mouseReleaseEvent(QMouseEvent *event)
{
    _iterate([&](int, int, int, int, sh::paneldetails::PanelDetailRow*, QString){
    }, [&](int x, int y, int w, int h, sh::paneldetails::PanelDetailRow*, sh::paneldetails::PanelDetailRowValueElement* element, std::shared_ptr<sh::paneldetails::PanelDetail> detail){
        if (!element->linktarget().isNull())
        {
            QRect rect(x, y, w, h);
            if (rect.contains(event->pos()))
                detail->_emit_linkTriggered(element->linktarget());

        }
    });
    QWidget::mouseReleaseEvent(event);
}

void sh::ui::qt::QtFileDetailsPanel::_iterate(
        std::function<void(int x, int y, int w, int h, sh::paneldetails::PanelDetailRow* row, QString text)> _key,
        std::function<void(int x, int y, int w, int h, sh::paneldetails::PanelDetailRow* row, sh::paneldetails::PanelDetailRowValueElement* element, std::shared_ptr<sh::paneldetails::PanelDetail> detail)> _element
        )
{
    foreach(DetailPlacement placement, _placements)
    {
        int y = 0;
        QList<DetailRowPlacement> rowplacements = placement.rowplacements;
        foreach(sh::paneldetails::PanelDetailRow row, placement.detail->rows())
        {
            DetailRowPlacement rowplacement = rowplacements.takeFirst();
            if (!row.key().isEmpty())
                _key(placement.x, placement.y + y, placement.w, placement.h, &row, row.key()+": ");
            int x = rowplacement.contentx;
            int _x = placement.x + x;
            foreach(sh::paneldetails::PanelDetailRowValueElement element, row.value())
            {
                int w = rowplacement.elementwidths.isEmpty() ? 0 : rowplacement.elementwidths.takeFirst();
                int _y = placement.y + y + rowplacement.contenty;
                _element(_x, _y, w, placement.h, &row, &element, placement.detail);
                _x += w;
                x += w;
                if (x > placement.w)
                    break;
            }
            y += rowplacement.h;
        }
    }
}

QSize sh::ui::qt::QtFileDetailsPanel::sizeHint() const
{
    return QSize(250, 80);
}

void sh::ui::qt::QtFileDetailsPanel::paintEvent(QPaintEvent*)
{
    QPainter painter(this);
    if (!_widgetimagecache.isNull())
        painter.drawPixmap(0, 0, _widgetimagecache);
    else
    {
        QPoint mousepos = mapFromGlobal(QCursor::pos());
        bool mouseinwidget = mousepos.x() >= 0 && mousepos.y() >= 0 && mousepos.x() < width() && mousepos.y() < height();
        QFontMetrics fmNormal(fontNormal);
        _iterate([&](int x, int y, int w, int h, sh::paneldetails::PanelDetailRow*, QString text){
            QRect rect(x, y, w, h);
            painter.setFont(fontBold);
            painter.setPen( QPalette().text().color() );
            painter.drawText(rect, text);
        }, [&](int x, int y, int w, int h, sh::paneldetails::PanelDetailRow*, sh::paneldetails::PanelDetailRowValueElement* element, std::shared_ptr<sh::paneldetails::PanelDetail>){
            if (!element->icon().isNull())
            {
                painter.drawPixmap(x, y, element->icon().pixmap(fmNormal.height()*0.9, fmNormal.height()*0.9) );
            }
            else if (!element->text().isNull())
            {
                QRect rect(x, y, w, h);
                QString rowval = fmNormal.elidedText(element->text(), Qt::ElideLeft, w);
                painter.setFont(fontNormal);
                bool drawtxt = true;
                if (element->linktarget().isNull())
                    painter.setPen( QPalette().text().color() );
                else
                {
                    if (element->linkautohide())
                        drawtxt=mouseinwidget;
                    painter.setPen( QPalette().link().color() );
                    if (rect.contains(mousepos))
                    {
                        QFont _f = fontNormal;
                        _f.setUnderline(true);
                        painter.setFont(_f);
                    }
                }
                if (drawtxt)
                    painter.drawText(rect, 0, rowval);
            }
        });
    }
}

void sh::ui::qt::QtFileDetailsPanel::resizeEvent(QResizeEvent*)
{
    refreshview();
}

sh::ui::qt::QtFileDetailsPanel::~QtFileDetailsPanel()
{
}

sh::ui::qt::QtFileDetailsPanel::DetailRowPlacement::DetailRowPlacement(int h, int contentx, int contenty, QList<int> elementwidths)
            : h(h), contentx(contentx), contenty(contenty), elementwidths(elementwidths)
{
}

sh::ui::qt::QtFileDetailsPanel::DetailPlacement::DetailPlacement(int x, int y, int w, int h, QList<DetailRowPlacement> rowplacements, std::shared_ptr<sh::paneldetails::PanelDetail> detail)
            : detail(detail), x(x), y(y), w(w), h(h), rowplacements(rowplacements)
{
}
