/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2020 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <QMenu>
#include <QDragEnterEvent>
#include <QApplication>
#include <QDrag>
#include "actions/actionactionitem.h"
#include "filesystem/detailcolumn.h"
#include "filesystem/eurl.h"
#include "filesystem/filesystemhandler.h"
#include "filesystem/filesystemhandlerregister.h"
#include "filesystem/filesystemmodel.h"
#include "filesystem/filesystemmodelfileviewproxy.h"
#include "settings/settingsmanager.h"
#include "base/threadpool.h"
#include "tools/dataexchange.h"
#include "ui/mainwindow.h"
#include "ui/fileview.h"
#include "qtfileiconview.h"
#include "qtfileviewcontrol.h"


sh::ui::qt::QtFileIconview::QtFileIconview(QWidget *parent) :
    QListView(parent), sh::ui::qt::QtFileView()
{
    _modelwithtooltip = true;
    setItemDelegate(new MyItemDelegate(this));
    setAcceptDrops(true);
    setSelectionMode(SelectionMode::ExtendedSelection);
    setFlow(QListView::LeftToRight);
    setViewMode(QListView::IconMode);
    setResizeMode(QListView::Adjust);
    setUniformItemSizes(true);
    setThumbDimension(1.0);
}

void sh::ui::qt::QtFileIconview::setThumbDimension(double size)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    size = std::max(0.5, size);
    _dsx = logicalDpiX() * size;
    _dsy = logicalDpiY() * size;
    setIconSize(QSize(_dsx, _dsy));
    int old_thumbnailsize = _thumbnailsize;
    _thumbnailsize = iconSize().width();
    bool doreissue = (old_thumbnailsize != -1 && old_thumbnailsize != _thumbnailsize);
    _lineheight = QFontMetrics(font()).height();
    _deleg_w = (int)(1.2*_dsx);
    _deleg_h = (int)(1.1*_dsy + 2*_lineheight);
    setGridSize(QSize(_deleg_w, _deleg_h));
    if (doreissue && control() && control()->filemodel())
        control()->filemodel()->setThumbnail(true, _thumbnailsize);
}

void sh::ui::qt::QtFileIconview::dragEnterEvent(QDragEnterEvent *event)
{
    dnd_dragEnterEvent(event);
}

void sh::ui::qt::QtFileIconview::mousePressEvent(QMouseEvent *event)
{
    if (!dnd_mousePressEvent(event))
        QListView::mousePressEvent(event);
}

void sh::ui::qt::QtFileIconview::mouseMoveEvent(QMouseEvent *event)
{
    if (!dnd_mouseMoveEvent(event))
        QListView::mouseMoveEvent(event);
}

void sh::ui::qt::QtFileIconview::dragMoveEvent(QDragMoveEvent *event)
{
    dnd_dragMoveEvent(event);
}

void sh::ui::qt::QtFileIconview::dragLeaveEvent(QDragLeaveEvent *event)
{
    dnd_dragLeaveEvent(event);
}

void sh::ui::qt::QtFileIconview::dropEvent(QDropEvent *event)
{
    dnd_dropEvent(event);
}

void sh::ui::qt::QtFileIconview::keyPressEvent(QKeyEvent *event)
{
    evaluateKeyPress(event);
    QListView::keyPressEvent(event);
}

void sh::ui::qt::QtFileIconview::onBackgroundColorSet()
{
    this_qwidget->setStyleSheet("QListView{background:" + bgColor + "}");
}

sh::ui::qt::QtFileIconview::MyItemDelegate::MyItemDelegate(QObject* parent)
    : QStyledItemDelegate(parent)
{

}

QSize sh::ui::qt::QtFileIconview::MyItemDelegate::sizeHint(const QStyleOptionViewItem& option, const QModelIndex&) const
{
    auto w = qobject_cast<sh::ui::qt::QtFileIconview*>(option.styleObject);
    return QSize(w->_deleg_w, w->_deleg_h);
}

int sh::ui::qt::QtFileIconview::MyItemDelegate::_getVisibleCharCount(QString text, int width, QFontMetrics* fm, bool forward) const
{
    if (text.isEmpty())
        return 0;
    double stepsize = 0;
    int chars = width / fm->averageCharWidth();
    while (true) {
        QString s1 = forward ? text.left(chars) : text.right(chars);
        int s1w = fm->width(s1 + (forward ? ELLIPSIS : ""));
        if (s1w <= width) {
            // we aren't too long
            if (s1 == text)
                break;  // we are even complete with it
            QString s2 = forward ? text.left(chars+1) : text.right(chars+1);
            int s2w = fm->width(s2 + (forward ? ELLIPSIS : ""));
            if (s2w > width) {
                // we aren't too short
                break;
            }  // otherwise we are too short
        }  // otherwise we are too long
        double avgcharwidth = s1.isEmpty() ? fm->averageCharWidth() : (s1w / s1.length());
        double prediction = ((double)(width - s1w)) / avgcharwidth;
        stepsize = 0.5 * stepsize + 0.5 * prediction;
        chars += ((stepsize<0)?-1:1) * std::max(1.0, std::abs(stepsize));
    }
    return std::min(chars, text.length());
}

void sh::ui::qt::QtFileIconview::MyItemDelegate::paint(QPainter *painter, const QStyleOptionViewItem &option, const QModelIndex &index) const
{
    auto w = qobject_cast<sh::ui::qt::QtFileIconview*>(option.styleObject);
    painter->save();
    painter->setClipRect(option.rect);
    bool selected = option.state & QStyle::State_Selected;
    QColor hicol = QApplication::palette().color(QPalette::Normal, QPalette::Highlight);
    if (selected)
        painter->fillRect(option.rect, hicol);
    if (option.state & QStyle::State_MouseOver) {
        hicol.setAlpha(100);
        painter->fillRect(option.rect, hicol);
    }
    auto model = index.model();
    QString text = model->data(index).toString();
    QVariant vicon = model->data(index, Qt::DecorationRole);
    QIcon icon = *(QIcon*)vicon.data();
    icon.paint(painter, option.rect.left()+1, option.rect.top()+1, w->_deleg_w-2, w->_dsy, Qt::AlignCenter, selected?QIcon::Selected:QIcon::Normal, QIcon::On);
    if (!text.isEmpty())
    {
        int txty1 = (int)(1 + w->_dsy + 0.05*w->_dsy);
        int txty2 = txty1 + w->_lineheight;
        QFontMetrics fm(w->font());
        int charsline1 = _getVisibleCharCount(text, w->_deleg_w-4, &fm, true);
        int charsline2 = _getVisibleCharCount(text, w->_deleg_w-4, &fm, false);
        QString line1;
        QString line2;
        if (charsline2 == text.length())  // complete text goes in one line
            line1 = text;
        else {
            if (charsline1 + charsline2 > text.length()) {  // we have more room than characters
                QString lline1 = text.left(charsline1);
                for(int j=lline1.length()-1; j>=0; j--) {
                    if (!lline1.at(j).isLetterOrNumber()) {
                        if (j + 1 + charsline2 >= text.length()) {
                            line1 = text.left(j+1);
                            line2 = text.mid(j+1);
                        }
                        break;
                    }
                }
            }
            if (line1.isNull()) {
                line1 = text.left(charsline1) + ((charsline1 + charsline2 >= text.length()) ? "" : ELLIPSIS);
                line2 = text.mid(charsline1).right(charsline2);
            }
        }
        painter->setPen(QApplication::palette().color(QPalette::Normal, selected?QPalette::HighlightedText:QPalette::Text));
        painter->drawText(option.rect.left()+2, option.rect.top()+txty1, w->_deleg_w-4, w->_lineheight, Qt::AlignCenter, line1);
        painter->drawText(option.rect.left()+2, option.rect.top()+txty2, w->_deleg_w-4, w->_lineheight, Qt::AlignCenter, line2);
    }
    painter->restore();
}
