/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2020 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <QScrollBar>
#include "filesystem/filesystemnode.h"
#include "filesystem/eurl.h"
#include "filesystem/operation.h"
#include "base/iconmanager.h"
#include "base/singletoninitializer.h"
#include "base/threadpool.h"
#include "base/mainthread.h"
#include "qtfilepropertydialogtabactionsview.h"
#include "qtfilepropertydialogtabtableview.h"
#include "qtfilepropertydialogtabtextview.h"
#include "qtfilepropertydialogtabicontextbannerview.h"
#include "qtmainwindow.h"
#include "qtuistyle.h"
#include "qtfilepropertydialog.h"
#include "ui_qtfilepropertydialog.h"


sh::ui::qt::QtFilePropertyDialog::QtFilePropertyDialog(QList<std::shared_ptr<sh::filesystem::FilesystemNode> > nodes) :
    FilePropertyDialog(nodes), ui(new Ui::QtFilePropertyDialog)
{
    ui->setupUi(this);
    scrollAreaLayout = qobject_cast<QVBoxLayout*>(ui->scrollAreaWidgetContents->layout());
    ui->btnClose->setSizeByFactor(120);
    ui->btnClose->setIcon(sh::base::IconManager::instance()->getIcon("window-close"));
    auto uistyle = sh::ui::qt::QtMainWindow::mainWindow()->uiStyle();
    uistyle->createSheet().background_liketitlebar().applyTo(ui->panelHead);
    uistyle->createSheet().fontsize_header().applyTo(ui->label);
    uistyle->createSheet().background_highlighted().applyTo(ui->scrollAreaWidgetContents);
    connect(ui->scrollArea->verticalScrollBar(), &QScrollBar::valueChanged, this, &QtFilePropertyDialog::selectTabByScrollPosition);
    ui->scrollAreaWidgetContents->setFocus();
}

void sh::ui::qt::QtFilePropertyDialog::refresh()
{
    updateWidgets();
}

sh::ui::FilePropertyDialogTabTableView* sh::ui::qt::QtFilePropertyDialog::createTabViewTable()
{
    return new sh::ui::qt::QtFilePropertyDialogTabTableView(this);
}

sh::ui::FilePropertyDialogTabTextView* sh::ui::qt::QtFilePropertyDialog::createTabViewText()
{
    return new sh::ui::qt::QtFilePropertyDialogTabTextView(this);
}

sh::ui::FilePropertyDialogTabIconTextBannerView* sh::ui::qt::QtFilePropertyDialog::createTabViewIconTextBanner()
{
    return new sh::ui::qt::QtFilePropertyDialogTabIconTextBannerView(this);
}

void sh::ui::qt::QtFilePropertyDialog::init(std::shared_ptr<Dialog>)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    auto _nodes = nodes();
    setWindowTitle(_nodes.length()==1 ? tr("Properties: %1").arg(_nodes.first()->eurl()->basename()) : tr("Properties: %1 items").arg(_nodes.length()));
    ui->label->setText(_nodes.length()==1 ? tr("Properties of '%1'").arg(_nodes.first()->eurl()->basename()) : tr("Properties of %1 items").arg(_nodes.length()));
    auto uistyle = sh::ui::qt::QtMainWindow::mainWindow()->uiStyle();
    foreach (FilePropertyDialogTabFactory* tabfactory, getPropertyTabs()) {
        std::shared_ptr<sh::ui::FilePropertyDialogTab> tab = tabfactory->construct(std::dynamic_pointer_cast<FilePropertyDialog>(shared_from_this()));
        if (tab) {
            QList<QString> properties = tab->properties();
            TabViewStructWidgets widgets;
            if (!properties.isEmpty()) {
                QWidget* widget = new QWidget(this);
                QVBoxLayout* layout = new QVBoxLayout(widget);
                widget->setLayout(layout);
                QLabel* headlabel = new QLabel(this);
                headlabel->setText(tab->titleWithoutMnemonic());
                headlabel->setWordWrap(true);
                headlabel->setAlignment(Qt::AlignHCenter);
                uistyle->createSheet().fontsize_byFactor(150).applyTo(headlabel);
                layout->addWidget(headlabel);
                for (int i=0; i<properties.length(); i++) {
                    QLabel* label = new QLabel(this);
                    label->setText(properties[i]);
                    label->setWordWrap(true);
                    uistyle->createSheet().fontsize_byFactor(130).customcss("padding:0px 0px 0px 30px").applyTo(label);
                    layout->addWidget(label);
                    sh::ui::qt::QtFilePropertyDialogTabActionsView* pwidget = new sh::ui::qt::QtFilePropertyDialogTabActionsView(this);
                    tab->populateWidget(i, pwidget);
                    pwidget->setParent(this);
                    pwidget->installEventFilter(this);
                    layout->addWidget(pwidget);
                    widgets.append(pwidget);
                    layout->addSpacing(1); //will be set later
                }
                scrollAreaLayout->addWidget(widget);
                QWidget* internaltabwidget = new QWidget(this);
                _internaltabwidgets.append(internaltabwidget);
                _tabheads.append(tab->title());
            }
            addTabstruct(tab, widgets);
        }
    }
    scrollAreaLayout->addStretch(1);
    QDialog::show();
    foreach(auto tab, tabstructs())
        foreach(auto w, tab.second)
            dynamic_cast<sh::ui::qt::QtFilePropertyDialogTabActionsView*>(w)->hide();
    updateWidgets();
}

bool sh::ui::qt::QtFilePropertyDialog::eventFilter(QObject *obj, QEvent *event)
{
    if (isVisible()/*otherwise dialog is closing*/
            && (event->type() == QEvent::ShowToParent || event->type() == QEvent::HideToParent)) { // a child widget has been shown/hidden
        bool newVisibility = (event->type() == QEvent::ShowToParent);
        QWidget* wobj = qobject_cast<QWidget*>(obj);
        QWidget* wpar = wobj->parentWidget();
        QVBoxLayout* wparlayout = qobject_cast<QVBoxLayout*>(wpar->layout());
        int i = wparlayout->indexOf(wobj);
        // handle the label visibility
        QWidget* lbl = qobject_cast<QWidget*>(wparlayout->itemAt(i-1)->widget());
        QSpacerItem* spacer = (QSpacerItem*)(wparlayout->itemAt(i+1));
        lbl->setVisible(newVisibility);
        spacer->changeSize(1, newVisibility?20:0); // this fails sometimes :-/ but not a major thing...
        // check if the parent box should be visible
        bool boxvisible = false;
        wpar->setVisible(true); // otherwise our visibility tests some lines below would fail
        for (int i=1; i<wparlayout->count(); i++) {
            QWidget* w = qobject_cast<QWidget*>(wparlayout->itemAt(i)->widget());
            if (w && w->isVisible()) {
                boxvisible = true;
                break;
            }
        }
        wpar->setVisible(boxvisible);
        // handle the tab visibility
        int tabidx = scrollAreaLayout->indexOf(wpar);
        if (boxvisible) {
            int wtabidx = 0;
            bool insert = true;
            while (wtabidx < ui->tabWidget->count()) {
                QWidget* w = ui->tabWidget->widget(wtabidx);
                int _tabidx = _internaltabwidgets.indexOf(w);
                if (_tabidx == tabidx) {
                    insert = false;
                    break;
                }
                else if (_tabidx > tabidx)
                    break;
                wtabidx++;
            }
            if (insert) {
                int _i = ui->tabWidget->currentIndex();
                ui->tabWidget->insertTab(wtabidx, _internaltabwidgets[tabidx], _tabheads[tabidx]);
                if (_i == 0)
                    ui->tabWidget->setCurrentIndex(_i);
            }
        }
        else {
            QWidget* w =  _internaltabwidgets[tabidx];
            int i = ui->tabWidget->indexOf(w);
            if (i!=-1)
                ui->tabWidget->removeTab(i);
        }
    }
    return QObject::eventFilter(obj, event);
}

sh::ui::qt::QtFilePropertyDialog::~QtFilePropertyDialog()
{
    delete ui;
}

void sh::ui::qt::QtFilePropertyDialog::on_btnClose_clicked()
{
    reject();
}

void sh::ui::qt::QtFilePropertyDialog::on_btnRefresh_clicked()
{
    if (!isVisible())
        return;
    updateWidgets();
}

void sh::ui::qt::QtFilePropertyDialog::on_tabWidget_currentChanged(int index)
{
    if (_skip_on_tabWidget_currentChanged)
        return;
    _skip_selectTabByScrollPosition++;
    int tindex = _internaltabwidgets.indexOf(ui->tabWidget->widget(index));
    QLayoutItem* li = scrollAreaLayout->itemAt(tindex);
    if (li) {
        int ypos = li->geometry().y();
        ui->scrollArea->verticalScrollBar()->setValue(ypos * (ui->scrollArea->verticalScrollBar()->maximum()) / std::max(1, ui->scrollAreaWidgetContents->height()-ui->scrollArea->height()));
    }
    _skip_selectTabByScrollPosition--;
}

void sh::ui::qt::QtFilePropertyDialog::selectTabByScrollPosition()
{
    if (_skip_selectTabByScrollPosition)
        return;
    QWidget* w = ui->scrollArea->childAt(10, 10);
    int iw = scrollAreaLayout->indexOf(w);
    if (iw != -1) {
        int i = ui->tabWidget->indexOf(_internaltabwidgets[iw]);
        _skip_on_tabWidget_currentChanged++;
        ui->tabWidget->setCurrentIndex(i);
        _skip_on_tabWidget_currentChanged--;
    }
}
