/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2020 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <math.h>
#include <QKeyEvent>
#include "actions/actionsmanager.h"
#include "actions/abstractactionitem.h"
#include "actions/actionactionitem.h"
#include "filesystem/filesystemnode.h"
#include "filesystem/filesystemmodel.h"
#include "filesystem/filesystemmodelfileviewproxy.h"
#include "exceptions/argumentexception.h"
#include "settings/settingsregistration.h"
#include "settings/settingsmanager.h"
#include "base/mainthread.h"
#include "base/threadpool.h"
#include "qtmainwindow.h"
#include "qtuistyle.h"
#include "qtfilelist.h"
#include "qtfileview.h"
#include "qtfileviewcontrol.h"
#include "qtfileiconview.h"
#include "qtfilesystempanel.h"
#include "ui_qtfilesystempanel.h"


sh::ui::qt::QtFilesystemPanel::QtFilesystemPanel(QWidget *parent) :
    QWidget(parent),
    ui(new Ui::QtFilesystemPanel)
{
    ui->setupUi(this);
    auto uistyle = sh::ui::qt::QtMainWindow::mainWindow()->uiStyle();
    activeColor = uistyle->backgroundColor().name();
    inactiveColor = uistyle->inactiveBackgroundColor().name();
    addView();
    activeView = _views[0];
    activeView->focus();
}

sh::ui::qt::QtFilesystemPanel::~QtFilesystemPanel()
{
    delete ui;
    _views.clear();
    foreach(sh::ui::FileView* o, _viewcontrols)
        delete o;
}

void sh::ui::qt::QtFilesystemPanel::addView(bool reinitialize)
{
    sh::ui::qt::QtFileViewControl* viewctrl = new sh::ui::qt::QtFileViewControl(this, _views.count());
    if (reinitialize)
        viewctrl->columndimensions()->remove();
    _viewcontrols.append(viewctrl);
    if (reinitialize)
        viewctrl->columndimensions()->remove();
    std::shared_ptr<sh::ui::qt::QtFileView> list;
    buildView(&list, viewctrl, true);
    connect(viewctrl, &sh::ui::qt::QtFileViewControl::nodesActivated, this, [=](QList<std::shared_ptr<sh::filesystem::FilesystemNode>> nodes){
        sh::ui::qt::QtMainWindow::mainWindow()->activateNodes(nodes, [&](std::shared_ptr<sh::filesystem::FilesystemNode> nn){
            emit panelFolderActivated(nn, true);
        });
    });
    connect(viewctrl, &sh::ui::FileView::viewOptionChanged, this, [=](){
        int index = qobject_cast<sh::ui::FileView*>(sender())->index();
        emit panelViewOptionChanged(index);
    });
    connect(viewctrl, &sh::ui::FileView::selectionChanged, this, [=](){
        emit panelSelectionChanged();
    });
    emit panelViewCountChanged();
    sh::settings::SettingsManager::instance()->applyGlobalSettingsToFileView(viewctrl);
    sh::settings::SettingsManager::instance()->applyPerEurlSettingsToFileView(viewctrl);
    list->focus();
}

void sh::ui::qt::QtFilesystemPanel::buildView(std::shared_ptr<sh::ui::qt::QtFileView>* list, sh::ui::qt::QtFileViewControl* viewctrl, bool isNew)
{
    QWidget* main;
    if (isNew)
    {
        main = new QWidget(ui->splitter);
        QVBoxLayout* mainlayout = new QVBoxLayout(main);
        mainlayout->setMargin(0);
        main->setLayout(mainlayout);
        _mainviews.insert(viewctrl->index(), main);
        ui->splitter->insertWidget(viewctrl->index(), main);
    }
    else
        main = _mainviews[viewctrl->index()];
    if (viewctrl->viewmode() == sh::ui::FileViewMode::ListView)
        *list = std::shared_ptr<sh::ui::qt::QtFileView>(new sh::ui::qt::QtFileList(main));
    else if (viewctrl->viewmode() == sh::ui::FileViewMode::IconView)
        *list = std::shared_ptr<sh::ui::qt::QtFileView>(new sh::ui::qt::QtFileIconview(main));
    else
        sh::exceptions::Exception::executeGuarded([&](){
            throw sh::exceptions::ProgramException(sh::exceptions::ExceptionData().details(QString("%1 is not a valid mode.").arg((int)viewctrl->viewmode())));
        });
    if (!isNew)
        _views.removeAt(viewctrl->index());
    _views.insert(viewctrl->index(), *list);
    (*list)->configure(viewctrl);
    (*list)->setBackgroundColor(inactiveColor);
    QWidget *llst = (*list)->as_qwidget();
    dynamic_cast<QVBoxLayout*>(main->layout())->insertWidget(0, (*list)->as_qwidget());
    llst->installEventFilter(this);
}

void sh::ui::qt::QtFilesystemPanel::removeView(int i)
{
    std::shared_ptr<sh::ui::qt::QtFileView> ov = _views.takeAt(i);
    dynamic_cast<QWidget*>(ov.get())->setParent(0);
    delete _viewcontrols.takeAt(i);
    QWidget* mw = _mainviews.takeAt(i);
    _customwidgets.remove(mw);
    delete mw;
    _viewcontrols.last()->columndimensions()->remove(); // this just looks weird ;)
    for(int j = i; j< _viewcontrols.length(); j++)
        _viewcontrols[j]->setIndex(j);
    if (ov == activeView)
        _views[std::min(i, _views.length()-1)]->focus();
    emit panelViewCountChanged();
}

sh::ui::FileView* sh::ui::qt::QtFilesystemPanel::currentView()
{
    return activeView->control();
}

int sh::ui::qt::QtFilesystemPanel::currentViewIndex()
{
    return activeView->control()->index();
}

void sh::ui::qt::QtFilesystemPanel::selectFolder(std::shared_ptr<filesystem::FilesystemNode> folder, int index)
{
    CHECK_CORRECT_CALLER_THREAD
    sh::ui::FileView* v = (index == -1) ? activeView->control() : _viewcontrols[index];
    v->gotoDir(folder);
}

void sh::ui::qt::QtFilesystemPanel::setCustomWidget(int i, std::shared_ptr<QWidget> w)
{
    QWidget* main = _mainviews[i];
    std::shared_ptr<QWidget> custom = _customwidgets.take(main);
    if (custom)
        main->layout()->removeWidget(custom.get());
    if (w)
    {
        main->layout()->addWidget(w.get());
        QVBoxLayout* l = dynamic_cast<QVBoxLayout*>(main->layout());
        l->setStretch(0, 1);
        l->setStretch(1, 0);
        _customwidgets[main] = w;
    }
}

int sh::ui::qt::QtFilesystemPanel::viewsCount()
{
    CHECK_CORRECT_CALLER_THREAD
    return _views.length();
}

sh::ui::FileView *sh::ui::qt::QtFilesystemPanel::view(int i)
{
    CHECK_CORRECT_CALLER_THREAD
    return _viewcontrols[i];
}

bool sh::ui::qt::QtFilesystemPanel::eventFilter(QObject *object, QEvent *event)
{
    if (event->type() == QEvent::FocusIn && !_skip_eventfilter)
    {
        sh::ui::qt::QtFileView* _w = dynamic_cast<sh::ui::qt::QtFileView*>(object);
        if (_w) {
            activeView = _w->shared_from_this();
            for(int i=0; i<_views.length();i++)
                _views[i]->setBackgroundColor(inactiveColor);
            _w->setBackgroundColor(activeColor);
            emit panelCurrentViewChanged();
            emit panelFolderActivated(_w->node(), false);
            emit panelSelectionChanged();
        }
    }
    return QWidget::eventFilter(object, event);
}

std::shared_ptr<sh::ui::qt::QtFileView> sh::ui::qt::QtFilesystemPanel::viewwidget(int i)
{
    CHECK_CORRECT_CALLER_THREAD
    return _views[i];
}

QList<std::shared_ptr<sh::filesystem::FilesystemNode> > sh::ui::qt::QtFilesystemPanel::selectedNodes()
{
    CHECK_CORRECT_CALLER_THREAD
    return activeView->selectedNodes();
}

void sh::ui::qt::QtFilesystemPanel::_emit_viewmodeChanged(int i)
{
    emit panelViewmodeChanged(i);
}
