/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2020 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <QMenu>
#include <QApplication>
#include <QScrollBar>
#include "actions/actionactionitem.h"
#include "actions/actionsmanager.h"
#include "filesystem/detailcolumn.h"
#include "filesystem/eurl.h"
#include "filesystem/filesystemhandler.h"
#include "filesystem/filesystemhandlerregister.h"
#include "filesystem/filesystemmodel.h"
#include "filesystem/filesystemmodelfileviewproxy.h"
#include "configuration/configurationmanager.h"
#include "settings/settingsmanager.h"
#include "base/threadpool.h"
#include "base/mainthread.h"
#include "tools/dataexchange.h"
#include "tools/visibleviews.h"
#include "ui/qt/qtactionmenu.h"
#include "ui/fileview.h"
#include "ui/columndimensions.h"
#include "qtactionmenuhandler.h"
#include "qtfileview.h"
#include "qtfileviewcontrol.h"
#include "qtmainwindow.h"


sh::ui::qt::QtFileView::QtFileView()
{
    QColor _actColor = sh::ui::qt::QtMainWindow::mainWindow()->palette().color(QPalette::Active, QPalette::Base);
    QColor _dndTargetColor = _actColor.darker(130);
    dndTargetColor = _dndTargetColor.name();
}

sh::ui::qt::QtFileView::~QtFileView() {
}

void sh::ui::qt::QtFileView::configure(sh::ui::qt::QtFileViewControl *viewctrl)
{
    this->_viewctrl = viewctrl;
    this_qobject = dynamic_cast<QObject*>(this);
    this_qwidget = dynamic_cast<QWidget*>(this);
    this_qabstractitemview = dynamic_cast<QAbstractItemView*>(this);
    this_qwidget->installEventFilter(new sh::ui::qt::QtFileView::EventFilter(this_qobject, this, control()));
    this_qabstractitemview->setFrameShape(QFrame::NoFrame);
    if (this_qabstractitemview)
    {
        this_qabstractitemview->setContextMenuPolicy(Qt::CustomContextMenu);
        QObject::connect(this_qabstractitemview, &QWidget::customContextMenuRequested, this_qabstractitemview, [=](QPoint pos){
            auto instantiation = std::make_shared<sh::actions::ActionInstantiation>(selectedNodes(), node());
            auto menu = std::make_shared<QtActionMenu>();
            sh::ui::qt::QtActionMenuHandler menuhandler(instantiation, menu);
            menu->exec(this_qabstractitemview->viewport()->mapToGlobal(pos));
        });
        QObject::connect(this_qabstractitemview, &QAbstractItemView::activated, this_qabstractitemview, [=](){
            control()->emit_nodesActivated(selectedNodes());
        });
    }
    this_qwidget->setSizePolicy(QSizePolicy::Expanding, QSizePolicy::Expanding);
    setSort(viewctrl->sortColumn(), viewctrl->sortOrder());
    viewctrl->gotoDir(viewctrl->node());
}

sh::ui::qt::QtFileView::EventFilter::EventFilter(QObject *parent, QtFileView *fileview, sh::ui::qt::QtFileViewControl* fileviewctrl)
    : QObject(parent), fileview(fileview), fileviewctrl(fileviewctrl)
{
}

bool sh::ui::qt::QtFileView::EventFilter::eventFilter(QObject*, QEvent *event)
{
    if (event->type() == QEvent::KeyPress) {
        QKeyEvent* keyevent = (QKeyEvent*) event;
        if (keyevent->key() == Qt::Key_Backspace) {
            std::shared_ptr<sh::filesystem::FilesystemNode> node = fileview->node();
            if (node->parentnode())
                fileviewctrl->emit_nodesActivated(QList<std::shared_ptr<sh::filesystem::FilesystemNode>>({node->parentnode()}));
        }
        else if (keyevent->key() == Qt::Key_Down && fileview->control()->filemodel()->rowCount() == 1 && !fileview->this_qabstractitemview->selectionModel()->hasSelection()) {
            fileview->this_qabstractitemview->selectAll();
        }
    }
    return false;
}

QThread* sh::ui::qt::QtFileView::thread()
{
    return this_qobject->thread();
}

void sh::ui::qt::QtFileView::gotoDir(std::shared_ptr<sh::filesystem::FilesystemNode> n)
{
    if (_node != n) {
        _goto(n);
        control()->reload(true);
        if (n != sh::filesystem::FilesystemModel::instance()->rootNode())
            _viewctrl->historyTracker()->visitValue(n->eurl());
    }
}

void sh::ui::qt::QtFileView::_goto(std::shared_ptr<filesystem::FilesystemNode> node)
{
    CHECK_CORRECT_CALLER_THREAD
    QAbstractItemModel* oldmodel = control()->filemodel();
    _node = node;
    int myasynccounter = ++_async_gotodir_index;
    if (node) {
        std::weak_ptr<sh::ui::qt::QtFileView> _canary = shared_from_this();
        sh::base::MainThread::dispatcher()->invokeAsync([=](){ // we must do that async, otherwise model machinery kills us in a subtile way (when this is called between beginRemoveRows and endRemoveRows)
            std::shared_ptr<sh::ui::qt::QtFileView> _scanary = _canary.lock();
            if (!_scanary)
                return;
            if (myasynccounter != _async_gotodir_index)
                return;
            QModelIndex index = sh::filesystem::FilesystemModel::instance()->getIndexForNode( node );
            control()->setFilemodel(sh::filesystem::FilesystemModel::instance()->createFileviewProxy(index, _modelwithtooltip, [=](std::shared_ptr<sh::filesystem::FilesystemNode> bestValid) {
                // this happens when this model dies (i.e. directory was removed)
                _goto(nullptr);
                sh::base::MainThread::dispatcher()->invokeAsync([=](){
                    _viewctrl->gotoDir(bestValid);
                });
            }));
            auto _model = control()->filemodel();
            if (_thumbnailsize != -1)
                _model->setThumbnail(true, _thumbnailsize);
            QObject::connect(sh::filesystem::FilesystemModel::instance().get(), &QAbstractItemModel::rowsInserted, _model,
                             [=](const QModelIndex& parent, int a, int b){
                if (index==parent) {
                    for (int i=a;i<=b;i++)
                        node->childnode(i)->requestDetails();
                }
            });
            _model->setSizeFormattingMode(control()->getSizeFormattingMode());
            _model->setHiddenFilesVisible(control()->hiddenFilesVisible());
            QItemSelectionModel* oldsm = this_qabstractitemview->selectionModel();

            _model->sort(_sort_column, _sort_order);
            QObject::connect(_model, &QAbstractItemModel::columnsInserted, [=](QModelIndex, int, int){  // we must reset the sorting, since new columns may change it (within qt)
                sh::base::MainThread::dispatcher()->invokeAsync([=](){
                    if (_model)
                        _model->sort(_sort_column, _sort_order);
                });
            });
            this_qabstractitemview->setModel(_model);
            delete oldsm;
            _goto_modelloaded();
            QObject::connect(this_qabstractitemview->selectionModel(), &QItemSelectionModel::selectionChanged, this_qabstractitemview, [=](const QItemSelection&, const QItemSelection&){
                control()->emit_selectionChanged();
            });
            control()->emit_selectionChanged();
        });
    }
    _goto_flushmodel();
    if (oldmodel)
        delete oldmodel;
}

void sh::ui::qt::QtFileView::setBackgroundColor(QString c)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    bgColor = c;
    onBackgroundColorSet();
}

QObject* sh::ui::qt::QtFileView::as_qobject()
{
    return this_qobject;
}

QWidget* sh::ui::qt::QtFileView::as_qwidget()
{
    return this_qwidget;
}

QAbstractItemView* sh::ui::qt::QtFileView::as_qabstractitemview()
{
    return this_qabstractitemview;
}

QString sh::ui::qt::QtFileView::backgroundColor()
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    return bgColor;
}

void sh::ui::qt::QtFileView::setSizeFormatting(sh::filesystem::SizeFormatting v)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    auto _model = control()->filemodel();
    if (_model)
        _model->setSizeFormattingMode(v);
}

QList<std::shared_ptr<sh::filesystem::FilesystemNode>> sh::ui::qt::QtFileView::selectedNodes()
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    auto _model = control()->filemodel();
    QList<std::shared_ptr<sh::filesystem::FilesystemNode>> nodes;
    foreach (const QModelIndex idx, selectedModelIndices()) {
        if (idx.column()==0)  {
            QModelIndex didx = _model->mapToSource(idx);
            std::shared_ptr<sh::filesystem::FilesystemNode> node = sh::filesystem::FilesystemModel::instance()->getNodeForIndex(didx);
            nodes.append(node);
        }
    }
    return nodes;
}

void sh::ui::qt::QtFileView::setSelection(const QList<std::shared_ptr<sh::filesystem::FilesystemNode>> nodes)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    auto _model = control()->filemodel();
    auto smod = as_qabstractitemview()->selectionModel();
    QItemSelection selection;
    if (_model) {
        int lr = _model->rowCount();
        int lc = _model->columnCount();
        for (int ir=0; ir<lr; ir++)
        {
            QModelIndex mi = _model->index(ir, 0);
            QModelIndex mi2 = _model->index(ir, lc-1);
            QModelIndex nmi = _model->mapToSource(mi);
            if (nmi.isValid()) {
                auto minode = sh::filesystem::FilesystemModel::instance()->getNodeForIndex(nmi);
                if (minode && minode->eurl() && nodes.contains(minode))
                    selection.select(mi, mi2);
            }
        }
    }
    smod->clear();
    smod->select(selection, QItemSelectionModel::Select);
}

void sh::ui::qt::QtFileView::setHiddenFilesVisible(bool value)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    auto _model = control()->filemodel();
    if (_model)
        _model->setHiddenFilesVisible(value);
}

std::shared_ptr<sh::filesystem::FilesystemNode> sh::ui::qt::QtFileView::node()
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    return _node;
}

QList<std::shared_ptr<sh::filesystem::FilesystemNode>> sh::ui::qt::QtFileView::getAllVisibleNodes()
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    auto _model = control()->filemodel();
    QList<std::shared_ptr<sh::filesystem::FilesystemNode>> result;
    if (_model) {
        int lr = _model->rowCount();
        for (int ir=0; ir<lr; ir++) {
            QModelIndex mi = _model->index(ir, 0);
            QModelIndex nmi = _model->mapToSource(mi);
            if (nmi.isValid()) {
                auto minode = sh::filesystem::FilesystemModel::instance()->getNodeForIndex(nmi);
                if (minode && minode->eurl())
                    result.append(minode);
            }
        }
    }
    return result;
}

void sh::ui::qt::QtFileView::evaluateKeyPress(QKeyEvent *event)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    int key = event->key();
    if (key != Qt::Key_Down && key != Qt::Key_Up && key != Qt::Key_Right &&
                    key != Qt::Key_Left && key != Qt::Key_PageDown && key != Qt::Key_PageUp &&
                    key != Qt::Key_Home && key != Qt::Key_End && key != Qt::Key_Control &&
                    key != Qt::Key_Meta && key != Qt::Key_Alt && key != Qt::Key_Shift && key != Qt::Key_Space) {
        sh::actions::ActionsManager::instance()->getActionForShortcut(selectedNodes(), _node, event->key(), event->modifiers(), [=](std::shared_ptr<sh::actions::AbstractActionItem> a){
            sh::actions::ActionActionItem* aaction = qobject_cast<sh::actions::ActionActionItem*>(a.get());
            if (aaction)
                aaction->execute();  // a noop for disabled actions
        });
    }
}

bool sh::ui::qt::QtFileView::dnd_dragEnterEvent(QDragEnterEvent *event)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    if (sh::tools::DataExchange::instance()->containsFileEntries(event->mimeData())) {
        _tmp_bgColor = backgroundColor();
        setBackgroundColor(dndTargetColor);
        event->acceptProposedAction();
        return true;
    }
    else {
        event->dropAction();
        return false;
    }
}

bool sh::ui::qt::QtFileView::dnd_mousePressEvent(QMouseEvent *event)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    int btn = event->button();
    if (btn == Qt::LeftButton) {
        _dragStartPosition = event->pos();
        _dragIsValid = this_qabstractitemview->indexAt(event->pos()).isValid();
    }
    else if (btn == Qt::BackButton || btn == Qt::ForwardButton) {
        auto lst = (btn == Qt::BackButton) ? _viewctrl->historyTracker()->backwardList() : _viewctrl->historyTracker()->forwardList();
        if (!lst.isEmpty()) {
            auto e = lst.first();
            control()->historyTracker()->revisitValue(e.index());
            sh::ui::MainWindow::mainWindow()->jumpToEurl(e.value());
        }
        return true;
    }
    return false;
}

bool sh::ui::qt::QtFileView::dnd_mouseMoveEvent(QMouseEvent *event)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    if ((event->buttons() & Qt::LeftButton)) {
        if ((event->pos() - _dragStartPosition).manhattanLength() >= QApplication::startDragDistance()) {
            QList<std::shared_ptr<sh::filesystem::FilesystemNode>> nodes = selectedNodes();
            if (nodes.length()>0 && _dragIsValid) {
                QDrag *drag = new QDrag(this_qobject);
                QMimeData* mimeData = sh::tools::DataExchange::instance()->getMimeDataFromFilesystemNodes(nodes, sh::tools::DataExchange::DataExchangeType::MOVE);
                drag->setMimeData(mimeData);
                drag->exec(Qt::CopyAction | Qt::MoveAction);
            }
        }
        return true;
    }
    else
        return false;
}

void sh::ui::qt::QtFileView::dnd_dragMoveEvent(QDragMoveEvent *event)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    event->acceptProposedAction();
}

void sh::ui::qt::QtFileView::dnd_dragLeaveEvent(QDragLeaveEvent *event)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    setBackgroundColor(_tmp_bgColor);
    event->accept();
}

bool sh::ui::qt::QtFileView::dnd_dropEvent(QDropEvent *event)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    setBackgroundColor(_tmp_bgColor);
    if (event->source() != this_qobject) {
        std::shared_ptr<sh::actions::AbstractActionItem> action;
        if (event->dropAction()==Qt::CopyAction)
            action = sh::tools::DataExchange::instance()->createCopyAction( event->mimeData(), node()->eurl());
        else // if (event->dropAction()==Qt::MoveAction)
            action = sh::tools::DataExchange::instance()->createMoveAction( event->mimeData(), node()->eurl());
        sh::base::ThreadPool::enqueueForce([=](){
            sh::actions::ActionActionItem* a = qobject_cast<sh::actions::ActionActionItem*>(action.get());
            a->initializeSync();
            a->execute();
        });
        return true;
    }
    return false;
}

QModelIndexList sh::ui::qt::QtFileView::selectedModelIndices()
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    return this_qabstractitemview->selectionModel()->selection().indexes();
}

void sh::ui::qt::QtFileView::_goto_modelloaded()
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
}

void sh::ui::qt::QtFileView::_goto_flushmodel()
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    QItemSelectionModel* oldsm = this_qabstractitemview->selectionModel();
    control()->setFilemodel(nullptr);
    this_qabstractitemview->setModel(nullptr);
    delete oldsm;
}

void sh::ui::qt::QtFileView::focus()
{
    this_qwidget->setFocus();
}

sh::ui::qt::QtFileViewControl* sh::ui::qt::QtFileView::control()
{
    return _viewctrl;
}

void sh::ui::qt::QtFileView::setSort(int column, Qt::SortOrder order)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    auto _model = control()->filemodel();
    _sort_column = column;
    _sort_order = order;
    if (_model)
        _model->sort(column, order);
}
