/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2020 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef QTFILEVIEW_H
#define QTFILEVIEW_H

#include <memory>
#include <QAbstractItemModel>
#include <QDragEnterEvent>
#include <QDrag>
#include <QHeaderView>
#include <QObject>
#include "filesystem/filesystemmodelfileviewproxy.h"


namespace sh{
    namespace filesystem{class FilesystemNode;}
    namespace configuration{class ConfigurationValue;}
    namespace actions{
        class AbstractActionItem;
        namespace common{
            class ActionHistoryNavigateForward;
            class ActionHistoryNavigateBackward;
            class ActionNavigateInHistory;
        }
    }
    namespace ui {
        namespace qt {
            class QtFileViewControl;
            class QtActionMenu;

            /** \brief
             * Abstract base class for views for the contents of one directory.
             */
            class QtFileView : public std::enable_shared_from_this<QtFileView>
            {
                friend class sh::actions::common::ActionHistoryNavigateForward;
                friend class sh::actions::common::ActionHistoryNavigateBackward;
                friend class sh::actions::common::ActionNavigateInHistory;
                friend class sh::ui::qt::QtFileViewControl;
            public:
                /** \brief Constructed only by the infrastructure and made available otherwise.
                 */
                QtFileView();
                virtual ~QtFileView();
                void setBackgroundColor(QString c);
                QString backgroundColor();
                QList<std::shared_ptr<sh::filesystem::FilesystemNode> > selectedNodes();
                void setSelection(const QList<std::shared_ptr<sh::filesystem::FilesystemNode> > nodes);
                std::shared_ptr<sh::filesystem::FilesystemNode> node();
                QList<std::shared_ptr<sh::filesystem::FilesystemNode> > getAllVisibleNodes();
                void setSizeFormatting(sh::filesystem::SizeFormatting v);
                void gotoDir(std::shared_ptr<sh::filesystem::FilesystemNode> n);
                void setHiddenFilesVisible(bool value);
                QThread* thread();
                virtual void configure(sh::ui::qt::QtFileViewControl *viewctrl);
                virtual void focus();
                virtual void setSort(int column, Qt::SortOrder order);
                QObject* as_qobject();
                QWidget* as_qwidget();
                QAbstractItemView* as_qabstractitemview();
                sh::ui::qt::QtFileViewControl* control();

            protected:
                virtual QModelIndexList selectedModelIndices();
                virtual void _goto(std::shared_ptr<sh::filesystem::FilesystemNode> node);
                virtual void _goto_modelloaded();
                virtual void _goto_flushmodel();
                virtual void onBackgroundColorSet() =0;
                void evaluateKeyPress(QKeyEvent *event);
                bool dnd_dragEnterEvent(QDragEnterEvent *event);
                bool dnd_mousePressEvent(QMouseEvent *event);
                bool dnd_mouseMoveEvent(QMouseEvent *event);
                void dnd_dragMoveEvent(QDragMoveEvent *event);
                void dnd_dragLeaveEvent(QDragLeaveEvent *event);
                bool dnd_dropEvent(QDropEvent *event);

                QString bgColor;
                QString dndTargetColor;
                std::shared_ptr<sh::filesystem::FilesystemNode> _node = nullptr;
                QObject* this_qobject;
                QWidget* this_qwidget;
                QAbstractItemView* this_qabstractitemview;
                int _thumbnailsize = -1;
                bool _modelwithtooltip = false;

            private:
                class EventFilter : public QObject
                {
                public:
                    EventFilter(QObject *parent, QtFileView* fileview, sh::ui::qt::QtFileViewControl *fileviewctrl);
                    bool eventFilter(QObject *object, QEvent *event);
                private:
                    QtFileView* fileview;
                    sh::ui::qt::QtFileViewControl* fileviewctrl;
                };

                sh::ui::qt::QtFileViewControl *_viewctrl = nullptr;
                QString _tmp_bgColor;
                QPoint _dragStartPosition;
                int _async_gotodir_index = 0;
                bool _dragIsValid = false;
                int _sort_column = 0;
                Qt::SortOrder _sort_order = Qt::AscendingOrder;
            };

        }
    }
}

#endif // QTFILEVIEW_H
