/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2020 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <QDockWidget>
#include "filesystem/eurl.h"
#include "filesystem/filesystemmodel.h"
#include "filesystem/filesystemhandler.h"
#include "settings/settingsmanager.h"
#include "base/mainthread.h"
#include "base/threaddispatcher.h"
#include "base/threadpool.h"
#include "ui/columndimensions.h"
#include "qtfileiconview.h"
#include "qtfilesystempanel.h"
#include "qtfileview.h"
#include "qtfileviewcontrol.h"
#include "qtsearchpanel.h"


sh::ui::qt::QtFileViewControl::QtFileViewControl(sh::ui::qt::QtFilesystemPanel* panel, int i) :
    sh::ui::FileView(), _panel(panel), i(i)
{
    CHECK_CORRECT_CALLER_THREAD
    _columndimensions = new sh::ui::ColumnDimensions(i);
    _historyTracker = new sh::tools::HistoryTracker<std::shared_ptr<const sh::filesystem::Eurl>>();
    _sortColumn = 0;
    _sortOrder = Qt::AscendingOrder;
}

sh::ui::qt::QtFileViewControl::~QtFileViewControl()
{
    CHECK_CORRECT_CALLER_THREAD
    delete _columndimensions;
    delete _historyTracker;
}

sh::ui::ColumnDimensions* sh::ui::qt::QtFileViewControl::columndimensions()
{
    CHECK_CORRECT_CALLER_THREAD
    return _columndimensions;
}

sh::tools::HistoryTracker<std::shared_ptr<const sh::filesystem::Eurl>>* sh::ui::qt::QtFileViewControl::historyTracker()
{
    CHECK_CORRECT_CALLER_THREAD
    return _historyTracker;
}

sh::ui::FileViewMode sh::ui::qt::QtFileViewControl::viewmode()
{
    CHECK_CORRECT_CALLER_THREAD
    return _viewmode;
}

void sh::ui::qt::QtFileViewControl::setViewmode(FileViewMode mode)
{
    if (_viewmode != mode)
    {
        bool isActive = i == _panel->currentView()->index();
        _viewmode = mode;
        std::shared_ptr<sh::ui::qt::QtFileView> v;
        {
            IGNORE_FORBIDDEN_ASYNC_CALLS
            _panel->buildView(&v, this, false);
        }
        if (isActive)
        {
            sh::base::MainThread::dispatcher()->invokeAsync([=](){
                v->focus(); // seems to fail sometimes without async
            });
        }
        _panel->_emit_viewmodeChanged(i);
        emit viewOptionChanged();
    }
}

void sh::ui::qt::QtFileViewControl::setIndex(int i)
{
    CHECK_CORRECT_CALLER_THREAD
    this->i = i;
    _columndimensions->index = i;
    _columndimensions->store();
}

int sh::ui::qt::QtFileViewControl::index()
{
    CHECK_CORRECT_CALLER_THREAD
    return i;
}

void sh::ui::qt::QtFileViewControl::gotoDir(std::shared_ptr<sh::filesystem::FilesystemNode> n)
{
    CHECK_CORRECT_CALLER_THREAD
    FileView::gotoDir(n);
    bool showSearchPanel = false;
    if (n && n->handler())
        n->handler()->customizeUi(n, &showSearchPanel);
    if (showSearchPanel) {  // TODO # this code currently only supports one additional panel. currently we have just one. we can enhance it when needed.
        QDockWidget* dcustomui = new QDockWidget();
        auto searchpanel = new sh::ui::qt::QtSearchPanel(n);
        dcustomui->setFeatures(QDockWidget::DockWidgetVerticalTitleBar);
        dcustomui->setWindowTitle(tr("Search"));
        dcustomui->setWidget(searchpanel);
        searchpanel->setSizePolicy(QSizePolicy::Ignored, QSizePolicy::Fixed);
        connect(dcustomui, &QObject::destroyed, [=](){
            searchpanel->setParent(0);
        });
        _panel->setCustomWidget(i, std::shared_ptr<QWidget>(dcustomui));
    }
    else
        _panel->setCustomWidget(i, 0);
    _panel->viewwidget(i)->gotoDir(n);
    sh::settings::SettingsManager::instance()->applyPerEurlSettingsToFileView(this);
}

sh::filesystem::SizeFormatting sh::ui::qt::QtFileViewControl::getSizeFormattingMode()
{
    CHECK_CORRECT_CALLER_THREAD
    return _sizeformattingmode;
}

void sh::ui::qt::QtFileViewControl::setSizeFormattingMode(sh::filesystem::SizeFormatting mode)
{
    CHECK_CORRECT_CALLER_THREAD
    _sizeformattingmode = mode;
    _panel->viewwidget(i)->setSizeFormatting(mode);
    emit viewOptionChanged();
}

bool sh::ui::qt::QtFileViewControl::hiddenFilesVisible()
{
    CHECK_CORRECT_CALLER_THREAD
    return _hiddenfilesvisible;
}

void sh::ui::qt::QtFileViewControl::setHiddenFilesVisible(bool v)
{
    CHECK_CORRECT_CALLER_THREAD
    _hiddenfilesvisible = v;
    _panel->viewwidget(i)->setHiddenFilesVisible(v);
    emit viewOptionChanged();
}

void sh::ui::qt::QtFileViewControl::emit_nodesActivated(QList<std::shared_ptr<sh::filesystem::FilesystemNode>> nodes)
{
    emit nodesActivated(nodes);
}

void sh::ui::qt::QtFileViewControl::emit_selectionChanged()
{
    emit selectionChanged();
}

int sh::ui::qt::QtFileViewControl::iconDimension()
{
    CHECK_CORRECT_CALLER_THREAD
    return _icondimension;
}

void sh::ui::qt::QtFileViewControl::setIconDimension(int v)
{
    CHECK_CORRECT_CALLER_THREAD
    _icondimension = v;
    sh::ui::qt::QtFileIconview* fiv = qobject_cast<sh::ui::qt::QtFileIconview*>(_panel->viewwidget(i)->as_qwidget());
    if (fiv)
        fiv->setThumbDimension(1.0 + 0.15 * v);
}

void sh::ui::qt::QtFileViewControl::setSort(int column, Qt::SortOrder order)
{
    _sortColumn = column;
    _sortOrder = order;
    _panel->viewwidget(i)->setSort(column, order);
}

int sh::ui::qt::QtFileViewControl::sortColumn()
{
    return _sortColumn;
}

Qt::SortOrder sh::ui::qt::QtFileViewControl::sortOrder()
{
    return _sortOrder;
}

QList<std::shared_ptr<sh::filesystem::FilesystemNode>> sh::ui::qt::QtFileViewControl::selectedNodes()
{
    CHECK_CORRECT_CALLER_THREAD
    return _panel->viewwidget(i)->selectedNodes();
}

void sh::ui::qt::QtFileViewControl::setSelection(const QList<std::shared_ptr<sh::filesystem::FilesystemNode>> nodes)
{
    CHECK_CORRECT_CALLER_THREAD
    _panel->viewwidget(i)->setSelection(nodes);
}

QList<std::shared_ptr<sh::filesystem::FilesystemNode>> sh::ui::qt::QtFileViewControl::getAllVisibleNodes()
{
    CHECK_CORRECT_CALLER_THREAD
    return _panel->viewwidget(i)->getAllVisibleNodes();
}
