/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2020 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "actions/actionsmanager.h"
#include "actions/separatoractionitem.h"
#include "actions/common/actionbookmarks.h"
#include "actions/common/actiongroups.h"
#include "actions/common/actionnavigation.h"
#include "actions/mainmenu/actionmain.h"
#include "filesystem/filesystemmodeldirectorytreeproxy.h"
#include "filesystem/filesystemmodeldirectorytreeproxyvisibilityenforcements.h"
#include "configuration/configurationmanager.h"
#include "filesystemhandlers/localfilesystemhandler.h"
#include "settings/settingsmanager.h"
#include "base/mainthread.h"
#include "base/shallotprocess.h"
#include "base/threadpool.h"
#include "ui/fileview.h"
#include "qtactionexecutioninfodialog.h"
#include "qtactionexecutioninfopanel.h"
#include "qtactionmenu.h"
#include "qtdialog.h"
#include "qtfiledetailspanel.h"
#include "qtfileview.h"
#include "qtjumpbar.h"
#include "qtmainwindow.h"
#include "qtaboutdialog.h"
#include "qtexceptiondialog.h"
#include "qtfilesystempanel.h"
#include "qtlogviewdialog.h"
#include "qtmanagebookmarksdialog.h"
#include "qtmanageprofilesdialog.h"
#include "qtopenwithdialog.h"
#include "qtstoreprofiledialog.h"
#include "qttoolbarbutton.h"
#include "qttoolbarbuttonhandler.h"
#include "qttuningdialog.h"
#include "qtfilepropertydialog.h"
#include "ui_qtmainwindow.h"
#include "qtuistyle.h"

REGISTER_UI_MAINWINDOW(sh::ui::qt, QtMainWindow)


sh::ui::qt::QtMainWindow::QtMainWindow(QWidget *parent) :
    QMainWindow(parent), sh::ui::MainWindow(),
    ui(new Ui::QtMainWindow)
{
    _qtMainWindow = this;
    ui->setupUi(this);
    QWidget* hspacer = new QWidget(ui->mainToolBar);
    QBoxLayout* hlayout = new QHBoxLayout(hspacer);
    hlayout->addStretch(1);
    hlayout->setMargin(0);
    hspacer->setLayout(hlayout);
    QWidget* vspacer = new QWidget(ui->mainToolBar);
    QBoxLayout* vlayout = new QVBoxLayout(vspacer);
    vlayout->addStretch(1);
    vlayout->setMargin(0);
    vspacer->setLayout(vlayout);
    ui->mainToolBar->addWidget(vspacer);
    _toolbarLeftRightSplitSeparator = ui->mainToolBar->addWidget(hspacer);
    QSize treeViewDefaultIconSize = ui->treeView->iconSize();
    _thumbnailwidth = -1;
    cfgvalFolderTreeIconSize->consumeValue(this, [=](){
        int s = cfgvalFolderTreeIconSize->getConfigValueInt();
        if (s<0)
            ui->treeView->setIconSize(treeViewDefaultIconSize);
        else
            ui->treeView->setIconSize(QSize(s,s));
    });
    cfgvalToolbarLocation->consumeValue(this, [=](){
        QString s = cfgvalToolbarLocation->getConfigValueString();
        if (s != toolbarPosition()) {
            setToolbarPosition(s);
            slot_toolbar_moved();
        }
    });
    cfgvalDetailPanelLocation->consumeValue(this, [=](){
        QString s = cfgvalDetailPanelLocation->getConfigValueString();
        if (s != detailPanelPosition()) {
            setDetailPanelPosition(s);
            slot_detailbar_moved();
        }
    });
    cfgvalDetailPanelDimensions->consumeValue(this, [=](){
        _resizedetailpanel();
    });
    cfgvalDetailPanelThumbnailDimensions->consumeValue(this, [=](){
        _resizethumbnail();
    });
    fspanel = new sh::ui::qt::QtFilesystemPanel(ui->filesystempanel);
    ui->filesystempanel->layout()->addWidget(fspanel);
    jumpbar = new sh::ui::qt::QtJumpBar(this);
    static_cast<QVBoxLayout*>( ui->centralWidget->layout() )->insertWidget(0,jumpbar);
    connect(ui->mainToolBar, SIGNAL(topLevelChanged(bool)), this, SLOT(slot_toolbar_moved()));
    connect(ui->filedetailsdockwidget, SIGNAL(topLevelChanged(bool)), this, SLOT(slot_detailbar_moved()));
    connect(ui->treeView, SIGNAL(collapsed(QModelIndex)), this, SLOT(slot_treeview_collapsedexpanded(QModelIndex)));
    connect(ui->treeView, SIGNAL(expanded(QModelIndex)), this, SLOT(slot_treeview_collapsedexpanded(QModelIndex)));

    ui->splitter->setStretchFactor(0,0);
    ui->splitter->setStretchFactor(1,1);

    _statusbar_fullheight = ui->statusBar->height();
    ui->statusBar->setMinimumHeight(0);

    connect(&_statusbar_timer, &QTimer::timeout, this, &sh::ui::qt::QtMainWindow::slot_statusbartimer);

    ui->filedetailsdockwidget->setAutoFillBackground(true);
    QPalette pal;
    pal.setColor(backgroundRole(), uiStyle()->windowColorHighlighted());
    ui->filedetailsdockwidget->setPalette(pal);
    _dialogManager = std::make_shared<QtDialogManager>();
}

void sh::ui::qt::QtMainWindow::_addPermanentActionToToolbar(std::shared_ptr<sh::actions::SubmenuActionItem> a, bool rightSide, bool preventDefaultAction)
{
    auto btn = new QtToolbarButton(ui->mainToolBar);
    QAction* action;
    if (rightSide)
        action = ui->mainToolBar->addWidget(btn);
    else
        action = ui->mainToolBar->insertWidget(_toolbarLeftRightSplitSeparator, btn);
    auto btnhandler = new sh::ui::qt::QtToolbarButtonHandler(a, preventDefaultAction, btn, action);
    toolbarButtonHandlers.append(btnhandler);
}

void sh::ui::qt::QtMainWindow::_refreshToolbar(std::shared_ptr<sh::actions::ActionInstantiation> instantiation)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    static int _requestcounter = 0;
    int _myrequest = ++_requestcounter;
    sh::actions::ActionsManager::instance()->getActions(instantiation, [=](QList<std::shared_ptr<sh::actions::ActionInstantiation>> selacts, QList<std::shared_ptr<sh::actions::ActionInstantiation>> diracts){
        sh::base::MainThread::dispatcher()->invokeSync([=]() {
            if (_myrequest != _requestcounter)
                return;
            foreach (auto category, sh::actions::ActionCategory::categories())
                actionGroupsActions[category->name()]->setGroupActions(instantiation, selacts, diracts);
        });
    });
}

void sh::ui::qt::QtMainWindow::_setupGlobalShortcut(std::shared_ptr<sh::actions::AbstractActionItem> action)
{
    CHECK_CORRECT_CALLER_THREAD
    sh::actions::ActionActionItem* _action = qobject_cast<sh::actions::ActionActionItem*>(action.get());
    QKeySequence seq = _action->shortcuthint();
    if (!seq.isEmpty()) {
        QShortcut* shortcut = new QShortcut(seq, this);
        shortcut->setContext(Qt::WindowShortcut);
        connect(shortcut, &QShortcut::activated, action.get(), [=](){
            _action->execute();
        });
        connect(action.get(), &QWidget::destroyed, this, [=](){
            delete shortcut;
        });
    }
}

void sh::ui::qt::QtMainWindow::_setupGlobalShortcut_recursive(std::shared_ptr<sh::actions::SubmenuActionItem> submenu)
{
    CHECK_CORRECT_CALLER_THREAD
    foreach(std::shared_ptr<sh::actions::AbstractActionItem> aai, submenu->subitems()) {
        auto itmSubmenu = std::dynamic_pointer_cast<sh::actions::SubmenuActionItem>(aai);
        auto itmAction = std::dynamic_pointer_cast<sh::actions::ActionActionItem>(aai);
        if (itmSubmenu)
            _setupGlobalShortcut_recursive(itmSubmenu);
        else if (itmAction)
            _setupGlobalShortcut(itmAction);
    }
}

void sh::ui::qt::QtMainWindow::initialize()
{
    connect(fspanel, &sh::ui::qt::QtFilesystemPanel::panelViewmodeChanged, this, [=](int){
        _refresh_detailthumbnailvisibility();
    });
    connect(fspanel, &sh::ui::qt::QtFilesystemPanel::panelCurrentViewChanged, this, [=](){
        _refresh_detailthumbnailvisibility();
    });
    connect(fspanel, &sh::ui::qt::QtFilesystemPanel::panelViewOptionChanged, this, [=](int i){
        _emit_fileViewOptionsChanged(i);
    });
    connect(fspanel, &sh::ui::qt::QtFilesystemPanel::panelCurrentViewChanged, this, [=](){
        _emit_currentFileViewChanged();
    });
    connect(fspanel, &sh::ui::qt::QtFilesystemPanel::panelViewCountChanged, this, [=](){
        _emit_fileViewCountChanged();
    });
    _addPermanentActionToToolbar(std::make_shared<sh::actions::common::ActionHistoryNavigateBackward>());
    _addPermanentActionToToolbar(std::make_shared<sh::actions::common::ActionHistoryNavigateForward>());
    _addPermanentActionToToolbar(std::make_shared<sh::actions::common::ActionBookmarks>());
    ui->mainToolBar->insertSeparator(_toolbarLeftRightSplitSeparator);
    foreach (auto category, sh::actions::ActionCategory::categories()) {
        auto agroup = std::make_shared<sh::actions::common::ActionGroups>(category);
        actionGroupsActions[category->name()] = agroup;
        _addPermanentActionToToolbar(agroup, false, true);
    }
    auto toplvlmenu = sh::actions::mainmenu::ActionMain::instance();
    toplvlmenu->initializeAsync([=](){
        sh::base::MainThread::dispatcher()->invokeSync([&](){
            _setupGlobalShortcut_recursive(toplvlmenu);
        });
    });
    _addPermanentActionToToolbar(toplvlmenu, true);
    treemodel = sh::filesystem::FilesystemModelDirectoryTreeProxy::instance().get();
    ui->treeView->setModel(treemodel);
    connect(fspanel, &sh::ui::qt::QtFilesystemPanel::panelFolderActivated, [=](std::shared_ptr<sh::filesystem::FilesystemNode> node, bool realSwitch){
        QModelIndex mi = sh::filesystem::FilesystemModel::instance()->getIndexForNode(node);
        _jumpToIndex(mi, (realSwitch?-1:1));
    });
    connect(fspanel, &sh::ui::qt::QtFilesystemPanel::panelSelectionChanged , [=](){
        auto selectednodes = fspanel->selectedNodes();
        ui->filedetailspanel->setNodes(selectednodes);
        requestDetailThumbnail();
        auto instantiation = std::make_shared<sh::actions::ActionInstantiation>(selectednodes, currentDirectory());
        _refreshToolbar(instantiation);
    });
    connect(ui->treeView->selectionModel(), &QItemSelectionModel::currentChanged , [=](QModelIndex a, QModelIndex){
        if (sh::filesystem::FilesystemModel::instance()->getNodeForIndex(a) != currentDirectory()) {
            QModelIndex t = treemodel->mapToSource(a);
            _jumpToIndex(t, 2);
        }
    });
    connect(jumpbar, &sh::ui::qt::QtJumpBar::nodeChanged, [=](){
        QModelIndex mi = sh::filesystem::FilesystemModel::instance()->getIndexForNode( jumpbar->node() );
        _jumpToIndex(mi, 3);
    });
    connect(jumpbar, &sh::ui::qt::QtJumpBar::eurlRequested , [=](std::shared_ptr<const sh::filesystem::Eurl> eurl){
        jumpToEurl(eurl);
        fspanel->_skip_eventfilter = true;
        jumpbar->setButtonMode();
        fspanel->_skip_eventfilter = false;
        fspanel->viewwidget(fspanel->currentViewIndex())->focus();
    });
    connect(ui->detailssplitter, &QSplitter::splitterMoved, [=](int, int){
        _thumbnailwidth = ui->detailssplitter->sizes().first();
        _thumbnail_resized();
    });
    ui->detailsThumbnail->installEventFilter(this);
    ui->filedetailsdockwidget->installEventFilter(this);
    setEnabled(true);
    MainWindow::initialize();
}

std::shared_ptr<sh::ui::DialogManager> sh::ui::qt::QtMainWindow::dialogManager()
{
    return _dialogManager;
}

bool sh::ui::qt::QtMainWindow::eventFilter(QObject *obj, QEvent *event)
{
    if (obj == ui->detailsThumbnail) {
        if (event->type() == QEvent::Resize && !_request_detailthumbnail_skip)
            requestDetailThumbnail(false);
    }
    else if (obj == ui->filedetailsdockwidget) {
        if (event->type() == QEvent::Resize)
            _detailpanel_resized();
    }
    return QObject::eventFilter(obj, event);
}

void sh::ui::qt::QtMainWindow::_refresh_detailthumbnailvisibility()
{
    if (fspanel->currentView()->viewmode() == sh::ui::FileViewMode::ListView) {
        _thumbnailwidth = std::abs(_thumbnailwidth);
        ui->detailssplitter->setSizes({_thumbnailwidth, 1});
    }
    else {
        _thumbnailwidth = -std::abs(_thumbnailwidth);
        ui->detailssplitter->setSizes({0, 1});
    }
}

void sh::ui::qt::QtMainWindow::requestDetailThumbnail(bool force) {
    QSize dtsize = ui->detailsThumbnail->size();
    int width = dtsize.width();
    int height = dtsize.height();
    if (ui->detailssplitter->orientation() == Qt::Horizontal)
        width = _thumbnailwidth;
    else
        height = _thumbnailwidth;
    sh::ui::MainWindow::requestDetailThumbnail(width, height, [=]() { /* onBeforeRequest */
        ui->detailsThumbnail->setPixmap(QPixmap());
    }, [=](QIcon result){ /* onArrived */
        _request_detailthumbnail_skip = true;
        ui->detailssplitter->setSizes({0, 1});
        _request_detailthumbnail_skip = false;
        if (!result.isNull()) {
            _request_detailthumbnail_skip = true;
            ui->detailssplitter->setSizes({_thumbnailwidth, 1});
            ui->detailsThumbnail->setPixmap(result.pixmap(QSize(width, height)));
            _request_detailthumbnail_skip = false;
        }
    }, force);
}

void sh::ui::qt::QtMainWindow::slot_statusbartimer()
{
    int h = ui->statusBar->minimumHeight();
    int step = _statusbar_targetheight - h;
    if (step == 0)
        _statusbar_timer.stop();
    else {
        int ostep = step;
        step /= 2;
        if (step == 0)
            step= (ostep>0)?1:-1;
        ui->statusBar->setMinimumHeight(h+step);
    }
}

void sh::ui::qt::QtMainWindow::_newToStatusbar_helper(sh::ui::qt::QtActionExecutionInfoPanel* w)
{
    CHECK_CORRECT_CALLER_THREAD
    setStatusBarVisibility(true);
    _statusbar_childs.insert(w);
    connect(w, &QObject::destroyed, this, [=](){
        _statusbar_childs.remove(w);
        setStatusBarVisibility(!_statusbar_childs.isEmpty());
    });
    w->onVisibilityChanged([=](){
        if (w->isVisible())
            _statusbar_childs.insert(w);
        else
            _statusbar_childs.remove(w);
        setStatusBarVisibility(!_statusbar_childs.isEmpty());
    });
}

std::shared_ptr<sh::ui::ActionExecutionInfoPanel> sh::ui::qt::QtMainWindow::addInfoPanel(int position, sh::actions::ActionExecutionInfo* info, QColor color)
{
    auto panel = std::make_shared<sh::ui::qt::QtActionExecutionInfoPanel>(info, color);
    if (position == -1)
        position = statusBar()->actions().count();
    statusBar()->insertWidget(position, panel.get());
    _newToStatusbar_helper(panel.get());
    return panel;
}

void sh::ui::qt::QtMainWindow::setStatusBarVisibility(bool v)
{
    CHECK_CORRECT_CALLER_THREAD
    if ((v && (_statusbar_targetheight == 0)) || (!v && (_statusbar_targetheight != 0))) {
        _statusbar_targetheight = v?_statusbar_fullheight:0;
        if (!_statusbar_timer.isActive())
            _statusbar_timer.start(80);
    }
}

void sh::ui::qt::QtMainWindow::closeEvent (QCloseEvent * event)
{
    if (closeApp())
        event->accept();
    else
        event->ignore();
}

void sh::ui::qt::QtMainWindow::handleCloseAppRejected(QString rejectmsg)
{
    QMessageBox::information(this, tr("Shallot"), rejectmsg);
}

void sh::ui::qt::QtMainWindow::handleClosed()
{
    MainWindow::handleClosed();
    if (fspanel) {
        delete fspanel;
        fspanel = 0;
    }
}

void sh::ui::qt::QtMainWindow::_jumpToIndex(QModelIndex index, int skip)
{
    CHECK_CORRECT_CALLER_THREAD
    std::shared_ptr<sh::filesystem::FilesystemNode> node = _currentDirectory = sh::filesystem::FilesystemModel::instance()->getNodeForIndex(index);
    if (skip != 1) {
        // update file panel
        fspanel->selectFolder(node);
    }
    if (skip != 2) {
        // update tree
        if (treeSticky()) {
            QModelIndex tindex = treemodel->mapFromSource(index);
            if (tindex.isValid())
                ui->treeView->selectionModel()->setCurrentIndex(tindex, QItemSelectionModel::ClearAndSelect);
            else
                ui->treeView->clearSelection();
        }
    }
    if (skip != 3) {
        // update jumpbar
        jumpbar->setNode(node);
    }
    // update actions
    sh::base::MainThread::dispatcher()->invokeAsync([=](){ //async is important here (error handling, ...)
        if (!isClosing()) {
            auto instantiation = std::make_shared<sh::actions::ActionInstantiation>(fspanel->activeView->selectedNodes(), node);
            _refreshToolbar(instantiation);
            slot_toolbar_moved();
        }
    });
    // global stuff
    sh::settings::SettingsManager::instance()->applyPerEurlSettingsToMainWindow();
    ui->filedetailspanel->setNodes(QList<std::shared_ptr<sh::filesystem::FilesystemNode>>());
    setWindowTitle(_getWindowTitle());
    _emit_currentDirectoryChanged();
}

int sh::ui::qt::QtMainWindow::fileViewCount()
{
    return fspanel->viewsCount();
}

void sh::ui::qt::QtMainWindow::addFileView(bool reinitialize)
{
    fspanel->addView(reinitialize);
}

void sh::ui::qt::QtMainWindow::removeFileView(int i)
{
    fspanel->removeView(i);
}

sh::ui::FileView* sh::ui::qt::QtMainWindow::currentFileView()
{
    return fspanel->currentView();
}

sh::ui::FileView* sh::ui::qt::QtMainWindow::getFileView(int i)
{
    return fspanel->view(i);
}

void sh::ui::qt::QtMainWindow::jumpToNode(std::shared_ptr<sh::filesystem::FilesystemNode> node)
{
    CHECK_CORRECT_CALLER_THREAD
    if (node->isAlive()) {
        QModelIndex idx = sh::filesystem::FilesystemModel::instance()->getIndexForNode(node);
        _jumpToIndex(idx, -1);
        fspanel->viewwidget(fspanel->currentViewIndex())->focus();
    }
}

QString sh::ui::qt::QtMainWindow::toolbarPosition()
{
    CHECK_CORRECT_CALLER_THREAD
    Qt::ToolBarArea area = toolBarArea(ui->mainToolBar);
    QString location = "N";
    switch (area) {
    case Qt::LeftToolBarArea:
        location = "W";
        break;
    case Qt::RightToolBarArea:
        location = "E";
        break;
    case Qt::TopToolBarArea:
        location = "N";
        break;
    case Qt::BottomToolBarArea:
        location = "S";
        break;
    default:
        break;
    }
    return location;
}

void sh::ui::qt::QtMainWindow::setToolbarPosition(QString pos)
{
    CHECK_CORRECT_CALLER_THREAD
    Qt::ToolBarArea area;
    if (pos == "S")
        area = Qt::BottomToolBarArea;
    else if (pos == "W")
        area = Qt::LeftToolBarArea;
    else if (pos == "E")
        area = Qt::RightToolBarArea;
    else /* if (pos == "N") */
        area = Qt::TopToolBarArea;
    addToolBar(area, ui->mainToolBar);
}

QString sh::ui::qt::QtMainWindow::detailPanelPosition()
{
    CHECK_CORRECT_CALLER_THREAD
    Qt::DockWidgetArea area = dockWidgetArea(ui->filedetailsdockwidget);
    QString location = "N";
    switch (area) {
    case Qt::LeftDockWidgetArea:
        location = "W";
        break;
    case Qt::RightDockWidgetArea:
        location = "E";
        break;
    case Qt::TopDockWidgetArea:
        location = "N";
        break;
    case Qt::BottomDockWidgetArea:
        location = "S";
        break;
    default:
        break;
    }
    return location;
}

void sh::ui::qt::QtMainWindow::setDetailPanelPosition(QString pos)
{
    CHECK_CORRECT_CALLER_THREAD
    Qt::DockWidgetArea area;
    if (pos == "S")
        area = Qt::BottomDockWidgetArea;
    else if (pos == "W")
        area = Qt::LeftDockWidgetArea;
    else if (pos == "E")
        area = Qt::RightDockWidgetArea;
    else  // if (pos == "N")
        area = Qt::TopDockWidgetArea;
    addDockWidget(area, ui->filedetailsdockwidget);
}

void sh::ui::qt::QtMainWindow::slot_detailbar_moved()
{
    Qt::DockWidgetArea area = dockWidgetArea(ui->filedetailsdockwidget);
    Qt::Orientation orientation = (area == Qt::LeftDockWidgetArea || area == Qt::RightDockWidgetArea)
            ? Qt::Vertical : Qt::Horizontal;
    QString location = detailPanelPosition();
    cfgvalDetailPanelLocation->setConfigValue(location);
    ui->detailssplitter->setOrientation(orientation);
    ui->filedetailspanel->setOrientation(orientation);
    _resizethumbnail();
    _resizedetailpanel();
    _myorientation = orientation;
}

void sh::ui::qt::QtMainWindow::slot_toolbar_moved()
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    QString location = toolbarPosition();
    cfgvalToolbarLocation->setConfigValue(location);
    sh::ui::qt::QtToolbarButton::Position plocation = sh::ui::qt::QtToolbarButton::Position::TOP;
    if (location == "N")
        plocation = sh::ui::qt::QtToolbarButton::Position::TOP;
    else if (location == "E")
        plocation = sh::ui::qt::QtToolbarButton::Position::RIGHT;
    else if (location == "S")
        plocation = sh::ui::qt::QtToolbarButton::Position::BOTTOM;
    else if (location == "W")
        plocation = sh::ui::qt::QtToolbarButton::Position::LEFT;
    foreach(QObject* w, ui->mainToolBar->children()) {
        sh::ui::qt::QtToolbarButton* btn = qobject_cast<sh::ui::qt::QtToolbarButton*>(w);
        if (btn != 0)
            btn->setLocation(plocation);
    }
}

void sh::ui::qt::QtMainWindow::slot_treeview_collapsedexpanded(const QModelIndex &index)
{
    QModelIndex fmindex = qobject_cast<const QAbstractProxyModel*>(index.model())->mapToSource(index);
    std::shared_ptr<sh::filesystem::FilesystemNode> node = sh::filesystem::FilesystemModel::instance()->getNodeForIndex(fmindex);
    sh::filesystem::FilesystemModelDirectoryTreeProxyVisibilityEnforcements::instance()->nodeCollapsedInTree(node);
}

void sh::ui::qt::QtMainWindow::_thumbnail_resized()
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    int w, h;
    _cfgvalDetailPanelThumbnailDimensionsValueDimensions(w, h);
    if (ui->detailssplitter->orientation() == Qt::Horizontal)
        w = _thumbnailwidth;
    else
        h = _thumbnailwidth;
    cfgvalDetailPanelThumbnailDimensions->setConfigValue(QString("%1;%2").arg(w).arg(h));
}

void sh::ui::qt::QtMainWindow::_resizethumbnail()
{
    QString s = cfgvalDetailPanelThumbnailDimensions->getConfigValueString();
    QStringList sl = s.split(";");
    if (sl.length() == 2) {
        bool ok = true;
        int x = (ui->detailssplitter->orientation() == Qt::Horizontal ? sl[0] : sl[1]).toInt(&ok);
        if (ok && x != -1) {
            _thumbnailwidth = x;
            ui->detailssplitter->setSizes({x, 1});
        }
    }
}

void sh::ui::qt::QtMainWindow::_detailpanel_resized()
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    Qt::DockWidgetArea dockarea = dockWidgetArea(ui->filedetailsdockwidget);
    if ((dockarea == Qt::TopDockWidgetArea || dockarea == Qt::BottomDockWidgetArea) != (_myorientation == Qt::Horizontal))
        return;
    QString x = cfgvalDetailPanelDimensions->getConfigValueString();
    int w = -1;
    int h = -1;
    QStringList lx = x.split(";");
    if (lx.length() == 2) {
        bool ok = true;
        w = lx[0].toInt(&ok);
        if (!ok)
            w = -1;
        h = lx[1].toInt(&ok);
        if (!ok)
            h = -1;
    }
    QSize dss = ui->filedetailsdockwidget->size();
    if (dockarea == Qt::TopDockWidgetArea || dockarea == Qt::BottomDockWidgetArea)
        h = dss.height();
    else
        w = dss.width();
    cfgvalDetailPanelDimensions->setConfigValue(QString("%1;%2").arg(w).arg(h));
}

void sh::ui::qt::QtMainWindow::_resizedetailpanel()
{
    QString s = cfgvalDetailPanelDimensions->getConfigValueString();
    QStringList sl = s.split(";");
    if (sl.length() == 2) {
        bool ok = true;
        int x;
        Qt::Orientation ortho;
        Qt::DockWidgetArea dockarea = dockWidgetArea(ui->filedetailsdockwidget);
        if (dockarea == Qt::TopDockWidgetArea || dockarea == Qt::BottomDockWidgetArea) {
            x = sl[1].toInt(&ok);
            ortho = Qt::Vertical;
        }
        else {
            x = sl[0].toInt(&ok);
            ortho = Qt::Horizontal;
        }
        if (ok && x != -1)
            resizeDocks({ui->filedetailsdockwidget}, {x}, ortho);
    }
}

sh::ui::qt::QtMainWindow::~QtMainWindow()
{
    CHECK_CORRECT_CALLER_THREAD
    delete ui;
    foreach(auto tbbh, toolbarButtonHandlers)
        delete tbbh;
    if (_uistyle)
        delete _uistyle;
}

sh::ui::qt::QtMainWindow* sh::ui::qt::QtMainWindow::_qtMainWindow = nullptr;

bool sh::ui::qt::QtMainWindow::tryCreateMainWindow(sh::ui::MainWindow*& mainWindow)
{
    if (sh::base::ShallotProcess::instance()->uiMode() == "qt") {
        sh::ui::qt::QtMainWindow* w = new sh::ui::qt::QtMainWindow();
        w->setAttribute(Qt::WA_DeleteOnClose, false);
        w->setAttribute(Qt::WA_QuitOnClose, false);
        w->show();
        w->setEnabled(false);
        mainWindow = w;
        return true;
    }
    return false;
}

sh::ui::qt::QtMainWindow* sh::ui::qt::QtMainWindow::mainWindow()
{
    return _qtMainWindow;
}

void sh::ui::qt::QtMainWindow::setFileDetailsPanelVisible(bool v)
{
    sh::ui::MainWindow::setFileDetailsPanelVisible(v);
    ui->filedetailsdockwidget->setVisible(v);
}

std::shared_ptr<sh::filesystem::FilesystemNode> sh::ui::qt::QtMainWindow::currentDirectory()
{
    CHECK_CORRECT_CALLER_THREAD
    return _currentDirectory ? _currentDirectory : sh::filesystem::FilesystemModel::instance()->rootNode();
}

void sh::ui::qt::QtMainWindow::setTitlePattern(QString value)
{
    sh::ui::MainWindow::setTitlePattern(value);
    setWindowTitle(_getWindowTitle());
}


void sh::ui::qt::QtMainWindow::setTreeVisible(bool v)
{
    sh::ui::MainWindow::setTreeVisible(v);
    ui->treeView->setVisible(v);
}

void sh::ui::qt::QtMainWindow::jumpbarSetTextMode()
{
    jumpbar->setTextMode();
}

void sh::ui::qt::QtMainWindow::jumpbarSetButtonMode()
{
    jumpbar->setButtonMode();
}

bool sh::ui::qt::QtMainWindow::jumpbarIsTextMode()
{
    return jumpbar->isTextMode();
}

std::shared_ptr<sh::ui::AboutDialog> sh::ui::qt::QtMainWindow::showAboutDialog()
{
    return _dialogManager->createAndShowDialog<QtAboutDialog>();
}

std::shared_ptr<sh::ui::ManageProfilesDialog> sh::ui::qt::QtMainWindow::showManageProfilesDialog()
{
    return _dialogManager->createAndShowDialog<QtManageProfilesDialog>();
}

std::shared_ptr<sh::ui::OpenWithDialog> sh::ui::qt::QtMainWindow::showOpenWithDialog(std::shared_ptr<sh::filesystem::FilesystemNode> node, QList<std::shared_ptr<sh::tools::filetypes::OpenMethod>> openMethods)
{
    return _dialogManager->createAndShowDialog<QtOpenWithDialog>(node, openMethods);
}

std::shared_ptr<sh::ui::StoreProfileDialog> sh::ui::qt::QtMainWindow::showStoreProfileDialog(std::shared_ptr<const sh::filesystem::Eurl> eurl)
{
    return _dialogManager->createAndShowDialog<QtStoreProfileDialog>(eurl);
}

std::shared_ptr<sh::ui::TuningDialog> sh::ui::qt::QtMainWindow::showTuningDialog()
{
    return _dialogManager->createAndShowDialog<QtTuningDialog>();
}

std::shared_ptr<sh::ui::ExceptionDialog> sh::ui::qt::QtMainWindow::showExceptionDialog(QString error1, QString error2, QString details, QString icon, bool mayRetry, bool mayClose, bool mayCancel, bool showLoglabelAndDetails)
{
    return _dialogManager->createAndShowDialog<QtExceptionDialog>(error1, error2, details, icon, mayRetry, mayClose, mayCancel, showLoglabelAndDetails);
}

std::shared_ptr<sh::ui::LogViewDialog> sh::ui::qt::QtMainWindow::showLogViewDialog(QString headtext, QString subheadtxt, sh::base::LogSeverity minseverity)
{
    return _dialogManager->createAndShowDialog<QtLogViewDialog>(headtext, subheadtxt, minseverity);
}

std::shared_ptr<sh::ui::ManageBookmarksDialog> sh::ui::qt::QtMainWindow::showManageBookmarksDialog()
{
    return _dialogManager->createAndShowDialog<QtManageBookmarksDialog>();
}

std::shared_ptr<sh::ui::FilePropertyDialog> sh::ui::qt::QtMainWindow::showFilePropertyDialog(QList<std::shared_ptr<sh::filesystem::FilesystemNode>> nodes)
{
    return _dialogManager->createAndShowDialog<QtFilePropertyDialog>(nodes);
}

std::shared_ptr<sh::ui::ActionExecutionInfoDialog> sh::ui::qt::QtMainWindow::createActionExecutionInfoDialog(sh::actions::ActionExecutionInfo* info)
{
    return std::make_shared<QtActionExecutionInfoDialog>(info);
}

sh::ui::qt::QtUIStyle* sh::ui::qt::QtMainWindow::uiStyle()
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    if (_uistyle == 0)
        _uistyle = new sh::ui::qt::QtUIStyle(this);
    return _uistyle;
}

sh::ui::qt::QtMainWindow::MyFlexibleLabel::MyFlexibleLabel(QWidget* parent)
                : QLabel(parent)
{
}

QSize sh::ui::qt::QtMainWindow::MyFlexibleLabel::sizeHint() const
{
    return QSize(1, 1);
}
