/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2020 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef QTMAINWINDOW_H
#define QTMAINWINDOW_H

#include <functional>
#include <QLabel>
#include <QMainWindow>
#include <QModelIndex>
#include <QSet>
#include <QTimer>
#include "exceptions/exception.h"
#include "base/threaddispatcher.h"
#include "ui/mainwindow.h"
#include "qtdialog.h"
#include "qttoolbarbutton.h"


namespace Ui {
class MainWindow;
}

namespace sh{
    namespace actions {
        namespace common { class ActionGroups; }
        class AbstractActionItem;
        class SubmenuActionItem;
        class ActionInstantiation;
    }
    namespace ui{
        namespace qt{
            class QtUIStyle;
            class QtJumpBar;
            class QtFilesystemPanel;
            class QtActionExecutionInfoPanel;
            class QtToolbarButtonHandler;

            /** \brief
             * The qt main window implementation.
             */
            class QtMainWindow : public QMainWindow, public sh::ui::MainWindow
            {
                friend class QtToolbarButton;
                friend class QtFilesystemPanel;
                Q_OBJECT

            public:
                static bool tryCreateMainWindow(MainWindow*& mainWindow);
                static QtMainWindow* mainWindow();
                explicit QtMainWindow(QWidget *parent = 0);
                ~QtMainWindow();
                void initialize() override;
                std::shared_ptr<DialogManager> dialogManager() override;
                int fileViewCount() override;
                void addFileView(bool reinitialize=false) override;
                void removeFileView(int i) override;
                sh::ui::FileView* currentFileView() override;
                sh::ui::FileView* getFileView(int i) override;
                void jumpToNode(std::shared_ptr<sh::filesystem::FilesystemNode> node) override;
                void setTitlePattern(QString value) override;
                void setTreeVisible(bool v) override;
                void setFileDetailsPanelVisible(bool v) override;
                std::shared_ptr<sh::filesystem::FilesystemNode> currentDirectory() override;
                std::shared_ptr<sh::ui::ActionExecutionInfoPanel> addInfoPanel(int position, sh::actions::ActionExecutionInfo* info, QColor color=QColor()) override;
                void jumpbarSetTextMode() override;
                void jumpbarSetButtonMode() override;
                bool jumpbarIsTextMode() override;
                QString toolbarPosition() override;
                void setToolbarPosition(QString pos) override;
                QString detailPanelPosition() override;
                void setDetailPanelPosition(QString pos) override;
                std::shared_ptr<AboutDialog> showAboutDialog() override;
                std::shared_ptr<ManageProfilesDialog> showManageProfilesDialog() override;
                std::shared_ptr<OpenWithDialog> showOpenWithDialog(std::shared_ptr<sh::filesystem::FilesystemNode> node, QList<std::shared_ptr<sh::tools::filetypes::OpenMethod>> openMethods) override;
                std::shared_ptr<StoreProfileDialog> showStoreProfileDialog(std::shared_ptr<const sh::filesystem::Eurl> eurl) override;
                std::shared_ptr<TuningDialog> showTuningDialog() override;
                std::shared_ptr<LogViewDialog> showLogViewDialog(QString headtext=QString(), QString subheadtxt=QString(), sh::base::LogSeverity minseverity=sh::base::LogSeverity::_DEFAULT) override;
                std::shared_ptr<ManageBookmarksDialog> showManageBookmarksDialog() override;
                std::shared_ptr<FilePropertyDialog> showFilePropertyDialog(QList<std::shared_ptr<sh::filesystem::FilesystemNode>> nodes) override;
                std::shared_ptr<ActionExecutionInfoDialog> createActionExecutionInfoDialog(sh::actions::ActionExecutionInfo* info) override;
                void handleCloseAppRejected(QString rejectmsg) override;
                void handleClosed() override;

                /** \brief
                 * Gets a QtUIStyle instance, which helps for common ui styling.
                 */
                sh::ui::qt::QtUIStyle* uiStyle();

                /** \brief
                 * Needed internally in main window in order to have a label which really does not request any sizes.
                 */
                class MyFlexibleLabel : public QLabel {
                public:
                    MyFlexibleLabel(QWidget* parent = 0);
                    QSize sizeHint() const override;
                };

            protected:
                std::shared_ptr<ExceptionDialog> showExceptionDialog(QString error1, QString error2, QString details, QString icon, bool mayRetry, bool mayClose, bool mayCancel, bool showLoglabelAndDetails) override;
                virtual void closeEvent ( QCloseEvent * event );
                bool eventFilter(QObject *obj, QEvent *event);

            private slots:
                void slot_detailbar_moved();
                void slot_toolbar_moved();
                void slot_statusbartimer();
                void slot_treeview_collapsedexpanded(const QModelIndex &index);

            private:
                void _setupGlobalShortcut(std::shared_ptr<sh::actions::AbstractActionItem> action);
                void _setupGlobalShortcut_recursive(std::shared_ptr<sh::actions::SubmenuActionItem> submenu);
                void _addPermanentActionToToolbar(std::shared_ptr<sh::actions::SubmenuActionItem> a, bool rightSide=false, bool preventDefaultAction=false);
                void _refreshToolbar(std::shared_ptr<sh::actions::ActionInstantiation>);
                void _jumpToNode(std::shared_ptr<sh::filesystem::FilesystemNode> node, int skip);
                void _jumpToIndex(QModelIndex idx, int skip);
                void _newToStatusbar_helper(sh::ui::qt::QtActionExecutionInfoPanel *w);
                void _refresh_detailthumbnailvisibility();
                void _thumbnail_resized();
                void _resizethumbnail();
                void _detailpanel_resized();
                void _resizedetailpanel();
                void requestDetailThumbnail(bool force=true);
                void setStatusBarVisibility(bool v);
                static QtMainWindow* _qtMainWindow;
                Ui::QtMainWindow *ui;
                sh::ui::qt::QtFilesystemPanel* fspanel;
                sh::filesystem::FilesystemModelDirectoryTreeProxy* treemodel = 0;
                sh::ui::qt::QtJumpBar* jumpbar;
                QTimer _statusbar_timer;
                int _statusbar_fullheight;
                int _statusbar_targetheight=0;
                QSet<sh::ui::qt::QtActionExecutionInfoPanel*> _statusbar_childs;
                std::shared_ptr<sh::filesystem::FilesystemNode> _currentDirectory = 0;
                bool _request_detailthumbnail_skip = false;
                int _thumbnailwidth = 0;
                Qt::Orientation _myorientation = Qt::Horizontal;
                sh::ui::qt::QtUIStyle* _uistyle = 0;
                QAction* _toolbarLeftRightSplitSeparator;
                QHash<QString, std::shared_ptr<sh::actions::common::ActionGroups>> actionGroupsActions;
                QList<QtToolbarButtonHandler*> toolbarButtonHandlers;
                std::shared_ptr<QtDialogManager> _dialogManager;

            };

        }
    }
}

#endif // QTMAINWINDOW_H
