/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2020 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <QInputDialog>
#include "filesystem/eurl.h"
#include "filesystem/filesystemnode.h"
#include "base/iconmanager.h"
#include "tools/bookmarkmanager.h"
#include "qtmainwindow.h"
#include "qtuistyle.h"
#include "qtmanagebookmarksdialog.h"
#include "ui_qtmanagebookmarksdialog.h"


sh::ui::qt::QtManageBookmarksDialog::QtManageBookmarksDialog() :
    ui(new Ui::QtManageBookmarksDialog)
{
    ui->setupUi(this);
    ui->btnClose->setSizeByFactor(120);
    ui->btnClose->setIcon(sh::base::IconManager::instance()->getIcon("dialog-close"));
    ui->treeWidget->setFocus();
    ui->btnDelete->setIcon(sh::base::IconManager::instance()->getIcon("list-remove"));
    ui->btnNew->setIcon(sh::base::IconManager::instance()->getIcon("list-add"));
    ui->btnUp->setIcon(sh::base::IconManager::instance()->getIcon("arrow-up"));
    ui->btnDown->setIcon(sh::base::IconManager::instance()->getIcon("arrow-down"));
    ui->btnMove->setIcon(sh::base::IconManager::instance()->getIcon("transform-move"));
    sh::ui::qt::QtMainWindow::mainWindow()->uiStyle()->createSheet().background_liketitlebar().applyTo(ui->panelHead);
    connect(sh::tools::BookmarkManager::instance().get(), &sh::tools::BookmarkManager::changed, this, [=](){
        readBookmarks();
    });
    readBookmarks();
    _refresh_values();
}

sh::ui::qt::QtManageBookmarksDialog::~QtManageBookmarksDialog()
{
    delete ui;
}

void sh::ui::qt::QtManageBookmarksDialog::on_btnClose_clicked()
{
    accept();
}

void sh::ui::qt::QtManageBookmarksDialog::readBookmarks()
{
    ui->treeWidget->clear();
    QTreeWidgetItem* root = new QTreeWidgetItem(ui->treeWidget, QStringList({tr("Bookmarks"), "", "", ""}));
    root->setIcon(0, sh::base::IconManager::instance()->getIcon("folder-bookmark"));
    std::function<void(QList<std::shared_ptr<sh::ui::ManageBookmarksDialog::BM>>, QTreeWidgetItem*)> _readbookmarks = [&](QList<std::shared_ptr<sh::ui::ManageBookmarksDialog::BM>> cbookmarks, QTreeWidgetItem* qnode) {
        foreach (std::shared_ptr<sh::ui::ManageBookmarksDialog::BM> bookmark, cbookmarks) {
            if (bookmark->isFolder()) {
                auto fbookmark = bookmark->asFolder();
                QTreeWidgetItem* qcnode = new QTreeWidgetItem(qnode, QStringList({fbookmark->name, "", "", fbookmark->name}));
                qcnode->setIcon(0, sh::base::IconManager::instance()->getIcon("folder-bookmark"));
                _readbookmarks(fbookmark->children, qcnode);
            }
            else {
                auto bbbookmark = bookmark->asBookmark()->bookmark;
                new QTreeWidgetItem(qnode, QStringList({bbbookmark->label().isEmpty()?bbbookmark->eurl()->asString():bbbookmark->label(),
                                                        bbbookmark->id(), bbbookmark->eurl()->asString(), bbbookmark->label()}));

            }
        }
    };
    _readbookmarks(getBookmarksStructure(), root);
    ui->treeWidget->expandAll();
}

void sh::ui::qt::QtManageBookmarksDialog::_refresh_values()
{
    bool selected = !ui->treeWidget->selectedItems().isEmpty();
    bool leafselected = selected && !ui->treeWidget->selectedItems().first()->data(2, Qt::DisplayRole).toString().isEmpty();
    bool rootselected = selected && ui->treeWidget->selectedItems().first() == ui->treeWidget->topLevelItem(0);
    ui->panelValues->setEnabled(leafselected);
    ui->btnDelete->setEnabled(selected && !rootselected);
    ui->btnUp->setEnabled(leafselected);
    ui->btnDown->setEnabled(leafselected);
    ui->btnMove->setEnabled(leafselected);
    if (!leafselected) {
        ui->edtLabel->clear();
        ui->edtLocation->clear();
    }
}

void sh::ui::qt::QtManageBookmarksDialog::on_treeWidget_itemSelectionChanged()
{
    _refresh_values();
    if (ui->edtLabel->isEnabled()) {
        ui->edtLabel->setText(ui->treeWidget->selectedItems().first()->data(3, Qt::DisplayRole).toString());
        ui->edtLocation->setText(ui->treeWidget->selectedItems().first()->data(2, Qt::DisplayRole).toString());
        ui->btnStore->setEnabled(false);
    }
}

void sh::ui::qt::QtManageBookmarksDialog::on_btnDelete_clicked()
{
    QTreeWidgetItem* item = ui->treeWidget->selectedItems().first();
    if (item->childCount() > 0) {
        QMessageBox::warning(this, tr("Remove entry"), tr("Collections can only be removed if empty."));
    }
    else {
        int btn = QMessageBox::question(this, tr("Remove entry"), tr("Do you really want to delete '%1'?").arg(item->data(0, Qt::DisplayRole).toString()));
        if (btn == QMessageBox::Yes) {
            if (!item->data(2, Qt::DisplayRole).toString().isEmpty())
                sh::tools::BookmarkManager::instance()->removeBookmark(item->data(1, Qt::DisplayRole).toString());
            else
                delete item;
        }
    }
}

void sh::ui::qt::QtManageBookmarksDialog::on_edtLabel_textChanged(const QString&)
{
    ui->btnStore->setEnabled(true);
}

void sh::ui::qt::QtManageBookmarksDialog::on_edtLocation_textChanged(const QString&)
{
    ui->btnStore->setEnabled(true);
}

void sh::ui::qt::QtManageBookmarksDialog::on_btnStore_clicked()
{
    QString id = ui->treeWidget->selectedItems().first()->data(1, Qt::DisplayRole).toString();
    sh::exceptions::Exception::executeGuarded([&](){
        sh::tools::BookmarkManager::instance()->changeBookmark(id, ui->edtLabel->text(), sh::filesystem::Eurl::fromString(ui->edtLocation->text()));
        _selectbookmark(id);
    }, sh::exceptions::ExecuteGuardFlag::RESUMEABLE_PROGRAM_ERROR_END_WITH_USER_FEEDBACK_HERE);
}

void sh::ui::qt::QtManageBookmarksDialog::on_btnNew_clicked()
{
    QMenu menu;
    QFont font = menu.font();
    font.setBold(true);
    QAction* actNewBookmark = menu.addAction(tr("New bookmark"));
    actNewBookmark->setFont(font);
    QAction* actNewCollection = menu.addAction(tr("New collection of bookmarks"));
    QAction* a = menu.exec(ui->btnNew->mapToGlobal(QPoint(0, ui->btnNew->height())));
    QStringList folder;
    QTreeWidgetItem* selecteditem;
    if (!ui->treeWidget->selectedItems().isEmpty()) {
        selecteditem = ui->treeWidget->selectedItems().first();
        if (!selecteditem->data(2, Qt::DisplayRole).toString().isEmpty())
            selecteditem = selecteditem->parent();
        QTreeWidgetItem* _item = selecteditem;
        while (_item != ui->treeWidget->topLevelItem(0)) {
            folder.insert(0, _item->data(3, Qt::DisplayRole).toString());
            _item = _item->parent();
        }
    }
    else
        selecteditem = ui->treeWidget->topLevelItem(0);
    if (a == actNewBookmark) {
        QString id = sh::tools::BookmarkManager::instance()->addBookmark(folder, "", sh::ui::MainWindow::mainWindow()->currentDirectory()->eurl());
        _selectbookmark(id);
        ui->edtLabel->setFocus();
    }
    else if (a == actNewCollection) {
        QString name = QInputDialog::getText(this, tr("New collection"), tr("Please enter the name for the new collection."), QLineEdit::Normal, "");
        name = name.replace("/", "_");
        if (!name.isEmpty()) {
            bool exists = false;
            for (int i=0;i<selecteditem->childCount();i++) {
                if (selecteditem->child(i)->data(3, Qt::DisplayRole).toString() == name) {
                    exists = true;
                    break;
                }
            }
            if (!exists) {
                QTreeWidgetItem* firstNonfolder = 0;
                for(int i=0;i<selecteditem->childCount();i++) {
                    if (!selecteditem->child(i)->data(2, Qt::DisplayRole).toString().isEmpty()) {
                        firstNonfolder = selecteditem->child(i);
                        break;
                    }
                }
                QTreeWidgetItem* newitem;
                if (firstNonfolder) {
                    newitem = new QTreeWidgetItem(QStringList({name, "", "", name}));
                    selecteditem->insertChild(selecteditem->indexOfChild(firstNonfolder), newitem);
                }
                else
                    newitem = new QTreeWidgetItem(selecteditem, QStringList({name, "", "", name}));
                newitem->setIcon(0, sh::base::IconManager::instance()->getIcon("folder-bookmark"));
            }
            QStringList folder;
            folder << name;
            QTreeWidgetItem* x = selecteditem;
            while (x != ui->treeWidget->topLevelItem(0)) {
                folder.insert(0, x->data(3, Qt::DisplayRole).toString());
                x = x->parent();
            }
            ui->treeWidget->expandItem(selecteditem);
            _selectfolder(folder);
        }
    }
}

void sh::ui::qt::QtManageBookmarksDialog::_selectfolder(QStringList name)
{
    QTreeWidgetItem* node = ui->treeWidget->topLevelItem(0);
    while (!name.isEmpty()) {
        QString segm = name.takeFirst();
        bool found = false;
        for (int i=0; i<node->childCount(); i++) {
            QTreeWidgetItem* itm = node->child(i);
            if (itm->data(3, Qt::DisplayRole).toString() == segm) {
                found = true;
                node = itm;
                break;
            }
        }
        if (!found)
            return;
    }
    if (!ui->treeWidget->selectedItems().isEmpty())
        ui->treeWidget->selectedItems().first()->setSelected(false);
    node->setSelected(true);
}

void sh::ui::qt::QtManageBookmarksDialog::_selectbookmark(QString id)
{
    QList<QTreeWidgetItem*> stack;
    stack.append(ui->treeWidget->topLevelItem(0));
    while (!stack.isEmpty()) {
        QTreeWidgetItem* itm = stack.takeFirst();
        if (itm->data(1, Qt::DisplayRole).toString() == id) {
            if (!ui->treeWidget->selectedItems().isEmpty())
                ui->treeWidget->selectedItems().first()->setSelected(false);
            itm->setSelected(true);
            break;
        }
        for(int i=0; i<itm->childCount(); i++)
            stack.append(itm->child(i));
    }
}

void sh::ui::qt::QtManageBookmarksDialog::on_btnUp_clicked()
{
    QString id = ui->treeWidget->selectedItems().first()->data(1, Qt::DisplayRole).toString();
    sh::tools::BookmarkManager::instance()->moveBookmarkUp(id);
    _selectbookmark(id);
}

void sh::ui::qt::QtManageBookmarksDialog::on_btnDown_clicked()
{
    QString id = ui->treeWidget->selectedItems().first()->data(1, Qt::DisplayRole).toString();
    sh::tools::BookmarkManager::instance()->moveBookmarkDown(id);
    _selectbookmark(id);
}

void sh::ui::qt::QtManageBookmarksDialog::on_btnMove_clicked()
{
    QString id = ui->treeWidget->selectedItems().first()->data(1, Qt::DisplayRole).toString();
    QMenu menu;
    QSet<QString> folders;
    folders.insert("/");
    typedef QPair<QTreeWidgetItem*, QString> StackItem;
    QList<StackItem> stack;
    stack.append(StackItem(ui->treeWidget->topLevelItem(0), ""));
    while (!stack.isEmpty()) {
        StackItem stackitem = stack.takeLast();
        QTreeWidgetItem* witem = stackitem.first;
        for (int i=0; i<witem->childCount(); i++) {
            QTreeWidgetItem* cwitem = witem->child(i);
            if (cwitem->data(2, Qt::DisplayRole).toString().isEmpty()) {
                QString folder = stackitem.second+"/"+cwitem->data(3, Qt::DisplayRole).toString();
                stack.append(StackItem(cwitem, folder));
                folders.insert(folder);
            }
        }
    }
    QStringList lfolders = folders.toList();
    qSort(lfolders.begin(), lfolders.end());
    foreach (QString f, lfolders)
        menu.addAction(f);
    QAction* a = menu.exec(ui->btnMove->mapToGlobal(QPoint(0, ui->btnMove->height())));
    if (a) {
        QStringList newfolder = a->text().split('/');
        newfolder.takeFirst();
        if (newfolder.first().isEmpty())
            newfolder.takeFirst();
        sh::tools::BookmarkManager::instance()->moveBookmarkToFolder(id, newfolder);
        _selectbookmark(id);
    }
}
