/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2020 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "filesystem/eurl.h"
#include "settings/settingsmanager.h"
#include "settings/profilenode.h"
#include "settings/setting.h"
#include "base/iconmanager.h"
#include "qtmainwindow.h"
#include "qtuistyle.h"
#include "qtsettinguiframe.h"
#include "qtmanageprofilesdialog.h"
#include "ui_qtmanageprofilesdialog.h"


sh::ui::qt::QtManageProfilesDialog::QtManageProfilesDialog() :
    ui(new Ui::QtManageProfilesDialog)
{
    ui->setupUi(this);
    ui->listView->setEditTriggers(QAbstractItemView::NoEditTriggers);
    ui->listView->setSelectionBehavior(QAbstractItemView::SelectRows);
    ui->listView->setSelectionMode(QAbstractItemView::SingleSelection);
    ui->btnClose->setSizeByFactor(120);
    ui->btnClose->setIcon(sh::base::IconManager::instance()->getIcon("dialog-close"));
    connect (ui->listView, SIGNAL(currentItemChanged(QListWidgetItem*,QListWidgetItem*)), this, SLOT(slot_listViewactivated(QListWidgetItem*,QListWidgetItem*)));
    _createSettingsList();
    _createProfileList();
    ui->listView->setFocus();
    ui->btnDelNode->setIcon(sh::base::IconManager::instance()->getIcon("format-remove-node"));
    ui->btnDelProfile->setIcon(sh::base::IconManager::instance()->getIcon("edit-clear-list"));
    auto uistyle = sh::ui::qt::QtMainWindow::mainWindow()->uiStyle();
    QString wc = uistyle->backgroundColor().name();
    ui->listFrame->setStyleSheet(QString("QFrame{background:%1;}").arg(wc));
    ui->scrollAreaWidgetContents->setStyleSheet(QString("background:%1;").arg(wc));
    ui->label_2->setStyleSheet(QString("font-size:%1pt; margin:8px; color: %2;")
                        .arg(uistyle->fontSizeInPt(115))
                        .arg(uistyle->foregroundColorLighter1().name())
                             );
    ui->panelSelectProfile->setStyleSheet(QString(".QWidget{background:%1;}")
                        .arg(uistyle->windowColorLikeTitlebar().name()));
    ui->lblHead->setStyleSheet(QString("font-size:%1pt;")
                        .arg(uistyle->fontSizeInPt(130)));
}

void sh::ui::qt::QtManageProfilesDialog::keyPressEvent(QKeyEvent *event)
{
    if (event->key() == Qt::Key_Delete)
        on_btnDelNode_clicked();
    else
        QDialog::keyPressEvent(event);
}

void sh::ui::qt::QtManageProfilesDialog::resizeEvent(QResizeEvent* event)
{
    if (event)
        QDialog::resizeEvent(event);
    int w1=1, w2=1;
    if (ui->listView->count()>0) {
        w1=ui->listView->sizeHintForColumn(0);
        w2=width() - w1 - 130;
    }
    ui->splitter->setSizes(QList<int>({w1,w2}));
}

void sh::ui::qt::QtManageProfilesDialog::_createProfileList()
{
    int iselected = -1;
    QString currentProfile = ui->comboBox->currentText();
    QStringList profilelist;
    getProfileList(currentProfile, profilelist, iselected);
    ui->comboBox->clear();
    ui->comboBox->addItems(profilelist);
    if (iselected > -1)
        ui->comboBox->setCurrentText(profilelist[iselected]);
}

sh::ui::qt::QtManageProfilesDialog::~QtManageProfilesDialog()
{
    delete ui;
}

void sh::ui::qt::QtManageProfilesDialog::_createNodesList()
{
    QString profilename = ui->comboBox->currentText();
    _nodes.clear();
    ui->listView->clear();
    foreach(const Node& n, getNodesList(profilename)) {
        QListWidgetItem *item = new QListWidgetItem(n.text, ui->listView);
        ui->listView->addItem(item);
        _nodes[item] = n.nodeId;
        item->setIcon(sh::base::IconManager::instance()->getIcon(n.icon));
    }
    if (_nodes.keys().length() > 0) {
        ui->listView->selectionModel()->setCurrentIndex(ui->listView->model()->index(0, 0), QItemSelectionModel::SelectCurrent);
        resizeEvent(0);
    }
}

void sh::ui::qt::QtManageProfilesDialog::on_comboBox_currentTextChanged(const QString &)
{
    _createNodesList();
}

void sh::ui::qt::QtManageProfilesDialog::_createSettingsList()
{
    QWidget* l = ui->scrollAreaWidgetContents;
    QVBoxLayout* lo = (QVBoxLayout*)l->layout();

    QLayoutItem* it;
    while ((it = lo->takeAt(0)) != 0) {
        if (it->widget())
            delete it->widget();
        delete it;
    }
    auto uistyle = sh::ui::qt::QtMainWindow::mainWindow()->uiStyle();
    int _fontSize1 = uistyle->fontSizeInPt(90);
    QString _fgColor1 = uistyle->foregroundColor().name();
    QString _fgColor2 = uistyle->foregroundColorLighter1().name();
    if (_currentNode)
    {
        QLabel* lbl = new QLabel(l);
        lbl->setWordWrap(true);
        lbl->setText(tr("Those settings are stored in the selected node:"));
        lbl->setStyleSheet(QString("font-size:%1pt; color:%2;").arg(_fontSize1).arg(_fgColor1) );
        lo->addWidget(lbl);
        bool empty = true;
        foreach(QString settingname, _currentNode->getSettingNames())
        {
            std::shared_ptr<sh::settings::Setting> setting = sh::settings::SettingsManager::instance()->getSettingByName(settingname);
            if (setting != 0)
            {
                empty = false;
                QString value = _currentNode->getSettingValue(settingname);
                int onlyinfileview = _currentNode->getSettingOnlyInFileviewIndex(settingname);
                sh::ui::qt::QtSettingUIFrame* frame = new sh::ui::qt::QtSettingUIFrame(setting.get(), false, value,
                                                onlyinfileview!=-1 ? tr("only for file list #%1").arg(onlyinfileview+1) : "", "", l);
                l->layout()->addWidget(frame);
            }
        }
        if (empty)
        {
            lbl = new QLabel(l);
            lbl->setWordWrap(true);
            lbl->setText(tr("(none)"));
            lbl->setStyleSheet(QString("font-size:%1pt; color:%2; margin:4px;").arg(_fontSize1).arg(_fgColor2));
            lo->addWidget(lbl);
        }
        lbl = new QLabel(l);
        lbl->setWordWrap(true);
        lbl->setText(tr("also applies on subfolders:"));
        lbl->setStyleSheet(QString("font-size:%1pt; color:%2;").arg(_fontSize1).arg(_fgColor1));
        lo->addWidget(lbl);
        lbl = new QLabel(l);
        lbl->setWordWrap(true);
        lbl->setText((_currentNode->withSubfolders() ? tr("yes"): tr("no")));
        lbl->setStyleSheet(QString("font-size:%1pt; color:%2; margin:4px;").arg(_fontSize1).arg(_fgColor2));
        lo->addWidget(lbl);
        lbl = new QLabel(l);
        lbl->setWordWrap(true);
        lbl->setText(tr("inherits from other profiles:"));
        lbl->setStyleSheet(QString("font-size:%1pt; color:%2;").arg(_fontSize1).arg(_fgColor1));
        lo->addWidget(lbl);
        lbl = new QLabel(l);
        lbl->setWordWrap(true);
        lbl->setText(_generateProfileListString(_currentNode->inheritsFrom()));
        lbl->setStyleSheet(QString("font-size:%1pt; color:%2; margin:4px;").arg(_fontSize1).arg(_fgColor2));
        lo->addWidget(lbl);
        lo->addStretch(1);
    }
    else
    {
        QLabel* lbl = new QLabel(l);
        lbl->setWordWrap(true);
        lbl->setText(tr("Nothing selected."));
        lbl->setStyleSheet(QString("font-size:%1pt; color:%2;").arg(_fontSize1).arg(_fgColor2));
        lo->addWidget(lbl);
    }
}

void sh::ui::qt::QtManageProfilesDialog::slot_listViewactivated(QListWidgetItem* current,QListWidgetItem*)
{
    if (current)
        _currentNode = sh::settings::SettingsManager::instance()->getNodeByNodeId( _nodes[current] );
    else
        _currentNode = 0;
    ui->btnDelNode->setEnabled(current!=0);
    _createSettingsList();
}

void sh::ui::qt::QtManageProfilesDialog::on_btnClose_clicked()
{
    QtDialog::close();
}

void sh::ui::qt::QtManageProfilesDialog::on_btnDelNode_clicked()
{
    sh::exceptions::Exception::executeGuarded([&](){
        if (_currentNode){
            QString nodeId = _currentNode->nodeId();
            int btn = QMessageBox::question(this, tr("Remove entry"), tr("Do you really want to delete the selected entry from the profile?"));
            if (btn == QMessageBox::Yes) {
                sh::settings::SettingsManager::instance()->removeNode(nodeId);
                _createProfileList();
            }
        }
    }, sh::exceptions::ExecuteGuardFlag::RESUMEABLE_PROGRAM_ERROR_END_WITH_USER_FEEDBACK_HERE);
}

void sh::ui::qt::QtManageProfilesDialog::on_btnDelProfile_clicked()
{
    sh::exceptions::Exception::executeGuarded([&](){
        QString profile = ui->comboBox->currentText();
        int nodecount = _nodes.keys().count();
        QString _q;
        if (nodecount != 1)
            _q = tr("Do you really want to delete the complete profile '%2' with its %1 entries?").arg(nodecount);
        else
            _q = tr("Do you really want to delete the complete profile '%1' with its one entry?");
        int btn = QMessageBox::question(this, tr("Remove profile"), _q.arg(profile));
        if (btn == QMessageBox::Yes) {
            sh::settings::SettingsManager::instance()->removeProfile(profile);
            _createProfileList();
        }
    }, sh::exceptions::ExecuteGuardFlag::RESUMEABLE_PROGRAM_ERROR_END_WITH_USER_FEEDBACK_HERE);
}
