/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2020 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <QInputDialog>
#include "filesystem/eurl.h"
#include "filesystem/filesystemnode.h"
#include "tools/filetypes/filetypemanager.h"
#include "base/iconmanager.h"
#include "qtmainwindow.h"
#include "qtuistyle.h"
#include "qtopenwithdialog.h"
#include "ui_qtopenwithdialog.h"


sh::ui::qt::QtOpenWithDialog::QtOpenWithDialog(std::shared_ptr<filesystem::FilesystemNode> node, QList<std::shared_ptr<sh::tools::filetypes::OpenMethod>> openMethods) :
    OpenWithDialog(node, openMethods), _rememberfordirectory(node->eurl()), ui(new Ui::QtOpenWithDialog)
{
    ui->setupUi(this);
    on_chkRememberDir_toggled(0);
    ui->btnOk->setSizeByFactor(120);
    ui->btnCancel->setIcon(sh::base::IconManager::instance()->getIcon("dialog-close"));
    ui->btnOk->setIcon(sh::base::IconManager::instance()->getIcon("dialog-ok"));
    auto uistyle = sh::ui::qt::QtMainWindow::mainWindow()->uiStyle();
    uistyle->createSheet().background_highlighted().applyTo(ui->listView);
    uistyle->createSheet().background_liketitlebar().applyTo(ui->panelHead);
    uistyle->createSheet().fontsize_header().applyTo(ui->lblHead);
    uistyle->createSheet().fontsize_smaller().applyTo(ui->lblName);
    ui->btnAnotherAncestor->setSizeByFactor(80);
    _listmodel = new QtOpenWithDialogModel(openMethods, this);
    ui->listView->setModel(_listmodel);
    connect(ui->listView->selectionModel(), &QItemSelectionModel::selectionChanged, this, [=](const QItemSelection& sel){
        ui->btnOk->setEnabled(true);
        _chosenMethod = openMethods[sel.indexes().first().row()];
        ui->lblName->setText(_chosenMethod->name);
    });
    ui->listView->setFocus();
}

std::shared_ptr<sh::tools::filetypes::OpenMethod> sh::ui::qt::QtOpenWithDialog::chosenMethod()
{
    return _chosenMethod;
}

bool sh::ui::qt::QtOpenWithDialog::rememberForMimetype()
{
    return ui->chkRememberType->isChecked();
}

std::shared_ptr<const sh::filesystem::Eurl> sh::ui::qt::QtOpenWithDialog::rememberForDirectory()
{
    return (ui->chkRememberDir->isChecked()) ? _rememberfordirectory : 0;
}

bool sh::ui::qt::QtOpenWithDialog::rememberForFile()
{
    return ui->chkRememberFile->isChecked();
}

sh::ui::qt::QtOpenWithDialog::~QtOpenWithDialog()
{
    delete ui;
}

void sh::ui::qt::QtOpenWithDialog::on_btnCancel_clicked()
{
    reject();
}

void sh::ui::qt::QtOpenWithDialog::on_btnOk_clicked()
{
    accept();
}

void sh::ui::qt::QtOpenWithDialog::on_btnSelectOther_clicked()
{
    QString filename = QFileDialog::getOpenFileName(this, tr("Open with"));
    if (!filename.isEmpty()) {
        _chosenMethod = _openMethodForExecutable(filename);
        ui->btnOk->setEnabled(true);
        ui->lblName->setText(_chosenMethod->name);
    }
}

void sh::ui::qt::QtOpenWithDialog::on_listView_activated(const QModelIndex&)
{
    if (ui->listView->selectionModel()->hasSelection())
        accept();
}

sh::ui::qt::QtOpenWithDialogModel::QtOpenWithDialogModel(QList<std::shared_ptr<tools::filetypes::OpenMethod> > methods, QObject *parent)
        : QStringListModel(parent), _methods(methods)
{
    QStringList names;
    foreach(std::shared_ptr<sh::tools::filetypes::OpenMethod> m, _methods)
        names.append(m->name);
    setStringList(names);
}

QVariant sh::ui::qt::QtOpenWithDialogModel::data(const QModelIndex &index, int role) const
{
    if (index.isValid() && role == Qt::DecorationRole) {
        QIcon i;
        i.addPixmap(_methods[index.row()]->icon.pixmap(32, 32, QIcon::Normal, QIcon::On), QIcon::Normal, QIcon::On);
        i.addPixmap(_methods[index.row()]->icon.pixmap(32, 32, QIcon::Active, QIcon::On), QIcon::Active, QIcon::On);
        i.addPixmap(_methods[index.row()]->icon.pixmap(32, 32, QIcon::Selected, QIcon::On), QIcon::Selected, QIcon::On);
        return i;
    }
    else
        return QStringListModel::data(index, role);
}

void sh::ui::qt::QtOpenWithDialog::on_btnAnotherAncestor_clicked()
{
    QString ndir = QInputDialog::getText(this, tr("Another ancestor"), tr("Please specify which directory to store this application choice for."), QLineEdit::Normal, _rememberfordirectory?_rememberfordirectory->asString():"");
    try {
        _rememberfordirectory = sh::filesystem::Eurl::fromString(ndir);
        QString btntxt = _rememberfordirectory->basename();
        if (btntxt.length() > 20)
            btntxt = "..." + btntxt.right(17);
        ui->btnAnotherAncestor->setText(btntxt);
    }
    catch (sh::filesystem::EurlMisformattedException&) {
    }
}

void sh::ui::qt::QtOpenWithDialog::on_chkRememberDir_toggled(bool)
{
    ui->btnAnotherAncestor->setVisible(ui->chkRememberDir->isChecked());
}
