/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2020 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "configuration/configurationmanager.h"
#include "configuration/configurationvaluetype.h"
#include "base/iconmanager.h"
#include "qtlinkbutton.h"
#include "qtmainwindow.h"
#include "qtuistyle.h"
#include "qttuningdialog.h"
#include "ui_qttuningdialog.h"


sh::ui::qt::QtTuningDialog::QtTuningDialog() :
    ui(new Ui::QtTuningDialog)
{
    ui->setupUi(this);
    ui->btnCancel->setSizeByFactor(120);
    auto uistyle = sh::ui::qt::QtMainWindow::mainWindow()->uiStyle();

    ui->btnCancel->setIcon(sh::base::IconManager::instance()->getIcon("dialog-close"));

    uistyle->createSheet()
            .background_highlighted()
            .applyTo(ui->layoutAppear)
            .applyTo(ui->layoutBehav)
            .applyTo(ui->layoutExt)
            .applyTo(ui->layoutInt);
    uistyle->createSheet()
            .background_liketitlebar()
            .applyTo(ui->panelHead);
    uistyle->createSheet()
            .fontsize_header()
            .applyTo(ui->label);
    uistyle->createSheet()
            .fontsize_smaller()
            .applyTo(ui->label_2);

    foreach(std::shared_ptr<sh::configuration::ConfigurationValue> cv, sh::configuration::ConfigurationManager::instance()->getAllConfigurationValues())
    {
        QWidget* par = ui->layoutInt;
        if (cv->category()==sh::configuration::ConfigurationCategory::CategoryGUI)
            par = ui->layoutAppear;
        if (cv->category()==sh::configuration::ConfigurationCategory::CategoryBehavior)
            par = ui->layoutBehav;
        if (cv->category()==sh::configuration::ConfigurationCategory::CategoryExternalTools)
            par = ui->layoutExt;
        if (par) {
            QBoxLayout* layout = static_cast<QBoxLayout*>(par->layout());
            QFrame* box = new QFrame(par);
            boxWidgetsByDesc.append(BoxWidgetByDesc(cv->description(), box));
            box->setStyleSheet(QString(".QFrame{border-radius:10px; background:%1;} .QWidget{background:%1;}").arg(uistyle->windowColorHalfHighlighted().name()));
            QVBoxLayout* blayout = new QVBoxLayout(box);
            QLabel* t = new QLabel(box);
            t->setText(cv->description());
            QWidget* w = new QWidget(box);
            QHBoxLayout* l = new QHBoxLayout();
            QWidget* iw = new QWidget(box);
            QVBoxLayout* il = new QVBoxLayout();
            QLabel* lbl = new QLabel(iw);
            lbl->setWordWrap(true);
            lbl->setSizePolicy(QSizePolicy::Ignored, QSizePolicy::Fixed);
            cv->consumeValue(box, [=](){
                lbl->setText(cv->valueType()->valueDescription(cv->getConfigValueVariant()));
            });
            lbl->setStyleSheet(QString("color: %2;")
                                .arg(uistyle->foregroundColorLighter1().name()));
            sh::ui::qt::QtLinkButton* btnch = new sh::ui::qt::QtLinkButton(w);
            btnch->setText(QObject::tr("change"));
            QObject::connect(btnch, &QToolButton::clicked, [=](){
                do_change(cv);
            });
            if (!cv->longDescription().isEmpty()){
                QLabel* lbl2 = new QLabel(iw);
                lbl2->setText(cv->longDescription());
                lbl2->setStyleSheet(QString("font-size:%1pt;").arg(uistyle->fontSizeInPt(90)));
                lbl2->setWordWrap(true);
                il->addWidget(lbl2);
            }
            bool showChangeHintInOverview = dynamic_cast<sh::configuration::ConfigurationValueTypeLocalFilePath*>(cv->valueType())
                                         || dynamic_cast<sh::configuration::ConfigurationValueTypeBoolean*>(cv->valueType());
            if (showChangeHintInOverview && !cv->changeHint().isEmpty()){
                QLabel* lbl2 = new QLabel(iw);
                lbl2->setText(cv->changeHint());
                lbl2->setStyleSheet(QString("font-size:%1pt;").arg(uistyle->fontSizeInPt(90)));
                lbl2->setWordWrap(true);
                il->addWidget(lbl2);
            }
            il->addWidget(lbl);
            l->addWidget(iw);
            if (cv->mayChange())
                l->addWidget(btnch);
            l->setStretch(0,1);
            l->setStretch(1,0);
            iw->setLayout(il);
            w->setLayout(l);
            l->setMargin(2);
            il->setMargin(0);
            il->setSpacing(5);
            blayout->addWidget(t);
            blayout->addWidget(w);
            blayout->setMargin(10);
            blayout->setSpacing(0);
            box->setLayout(blayout);
            layout->addWidget(box);
        }
    }
    static_cast<QBoxLayout*>(ui->layoutAppear->layout())->addStretch(1);
    static_cast<QBoxLayout*>(ui->layoutBehav->layout())->addStretch(1);
    static_cast<QBoxLayout*>(ui->layoutExt->layout())->addStretch(1);
    static_cast<QBoxLayout*>(ui->layoutInt->layout())->addStretch(1);
    ui->tabWidget->setCurrentIndex(0);
}

QString sh::ui::qt::QtTuningDialog::_changedlglabel(std::shared_ptr<sh::configuration::ConfigurationValue> cv)
{
    QString label = cv->description();
    if (!cv->changeHint().isEmpty())
        label += "\n\n" + cv->changeHint();
    else if (!cv->longDescription().isEmpty())
        label += "\n\n" + cv->longDescription();
    return label;
}

void sh::ui::qt::QtTuningDialog::do_change(std::shared_ptr<sh::configuration::ConfigurationValue> cv)
{
    auto cvt = cv->valueType();
    if (dynamic_cast<sh::configuration::ConfigurationValueTypeLocalFilePath*>(cvt)) {
        QString text = cv->getConfigValueString();
        QString n = QFileDialog::getOpenFileName(this, QObject::tr("Change"), text, QString());
        if (!n.isNull())
            cv->setConfigValue(n);
    }
    else if (dynamic_cast<sh::configuration::ConfigurationValueTypeInteger*>(cvt)) {
        int val = cv->getConfigValueInt();
        QString label = _changedlglabel(cv);
        bool _ok;
        val = QInputDialog::getInt(this, QObject::tr("Change"), label, val, std::numeric_limits<int>::min(), std::numeric_limits<int>::max(), 1, &_ok);
        if (_ok)
            cv->setConfigValue(val);
    }
    else if (dynamic_cast<sh::configuration::ConfigurationValueTypeFloat*>(cvt)) {
        double val = cv->getConfigValueFloat();
        QString label = _changedlglabel(cv);
        bool _ok;
        val = QInputDialog::getDouble(this, QObject::tr("Change"), label, val, std::numeric_limits<double>::min(), std::numeric_limits<double>::max(), 999, &_ok);
        if (_ok)
            cv->setConfigValue(val);
    }
    else if (dynamic_cast<sh::configuration::ConfigurationValueTypeBoolean*>(cvt)) {
        cv->setConfigValue(!cv->getConfigValueBool());
    }
    else {
        QString text = cvt->valueDescription(cv->getConfigValueVariant());
        QString label = _changedlglabel(cv);
        QString n = QInputDialog::getText(this, QObject::tr("Change"), label, QLineEdit::Normal, text);
        if (!n.isNull()){
            QVariant val = cvt->parse(n);
            if (!val.isNull())
                cv->setConfigValue(val);
        }
    }
}

sh::ui::qt::QtTuningDialog::~QtTuningDialog()
{
    delete ui;
}

void sh::ui::qt::QtTuningDialog::on_btnCancel_clicked()
{
    QtDialog::close();
}

void sh::ui::qt::QtTuningDialog::filterVisibility(QString s)
{
    QStringList ls = s.toLower().replace('.', ' ').replace(':', ' ').split(' ');
    foreach(BoxWidgetByDesc bwd, boxWidgetsByDesc) {
        bool vis = true;
        if (!s.isEmpty()) {
            QString klwo = bwd.first.toLower();
            foreach(QString ssp,  ls) {
                if (!ssp.isEmpty() && !klwo.contains(ssp)) {
                    vis = false;
                    break;
                }
            }
        }
        bwd.second->setVisible(vis);
    }
}

void sh::ui::qt::QtTuningDialog::on_lineEditSearchInt_textChanged(const QString &text)
{
    filterVisibility(text);
}

void sh::ui::qt::QtTuningDialog::on_lineEditSearchExt_textChanged(const QString &text)
{
    filterVisibility(text);
}

void sh::ui::qt::QtTuningDialog::on_lineEditSearchBehav_textChanged(const QString &text)
{
    filterVisibility(text);
}

void sh::ui::qt::QtTuningDialog::on_lineEditSearchAppear_textChanged(const QString &text)
{
    filterVisibility(text);
}

void sh::ui::qt::QtTuningDialog::on_tabWidget_currentChanged(int)
{
    ui->lineEditSearchInt->setText("");
    ui->lineEditSearchExt->setText("");
    ui->lineEditSearchBehav->setText("");
    ui->lineEditSearchAppear->setText("");
}
