/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2020 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef SEARCHPANELCONFIGURATION_H
#define SEARCHPANELCONFIGURATION_H

#include <functional>
#include <QObject>


namespace sh {
    namespace ui {

        /** \brief A button in the button bar of a search panel.
         */
        class SearchPanelButton
        {
        public:
            /** \brief
             * @a_infrastructure
             */
            SearchPanelButton();
            virtual ~SearchPanelButton();
            /** \brief Sets the button text.
             * @a_dooverride
             */
            virtual void setButtonText(QString txt) =0;
            /** \brief Sets the currently selected menu item (for buttons with menus).
             * @a_dooverride
             */
            virtual void setMenuSelection(int i) =0;
        };

        /** \brief Abstract base class for editor widgets in a search panel.
         */
        class SearchPanelAbstractEditor
        {
        public:
            /** \brief
             * @a_infrastructure
             */
            SearchPanelAbstractEditor() = default;
            virtual ~SearchPanelAbstractEditor() = default;
            /** \brief Returns if the widget is enabled.
             * @a_dooverride
             */
            virtual bool isWidgetEnabled() =0;
            /** \brief Sets if the widget is enabled.
             * @a_dooverride
             */
            virtual void setWidgetEnabled(bool v) =0;
        };

        /** \brief A single line text editor for search panel usage.
         */
        class SearchPanelTextEditor : public SearchPanelAbstractEditor
        {
        public:
            /** \brief
             * @a_infrastructure
             */
            SearchPanelTextEditor() = default;
            /** \brief Returns the text content.
             * @a_dooverride
             */
            virtual QString textContent() =0;
            /** \brief Sets the text content.
             * @a_dooverride
             */
            virtual void setTextContent(QString s) =0;
            /** \brief Returns the placeholder description text (visible if field is empty).
             * @a_dooverride
             */
            virtual QString placeholderDescription() =0;
            /** \brief Sets the placeholder description text (visible if field is empty).
             * @a_dooverride
             */
            virtual void setPlaceholderDescription(QString s) =0;
        };

        /** \brief A date/time picker for search panel usage.
         */
        class SearchPanelDateTimeEditor : public SearchPanelAbstractEditor
        {
        public:
            /** \brief
             * @a_infrastructure
             */
            SearchPanelDateTimeEditor() = default;
            /** \brief Returns the selected date/time.
             * @a_dooverride
             */
            virtual QDateTime datetime() =0;
            /** \brief Sets the selected date/time.
             */
            virtual void setDatetime(QDateTime s) =0;
        };

        /** \brief A text label for search panel usage.
         */
        class SearchPanelLabelEditor : public SearchPanelAbstractEditor
        {
        public:
            /** \brief
             * @a_infrastructure
             */
            SearchPanelLabelEditor() = default;
            /** \brief Returns the text content.
             * @a_dooverride
             */
            virtual QString textContent() =0;
            /** \brief Sets the text content.
             * @a_dooverride
             */
            virtual void setTextContent(QString s) =0;
            /** \brief Returns the focus proxy widget (which focus gets redirected to in some situations).
             * @a_dooverride
             */
            virtual SearchPanelAbstractEditor* focusProxyEditor() =0;
            /** \brief Sets the focus proxy widget (which focus gets redirected to in some situations).
             * @a_dooverride
             */
            virtual void setFocusProxyEditor(SearchPanelAbstractEditor* w) =0;
        };

        /** \brief A spacer for search panel usage.
         */
        class SearchPanelSpacerEditor : public SearchPanelAbstractEditor
        {
        public:
            /** \brief
             * @a_infrastructure
             */
            SearchPanelSpacerEditor() = default;
        };

        /** \brief A helper for some signal/slot stuff of in SearchPanelConfiguration.
         */
        class _SearchPanelConfiguration_HelperQObject : public QObject
        {
            Q_OBJECT
        };

        /** \brief Configuration for a search panel population.
         *
         * It is populated with ui controls by the chosen search criterion. Later on those controls are
         * used for updating the configuration data.
         */
        class SearchPanelConfiguration
        {
        public:
            /** \brief
             * @a_infrastructure
             */
            SearchPanelConfiguration();
            virtual ~SearchPanelConfiguration();
            /** \brief Adds and returns a button for a menu.
             * @a_dooverride
             */
            virtual SearchPanelButton* addMenuButton(QString text, QStringList menu, std::function<void(int)> onchanged) =0;
            /** \brief Adds and returns a button for an action.
             * @a_dooverride
             */
            virtual SearchPanelButton* addActionButton(QString text, std::function<void()> action) =0;
            /** \brief Adds and returns a text editor.
             * @a_dooverride
             */
            virtual SearchPanelTextEditor* addTextEditor() =0;
            /** \brief Adds and returns a date/time editor.
             * @a_dooverride
             */
            virtual SearchPanelDateTimeEditor* addDateTimeEditor() =0;
            /** \brief Adds and returns a label.
             * @a_dooverride
             */
            virtual SearchPanelLabelEditor* addLabel() =0;
            /** \brief Adds and returns a spacer.
             * @a_dooverride
             */
            virtual SearchPanelSpacerEditor* addSpacer() =0;
            /** \brief Returns the editor widget at a given position.
             * @a_dooverride
             */
            virtual SearchPanelAbstractEditor* getEditorAt(int i) =0;
            /** \brief Sets a handler for panel removal (optionally bound to an owner lifetime).
             */
            void onDestroyed(std::function<void()> fct, QObject* owner=0);
            _SearchPanelConfiguration_HelperQObject myqobject;
        };

    }
}

#endif // SEARCHPANELCONFIGURATION_H
