/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2020 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef STOREPROFILEDIALOG_H
#define STOREPROFILEDIALOG_H

#include "dialog.h"


namespace sh{
    namespace configuration { class ConfigurationValue; }
    namespace settings { class Setting; }
    namespace filesystem { class Eurl; }
    namespace ui {

        /** \brief The 'Save Settings' dialog.
         */
        class StoreProfileDialog : public Dialog
        {
        public:

            /** \brief A storable entry in a 'Save Settings' dialog.
             */
            class SettingEntry
            {
            public:

                /** @a_infrastructure
                 */
                SettingEntry(sh::settings::Setting* setting, int onlyInFileview=-1);

                /** \brief Returns the setting to store.
                 */
                sh::settings::Setting* setting();

                /** \brief Returns the index of the file view to store this setting for (or -1 for all).
                 */
                int fileview();

                /** \brief Returns if this setting is to be stored for a particular file view (instead of for all).
                 */
                bool isForFileview();

            private:
                sh::settings::Setting* _setting;
                int _fileview;
            };

            /**
             * \param eurl: The current directory this dialog shall work on.
             * @a_infrastructure
             */
            StoreProfileDialog(std::shared_ptr<const sh::filesystem::Eurl> eurl);

            /** \brief Returns the current directory this dialog shall work on.
             */
            std::shared_ptr<const sh::filesystem::Eurl> eurl();

            /** \brief Returns the list of settings the user has checked to store.
             */
            virtual QList<SettingEntry> checkedSettings() =0;

            /** \brief Returns the profile the user has chosen to store settings for.
             */
            virtual QString profileName() =0;

            /** \brief Returns if the user has chosen to apply the checked settings also for subdirectories.
             */
            virtual bool withSubDirectories() =0;

            /** \brief Returns the list of profiles the user has chosen to inherit from.
             */
            virtual QStringList inheritsFrom() =0;

            /** \brief Returns if the user has chosen to apply the checked settings for each directory (instead of the current directory).
             */
            virtual bool hasGlobal() =0;

        protected:
            static std::shared_ptr<sh::configuration::ConfigurationValue> cfgvalLevel;

        private:
            std::shared_ptr<const sh::filesystem::Eurl> _eurl;

        };

    }
}

#endif // STOREPROFILEDIALOG_H
