/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2020 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "base/logger.h"
#include "base/mainthread.h"
#include "exceptions/runtimeexception.h"
#include "pocowebserverengine.h"


#if WEBUI_ENGINE == 1

sh::ui::web::PocoNetWebServerEngine::PocoNetWebServerEngineRequest::PocoNetWebServerEngineRequest(QString url, QString clientHost, Poco::Net::HTTPServerRequest* req, Poco::Net::HTTPServerResponse* resp)
        : WebServerEngineRequest(url, clientHost), _req(req), _resp(resp)
{
}

QString sh::ui::web::PocoNetWebServerEngine::PocoNetWebServerEngineRequest::headerData(QString key)
{
    std::string skey = key.toStdString();
    return _req->has(skey) ? QString::fromStdString((*_req)[skey]) : QString();
}

QString sh::ui::web::PocoNetWebServerEngine::PocoNetWebServerEngineRequest::getCookie(QString key)
{
    Poco::Net::NameValueCollection pcookies;
    _req->getCookies(pcookies);
    auto ipc = pcookies.begin();
    auto skey = key.toStdString();
    while (ipc != pcookies.end()) {
        if (ipc->first == skey) {
            return QString::fromStdString(ipc->second);
        }
        ipc++;
    }
    return QString();
}

void sh::ui::web::PocoNetWebServerEngine::PocoNetWebServerEngineRequest::setCookie(QString key, QString value, int maxAgeSecs)
{
    Poco::Net::HTTPCookie pcookie(key.toStdString(), value.toStdString());
    pcookie.setPath("/"); // otherwise we get one cookie for each command
    if (maxAgeSecs >= 0)
        pcookie.setMaxAge(maxAgeSecs);
    _resp->addCookie(pcookie);
}

sh::ui::web::PocoNetWebServerEngine::PocoNetWebServerEngineRequestHandler::PocoNetWebServerEngineRequestHandler(PocoNetWebServerEngine* engine) : engine(engine)
{
}

void sh::ui::web::PocoNetWebServerEngine::PocoNetWebServerEngineRequestHandler::handleRequest(Poco::Net::HTTPServerRequest &req, Poco::Net::HTTPServerResponse &resp)
{
    PocoNetWebServerEngineRequest request(QString::fromStdString(req.getURI()), QString::fromStdString(req.clientAddress().host().toString()), &req, &resp);
    try {
        QString rpath = request.path();
        if (rpath == "/" || rpath == "")
            engine->rootCommand(&request, "poconet");
        else if (rpath.startsWith("/static/"))
            engine->getStaticFile(&request, rpath.mid(7));
        else {
            bool handled = engine->executeCommand(&request, rpath.mid(1));
            if (handled && !request.responseSet())
                request.setResponse(QJsonObject());
        }
    }
    catch(sh::exceptions::Exception& e) {
        request.setResponse((e.name() + ": " + e.message() + " - " + e.details() + "\n\n" + e.callstack()).toLocal8Bit(), "text/plain", sh::ui::web::WebServerEngineRequest::HTTP_INTERNAL_SERVER_ERROR);
    }
    resp.setStatus((Poco::Net::HTTPResponse::HTTPStatus)request.responseCode());
    resp.setContentType(request.responseMimeType().toStdString());
    std::ostream& out = resp.send();
    QByteArray rresponse = request.response();
    char* responsedata = rresponse.data();
    out.write(responsedata, rresponse.length());
    out.flush();
}

sh::ui::web::PocoNetWebServerEngine::PocoNetWebServerEngineRequestHandlerFactory::PocoNetWebServerEngineRequestHandlerFactory(PocoNetWebServerEngine* engine) : engine(engine)
{
}

Poco::Net::HTTPRequestHandler* sh::ui::web::PocoNetWebServerEngine::PocoNetWebServerEngineRequestHandlerFactory::createRequestHandler(const Poco::Net::HTTPServerRequest&)
{
    return new PocoNetWebServerEngineRequestHandler(engine);
}

sh::ui::web::PocoNetWebServerEngine::PocoNetWebServerEngineThread::PocoNetWebServerEngineThread(PocoNetWebServerEngine* engine) : engine(engine)
{
}

void sh::ui::web::PocoNetWebServerEngine::PocoNetWebServerEngineThread::run()
{
    sh::ui::web::PocoNetWebServerEngine::PocoNetWebServerEngineApplication app(engine, this);
    std::vector<std::string> args;
    args.push_back((QCoreApplication::arguments().first().toStdString()));
    app.run(args);
}

void sh::ui::web::PocoNetWebServerEngine::PocoNetWebServerEngineThread::stopServer()
{
    {
        QMutexLocker l1(&mutexRunning);
        running = false;
        cndRunning.wakeAll();
    }
    wait();
}

void sh::ui::web::PocoNetWebServerEngine::PocoNetWebServerEngineThread::waitWhileRunning()
{
    QMutexLocker l1(&mutexRunning);
    while (running)
        cndRunning.wait(&mutexRunning);
}

sh::ui::web::PocoNetWebServerEngine::PocoNetWebServerEngineApplication::PocoNetWebServerEngineApplication(PocoNetWebServerEngine* engine, PocoNetWebServerEngineThread *thread) : engine(engine), thread(thread)
{
}

int sh::ui::web::PocoNetWebServerEngine::PocoNetWebServerEngineApplication::main(const std::vector<std::string>&)
{
    auto _inited = std::function<void()>([&](){
        QMutexLocker _l1(&engine->_initedmutex);
        engine->_inited = true;
        engine->_initedcondition.wakeAll();
    });
    for (int port=engine->minimumPort(); port<=engine->maximumPort(); port++) {
        try {
            Poco::Net::ServerSocket socket;
            Poco::ThreadPool threadpool;
            socket.impl()->bind(Poco::Net::SocketAddress(Poco::Net::IPAddress(), port), false, false);
            socket.impl()->listen(64);
            sh::ui::web::PocoNetWebServerEngine::PocoNetWebServerEngineRequestHandlerFactory factory(engine);
            Poco::Net::HTTPServer server(&factory, threadpool, socket, new Poco::Net::HTTPServerParams);
            engine->_url = QString("http://localhost:%1/").arg(port);
            server.start();
            _inited();
            thread->waitWhileRunning();
            threadpool.stopAll();
            threadpool.joinAll();
            server.stop();
            return 0;
        }
        catch (Poco::Exception& e) {
            continue;
        }
    }
    SH_LOG_ERROR(QString("No free port found between %1 and %2.").arg(engine->minimumPort()).arg(engine->maximumPort()));
    _inited();
    return -1;
}//TODO # !! properly stop web app (see branch 1902_...)?!

sh::ui::web::PocoNetWebServerEngine::PocoNetWebServerEngine(bool withMainModule, int minport, int maxport, QString externalUrl, QString dispatcherUrl, QByteArray wtoken)
                : sh::ui::web::WebServerEngine(withMainModule, minport, maxport, externalUrl, dispatcherUrl, wtoken)
{
    thread = new sh::ui::web::PocoNetWebServerEngine::PocoNetWebServerEngineThread(this);
    thread->start();
    {
        QMutexLocker _l1(&_initedmutex);
        while (!_inited)
            _initedcondition.wait(&_initedmutex);
    }
}

sh::ui::web::PocoNetWebServerEngine::~PocoNetWebServerEngine()
{
    thread->stopServer();
    if (thread)
        delete thread;
}

QUrl sh::ui::web::PocoNetWebServerEngine::url()
{
    return _url;
}

#endif
