/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2020 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "actions/actionexecutioninfo.h"
#include "base/iconmanager.h"
#include "base/mainthread.h"
#include "tools/misc.h"
#include "webactionexecutioninfodialog.h"
#include "webactionmanager.h"
#include "webmainwindow.h"
#include "webserverengine.h"


sh::ui::web::WebActionExecutionInfoDialog::WebActionExecutionInfoDialog(sh::actions::ActionExecutionInfo *info) : sh::ui::ActionExecutionInfoDialog(info)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    static qint64 next_id = 0;
    _id = next_id++;
    _webts_created = sh::ui::web::WebMainWindow::mainWindow()->webserver()->currentTimestamp();
}

qint64 sh::ui::web::WebActionExecutionInfoDialog::id()
{
    return _id;
}

qint64 sh::ui::web::WebActionExecutionInfoDialog::webts_created()
{
    return _webts_created;
}

QString sh::ui::web::WebActionExecutionInfoDialog::details_fromverb()
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    return _details_fv;
}

QString sh::ui::web::WebActionExecutionInfoDialog::details_fromobject()
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    return _details_fob;
}

QString sh::ui::web::WebActionExecutionInfoDialog::details_toverb()
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    return _details_tv;
}

QString sh::ui::web::WebActionExecutionInfoDialog::details_toobject()
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    return _details_tob;
}

QString sh::ui::web::WebActionExecutionInfoDialog::head()
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    return _head;
}

bool sh::ui::web::WebActionExecutionInfoDialog::progress_isDeterminate()
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    return _progress_isDeterminate;
}

quint64 sh::ui::web::WebActionExecutionInfoDialog::progress_done()
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    return _progress_done;
}

quint64 sh::ui::web::WebActionExecutionInfoDialog::progress_all()
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    return _progress_all;
}

QString sh::ui::web::WebActionExecutionInfoDialog::progress_text()
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    return _progress_text;
}

QJsonValue sh::ui::web::WebActionExecutionInfoDialog::userFeedbackAsJsonValue()
{
    QMutexLocker _l1(&_currentUserFeedbackMutex);
    if (_currentUserFeedback) {
        QJsonObject result;
        result = QJsonObject::fromVariantMap(*_currentUserFeedback);
        result["id"] = _currentUserFeedback->id;
        return result;
    }
    else
        return QJsonValue::Undefined;
}

void sh::ui::web::WebActionExecutionInfoDialog::_triggerChanged() {
    sh::ui::web::WebMainWindow::mainWindow()->actionManager()->_triggerActionuiChanged();
}

void sh::ui::web::WebActionExecutionInfoDialog::setDetails(QString fv, QString fob, QString tv, QString tob)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    _details_fv = fv;
    _details_fob = fob;
    _details_tv = tv;
    _details_tob = tob;
    _triggerChanged();
}

void sh::ui::web::WebActionExecutionInfoDialog::setHead(QString txt)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    _head = txt;
    _triggerChanged();
}

void sh::ui::web::WebActionExecutionInfoDialog::setProgress(bool isDeterminate, quint64 done, quint64 all, QString text)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    _progress_isDeterminate = isDeterminate;
    _progress_done = done;
    _progress_all = all;
    _progress_text = text;
    _triggerChanged();
}

void sh::ui::web::WebActionExecutionInfoDialog::_handleUserFeedback(QString kind, std::shared_ptr<UserFeedback> userfeedback)
{
    static qint64 next_userfeedback_id = 0;
    {
        QMutexLocker _l1(&_currentUserFeedbackMutex);
        userfeedback->id = next_userfeedback_id++;
        userfeedback->insert("view", "shwebui.StatusBarPanelDialogUserFeedback_"+kind);
        _currentUserFeedback = userfeedback;
        bool manualInNeeded = _info->isManualInterventionNeeded();
        _info->setManualInterventionNeeded(true);
        _triggerChanged();
        while (!userfeedback->answered)
            _currentUserFeedbackChangedCondition.wait(&_currentUserFeedbackMutex);
        _info->setManualInterventionNeeded(manualInNeeded);
    }
}

int sh::ui::web::WebActionExecutionInfoDialog::messageBox(QString text, QList<QString> answers, QString icon, int defaultanswer, int cancelanswer, QList<QString> answericons)
{
    auto userfeedback = std::make_shared<UserFeedback>();
    userfeedback->insert("text", text);
    userfeedback->insert("answers", QJsonArray::fromStringList(answers));
    userfeedback->insert("icon", iconToBase64Src(icon, 64));
    userfeedback->insert("defaultanswer", defaultanswer);
    userfeedback->insert("cancelanswer", cancelanswer);
    userfeedback->insert("answericons", QJsonArray::fromVariantList(iconsToBase64Srcs(answericons, 64)));
    _handleUserFeedback("MessageBox", userfeedback);
    return userfeedback->value("answer").toInt();
}

int sh::ui::web::WebActionExecutionInfoDialog::inputBox(QString text, QList<QString> answers, QString *value, QString icon, int defaultanswer, int cancelanswer, int valuePreselectFrom, int valuePreselectTo)
{
    auto userfeedback = std::make_shared<UserFeedback>();
    userfeedback->insert("text", text);
    userfeedback->insert("answers", QJsonArray::fromStringList(answers));
    userfeedback->insert("value", *value);
    userfeedback->insert("valuePreselectFrom", valuePreselectFrom);
    userfeedback->insert("valuePreselectTo", valuePreselectTo);
    userfeedback->insert("icon", iconToBase64Src(icon, 64));
    userfeedback->insert("defaultanswer", defaultanswer);
    userfeedback->insert("cancelanswer", cancelanswer);
    userfeedback->insert("answericons", QJsonArray());
    _handleUserFeedback("InputBox", userfeedback);
    *value = userfeedback->value("value").toString();
    return userfeedback->value("answer").toInt();
}

int sh::ui::web::WebActionExecutionInfoDialog::multilineInputBox(QString text, QList<QString> answers, QString *value, QString icon, int defaultanswer, int cancelanswer)
{
    auto userfeedback = std::make_shared<UserFeedback>();
    userfeedback->insert("text", text);
    userfeedback->insert("answers", QJsonArray::fromStringList(answers));
    userfeedback->insert("value", *value);
    userfeedback->insert("icon", iconToBase64Src(icon, 64));
    userfeedback->insert("defaultanswer", defaultanswer);
    userfeedback->insert("cancelanswer", cancelanswer);
    userfeedback->insert("answericons", QJsonArray());
    _handleUserFeedback("MultilineInputBox", userfeedback);
    *value = userfeedback->value("value").toString();
    return userfeedback->value("answer").toInt();
}

int sh::ui::web::WebActionExecutionInfoDialog::simpleChooserGridform(QString text, GridformEntries* entries, QStringList answers, int defaultanswer, int cancelanswer)
{
    auto userfeedback = std::make_shared<UserFeedback>();
    userfeedback->insert("text", text);
    userfeedback->insert("answers", QJsonArray::fromStringList(answers));
    userfeedback->insert("defaultanswer", defaultanswer);
    userfeedback->insert("cancelanswer", cancelanswer);
    userfeedback->insert("answericons", QJsonArray());
    QJsonArray jentries;
    foreach(auto entry, entries->list) {
        QJsonObject jentry;
        jentry["label"] = entry->label;
        QJsonArray jchoices;
        foreach(auto choice, entry->choices.list) {
            QJsonObject jchoice;
            jchoice["label"] = choice->label;
            jchoice["withText"] = !choice->text.isNull();
            jchoice["text"] = choice->text;
            jchoices.append(jchoice);
        }
        jentry["choices"] = jchoices;
        jentry["selected_choice"] = entry->choices.selected;
        jentries.append(jentry);
    }
    userfeedback->insert("entries", jentries);
    _handleUserFeedback("SimpleChooserGridform", userfeedback);
    auto jentries2 = userfeedback->value("entries").toList();
    for(int ie=0; ie<entries->list.length(); ie++) {
        auto nentry = entries->list[ie];
        auto jentry = jentries2[ie].toMap();
        for(int ic=0; ic<nentry->choices.list.length(); ic++) {
            auto nchoice = nentry->choices.list[ic];
            auto jchoice = jentry["choices"].toList()[ic].toMap();
            if (!nchoice->text.isNull())
                nchoice->text = jchoice["text"].toString();
        }
        nentry->choices.selected = jentry["selected_choice"].toInt();
    }
    return userfeedback->value("answer").toInt();
}

bool sh::ui::web::WebActionExecutionInfoDialog::credentialsDialog(QString text, bool showDomain, bool showUsername, bool showPassword, bool showAnonymous, bool showRemember, QString* domain, QString* username, QString* password, bool* isAnonymous, bool* isRemember)
{
    auto userfeedback = std::make_shared<UserFeedback>();
    userfeedback->insert("text", text);
    userfeedback->insert("showDomain", showDomain);
    userfeedback->insert("showUsername", showUsername);
    userfeedback->insert("showPassword", showPassword);
    userfeedback->insert("showAnonymous", showAnonymous);
    userfeedback->insert("showRemember", showRemember);
    userfeedback->insert("domain", *domain);
    userfeedback->insert("username", *username);
    userfeedback->insert("password", *password);
    userfeedback->insert("isAnonymous", *isAnonymous);
    userfeedback->insert("isRemember", *isRemember);
    _handleUserFeedback("CredentialsDialog", userfeedback);
    *domain = userfeedback->value("domain").toString();
    *username = userfeedback->value("username").toString();
    *password = userfeedback->value("password").toString();
    *isAnonymous = userfeedback->value("isAnonymous").toBool();
    *isRemember = userfeedback->value("isRemember").toBool();
    return userfeedback->value("answer").toBool();
}

bool sh::ui::web::WebActionExecutionInfoDialog::unixPermissionsDialog(bool* userMayRead, bool* userMayWrite, bool* userMayExecute, bool* groupMayRead, bool* groupMayWrite, bool* groupMayExecute, bool* othersMayRead, bool* othersMayWrite, bool* othersMayExecute, bool* sticky, bool* setuid, bool* setgid, QStringList users, QStringList groups, QString* ownerUser, QString* ownerGroup)
{
    auto userfeedback = std::make_shared<UserFeedback>();
    userfeedback->insert("userMayRead", *userMayRead);
    userfeedback->insert("userMayWrite", *userMayWrite);
    userfeedback->insert("userMayExecute", *userMayExecute);
    userfeedback->insert("groupMayRead", *groupMayRead);
    userfeedback->insert("groupMayWrite", *groupMayWrite);
    userfeedback->insert("groupMayExecute", *groupMayExecute);
    userfeedback->insert("othersMayRead", *othersMayRead);
    userfeedback->insert("othersMayWrite", *othersMayWrite);
    userfeedback->insert("othersMayExecute", *othersMayExecute);
    userfeedback->insert("sticky", *sticky);
    userfeedback->insert("setuid", *setuid);
    userfeedback->insert("setgid", *setgid);
    userfeedback->insert("users", users);
    userfeedback->insert("groups", groups);
    userfeedback->insert("ownerUser", *ownerUser);
    userfeedback->insert("ownerGroup", *ownerGroup);
    _handleUserFeedback("UnixPermissionsDialog", userfeedback);
    *userMayRead = userfeedback->value("userMayRead").toBool();
    *userMayWrite = userfeedback->value("userMayWrite").toBool();
    *userMayExecute = userfeedback->value("userMayExecute").toBool();
    *groupMayRead = userfeedback->value("groupMayRead").toBool();
    *groupMayWrite = userfeedback->value("groupMayWrite").toBool();
    *groupMayExecute = userfeedback->value("groupMayExecute").toBool();
    *othersMayRead = userfeedback->value("othersMayRead").toBool();
    *othersMayWrite = userfeedback->value("othersMayWrite").toBool();
    *othersMayExecute = userfeedback->value("othersMayExecute").toBool();
    *sticky = userfeedback->value("sticky").toBool();
    *setuid = userfeedback->value("setuid").toBool();
    *setgid = userfeedback->value("setgid").toBool();
    *ownerUser = userfeedback->value("ownerUser").toString();
    *ownerGroup = userfeedback->value("ownerGroup").toString();
    return userfeedback->value("answer").toBool();
}

QByteArray sh::ui::web::WebActionExecutionInfoDialog::iconToBase64Src(QString icon, int sizeInPt)
{
    QByteArray result;
    if (!icon.isEmpty())
        result = sh::tools::Misc::iconToBase64SrcEncoding(base::IconManager::instance()->getIcon(icon), sizeInPt);
    return result;
}

QList<QVariant> sh::ui::web::WebActionExecutionInfoDialog::iconsToBase64Srcs(QStringList icons, int sizeInPt)
{
    QList<QVariant> result;
    foreach (QString icon, icons)
        result.append(iconToBase64Src(icon, sizeInPt));
    return result;
}
