/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2020 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "actions/actioncategory.h"
#include "actions/actionsmanager.h"
#include "actions/headeractionitem.h"
#include "actions/separatoractionitem.h"
#include "actions/submenuactionitem.h"
#include "actions/common/actionbookmarks.h"
#include "actions/common/actiongroups.h"
#include "actions/mainmenu/actionmain.h"
#include "actions/actionexecutioninfo.h"
#include "base/iconmanager.h"
#include "base/mainthread.h"
#include "tools/misc.h"
#include "webactionmanager.h"
#include "webactionexecutioninfodialog.h"
#include "webactionexecutioninfopanel.h"
#include "webserverengine.h"


sh::ui::web::WebActionManager::WebActionManager()
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    _triggerActionuiChangedTimer.setInterval(70);
    _triggerActionuiChangedTimer.setSingleShot(true);
    connect(&_triggerActionuiChangedTimer, &QTimer::timeout, [=](){
        webserver()->triggerEvent("action/ui_changed");
    });
    _triggerToolbarRefreshTimer.setInterval(100);
    _triggerToolbarRefreshTimer.setSingleShot(true);
    connect(&_triggerToolbarRefreshTimer, &QTimer::timeout, [=](){
        webserver()->triggerEvent("toolbar/refresh");
    });
}

void sh::ui::web::WebActionManager::initialize()
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    auto bookmarkaction = std::make_shared<sh::actions::common::ActionBookmarks>();//TODO# adding a new bookmarks leads to a ghost icon in menu (maybe bad example once separators are implemented)
    _permanentToolbarActions["_bookmarks"] = bookmarkaction;
    bookmarkaction->initializeAsync();
    for(auto category : sh::actions::ActionCategory::categories()) {
        auto paction = _permanentToolbarActions[category->name()] = std::make_shared<sh::actions::common::ActionGroups>(category);
        paction->initializeAsync();
    }
    auto toplvlmenu = sh::actions::mainmenu::ActionMain::instance();
    toplvlmenu->initializeAsync([=](){ });
    _permanentToolbarActions[toplvlmenu->metaObject()->className()] = toplvlmenu;
    foreach (auto a, _permanentToolbarActions.values())
        _observeToolbarAction(a.get());
}

void sh::ui::web::WebActionManager::refreshToolbar(std::shared_ptr<sh::actions::ActionInstantiation> instantiation)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    static int _requestcounter = 0;
    int _myrequest = ++_requestcounter;
    sh::actions::ActionsManager::instance()->getActions(instantiation, [=](QList<std::shared_ptr<sh::actions::ActionInstantiation>> selacts, QList<std::shared_ptr<sh::actions::ActionInstantiation>> diracts){
        sh::base::MainThread::dispatcher()->invokeSync([=]() {
            if (_myrequest != _requestcounter)
                return;
            for (auto category : sh::actions::ActionCategory::categories())
                std::dynamic_pointer_cast<sh::actions::common::ActionGroups>(_permanentToolbarActions[category->name()])->setGroupActions(instantiation, selacts, diracts);
        });
    });
}

void sh::ui::web::WebActionManager::_observeToolbarAction(sh::actions::AbstractActionItem* a)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    disconnect(a, 0, this, 0);
    connect(a, &sh::actions::AbstractActionItem::changed, this, [=](){
        if (!_triggerToolbarRefreshTimer.isActive())
            _triggerToolbarRefreshTimer.start();
    });
    auto sa = dynamic_cast<sh::actions::SubmenuActionItem*>(a);
    if (sa) {
        foreach (auto ssa, sa->subitems())
            _observeToolbarAction(ssa.get());
        connect(sa, &sh::actions::SubmenuActionItem::subitemsChanged, this, [=](){
            if (!_triggerToolbarRefreshTimer.isActive())
                _triggerToolbarRefreshTimer.start();
            foreach (auto ssa, sa->subitems())
                _observeToolbarAction(ssa.get());
        });
    }
}

bool sh::ui::web::WebActionManager::executeCommand(WebServerEngineRequest* request, QString command)
{
    return executeCommand_byQObjectReflection(this, request, command, "action");
}

std::shared_ptr<sh::ui::web::WebActionExecutionInfoPanel> sh::ui::web::WebActionManager::addInfoPanel(int position, sh::actions::ActionExecutionInfo* info, QColor color)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    auto panel = std::shared_ptr<WebActionExecutionInfoPanel>(new WebActionExecutionInfoPanel(info, color),
                                       [=](WebActionExecutionInfoPanel *pi) {
        sh::base::MainThread::dispatcher()->invokeSync([&](){
            delete pi;
            for(int i=0; i<_infopanels.length(); i++)
                if (_infopanels[i] == pi) {
                    _infopanels.removeAt(i);
                    _triggerActionuiChanged();
                    break;
                }
        });
    });
    panel->onVisibilityChanged([=](){
        _triggerActionuiChanged();
    });
    if (position == -1)
        position = _infopanels.length();
    _infopanels.insert(position, panel.get());
    _triggerActionuiChanged();
    return panel;
}

std::shared_ptr<sh::ui::web::WebActionExecutionInfoDialog> sh::ui::web::WebActionManager::addActionExecutionInfoDialog(sh::actions::ActionExecutionInfo* info)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    std::shared_ptr<WebActionExecutionInfoDialog> dialog = std::shared_ptr<WebActionExecutionInfoDialog>(new WebActionExecutionInfoDialog(info),
                                       [=](WebActionExecutionInfoDialog *pi) {
        sh::base::MainThread::dispatcher()->invokeSync([&](){
            delete pi;
            for(int i=0; i<_infodialogs.length(); i++)
                if (_infodialogs[i] == pi) {
                    _infodialogs.removeAt(i);
                    _triggerActionuiChanged();
                    break;
                }
        });
    });
    int position = _infodialogs.length();
    _infodialogs.insert(position, dialog.get());
    _triggerActionuiChanged();
    return dialog;
}

void sh::ui::web::WebActionManager::_triggerActionuiChanged(/*TODO# this single and id-less event is too coarse, since this can break usage in some situations by havoc refreshing everything)*/)
{
    sh::base::MainThread::dispatcher()->invokeSync([&](){
        if (!_triggerActionuiChangedTimer.isActive())
            _triggerActionuiChangedTimer.start();
    });
}

void sh::ui::web::WebActionManager::cmd__answer_userfeedback__M(WebServerEngineRequest* request)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    QString updates = request->data("updates");
    auto jupdates = QJsonDocument::fromJson(updates.toLocal8Bit()).object();
    qint64 id = jupdates["id"].toVariant().toLongLong();
    foreach (auto dlg, _infodialogs) {
        QMutexLocker _l1(&dlg->_currentUserFeedbackMutex);
        if (dlg->_currentUserFeedback && dlg->_currentUserFeedback->id == id) {
            foreach (QString kupdate, jupdates.keys())
                dlg->_currentUserFeedback->insert(kupdate, jupdates[kupdate].toVariant());
            dlg->_currentUserFeedback->answered = true;
            dlg->_currentUserFeedbackChangedCondition.wakeAll();
            break;
        }
    }
}


void sh::ui::web::WebActionManager::cmd__execute(WebServerEngineRequest* request)
{
    QStringList actionpath;
    QList<QPair<QString, std::shared_ptr<sh::actions::AbstractActionItem>>> subactions;
    auto act = resolveActionpath(request->data("actionpath"), &actionpath, &subactions);
    std::shared_ptr<sh::actions::ActionActionItem> actaa = std::dynamic_pointer_cast<sh::actions::ActionActionItem>(act);
    if (actaa)
        actaa->execute();
}

QRegularExpression sh::ui::web::WebActionManager::reActionTextAccel("&([^&])");

void sh::ui::web::WebActionManager::cmd__get(WebServerEngineRequest* request)
{
    QJsonArray result;
    QStringList actionpath;
    QList<QPair<QString, std::shared_ptr<sh::actions::AbstractActionItem>>> subactions;
    resolveActionpath(request->data("actionpath"), &actionpath, &subactions);
    foreach (auto pact, subactions) {
        auto aact = pact.second;
        QStringList lnactionpath = actionpath;
        lnactionpath << pact.first;
        QString nactionpath = lnactionpath.join(' ');
        QJsonObject jaee;
        jaee["name"] = nactionpath;
        jaee["label"] = aact->text().replace(reActionTextAccel, "\\1");
        if (!aact->enabled())
            jaee["disabled"] = true;
        if (!aact->visible())
            jaee["invisible"] = true;
        auto headpact = dynamic_cast<sh::actions::HeaderActionItem*>(aact.get());
        auto seppact = dynamic_cast<sh::actions::SeparatorActionItem*>(aact.get());
        if (headpact)
            jaee["header"] = true;
        else if (seppact)
            jaee["separator"] = true;
        QJsonObject jee;
        jee["actionpath"] = nactionpath;
        jee["value"] = jaee;
        result.append(jee);
    }
    request->setResponse(result);
}

void sh::ui::web::WebActionManager::cmd__get_ui__M(WebServerEngineRequest* request)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    QJsonObject result;
    QJsonArray jInfoPanels;
    QHash<sh::actions::ActionExecutionInfo*, WebActionExecutionInfoPanel*> info2panel;
    foreach (auto infopanel, _infopanels) {
        if (!infopanel->info() || infopanel->info()->isManualInterventionNeeded() || infopanel->info()->isVisualProcessFeedbackActive()) {
            QJsonObject jno;
            if (infopanel->info())
                info2panel[infopanel->info()] = infopanel;
            else
                jno["handlesClick"] = true;
            jno["id"] = infopanel->id();
            jno["label"] = infopanel->label();
            jno["isProgressDeterminate"] = infopanel->isProgressDeterminate();
            jno["progressDone"] = (qint64) infopanel->progressDone();
            jno["progressAll"] = (qint64) infopanel->progressAll();
            jno["forceForeground"] = infopanel->forceForeground();
            jno["panelWidth"] = infopanel->width();
            if (infopanel->color().isValid())
                jno["color"] = infopanel->color().name();
            jInfoPanels.append(jno);
        }
    }
    result["infoPanels"] = jInfoPanels;
    QJsonArray jDialogs;
    foreach (auto infodialog, _infodialogs) {
        if (!infodialog->info() || infodialog->info()->isManualInterventionNeeded() || infodialog->info()->isVisualProcessFeedbackActive()) {
            QJsonObject jno;
            jno["id"] = infodialog->id();
            jno["webts_created"] = infodialog->webts_created();
            auto ipnl = info2panel.find(infodialog->info());
            if (ipnl != info2panel.end())
                jno["panel_id"] = (*ipnl)->id();
            jno["details_fromverb"] = infodialog->details_fromverb();
            jno["details_fromobject"] = infodialog->details_fromobject();
            jno["details_toverb"] = infodialog->details_toverb();
            jno["details_toobject"] = infodialog->details_toobject();
            jno["head"] = infodialog->head();
            jno["progress_isDeterminate"] = infodialog->progress_isDeterminate();
            jno["progress_done"] = (qint64) infodialog->progress_done();
            jno["progress_all"] = (qint64) infodialog->progress_all();
            jno["progress_text"] = infodialog->progress_text();
            jno["userfeedback"] = infodialog->userFeedbackAsJsonValue();
            jDialogs.append(jno);
        }
    }
    result["dialogs"] = jDialogs;
    request->setResponse(result);
}

void sh::ui::web::WebActionManager::cmd__icon(WebServerEngineRequest* request)
{
    QStringList actionpath;
    QList<QPair<QString, std::shared_ptr<sh::actions::AbstractActionItem>>> subactions;
    std::shared_ptr<sh::actions::AbstractActionItem> action = resolveActionpath(request->data("actionpath"), &actionpath, &subactions);
    QIcon icon = sh::base::IconManager::instance()->getIcon(action ? action->icon() : "");
    request->setResponse(sh::tools::Misc::iconToPngByteArray(icon, 24), "image/png");
}

void sh::ui::web::WebActionManager::cmd__panel_clicked__M(WebServerEngineRequest* request)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    qint64 id = request->data("id").toLongLong();
    WebActionExecutionInfoPanel* panel = nullptr;
    foreach (auto _panel, _infopanels) {
        if (_panel->id() == id) {
            panel = _panel;
            break;
        }
    }
    if (panel)
        panel->_emit_clicked();
}

std::shared_ptr<sh::actions::AbstractActionItem> sh::ui::web::WebActionManager::resolveActionpath(QString sactionpath, QStringList* actionpath, QList<QPair<QString, std::shared_ptr<sh::actions::AbstractActionItem>>>* subactions)
{
    std::shared_ptr<sh::actions::AbstractActionItem> result = 0;
    sh::base::MainThread::dispatcher()->invokeSync([&](){
        foreach (QString ktoplvlact, _permanentToolbarActions.keys())
            subactions->append(QPair<QString, std::shared_ptr<sh::actions::AbstractActionItem>>(ktoplvlact, _permanentToolbarActions[ktoplvlact]));
    });
    foreach(QString apn, sactionpath.split(' ')) {
        if (!apn.isEmpty()) {
            foreach (auto psa, *subactions) {
                if (psa.first == apn) {
                    result = psa.second;
                    if (result)
                        break;
                }
            }
            subactions->clear();
            if (result) {
                std::shared_ptr<sh::actions::SubmenuActionItem> resultsa = std::dynamic_pointer_cast<sh::actions::SubmenuActionItem>(result);
                if (resultsa) {
                    auto tactsubs = resultsa->subitems();
                    for(int i=0; i<tactsubs.length(); i++)
                        subactions->append(QPair<QString, std::shared_ptr<sh::actions::AbstractActionItem>>(QString("%1").arg(i), tactsubs[i]));
                }
            }
            else {
                result = 0;
                break;
            }
            *actionpath << apn;
        }
    }
    return result;
}
