/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2020 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "base/mainthread.h"
#include "base/threadpool.h"
#include "configuration/configurationvalue.h"
#include "paneldetails/paneldetailmanager.h"
#include "tools/misc.h"
#include "ui/web/webmainwindow.h"
#include "webdetailpanelmanager.h"
#include "webserverengine.h"


sh::ui::web::WebDetailPanelManager::WebDetailPanelManager()
{
    auto mainwnd = sh::ui::web::WebMainWindow::mainWindow();
    int w, h;
    mainwnd->_cfgvalDetailPanelThumbnailDimensionsValueDimensions(w, h);
    mainwnd->webserver()->setConfigValue("detailPanelThumbnailWidth", QString("%1").arg(w)); //TODO # also store panel height same way?!
}

bool sh::ui::web::WebDetailPanelManager::executeCommand(WebServerEngineRequest* request, QString command)
{
    return executeCommand_byQObjectReflection(this, request, command, "detailpanel");
}

void sh::ui::web::WebDetailPanelManager::cmd__get_details(WebServerEngineRequest* request)
{
    QMutexLocker _l1(&_mutex);
    QJsonArray result;
    sh::base::MainThread::dispatcher()->invokeSync([&](){
        _webts_lastupdate = webserver()->currentTimestamp();
    });
    foreach (std::shared_ptr<sh::paneldetails::PanelDetail> paneldetail, _paneldetails) {
        if (!paneldetail->rows().isEmpty()) {
            QJsonObject opd;
            opd["valueWidthHint"] = paneldetail->valueWidthHint();
            QJsonArray rows;
            foreach (sh::paneldetails::PanelDetailRow paneldetailrow, paneldetail->rows()) {
                QJsonObject opdr;
                opdr["key"] = paneldetailrow.key();
                QJsonArray orowvalue;
                foreach (sh::paneldetails::PanelDetailRowValueElement rowvalue, paneldetailrow.value()) {
                    QJsonObject orowvalueelement;
                    orowvalueelement["text"] = rowvalue.text();
                    if (!rowvalue.icon().isNull())
                        orowvalueelement["icon"] = QString::fromLatin1(sh::tools::Misc::iconToBase64SrcEncoding(rowvalue.icon(), 16));
                    orowvalueelement["linktarget"] = rowvalue.linktarget();
                    orowvalueelement["linkautohide"] = rowvalue.linkautohide();
                    orowvalue.append(orowvalueelement);
                }
                opdr["value"] = orowvalue;
                rows.append(opdr);
            }
            opd["rows"] = rows;
            opd["webts_lastupdate"] = _webts_lastupdate;
            result.append(opd);
        }
    }
    request->setResponse(result);
}

void sh::ui::web::WebDetailPanelManager::cmd__get_thumbnail(WebServerEngineRequest* request)
{
    int width = request->data("width").toInt();
    int height = request->data("height").toInt();
    QByteArray result;
    bool done = false;
    QMutex m;
    QWaitCondition wc;
    auto mainwnd = sh::ui::web::WebMainWindow::mainWindow();
    sh::base::MainThread::dispatcher()->invokeSync([&](){
        mainwnd->requestDetailThumbnail(width, height, [](){}, [&](QIcon qicon){
            result = sh::tools::Misc::iconToPngByteArray(qicon, std::max(width, height)/*TODO# px vs pt*/);
            {
                QMutexLocker _l1(&m);
                done = true;
                wc.wakeAll();
            }
        }, true);
    });
    int w, h;
    mainwnd->_cfgvalDetailPanelThumbnailDimensionsValueDimensions(w, h);
    if (w != width)
        mainwnd->cfgvalDetailPanelThumbnailDimensions->setConfigValue(QString("%1;%2").arg(width).arg(h));
    {
        QMutexLocker _l1(&m);
        while (!done)
            wc.wait(&m);
    }
    request->setResponse(result, "image/png");
}

void sh::ui::web::WebDetailPanelManager::cmd__link_clicked(WebServerEngineRequest* request)
{
    std::shared_ptr<sh::paneldetails::PanelDetail> detail = nullptr;
    qint64 idx = request->data("idx").toLongLong();
    qint64 webts_lastupdate = request->data("webts_lastupdate").toLongLong();
    if (webts_lastupdate != _webts_lastupdate)
        return;
    QString linktarget = request->data("linktarget");
    {
        QMutexLocker _l1(&_mutex);
        if (idx < _paneldetails.count())
            detail = _paneldetails[idx];
    }
    if (detail)
        emit detail->linkTriggered(linktarget);
}

void sh::ui::web::WebDetailPanelManager::load(QList<std::shared_ptr<sh::filesystem::FilesystemNode>> nodes)
{
    sh::base::ThreadPool::enqueueForce([=](){
        QMutexLocker _l1(&_mutex);
        _paneldetails = sh::paneldetails::PanelDetailManager::instance()->getDetails(nodes);
        foreach (std::shared_ptr<sh::paneldetails::PanelDetail> paneldetail, _paneldetails) {
            QObject::connect(paneldetail.get(), &sh::paneldetails::PanelDetail::changed, this, [=](){
                webserver()->triggerEvent("detailpanel/changed");
            });
        }
        webserver()->triggerEvent("detailpanel/changed");
    });
}

