/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2020 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "webdialog.h"
#include "webmainwindow.h"
#include "webserverengine.h"


sh::ui::web::WebDialogManager::WebDialogManager()
{
}

sh::ui::web::WebServerEngine* sh::ui::web::WebDialogManager::webserver()
{
    return sh::ui::web::WebMainWindow::mainWindow()->webserver();
}

bool sh::ui::web::WebDialogManager::wasAccepted(std::shared_ptr<Dialog> dialog)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    auto wdialog = std::dynamic_pointer_cast<WebDialog>(dialog);
    return !wdialog->dialogResult().isEmpty();
}

void sh::ui::web::WebDialogManager::waitClosed(std::shared_ptr<Dialog> dialog)
{
    bool ismainthread = QThread::currentThread() == QCoreApplication::instance()->thread();
    bool dlgisopen = true;
    while (dlgisopen) {
        QThread::msleep(40);
        sh::base::MainThread::dispatcher()->invokeSync([&](){
            if (ismainthread)
                QCoreApplication::processEvents();
            dlgisopen = !dialog->wasClosed();
        });
    }
}

void sh::ui::web::WebDialogManager::stopDialog(std::shared_ptr<Dialog> dialog)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    webserver()->triggerEvent("dialog/closed", dialog->dialogId());
}

void sh::ui::web::WebDialogManager::showDialog(std::shared_ptr<Dialog> dialog)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    QJsonObject jdata = std::dynamic_pointer_cast<WebDialog>(dialog)->toJson();
    webserver()->triggerEvent("dialog/opened", jdata);
}

bool sh::ui::web::WebDialogManager::executeCommand(WebServerEngineRequest* request, QString command)
{
    static QJsonObject jerrOrphaned{{"_dialog_error", QJsonObject{{"_dialog_orphaned", true}}}};
    if (command.startsWith("dialog/")) {
        QString scommand = command.mid(7);
        if (scommand.indexOf("/") == -1)
            return executeCommand_byQObjectReflection(this, request, command, "dialog");
        else {
            auto dialog = std::dynamic_pointer_cast<WebDialog>(getDialogForRequest(request));
            if (dialog) {
                bool handled = dialog->executeDialogCommand(request, scommand);
                if (!handled)
                    request->setResponse("", "", WebServerEngineRequest::HTTP_NOT_FOUND);
            }
            else
                request->setResponse(jerrOrphaned);
        }
        return true;
    }
    return false;
}

std::shared_ptr<sh::ui::Dialog> sh::ui::web::WebDialogManager::getDialogForRequest(WebServerEngineRequest* request)
{
    bool ok = false;
    qint64 dialogId = request->data("dialogId").toLongLong(&ok);
    std::shared_ptr<sh::ui::Dialog> result;
    sh::base::MainThread::dispatcher()->invokeSync([&](){
        result = (ok ? getDialogById(dialogId) : nullptr);
    });
    return result;
}

void sh::ui::web::WebDialogManager::cmd__change_dialog_property__M(WebServerEngineRequest* request)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    auto dialog = std::dynamic_pointer_cast<WebDialog>(getDialogForRequest(request));
    if (dialog) {
        QString dlgPropertyKey = request->data("key");
        QString sjvalue = request->data("value");
        QVariant value;
        if (!sjvalue.isEmpty()) {
            QJsonDocument jd = QJsonDocument::fromJson(sjvalue.toLocal8Bit());
            value = jd.object();
        }
        dialog->_dohandleBrowserSideDidChangeProperty(dlgPropertyKey, value);
    }
}

void sh::ui::web::WebDialogManager::cmd__closed_in_browser__M(WebServerEngineRequest* request)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    auto dialog = getDialogForRequest(request);
    if (dialog)
        dialog->close();
}

void sh::ui::web::WebDialogManager::cmd__list__M(WebServerEngineRequest* request)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    QJsonArray jresult;
    foreach (auto dlg, getAllDialogs()) // sorted by dialogId
        jresult.append(std::dynamic_pointer_cast<WebDialog>(dlg)->toJson());
    request->setResponse(jresult);
}

sh::ui::web::WebDialog::WebDialog(QString dialogClassName) : _dialogClassName(dialogClassName)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    _webts_created = WebMainWindow::mainWindow()->webserver()->currentTimestamp();
}

sh::ui::web::WebDialog::~WebDialog()
{
}

void sh::ui::web::WebDialog::_dohandleBrowserSideDidChangeProperty(QString dlgPropertyKey, QVariant value)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    setDialogProperty(dlgPropertyKey, value);
    handleBrowserSideDidChangeProperty(dlgPropertyKey);
}

void sh::ui::web::WebDialog::handleBrowserSideDidChangeProperty(QString)
{
}

bool sh::ui::web::WebDialog::executeDialogCommand(WebServerEngineRequest*, QString)
{
    return false;
}

QVariant sh::ui::web::WebDialog::dialogProperty(QString dlgPropertyKey, QVariant deflt)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    return _properties.value(dlgPropertyKey, deflt);
}

void sh::ui::web::WebDialog::setDialogProperty(QString dlgPropertyKey, QVariant value)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    Dialog* asdlg = dynamic_cast<Dialog*>(this);
    _properties[dlgPropertyKey] = value;
    if (asdlg->isInited()) {
        QJsonObject jc;
        jc["dialogId"] = asdlg->dialogId();
        jc["webts"] = WebMainWindow::mainWindow()->webserver()->currentTimestamp();
        jc["key"] = dlgPropertyKey;
        jc["value"] = QJsonValue::fromVariant(value);
        WebMainWindow::mainWindow()->webserver()->triggerEvent("dialog/changed", jc);
    }
}

QVariantHash sh::ui::web::WebDialog::dialogResult()
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    QVariant dres = dialogProperty("_dialog_result");
    if (!dres.isNull()) {
        if (dres.canConvert<QVariantHash>())
            return dres.toHash();
        else
            return QVariantHash({{"value", dres}});
    }
    return QVariantHash();
}

QJsonObject sh::ui::web::WebDialog::toJson()
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    Dialog* asdlg = dynamic_cast<Dialog*>(this);
    QJsonObject result;
    result["dialogId"] = asdlg->dialogId();
    result["webts_created"] = _webts_created;
    result["webts"] = WebMainWindow::mainWindow()->webserver()->currentTimestamp();
    result["view"] = "shwebui." + _dialogClassName;
    foreach(QString kppr, _properties.keys())
        result[kppr] = QJsonValue::fromVariant(_properties[kppr]);
    return result;
}

void sh::ui::web::WebDialog::triggerDialogEvent(QString name, QJsonObject data)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    Dialog* asdlg = dynamic_cast<Dialog*>(this);
    data["dialogId"] = asdlg->dialogId();
    WebMainWindow::mainWindow()->webserver()->triggerEvent("dialog/"+name, data);
}

void sh::ui::web::WebDialog::setCloseableByUser(bool v)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    Dialog* asdlg = dynamic_cast<Dialog*>(this);
    if (asdlg->isInited())
        throw new sh::exceptions::ProgramException(sh::exceptions::ExceptionData().details("setCloseableByUser called too late"));
    setDialogProperty("closeable", v);
}

bool sh::ui::web::WebDialog::isCloseableByUser()
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    return dialogProperty("closeable", true).toBool();
}
