/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2020 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef WEBDIALOG_H
#define WEBDIALOG_H

#include <future>
#include <memory>
#include "base/mainthread.h"
#include "tools/misc.h"
#include "ui/dialog.h"
#include "webmodule.h"
#include "webserverengine.h"


namespace sh{
    namespace ui{
        class Dialog;
        namespace web{

            /** @brief A DialogManager used in Web ui.
             */
            class WebDialogManager : public QObject, public WebModule, public sh::ui::DialogManager
            {
                Q_OBJECT
            public:

                /** @a_infrastructure
                 */
                WebDialogManager();

                /** \brief Return the WebServerEngine hosting this dialog.
                 */
                WebServerEngine* webserver();

            protected:
                bool wasAccepted(std::shared_ptr<Dialog> dialog) override;
                void waitClosed(std::shared_ptr<Dialog> dialog) override;
                void stopDialog(std::shared_ptr<Dialog> dialog) override;

            private slots:
                void cmd__change_dialog_property__M(WebServerEngineRequest* request);
                void cmd__closed_in_browser__M(WebServerEngineRequest* request);
                void cmd__list__M(WebServerEngineRequest* request);

            private:

                /** \brief Returns the WebDialog referred to the `request` (i.e. its `dialogId` parameter).
                 */
                std::shared_ptr<Dialog> getDialogForRequest(WebServerEngineRequest* request);

                bool executeCommand(WebServerEngineRequest* request, QString command) override;
                void showDialog(std::shared_ptr<Dialog> dialog) override;
            };

            /** \brief Abstract base class for a web based dialog. Typically used for also implementing some sh::ui::Dialog.
             *
             * A web dialog is a dialog window presented to the user on browser side.
             * A web dialog implementation consists of a backend part (a subclass of WebDialog)
             * and a browser side part (basically a subclass of shwebui.Dialog).
             *
             * For implementing a web dialog, implement a subclass of WebDialog (which in turn will
             * specify the browser side part) and sh::ui::Dialog. Create such dialogs by means of WebDialogManager.
             *
             * Communicating values between backend and browser typically works by dialogProperty() and setDialogProperty(),
             * also dialogResult() for the dialog result after closing.
             */
            class WebDialog : public sh::tools::Jsonable
            {
                friend class WebDialogManager;
            public:
                /** @a_infrastructure
                 * See WebDialogManager.
                 * \param dialogClassName: The name of the shwebui.Dialog subclass which implements the web dialog on browser side.
                 */
                WebDialog(QString dialogClassName);

                /** \brief Returns a dialog property value by key.
                 */
                QVariant dialogProperty(QString dlgPropertyKey, QVariant deflt=QVariant());

                /** \brief Sets a dialog property value for a key.
                 */
                void setDialogProperty(QString dlgPropertyKey, QVariant value);

                /** \brief Returns the dialog result as hash.
                 *
                 * The browser side of a web dialog can 'accept' a dialog while closing by setting a non-empty dialog result.
                 * This mechanism allows browser side dialogs to return data to the backend. It's up to the backend
                 * to interpret the content of this hash.
                 *
                 * If the dialog was cancelled, or is still open, the returned hash is empty. See also wasClosed().
                 */
                QVariantHash dialogResult();

                /** \brief Triggers a dialog event (at the browsers).
                 *
                 * They in turn typically fetch some data again in order to refresh the ui.
                 */
                void triggerDialogEvent(QString name, QJsonObject data=QJsonObject());

                /** \brief Sets if this dialog shall be directly closeable by the user by the window
                 * decoration. Set to `false` if the user has to react on this dialog by clicking on some
                 * controls in the dialog body.
                 *
                 * This method must be called during construction.
                 */
                void setCloseableByUser(bool v=true);

                /** \brief Returns if this dialog shall be directly closeable by the user by the window
                 * decoration.
                 */
                bool isCloseableByUser();

                virtual ~WebDialog();
                QJsonObject toJson() override;

            protected:

                /** \brief Internally used for handling dialog property changes notified from browser side.
                 *
                 * @a_infrastructure
                 */
                void _dohandleBrowserSideDidChangeProperty(QString dlgPropertyKey, QVariant value);

                /** \brief Executes custom steps reacting on dialog property changes notified from browser side.
                 *
                 * This mechanism helps updating data on backend side with current ui state.
                 *
                 * @a_dooverride_optional
                 */
                virtual void handleBrowserSideDidChangeProperty(QString dlgPropertyKey);

                /** \brief Executes `command` on this dialog as requested by the browser, if responsible.
                 *
                 * Read parameters, check if this module is responsible for answering, and then write an answer with `request`.
                 * Note: this module should typically be responsible, because the original request specified its dialog id.
                 *
                 * The concept is similar to sh::ui::web::WebModule::executeCommand().
                 * Implementations often use sh::ui::web::WebModule::executeCommand_byQObjectReflection() for convenience.
                 *
                 * @a_dooverride_optional
                 */
                virtual bool executeDialogCommand(WebServerEngineRequest* request, QString command);

            private:
                QVariantHash _properties;
                QString _dialogClassName;
                qint64 _webts_created;

            };

        }
    }
}

#endif // WEBDIALOG_H
