/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2020 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "filesystem/filesystemmodel.h"
#include "filesystem/filesystemnode.h"
#include "base/mainthread.h"
#include "tools/misc.h"
#include "webfilepropertydialog.h"
#include "webmainwindow.h"
#include "webserverengine.h"


sh::ui::web::WebFilePropertyDialog::WebFilePropertyDialog(QList<std::shared_ptr<filesystem::FilesystemNode>> nodes)
        : WebDialog("FilePropertyDialog"), FilePropertyDialog(nodes)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    _refreshTimer.setSingleShot(true);
    QObject::connect(&_refreshTimer, &QTimer::timeout, [=](){
        triggerDialogEvent("properties/changed");
    });
    QStringList lnodes;
    foreach(auto n, nodes)
        lnodes.append(n->eurl()->asString());
    setDialogProperty("nodes", lnodes);
}

sh::ui::web::WebFilePropertyDialog::~WebFilePropertyDialog()
{
    foreach (WebFilePropertyDialogChild* child, _childs)
        delete child;
}

void sh::ui::web::WebFilePropertyDialog::init(std::shared_ptr<Dialog> dialog)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    sh::base::MainThread::dispatcher()->invokeAsync([=](){
        auto dlg = std::dynamic_pointer_cast<FilePropertyDialog>(dialog);
        foreach (FilePropertyDialogTabFactory* tabfactory, getPropertyTabs()) {
            std::shared_ptr<sh::ui::FilePropertyDialogTab> tab = tabfactory->construct(dlg);
            if (tab) {
                QList<QString> properties = tab->properties();
                TabViewStructWidgets widgets;
                for (int i=0; i<properties.length(); i++) {
                    auto pwidget = new WebFilePropertyDialogTabActionsView(this);
                    _childs.append(pwidget);
                    tab->populateWidget(i, pwidget);
                    widgets.append(pwidget);
                    pwidget->setLabelText(properties[i]);
                };
                addTabstruct(tab, widgets);
            }
        }
        updateWidgets();
    }, dialog);
}

void sh::ui::web::WebFilePropertyDialog::refresh()
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    if (!_refreshTimer.isActive())
        _refreshTimer.start(200);
}

sh::ui::web::WebFilePropertyDialog::WebFilePropertyDialogChild::WebFilePropertyDialogChild(WebFilePropertyDialog* dlg)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    _dlg = dlg;
}

sh::ui::web::WebFilePropertyDialog::WebFilePropertyDialogChild::~WebFilePropertyDialogChild()
{
}

void sh::ui::web::WebFilePropertyDialog::WebFilePropertyDialogChild::requestClientRefresh()
{
    _dlg->refresh();
}

sh::ui::web::WebFilePropertyDialog::WebFilePropertyDialogTabActionsView::WebFilePropertyDialogTabActionsView(WebFilePropertyDialog* dlg) : WebFilePropertyDialogChild(dlg)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
}

QJsonObject sh::ui::web::WebFilePropertyDialog::WebFilePropertyDialogTabActionsView::toJson()
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    QJsonObject result;
    result["type"] = "Actions";
    result["content"] = dynamic_cast<Jsonable*>(_content)->toJson();
    result["visible"] = _visible;
    result["labelText"] = _labelText;
    QJsonArray jbuttons;
    foreach (auto btn, _buttons)
        jbuttons.append(btn);
    result["buttons"] = jbuttons;
    return result;
}

void sh::ui::web::WebFilePropertyDialog::WebFilePropertyDialogTabActionsView::setVisible(bool v)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    _visible = v;
    requestClientRefresh();
}

void sh::ui::web::WebFilePropertyDialog::WebFilePropertyDialogTabActionsView::setContent(FilePropertyDialogTabViewContent* cnt)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    sh::ui::FilePropertyDialogTabActionsView::setContent(cnt);
    requestClientRefresh();
}

void sh::ui::web::WebFilePropertyDialog::WebFilePropertyDialogTabActionsView::setButtons(QList<QString> buttons)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    sh::ui::FilePropertyDialogTabActionsView::setButtons(buttons);
    requestClientRefresh();
}

void sh::ui::web::WebFilePropertyDialog::WebFilePropertyDialogTabActionsView::setLabelText(QString text)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    _labelText = text;
    requestClientRefresh();
}

sh::ui::FilePropertyDialogTabTableView* sh::ui::web::WebFilePropertyDialog::createTabViewTable()
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    auto r = new WebFilePropertyDialogTabTableView(this);
    _childs.append(r);
    return r;
}

sh::ui::web::WebFilePropertyDialog::WebFilePropertyDialogTabTableView::WebFilePropertyDialogTabTableView(WebFilePropertyDialog* dlg) : WebFilePropertyDialogChild(dlg)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
}

QJsonObject sh::ui::web::WebFilePropertyDialog::WebFilePropertyDialogTabTableView::toJson()
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    QJsonObject result;
    result["type"] = "Table";
    QJsonArray jcontent;
    foreach(auto x, _content) {
        QJsonObject jx;
        jx["0"] = x.first;
        jx["1"] = x.second;
        jcontent.append(jx);
    }
    result["content"] = jcontent;
    return result;
}

void sh::ui::web::WebFilePropertyDialog::WebFilePropertyDialogTabTableView::setContent(QList<QPair<QString, QString>> content)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    _content = content;
    requestClientRefresh();
}

QList<QPair<int, int>> sh::ui::web::WebFilePropertyDialog::WebFilePropertyDialogTabTableView::getSelectedItems()
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    QList<QPair<int, int>> result;
    foreach(int slidx, _selectedItems)
        result.append(QPair<int, int>(slidx, 1));
    return result;
}

void sh::ui::web::WebFilePropertyDialog::WebFilePropertyDialogTabTableView::setSelectedItems(QList<int> selitms)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    _selectedItems = selitms;
}

QVariant sh::ui::web::WebFilePropertyDialog::WebFilePropertyDialogTabTableView::getItemValue(int r, int c)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    auto o = _content[r];
    if (c == 0)
        return o.first;
    else /* if (c == 1) */
        return o.second;
}

sh::ui::FilePropertyDialogTabTextView* sh::ui::web::WebFilePropertyDialog::createTabViewText()
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    auto r = new WebFilePropertyDialogTabTextView(this);
    _childs.append(r);
    return r;
}

sh::ui::web::WebFilePropertyDialog::WebFilePropertyDialogTabTextView::WebFilePropertyDialogTabTextView(WebFilePropertyDialog* dlg) : WebFilePropertyDialogChild(dlg)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
}

QJsonObject sh::ui::web::WebFilePropertyDialog::WebFilePropertyDialogTabTextView::toJson()
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    QJsonObject result;
    result["type"] = "Text";
    result["content"] = _content;
    return result;
}

void sh::ui::web::WebFilePropertyDialog::WebFilePropertyDialogTabTextView::setContent(QString content)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    _content = content;
    requestClientRefresh();
}

sh::ui::FilePropertyDialogTabIconTextBannerView* sh::ui::web::WebFilePropertyDialog::createTabViewIconTextBanner()
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    auto r = new WebFilePropertyDialogTabIconTextBannerView(this);
    _childs.append(r);
    return r;
}

sh::ui::web::WebFilePropertyDialog::WebFilePropertyDialogTabIconTextBannerView::WebFilePropertyDialogTabIconTextBannerView(WebFilePropertyDialog* dlg) : WebFilePropertyDialogChild(dlg)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
}

QJsonObject sh::ui::web::WebFilePropertyDialog::WebFilePropertyDialogTabIconTextBannerView::toJson()
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    QJsonObject result;
    result["type"] = "IconTextBanner";
    QJsonArray jcontent;
    foreach (WebIconTextBannerItem itm, _content) {
        QJsonObject jitm;
        if (!itm.second.isNull())
            jitm["text"] = itm.second;
        else {
            QJsonObject jitmicon;
            QString itmsrc = QString::fromLatin1(sh::tools::Misc::iconToBase64SrcEncoding(itm.first.first, itm.first.second));
            jitmicon["src"] = itmsrc;
            jitmicon["sizeInPt"] = itm.first.second;
            jitm["icon"] = jitmicon;
        }
        jcontent.append(jitm);
    }
    result["content"] = jcontent;
    return result;
}

void sh::ui::web::WebFilePropertyDialog::WebFilePropertyDialogTabIconTextBannerView::clear()
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    _content.clear();
    requestClientRefresh();
}

void sh::ui::web::WebFilePropertyDialog::WebFilePropertyDialogTabIconTextBannerView::insertIcon(QIcon icon, int sizePt)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    WebIconTextBannerItemIcon pico(icon, sizePt);
    _content.append(WebIconTextBannerItem(pico, QString()));
    requestClientRefresh();
}

void sh::ui::web::WebFilePropertyDialog::WebFilePropertyDialogTabIconTextBannerView::insertText(QString text)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    WebIconTextBannerItemIcon pico(QIcon(), 0);
    _content.append(WebIconTextBannerItem(pico, text));
    requestClientRefresh();
}

bool sh::ui::web::WebFilePropertyDialog::executeDialogCommand(WebServerEngineRequest* request, QString command)
{
    return WebModule::executeCommand_byQObjectReflection(this, request, command, "properties");
}

void sh::ui::web::WebFilePropertyDialog::cmd__get_properties__M(WebServerEngineRequest* request)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    request->setResponse(getProperties());
}

void sh::ui::web::WebFilePropertyDialog::cmd__refresh__M(WebServerEngineRequest*)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    updateWidgets();
}

void sh::ui::web::WebFilePropertyDialog::cmd__trigger_action__M(WebServerEngineRequest* request)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    int cookie = request->data("cookie").toInt();
    int tabidx = request->data("tabidx").toInt();
    int widgetidx = request->data("widgetidx").toInt();
    int btnidx = request->data("btnidx").toInt();
    triggerAction(cookie, tabidx, widgetidx, btnidx);
}

void sh::ui::web::WebFilePropertyDialog::cmd__tabwidget_table_select__M(WebServerEngineRequest* request)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    int cookie = request->data("cookie").toInt();
    int tabidx = request->data("tabidx").toInt();
    int widgetidx = request->data("widgetidx").toInt();
    QString ssel =  request->data("selected");
    auto jsel = QJsonDocument::fromJson(ssel.toLocal8Bit());
    QList<int> lsel;
    foreach(auto jcsel, jsel.array())
        lsel.append(jcsel.toInt());
    tabwidgetTableSelect(cookie, tabidx, widgetidx, lsel);
}

void sh::ui::web::WebFilePropertyDialog::updateWidgets()
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    _cookie++;
    sh::ui::FilePropertyDialog::updateWidgets();
}

void sh::ui::web::WebFilePropertyDialog::triggerAction(int cookie, int tabidx, int widgetidx, int btnidx)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    if (_cookie != cookie)
        return;
    auto _tabstructs = tabstructs();
    if (tabidx < _tabstructs.length()) {
        auto ttab = _tabstructs[tabidx].first;
        auto twidget = dynamic_cast<WebFilePropertyDialogTabActionsView*>(ttab->widgetAt(widgetidx));
        if (twidget)
            twidget->_emit_buttonTriggered(btnidx);
    }
}

QJsonArray sh::ui::web::WebFilePropertyDialog::getProperties()
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    QJsonArray result;
    foreach (auto xtab, tabstructs()) {
        QJsonObject jtab;
        jtab["cookie"] = _cookie;
        jtab["title"] = xtab.first->title();
        jtab["titleWithoutMnemonic"] = xtab.first->titleWithoutMnemonic();
        TabViewStructWidgets htab = xtab.second;
        QJsonArray jtabwidgets;
        for (int itw=0; itw<htab.count(); itw++) {
            auto tw = htab[itw];
            QJsonObject jtabwidget = dynamic_cast<Jsonable*>(tw)->toJson();
            jtabwidget["idx"] = itw;
            jtabwidgets.append(jtabwidget);
        }
        jtab["widgets"] = jtabwidgets;
        result.append(jtab);
    }
    return result;
}

void sh::ui::web::WebFilePropertyDialog::tabwidgetTableSelect(int cookie, int tabidx, int widgetidx, QList<int> lsel)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    if (_cookie != cookie)
        return;
    auto _tabstructs = tabstructs();
    if (tabidx < _tabstructs.length()) {
        auto ttab = _tabstructs[tabidx].first;
        auto twidget = dynamic_cast<WebFilePropertyDialogTabActionsView*>(ttab->widgetAt(widgetidx));
        if (twidget) {
            auto tblwidget = dynamic_cast<WebFilePropertyDialogTabTableView*>(twidget->content());
            tblwidget->setSelectedItems(lsel);
        }
    }
}
