/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2020 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "filesystem/filesystemmodel.h"
#include "base/mainthread.h"
#include "webmainwindow.h"
#include "webfileview.h"


bool sh::ui::web::WebFileViewManager::executeCommand(WebServerEngineRequest* request, QString command)
{
    return WebModule::executeCommand_byQObjectReflection(this, request, command, "fileview");
}

std::shared_ptr<sh::ui::web::WebFileView> sh::ui::web::WebFileViewManager::addFileView(bool reinitialize/*TODO # use, once columndimensions are there?*/)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    auto fileview = std::make_shared<WebFileView>(this);
    auto sfileview = fileview.get();
    connect(sfileview, &WebFileView::viewOptionChanged, this, [=](){ //TODO# anders
        webserver()->triggerEvent("fileview/fileview_config_changed", sfileview->id());
        emit fileViewOptionsChanged(sfileview->index());
    });
    connect(sfileview, &WebFileView::currentNodeChanged, this, [=](){ //TODO# anders
        webserver()->triggerEvent("fileview/fileview_currentnode_changed", sfileview->id());
    });
    QTimer* triggerItemsChangedTimer = new QTimer(sfileview);
    triggerItemsChangedTimer->setInterval(500);
    triggerItemsChangedTimer->setSingleShot(true);
    auto triggerItemsChangedFct = [=](){
        webserver()->triggerEvent("fileview/items_changed", sfileview->id());
    };
    connect(triggerItemsChangedTimer, &QTimer::timeout, triggerItemsChangedFct);
    auto connectmodelfct = [=](){
        //TODO# disconnect old! (actually needed?)
        auto enableTriggerItemsChangedTimer = [=](){
            if (!triggerItemsChangedTimer->isActive())
                triggerItemsChangedTimer->start();
        };
        auto enableTriggerItemsChangedTimer1 = [=](const QModelIndex&, int, int){ enableTriggerItemsChangedTimer(); };
        auto enableTriggerItemsChangedTimer2 = [=](const QModelIndex&, int, int, const QModelIndex&, int){ enableTriggerItemsChangedTimer(); };
        auto enableTriggerItemsChangedTimer3 = [=](const QModelIndex&, const QModelIndex&, const QVector<int>&){ enableTriggerItemsChangedTimer(); };
        triggerItemsChangedFct();
        auto model = sfileview->model().get();
        if (model) {
            connect(model, &QSortFilterProxyModel::rowsInserted, this, enableTriggerItemsChangedTimer1);
            connect(model, &QSortFilterProxyModel::rowsMoved, this, enableTriggerItemsChangedTimer2);
            connect(model, &QSortFilterProxyModel::rowsRemoved, this, enableTriggerItemsChangedTimer1);
            connect(model, &QSortFilterProxyModel::columnsInserted, this, enableTriggerItemsChangedTimer1);
            connect(model, &QSortFilterProxyModel::columnsMoved, this, enableTriggerItemsChangedTimer2);
            connect(model, &QSortFilterProxyModel::columnsRemoved, this, enableTriggerItemsChangedTimer1);
            connect(model, &QSortFilterProxyModel::modelReset, this, enableTriggerItemsChangedTimer);
            connect(model, &QSortFilterProxyModel::dataChanged, this, enableTriggerItemsChangedTimer3);
        }
    };
    connect(sfileview, &WebFileView::modelChanged, this, connectmodelfct);
    connectmodelfct();
    _fileviews.append(fileview);
    webserver()->triggerEvent("fileview/number_changed");
    emit fileViewCountChanged();
    return fileview;
}

void sh::ui::web::WebFileViewManager::setCurrentFileViewByIndex(int idx)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    _icurrentfileview = idx;
    emit currentFileViewChanged();
}

sh::ui::web::WebFileView* sh::ui::web::WebFileViewManager::currentFileView()
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    if (_icurrentfileview >= 0 && _icurrentfileview < fileViewCount())
        return getFileView(_icurrentfileview);
    else if (fileViewCount() > 0)
        return getFileView(0);
    else
        return 0;
}

int sh::ui::web::WebFileViewManager::fileViewCount()
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    return _fileviews.count();
}

void sh::ui::web::WebFileViewManager::removeFileView(int i)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    _fileviews.removeAt(i);
    webserver()->triggerEvent("fileview/number_changed");
    emit fileViewCountChanged();
}

sh::ui::web::WebFileView *sh::ui::web::WebFileViewManager::getFileView(int i)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    return _fileviews[i].get();
}

int sh::ui::web::WebFileViewManager::getFileViewIndex(sh::ui::FileView* fileview)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    for (int i=0; i<_fileviews.length(); i++)
        if (_fileviews[i].get() == fileview)
            return i;
    return -1;
}

void sh::ui::web::WebFileViewManager::cmd__checkednodes_changed__M(WebServerEngineRequest* request)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    int idx = request->data("index").toInt();
    if (idx >= fileViewCount())
        return;
    QString eurls = request->data("eurls");
    QStringList leurls;
    auto jd = QJsonDocument::fromJson(eurls.toLocal8Bit());
    foreach(auto x, jd.array())
        leurls.append(x.toString());
    auto fv = qobject_cast<sh::ui::web::WebFileView*>(getFileView(idx));
    QList<std::shared_ptr<sh::filesystem::FilesystemNode>> nodes;
    foreach(auto n, fv->getAllVisibleNodes()) {
        QString nseurl = n->eurl()->asString();
        if (eurls.contains(nseurl)) {
            nodes.append(n);
            eurls.remove(nseurl);
        }
    }
    fv->setCheckedNodes(nodes);
    emit activeSelectionChanged();
}

void sh::ui::web::WebFileViewManager::cmd__focus_current__M(WebServerEngineRequest* request)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    int idx = request->data("index").toInt();
    if (idx >= fileViewCount())
        return;
    setCurrentFileViewByIndex(idx);
    emit activeSelectionChanged();
}

void sh::ui::web::WebFileViewManager::cmd__get__M(WebServerEngineRequest* request)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    QJsonArray result;
    foreach(std::shared_ptr<WebFileView> webfileview, _fileviews) {
        QJsonObject jno;
        jno["id"] = webfileview->id();
        jno["sizeFormattingMode"] = webfileview->getSizeFormattingMode();
        jno["hiddenFilesVisible"] = webfileview->hiddenFilesVisible();
        jno["iconDimension"] = webfileview->iconDimension();
        jno["sortColumn"] = webfileview->sortColumn();
        jno["sortOrder"] = webfileview->sortOrder();
        jno["node"] = webfileview->node()->eurl()->asString();
        jno["viewmode"] = (int) webfileview->viewmode();
        result.append(jno);
    }
    request->setResponse(result);
}

void sh::ui::web::WebFileViewManager::cmd__list_view_items__M(WebServerEngineRequest* request)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    int listid = request->data("listid").toInt();
    std::shared_ptr<WebFileView> fv = 0;
    foreach(auto _fv, _fileviews) {
        if (_fv->id() == listid) {
            fv = _fv;
            break;
        }
    }
    QJsonObject result;
    QJsonArray resultData;
    QJsonArray resultHeaderData;
    if (fv) {
        auto model = fv->model();
        int rc = model->rowCount();
        int cc = model->columnCount();
        if (cc > 0) //TODO # code dup (see filesystem/list_items)
            for(int ir=0; ir<rc; ir++) {
                QModelIndex mii = model->mapToSource(model->index(ir, 0));
                auto fsmodel = sh::filesystem::FilesystemModel::instance();
                auto nl = fsmodel->getNodeForIndex(mii);
                if (!nl->eurl())
                    continue; //TODO# instead support the loading... node
                QJsonObject jno;
                jno["displayName"] = nl->displayName();
                jno["eurl"] = nl->eurl()->asString();
                jno["nodetype"] = QString("%1").arg((int)nl->nodetype());
                jno["isHidden"] = nl->isHidden();
                for(int ic=1; ic<cc; ic++)
                    jno[QString("col_%1").arg(ic)] = model->data(model->index(ir, ic)).toString();
                if (ir == 0) {
                    for(int ic=1; ic<cc; ic++)
                        resultHeaderData.append(model->headerData(ic, Qt::Horizontal).toString());
                }
                resultData.append(jno);
            }
    }
    result["data"] = resultData;
    result["headerData"] = resultHeaderData;
    request->setResponse(result);
}

void sh::ui::web::WebFileViewManager::cmd__node_activated__M(WebServerEngineRequest* request)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    int idx = request->data("index").toInt();
    QString eurls = request->data("eurls");
    auto jd = QJsonDocument::fromJson(eurls.toLocal8Bit());
    QList<std::shared_ptr<sh::filesystem::FilesystemNode>> nodes;
    auto fv = qobject_cast<sh::ui::web::WebFileView*>(getFileView(idx));
    foreach(auto x, jd.array()) {
        foreach(auto n, fv->getAllVisibleNodes())
            if (n->eurl()->asString() == x.toString()) {
                nodes.append(n);
                break;
            }
    }
    sh::ui::web::WebMainWindow::mainWindow()->activateNodes(nodes, [&](std::shared_ptr<sh::filesystem::FilesystemNode> nn){
        sh::ui::web::WebMainWindow::mainWindow()->jumpToNode(nn);
    });
}

void sh::ui::web::WebFileViewManager::cmd__selection_changed__M(WebServerEngineRequest* request)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    int idx = request->data("index").toInt();
    if (idx >= fileViewCount())
        return;
    QString eurl = request->data("eurl");
    auto fv = qobject_cast<sh::ui::web::WebFileView*>(getFileView(idx));
    std::shared_ptr<sh::filesystem::FilesystemNode> node = 0;
    foreach(auto n, fv->getAllVisibleNodes())
        if (n->eurl()->asString() == eurl) {
            node = n;
            break;
        }
    if (node)
        fv->setSelectedNode(node);
    emit activeSelectionChanged();
}

sh::ui::web::WebFileView::WebFileView(WebFileViewManager *manager) : _manager(manager)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    static qint64 next_id = 0;
    _id = next_id++;
    _set_model();
}

sh::ui::ColumnDimensions* sh::ui::web::WebFileView::columndimensions()
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    //TODO#
    return 0;
}

sh::tools::HistoryTracker<std::shared_ptr<const sh::filesystem::Eurl>>* sh::ui::web::WebFileView::historyTracker()
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    //TODO #xx
    return 0; //TODO# weird api (see return type)
}

sh::ui::FileViewMode sh::ui::web::WebFileView::viewmode()
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    return _viewmode;
}

void sh::ui::web::WebFileView::setViewmode(FileViewMode m)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    _viewmode = m;
    emit viewOptionChanged();//TODO #zz implement iconview
}

int sh::ui::web::WebFileView::index()
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    return _manager->getFileViewIndex(this);
}

void sh::ui::web::WebFileView::gotoDir(std::shared_ptr<sh::filesystem::FilesystemNode> n)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    if (node() != n) {
        FileView::gotoDir(n);
        _set_model();
        emit currentNodeChanged();
    }
}

sh::filesystem::SizeFormatting sh::ui::web::WebFileView::getSizeFormattingMode()
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    return _sizeformatting;
}

void sh::ui::web::WebFileView::setSizeFormattingMode(sh::filesystem::SizeFormatting mode)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    _sizeformatting = mode;
    emit viewOptionChanged();
}

bool sh::ui::web::WebFileView::hiddenFilesVisible()
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    return _hiddenFilesVisible;
}

void sh::ui::web::WebFileView::setHiddenFilesVisible(bool v)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    _hiddenFilesVisible = v;
    emit viewOptionChanged();
}

int sh::ui::web::WebFileView::iconDimension()
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    return _iconDimension;
}

void sh::ui::web::WebFileView::setIconDimension(int v)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    _iconDimension = v;
    emit viewOptionChanged();
}

void sh::ui::web::WebFileView::setSort(int column, Qt::SortOrder order)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    _sortColumn = column;
    _sortOrder = order;
    emit viewOptionChanged();
}

int sh::ui::web::WebFileView::sortColumn()
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    return _sortColumn;
}

Qt::SortOrder sh::ui::web::WebFileView::sortOrder()
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    return _sortOrder;
}

QList<std::shared_ptr<sh::filesystem::FilesystemNode>> sh::ui::web::WebFileView::selectedNodes()
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    if (_checkedNodes.isEmpty()) {
        if (_selectedNode)
            return QList<std::shared_ptr<sh::filesystem::FilesystemNode>>({_selectedNode});
        else
            return QList<std::shared_ptr<sh::filesystem::FilesystemNode>>();
    }
    else
        return _checkedNodes;
}

void sh::ui::web::WebFileView::setSelection(const QList<std::shared_ptr<sh::filesystem::FilesystemNode>> nodes)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    _selectedNode = (nodes.length() == 1) ? nodes.first() : nullptr;
    _checkedNodes = nodes;
    QJsonObject jdata;
    jdata["id"] = id();
    if (_selectedNode)
        jdata["selectedNode"] = _selectedNode->eurl()->asString();
    QJsonArray jcheckednodes;
    foreach (auto cnode, _checkedNodes)
        jcheckednodes.append(cnode->eurl()->asString());
    jdata["checkedNodes"] = jcheckednodes;
    _manager->webserver()->triggerEvent("fileview/selection_changed", jdata);
}

QList<std::shared_ptr<sh::filesystem::FilesystemNode>> sh::ui::web::WebFileView::getAllVisibleNodes()
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    QList<std::shared_ptr<sh::filesystem::FilesystemNode>> result;
    if (_model) {
        int lr = _model->rowCount();
        for (int ir=0; ir<lr; ir++)
        {
            QModelIndex mi = _model->index(ir, 0);
            QModelIndex nmi = _model->mapToSource(mi);
            if (nmi.isValid()) {
                auto minode = sh::filesystem::FilesystemModel::instance()->getNodeForIndex(nmi);
                if (minode && minode->eurl())
                    result.append(minode);
            }
        }
    }
    return result;
}

qint64 sh::ui::web::WebFileView::id()
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    return _id;
}

std::shared_ptr<sh::filesystem::FilesystemModelFileviewProxy> sh::ui::web::WebFileView::model()
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    return _model;
}

void sh::ui::web::WebFileView::_set_model()
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    QModelIndex mi = sh::filesystem::FilesystemModel::instance()->getIndexForNode(node());
    _model = std::shared_ptr<sh::filesystem::FilesystemModelFileviewProxy>(sh::filesystem::FilesystemModel::instance()->createFileviewProxy(mi, false, [=](std::shared_ptr<sh::filesystem::FilesystemNode> bestValid) {
        // this happens when this model dies (i.e. directory was removed)
        //TODO#
/*        _reload_goto(true, true, 0); (now: _goto)
        sh::base::MainThread::dispatcher()->invokeAsync([=](){
            viewctrl->gotoDir(bestValid);
        });*/
    }));
    _model->setHiddenFilesVisible(_hiddenFilesVisible);
    _model->setSizeFormattingMode(_sizeformatting);
    _model->sort(_sortColumn, _sortOrder);
    emit modelChanged();
}

void sh::ui::web::WebFileView::setSelectedNode(std::shared_ptr<sh::filesystem::FilesystemNode> node)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    _selectedNode = node;
    if (_checkedNodes.isEmpty())
        emit selectionChanged();
}

void sh::ui::web::WebFileView::setCheckedNodes(QList<std::shared_ptr<sh::filesystem::FilesystemNode>> nodes)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    _checkedNodes = nodes;
    emit selectionChanged();
}
