/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2020 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef WEBFILEVIEW_H
#define WEBFILEVIEW_H

#include "ui/fileview.h"
#include "webmodule.h"


namespace sh{
    namespace filesystem { class FilesystemModelFileviewProxy; }
    namespace ui{
        namespace web{
            class WebFileView;

            /** \brief Manages WebFileView instances.
             */
            class WebFileViewManager : public QObject, public WebModule
            {
                Q_OBJECT
            public:
                bool executeCommand(WebServerEngineRequest* request, QString command) override;
                std::shared_ptr<WebFileView> addFileView(bool reinitialize);
                void setCurrentFileViewByIndex(int idx);
                sh::ui::web::WebFileView* currentFileView();
                int fileViewCount();
                void removeFileView(int i);
                sh::ui::web::WebFileView* getFileView(int i);
                int getFileViewIndex(sh::ui::FileView* fileview);
            signals:
                void activeSelectionChanged();
                void fileViewOptionsChanged(int i);
                void fileViewCountChanged();
                void currentFileViewChanged();
            private slots:
                void cmd__checkednodes_changed__M(WebServerEngineRequest* request);
                void cmd__focus_current__M(WebServerEngineRequest* request);
                void cmd__get__M(WebServerEngineRequest* request);
                void cmd__list_view_items__M(WebServerEngineRequest* request);
                void cmd__node_activated__M(WebServerEngineRequest* request);
                void cmd__selection_changed__M(WebServerEngineRequest* request);
            private:
                QList<std::shared_ptr<WebFileView>> _fileviews;
                int _icurrentfileview = 0;
            };

            /** \brief Web based file view.
             */
            class WebFileView : public FileView //TODO #xx support sorting by columns in ui
            {
                friend class WebFileViewManager;
                Q_OBJECT
            public:
                WebFileView(WebFileViewManager* manager);
                sh::ui::ColumnDimensions* columndimensions() override;
                sh::tools::HistoryTracker<std::shared_ptr<const sh::filesystem::Eurl>>* historyTracker() override;
                FileViewMode viewmode() override;
                void setViewmode(FileViewMode m) override;
                int index() override;
                void gotoDir(std::shared_ptr<sh::filesystem::FilesystemNode> n) override;
                sh::filesystem::SizeFormatting getSizeFormattingMode() override;
                void setSizeFormattingMode(sh::filesystem::SizeFormatting mode) override;
                bool hiddenFilesVisible() override;
                void setHiddenFilesVisible(bool v) override;
                int iconDimension() override;
                void setIconDimension(int v) override;
                void setSort(int column, Qt::SortOrder order) override;
                int sortColumn() override;
                Qt::SortOrder sortOrder() override;
                QList<std::shared_ptr<sh::filesystem::FilesystemNode>> selectedNodes() override;
                void setSelection(const QList<std::shared_ptr<sh::filesystem::FilesystemNode>> nodes) override;
                QList<std::shared_ptr<sh::filesystem::FilesystemNode>> getAllVisibleNodes() override;

                /**
                 * @brief Returns the unique identifier.
                 */
                qint64 id();

                /**
                 * @brief Returns the filesystem model bound to this view.
                 */
                std::shared_ptr<sh::filesystem::FilesystemModelFileviewProxy> model();

            signals:
                void currentNodeChanged();
                void modelChanged();
            private:
                void _set_model();
                void setSelectedNode(std::shared_ptr<sh::filesystem::FilesystemNode> node);
                void setCheckedNodes(QList<std::shared_ptr<sh::filesystem::FilesystemNode>> nodes);
                std::shared_ptr<sh::filesystem::FilesystemModelFileviewProxy> _model;
                qint64 _id;
                FileViewMode _viewmode = FileViewMode::ListView;
                sh::filesystem::SizeFormatting _sizeformatting = sh::filesystem::SizeFormatting::SizeFormattingModePrefixes;
                bool _hiddenFilesVisible = false;
                int _iconDimension = 20;
                int _sortColumn = 0;
                Qt::SortOrder _sortOrder = Qt::SortOrder::AscendingOrder;
                std::shared_ptr<sh::filesystem::FilesystemNode> _selectedNode;  // the node the user has selected in the list
                QList<std::shared_ptr<sh::filesystem::FilesystemNode>> _checkedNodes;  // the nodes the user have checked in the list
                WebFileViewManager* _manager;
            };

        }
    }
}

#endif // WEBFILEVIEW_H
