/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2020 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "actions/actionsmanager.h"
#include "filesystem/filesystemmodeldirectorytreeproxy.h"
#include "filesystem/filesystemmodel.h"
#include "filesystem/filesystemnode.h"
#include "exceptions/runtimeexception.h"
#include "paneldetails/paneldetailmanager.h"
#include "base/shallotprocess.h"
#include "base/singletoninitializer.h"
#include "base/threaddispatcher.h"
#include "base/threadpool.h"
#include "base/mainthread.h"
#include "tools/misc.h"
#include "webaboutdialog.h"
#include "webactionexecutioninfodialog.h"
#include "webactionexecutioninfopanel.h"
#include "webactionmanager.h"
#include "webdetailpanelmanager.h"
#include "webexceptiondialog.h"
#include "webfilepropertydialog.h"
#include "webfileview.h"
#include "weblogviewdialog.h"
#include "webmainwindow.h"
#include "webmanagebookmarksdialog.h"
#include "webmanageprofilesdialog.h"
#include "webmodule.h"
#include "webopenwithdialog.h"
#include "webserverengine.h"
#include "webserverenginedispatcher.h"
#include "webstoreprofiledialog.h"
#include "webtuningdialog.h"

REGISTER_UI_MAINWINDOW(sh::ui::web, WebMainWindow)


bool sh::ui::web::WebMainWindow::tryCreateMainWindow(sh::ui::MainWindow*& mainWindow)
{
#ifdef WEBUI
    if (sh::base::ShallotProcess::instance()->uiMode() == "web") {
        mainWindow = WebServerEngineDispatcher::handleDispatching([](QString dispatcherUrl, QByteArray wtoken){ return new WebMainWindow(dispatcherUrl, wtoken); });
        return true;
    }
#endif
    return false;
}

sh::ui::web::WebMainWindow* sh::ui::web::WebMainWindow::_webMainWindow = nullptr;

sh::ui::web::WebMainWindow* sh::ui::web::WebMainWindow::mainWindow()
{
    return _webMainWindow;
}

sh::ui::web::WebMainWindow::WebMainWindow(QString dispatcherUrl, QByteArray wtoken)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    _webMainWindow = this;
    _webserver = sh::ui::web::WebServerEngine::createWorkerEngine(dispatcherUrl, wtoken);
    _webserver->addAndPreserveModule(std::shared_ptr<WebModule>(this, [](WebModule*){}));
    std::shared_ptr<sh::filesystem::FilesystemNode> rootnode = sh::filesystem::FilesystemModel::instance()->rootNode();
    QString rooteurl = rootnode->eurl()->asString();
    _webserver->setConfigValue("rootEurl", rooteurl);
    _webserver->setConfigValue("windowTitle", QChar(160));
    _webserver->setConfigValue("logseverity_debug", QString("%1").arg((int)sh::base::LogSeverity::DEBUG));
    _webserver->setConfigValue("logseverity_info", QString("%1").arg((int)sh::base::LogSeverity::INFO));
    _webserver->setConfigValue("logseverity_warning", QString("%1").arg((int)sh::base::LogSeverity::WARNING));
    _webserver->setConfigValue("logseverity_error", QString("%1").arg((int)sh::base::LogSeverity::ERROR));
    setCurrentEurlByNode(rootnode);
    setFileDetailsPanelVisible(true);
    _actionManager = std::make_shared<WebActionManager>();
    _detailPanelManager = std::make_shared<WebDetailPanelManager>();
    _dialogManager = std::make_shared<WebDialogManager>();
    _fileViewManager = std::make_shared<WebFileViewManager>();
    _webserver->addModule(_actionManager);
    _webserver->addModule(_detailPanelManager);
    _webserver->addModule(_dialogManager);
    _webserver->addModule(_fileViewManager);
    connect(_fileViewManager.get(), &WebFileViewManager::activeSelectionChanged, this, [=](){
        QList<std::shared_ptr<sh::filesystem::FilesystemNode>> nodes;
        auto currfileview = sh::ui::MainWindow::mainWindow()->currentFileView();
        if (currfileview)
            nodes = currfileview->selectedNodes();
        _detailPanelManager->load(nodes);
    });
    connect(_fileViewManager.get(), &WebFileViewManager::fileViewOptionsChanged, this, [=](int i){
        _emit_fileViewOptionsChanged(i);
        _emit_currentDirectoryChanged();
    });
    connect(_fileViewManager.get(), &WebFileViewManager::fileViewCountChanged, this, [=](){
        _emit_fileViewCountChanged();
    });
    connect(_fileViewManager.get(), &WebFileViewManager::currentFileViewChanged, this, [=](){
        _emit_currentFileViewChanged();
    });
    SH_LOG_INFO(tr("The Shallot worker web server is now listening on %1.").arg(_webserver->url().toString()));
    openUrlOnDesktop(_webserver->url());
}

void sh::ui::web::WebMainWindow::initialize()
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    addFileView();
    _actionManager->initialize();
    MainWindow::initialize();
}

std::shared_ptr<sh::ui::DialogManager> sh::ui::web::WebMainWindow::dialogManager()
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    return _dialogManager;
}

std::shared_ptr<sh::ui::web::WebActionManager> sh::ui::web::WebMainWindow::actionManager()
{
    return _actionManager;
}

sh::ui::web::WebMainWindow::~WebMainWindow()
{
    delete _webserver;
}

void sh::ui::web::WebMainWindow::openUrlOnDesktop(QUrl url)
{
    if (sh::base::ShallotProcess::instance()->parameterValueInt("openbrowser", 1))
        QDesktopServices::openUrl(url);
}

int sh::ui::web::WebMainWindow::fileViewCount()
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    return _fileViewManager->fileViewCount();
}

void sh::ui::web::WebMainWindow::addFileView(bool reinitialize)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    auto fileview = _fileViewManager->addFileView(reinitialize);
    auto sfileview = fileview.get();
    connect(sfileview, &WebFileView::selectionChanged, this, [=](){
        if (sfileview == _fileViewManager->currentFileView()) {
            auto currdir = currentDirectory();
            if (currdir) {
                auto instantiation = std::make_shared<sh::actions::ActionInstantiation>(sfileview->selectedNodes(), currdir);
                _actionManager->refreshToolbar(instantiation);
            }
        }
    });
}

sh::ui::FileView* sh::ui::web::WebMainWindow::currentFileView()
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    return _fileViewManager->currentFileView();
}

void sh::ui::web::WebMainWindow::removeFileView(int i)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    _fileViewManager->removeFileView(i);
}

sh::ui::FileView* sh::ui::web::WebMainWindow::getFileView(int i)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    return _fileViewManager->getFileView(i);
}

void sh::ui::web::WebMainWindow::jumpToNode(std::shared_ptr<sh::filesystem::FilesystemNode> node)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    if (_currentDirectory != node) {
        _currentDirectory = node;
        auto fileview = currentFileView();
        if (fileview)
            fileview->gotoDir(node);
        sh::base::ThreadPool::enqueueForce([=](){
            node->requestContainedItems();
            //this fetches all files in the view and their details in one for loop (on top of the armada of threads for the visible ones)
            sh::base::MainThread::dispatcher()->invokeSync([&](){
                int childnodecount = node->childnodeCount();
                for (int i=0; i<childnodecount; i++)
                    node->childnode(i)->requestDetails();
            });
        });
        _webserver->setConfigValue("windowTitle", _getWindowTitle());
        setCurrentEurlByNode(node);
        auto instantiation = std::make_shared<sh::actions::ActionInstantiation>(QList<std::shared_ptr<sh::filesystem::FilesystemNode>>(), node);
        _actionManager->refreshToolbar(instantiation);
    }
}

void sh::ui::web::WebMainWindow::setTreeVisible(bool v)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    sh::ui::MainWindow::setTreeVisible(v);
}

void sh::ui::web::WebMainWindow::setTreeSticky(bool v)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    sh::ui::MainWindow::setTreeSticky(v);
}

void sh::ui::web::WebMainWindow::setFileDetailsPanelVisible(bool v)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    sh::ui::MainWindow::setFileDetailsPanelVisible(v);
    _webserver->setConfigValue("detailPanelVisible", v?"1":"0");
}

std::shared_ptr<sh::filesystem::FilesystemNode> sh::ui::web::WebMainWindow::currentDirectory()
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    return _currentDirectory;
}

std::shared_ptr<sh::ui::ActionExecutionInfoPanel> sh::ui::web::WebMainWindow::addInfoPanel(int position, sh::actions::ActionExecutionInfo* info, QColor color)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    return _actionManager->addInfoPanel(position, info, color);
}

void sh::ui::web::WebMainWindow::jumpbarSetTextMode()
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    _webserver->setConfigValue("jumpbarMode", "1");
}

void sh::ui::web::WebMainWindow::jumpbarSetButtonMode()
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    _webserver->setConfigValue("jumpbarMode", "0");
}

bool sh::ui::web::WebMainWindow::jumpbarIsTextMode()
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    return _webserver->getConfigValue("jumpbarMode") == "1";
}

QString sh::ui::web::WebMainWindow::toolbarPosition()
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    return "N";
}

void sh::ui::web::WebMainWindow::setToolbarPosition(QString)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    // not supported
}

QString sh::ui::web::WebMainWindow::detailPanelPosition()
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    return "S";
}

void sh::ui::web::WebMainWindow::setDetailPanelPosition(QString)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    // not supported
}

std::shared_ptr<sh::ui::AboutDialog> sh::ui::web::WebMainWindow::showAboutDialog()
{
    return _dialogManager->createAndShowDialog<WebAboutDialog>();
}

std::shared_ptr<sh::ui::ExceptionDialog> sh::ui::web::WebMainWindow::showExceptionDialog(QString error1, QString error2, QString details, QString icon, bool mayRetry, bool mayClose, bool mayCancel, bool showLoglabelAndDetails)
{
    return _dialogManager->createAndShowDialog<WebExceptionDialog>(error1, error2, details, icon, mayRetry, mayClose, mayCancel, showLoglabelAndDetails);
}

std::shared_ptr<sh::ui::ManageProfilesDialog> sh::ui::web::WebMainWindow::showManageProfilesDialog()
{
    return _dialogManager->createAndShowDialog<WebManageProfilesDialog>();
}

std::shared_ptr<sh::ui::OpenWithDialog> sh::ui::web::WebMainWindow::showOpenWithDialog(std::shared_ptr<sh::filesystem::FilesystemNode> node, QList<std::shared_ptr<sh::tools::filetypes::OpenMethod>> openMethods)
{
    return _dialogManager->createAndShowDialog<WebOpenWithDialog>(node, openMethods);
}

std::shared_ptr<sh::ui::StoreProfileDialog> sh::ui::web::WebMainWindow::showStoreProfileDialog(std::shared_ptr<const sh::filesystem::Eurl> eurl)
{
    return _dialogManager->createAndShowDialog<WebStoreProfileDialog>(eurl);
}

std::shared_ptr<sh::ui::TuningDialog> sh::ui::web::WebMainWindow::showTuningDialog()
{
    return _dialogManager->createAndShowDialog<WebTuningDialog>();
}

std::shared_ptr<sh::ui::LogViewDialog> sh::ui::web::WebMainWindow::showLogViewDialog(QString headtext, QString subheadtxt, sh::base::LogSeverity minseverity)
{
    return _dialogManager->createAndShowDialog<WebLogViewDialog>(headtext, subheadtxt, minseverity);
}

std::shared_ptr<sh::ui::ManageBookmarksDialog> sh::ui::web::WebMainWindow::showManageBookmarksDialog()
{
    return _dialogManager->createAndShowDialog<WebManageBookmarksDialog>();
}

std::shared_ptr<sh::ui::FilePropertyDialog> sh::ui::web::WebMainWindow::showFilePropertyDialog(QList<std::shared_ptr<sh::filesystem::FilesystemNode>> nodes)
{
    return _dialogManager->createAndShowDialog<WebFilePropertyDialog>(nodes);
}

std::shared_ptr<sh::ui::ActionExecutionInfoDialog> sh::ui::web::WebMainWindow::createActionExecutionInfoDialog(sh::actions::ActionExecutionInfo* info)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    return _actionManager->addActionExecutionInfoDialog(info);
}

void sh::ui::web::WebMainWindow::handleCloseAppRejected(QString)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    // handled differently
}

void sh::ui::web::WebMainWindow::setTitlePattern(QString value)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    sh::ui::MainWindow::setTitlePattern(value);
    _webserver->setConfigValue("windowTitle", _getWindowTitle());
}

sh::ui::web::WebServerEngine* sh::ui::web::WebMainWindow::webserver()
{
    return _webserver;
}

bool sh::ui::web::WebMainWindow::executeCommand(WebServerEngineRequest* request, QString command)
{
    return executeCommand_byQObjectReflection(this, request, command);
}

void sh::ui::web::WebMainWindow::setCurrentEurlByNode(std::shared_ptr<sh::filesystem::FilesystemNode> node)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    QJsonArray neurlpath;
    auto eurl = node->eurl();
    while (node) {
        QJsonObject p;
        p["displayName"] = node->displayName();
        p["eurl"] = node->eurl()->asString();
        neurlpath.append(p);
        node = node->parentnode();
    }
    QJsonObject res;
    res["npath"] = neurlpath;
    QJsonArray eurlpath;
    while (eurl) {
        QJsonObject p;
        p["displayName"] = eurl->basename();
        p["eurl"] = eurl->asString();
        eurlpath.append(p);
        eurl = eurl->parentSegment();
    }
    res["epath"] = eurlpath;
    _webserver->setConfigValue("currentEurl", QString::fromLatin1(sh::tools::Misc::qjsonToJson(res)));
}

void sh::ui::web::WebMainWindow::cmd_filesystem_get_icon(WebServerEngineRequest* request)//TODO #what if the icons changes?
{
    QString seurl = request->data("eurl");
    auto eurl = sh::filesystem::Eurl::fromString(seurl);
    auto nlst = sh::filesystem::FilesystemModel::instance()->findNodesForEurl(eurl);
    if (nlst.isEmpty())
        throw sh::exceptions::RuntimeException(sh::exceptions::ExceptionData().details("node not found for: " + eurl->asString()));
    else {
        std::shared_ptr<sh::filesystem::FilesystemNode> node = nlst.first();
        QByteArray bArray;
        sh::base::MainThread::dispatcher()->invokeSync([&](){
            bArray = sh::tools::Misc::iconToPngByteArray(node->icon(), 32);
        });
        request->setResponse(bArray, "image/png");
    }
}

void sh::ui::web::WebMainWindow::cmd_filesystem_list_items(WebServerEngineRequest* request)
{
    QString seurl = request->data("eurl");
    auto eurl = sh::filesystem::Eurl::fromString(seurl);
    auto nlst = sh::filesystem::FilesystemModel::instance()->findNodesForEurl(eurl);
    if (nlst.isEmpty())
        throw sh::exceptions::RuntimeException(sh::exceptions::ExceptionData().details("node not found for: " + eurl->asString()));
    else {
        std::shared_ptr<sh::filesystem::FilesystemNode> node = nlst.first();
        QJsonArray result;
        auto dtmodel = sh::filesystem::FilesystemModelDirectoryTreeProxy::instance();
        node->requestContainedItems();
        sh::base::MainThread::dispatcher()->invokeSync([&](){
            QModelIndex nodei = sh::filesystem::FilesystemModel::instance()->getIndexForNode(node);
            QModelIndex pnodei = dtmodel->mapFromSource(nodei);
            int ccnt = dtmodel->rowCount(pnodei);
            for (int ic=0; ic<ccnt; ic++) {
                QModelIndex cpnodei = dtmodel->index(ic, 0, pnodei);
                QModelIndex cnodei = dtmodel->mapToSource(cpnodei);
                auto nl = sh::filesystem::FilesystemModel::instance()->getNodeForIndex(cnodei);
                if (nl->eurl().get()) { //TODO# this just skips the LoadOnDemandPlaceholderFilesystemNode ; a bit too easy :)
                    QJsonObject jno;
                    jno["displayName"] = nl->displayName();
                    jno["eurl"] = nl->eurl()->asString();
                    jno["nodetype"] = QString("%1").arg((int)nl->nodetype());
                    jno["isHidden"] = nl->isHidden();
                    result.append(jno);
                }
            }
        });
        request->setResponse(result);
    }
}

void sh::ui::web::WebMainWindow::cmd_log_as_text(WebServerEngineRequest* request)
{
    request->setResponse(sh::base::Logger::instance()->getLogAsText().toLatin1(), "text/plain");
}

void sh::ui::web::WebMainWindow::cmd_log_log_message(WebServerEngineRequest* request)
{
    QString message = request->data("message");
    QString source = request->data("source");
    auto severity = (sh::base::LogSeverity) request->data("severity").toInt();
    QString ssource = "sh::ui::web/client";
    if (!source.isEmpty())
        ssource += ": " + source;
    sh::base::Logger::instance()->log(message, severity, ssource);
}

void sh::ui::web::WebMainWindow::cmd_mainwindow_set_current_directory(WebServerEngineRequest* request)
{
    jumpToEurl(sh::filesystem::Eurl::fromString(request->data("eurl")));
}

void sh::ui::web::WebMainWindow::cmd_mainwindow_shallot_icon(WebServerEngineRequest* request)
{
    webserver()->getStaticFile(request, "/webui_client/shallot.128.png");
}

void sh::ui::web::WebMainWindow::cmd_mainwindow_show_logview(WebServerEngineRequest* request)
{
    QString headtext = request->data("headtext");
    QString subheadtxt = request->data("subheadtxt");
    QString minseverity = request->data("minseverity");
    sh::base::LogSeverity ominseverity = minseverity.isEmpty() ? sh::base::LogSeverity::_DEFAULT : ((sh::base::LogSeverity)minseverity.toInt());
    showLogViewDialog(headtext, subheadtxt, ominseverity);
}

void sh::ui::web::WebMainWindow::setCloseable(bool closeable, QString message)
{
    MainWindow::setCloseable(closeable, message);
    closeable = isCloseable(&message); // apply corrections
    _webserver->setConfigValue("preventCloseMessage", message);
    QUrl spcmurl = _webserver->rebaseUrl(_webserver->dispatcherUrl(), webserver()->getFullCommandUrl("dispatcher/set_prevent_close_message", {{"message", message}, {"wtoken", _webserver->dispatcherWtoken()}}, false));
    sh::tools::Misc::makeHttpRequest(spcmurl);
}
