/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2020 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef WEBMAINWINDOW_H
#define WEBMAINWINDOW_H

#include "ui/mainwindow.h"
#include "webdialog.h"
#include "webmodule.h"


namespace sh {
    namespace actions {
        namespace common { class ActionGroups; }
        class AbstractActionItem;
        class ActionInstantiation;
        class SubmenuActionItem;
    }
    namespace paneldetails { class PanelDetail; }
    namespace ui {
        namespace web {
            class WebActionManager;
            class WebActionExecutionInfoPanel;
            class WebActionExecutionInfoDialog;
            class WebDetailPanelManager;
            class WebServerEngine;
            class WebFileView;
            class WebFileViewManager;

            /** \brief The web main window implementation.
             */
            class WebMainWindow : public QObject, public sh::ui::MainWindow, public WebModule
            {
                friend class WebFileViewManager;
                friend class WebActionExecutionInfoPanel;
                friend class WebActionExecutionInfoDialog;
                friend class WebDetailPanelManager;
                Q_OBJECT
            public:
                static bool tryCreateMainWindow(MainWindow*& mainWindow);
                static WebMainWindow* mainWindow();
                explicit WebMainWindow(QString dispatcherUrl, QByteArray wtoken);
                ~WebMainWindow();

                /** \brief Returns the WebServerEngine instance for this main window.
                 */
                WebServerEngine* webserver();

                void initialize() override;
                std::shared_ptr<DialogManager> dialogManager() override;
                int fileViewCount() override;
                void addFileView(bool reinitialize=false) override;
                void removeFileView(int i) override;
                sh::ui::FileView* currentFileView() override;
                sh::ui::FileView* getFileView(int i) override;
                void jumpToNode(std::shared_ptr<sh::filesystem::FilesystemNode> node) override;
                std::shared_ptr<sh::filesystem::FilesystemNode> currentDirectory() override;
                void setTitlePattern(QString value) override;
                void setTreeVisible(bool v) override;
                void setTreeSticky(bool v) override;
                void setFileDetailsPanelVisible(bool v) override;
                std::shared_ptr<sh::ui::ActionExecutionInfoPanel> addInfoPanel(int position, sh::actions::ActionExecutionInfo* info, QColor color=QColor()) override;
                void jumpbarSetTextMode() override;
                void jumpbarSetButtonMode() override;
                bool jumpbarIsTextMode() override;
                QString toolbarPosition() override;
                void setToolbarPosition(QString pos) override;
                QString detailPanelPosition() override;
                void setDetailPanelPosition(QString pos) override;
                std::shared_ptr<AboutDialog> showAboutDialog() override;
                std::shared_ptr<ManageProfilesDialog> showManageProfilesDialog() override;
                std::shared_ptr<OpenWithDialog> showOpenWithDialog(std::shared_ptr<sh::filesystem::FilesystemNode> node, QList<std::shared_ptr<sh::tools::filetypes::OpenMethod>> openMethods) override;
                std::shared_ptr<StoreProfileDialog> showStoreProfileDialog(std::shared_ptr<const sh::filesystem::Eurl> eurl) override;
                std::shared_ptr<TuningDialog> showTuningDialog() override;
                std::shared_ptr<LogViewDialog> showLogViewDialog(QString headtext=QString(), QString subheadtxt=QString(), sh::base::LogSeverity minseverity=sh::base::LogSeverity::_DEFAULT) override;
                std::shared_ptr<ManageBookmarksDialog> showManageBookmarksDialog() override;
                std::shared_ptr<FilePropertyDialog> showFilePropertyDialog(QList<std::shared_ptr<sh::filesystem::FilesystemNode>> nodes) override;
                std::shared_ptr<ActionExecutionInfoDialog> createActionExecutionInfoDialog(sh::actions::ActionExecutionInfo* info) override;
                void handleCloseAppRejected(QString rejectmsg) override;

                /** \brief Opens a url on the user desktop, typically in the browser.
                 *
                 * Note: It obeys the 'openbrowser' ui parameter.
                 */
                static void openUrlOnDesktop(QUrl url);

            protected:
                virtual bool executeCommand(WebServerEngineRequest* request, QString command) override;
                std::shared_ptr<ExceptionDialog> showExceptionDialog(QString error1, QString error2, QString details, QString icon, bool mayRetry, bool mayClose, bool mayCancel, bool showLoglabelAndDetails);

            private slots:
                void cmd_filesystem_get_icon(WebServerEngineRequest* request);
                void cmd_filesystem_list_items(WebServerEngineRequest* request);
                void cmd_log_as_text(WebServerEngineRequest* request);
                void cmd_log_log_message(WebServerEngineRequest* request);
                void cmd_mainwindow_set_current_directory(WebServerEngineRequest* request);
                void cmd_mainwindow_shallot_icon(WebServerEngineRequest* request);
                void cmd_mainwindow_show_logview(WebServerEngineRequest* request);

            private:
                std::shared_ptr<sh::ui::web::WebActionManager> actionManager();
                void setCloseable(bool closeable, QString message=QString()) override;
                void setCurrentEurlByNode(std::shared_ptr<sh::filesystem::FilesystemNode> node);
                static WebMainWindow* _webMainWindow;
                bool _treeSticky = true;
                bool _detailsPanelVisible = true;
                QString _currentProfile;
                WebServerEngine* _webserver;
                std::shared_ptr<sh::filesystem::FilesystemNode> _currentDirectory = 0;
                std::shared_ptr<WebActionManager> _actionManager;
                std::shared_ptr<WebDetailPanelManager> _detailPanelManager;
                std::shared_ptr<WebDialogManager> _dialogManager;
                std::shared_ptr<WebFileViewManager> _fileViewManager;
                QString _lastCloseableState;
            };

        }
    }
}

#endif // WEBMAINWINDOW_H
