/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2020 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "base/iconmanager.h"
#include "filesystem/eurl.h"
#include "filesystem/filesystemmodel.h"
#include "filesystem/filesystemnode.h"
#include "tools/misc.h"
#include "webmanagebookmarksdialog.h"
#include "webmainwindow.h"


sh::ui::web::WebManageBookmarksDialog::WebManageBookmarksDialog() : WebDialog("ManageBookmarksDialog")
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    setDialogProperty("folderIcon", sh::tools::Misc::iconToBase64SrcEncoding(base::IconManager::instance()->getIcon("folder-bookmark"), 24));
}

void sh::ui::web::WebManageBookmarksDialog::cmd__add_bookmark__M(WebServerEngineRequest* request)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    QString sfolder = request->data("folder");
    QStringList folder;
    QJsonDocument jfolder = QJsonDocument::fromJson(sfolder.toLocal8Bit());
    foreach (QJsonValue jv, jfolder.array())
        folder.append(jv.toString());
    auto currentNode = sh::ui::web::WebMainWindow::mainWindow()->currentDirectory();
    auto currentEurl = (currentNode ? currentNode : sh::filesystem::FilesystemModel::instance()->rootNode())->eurl();
    sh::tools::BookmarkManager::instance()->addBookmark(folder, "", currentEurl);
}

void sh::ui::web::WebManageBookmarksDialog::cmd__delete_bookmark(WebServerEngineRequest* request)
{
    QString id = request->data("id");
    sh::tools::BookmarkManager::instance()->removeBookmark(id);
}

void sh::ui::web::WebManageBookmarksDialog::cmd__get(WebServerEngineRequest* request)
{
    QJsonArray jresult;
    std::function<void(QList<std::shared_ptr<sh::ui::ManageBookmarksDialog::BM>>, QJsonArray*)> _readbookmarks = [&](QList<std::shared_ptr<sh::ui::ManageBookmarksDialog::BM>> cbookmarks, QJsonArray* jnode) {
        foreach (std::shared_ptr<sh::ui::ManageBookmarksDialog::BM> bookmark, cbookmarks) {
            QJsonObject jbookmark;
            if (bookmark->isFolder()) {
                auto fbookmark = bookmark->asFolder();
                jbookmark["label"] = fbookmark->name;
                jbookmark["treelabel"] = fbookmark->name;
                QJsonArray jcnode;
                _readbookmarks(fbookmark->children, &jcnode);
                jbookmark["children"] = jcnode;
                jbookmark["isFolder"] = true;
            }
            else {
                auto bbbookmark = bookmark->asBookmark()->bookmark;
                jbookmark["label"] = bbbookmark->label();
                jbookmark["treelabel"] = bbbookmark->label().isEmpty() ? bbbookmark->eurl()->asString() : bbbookmark->label();
                jbookmark["location"] = bbbookmark->eurl()->asString();
                jbookmark["isFolder"] = false;
                jbookmark["id"] = bbbookmark->id();
            }
            jnode->append(jbookmark);
        }
    };
    _readbookmarks(getBookmarksStructure(), &jresult);
    request->setResponse(jresult);
}

void sh::ui::web::WebManageBookmarksDialog::cmd__move_bookmark_down(WebServerEngineRequest* request)
{
    QString id = request->data("id");
    sh::tools::BookmarkManager::instance()->moveBookmarkDown(id);
}

void sh::ui::web::WebManageBookmarksDialog::cmd__move_bookmark_to_folder(WebServerEngineRequest* request)
{
    QString id = request->data("id");
    QString destination = request->data("destination");
    QStringList newfolder = destination.split('/');
    newfolder.takeFirst();
    if (newfolder.first().isEmpty())
        newfolder.takeFirst();
    sh::tools::BookmarkManager::instance()->moveBookmarkToFolder(id, newfolder);
}

void sh::ui::web::WebManageBookmarksDialog::cmd__move_bookmark_up(WebServerEngineRequest* request)
{
    QString id = request->data("id");
    sh::tools::BookmarkManager::instance()->moveBookmarkUp(id);
}

void sh::ui::web::WebManageBookmarksDialog::cmd__store(WebServerEngineRequest* request)
{
    QString label = request->data("label");
    QString location = request->data("location");
    QString id = request->data("id");
    auto eurl = sh::filesystem::Eurl::fromString(location);
    if (eurl)
        sh::tools::BookmarkManager::instance()->changeBookmark(id, label, eurl);
}

bool sh::ui::web::WebManageBookmarksDialog::executeDialogCommand(WebServerEngineRequest* request, QString command)
{
    return WebModule::executeCommand_byQObjectReflection(this, request, command, "managebookmarks");
}
