/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2020 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "settings/profilenode.h"
#include "settings/setting.h"
#include "settings/settingsmanager.h"
#include "base/iconmanager.h"
#include "tools/misc.h"
#include "webmanageprofilesdialog.h"


sh::ui::web::WebManageProfilesDialog::WebManageProfilesDialog() : WebDialog("ManageProfilesDialog")
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
}

bool sh::ui::web::WebManageProfilesDialog::executeDialogCommand(WebServerEngineRequest* request, QString command)
{
    return WebModule::executeCommand_byQObjectReflection(this, request, command, "manageprofiles");
}

void sh::ui::web::WebManageProfilesDialog::cmd__get__M(WebServerEngineRequest* request)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    QString currentprofile = request->data("currentprofile");
    QJsonObject result;
    QStringList profiles;
    int selected = -1;
    getProfileList(currentprofile, profiles, selected);
    QJsonArray jprofiles;
    foreach (QString profile, profiles) {
        QJsonObject jprofile;
        jprofile["name"] = profile;
        QJsonArray jnodes;
        foreach (Node node, getNodesList(profile)) {
            QJsonObject jnode;
            jnode["text"] = node.text;
            jnode["icon"] = node.icon;
            jnode["nodeId"] = node.nodeId;
            jnode["withSubfolders"] = node.pnode->withSubfolders();
            jnode["inheritsFromDescription"] = _generateProfileListString(node.pnode->inheritsFrom());
            QJsonArray jsettings;
            foreach (QString sname, node.pnode->getSettingNames()) {
                std::shared_ptr<sh::settings::Setting> setting = sh::settings::SettingsManager::instance()->getSettingByName(sname);
                if (setting) {
                    QJsonObject jsetting;
                    jsetting["description"] = setting->description();
                    sh::exceptions::Exception::executeGuarded_errorpanel([&](){
                        jsetting["valueDescription"] = setting->valueDescription(node.pnode->getSettingValue(sname));
                    });
                    jsetting["onlyInFileviewIndex"] = node.pnode->getSettingOnlyInFileviewIndex(sname);
                    jsettings.append(jsetting);
                }
            }
            jnode["settings"] = jsettings;
            jnodes.append(jnode);
        }
        jprofile["nodes"] = jnodes;
        jprofiles.append(jprofile);
    }
    result["profiles"] = jprofiles;
    result["selected"] = selected;
    request->setResponse(result);
}

void sh::ui::web::WebManageProfilesDialog::cmd__icon(WebServerEngineRequest* request)
{
    QString name = request->data("name");
    QByteArray result = sh::tools::Misc::iconToPngByteArray(sh::base::IconManager::instance()->getIcon(name), 32);
    request->setResponse(result, "image/png");
}

void sh::ui::web::WebManageProfilesDialog::cmd__remove_node__M(WebServerEngineRequest* request)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    QString nodeId = request->data("nodeId");
    if (!nodeId.isEmpty())
        sh::exceptions::Exception::executeGuarded([&](){
            sh::settings::SettingsManager::instance()->removeNode(nodeId);
        }, sh::exceptions::ExecuteGuardFlag::RESUMEABLE_PROGRAM_ERROR_END_WITH_USER_FEEDBACK_HERE);
}

void sh::ui::web::WebManageProfilesDialog::cmd__remove_profile__M(WebServerEngineRequest* request)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    QString profile = request->data("profile");
    if (!profile.isEmpty())
        sh::exceptions::Exception::executeGuarded([&](){
            sh::settings::SettingsManager::instance()->removeProfile(profile);
        }, sh::exceptions::ExecuteGuardFlag::RESUMEABLE_PROGRAM_ERROR_END_WITH_USER_FEEDBACK_HERE);
}
