/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2020 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef WEBMODULE_H
#define WEBMODULE_H

#include <functional>
#include <memory>
#include <QObject>


namespace sh{
    namespace ui{
        namespace web{
            class WebServerEngine;
            class WebServerEngineRequest;

            /** \brief Base class for modules, which provide some functionaly on top of
             * a web server engine.
             */
            class WebModule
            {
            public:
                WebModule();
                virtual ~WebModule();

                /** \brief Executes `command` as requested by the browser, if responsible.
                 *
                 * Read parameters, check if this module is responsible for answering, and then write an answer with `request`.
                 *
                 * Implementations often use executeCommand_byQObjectReflection() for convenience.
                 * @a_dooverride_optional
                 * @returns If it handled (answered) the command.
                 */
                virtual bool executeCommand(WebServerEngineRequest* request, QString command);

                /** \brief Returns the root ('index.html') content.
                 * @a_dooverride_optional
                 */
                virtual bool rootCommand(WebServerEngineRequest* request, QString webadapter);

                /**
                 * \brief Assigns this web module to a sh::ui::web::WebServerEngine.
                 * @a_infrastructure
                 */
                void setEngine(WebServerEngine* webserver);

                /**
                 * \brief Returns the sh::ui::web::WebServerEngine this web module is assigned to (may be nullptr).
                 */
                WebServerEngine* webserver();

                /**
                 * \brief Convenience function often used by executeCommand().
                 *
                 * For a QObject, it searches there for a slot with the name `cmd_{command}`, with `'/'`s replaced by `'_'`s
                 * (so, for the command `"foo/get_bars"`, it searches for slot `cmd_foo_get_bars`). If it finds one,
                 * it gets executed (in a worker thread). The request is considered as handled, if so.
                 *
                 * If there is such a slot, but with appended `__M` in name, it is called in main thread!
                 * @returns If it handled (answered) the command.
                 */
                static bool executeCommand_byQObjectReflection(QObject* o, WebServerEngineRequest* request, QString command, QString justLeftside=QString());

            private:
                WebServerEngine* _webserver = nullptr;

            };

        }
    }
}

#endif // WEBMODULE_H
