/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2020 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "actions/actionactionitem.h"
#include "actions/submenuactionitem.h"
#include "base/shallotprocess.h"
#include "filesystem/filesystemmodel.h"
#include "filesystem/filesystemnode.h"
#include "filesystem/filesystemnodelist.h"
#include "exceptions/runtimeexception.h"
#include "paneldetails/paneldetailmanager.h"
#include "base/logger.h"
#include "base/mainthread.h"
#include "ui/web/engines/pocowebserverengine.h"
#include "webactionexecutioninfodialog.h"
#include "webactionexecutioninfopanel.h"
#include "webi18n.h"
#include "webfileview.h"
#include "webmainwindow.h"
#include "webmodule.h"
#include "webserverengine.h"


sh::ui::web::WebServerEngineMainModule::WebServerEngineMainModule()
            : _lastRequestTime(QDateTime::currentSecsSinceEpoch()), _eventEntryCleanupThread(this), _stopWhenIdleThread(this)
{
    _eventEntryCleanupThread.start();
    _stopWhenIdleThread.start();
}

sh::ui::web::WebServerEngineMainModule::~WebServerEngineMainModule()
{
    _eventEntryCleanupThread.requestInterruption();
    _stopWhenIdleThread.requestInterruption();
    {
        QMutexLocker _l1(&_mutex_eventEntries);
        foreach (EventEntry* k, _eventEntries)
            delete k;
    }
}

bool sh::ui::web::WebServerEngineMainModule::executeCommand(WebServerEngineRequest* request, QString command)
{
    return WebModule::executeCommand_byQObjectReflection(this, request, command);
}

void sh::ui::web::WebServerEngineMainModule::cmd_dispatcher_toclient_heartbeat(WebServerEngineRequest* request)
{
    QMutexLocker _l1(&_mutex_lastRequestTime);
    _lastRequestTime = QDateTime::currentSecsSinceEpoch();
    request->setResponse("", "text/html", WebServerEngineRequest::HTTP_OK);
}

void sh::ui::web::WebServerEngineMainModule::cmd_events_get_next(WebServerEngineRequest* request)
{
    qint64 lastid = (qint64) request->data("lastid").toLongLong();
    QJsonArray result;
    {
        for (int iti=29; iti>=0; iti--) {
            QMutexLocker _l1(&_mutex_eventEntries);
            qint64 i;
            for (i=_eventEntries.count()-1; i>0; i--) {
                if (_eventEntries[i]->id <= lastid)
                    break;
            }
            i++;
            for (; i<_eventEntries.count(); i++) {
                EventEntry* ee = _eventEntries[i];
                QJsonObject jee;
                jee["id"] = ee->id;
                jee["name"] = ee->name;
                jee["data"] = ee->data;
                result.append(jee);
            }
            if (result.isEmpty()) {
                if (iti > 0) /* there is no more iteration, so waiting is pointless */
                    _eventEntriesCondition.wait(&_mutex_eventEntries, 1000);
            }
            else
                break;
        }
    }
    request->setResponse(result);
}

qint64 sh::ui::web::WebServerEngineMainModule::lastRequestTime()
{
    QMutexLocker _l1(&_mutex_lastRequestTime);
    return _lastRequestTime;
}

QList<std::shared_ptr<sh::ui::web::WebModule>> sh::ui::web::WebServerEngine::modules()
{
    QMutexLocker _l1(&_modulesMutex);
    QList<std::shared_ptr<sh::ui::web::WebModule>> result;
    auto im = _modules.begin();
    while (im != _modules.end()) {
        std::weak_ptr<sh::ui::web::WebModule> module = *im;
        std::shared_ptr<sh::ui::web::WebModule> smodule = module.lock();
        if (smodule) {
            result.append(smodule);
            im++;
        }
        else // should not really happen ?!
            im = _modules.erase(im);
    }
    return result;
}

QUrl sh::ui::web::WebServerEngine::externalUrl()
{
    return _externalurl.isEmpty() ? url() : _externalurl;
}

QUrl sh::ui::web::WebServerEngine::dispatcherUrl()
{
    return _dispatcherurl;
}

QByteArray sh::ui::web::WebServerEngine::dispatcherWtoken()
{
    return _wtoken;
}

bool sh::ui::web::WebServerEngine::rootCommand(WebServerEngineRequest* request, QString webadapter)
{
    QList<std::shared_ptr<sh::ui::web::WebModule>> aliveModules = modules();
    bool result = false;
    sh::exceptions::Exception::executeGuarded([&](){
        foreach (std::shared_ptr<sh::ui::web::WebModule> module, aliveModules) {
            if ((result = module->rootCommand(request, webadapter)))
                break;
        }
    }, sh::exceptions::ExecuteGuardFlag::RESUMEABLE_PROGRAM_ERROR_END_WITH_USER_FEEDBACK_HERE);
    return result;
}

bool sh::ui::web::WebServerEngine::executeCommand(WebServerEngineRequest* request, QString command)
{
    QList<std::shared_ptr<sh::ui::web::WebModule>> aliveModules = modules();
    bool result = false;
    sh::exceptions::Exception::executeGuarded([&](){
        try {
            foreach (std::shared_ptr<sh::ui::web::WebModule> module, aliveModules) {
                if ((result = module->executeCommand(request, command)))
                    break;
            }
        }
        catch (...) {
            request->setResponse(QByteArray(), "application/octet-stream", WebServerEngineRequest::HTTP_INTERNAL_SERVER_ERROR);
            throw;
        }
    }, sh::exceptions::ExecuteGuardFlag::RESUMEABLE_PROGRAM_ERROR_END_WITH_USER_FEEDBACK_HERE);
    return result;
}

int sh::ui::web::WebServerEngine::minimumPort()
{
    return _minport;
}

int sh::ui::web::WebServerEngine::maximumPort()
{
    return _maxport;
}

sh::ui::web::WebServerEngineRequest::WebServerEngineRequest(QString url, QString clientHost) : _url(url), _clientHost(clientHost)
{
    _path = _url.path();
    QUrlQuery query(_url);
    foreach (auto p, query.queryItems(QUrl::FullyDecoded))
        _getdata[p.first] = p.second;
}

QUrl sh::ui::web::WebServerEngineRequest::url()
{
    return _url;
}

QString sh::ui::web::WebServerEngineRequest::clientHost()
{
    return _clientHost;
}

QString sh::ui::web::WebServerEngineRequest::path()
{
    return _path;
}

QString sh::ui::web::WebServerEngineRequest::data(QString key)
{
    return _getdata[key];
}

int sh::ui::web::WebServerEngineRequest::responseCode()
{
    return _responseCode;
}

QString sh::ui::web::WebServerEngineRequest::responseMimeType()
{
    return _responseMimeType;
}

QByteArray sh::ui::web::WebServerEngineRequest::response()
{
    return _response;
}

bool sh::ui::web::WebServerEngineRequest::responseSet()
{
    return _responseSet;
}

void sh::ui::web::WebServerEngineRequest::setResponse(QByteArray response, QString mimetype, int responseCode)
{
    _responseSet = true;
    _response = response;
    _responseMimeType = mimetype;
    _responseCode = responseCode;
}

void sh::ui::web::WebServerEngineRequest::setResponse(QJsonArray response, int responseCode)
{
    setResponse(sh::tools::Misc::qjsonToJson(response), "application/json", responseCode);
}

void sh::ui::web::WebServerEngineRequest::setResponse(QJsonObject response, int responseCode)
{
    setResponse(sh::tools::Misc::qjsonToJson(response), "application/json", responseCode);
}

sh::ui::web::WebServerEngine::WebServerEngine(bool withMainModule, int minport, int maxport, QString externalUrl, QString dispatcherUrl, QByteArray wtoken)
                    : _jok(QJsonObject({{"ok", true}})), _modulesMutex(QMutex::Recursive), _minport(minport), _maxport(maxport), _externalurl(externalUrl), _dispatcherurl(dispatcherUrl), _wtoken(wtoken)
{
    _webstaticpath = sh::base::ShallotProcess::shallotDataDir() + "/web_static";
    if (withMainModule)
        addModule(_mainmodule=std::make_shared<WebServerEngineMainModule>());
}

sh::ui::web::WebServerEngine::~WebServerEngine()
{
}

sh::ui::web::WebServerEngine* sh::ui::web::WebServerEngine::createDispatcherEngine()
{
    return _createEngine(false, QString(), QByteArray());
}

sh::ui::web::WebServerEngine* sh::ui::web::WebServerEngine::createWorkerEngine(QString dispatcherUrl, QByteArray wtoken)
{
    return _createEngine(true, dispatcherUrl, wtoken);
}

sh::ui::web::WebServerEngine* sh::ui::web::WebServerEngine::_createEngine(bool withMainModule, QString dispatcherUrl, QByteArray wtoken)
{
    auto minport = sh::base::ShallotProcess::instance()->minport();
    auto maxport = sh::base::ShallotProcess::instance()->maxport();
    QString externalBaseUrl = sh::base::ShallotProcess::instance()->parameterValue("externalBaseUrl");
    QString webserverenginename = "unknown";
    sh::ui::web::WebServerEngine* result = nullptr;
#if WEBUI_ENGINE == 1
    webserverenginename = "Poco::Net";
    result = new PocoNetWebServerEngine(withMainModule, minport, maxport, externalBaseUrl, dispatcherUrl, wtoken);
#endif
    if (result)
        SH_LOG_DEBUG(QString("Web ui server engine: %1.").arg(webserverenginename))
    else
        SH_LOG_DEBUG("No web server engine implementation found.")
    return result;
}

QString sh::ui::web::WebServerEngine::staticPathToMimeType(QString path)
{
    path = path.toLower();
    if (path.endsWith(".htm") || path.endsWith(".html"))
        return "text/html";
    else if (path.endsWith(".css"))
        return "text/css";
    else if (path.endsWith(".js"))
        return "text/javascript";
    else if (path.endsWith(".png"))
        return "image/png";
    return "application/octet-stream";
}

QString sh::ui::web::WebServerEngine::staticFilesPath()
{
    return _webstaticpath;
}

QUrl sh::ui::web::WebServerEngine::rebaseUrl(QUrl rootUrl, QUrl url)
{
    QUrl result = rootUrl;
    result.setPath(url.path());
    result.setQuery(url.query());
    return result;
}

sh::ui::web::WebServerEngineMainModule::EventEntry::EventEntry(qint64 id, qint64 time, QString name, QString data) : id(id), time(time), name(name), data(data)
{
}

sh::ui::web::WebServerEngineMainModule::EventEntryCleanupThread::EventEntryCleanupThread(WebServerEngineMainModule* module): module(module)
{
}

void sh::ui::web::WebServerEngineMainModule::EventEntryCleanupThread::run()
{
    int clind = 0;
    while (!isInterruptionRequested()) {
        QThread::msleep(1000);
        clind--;
        if (clind < 0) {
            clind = 30;
            qint64 now = QDateTime::currentSecsSinceEpoch();
            QMutexLocker _l1(&module->_mutex_eventEntries);
            int irmto;
            for (irmto=0; irmto<module->_eventEntries.count(); irmto++) {
                EventEntry* ee = module->_eventEntries[irmto];
                if (now - ee->time < 60)
                    break;
            }
            for (int i=0; i<irmto; i++) {
                delete module->_eventEntries.first();
                module->_eventEntries.removeFirst();
            }
        }
    }
}

sh::ui::web::WebServerEngineMainModule::StopWhenIdleThread::StopWhenIdleThread(WebServerEngineMainModule* module): module(module)
{
}

void sh::ui::web::WebServerEngineMainModule::StopWhenIdleThread::run()
{
    int i = 0;
    while (!isInterruptionRequested()) {
        QThread::msleep(1000);
        if (!(i = (i + 1) % 20)) {
            if (module->lastRequestTime() < QDateTime::currentSecsSinceEpoch() - 20) {
                SH_LOG_WARNING(QString("Worker process %1 stops due to communication inactivity.").arg(QCoreApplication::applicationPid()));
                try {
                    sh::base::MainThread::dispatcher()->invokeSync([&](){
                        sh::ui::MainWindow::mainWindow()->_closeDirectly();
                    });
                }
                catch(...) {}
            }
        }
    }
}

QString sh::ui::web::WebServerEngine::staticPathToUrl(QString path)
{
    QString result = "static";
    if (!path.startsWith("/"))
        result += "/";
    result += path;
    return result;
}

QUrl sh::ui::web::WebServerEngine::getFullCommandUrl(QString command, WebCommandData data, bool external)
{
    QList<QPair<QString, QString>> ldata;
    foreach(auto x, data.toStdMap())
        ldata.append(QPair<QString, QString>(x.first, x.second));
    QUrlQuery urlq;
    urlq.setQueryItems(ldata);
    QUrl res = external ? externalUrl() : url();
    QString pcmd = res.path();
    if (!pcmd.endsWith("/"))
        pcmd += "/";
    pcmd += command;
    res.setPath(pcmd);
    res.setQuery(urlq);
    return res;
}

bool sh::ui::web::WebServerEngineMainModule::rootCommand(WebServerEngineRequest* request, QString webadapter)
{
    webserver()->getStaticFile(request, "/webui_client/index.html");
    QString scontent = QString::fromLocal8Bit(request->response());
    scontent = scontent.replace("//webadapter//", webserver()->staticPathToUrl("webui_client/webadapter." + webadapter + ".js"));
    QString jsconfiguration;
    {
        QMutexLocker _l1(&_mutex_eventEntries);
        QMutexLocker _l2(&webserver()->_configValuesMutex);
        foreach(QString kcfgitm, webserver()->_configValues.keys())
            jsconfiguration += QString("shwebui.config._s('%1',decodeURIComponent('%2'));")
                                           .arg(kcfgitm).arg(QString::fromLocal8Bit(QUrl::toPercentEncoding(webserver()->_configValues[kcfgitm])));
       jsconfiguration += QString("shwebui.eventfetcher.skipIds(%1);").arg(_eventEntriesLastId);
    }
    scontent = scontent.replace("//configuration//", jsconfiguration);
    scontent = scontent.replace("//i18n//", WebI18n::get());
    QString nscontent;
    int inscontent = 0;
    QRegularExpressionMatch m;
    for(auto gm = _restatic.globalMatch(scontent); gm.hasNext(); m=gm.next()) {
        auto m = gm.peekNext();
        nscontent += scontent.mid(inscontent, m.capturedStart(0)-inscontent) + "\"" + webserver()->staticPathToUrl(m.captured(1)) + "\"";
        inscontent = m.capturedEnd(0);
    }
    nscontent += scontent.mid(inscontent);
    request->setResponse(nscontent.toLocal8Bit(), request->responseMimeType());
    return true;
}

QString sh::ui::web::WebServerEngine::getConfigValue(QString key)
{
    QMutexLocker _l1(&_configValuesMutex);
    return _configValues.value(key);
}

void sh::ui::web::WebServerEngine::setConfigValue(QString key, QString value)
{
    bool changed = false;
    {
        QMutexLocker _l1(&_configValuesMutex);
        changed = (!_configValues.contains(key)) || (_configValues.value(key) != value);
        _configValues[key] = value;
    }
    if (changed) {
        QJsonObject jee;
        jee["key"] = key;
        jee["value"] = value;
        triggerEvent("configuration/changed", jee);
    }
}

void sh::ui::web::WebServerEngine::getStaticFile(WebServerEngineRequest* request, QString path)
{
    QFile fpath(staticFilesPath() + "/" + path);
    fpath.open(QFile::ReadOnly);
    if (!fpath.isOpen())
        request->setResponse(QByteArray(), "application/octet-stream", WebServerEngineRequest::HTTP_NOT_FOUND);
    else
        request->setResponse(fpath.readAll(), staticPathToMimeType(path));
}

void sh::ui::web::WebServerEngine::triggerEvent(QString name, QJsonObject data)
{
    _mainmodule->triggerEvent(name, sh::tools::Misc::qjsonToJson(data));
}

void sh::ui::web::WebServerEngine::triggerEvent(QString name, QJsonValue data)
{
    _mainmodule->triggerEvent(name, sh::tools::Misc::qjsonToJson(data));
}

void sh::ui::web::WebServerEngineMainModule::triggerEvent(QString name, QString data)
{
    QMutexLocker _l1(&_mutex_eventEntries);
    _eventEntries.append(new EventEntry(++_eventEntriesLastId, QDateTime::currentSecsSinceEpoch(), name, data));
    _eventEntriesCondition.wakeAll();
}

void sh::ui::web::WebServerEngine::addAndPreserveModule(std::shared_ptr<sh::ui::web::WebModule> module)
{
    QMutexLocker _l1(&_modulesMutex);
    _modules.append(module);
    _smodules.append(module);
    module->setEngine(this);
}

void sh::ui::web::WebServerEngine::addModule(std::shared_ptr<sh::ui::web::WebModule> module)
{
    QMutexLocker _l1(&_modulesMutex);
    _modules.append(module);
    module->setEngine(this);
}

void sh::ui::web::WebServerEngine::removeModule(WebModule* module)
{
    QMutexLocker _l1(&_modulesMutex);
    auto im = _modules.begin();
    while (im != _modules.end()) {
        std::shared_ptr<WebModule> simm = (*im).lock();
        if (simm) {
            if (simm.get() == module)
                im = _modules.erase(im);
            else
                im++;
        }
        else // should not really happen ?!
            im = _modules.erase(im);
    }
    auto ism = _smodules.begin();
    while (ism != _smodules.end()) {
        std::shared_ptr<WebModule> sismm = (*ism);
        if (sismm.get() == module)
            ism = _smodules.erase(ism);
        else
            ism++;
    }
}

qint64 sh::ui::web::WebServerEngine::currentTimestamp()
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    return _currentTimestamp++;
}

QRegularExpression sh::ui::web::WebServerEngineMainModule::_restatic("['\"]//static//([^'\"]+)['\"]");


//TODO# shallot web workers (also the redirector?) crash on exit (maybe it even crashes with a null mainwindow and no web?! could be...)
//TODO# react on Ctrl+C on terminal
