/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2020 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef WEBSERVERENGINE_H
#define WEBSERVERENGINE_H

#include <memory>
#include <QCoreApplication>
#include <QFile>
#include <QHostAddress>
#include <QJsonObject>
#include <QMutex>
#include <QRegularExpression>
#include <QThread>
#include <QUrl>
#include <QUrlQuery>
#include <QVector>
#include <QWaitCondition>
#include "ui/web/webmodule.h"


namespace sh {
    namespace filesystem { class FilesystemNode; }
    namespace actions { class AbstractActionItem; }
    namespace ui {
        namespace web {
            class WebServerEngineMainModule;

            typedef QMap<QString, QString> WebCommandData;

            /** \brief A web request in a WebServerEngine.
             *
             * Override this together with WebServerEngine.
             */
            class WebServerEngineRequest
            {
            public:
                static const int HTTP_OK = 200;
                static const int HTTP_SEE_OTHER = 303;
                static const int HTTP_NOT_FOUND = 404;
                static const int HTTP_INTERNAL_SERVER_ERROR = 500;
                static const int HTTP_BAD_GATEWAY = 502;

                /** @a_infrastructure
                 */
                WebServerEngineRequest(QString url, QString clientHost);

                /** \brief Returns the requested url.
                 *
                 * Note: The url should be considered as opaque, since the exact structure depends on the engine implementation.
                 */
                QUrl url();

                /** \brief Returns the client host address.
                 */
                QString clientHost();

                /** \brief Returns the path part of the requested url.
                 *
                 * Note: The url should be considered as opaque, since the exact structure depends on the engine implementation.
                 */
                QString path();

                /** \brief Returns data passed as parameter along with the request by `key`.
                 */
                QString data(QString key);

                /** \brief Returns the answered (http) response code.
                 */
                int responseCode();

                /** \brief Returns the mimetype of the answer content.
                 */
                QString responseMimeType();

                /** \brief Returns the answer content.
                 */
                QByteArray response();

                /** \brief Returns if a response was set.
                 */
                bool responseSet();

                /** \brief Sets the response.
                 */
                void setResponse(QByteArray response, QString mimetype, int responseCode=HTTP_OK);
                void setResponse(QJsonArray response, int responseCode=HTTP_OK);
                void setResponse(QJsonObject response, int responseCode=HTTP_OK);

                /** \brief Returns an http header value by key.
                 * @a_dooverride
                 */
                virtual QString headerData(QString key) =0;

                /** \brief Returns an http cookie stored on browser side.
                 *
                 * Note: This only works on dispatcher level and not in usual WebModule commands.
                 * @a_dooverride
                 * @a_infrastructure
                 */
                virtual QString getCookie(QString key) =0;

                /** \brief Sets an http cookie to be returned to the browser side.
                 *
                 * Note: This only works on dispatcher level and not in usual WebModule commands.
                 * @a_dooverride
                 * @a_infrastructure
                 */
                virtual void setCookie(QString key, QString value, int maxAgeSecs=-1) =0;

            private:
                QUrl _url;
                QString _clientHost;
                QString _path;
                QMap<QString, QString> _getdata;
                int _responseCode = HTTP_NOT_FOUND;
                QString _responseMimeType = "application/octet-stream";
                QByteArray _response;
                bool _responseSet = false;
            };

            /** \brief Abstract base class for a web server engine.
             *
             * Such an engine implements an http server which a browser can connect to.
             */
            class WebServerEngine
            {
                friend class WebServerEngineMainModule;
            public:

                /** @a_infrastructure
                 */
                WebServerEngine(bool withMainModule, int minport, int maxport, QString externalUrl, QString dispatcherUrl, QByteArray wtoken);

                virtual ~WebServerEngine();

                /** \brief Creates and returns a web engine for a dispatcher.
                 *
                 */
                static WebServerEngine* createDispatcherEngine();

                /** \brief Creates and returns a web engine for a worker.
                 *
                 * \param dispatcherUrl: The root url of the dispatcher.
                 */
                static WebServerEngine* createWorkerEngine(QString dispatcherUrl, QByteArray wtoken);

                /** \brief Returns the home url of this engine (where the end user can point the browser to).
                 * @a_dooverride
                 */
                virtual QUrl url() =0;

                /** \brief Returns the external root url for creating full externally valid urls (even behind reverse proxies).
                 *
                 * Returns url() if no external root url is specified.
                 *
                 * Change this by the cmdline parameter `'externalUrl'`.
                 */
                QUrl externalUrl();

                /** \brief Returns the internal root url of the dispatcher (workers only).
                 */
                QUrl dispatcherUrl();

                /** \brief Returns the wtoken got from the dispatcher (workers only).
                 */
                QByteArray dispatcherWtoken();

                /** \brief Generates a url path for a command.
                 *
                 * Typically only needed for full url generation on server side.
                 * @a_dooverride_optional
                 */
                virtual QUrl getFullCommandUrl(QString command, WebCommandData data, bool external);

                /** \brief Triggers an event (at the clients).
                 */
                void triggerEvent(QString name, QJsonObject data);
                void triggerEvent(QString name, QJsonValue data=QJsonValue::Undefined);

                /** \brief Returns the value for a runtime configuration key.
                 */
                QString getConfigValue(QString key);

                /** \brief Sets the runtime configuration key to a value (triggering a event updating the clients).
                 */
                void setConfigValue(QString key, QString value);

                /** \brief Returns a static file content (for the app itself) on a request.
                 */
                void getStaticFile(WebServerEngineRequest* request, QString path);

                /** \brief Returns a url rebased to another engine's root url.
                 *
                 * Used for dispatching to workers.
                 * @a_dooverride_optional
                 */
                virtual QUrl rebaseUrl(QUrl rootUrl, QUrl url);

                /** \brief Plugs a WebModule into the engine and holds a pointer to it.
                 * Unplug it by calling removeModule().
                 */
                void addAndPreserveModule(std::shared_ptr<sh::ui::web::WebModule> module);

                /** \brief Plugs a WebModule into the engine. Unplug it by either delete `module`
                 * (drop all pointers to it) or by calling removeModule().
                 */
                void addModule(std::shared_ptr<sh::ui::web::WebModule> module);

                /** \brief Unplugs a WebModule from the engine.
                 */
                void removeModule(sh::ui::web::WebModule* module);

                /** \brief Returns the current engine timestamp.
                 *
                 * It's not related to real time, but just a value which is higher
                 * each time you query it.
                 *
                 * This is used for coordination of some time-order sensitive operations on browser side.
                 */
                qint64 currentTimestamp();

            protected:

                /** \brief Returns a (relative) url for a static file path.
                 * @a_infrastructure
                 */
                virtual QString staticPathToUrl(QString path);

                /**
                 * \brief Executes the root command (root url) on a request.
                 */
                bool rootCommand(WebServerEngineRequest* request, QString webadapter);

                /** \brief Executes a command on a request.
                 */
                bool executeCommand(WebServerEngineRequest* request, QString command);

                /** \brief The lowest tcp port number this web engine tries to bind to.
                 *
                 * Change this by the cmdline parameter `'maxport'`, `'port'` or ( `'workermaxport'` for some situations).
                 */
                int minimumPort();

                /** \brief The largest tcp port number this web engine tries to bind to.
                 *
                 * Change this by the cmdline parameter `'maxport'`, `'port'` or ( `'workermaxport'` for some situations).
                 */
                int maximumPort();

                /** \brief Determines the mimetype for a static file path.
                 */
                QString staticPathToMimeType(QString path);

                /** \brief Returns the path to the static web files.
                 */
                QString staticFilesPath();

                /** \brief Returns a list of all added `WebModule`s.
                 */
                QList<std::shared_ptr<WebModule> > modules();

            private:
                static WebServerEngine* _createEngine(bool withMainModule, QString dispatcherUrl, QByteArray wtoken);
                const QJsonObject _jok;
                QMutex _configValuesMutex;
                QMutex _modulesMutex;
                int _minport;
                int _maxport;
                QUrl _externalurl;
                QUrl _dispatcherurl;
                QMap<QString, QString> _configValues;
                QList<std::weak_ptr<sh::ui::web::WebModule>> _modules;
                QList<std::shared_ptr<sh::ui::web::WebModule>> _smodules;
                std::shared_ptr<sh::ui::web::WebServerEngineMainModule> _mainmodule;
                QString _webstaticpath;
                qint64 _currentTimestamp = 1;
                QByteArray _wtoken;

            };

            /** \brief The web serve engine main module, providing some base services like events.
             */
            class WebServerEngineMainModule : public QObject, public WebModule
            {
                Q_OBJECT
            public:
                WebServerEngineMainModule();
                ~WebServerEngineMainModule();
                bool executeCommand(WebServerEngineRequest* request, QString command) override;
                bool rootCommand(WebServerEngineRequest* request, QString webadapter) override;

                /** \brief Triggers an event on browser side.
                 */
                void triggerEvent(QString name, QString data);

            private slots:
                void cmd_dispatcher_toclient_heartbeat(WebServerEngineRequest* request);
                void cmd_events_get_next(WebServerEngineRequest* request);

            private:
                qint64 lastRequestTime();

                /** \brief Data for one occurred event.
                 */
                class EventEntry {
                    friend class EventEntryCleanupThread;
                public:
                    EventEntry(qint64 id, qint64 time, QString name, QString data);
                    const qint64 id;
                    const qint64 time;
                    const QString name;
                    const QString data;
                };

                /** \brief Thread for cleaning up old EventEntry instances.
                 */
                class EventEntryCleanupThread : public QThread {
                public:
                    EventEntryCleanupThread(WebServerEngineMainModule* module);
                protected:
                    void run() override;
                private:
                    WebServerEngineMainModule* module;
                };

                /** \brief Thread for stopping the application if the remote side is gone.
                 */
                class StopWhenIdleThread : public QThread {
                public:
                    StopWhenIdleThread(WebServerEngineMainModule* module);
                protected:
                    void run() override;
                private:
                    WebServerEngineMainModule* module;
                };

                qint64 _lastRequestTime;
                QMutex _mutex_lastRequestTime;
                QList<EventEntry*> _eventEntries;
                quint64 _eventEntriesLastId = 0;
                QMutex _mutex_eventEntries;
                QWaitCondition _eventEntriesCondition;
                EventEntryCleanupThread _eventEntryCleanupThread;
                StopWhenIdleThread _stopWhenIdleThread;
                static QRegularExpression _restatic;
            };

        }
    }
}

#endif // WEBSERVERENGINE_H
