/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2020 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "base/logger.h"
#include "base/mainthread.h"
#include "base/shallotprocess.h"
#include "base/singletoninitializer.h"
#include "base/threadpool.h"
#include "tools/misc.h"
#include "webmainwindow.h"
#include "webserverengine.h"
#include "webserverenginedispatcher.h"
#ifdef Q_OS_UNIX
#include <unistd.h>
#elif defined(Q_OS_WIN)
#define _WIN32_WINNT 0x0601
#include <SdkDdkVer.h>
#include <windows.h>
#include <psapi.h>
#endif

REGISTER_STATICINIT(sh::ui::web, WebServerEngineDispatcher, "early")


sh::ui::web::WebServerEngineDispatcher::WebServerEngineDispatcher()
            : _janitor(this)
{
    _webserver = std::shared_ptr<WebServerEngine>(sh::ui::web::WebServerEngine::createDispatcherEngine());
    _janitor.start();
}

void sh::ui::web::WebServerEngineDispatcher::stop()
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    if (_janitor.isRunning()) {
        _janitor.requestInterruption();
        for (int itry=0; itry<60*5; itry++) {
            if (itry % 30 == 0) {
                QMutexLocker _l1(&_workersmutex);
                foreach (WorkerProcess* p, _workers)
                    p->terminate();
            }
            QThread::msleep(1000);
            QCoreApplication::processEvents();
            {
                QMutexLocker _l1(&_workersmutex);
                if (_workers.isEmpty())
                    break;
            }
        }
        QMutexLocker _l1(&_workersmutex);
        {
            foreach (WorkerProcess* p, _workers) {
                p->kill();
                delete p;
            }
        }
        _janitor.wait();
    }
}

sh::ui::web::WebServerEngineDispatcher::~WebServerEngineDispatcher()
{ //TODO# test (when program shutdown works): start long copy process, then sigterm dispatcher: it should stay until copy is finished, then BOTH processes must be gone; test2: sigterm with an open rename action and wait 5 min for killing ; (we must react on SIGTERM for that i think, printing preventClosedMessage)
    stop();
}

sh::ui::web::WebMainWindow* sh::ui::web::WebServerEngineDispatcher::handleDispatching(std::function<WebMainWindow*(QString dispatcherUrl, QByteArray wtoken)> createWndFct)
{
    WebMainWindow* mainwnd = nullptr;
    QString webworker = sh::base::ShallotProcess::instance()->parameterValue("webworker").toLatin1();
    if (!webworker.isEmpty()) {
        QTextStream qstdin(stdin);
        QByteArray wtoken = qstdin.readLine().toLocal8Bit();
        mainwnd = createWndFct(webworker, wtoken);
        QString surl = mainwnd->webserver()->url().toString();
        QString otoken = hashRedirectorWebworkerUrl(surl, wtoken);
        QUrl wcurl = mainwnd->webserver()->rebaseUrl(webworker, mainwnd->webserver()->getFullCommandUrl("dispatcher/working_for",
                            WebCommandData({{"url", surl}, {"otoken", QUrl::toPercentEncoding(otoken)}, {"wtoken", QUrl::toPercentEncoding(wtoken)}}), false));
        sh::tools::Misc::makeHttpRequest(wcurl);
        //TODO # test browser stop detection:      QTimer* TODO = new QTimer();connect(TODO,&QTimer::timeout,[=](){sh::ui::MainWindow::mainWindow()->closeApp();});TODO->start(5000);
        //TODO # test proper shutdown with retcode 0:         QTimer* TODO = new QTimer();connect(TODO,&QTimer::timeout,[=](){sh::ui::MainWindow::mainWindow()->closeApp();});TODO->start(5000);
        return mainwnd;
    }
    else {

        //TODO # QTimer* TODO = new QTimer();connect(TODO,&QTimer::timeout,[=](){   QCoreApplication::exit(0);});TODO->start(5000);

        QUrl url;
        WebServerEngineDispatcher::start(url);
        SH_LOG_INFO(QObject::tr("The Shallot web server is now listening on %1.").arg(url.toString()));
        WebMainWindow::openUrlOnDesktop(url);
    }
    return mainwnd;
}

void sh::ui::web::WebServerEngineDispatcher::doInitialize()
{
}

void sh::ui::web::WebServerEngineDispatcher::doShutdown()
{
    if (_dispatcher) {
        _dispatcher->stop();
        auto ws = _dispatcher->_webserver;
        _dispatcher = nullptr;
    }
}

void sh::ui::web::WebServerEngineDispatcher::start(QUrl &url)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    _dispatcher = std::shared_ptr<WebServerEngineDispatcher>(new WebServerEngineDispatcher(), [](WebServerEngineDispatcher* x){
        sh::base::MainThread::dispatcher()->invokeSync([&](){
            delete x;
        });
    });
    _dispatcher->_webserver->addModule(_dispatcher);
    url = _dispatcher->_webserver->url();
}

QString sh::ui::web::WebServerEngineDispatcher::wtokenFromRequestCookie(WebServerEngineRequest* request)
{
    return QUrl::fromPercentEncoding(request->getCookie(_dispatcherCookieName).toLocal8Bit());
}

void sh::ui::web::WebServerEngineDispatcher::cmd__drop_worker__M(WebServerEngineRequest* request)
{
    request->setCookie(_dispatcherCookieName, "");
    QString wtoken = wtokenFromRequestCookie(request);
    WorkerProcess* worker = findWorkerForWtoken(wtoken);
    if (worker)
        worker->kill();
}

void sh::ui::web::WebServerEngineDispatcher::cmd__set_last_user_activity(WebServerEngineRequest* request)
{
    int value = request->data("value").toInt();
    QString wtoken = wtokenFromRequestCookie(request);
    WorkerProcess* worker = findWorkerForWtoken(wtoken);
    if (worker)
        worker->gotUserInteraction(value);
}

void sh::ui::web::WebServerEngineDispatcher::cmd__set_prevent_close_message(WebServerEngineRequest* request)
{
    QString message = request->data("message");
    QString wtoken = request->data("wtoken");
    WorkerProcess* worker = findWorkerForWtoken(wtoken);
    if (worker)
        worker->setPreventCloseMessage(message);
}

void sh::ui::web::WebServerEngineDispatcher::cmd__working_for(WebServerEngineRequest* request)
{
    QString workerurl = QUrl::fromPercentEncoding(request->data("url").toLatin1());
    QString otoken = request->data("otoken");
    QString wtoken = request->data("wtoken");
    if (compareHashRedirectorWebworkerUrl(otoken.toLocal8Bit(), workerurl, wtoken)) {
        QMutexLocker _l1(&_workersmutex);
        WorkerProcess* p = _workers.value(wtoken, nullptr);
        if (p) {
            p->waitForStarted(5000);
            if (p)
                p->initialize(workerurl, p->processId());
        }
    }
    else
        QThread::msleep(2000);
}

bool sh::ui::web::WebServerEngineDispatcher::executeCommand(WebServerEngineRequest* request, QString command)
{
    if (executeCommand_byQObjectReflection(this, request, command, "dispatcher"))
        return true;
    else {
        bool isroot = command.isEmpty() || command == "/";
        QString wtoken = wtokenFromRequestCookie(request);
        WorkerProcess* worker = nullptr;
        if (!wtoken.isEmpty()) {
            worker = findWorkerForWtoken(wtoken);
            if (!worker && isroot) {
                worker = spawnNewWorker(request);
                wtoken = worker->wtoken();
            }
        }
        else if (isroot) {
            worker = spawnNewWorker(request);
            wtoken = worker->wtoken();
        }
        if (worker) {
            if (isroot || command == "events/get_next") { // this is somewhat arbitrary, but by far the best choice for a heartbeat, which is reliable but not too often
                request->setCookie(_dispatcherCookieName, QUrl::toPercentEncoding(wtoken), 60*60*24*2/*should be longer than real idle session lifetime*/);
                worker->gotBrowserHeartbeat();
            }
            worker->request(request);
        }
        else {
            QThread::msleep(500);
            request->setResponse(QByteArray(), "text/html", WebServerEngineRequest::HTTP_INTERNAL_SERVER_ERROR);
        }
    }
    return true;
}

bool sh::ui::web::WebServerEngineDispatcher::rootCommand(WebServerEngineRequest* request, QString)
{
    return executeCommand(request, "");
}

QString sh::ui::web::WebServerEngineDispatcher::_dispatcherId = QString::fromLatin1(sh::tools::Misc::generateUniqueHash());

QString sh::ui::web::WebServerEngineDispatcher::_dispatcherCookieName = QUrl::toPercentEncoding(QString("shallot_%1").arg(_dispatcherId).toLocal8Bit());

QByteArray sh::ui::web::WebServerEngineDispatcher::hashRedirectorWebworkerUrl(QString url, QString wtoken)
{
    QStringList l;
    l << url << wtoken;
    return sh::tools::Misc::hashUnsalted(l);
}

bool sh::ui::web::WebServerEngineDispatcher::compareHashRedirectorWebworkerUrl(QByteArray hash, QString url, QString wtoken)
{
    QStringList l;
    l << url << wtoken;
    return sh::tools::Misc::compareHash(hash, l);
}

sh::ui::web::WebServerEngineDispatcher::WorkerProcess::WorkerProcess(sh::ui::web::WebServerEngineDispatcher* dispatcher, QByteArray wtoken, QString clientHost, QStringList acceptedLanguages)
            : _dispatcher(dispatcher), _wtoken(wtoken), _clientHost(clientHost), _reProcStatus(QRegularExpression("^([^:]+):\\s*([0-9]+)\\skB\\s*$"))
{
    setInputChannelMode(QProcess::ManagedInputChannel);
    setProcessChannelMode(QProcess::ForwardedChannels);
    setProgram(QCoreApplication::applicationFilePath());
    QStringList pargs = QCoreApplication::arguments();
    pargs.removeFirst();
    pargs << "--webworker=" + dispatcher->webserver()->url().toString();
    pargs << "--openbrowser=0";
    pargs << QString("--minport=%1").arg(sh::base::ShallotProcess::instance()->workerminport());
    pargs << QString("--maxport=%1").arg(sh::base::ShallotProcess::instance()->workermaxport());
    pargs << "--externalBaseUrl=" + dispatcher->webserver()->externalUrl().toString();
    foreach (QString acceptlanguage, acceptedLanguages) {
        if (!acceptlanguage.isEmpty())
            pargs << "--lang=" + acceptlanguage;
    }
    pargs << "--lang=en"; // fallback if no other language was given (otherwise server default will be used)
    QString wlogprefix = "w{{pid}}";
    QString logprefix = sh::base::Logger::instance()->logPrefix();
    if (!logprefix.isEmpty())
        wlogprefix = logprefix + " " + wlogprefix;
    pargs << ("--logprefix=" + wlogprefix);
    setArguments(pargs);
    gotBrowserHeartbeat();
    gotUserInteraction(0);
}

QByteArray sh::ui::web::WebServerEngineDispatcher::WorkerProcess::wtoken()
{
    return _wtoken;
}

QUrl sh::ui::web::WebServerEngineDispatcher::WorkerProcess::rootUrl()
{
    QMutexLocker _l1(&_mutex);
    return _rooturl;
}

void sh::ui::web::WebServerEngineDispatcher::WorkerProcess::request(WebServerEngineRequest *request)
{
    QUrl wurl = _dispatcher->webserver()->rebaseUrl(rootUrl(), request->url());
    QByteArray wresponseBody;
    QString wmimeType;
    int whttpStatus = WebServerEngineRequest::HTTP_NOT_FOUND;
    sh::tools::Misc::makeHttpRequest(wurl.toString(), &wresponseBody, &wmimeType, &whttpStatus);
    if (whttpStatus > 0)
        request->setResponse(wresponseBody, wmimeType, whttpStatus);
    else
        request->setResponse("", "text/html", WebServerEngineRequest::HTTP_BAD_GATEWAY);
}

bool sh::ui::web::WebServerEngineDispatcher::WorkerProcess::waitOnline()
{
    QMutexLocker _l1(&_mutex);
    int i = 0;
    while (_pid == -1) {
        _cnd_inited.wait(&_mutex, 1000);
        if (++i == 60) {
            kill();
            return false;
        }
    }
    return true;
}

qint64 sh::ui::web::WebServerEngineDispatcher::WorkerProcess::workerPid()
{
    QMutexLocker _l1(&_mutex);
    if (_pid == -1)
        return processId();
    else
        return _pid;
}

qint64 sh::ui::web::WebServerEngineDispatcher::WorkerProcess::memoryUsage()
{
#ifdef Q_OS_LINUX
    QFile inputFile(QString("/proc/%1/status").arg(workerPid()));
    if (inputFile.open(QIODevice::ReadOnly|QIODevice::Text)) {
        qint64 rssanon = 0;
        qint64 vmswap = 0;
        foreach (QByteArray line, inputFile.readAll().split('\n')) {
            QRegularExpressionMatch m = _reProcStatus.match(QString::fromLatin1(line));
            if (m.hasMatch()) {
                QString key = m.captured(1);
                if (key == "RssAnon")
                    rssanon = m.captured(2).toLongLong();
                else if (key == "VmSwap")
                    vmswap = m.captured(2).toLongLong();
            }
        }
        inputFile.close();
        return (rssanon + vmswap) * 1024;
    }
#elif defined(Q_OS_WIN)
    PROCESS_MEMORY_COUNTERS_EX pmc;
    GetProcessMemoryInfo(GetCurrentProcess(), (PROCESS_MEMORY_COUNTERS*)(&pmc), sizeof(pmc));
    return pmc.PrivateUsage;
#endif
    return -1;
}

QDateTime sh::ui::web::WebServerEngineDispatcher::WorkerProcess::lastBrowserHeartbeat()
{
    QMutexLocker _l1(&_mutex);
    return _lastBrowserHeartbeat;
}

QDateTime sh::ui::web::WebServerEngineDispatcher::WorkerProcess::lastUserInteraction()
{
    QMutexLocker _l1(&_mutex);
    return _lastUserInteraction;
}

QString sh::ui::web::WebServerEngineDispatcher::WorkerProcess::clientHost()
{
    return _clientHost;
}

QString sh::ui::web::WebServerEngineDispatcher::WorkerProcess::preventCloseMessage()
{
    QMutexLocker _l1(&_mutex);
    return _preventCloseMessage;
}

bool sh::ui::web::WebServerEngineDispatcher::WorkerProcess::isClosingFine()
{
    return preventCloseMessage().isEmpty();
}

void sh::ui::web::WebServerEngineDispatcher::WorkerProcess::initialize(QString rooturl, qint64 pid)
{
    QMutexLocker _l1(&_mutex);
    if (_pid == -1) {
        _rooturl = QUrl(rooturl);
        _pid = pid;
    }
    _cnd_inited.wakeAll();
}

void sh::ui::web::WebServerEngineDispatcher::WorkerProcess::gotBrowserHeartbeat()
{
    QMutexLocker _l1(&_mutex);
    _lastBrowserHeartbeat = QDateTime::currentDateTime();
}

void sh::ui::web::WebServerEngineDispatcher::WorkerProcess::gotUserInteraction(int value)
{
    QMutexLocker _l1(&_mutex);
    QDateTime t = QDateTime::currentDateTime().addSecs(value);
    if (_lastUserInteraction < t)
        _lastUserInteraction = t;
}

void sh::ui::web::WebServerEngineDispatcher::WorkerProcess::setPreventCloseMessage(QString message)
{
    QMutexLocker _l1(&_mutex);
    _preventCloseMessage = message;
}

sh::ui::web::WebServerEngineDispatcher::WorkerProcess* sh::ui::web::WebServerEngineDispatcher::spawnNewWorker(WebServerEngineRequest* request)
{
    WorkerProcess* p = nullptr;
    for (int wtry=0; wtry<3; wtry++) {
        sh::base::MainThread::dispatcher()->invokeSync([&](){
            QByteArray wtoken = sh::tools::Misc::generateUniqueHash();
            QStringList lsacceptlanguages = request->headerData("Accept-Language").split(';');
            QStringList acceptLanguages = lsacceptlanguages.isEmpty() ? QStringList() : lsacceptlanguages.first().split(",");
            p = new WorkerProcess(this, wtoken, request->clientHost(), acceptLanguages);
            // configure and start a worker
            QObject::connect(p, QOverload<int, QProcess::ExitStatus>::of(&QProcess::finished), p, [=](int exitCode, QProcess::ExitStatus) {
                QMutexLocker _l1(&_workersmutex);
                _workers.remove(p->wtoken());
                p->deleteLater();
                if (exitCode != 0)
                    SH_LOG_ERROR(QString("Worker process %1 crashed with return code %2.").arg(p->workerPid()).arg(exitCode))
                else
                    SH_LOG_INFO(QString("Worker process %1 closed.").arg(p->workerPid()))
            });
            QObject::connect(p, &QObject::destroyed, p, [=](){
                QMutexLocker _l1(&_workersmutex);
                _workers.remove(p->wtoken());
            });
            p->start();
            p->waitForStarted();
            SH_LOG_INFO(QString("Started new worker process %1.").arg(p->processId()))
            p->write(wtoken);
            p->write("\n");
            p->waitForBytesWritten();
            {
                QMutexLocker _l1(&_workersmutex);
                _workers[wtoken] = p;
            }
        });
        if (p->waitOnline()) {
            sh::base::MainThread::dispatcher()->invokeSync([&](){
                QTimer* timer_toclient_heartbeat = new QTimer(p);
                QUrl utchburl = webserver()->rebaseUrl(p->rootUrl(), webserver()->getFullCommandUrl("dispatcher/toclient_heartbeat", {}, false));
                QObject::connect(timer_toclient_heartbeat, &QTimer::timeout, p, [=](){
                    sh::tools::Misc::makeHttpRequest(utchburl);
                });
                timer_toclient_heartbeat->start(1000*10);
            });
            break;
        }
        else {
            sh::base::MainThread::dispatcher()->invokeSync([&](){
                QMutexLocker _l1(&_workersmutex);
                if (_workers.contains(p->wtoken())) {
                    p->deleteLater();
                    p = nullptr;
                }
            });
        }
    }
    return p;
}

sh::ui::web::WebServerEngineDispatcher::WorkerProcess* sh::ui::web::WebServerEngineDispatcher::findWorkerForWtoken(QString wtoken)
{
    QMutexLocker _l1(&_workersmutex);
    if (wtoken.isEmpty()) // just for a better feeling :)
        return nullptr;
    return _workers.value(wtoken, nullptr);
}

sh::ui::web::WebServerEngineDispatcher::JanitorThread::JanitorThread(WebServerEngineDispatcher* dispatcher) : _dispatcher(dispatcher)
{
}

qint64 sh::ui::web::WebServerEngineDispatcher::JanitorThread::machineMemory()
{
#ifdef Q_OS_UNIX
    long pages = sysconf(_SC_PHYS_PAGES);
    long page_size = sysconf(_SC_PAGE_SIZE);
    return pages * page_size;
#elif defined(Q_OS_WIN)
    MEMORYSTATUSEX status;
    status.dwLength = sizeof(status);
    GlobalMemoryStatusEx(&status);
    return status.ullTotalPhys;
#endif
    return -1;
}

void sh::ui::web::WebServerEngineDispatcher::JanitorThread::stopWorker(WorkerProcess* worker, QString reason)
{
    NOT_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    SH_LOG_WARNING(tr("Terminating worker process %1 because of '%2'").arg(worker->workerPid()).arg(reason))
    sh::base::MainThread::dispatcher()->invokeSync([&](){
        worker->terminate();
        QTimer* t = new QTimer(worker);
        connect(t, &QTimer::timeout, worker, [=](){
            worker->kill();
        });
        t->start(3000);
    });
    worker->kill();
}

QList<sh::ui::web::WebServerEngineDispatcher::WorkerProcess*> sh::ui::web::WebServerEngineDispatcher::JanitorThread::stopWorkers(QList<WorkerProcess*> workers, int stopcount, QString reason)
{
    NOT_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    QList<WorkerProcess*> lstopped;
    if (stopcount > 0) {
        QHash<WorkerProcess*, qint64> workersScores;
        QHash<QString, int> hostsSeen;
        const qint64 SCORE_FULL_MEMORY_USAGE = 1000 * 1000 * 2;
        const qint64 SCORE_HOST_SEEN = 1000 * 300;
        const qint64 SCORE_1MIN_USER_INTERACTION_AGE = 1;
        const qint64 SCORE_1MIN_BROWSER_HEARTBEAT_AGE = 500;
        const qint64 SCORE_IS_CLOSEABLE = 1000 * 1000;
        qint64 now = QDateTime::currentDateTime().toSecsSinceEpoch();
        foreach (WorkerProcess* worker, workers)
            hostsSeen[worker->clientHost()]++;
        foreach (WorkerProcess* worker, workers) {
            qint64 score = 0;
            // memory
            if (_allMemory > 0) // -1 -> unsupported
                score += _workerMemoryUsagesLastSeconds.value(worker) * SCORE_FULL_MEMORY_USAGE;
            // workers per host
            score += (hostsSeen[worker->clientHost()]) * SCORE_HOST_SEEN;
            // user activity
            score += ((now - worker->lastUserInteraction().toSecsSinceEpoch()) / 60) * SCORE_1MIN_USER_INTERACTION_AGE;
            // browser heartbeat
            score += ((now - worker->lastBrowserHeartbeat().toSecsSinceEpoch()) / 60) * SCORE_1MIN_BROWSER_HEARTBEAT_AGE;
            // is closeable
            score += worker->isClosingFine() ? SCORE_IS_CLOSEABLE : 0;
            workersScores[worker] = score;
        }
        QMultiMap<qint64, WorkerProcess*> workersByScore;
        foreach (WorkerProcess* worker, workers)
            workersByScore.insertMulti(-workersScores[worker], worker);
        auto iwbs = workersByScore.begin();
        while (stopcount > 0 && iwbs != workersByScore.end()) {
            WorkerProcess* w = *iwbs;
            iwbs == workersByScore.erase(iwbs);
            stopWorker(w, reason);
            lstopped.append(w);
            stopcount--;
        }
    }
    return lstopped;
}

int sh::ui::web::WebServerEngineDispatcher::JanitorThread::detectGoodMaxNumberWorkersByMachineCapabilities()
{
    const qint64 MAGIC_CONSTANT = 150000000;
    const qint64 CONSERVATIVE_DEFAULT = 20;
    const qint64 MINIMUM = 10;
    const qint64 MAXIMUM = 1000;
    qint64 availmem = machineMemory();
    qint64 result;
    if (availmem == -1)
        result = CONSERVATIVE_DEFAULT;
    else
        result = std::max(availmem / MAGIC_CONSTANT, MINIMUM);
    return std::max(std::min(result, MAXIMUM), MINIMUM);
}

void sh::ui::web::WebServerEngineDispatcher::JanitorThread::run()  // also prevents DoS a bit, but assumes to be behind authentication, so not to fight against crazily bad things; if you have such enemies, tune the values or take additional countermeasures.
{
    // there is also a quit-on-browser-away-for-too-long handling on worker side
    auto shap = sh::base::ShallotProcess::instance();
    const int maxNumberWorkers = shap->maxnumberworkers(detectGoodMaxNumberWorkersByMachineCapabilities());
    const int maxNumberWorkersPerHost = shap->maxnumberworkersperhost(std::max(1, maxNumberWorkers/4));
    const int maxMemoryPercent = shap->maxmemorypercent(80); // we consider machine ram, so the OS can actually provide more, the memory usage is also maybe reported to high, so use high enough values here
    const int maxMemoryPercentGlobal = shap->maxmemorypercentglobal(100);
    const int maxBrowserAwaySeconds = shap->maxbrowserawayseconds(60 * 60 * 24 * 2);
    const int maxIdlenessSeconds = shap->maxidlenessseconds(60 * 60 * 24 * 14);
    qint64 i = 0; // not exactly required that large ^^
    const int STEPSIZE1 = 3;
    const int STEPSIZE2 = 5;
    _allMemory = machineMemory();
    if (_allMemory <= 0) // -1 -> unsupported
        SH_LOG_WARNING("Worker memory monitoring is not supported on this platform.")
    while (!isInterruptionRequested()) {
        bool fullCheckNext = false; // set true for early make a full check next time
        qint64 now = QDateTime::currentDateTime().toSecsSinceEpoch();
        {
            QMutexLocker _l1(&_dispatcher->_workersmutex);
            QList<WorkerProcess*> workers = _dispatcher->_workers.values();
            // some of those variables are only populated in some steps!
            QHash<WorkerProcess*, qint64> memoryUsages;
            if (i % STEPSIZE2 == 0) {
                _allMemory = machineMemory();
                auto iworker = workers.begin();
                while (iworker != workers.end()) {
                    WorkerProcess* worker = *iworker;
                    qint64 workermemoryusage = worker->memoryUsage();
                    memoryUsages[worker] = workermemoryusage;
                    QString _stopWorker;
                    // maxBrowserAwaySeconds
                    if ((now - worker->lastBrowserHeartbeat().toSecsSinceEpoch()) > maxBrowserAwaySeconds)
                        _stopWorker = "maxBrowserAwaySeconds";
                    // maxMemoryPercent
                    if (_allMemory > 0) { // -1 -> unsupported
                        if ((workermemoryusage / _allMemory) > (maxMemoryPercent / 100.0))
                            _stopWorker = "maxMemoryPercent";
                    }
                    //next
                    if (_stopWorker.isNull())
                        iworker++;
                    else {
                        stopWorker(worker, _stopWorker);
                        iworker = workers.erase(iworker);
                    }
                }
                QHash<WorkerProcess*, double> workerMemoryUsagesLastSeconds;
                const double wmulm_learnrate = 0.1;
                foreach (WorkerProcess* worker, workers)
                    workerMemoryUsagesLastSeconds[worker] = (wmulm_learnrate * worker->memoryUsage() / _allMemory) + (1 - wmulm_learnrate) * _workerMemoryUsagesLastSeconds[worker];
                _workerMemoryUsagesLastSeconds = workerMemoryUsagesLastSeconds;
            }
            if (i % STEPSIZE1 == 0) {
                // maxNumberWorkersPerHost
                QHash<QString, QList<WorkerProcess*>> workersByHost;
                foreach (WorkerProcess* worker, workers) {
                    QString workerclientHost = worker->clientHost();
                    if (!workerclientHost.isEmpty())
                        workersByHost[workerclientHost].append(worker);
                }
                foreach (QString host, workersByHost.keys()) {
                    QList<WorkerProcess*> hworkers = workersByHost[host];
                    foreach (WorkerProcess* remp, stopWorkers(hworkers, hworkers.length()-maxNumberWorkersPerHost, "maxNumberWorkersPerHost"))
                        workers.removeAll(remp);
                }
                auto iworker = workers.begin();
                while (iworker != workers.end()) {
                    WorkerProcess* worker = *iworker;
                    QString _stopWorker;
                    // maxIdlenessSeconds
                    if ((now - worker->lastUserInteraction().toSecsSinceEpoch()) > maxIdlenessSeconds)
                        _stopWorker = "maxIdlenessSeconds";
                    // maxNumberWorkers
                    foreach (WorkerProcess* remp, stopWorkers(workers, workers.length()-maxNumberWorkers, "maxNumberWorkers"))
                        workers.removeAll(remp);
                    //next
                    if (_stopWorker.isNull())
                        iworker++;
                    else {
                        stopWorker(worker, _stopWorker);
                        iworker = workers.erase(iworker);
                    }
                }
            }
            if (i % STEPSIZE2 == 0) {
                // maxMemoryPercentGlobal
                if (_allMemory > 0) { // -1 -> unsupported
                    qint64 memoryUsageSum = 0;
                    foreach (WorkerProcess* p, workers)
                        memoryUsageSum += memoryUsages.value(p);
                    if ((memoryUsageSum / _allMemory) > (maxMemoryPercentGlobal / 100.0)) {
                        fullCheckNext = true;
                        foreach (WorkerProcess* remp, stopWorkers(workers, 1, "maxMemoryPercentGlobal"))
                            workers.removeAll(remp);
                    }
                }
            }
        }
        if (fullCheckNext)
            i = 0;
        else
            i++;
        QThread::msleep(1000);
    }
}

std::shared_ptr<sh::ui::web::WebServerEngineDispatcher> sh::ui::web::WebServerEngineDispatcher::_dispatcher = nullptr;

