/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2020 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef WEBSERVERENGINEDISPATCHER_H
#define WEBSERVERENGINEDISPATCHER_H

#include <QDateTime>
#include <QMutex>
#include <QProcess>
#include <QRegularExpression>
#include <QThread>
#include <QUrl>
#include <QUuid>
#include <QWaitCondition>
#include "webmodule.h"


namespace sh {
    namespace ui {
        namespace web {
            class WebMainWindow;

            /** \brief The web serve engine dispatcher module, used in dispatcher mode for
             * providing a portal url which starts Shallot instances on request and internally redirects to them.
             */
            class WebServerEngineDispatcher : public QObject, public WebModule
            {
                Q_OBJECT
            public:

                /** @brief Handles dispatching and creates a new main window (in child processes) on demand via a given factory function.
                 */
                static WebMainWindow* handleDispatching(std::function<WebMainWindow*(QString dispatcherUrl, QByteArray wtoken)> createWndFct);

                /** @a_infrastructure
                 */
                WebServerEngineDispatcher();

                ~WebServerEngineDispatcher();

                /** @brief Stops the dispatcher.
                 */
                void stop();

                /** @a_infrastructure
                 */
                static void doInitialize();

                /** @a_infrastructure
                 */
                static void doShutdown();

            private slots:
                void cmd__drop_worker__M(WebServerEngineRequest* request);
                void cmd__set_last_user_activity(WebServerEngineRequest* request);
                void cmd__set_prevent_close_message(WebServerEngineRequest* request);
                void cmd__working_for(WebServerEngineRequest* request);

            private:

                /** @brief A internal Shallot worker process driving one Shallot session.
                 */
                class WorkerProcess : public QProcess {
                    friend class WebServerEngineDispatcher;
                public:

                    /** @brief The wtoken (used for association and security).
                     */
                    QByteArray wtoken();

                    /** @brief The worker root url.
                     */
                    QUrl rootUrl();

                    /** @brief Make a request to the worker.
                     */
                    void request(WebServerEngineRequest* request);

                    /** @brief Returns the process id of the worker.
                     */
                    qint64 workerPid();

                    /** @brief Returns the current memory usage of this worker (in bytes).
                     *
                     * Returns -1 if this is not supported on the target platform or it failed.
                     */
                    qint64 memoryUsage();

                    /** @brief Returns when a browser was seen connected to this worker last time.
                     */
                    QDateTime lastBrowserHeartbeat();

                    /** @brief Returns when a user has interacted with this worker last time.
                     */
                    QDateTime lastUserInteraction();

                    /** @brief Returns the client host this worker is serving.
                     *
                     * It's not guaranteed to have some specific content (maybe an ip address),
                     * but is equal for the same host, some one can count how many hosts work for one
                     * particular client host.
                     */
                    QString clientHost();

                    /** @brief Returns a reason message why this worker currently should not be closed
                     * (or `""` if closing is fine).
                     *
                     * See also isClosingInappropriate().
                     *
                     * Note: This is reported asynchronously. You must not rely on its precise information,
                     * but solely use it for monitoring.
                     */
                    QString preventCloseMessage();

                    /** @brief Returns if it is fine to stop this worker (or if there is a good reason to not do).
                     *
                     * See also preventCloseMessage().
                     *
                     * Note: This is reported asynchronously. You must not rely on its precise information,
                     * but solely use it for monitoring.
                     */
                    bool isClosingFine();

                private:

                    WorkerProcess(WebServerEngineDispatcher* dispatcher, QByteArray wtoken, QString clientHost, QStringList acceptedLanguages);

                    /** @brief Waits until the worker is online and ready for requests (or dead).
                     */
                    bool waitOnline();

                    /** @brief Initializes the worker instance (which exists on the dispatcher side!) by setting some final data.
                     */
                    void initialize(QString rooturl, qint64 pid);

                    /** @brief Refreshes the browser heartbeat timestamp.
                     *
                     * See lastBrowserHeartbeat().
                     */
                    void gotBrowserHeartbeat();

                    /** @brief Refreshes the user interaction timestamp.
                     *
                     * See lastUserInteraction().
                     */
                    void gotUserInteraction(int value);

                    /** @brief Sets the preventCloseMessage (empty: closeable).
                     */
                    void setPreventCloseMessage(QString message);

                    WebServerEngineDispatcher* _dispatcher;
                    QByteArray _wtoken;
                    QUrl _rooturl;
                    QMutex _mutex;
                    QWaitCondition _cnd_inited;
                    qint64 _pid = -1;
                    QDateTime _lastBrowserHeartbeat;
                    QDateTime _lastUserInteraction;
                    QString _clientHost;
                    QString _preventCloseMessage;
                    QRegularExpression _reProcStatus; // not static because that leads to crashes in shutdown

                };

                class JanitorThread : public QThread {
                public:
                    JanitorThread(WebServerEngineDispatcher* dispatcher);
                    void run() override;
                private:

                    /** @brief Returns the amount of memory this machine provides (in bytes).
                     *
                     * Returns -1 if this is not supported on the target platform or it failed.
                     */
                    qint64 machineMemory();

                    void stopWorker(WorkerProcess* worker, QString reason);
                    int detectGoodMaxNumberWorkersByMachineCapabilities();
                    QList<WorkerProcess*> stopWorkers(QList<WorkerProcess*> workers, int stopcount, QString reason);
                    WebServerEngineDispatcher* _dispatcher;
                    double _allMemory;
                    QHash<WorkerProcess*, double> _workerMemoryUsagesLastSeconds;
                };

                /** @brief Starts the dispatcher engine, accepting http connections and internally handling child processes.
                 */
                static void start(QUrl &url);

                /** \brief Compute a hash for a redirector webworker url.
                 *
                 * @a_infrastructure
                 */
                static QByteArray hashRedirectorWebworkerUrl(QString url, QString wtoken);

                /** \brief Compares a hash for a redirector webworker url to other data.
                 *
                 * @a_infrastructure
                 */
                bool compareHashRedirectorWebworkerUrl(QByteArray hash, QString url, QString wtoken);

                /** @brief Spawns a new worker.
                 *
                 * Does not execute any requests on it.
                 */
                WorkerProcess* spawnNewWorker(WebServerEngineRequest *request);

                /** @brief Returns the worker for a wtoken.
                 */
                WorkerProcess* findWorkerForWtoken(QString wtoken);

                /** @brief Returns the wtoken from a request (typically from cookie).
                 */
                QString wtokenFromRequestCookie(WebServerEngineRequest* request);

                bool executeCommand(WebServerEngineRequest* request, QString command) override;
                bool rootCommand(WebServerEngineRequest* request, QString webadapter) override;
                static std::shared_ptr<WebServerEngineDispatcher> _dispatcher;
                static QString _dispatcherId;
                static QString _dispatcherCookieName;
                std::shared_ptr<WebServerEngine> _webserver;
                QMutex _workersmutex;
                QHash<QString, WorkerProcess*> _workers;
                JanitorThread _janitor;

            };

        }
    }
}

#endif // WEBSERVERENGINEDISPATCHER_H
