/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2020 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "configuration/configurationvalue.h"
#include "filesystem/eurl.h"
#include "settings/setting.h"
#include "settings/settingsmanager.h"
#include "ui/fileview.h"
#include "ui/mainwindow.h"
#include "webstoreprofiledialog.h"


sh::ui::web::WebStoreProfileDialog::WebStoreProfileDialog(std::shared_ptr<const filesystem::Eurl> eurl)
    : WebDialog("StoreProfileDialog"), StoreProfileDialog(eurl)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    setDialogProperty("advancedMode", cfgvalLevel->getConfigValueInt() > 0);
    setDialogProperty("eurl", eurl->asString());
}

bool sh::ui::web::WebStoreProfileDialog::executeDialogCommand(WebServerEngineRequest* request, QString command)
{
    return WebModule::executeCommand_byQObjectReflection(this, request, command, "storeprofile");
}

void sh::ui::web::WebStoreProfileDialog::cmd__get__M(WebServerEngineRequest* request)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    QJsonObject result;
    QJsonArray jsettings;
    foreach (std::shared_ptr<sh::settings::Setting> setting, sh::settings::SettingsManager::instance()->getAllSettings()) {
        auto currentfv = sh::ui::MainWindow::mainWindow()->currentFileView();
        QString currval = setting->getValue(currentfv);
        if (!currval.isNull()) {
            QJsonObject jsetting;
            jsetting["name"] = setting->name();
            jsetting["description"] = setting->description();
            jsetting["groupDescription"] = sh::settings::Setting::getGroupDescription(setting->group());
            jsetting["valueDescription"] = setting->valueDescription(currval);
            jsetting["_additionalCheck"] = setting->isPerFileview() ? tr("only for file list #%1").arg(currentfv->index()+1) : "";
            jsetting["_additionalCheckValue"] = currentfv->index();
            jsetting["isAdvanced"] = setting->isAdvancedSetting();
            jsetting["isGlobal"] = setting->isGlobal();
            jsettings.append(jsetting);
        }
    }
    result["settings"] = jsettings;
    QJsonArray jprofiles;
    foreach (QString profile, sh::settings::SettingsManager::instance()->getProfileList())
        jprofiles.append(profile);
    result["profiles"] = jprofiles;
    result["currentProfile"] = sh::ui::MainWindow::mainWindow()->currentProfile();
    request->setResponse(result);
}

void sh::ui::web::WebStoreProfileDialog::cmd__store_mode__M(WebServerEngineRequest* request)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    int level = request->data("level").toInt();
    cfgvalLevel->setConfigValue(level);
}

QList<sh::ui::StoreProfileDialog::SettingEntry> sh::ui::web::WebStoreProfileDialog::checkedSettings()
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    QList<SettingEntry> result;
    QString jsettings = dialogResult()["settings"].toString();
    auto jd = QJsonDocument::fromJson(jsettings.toLatin1());
    foreach (auto jcd, jd.array()) {
        auto ojcd = jcd.toObject();
        QString name = ojcd["name"].toString();
        sh::settings::Setting* setting = sh::settings::SettingsManager::instance()->getSettingByName(name).get();
        if (setting) {
            int additionalCheckValue = ojcd["additionalCheckValue"].toInt(-1);
            result.append(SettingEntry(setting, additionalCheckValue));
        }
    }
    return result;
}

QString sh::ui::web::WebStoreProfileDialog::profileName()
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    return dialogResult()["profileName"].toString();
}

bool sh::ui::web::WebStoreProfileDialog::withSubDirectories()
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    return dialogResult()["withSubDirectories"].toBool();
}

QStringList sh::ui::web::WebStoreProfileDialog::inheritsFrom()
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    QStringList result;
    QString jinheritsFrom = dialogResult()["inheritsFrom"].toString();
    auto jd = QJsonDocument::fromJson(jinheritsFrom.toLatin1());
    foreach (auto jcd, jd.array())
        result << jcd.toString();
    return result;
}

bool sh::ui::web::WebStoreProfileDialog::hasGlobal()
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    return dialogResult()["hasGlobal"].toBool();
}
