/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2020 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "configuration/configurationmanager.h"
#include "configuration/configurationvalue.h"
#include "configuration/configurationvaluetype.h"
#include "base/shallotprocess.h"
#include "webtuningdialog.h"


sh::ui::web::WebTuningDialog::WebTuningDialog() : WebDialog("TuningDialog")
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
}

void sh::ui::web::WebTuningDialog::init(std::shared_ptr<Dialog> dialog)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    sh::base::MainThread::dispatcher()->invokeAsync([=](){
        QMutexLocker _l1(&_cfgvaluesmutex);
        _cfgvalues = sh::configuration::ConfigurationManager::instance()->getAllConfigurationValues();
        foreach(std::shared_ptr<sh::configuration::ConfigurationValue> cfgval, _cfgvalues)
        {
            cfgval->consumeValue(this, [=](){
                sh::base::MainThread::dispatcher()->invokeSync([&](){
                    triggerDialogEvent("tuning/configurationvalue_changed", configurationValueToJsonObject(cfgval));
                });
            });
        }
        _cfgvaluesinited = true;
        _cfgvaluescondition.wakeAll();
    }, dialog);
}

bool sh::ui::web::WebTuningDialog::executeDialogCommand(WebServerEngineRequest* request, QString command)
{
    return WebModule::executeCommand_byQObjectReflection(this, request, command, "tuning");
}

void sh::ui::web::WebTuningDialog::cmd__change_configurationvalue(WebServerEngineRequest* request)
{
    QString name = request->data("name");
    QString textual = request->data("textual");
    std::shared_ptr<sh::configuration::ConfigurationValue> cfgval = nullptr;
    foreach(std::shared_ptr<sh::configuration::ConfigurationValue> _cv, sh::configuration::ConfigurationManager::instance()->getAllConfigurationValues())
    {
        if (_cv->name() == name) {
            cfgval = _cv;
            break;
        }
    }
    if (cfgval) {
        if (dynamic_cast<sh::configuration::ConfigurationValueTypeLocalFilePath*>(cfgval->valueType()))
            cfgval->setConfigValue(cfgval->valueType()->parse(textual).toString());
        else if (dynamic_cast<sh::configuration::ConfigurationValueTypeString*>(cfgval->valueType()))
            cfgval->setConfigValue(cfgval->valueType()->parse(textual).toString());
        else if (dynamic_cast<sh::configuration::ConfigurationValueTypeInteger*>(cfgval->valueType()))
            cfgval->setConfigValue(cfgval->valueType()->parse(textual).toInt());
        else if (dynamic_cast<sh::configuration::ConfigurationValueTypeFloat*>(cfgval->valueType()))
            cfgval->setConfigValue(cfgval->valueType()->parse(textual).toFloat());
        else if (dynamic_cast<sh::configuration::ConfigurationValueTypeBoolean*>(cfgval->valueType()))
            cfgval->setConfigValue(!cfgval->getConfigValueBool());
        else
            throw sh::exceptions::ProgramException(sh::exceptions::ExceptionData().details("unknown valueType"));
    }
}

void sh::ui::web::WebTuningDialog::cmd__get_configurationvalues(WebServerEngineRequest* request)
{
    QJsonArray result;
    {
        QMutexLocker _l1(&_cfgvaluesmutex);
        while (!_cfgvaluesinited)
            _cfgvaluescondition.wait(&_cfgvaluesmutex, 5000);
    }
    foreach(std::shared_ptr<sh::configuration::ConfigurationValue> cfgval, _cfgvalues)
        result.append(configurationValueToJsonObject(cfgval));
    request->setResponse(result);
}

QString sh::ui::web::WebTuningDialog::categoryToName(sh::configuration::ConfigurationCategory category)
{
    if (category == sh::configuration::CategoryNone)
        return "None";
    else if (category == sh::configuration::CategoryGUI)
        return "GUI";
    else if (category == sh::configuration::CategoryBehavior)
        return "Behavior";
    else if (category == sh::configuration::CategoryExternalTools)
        return "ExternalTools";
    else
        throw sh::exceptions::ProgramException(sh::exceptions::ExceptionData().details("unknown category"));
}

QString sh::ui::web::WebTuningDialog::valueTypeToName(sh::configuration::ConfigurationValueType* valueType)
{
    if (dynamic_cast<sh::configuration::ConfigurationValueTypeLocalFilePath*>(valueType))
        return "LocalFilePath";
    else if (dynamic_cast<sh::configuration::ConfigurationValueTypeString*>(valueType))
        return "String";
    else if (dynamic_cast<sh::configuration::ConfigurationValueTypeInteger*>(valueType))
        return "Integer";
    else if (dynamic_cast<sh::configuration::ConfigurationValueTypeFloat*>(valueType))
        return "Float";
    else if (dynamic_cast<sh::configuration::ConfigurationValueTypeBoolean*>(valueType))
        return "Boolean";
    else
        throw sh::exceptions::ProgramException(sh::exceptions::ExceptionData().details("unknown valueType"));
}

QJsonObject sh::ui::web::WebTuningDialog::configurationValueToJsonObject(std::shared_ptr<sh::configuration::ConfigurationValue> cfgval)
{
    QJsonObject jcfgval;
    jcfgval["name"] = cfgval->name();
    jcfgval["valueDescription"] = cfgval->valueType()->valueDescription(cfgval->getConfigValueVariant());
    jcfgval["description"] = cfgval->description();
    jcfgval["longDescription"] = cfgval->longDescription();
    jcfgval["changeHint"] = cfgval->changeHint();
    jcfgval["category"] = categoryToName(cfgval->category());
    jcfgval["valueType"] = valueTypeToName(cfgval->valueType());
    jcfgval["mayChange"] = cfgval->mayChange();
    return jcfgval;
}
