/*
 @licstart  The following is the entire license notice for the JavaScript code in this page.

 Copyright (C) 2014-2020, Josef Hahn

 This file is part of Shallot.

 Shallot is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 Shallot is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with Shallot.  If not, see <http://www.gnu.org/licenses/>.

 @licend  The above is the entire license notice for the JavaScript code in this page.
*/


(function(){

    class FilesystemDatasource extends clove.AjaxAsyncDatasource {

        constructor(config) {
            config = config || [];
            super(clove.utils.applyDefaultsToConfig(config, {
                answerIdKeyName: 'eurl',
                requestIdKeyName: 'eurl',
                staticIcon: undefined,
                params: clove.utils.applyDefaultsToConfig(config.params),
                data_autoPullCharily: true,
                headerConfigs: {eurl: {invisible: true}, nodetype: {invisible: true}, isHidden: {invisible: true}, displayName: {label: String.fromCharCode(160)}}
            }));
        }

        _coldata(ptr, data) {
            return {cols: ["displayName", "nodetype", "isHidden", "eurl"], xmlatt2colidx:
                    {displayName: 0, nodetype: 1, isHidden: 2, eurl: 3}, idcolname: "eurl"};
        }

        getMetadata(ptr) {
            var result = {};
            if (ptr && ptr.icol == 0) {
                var itm = shwebui.ShallotApplication.getFileItemForValuePointer(ptr);
                if (itm.eurl)
                    result.icon = clove.Icon.byUrl(shwebui.actioncallUrl('filesystem/get_icon?eurl=' + encodeURIComponent(itm.eurl)));
            }
            return result;
        }

        _interpretdata(p, d) { //this is a hack. otherwise icons may move into wrong lines for new/removed items TODO# better?
            var xxx = clove.utils.suspendResizing();
            try {
                super._interpretdata(p, d);
                if (this.columnCount(p) > 0) {
                    var rcnt = this.rowCount(p);
                    if (rcnt > 0)
                        this.OnDataUpdate.trigger({parent: p, r1: 0, r2: rcnt-1, c1: 0, c2: 0});
                }
            }
            finally {
                clove.utils.resumeResizing(xxx);
            }
        }

    }

    class TreeviewFilesystemDatasource extends FilesystemDatasource {

        constructor(eurl) {
            super({data_pullUrl: 'filesystem/list_items', params: {onlydirs: "1", eurl: eurl}, staticIcon: clove.symbols.FolderClosed});
        }

    }

    class FileviewFilesystemDatasource extends FilesystemDatasource {

        constructor(listid) {
            super({data_pullUrl: 'fileview/list_view_items', params: {listid: listid}, rowHeadersVisible: false});
            this._baseHeaderConfigs = this.config.headerConfigs;
        }

        _datachildren(data) {
            return data.data;
        }

        _coldata(ptr, data) {
            var result = super._coldata(ptr, data.data);
            var cols = result.cols;
            var xmlatt2colidx = result.xmlatt2colidx;
            var dcio = cols.length;
            var hcfg = Object.assign({}, this._baseHeaderConfigs);
            var hdata = data.headerData;
            var hcookiestr = "";
            for (var ihd=0; ihd<hdata.length; ihd++) {
                var dcn = "col_" + (ihd + 1);
                var dhdn =
                cols.push(dcn);
                xmlatt2colidx[dcn] = ihd + dcio;
                var hdatai = hdata[ihd];
                hcfg[dcn] = {label: hdatai};
                hcookiestr += hdatai + "|";
            }
            if (this._coldatacookie != hcookiestr) {
                this._coldatacookie = hcookiestr;
                //TODO #xx ensure that the column header strings in ui get updated. (currently they go wrong when some new column gets inserted afterwards)
            }
            this.config.headerConfigs = hcfg;
            return result;
        }

    }

    class ToolbarActionsDatasource extends clove.AjaxAsyncDatasource {

        constructor(config) {
            config = config || [];
            super(clove.utils.applyDefaultsToConfig(config, {
                data_pullUrl: 'action/get',
                answerIdKeyName: 'actionpath',
                requestIdKeyName: 'actionpath',
                staticIcon: undefined,
                params: clove.utils.applyDefaultsToConfig(config.params),
            }));
        }

        setValue(ptr, value) {
            if (value.name)
                value.icon = clove.Icon.byUrl(shwebui.actioncallUrl('action/icon?actionpath=' + encodeURIComponent(value.name)));
            super.setValue(ptr, value);
        }

    }

    class DataManager {

        dirTreeDatasource() {
            return new TreeviewFilesystemDatasource(shwebui.config.rootEurl);
        }

        fileViewDatasource(listid) {
            return new FileviewFilesystemDatasource(listid);
        }

        toolbarActionsDatasource(owner) {
            var self = this;
            var result = new ToolbarActionsDatasource();
            var ontoolbarrefresh = shwebui.eventfetcher.getEvent("toolbar/refresh");
            var _timeoutid;
            ontoolbarrefresh.addHandler(function(e){
                if (_timeoutid === undefined)
                    _timeoutid = setTimeout(function(){
                        _timeoutid = undefined;
                        result.clearCache();
                        result.rowCount();
                    });
            }, owner);
            return result;
        }

    }

    shwebui.data = new DataManager();

})();
